//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "HistExplorerView.h"

#include <algorithm>

#include <QCheckBox>
#include <QComboBox>
#include <QDateTime>
#include <QPushButton>
#include <QListView>
#include <QVariant>
#include <QDebug>
#include <QStyle>

//#include "histdialog.h"
#include "HistPlotBase.h"
#include "qutil.h"
#include "dsplib/th1.h"
#include "th1_gui.h"
#include "dsplib/dsplib.h"
#include "dsplib/fitter.h"

HistExplorerView::HistExplorerView(QWidget *parent)
    : QWidget(parent)
{
    setupUi(this);

    pbShow->setIcon(style()->standardIcon(QStyle::SP_DialogOpenButton));
    pbHideAll->setIcon(style()->standardIcon(QStyle::SP_DialogCloseButton));
    pbSaveAll->setIcon(style()->standardIcon(QStyle::SP_DialogSaveButton));
    pbSave->setIcon(style()->standardIcon(QStyle::SP_DialogSaveButton));
    pbClear->setIcon(style()->standardIcon(QStyle::SP_DialogResetButton));
    pbClearAll->setIcon(style()->standardIcon(QStyle::SP_DialogResetButton));

    connect(pbClear, &QPushButton::clicked, this, &HistExplorerView::clearCurrentHist);
    connect(pbClearAll, &QPushButton::clicked, this, &HistExplorerView::clearAllHists);

    connect(checkBox_autoFit, &QCheckBox::toggled, this, &HistExplorerView::setAutoFit);
    connect(checkBox_autoRange, &QCheckBox::toggled, this, &HistExplorerView::setAutoRange);
    // connect(cb_hist, qOverload<int>(&QComboBox::activated),
    //         this, &HistExplorerView::setCurrentHistType); // requires c++14, CentOS 7 fails
    connect(cb_hist, SIGNAL(activated(int)),
            this, SLOT(setCurrentHistType(int)));
    connect(pbSave, &QPushButton::clicked, this, &HistExplorerView::save);
    connect(pbSaveAll, &QPushButton::clicked, this, &HistExplorerView::saveAll);
    // connect(spinBoxChannel, qOverload<int>(&QSpinBox::valueChanged),
    //         this, &HistExplorerView::setCurrentChannel);
    connect(spinBoxChannel, SIGNAL(valueChanged(int)),
            this, SLOT(setCurrentChannel(int)));
    labelInfo->clear();
}

void HistExplorerView::showEvent(QShowEvent *event)
{
    update_current_main_hist();
    QWidget::showEvent(event);
}

void HistExplorerView::onHistogramsChanged(const HistList &hl)
{
//    qDebug() << __PRETTY_FUNCTION__ << objectName() << hl.size() << "histograms";
    histList = hl;
    HistTypeList list;
    for (auto i : hl.histTypes()) {
        list.append(i);
    }
    updateComboBox(list);
    int max_channel = 0;
    for (const auto i : hl.keys()) {
        if (i.c > max_channel)
            max_channel = i.c;
    }
    spinBoxChannel->setMinimum(0);
    spinBoxChannel->setMaximum(max_channel);
    update_current_main_hist();
}

hist_type_t HistExplorerView::currentHistType() const
{
    int currentIndex = cb_hist->currentIndex();
    auto current_hist_type = qvariant_cast<hist_type_t>(cb_hist->itemData(currentIndex));
    return current_hist_type;
}

int HistExplorerView::currentHistChannel() const
{
    return spinBoxChannel->value();
}

th1 HistExplorerView::currentHist() const
{
    return histList[currentHistIndex()];
}

hist_index_t HistExplorerView::currentHistIndex() const
{
    const hist_type_t &currentType = currentHistType();
    if (hist_is_multichannel(currentType))
        return hist_index_t(currentType, currentHistChannel());
    return hist_index_t(currentType);
}

int HistExplorerView::findIndexByType(const hist_type_t &t) const
{
    for (int i=0; i<cb_hist->count(); i++) {
        const auto hist_type = qvariant_cast<hist_type_t>(cb_hist->itemData(i));
        if (t == hist_type)
            return i;
    }
    return -1;
}

void HistExplorerView::updateComboBox(const HistTypeList &list)
{
    // create map of currently set indices
    HistTypeList newList(list);
    HistTypeList origList;
    for (int i=0; i<cb_hist->count(); i++) {
        const auto hist_index = qvariant_cast<hist_type_t>(cb_hist->itemData(i));
        origList.append(hist_index);
    }
    // no ComboBox update if contents unchanged
    std::sort(origList.begin(), origList.end());
    std::sort(newList.begin(), newList.end());
    if (newList == origList)
        return;
    // save current index
    int oldIndex = cb_hist->currentIndex();
    const hist_type_t &current_hist_type = currentHistType();
    cb_hist->clear();
    for (auto & hist_type : newList) {
        cb_hist->addItem(hist_title(hist_type), hist_type);
    }
    int index = findIndexByType(current_hist_type);
    if (oldIndex == -1 && !newList.isEmpty()) // no previous index, assign first one
        index = 0;
    cb_hist->setCurrentIndex(index);
    cb_hist->adjustSize();

}

void HistExplorerView::recreateHistograms (const HistList &hl)
{
#if 0
    ScopedWidgetDisabler cb_hist_inhibit(cb_hist);
    //  qDebug() << "hl=" << hl;

    // save histogram window state
    QMap<QString, MyWidgetState> histWidgetState;
    for (auto & h, hists) {
        MyWidgetState ws;
        //    ws.rect = h->rect();
        ws.pos = h->pos();
        ws.size = h->size();
        ws.shown = h->isVisible();
        histWidgetState[h->getName()] = ws;
        delete h;
    }
    //  // create list of existing histogram names
    //  QStringList origNameList;
    //  foreach_const(hists, it) {
    //    origNameList.push_back((*it)->getName());
    //  }
    //
    //  // create list of new histogram names
    //  QStringList newNameList;
    //  for(HistList::ConstIterator it = hl.begin(); it!=hl.end(); ++it) {
    //    const HistPtr h1 = it.data();
    //    const QString name(h1->getname());
    //    newNameList.push_back(name);
    //  }
    //  // list of histograms to delete (exist in orig but not in new)
    //  QStringList toDeleteNames;
    //  foreach(origNameList, it) {
    //    const QString &name = *it;
    //    if (!newNameList.contains(name)) {
    //      toDeleteNames.push_back(name);
    //    }
    //  }
    hists.clear();

    QStringList strList;
    for(HistList::const_iterator histIt=hl.begin();histIt != hl.end(); histIt++) {
        th1Ptr h1 = histIt.value();
        const QString name(h1->getname());

        //    if (origNameList.contains(name)) {
        //      foreach(hists, it) {
        //        hist *h = *it;
        //        if (h->getName() == name) {
        //          h->set_h(h1);
        //          break;
        //        }
        //      }
        //    } else {
        HistDialog * h = new HistDialog (this, h1);
        h->setid (histIt.key());
        hists.push_back (h);
        if (histWidgetState.contains(name)) {
            const MyWidgetState &ws = histWidgetState[name];
            //        h->setGeometry(ws.rect);
            h->resize(ws.size);
            h->move(ws.pos);
            h->setVisible(ws.shown);
        }
        //    }
        strList.append(name);
    }

    //  // delete unused histograms
    //  for(hist* h = hists.first(); h; h = hists.next()) {
    //    const QString name(h->getName());
    //    if (toDeleteNames.contains(name)) {
    //      hists.remove(h); // iterator will point to next
    //    }
    //  }

    // update ComboBox
    QStringList origList;
    for(int i=0; i<cb_hist->count(); i++)
        origList << cb_hist->itemText(i);
    // no ComboBox update if contents unchanged
    if (origList == strList)
        return;
    QString selectionText = cb_hist->currentText();
    int selectionIndex = 0;
    if (int n = strList.indexOf(selectionText) != -1)
        selectionIndex = n;
    cb_hist->clear();
    cb_hist->addItems(strList);
    cb_hist->setCurrentIndex(selectionIndex);
    cb_hist->adjustSize();
    //  qDebug() << "strList=" << strList;
    //  qDebug() << "------------------------------------------";
#endif
}

void HistExplorerView::setCurrentHistType( int )
{
    updateView();
}

QString HistExplorerView::currentHistName() const
{
    return currentHist().getname();
}

void HistExplorerView::updateView()
{
    update_mainhist_controls();
    update_current_main_hist();
}

void HistExplorerView::update_mainhist_controls()
{
    const hist_index_t index = currentHistIndex();
    spinBoxChannel->setEnabled(hist_is_multichannel(index.t));
    update_checkBox_autoRange();
    update_checkBox_autoFit();
    //  update_current_main_hist(); -- don't!!!
}

//void HistExplorerView::copyHistData(const HistPlot *hp)
//{
//    if (hp) {
//        const auto &px = hp->getpointsx();
//        const auto &py = hp->getpointsy();
//        const auto &fpx = hp->getfitpointsx();
//        const auto &fpy = hp->getfitpointsy();
//        histPlot->setpoints(px, py);
//        histPlot->set_fit_points(fpx, fpy);
//        double xmin, xmax;
//        if (hp->get_x_range(&xmin, &xmax))
//            histPlot->set_x_range(xmin, xmax);
//    } else {
//        histPlot->clear();
//    }
//    histPlot->replot();
//}

void HistExplorerView::update_current_main_hist()
{
    if (!isVisible()) return;
    //    cb_hist->setCurrentText(hist_title(currentHistType()));
    const th1 &h = currentHist();
    histPlot->setHist(h);
    QString text;
    QStringList histLines;
    histLines << QString("<b>Histogram</b>");
    histLines << QString("Entries: %1").arg(h.get_nent());
    histLines << QString("Overflows: %1").arg(h.get_over());
    histLines << QString("Underflows: %1").arg(h.get_under());
    histLines << QString("Sum: %1").arg(h.get_sum());
    QStringList statLines;
    statLines << QString("<b>Data statistics</b>");
    {
        const auto &stats = h.simplestats();
        statLines << QString("Mean: %1 %2")
                         .arg(stats.mean())
                         .arg(h.get_x_label());
        statLines << QString("StdDev: %1 %2")
                         .arg(stats.stddev())
                         .arg(h.get_x_label());
    }
    if (0) {
        unsigned int bfirst = 0;
        unsigned int blast = h.get_nbins() -1;
        h.find_autorange_bins(&bfirst, &blast);
        const auto &py = h.get_py();
        double par[3];
        FindGausApproxD (&py[bfirst], blast-bfirst, &par[0]);
        statLines << QString("Value: %1").arg(par[0]);
        statLines << QString("Mean: %1").arg(par[1] * h.binwidth() + h.bincenter(0));
        statLines << QString("StdDev: %1").arg(par[2] * h.binwidth());
    }
    QStringList fitLines;
    if (h.fit_enabled() && h.get_nent() > 0) {
        double value=0;
        double mean=0;
        double sigma=0;
        double norm=0;
        h.update_autorange();
        h.fitgaus (&value, &mean, &sigma, &norm);
        fitLines << QString("<b>Gaussian fit</b>");
        fitLines << QString("Value: %1").arg(value);
        fitLines << QString("Mean: %1 %2").arg(mean).arg(h.get_x_label());
        fitLines << QString("Sigma: %1 %2").arg(sigma).arg(h.get_x_label());
        fitLines << QString("norm: %1").arg(norm);
    }
    labelInfo->setText(statLines.join("<br>") + "<p>"
                       + histLines.join("<br>") + "<p>"
                       + fitLines.join("<br>"));
#if 0

    for(int n=0; n < cb_hist->count(); n++)
    {
        //          bool enable = true;
        //	  if (decodethread)
        //		enable = decodethread->get_hist_enable (n);

        //          Q3ListBox *listbox = cb_hist_1->listBox(); // TODO
        //          Q3ListBoxItem *item = listbox->item(n);    // TODO
        //          if (item) item->setSelectable(enable);     // TODO
    }

    HistDialog * histDialog = current_hist();
    if (!histDialog) {
        clear_current_main_hist();
        return;
    }

    const HistPlot *hp = histDialog->getplot();
    const th1Ptr th = histDialog->get_h();
    if (!hp || !th) {
        clear_current_main_hist();
        return;
    }
    histDialog->update();
    copyHistData(hp);

//    double ar_min=0, ar_max=0;
//    th->update_autorange();
    //	th->update_autofit();
//    th->getrange(&ar_min, &ar_max);
    bool autofit = th->fit_enabled();
    if (autofit) {
        bool ok;
        double v, m, s, n;
        histDialog->get_fit_param(&ok, &v, &m, &s, &n);
        bool fit_res = th->get_options() & th1::FIT_RESOL;
        QString str;
        if (ok) {
            str = QString("Fit: value=%1, mean=%2, sigma=%3, FWHM=%4").arg(v).arg(m).arg(s).arg(s*GausFWHM);
            if (fit_res)
                str += QString(", resolution=%1%%").arg(100.*s*GausFWHM/m);
        } else
            str = "No fit parameters available";
        setStatus(str);
    } else {
        setStatus("");
    }
    update_current_main_hist_bh(true);
#endif
}

void HistExplorerView::update_current_main_hist_bh(bool enable_controls)
{
    histPlot->setEnabled(enable_controls);
    pbShow->setEnabled(enable_controls);
    checkBox_autoRange->setEnabled(enable_controls);
    checkBox_autoFit->setEnabled(enable_controls);
}

void HistExplorerView::update_checkBox_autoRange()
{
#if 0
    const HistDialog * h = current_hist();
    if (!h) return;
    const th1Ptr th = h->get_h();
    if (!th) return;
    bool state = th->autorange_enabled();
    //  std::cout<<"update "<<th->getname()<<" autoRange: "<<state<<std::endl;
    checkBox_autoRange->setChecked(state);
#endif
}

void HistExplorerView::setAutoRange( bool state )
{
#if 0
    const HistDialog * h = current_hist();
    if (!h) return;
    const th1Ptr th = h->get_h();
    if (!th) return;
    th->set_autorange_enable(state);
    //  std::cout<<"set "<<th->getname()<<" autoRange: "<<state<<std::endl;
    update_current_main_hist();
#endif
}

void HistExplorerView::update_checkBox_autoFit()
{
#if 0
    const HistDialog * h = current_hist();
    if (!h) return;
    const th1Ptr th = h->get_h();
    if (!th) return;
    bool state = th->fit_enabled();
    checkBox_autoFit->setChecked(state);
    //  std::cout<<"update "<<th->getname()<<" autoFit: "<<state<<std::endl;
#endif
}

void HistExplorerView::setAutoFit( bool state )
{
#if 0
    const HistDialog * h = current_hist();
    if (!h) return;
    const th1Ptr th = h->get_h();
    if (!th) return;
    //  std::cout<<"set "<<th->getname()<<" autoFit: "<<state<<std::endl;
    th->set_fit_enable(state);
    update_current_main_hist();
#endif
}

void HistExplorerView::clearCurrentHist()
{
    emit clearHist(currentHistIndex());
}

void HistExplorerView::save()
{
#if 0
    const HistDialog * histDialog = const_current_hist();
    if (histDialog){
        const th1Ptr histPtr = histDialog->get_h();
        if (histPtr)
            th1_gui::save_to_file(*histPtr, this);
    }
#endif
}


void HistExplorerView::saveAll()
{
#if 0
    QString subdir = QDateTime::currentDateTime().toString("yyyy-MM-dd_hh-mm-ss");
    for (auto & h : hists) {
        const th1Ptr hist = h->get_h();
        th1_gui::save_to_file(*hist, this, subdir);
    }
#endif
}

void HistExplorerView::setCurrentChannel(int)
{
    update_current_main_hist();
}
