#include <unistd.h>
#include <QTimer>
#include <QApplication>
#include <QList>
#include <QDebug>
#include <QSettings>

#include "runcontrol.h"
#include "adcmlib/daq_ipc.h"

const unsigned int ERROR_RETRY_COUNT = 3; // number of automatic initialization retries
const double periodicTask_time = 1.0;

QList<daq_state_t> RunControl::getAllStates()
{
    QList<daq_state_t> allStates;
    allStates += ST_OFF;
    allStates += ST_RESET;
    allStates += ST_RUN;
    allStates += ST_PAUSE;
    allStates += ST_ERROR;
    allStates += ST_FAILURE;
    return allStates;
}

//--------------------------------------------------------------------------
RunControl::RunControl(QObject *parent)
    : QObject(parent),
    periodicTaskTimer(new QTimer(this))
{
    rcState = ST_OFF;

    statusBgColor[ST_OFF] = Qt::gray;
    statusBgColor[ST_RESET] = Qt::cyan;
    statusBgColor[ST_RUN] = Qt::green;
    statusBgColor[ST_PAUSE] = Qt::yellow;
    statusBgColor[ST_ERROR] = Qt::red;
    statusBgColor[ST_FAILURE] = Qt::magenta;

    enabled = false; // state change prohibited until explicit enable
    updateStateChangeEnabled();
//    recovery_requested = false;
    setAutoRecover(true);
    setAutoRun(false);

    //Unused
//    connect (periodicTaskTimer, SIGNAL (timeout ()), SLOT (periodicTask ()));
//    periodicTaskTimer->start ((int)(periodicTask_time*1e3), FALSE);
}

// static
bool RunControl::isTransitionAllowed(daq_state_t fromState, daq_state_t toState)
{
    QMap<QPair <daq_state_t, daq_state_t>, bool> transitionAllowed; // fromState, toState
    QList<daq_state_t> allStates = getAllStates();
    for (auto state : allStates) {
        if (state != ST_OFF) transitionAllowed[qMakePair(state, ST_OFF)] = true;
        if (state != ST_RESET) transitionAllowed[qMakePair(state, ST_RESET)] = true;
        if (state != ST_ERROR) transitionAllowed[qMakePair(state, ST_ERROR)] = true;
        if (state != ST_FAILURE) transitionAllowed[qMakePair(state, ST_FAILURE)] = true;
    }
    transitionAllowed[qMakePair(ST_RESET, ST_RUN)] = true;
    transitionAllowed[qMakePair(ST_RUN, ST_PAUSE)] = true;
    transitionAllowed[qMakePair(ST_PAUSE, ST_RUN)] = true;

    return transitionAllowed[qMakePair(fromState, toState)];
}

// static
daq_state_t RunControl::nextState(daq_state_t fromState)
{
    QMap<daq_state_t, daq_state_t> transitionNext; // used to automatically change state
    transitionNext[ST_FAILURE] = ST_OFF;
    transitionNext[ST_ERROR] = ST_OFF;
    transitionNext[ST_OFF] = ST_RESET;
    transitionNext[ST_RESET] = ST_RUN;
    transitionNext[ST_RUN] = ST_RUN; // end state
    transitionNext[ST_PAUSE] = ST_PAUSE; // end state
    daq_state_t next = transitionNext[fromState];
    //  if (next == 0) // 0 is ST_OFF
    //	return fromState;
    return next;
}

// static
daq_state_t RunControl::highlightState(daq_state_t fromState)
{
    daq_state_t highlight = nextState(fromState);
    if (fromState == ST_RUN) highlight = ST_PAUSE;
    if (fromState == ST_PAUSE) highlight = ST_RUN;
    return highlight;
}

// static
QList<daq_state_t> RunControl::statePath(daq_state_t fromState, daq_state_t toState)
{
    int steps = 0;
    daq_state_t tmpState = fromState;
    QList<daq_state_t> path;
    while (steps < 10) {
        if (tmpState == toState) return path;
        tmpState = nextState(tmpState);
        path += tmpState;
    }
    qWarning() << QString("statePath: no path found from %1 to %2").arg(daq_state_str(fromState)).arg(daq_state_str(toState));
    return QList<daq_state_t>();
}

QColor RunControl::statusPaletteBackgroundColor() const
{
    QColor color = statusBgColor[rcState];
    if (!color.isValid()) color = Qt::red;
    return color;
}

void RunControl::transitionPreTask (daq_state_t toState)
{
//    qDebug() << QString("pre-task for state %1").arg(daq_state_str(toState));
    if (toState == ST_OFF || toState == ST_RESET) {
        //	clearAllErrors();
    }
    if (toState == ST_ERROR || toState == ST_FAILURE) {
        error_state_count++;
//        qDebug() << QString("error_state_count %1").arg(error_state_count);
    }
}

void RunControl::changeStateForced (daq_state_t toState)
{
    clearAllErrors();
    changeState(toState);
}

void RunControl::clearAllErrors()
{
    error_state_count = 0;
}

void RunControl::changeState (daq_state_t toState)
{
    if (!enabled) {
//        qDebug() << QString("skipped disabled state change request for %1").arg(daq_state_str(toState));
        return;
    }
//    qDebug() << QString("changeState requested: %1 -> %2").arg(daq_state_str(rcst)).arg(daq_state_str(toState));
    if (rcState == toState) return;
    if (!isTransitionAllowed(rcState, toState)) {
        qWarning() << (QString("Invalid state transition requested: %1 !-> %2").arg(daq_state_str(rcState)).arg(daq_state_str(toState)));
        return;
    }

    transitionPreTask(toState);
    taskRunning = true;
    updateStateChangeEnabled();
    emit transitionTask(toState);
}

void RunControl::transitionTaskComplete(daq_state_t toState)
{
    rcState = toState;
//    qDebug() << QString("task complete for state %1").arg(daq_state_str(rcst));
    emit stateChanged(rcState);
    taskRunning = false;
    updateStateChangeEnabled();
    usleep(100000); // FIXME "Exception in adc_get_counters(): MlinkDevice::ctrlExchangeSingle(): Receive timeout"
    autoRun();
}

void RunControl::transitionTaskFailed(daq_state_t toState)
{
    taskRunning = false;
    updateStateChangeEnabled();
    qWarning() <<  (QString("task failed for state %1").arg(daq_state_str(toState)));
    // simulate task_error
    transitionPreTask(ST_ERROR);
    transitionTaskComplete(ST_ERROR);
    //  if (rcst != ST_ERROR)
    //    changeState(ST_ERROR);
    //  autoRun();
}

void RunControl::periodicTask ()   // automatic state change (by timer)
{
    /*
  if (!mutex1.tryLock()) return;
  if (rcst == ST_OFF)
    changeState (ST_INIT);                 // initialize to known state after cold start
  if (rcst == ST_INIT)
    changeState(ST_RESET);                // reset after init is essential
  if (autorun && rcst == ST_RESET) changeState(ST_RUN); // run
  if (autorun && rcst == ST_ERROR) {
    if (autorecover && (error_state_count < ERROR_RETRY_COUNT)) {
          log_notice("autorecover");
      changeState(ST_INIT);                // try to recover - dangerous
    } else
      changeState(ST_FAILURE);
  }
//  if (rcst == ST_FAILURE) {
    if (recovery_requested) {
        error_state_count=0;
        changeState(ST_INIT);
        recovery_requested = false;
    }
//  }

//  if (rcst == ST_INIT || rcst == ST_RESET || rcst == ST_RUN)
//    error_state_count = 0;
  mutex1.unlock();
*/
}
/*
void RunControl::tryRecover()
{
  recovery_requested = true;
}
*/
void RunControl::setAutoRecover(bool value)
{
    autorecover = value;
}

void RunControl::setAutoRun(bool value)
{
    autorun = value;
    autoRun();
}

void RunControl::readSettings()
{
    QSettings settings;
    setAutoRun(settings.value("autostart").toBool());
    setAutoRecover(settings.value("autorecover").toBool());
}

void RunControl::writeSettings()
{
    QSettings settings;
    settings.setValue("autostart",autorun);
    settings.setValue("autorecover", autorecover);
}

void RunControl::enable()
{
    enabled = true;
    updateStateChangeEnabled();
    autoRun();
}

void RunControl::disable()
{
    enabled = false;
    updateStateChangeEnabled();
}

void RunControl::autoRun()
{
    if (!enabled)
        return;
    const daq_state_t next = nextState(rcState);
//    qDebug() << QString("autorun: error_state_count=%1, autoStateChangeEnabled=%2, this=%3, next=%4")
//                 .arg(error_state_count).arg(autoStateChangeEnabled())
//                 .arg(daq_state_str(rcst)).arg(daq_state_str(next));
    if ((rcState == ST_OFF) || autoStateChangeEnabled()) {
        if (next != rcState) changeState(next);
    }
}

bool RunControl::isStateChangeEnabled()
{
    return enabled && !taskRunning;
}

void RunControl::updateStateChangeEnabled()
{
    emit stateChangeEnabled(isStateChangeEnabled());
}

bool RunControl::isAutorunBlockedByError()
{
    return autorecover?(error_state_count >= ERROR_RETRY_COUNT):error_state_count;
}

bool RunControl::autoStateChangeEnabled()
{
    if (!isStateChangeEnabled()) return false;
    return autorun && !isAutorunBlockedByError();
}
