/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AdcmApp.h"

#include <chrono>

#include <QThread>
#include <QTimer>

#include "config/ConfigManager.h"
#include "adcinstance.h"
#include "DecodeManager.h"
#include "EventListFiller.h"
#include "RawDataWriter.h"
#include "RemoteControlServer.h"
#include "hist_manager/histmanager.h"
#include "hist_manager/HistMerger.h"
#include "adcmdevice/FrameList.h"
#include "adcm_df.h"
#include "util/TextFormatter.h"

using namespace std::chrono;

static const double threadWaitTimeMsec = 1000;

AdcmApp::AdcmApp(QObject *parent)
    : QObject(parent),
    configManager(new ConfigManager(this)),
    adcInstance(new AdcInstance(nullptr)),
    decodeManager(new DecodeManager(configManager, nullptr)),
    filler(decodeManager->createFiller()),
    writeInstance(new RawDataWriter (nullptr)),
    remoteControlServer(new RemoteControlServer()),
    histMerger(new HistMerger(this)),

    // threads
    adcThread(new QThread(this)),
    decodethread(new QThread(this)),
    fillerThread(new QThread(this)),
    writethread(new QThread(this)),
    remoteControlThread(new QThread(this))

{
    setObjectName("ADCM app");
    adcThread->setObjectName("adcInst");
    decodethread->setObjectName("decode Mgr");
    fillerThread->setObjectName("filler");
    writethread->setObjectName("writer");
    remoteControlThread->setObjectName("remControl");

    remoteControlServer->moveToThread(remoteControlThread);
    decodeManager->moveToThread(decodethread);
    filler->moveToThread(fillerThread);
    writeInstance->moveToThread(writethread);
    adcInstance->moveToThread(adcThread);

    // configuration updates
    qRegisterMetaType<conf_mem_t>();
    connect(adcInstance, &AdcInstance::confMemChanged, configManager, &ConfigManager::setAdcConfMem);
    connect(configManager, &ConfigManager::adcConfMemChanged, filler, &EventListFiller::setAdcConfMem);
    connect(configManager, &ConfigManager::adcConfMemChanged, decodeManager, &DecodeManager::setAdcConfMem);

    qRegisterMetaType<AdcConfig>();
    connect(configManager, &ConfigManager::adcConfigChanged, adcInstance, &AdcInstance::setAdcConfig);
    connect(configManager, &ConfigManager::adcConfigChanged, decodeManager, &DecodeManager::setAdcConfig);

    qRegisterMetaType<DecoderConfig>();
    connect(configManager, &ConfigManager::decoderConfigChanged, decodeManager, &DecodeManager::setDecoderConfig);
    connect(configManager, &ConfigManager::calZeroConfigChanged, adcInstance, &AdcInstance::setCalZeroConfig);

    qRegisterMetaType<StorageConfig>();
    connect(configManager, &ConfigManager::evStorageConfigChanged, decodeManager, &DecodeManager::setStorageConfig);
    connect(configManager, &ConfigManager::rawStorageConfigChanged, writeInstance, &RawDataWriter::setStorageConfig);

    // data pipeline
    qRegisterMetaType<FrameList>();
    connect(adcInstance, SIGNAL(frameReceived(FrameList)), filler, SLOT(appendFrameList(FrameList)));
    connect(filler, SIGNAL(completed(quint64)), adcInstance, SLOT(dataProgress(quint64)));
    connect(adcInstance, SIGNAL(frameReceived(FrameList)), writeInstance, SLOT(writeFrameList(FrameList)));

    qRegisterMetaType<adcm_counters_t>();
    connect(adcInstance, SIGNAL(hitCountsChanged(adcm_counters_t)), decodeManager, SLOT(appendChannelCounters(adcm_counters_t)));

    connect(decodeManager, &DecodeManager::decodedCalZero, configManager, &ConfigManager::addCalZero);
    connect(decodeManager, &DecodeManager::decodedCalAvg, configManager, &ConfigManager::addCalAvg);
//    connect(configManager, &ConfigManager::calAvgConfigChanged, decodeManager, &DecodeManager::setCalAvgConfig);
//    connect(configManager, &ConfigManager::calZeroConfigChanged, decodeManager, &DecodeManager::setCalZeroConfig);

    connect(decodeManager, &DecodeManager::histogramsChanged, histMerger, &HistMerger::addHistograms);

    connect(histMerger, &HistMerger::histogramsChanged, this, &AdcmApp::onHistogramsChanged);
    connect(this, &AdcmApp::clearAllHists, histMerger, &HistMerger::clearAllHists);
    connect(this, &AdcmApp::clearHist, histMerger, &HistMerger::clearHist);

    QTimer::singleShot(0, this, &AdcmApp::startThreads);
}

AdcmApp::~AdcmApp()
{
    stopThreads();
}

void AdcmApp::startThreads()
{
    decodethread->start();
    decodethread->setPriority(QThread::LowestPriority);

    //    QMetaObject::invokeMethod(decodeManager, "loop", Qt::QueuedConnection);

    fillerThread->start();
    fillerThread->setPriority(QThread::LowPriority);

    writethread->start();
    writethread->setPriority(QThread::LowPriority);

    adcThread->start();
    adcThread->setPriority(QThread::HighPriority);

    remoteControlThread->start();
    remoteControlThread->setPriority(QThread::LowPriority);

    int waitMs = 10000;
    QDateTime start(QDateTime::currentDateTime());
    while(!adcThread->isRunning() || !decodethread->isRunning() || !writethread->isRunning()) {
        QDateTime now(QDateTime::currentDateTime());
        int elapsed = now.toTime_t() - start.toTime_t();
    }
}

void AdcmApp::stopThreads()
{
    // ask threads to exit from run()
    QList<QThread *> threads = {
        adcThread,
        writethread,
        decodethread,
        fillerThread,
        remoteControlThread
    };
    for (auto i : threads)
        i->quit();

    auto start = steady_clock::now();
    while (!threads.empty()) {
        for (auto i : threads)
            if (! i->isRunning()) {
                threads.removeOne(i);
                qDebug().noquote() << i->objectName() << "thread exited";
                break;
            }
        auto now = steady_clock::now();
        double ms = duration_cast<milliseconds>(now - start).count();
        if (ms > threadWaitTimeMsec) {
            QStringList sl;
            for (auto i : threads)
                sl << i->objectName();
            qCritical().noquote() << "Timeout waiting threads to stop:" << sl.join(", ");
            break;
        }
    }
}

void AdcmApp::onHistogramsChanged(const HistList &r)
{
    emit histogramsChanged(r);
}
