#include "DecodeManager.h"

#include <unistd.h>

#include <QDebug>
#include <QDir>
#include <QSemaphore>
#include <QThread>
#include <QTime>
#include <QCoreApplication>

#include "hist_manager/histmanager.h"
#include "DecodeWorker.h"
#include "config/ConfigManager.h"
#include "config/adcconfig.h"
#include "adcmraw/adcm_mem.h"
#include "adcmlib/adcm_dec.h"
#include "adcmlib/adcm_decstat.h"
#include "adcm_df.h"
#include "adcmlib/adcm_ev.h"
#include "adcmraw/adcm_ids.h"
#include "adcmlib/timer.h"
#include "storage.h"
#include "adcmlib/adcm_el.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_types.h"
#include "EventListFiller.h"

const int DecodeManager::numberWorkers = QThread::idealThreadCount ();
//const int DecodeManager::numberWorkers = 1;

#define PREFIX "DecodeManager: "

namespace {
const int updateHistogramsTime = 700; // in ms

const int MIN_EVENTS_PACK_SIZE = 100;
const int EVENTS_PACK_SIZE = 1000;
const int EVENT_LIST_SIZE = EVENTS_PACK_SIZE*DecodeManager::numberWorkers*5;
const int RUN_LOOP_PERIOD_MS = 100;
const int RUN_LOOP_TIMER_MS = 1000;
//const size_t DECODE_MAX_EVENTS_ONCE = 100;
//const size_t DECODE_MAX_WORDS_ONCE = 1024*256; // 1MByte
//const size_t DECODE_MAX_FRAMES_ONCE = 1000;

// 4 threads, full mode, 1000 frames; debug:up to 100ms, release:up to 50ms
//const uint waitWorkersTimeoutframesPackSize =EVENTS_PACK_SIZE/DecodeManager::numberWorkers;
}

DecodeManager::DecodeManager(ConfigManager *configManager, QObject *parent):
    QObject(nullptr),
    adcConfig(new AdcConfig),
    storageConfig(new StorageConfig),
    decoderConfig(new DecoderConfig),
    hmgr(new HistManager(this)),
    counters(new adcm_counters_t),
    stor(new storage),
    idleWorkers(new QSemaphore(numberWorkers)),
    freeEvents(new QSemaphore(EVENT_LIST_SIZE)),
    managerEvents(new QSemaphore(0)),
    pauseSem(new QSemaphore(0)),
    lastEventListSize(EVENT_LIST_SIZE),
    eventList(new adcm_el_t(freeEvents, managerEvents, EVENT_LIST_SIZE)), // 4*numberWorkers*EVENTS_PACK_SIZE
    summedDecStat(new adcm_decstat_t),
    lastHistSumTime(new QTime),
    workStatTimer(new QTime),
    jobTimer(new QTimer(this))
{
    Q_ASSERT(numberWorkers>0);
    qRegisterMetaType<adcm_decstat_t>();
    qRegisterMetaType<EventListIt>();
    hmgr->setObjectName("HistManager[master]");
    for(int i=0; i<numberWorkers; ++i){
        auto worker = new DecodeWorker(i, idleWorkers, stor, decoderConfig, eventList, storageConfig->enable);
        workers.append(worker);
        auto workerThread = new QThread(this);
        workerThreads.append(workerThread);

        worker->moveToThread(workerThread);
        workerThread->setObjectName(QString("Worker:%1").arg(i));
        workerThread->start();
        workerThread->setPriority(QThread::LowestPriority);

        if(i==0){
            qRegisterMetaType<AdcmEvent>();
            connect(worker, &DecodeWorker::monitorEvent, this, &DecodeManager::on_monitorEvent); // FIXME
        }

        connect(worker, &DecodeWorker::statisticsUpdated, this, &DecodeManager::addWorkerStatistics);
//        qRegisterMetaType<CalZeroConfig>();
        connect(worker, &DecodeWorker::decodedCalZero, configManager, &ConfigManager::addCalZero, Qt::QueuedConnection);
//        qRegisterMetaType<CalAvgConfig>();
        connect(worker, &DecodeWorker::decodedCalAvg, configManager, &ConfigManager::addCalAvg, Qt::QueuedConnection);

        connect(worker, &DecodeWorker::histogramsChanged, this, &DecodeManager::on_histogramsChanged);

        connect(this, &DecodeManager::decoderConfigModified, worker, &DecodeWorker::setDecoderConfig);
        connect(this, &DecodeManager::adcConfigModified, worker, &DecodeWorker::setAdcConfig);
        connect(this, &DecodeManager::adcConfMemChanged, worker, &DecodeWorker::setAdcConfMem);

//        connect(this, &DecodeManager::clearAllHists, worker, &DecodeWorker::clearAllHists);
//        connect(this, &DecodeManager::clearHist, worker, &DecodeWorker::clearHist);
    }
//    connect(this, &DecodeManager::adcConfigModified, hmgr, &HistManager::setAdcConfig);
//    connect(this, &DecodeManager::adcConfMemChanged, hmgr, &HistManager::setAdcConfMem);

    qRegisterMetaType<HistList>();
    connect(hmgr, &HistManager::histListModified, this, &DecodeManager::on_histListModified);
//    connect(hmgr, &HistManager::histogramsChanged, this, &DecodeManager::on_histogramsChanged);

//    connect (&updateHistogramsTimer, SIGNAL (timeout ()), SLOT (update_mainhist_view ()));
//    updateHistogramsTimer.start (updateHistogramsTime);

//    connect(this, &DecodeManager::clearAllHists, hmgr, &HistManager::clearAllHists);
//    connect(this, &DecodeManager::clearHist, hmgr, &HistManager::clearHist);

    qInfo()<<QString("Decode manager has %1 worker(s); eventList size=%2; max pack size=%3; min pack size=%4")
              .arg(numberWorkers).arg(EVENT_LIST_SIZE)
              .arg(EVENTS_PACK_SIZE).arg(MIN_EVENTS_PACK_SIZE);

    lastHistSumTime->start();
    workStatTimer->start();
    jobTimer->setSingleShot(true);
    connect(jobTimer, &QTimer::timeout, this, &DecodeManager::loop);
    jobTimer->start(1);
}

DecodeManager::~DecodeManager()
{
    paused = true;
    foreach(QThread *workerThread, workerThreads){
        workerThread->quit();
    }
    foreach(QThread *workerThread, workerThreads){
        if(!workerThread->wait(100/numberWorkers))
            qWarning()<<"worker aborted";
        delete workerThread;
    }
    workerThreads.clear();

    delete adcConfig;
    delete storageConfig;
    delete decoderConfig;
    delete hmgr;
    delete counters;
    delete stor;

    delete idleWorkers;
    delete freeEvents;
    delete managerEvents;
    delete eventList;
    delete lastHistSumTime;
    delete workStatTimer;
}

void DecodeManager::on_histListModified(const HistList &r)
{
    emit histListModified(r);
}

void DecodeManager::on_histogramsChanged(const HistList &r)
{
    emit histogramsChanged(r);
}

void DecodeManager::on_monitorEvent(const AdcmEvent &ev)
{
    emit monitorEvent(ev);
}

EventListFiller * DecodeManager::createFiller()
{
    auto filler = new EventListFiller(eventList, this);
    connect(filler, SIGNAL(statisticsUpdated(adcm_decstat_t)), SLOT(addWorkerStatistics(adcm_decstat_t)));
    return filler;
}


void DecodeManager::pause()
{
    paused = true;
    pauseSem->acquire(); // Enshure that loop is not running
    idleWorkers->acquire(numberWorkers); // Wait worker finished their jobs
    idleWorkers->release(numberWorkers);
    clenUpCompletedJobs();
    if(!currentJobs.empty()){
        qWarning()<<"missing events detected";
        info();
        Q_ASSERT(false);
    }
}


void DecodeManager::resume()
{
    paused = false;
}


void DecodeManager::setAdcConfMem(const conf_mem_t &conf_mem)
{
    idleWorkers->acquire(numberWorkers);
    nch = conf_mem.nch;
    emit adcConfMemChanged(conf_mem);
    clearCounters();
    idleWorkers->release(numberWorkers);
}


void DecodeManager::reset()
{
    pause();

    int eventSize = managerEvents->available();
    if(eventSize){
        managerEvents->acquire(eventSize);
        freeEvents->release(eventSize);

        Q_ASSERT(eventList->getStartWorkerPos()==eventList->getEndWorkerPos());

        EventListIt endIt=eventList->getEndWorkerPos();
        eventList->next(endIt, eventSize);
        eventList->setStartWorkerPos(endIt);
        eventList->setEndWorkerPos(endIt);
        Q_ASSERT(eventList->getManagerEvNum()==endIt->first);
    }

    workersDecStat.clear();
    summedDecStat->clear();
    hmgr->clearAllHists();
    stor->reset ();

    if(storageConfig->enable)
        writeChMap ();

    clearCounters();

    checkEvListSize();
    foreach(DecodeWorker *worker, workers){
        worker->reset();
    }

    resume();
}


void DecodeManager::clear()
{
    pause();
    foreach(DecodeWorker *worker, workers){
        worker->lockWorker();
        worker->clear();
        worker->unlockWorker();
    }

    workersDecStat.clear();
    summedDecStat->clear();
    hmgr->clearAllHists();

    endBurst();
    resume();
}


void DecodeManager::loop()
{
    if(paused){
        pauseSem->release();
        jobTimer->start(1);
        return;
    }
    workStatTimer->restart();
    uint loopCount=0;
    while(!paused){
        ++loopCount;

        if(!idleWorkers->tryAcquire(1, RUN_LOOP_PERIOD_MS)){
            qWarning()<<"Workers work too long";
            if(workStatTimer->elapsed() > RUN_LOOP_TIMER_MS)
                break;
            continue;
        }
        idleWorkers->release();

        if( paused || ((!startJob() || loopCount%10 == 0) &&
                       (workStatTimer->elapsed() > RUN_LOOP_TIMER_MS)) ){
                break;
        }
        QCoreApplication::processEvents();
    }

//    qDebug()<<"loop count:"<<loopCount;
//    info();
//    if(!paused) checkEvListSize();

//    if(histEnabled && lastHistSumTime->elapsed()>updateHistogramsTime){
//        hmgr->lock();
//        hmgr->clearAll();
//        foreach (DecodeWorker* worker, workers){
//            worker->lockWorker();
//            hmgr->addHist(worker->getHistManager());
//            worker->unlockWorker();
//        }
//        hmgr->unlock();
//        lastHistSumTime->restart();
//    }

    emit summedStatisticsUpdated(*summedDecStat);
    summedDecStat->clear();
    jobTimer->start(0);
}


void DecodeManager::setAdcConfig(const AdcConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    bool chMemChanged = false;
    if (adcConfig->channels.keys() != r.channels.keys())
        chMemChanged = true;
    for (int i : adcConfig->channels.keys()) {
        if (adcConfig->channels[i].flags.all != r.channels[i].flags.all) {
            chMemChanged = true;
            break;
        }
    }
//    adcConfig->size() != r.size() || adcConfig->trig != r.trig;
    *adcConfig = r;
    emit adcConfigModified(r);
    idleWorkers->release(numberWorkers);

    if(storageConfig->enable && chMemChanged)
        writeChMap ();
}

void DecodeManager::setCalAvgConfig(const CalAvgConfig &r)
{
    DecoderConfig tmp(*decoderConfig);
    tmp.calavg = r;
    setDecoderConfig(tmp);
}

void DecodeManager::setCalZeroConfig(const CalZeroConfig &r)
{
    DecoderConfig tmp(*decoderConfig);
    tmp.calzero = r;
    setDecoderConfig(tmp);
}

void DecodeManager::setDecoderConfig(const DecoderConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    *decoderConfig = r;
    emit decoderConfigModified(r);
    idleWorkers->release(numberWorkers);
}

void DecodeManager::setStorageConfig(const StorageConfig &r)
{
    idleWorkers->acquire(numberWorkers);
    *storageConfig = r;
    stor->setFileName(storageConfig->fileName);
    stor->setopt(storageConfig->enable, storageConfig->mode);
    stor->reset ();

    foreach(DecodeWorker *worker, workers)
        worker->setStorageConfig(storageConfig->enable);
    idleWorkers->release(numberWorkers);
}


void DecodeManager::appendChannelCounters(const adcm_counters_t &cnt)
{
    if (counters->n != cnt.n) {
        counters->clear();
        counters->resize(cnt.n);
    }
    for (int i=0; i<cnt.n; i++)
        counters->rawhits[i] += cnt.rawhits[i];
    counters->time += cnt.time;
}

bool DecodeManager::endBurst()
{
    stor->lock();
    bool ret = true;
    quint16 d;
    d = STOR_ID_CNTR;
    ret &= stor->write (&d, sizeof(d)); // write type
    d = 2 * sizeof(d) + sizeof(adcm_counters_t);
    ret &= stor->write (&d, sizeof(d)); // write total length=size(type)+size(length)+size(data)
    ret &= stor->write (counters, sizeof(adcm_counters_t)); // write data
    clearCounters();
    writeChMap();

    /*
  {
    int i;
    double t;

    t = 1e-8 * ts_last;
    for (i = 0; i < ADC_NCH; i++)
      if (cnt->rawhits[i])
        printf ("Hits [%2d]: %10u   %10.3f kHz\n", i, cnt->rawhits[i], 1e-3 * cnt->rawhits[i] / t);
  }
*/

    stor->flush();
    stor->unlock();
    return ret;
}

void DecodeManager::setHistEnabled(bool val)
{
    histEnabled = val;
    foreach (DecodeWorker* worker, workers) {
        worker->setHistEnabled(val);
    }
}


// Gets workers and filler stats
void DecodeManager::addWorkerStatistics(const adcm_decstat_t &stat, int index)
{
    *summedDecStat += stat;
//    qDebug()<<"Got stat:"<<index;
//    if(workersDecStat.contains(index)) {
//        qDebug()<<"Double statistic for worker:"<<index;
//        workersDecStat[index].add(stat);
//    } else
//        workersDecStat[index] = stat;

//    if(workersDecStat.size() == numberWorkers+1) {
//        adcm_decstat_t summaryStat;
//        foreach (adcm_decstat_t var, workersDecStat.values()) {
//            summaryStat.add(var);
//        }
//        workersDecStat.clear();
//        emit summedStatisticsUpdated(summaryStat);
//    }
}


void DecodeManager::clearCounters()
{
    counters->clear();
}


bool DecodeManager::startJob()
{
    bool noDelay=true;
    uint eventSize;
//    info();

    if(managerEvents->tryAcquire(MIN_EVENTS_PACK_SIZE, RUN_LOOP_PERIOD_MS)){
        eventSize = std::min(managerEvents->available(), EVENTS_PACK_SIZE-MIN_EVENTS_PACK_SIZE);
        managerEvents->acquire(eventSize);
        eventSize += MIN_EVENTS_PACK_SIZE;
    } else {
        clenUpCompletedJobs();
        eventSize = managerEvents->available();
        if(eventSize==0){
//            info();
            return false;
        }
        managerEvents->acquire(eventSize);
        noDelay = false;
    }

    Q_ASSERT(idleWorkers->available());
    idleWorkers->acquire();

    DecodeWorker* idleWorker=getIdleWorker();
    Q_ASSERT(idleWorker);
    idleWorker->lockWorker();

    EventListIt startIt=eventList->getEndWorkerPos();
    EventListIt endIt=startIt;
    eventList->next(endIt, eventSize);

    if(noDelay) clenUpCompletedJobs(idleWorker->getWorkerIndex());

    if(eventList->getStartWorkerPos() != startIt){
        Q_ASSERT(currentJobs.size());
        bool found=false;
        foreach (WorkerJob job, currentJobs) {
            if(job.endPos==startIt){
                found=true;
                break;
            }
        }
        Q_ASSERT(found);
    }
    eventList->setEndWorkerPos(endIt);
    currentJobs.append(WorkerJob(idleWorker->getWorkerIndex(), startIt, endIt, eventSize));
//    qDebug()<<"Add job #"<<idleWorker->getWorkerIndex()<<":" << eventSize << "; " << startIt->first << "-" << endIt->first;
    Q_ASSERT(startIt->first<=0xFFFF);

//    info();
//    qDebug()<<"\n\n\n";
    QMetaObject::invokeMethod(idleWorker, "decodeEventList",
                              Qt::QueuedConnection,
                              Q_ARG(EventListIt, startIt),
                              Q_ARG(EventListIt, endIt));
    return noDelay;
}


void DecodeManager::clenUpCompletedJobs(int index)
{
    uint workerStatus=0;
    foreach (DecodeWorker *worker, workers) {
        if(worker->getWorkerIndex()==index || worker->isIdle())
            workerStatus |= 1<<worker->getWorkerIndex();
    }
    EventListIt workerPos = eventList->getStartWorkerPos();
    int completedEvents=0;
    for(int i=0;i<currentJobs.size();){
        WorkerJob &job = currentJobs[i];
//        if(index>=0 && job.workerIndex == (uint)index){
//            job.completed = true;
//        } else {
            if(1<<job.workerIndex & workerStatus)
                job.completed = true;
//        }


        if(job.startPos==workerPos && job.completed){
            workerPos=job.endPos;
            completedEvents += job.size;
//            qDebug()<<"\t\t\t\tClear   #"<<job.workerIndex<<":"<<job.startPos->first<<"-"<<job.endPos->first
//                   <<" "<< job.size;
            currentJobs.removeAt(i);
            continue;
        }
        ++i;
    }

    if(completedEvents){
        //        qDebug()<<"================";
        //        info();

        eventList->setStartWorkerPos(workerPos);
//        qDebug()<<"\t\t\t\tStart worker:"<<workerPos->first;
        freeEvents->release(completedEvents);
        Q_ASSERT(eventList->getStartWorkerPos() == workerPos);
        //        qDebug()<<"Clean size:"<<completedEvents;
        //        info();

        //        qDebug()<<"================";
    }
    Q_ASSERT(currentJobs.size()==0 || currentJobs[0].startPos == workerPos);
}


void DecodeManager::info()
{
    EventListIt sW = eventList->getStartWorkerPos();
    EventListIt eW = eventList->getEndWorkerPos();
    uint mP = eventList->getManagerEvNum();
    uint begin = eventList->getBeginEvNum();
    qInfo()<<QString("eventList begin:%1; workers:%2-%3; manager:%4")
              .arg(begin)
              .arg(sW->first).arg(eW->first)
              .arg(mP);
    int m = managerEvents->available();
    int f = freeEvents->available();
    int jobs=0;
    if(!currentJobs.empty()){
    foreach (WorkerJob job, currentJobs) {
        jobs += job.size;
        qInfo()<<QString("job #%1: %2-%3 %4").arg(job.workerIndex)
                  .arg(job.startPos->first).arg(job.endPos->first)
                  .arg(job.size);
    }
    } else {
        qInfo()<<"No pending jobs";
    }
    qInfo()<<QString("man ev:%1; free ev:%2; jobs:%3; sum:%4")
              .arg(m).arg(f).arg(jobs).arg(m+f+jobs);
}


void DecodeManager::checkEvListSize()
{
    return;
    {
    QList<WorkerJob> jobs = currentJobs;
    EventListIt sp = eventList->getStartWorkerPos();
    EventListIt ep = eventList->getEndWorkerPos();
    while(!jobs.empty()){
        int i;
        for(i=0;i<jobs.size();++i){
            if(sp==jobs[i].startPos){
                sp=jobs[i].endPos;
                break;
            }
        }
        if(i==jobs.size()){
            info();
            qDebug()<<"NotFound";
            break;
        } else {
            jobs.removeAt(i);
        }
    }
    if(sp!=ep){
        info();
        qDebug()<<"chain is broken";
    }
    }

    uint freeSize=0;
    freeSize=std::max(freeEvents->available()-10,0);
    while(!freeEvents->tryAcquire(freeSize)){
        freeSize-=10;
        if(freeSize<=0){
            freeSize=0;
            break;
        }
    }
    while(freeEvents->tryAcquire()){
        ++freeSize;
    }
    usleep(10);
    eventList->flushManEvents();
    uint manSize=managerEvents->available();
    uint jobs=0;
    foreach(WorkerJob job, currentJobs){
        jobs+=job.size;
    }
    uint sum = freeSize+manSize+jobs;
    if(qAbs(sum-lastEventListSize)>0){
        qDebug()<<"================"<<lastEventListSize<<"-"<<sum<<"="<<lastEventListSize-sum<<"================";
        lastEventListSize=sum;
    }
    freeEvents->release(freeSize);
}


DecodeWorker* DecodeManager::getIdleWorker()
{
    for (auto & worker : workers)
        if (worker->isIdle())
            return worker;
    return nullptr;
}


void DecodeManager::writeChMap()
{
    QMap<int, quint8> chMap;

    for (auto c : adcConfig->channels.keys()) {
        adcm_channel_flags_t trig = adcConfig->channels[c].flags;
        trig.bit.test = (decoderConfig->mode == DM_ZERO);
        chMap[c] = trig.all;
    }
    int max_c = nch;
    for (auto c : adcConfig->channels.keys()) {
        if (max_c < c)
            max_c = c;
    }

    bool ret = true;
    quint16 blk_id = STOR_ID_CMAP;
    ret &= stor->write (&blk_id, 2);
    quint16 blk_size = 4 + 4 + max_c;
    ret &= stor->write (&blk_size, 2);
    quint32 cmap_n = max_c;
    ret &= stor->write (&cmap_n, 4);
    for (int c=0; c<max_c; c++) {
        quint8 map = chMap[c];
        ret &= stor->write (&map, sizeof(map));
    }
}
