/*
**    Copyright 2014-2017 Ivan Filippov
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "EventListFiller.h"

#include <QDebug>
#include <QSemaphore>

#include "DecodeManager.h"
#include "adcmdevice/FrameList.h"
#include "decoder/AdcmDataBlockDecoder.h"
#include "adcmlib/adcm_el.h"
#include "adcmraw/adcm_ids.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmraw/adcm_mem.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_types.h"
#include "adcmlib/timer.h"

#ifndef DEBUG_PRINT_DATA_RATE
static const int DEBUG_PRINT_DATA_RATE = 0;
#endif

namespace{
const int updateProgressTimeout = 300; // in ms
}

EventListFiller::EventListFiller(adcm_el_t *evList, DecodeManager * decodeManager) :
    QObject(nullptr),
    eventList(evList),
    adcmDataBlockDecoder(new AdcmDataBlockDecoder(0)),
    fillerDecStat(new adcm_decstat_t),
    manager(decodeManager),
    lastProgressTime (new QTime),
    workTime(new QTime),
    statisticTimer(new QTimer(this))
{
    lastProgressTime->start();
    workTime->start();
    connect(statisticTimer, SIGNAL(timeout()), SLOT(sendStatistics()));
    statisticTimer->setInterval(1000);
    statisticTimer->start();
}

EventListFiller::~EventListFiller()
{
    delete adcmDataBlockDecoder;
    delete fillerDecStat;
    delete lastProgressTime;
    delete workTime;
    delete statisticTimer;
}

void EventListFiller::appendFrameList(const FrameList &frameList)
{
    if (DEBUG_PRINT_DATA_RATE){
        workTime->start();
    }

    quint64 bytesDecoded=0;
    framesDecoded=0;
    frameListSize=frameList.size();
    FrameList::const_iterator currentFrame = frameList.begin();
    FrameList::const_iterator end = frameList.end();

    updateProgress(false);
    while(currentFrame != end){
        appendFrame(*currentFrame);
        if (DEBUG_PRINT_DATA_RATE) bytesDecoded += currentFrame->size();
        ++currentFrame;
        if(++framesDecoded%500==0){
            updateProgress();
        }
    }
    eventList->flushManEvents();
    updateProgress(false);
    if (DEBUG_PRINT_DATA_RATE){
        double rate=workTime->elapsed();
        if(rate){
            rate=bytesDecoded/1024/1024 / (rate*1e-3);
        } else {
            rate=-1;
        }
        qInfo() << "Decoder throughput:" << rate << "MB/s";
    }

//    qDebug()<<"Frame list completed; size="<<frameList.size();
    sendStatistics();
    emit completed(frameList.seq);
}


void EventListFiller::setAdcConfMem(const conf_mem_t &conf_mem)
{
    manager->pause();
    adcmDataBlockDecoder->setAdcConfMem(conf_mem);
    manager->resume();
}

void EventListFiller::clear()
{
    fillerDecStat->clear();
    manager->clear();
}


void EventListFiller::reset()
{
    manager->reset();
    statisticTimer->start();
    eventList->reset();

    adcmDataBlockDecoder->reset();

    fillerDecStat->clear();

    lastProgressTime->restart();
    workTime->restart();
}

void EventListFiller::sendStatistics()
{
//    emit requestWorkerStatistic();
    emit statisticsUpdated(*fillerDecStat);
    emit eventsProceedUpdated(eventList->getEventsProceed());
    fillerDecStat->clear();
}


void EventListFiller::appendFrame(const QByteArray &buf)
{
    // Sort Frames to Data Blocks, fill Event List
#if DECODER_PROFILING
    ftimer_t timerRawDecode;
    timerRawDecode.start();
#endif
    decodeRawFrame(reinterpret_cast<const unsigned int *>(buf.constData()), buf.size()/4);
#if DECODER_PROFILING
    fillerDecStat->prof.ta[decst_n_rawdec] += timerRawDecode.read_clear();
#endif
//    qDebug() << QString("process_raw_events: %1 data block(s)").arg(ret);
}


int EventListFiller::decodeRawFrame(const unsigned int *frame, size_t flen)
{
    fillerDecStat->nw += flen;
    int decoded_data_blocks=0;
    const frame_hdr_t * const fh = reinterpret_cast<const frame_hdr_t *>(frame);

    size_t min_len = ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS;
    switch (fh->type) {
    case FRAME_TYPE_ECHO:
        fillerDecStat->nf_echo++;
        break;
    case FRAME_TYPE_ADC_DATA:
    {
        ++fillerDecStat->nf_dadc;
        const int dlen = flen - min_len;
        if (dlen > 0) {
            const unsigned int *dbuf = &frame[ML_FRAME_HEADER_WORDS];
            int ret = adcmDataBlockDecoder->adcm_decode_data_blocks (dbuf, dlen, eventList, fillerDecStat);
            if (ret > 0) {
                decoded_data_blocks += ret;
            }
        }
    }
        break;
    case FRAME_TYPE_CTRL_REQ:
        fillerDecStat->nf_creq++;
        break;
    case FRAME_TYPE_CTRL_ACK:
        ++fillerDecStat->nf_cack;
        //    printf("decode_data_blocks: frame CTRL_ACK\n");
        break;
    case FRAME_TYPE_STAT:
        ++fillerDecStat->nf_stat;
        //    printf("decode_data_blocks: frame STAT\n");
        break;
    default:
        ++fillerDecStat->nf_unkn;
        qWarning() << QString("unexpected frame type 0x%1").arg(fh->type, 0, 16);
//        adcm_frame_print (frame, flen);
        break;
    }
    return decoded_data_blocks;
}


void EventListFiller::updateProgress(bool force)
{
    if(!force && lastProgressTime->elapsed() < updateProgressTimeout)
        return;
    lastProgressTime->restart();
    emit progressChanged(frameListSize ? 1.*framesDecoded/frameListSize : 1);
}
