#ifndef ADCINST_H
#define ADCINST_H

#include <QObject>
#include <QByteArray>
#include <QSharedPointer>

#include "adcm_df.h"
#include "config/allconfig.h"

class AdcmDevice;
class QTimer;
class QUrl;
class FrameList;
class conf_mem_t;
class StopWatchTimer;
class AdcHwCounters;

struct ReadoutStats
{
    quint64 statsNumBytesReceived = 0;
    double readout_time_poll_ms = 0;
    double readout_time_framelist_ms = 0;
    double readout_time_emit_data_ms = 0;
    double throughput_Bps = 0;
};

Q_DECLARE_METATYPE(ReadoutStats)

namespace mlink {
class RegIOMLink;
}

class AdcInstance : public QObject
{
    Q_OBJECT
public:
    Q_DISABLE_COPY(AdcInstance)
    explicit AdcInstance(QObject *parent = nullptr);
    ~AdcInstance() final;
    AdcmDevice *dev() { return adcmDevice; }
    float getMinGain() const;
    float getMaxGain() const;
    float getGainStep() const;
    bool hasGainControl() const;
    bool hasTriggerControl() const;

//    bool waitForConnected(int msecs = 30000);
    double runtime () const;
    QString errorString() const { return p_errorString; }
    QStringList getDeviceIdStrList() const;

Q_SIGNALS:
    void confMemChanged(const conf_mem_t &);
//    void bytesReceived(QByteArray);
    void frameReceived(const FrameList &r);
    void connected();
    void disconnected();
    void statusTextChanged(QString);
    void readoutStateChanged(bool);
    void timeElapsed(double seconds);
    void byteCountChanged(quint64 bytes);
    void hitCountsChanged(adcm_counters_t);
    void tempChanged(QVector<double>);
    void readoutStatsChanged(ReadoutStats);

public Q_SLOTS:
    void setCalZeroConfig(const CalZeroConfig &r);
    void setAdcConfig(const AdcConfig &r);
    void startReadout();
    void stopReadout();
    void reset();
    void dataProgress(quint64 value);

protected:
    void setErrorString(const QString &str);
    void fatalError(const QString &str);

private:
    void setup();
    float adc_gain_val_to_db(int value) const;
    quint64 queueSize() const;
    bool isQueueFull() const;
    adcm_counters_t makeAdcmCounters(const AdcHwCounters &adcHwCounters) const;

    float min_gain_db = 0;
    float max_gain_db = 0;
    float gain_step_db = 3;

    AdcmDevice * adcmDevice = nullptr;
    QTimer * const reconnectTimer;
    QTimer * const pollDataTimer;
    QTimer * const pollStatusTimer;
    QTimer * const hitcountsTimer;
    quint64 prev_pos = 0;
    quint64 decoder_pos = 0;
    bool dma_running = false;
    bool desiredReadoutState = false;

    StopWatchTimer * const runTimer;
    quint64 numBytesReceived = 0;
    adcm_counters_t adcm_counters;
//    QByteArray rawdata;
    QString p_errorString;
    quint64 lastSentSeq = 0;
    quint64 lastReceivedSeq = 0;
    // statistics
    double statsTimestamp = 0;
    QVector<double> temperatures;
    AdcHwCounters * const prevAdcHwCounters;
//    AllConfig currentCfg;
    CalZeroConfig calZero;
    AdcConfig adcConfig;
    ReadoutStats readoutStats;
    QSharedPointer<mlink::RegIOMLink> adcmLink;

private Q_SLOTS:
    void disconnectDevice();
    void connectDevice();
    void pollData();
    void pollStatus();
    void readHitCounts();
    void on_connected();
    void on_disconnected();
    void on_readoutStateChanged(bool);
    void updateStatistics();
};

#endif /* ADCINST_H */
