
#include "storage.h"

#include <QFileInfo>
#include <QDir>
#include <QDebug>

QString storage::DEFAULT_STORAGE_FILENAME()
{
    return "data.dat";
}

QString storage::DEFAULT_STORAGE_FILENAME_TMP()
{
    return "data.tmp";
}

storage::storage(QObject *parent)
    : QObject(parent), fileMutex(QMutex::Recursive)
{
    enable_storage = true;
    storage_nwr = 0;
    storage_total = 0;
    mode = MODE_RECYCLE;
    STORAGE_FILENAME = DEFAULT_STORAGE_FILENAME();
    STORAGE_FILENAME_TMP = DEFAULT_STORAGE_FILENAME_TMP();
}
storage::~storage()
{
}

QString storage::mode_str(int mode) const
{
    if (mode == MODE_RECYCLE) return "recycle";
    if (mode == MODE_CONT) return "continuous";
    return "unknown";
}

void storage::print (const char *prefix = "") const
{
    qInfo() << QString("[%1] enable=%2, mode %3, file '%4'").arg(prefix).arg(enable_storage).arg(mode_str(mode)).arg(STORAGE_FILENAME);
}

int storage::setopt (bool fenable, int fmode) {
    enable_storage = fenable;
    mode = fmode;
//    print("setopt");
    return(0);
}

void storage::setFileName (const QString &fn)
{

    if (!fn.isEmpty()) { // not empty
        STORAGE_FILENAME = fn;
    } else {
        STORAGE_FILENAME = DEFAULT_STORAGE_FILENAME();
    }

    STORAGE_FILENAME_TMP = STORAGE_FILENAME + ".tmp";

//    qDebug() << QString("storage: data file '%1'").arg(STORAGE_FILENAME);
//    qDebug() << QString("storage: tmp  file '%1'").arg(STORAGE_FILENAME_TMP);
}

bool storage::prepare_files()
{
    QMutexLocker locker(&fileMutex);
    if (!enable_storage)
        return true;
//    print("prepare_files...");
    if (dataFile.isOpen())
        dataFile.close();

    if (mode == MODE_RECYCLE) {
        QFile::remove(STORAGE_FILENAME_TMP);
        storage_nwr = 0;
        dataFile.setFileName(STORAGE_FILENAME_TMP);
    }
    if (mode == MODE_CONT) {
        dataFile.setFileName(STORAGE_FILENAME);
    }
    dataFile.open( QIODevice::WriteOnly | ( MODE_RECYCLE ? QIODevice::Truncate : QIODevice::Append ) );
    if (!dataFile.isOpen()) {
        qWarning() << QString("prepare_files: Can't open file '%1' for writing: %2").arg(dataFile.fileName()).arg(dataFile.errorString());
        return false;
    }
//    print("prepare_files ok");
    return true;
}

bool storage::reset ()
{
    QMutexLocker locker(&fileMutex);
//    print("reset...");

    storage_nwr = 0;
    storage_total = 0;

    if (!prepare_files())
        return false;

//    print("reset ok");
    return true;
}

bool storage::flush ()
{
    QMutexLocker locker(&fileMutex);
//    print("flush....");

    if (dataFile.isOpen()) {
        dataFile.close();
    }

    if (!enable_storage)
        return true;

    if (mode == MODE_RECYCLE) {
        QFile::remove(STORAGE_FILENAME);
        QFileInfo fi(STORAGE_FILENAME);
        QDir dir(fi.dir().path());
        if (!dir.rename(STORAGE_FILENAME_TMP, STORAGE_FILENAME)) {
            qCritical() << QString("Failed to rename %1 to %2").arg(STORAGE_FILENAME_TMP).arg(STORAGE_FILENAME);
        } else {
//            qInfo() << QString("storage_flush: %1 MB written to '%2'")
//                          .arg(storage_nwr * 1. / 1024 / 1024).arg(STORAGE_FILENAME);
        }
        //  if ((ret=unlink(STORAGE_FILENAME_TMP))) goto err;
        //  storage_nwr = 0;

    }
    return prepare_files();
}

bool storage::write(const void *buf, quint64 count)
{
    if (!enable_storage)
        return true;

    QMutexLocker locker(&fileMutex);
    if (!dataFile.isOpen())
        return false;
//    qDebug() << QString("write: enable=%1, file '%2' %3")
//                .arg(enable_storage).arg(dataFile.name()).arg(dataFile.isOpen()?"open":"closed");
    qint64 numWritten = dataFile.write(static_cast<const char *>(buf), count);
    if ((quint64)numWritten != count) {
        qCritical() << QString("storage_write: Can't write %1 bytes to file '%2': %3")
                       .arg(count).arg(dataFile.fileName()).arg(dataFile.errorString());
        return false;
    }
    storage_nwr += count;
    storage_total += count;

//      qDebug() << QString("storage_write: %1 MB written to %2")
//            .arg((double)storage_nwr / 1024. / 1024.)
//              .arg(STORAGE_FILENAME_TMP);

    return true;
}

void storage::close ()
{
    QMutexLocker locker(&fileMutex);
//    print ("close...");
    if (dataFile.isOpen())
        dataFile.close();
}

void storage::lock()
{
    fileMutex.lock();
}


void storage::unlock()
{
    fileMutex.unlock();
}
