/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AdcmEventDebug.h"
#include <cmath>
#include "adcm_ev.h"
#include <QString>
#include <QTextStream>
#include "util/AsciiGraphics.h"

constexpr bool EV_PRINT_FULL_WAVEFORM = true;
constexpr unsigned int print_max_samples = 10;

QTextStream & operator<<(QTextStream &stream, const AdcmConvolvedData &conv)
{
    stream << "  # --- AdcmConvolvedData --- " << endl;
    stream << "    offs: " << conv.offs << ", i1: " << conv.i1 << ", i2: " << conv.i2 << endl;
    double wmin = 1e100;
    double wmax = -1e100;
    for (int s : conv.w) {
        if (s < wmin)
            wmin = s;
        if (s > wmax)
            wmax = s;
    }
    for (int i=0; i < conv.w.size(); i++) {
        bool is_peak = false;
        for (auto &peak : conv.peaks)
            if (abs(peak.pt_coarse - i) < 0.5)
                is_peak = true;
        stream << QString("      conv[%1]: %2 %3 %4 %5 %6")
                      .arg(i, 3)
                      .arg(conv.w[i], 8)
                      .arg((i == conv.i1) ? "1" : " ")
                      .arg((i == conv.i2) ? "2" : " ")
                      .arg(is_peak ? "p" : " ")
                      .arg(AsciiGraphics::ascii_bar(40, 1.0 * (conv.w[i] - wmin) / (wmax-wmin)))
            ;
        stream << endl;
    }
    for (auto i : conv.peaks) {
        stream << "    Peak time: " << i.pt_coarse << " / " << i.pt
               << ", value: " << i.pa_coarse << " / " << i.pa
               << ", width: " << i.pw_coarse << " / " << i.pw << ", chi2: " << i.pchi2;
    }
    return stream;
}

QTextStream & operator<<(QTextStream &stream, const adcm_peakbuf_t &fit)
{
    stream << "  # --- adcm_peakbuf_t --- " << endl;
//    stream << "    offs: " << fit.offs << endl;
//    double wmin = 1e100;
//    double wmax = -1e100;
//    for (int s : fit.w) {
//        if (s < wmin)
//            wmin = s;
//        if (s > wmax)
//            wmax = s;
//    }
//    for (int i=0; i < fit.w.size(); i++) {
//        stream << QString("      fit[%1]: %2 %3")
//                      .arg(i, 3)
//                      .arg(fit.w[i], 8)
//                      .arg(AsciiGraphics::ascii_bar(40, 1.0 * (fit.w[i] - wmin) / (wmax-wmin)))
//            ;
//        stream << endl;
//    }
    return stream;
}

QTextStream &operator<<(QTextStream &stream, const adcm_chbuf_t &cb)
{
    const auto & cf = cb.flags;

    if (cf.bit.has_data) {

        const unsigned int len = cb.len();
        quint64 ts = cb.ts;
        int bl = 0; // already compensated // (1<<(cb.w_bits-1));
        stream << "  # --- channel " << cb.ch << endl;
        stream << "    ch:" << cb.ch << ", ts:" << ts << ", s[" << len << "]:";
        if (! EV_PRINT_FULL_WAVEFORM) {
            if (len >= print_max_samples) {
                for (int s = 0; s < print_max_samples / 2; s++)
                    stream << " " << cb.w[s] - bl;
                stream << " ...";
                for (int s = len - print_max_samples / 2; s < len; s++)
                    stream << " " << cb.w[s] - bl;
            } else {
                for (int s = 0; s < len; s++)
                    stream << " " << cb.w[s] - bl;
            }
        } else {
            for (unsigned int s = 0; s < len; s++) {
                stream << "    [" << s << "]: " << cb.w[s] - bl << " ";
                if (cf.bit.has_thr && (s==cb.tp))
                    stream << " <";
                stream << endl;
            }
        }
        stream << "    flags=" << hex << cf << dec << endl;

        // printf("", ev->ch[c].flags);
        if (cf.bit.has_thr)
            stream << "    Thr pos " << cb.tp << ": " << cb.w[cb.tp] << endl;
        if (cf.bit.eval_bl)
            stream << "    Baseline: " << cb.b << endl;
        if (cf.bit.eval_int)
            stream << "    Integral: " << cb.i << endl;
        if (cf.bit.eval_rms)
            stream << "    DC:" << cb.dc << ", RMS:" << cb.rms << endl;
        if (cf.bit.has_pu)
            stream << "    Pile-up detected" << endl;
        if (cf.bit.has_clip)
            stream << "    Clipping detected" << endl;
        stream << endl;
        stream << cb.conv;
        stream << endl;
    }
    return stream;
}

QTextStream & operator<<(QTextStream &stream, const AdcmEvent &ev)
{
    int nc=0;
    for (auto c : ev.channel.keys()) {
        const adcm_chbuf_t &cb = ev.channel[c];
        if (cb.flags.bit.has_data) nc++;
    }

    stream << "EV #" << ev.getEventNumber() << ", burst " << ev.getBurstNumber()
           << ", ts:" << ev.getEventTimestamp() << ", nc:" << nc << endl;
    stream << "  chbuf contents: ";
    for (int c=0; c < ADCM_N_CH; c++) {
        if (!ev.channel.contains(c)) {
            stream << '.';
            continue;
        }
        const adcm_chbuf_t &cb = ev.channel[c];
        stream << (cb.flags.bit.has_data ? "D" : "e");
    }
    stream << endl;
    for (auto c : ev.channel.keys()) {
        const adcm_chbuf_t &cb = ev.channel[c];
        stream << cb;
    }
    return stream;
}
