#include <QString>
#include <QMutex>
#include <QPair>
#include <QSemaphore>
//#include <QDebug>

#include "adcm_ev.h"
#include "adcm_el.h"

//#define USE_FREE_EV_BUF

class AdcmEventListPrivate
{
public:
    AdcmEventListPrivate(QSemaphore * const freeEv, QSemaphore *const managerEv, uint _size);
    ~AdcmEventListPrivate();
    EventListIt nextPos(EventListIt &it, uint shift=1);

    QVector< QPair<uint,AdcmEvent*> > events;
    EventListIt managerPos; // Points to the last event pair
    EventListIt startWorkersPos; // First event pair sent to a worker (uses to know is there any free space)
    EventListIt endWorkersPos; // First event pair that will be send to a worker
    QSemaphore * const freeEvents;
    QSemaphore *const managerEvents;
    uint manEvBuf; // "buffer"
    uint freeEvBuf; // "buffer"
    quint64 eventsProceed;
};

AdcmEventListPrivate::AdcmEventListPrivate(QSemaphore * const freeEv, QSemaphore *const managerEv, uint _size)
    : events(_size+1),
      freeEvents(freeEv),
      managerEvents(managerEv),
      manEvBuf(0),
      freeEvBuf(0),
      eventsProceed(0)
{
    for(uint i=0;i<_size+1;++i)
        events[i] = QPair<uint,AdcmEvent*>(-1, new AdcmEvent);
    managerPos = events.end();
    startWorkersPos=events.begin();
    endWorkersPos=startWorkersPos;
}

AdcmEventListPrivate::~AdcmEventListPrivate()
{
    for (auto & event : events)
        delete event.second;
}

EventListIt AdcmEventListPrivate::nextPos(EventListIt &it, uint shift)
{
    while(shift){
        if(++it==events.end())
            it=events.begin();
        --shift;
    }
    return it;
}

//  ============================

adcm_el_t::adcm_el_t (QSemaphore * const freeEvents, QSemaphore *const managerEvents, uint size)
    : d(new AdcmEventListPrivate(freeEvents, managerEvents, size))
{
}

adcm_el_t::~adcm_el_t ()
{
    delete d;
}

void adcm_el_t::reset ()
{
    d->managerPos=d->events.end();
    d->startWorkersPos=d->events.begin();
    d->endWorkersPos=d->startWorkersPos;
    d->eventsProceed=0;
    flushManEvents();
#ifdef USE_FREE_EV_BUF
    if(d->freeEvBuf){
        d->freeEvents->release(d->freeEvBuf);
    }
#endif
    //    qDebug()<<"free="<<d->freeEvents->available()<<" man="<<d->managerEvents->available();
}

AdcmEvent *adcm_el_t::getEvent (unsigned int eventNumber)
{
    const unsigned int EC_MASK = 0x0000FFFF;
    eventNumber &= EC_MASK;
    AdcmEvent* ev=nullptr;

    if(d->managerPos == d->events.end()) {
        d->managerPos = d->events.begin();
    } else {
        if(d->managerPos->first==eventNumber)
            ev = d->managerPos->second;
        else {
#ifdef USE_FREE_EV_BUF
            if(d->freeEvBuf==0){
                d->freeEvBuf = d->freeEvents->available();
                if(d->freeEvBuf){
                    d->freeEvents->acquire(d->freeEvBuf);
                } else {
                    d->freeEvents->acquire();
                    d->freeEvBuf=1;
                }
            }
            --d->freeEvBuf;
#else
            d->freeEvents->acquire();
#endif
            d->manEvBuf++; // flush should be after setting eventNumber
            d->nextPos(d->managerPos);
            d->eventsProceed++;
        }
    }

    if(ev==nullptr){
        d->managerPos->first = eventNumber;
//        if(d->managerPlusOnePos == d->events.begin())
//            qDebug()<<"Loop in filler:"<<eventNumber<<" "<<d->managerPlusOnePos->first;

        ev = d->managerPos->second;
        ev->clear();
        if(d->manEvBuf>500)
            flushManEvents();
//        qDebug()<<"New event:"<<eventNumber;
    }

    return ev;
}

void adcm_el_t::flushManEvents()
{
    if(d->manEvBuf){
        d->managerEvents->release(d->manEvBuf);
        d->manEvBuf = 0;
    }
}

EventListIt adcm_el_t::next(EventListIt &pairIt, uint shift)
{
    return d->nextPos(pairIt, shift);
}

uint adcm_el_t::getManagerEvNum()
{
    if(d->managerPos==d->events.end())
        return 0;
    else
        return d->managerPos->first;
}

EventListIt adcm_el_t::getStartWorkerPos()
{
    return d->startWorkersPos;
}

void adcm_el_t::setStartWorkerPos(EventListIt pos)
{
    d->startWorkersPos = pos;
}

EventListIt adcm_el_t::getEndWorkerPos()
{
    return d->endWorkersPos;
}

void adcm_el_t::setEndWorkerPos(EventListIt pos)
{
    d->endWorkersPos = pos;
}

uint adcm_el_t::getBeginEvNum()
{
    return d->events.begin()->first;
}

quint64 adcm_el_t::getEventsProceed()
{
    return d->eventsProceed;
}
