
#include "adcm_mem.h"

#include <cmath>
#include <QDebug>

#include "adcm_ids.h"

static constexpr double eps = 1e-15;

bool AdcmHardwareConfig::operator==(const AdcmHardwareConfig &rhs) const
{
    return true
            && fw_ver == rhs.fw_ver
            && fw_rev == rhs.fw_rev
            && serial_id == rhs.serial_id
            && ch_dpm_ks == rhs.ch_dpm_ks
            && flen_max == rhs.flen_max
            && fcnt_max == rhs.fcnt_max
            && prop_reg.all == rhs.prop_reg.all
            ;
}

bool AdcmHardwareConfig::validateAny(int i)
{
    if (ch_dpm_ks==0 || ch_dpm_ks == 0xFFFFu) {
        qWarning() << QString("Board %1: bad ch_dpm_ks=%2, set to 1").arg(i).arg(ch_dpm_ks);
        ch_dpm_ks = 1;
    }
    if (flen_max < 8 || flen_max == 0xFFFFu) {
        qWarning() << QString("Board %1: bad flen_max=%2, set to 120").arg(i).arg(flen_max);
        flen_max = 120;
    } else {
        flen_max -= 8; // FIXME: 8 samples - data block header/trailer overhead
    }
    if (fcnt_max == 0 || fcnt_max == 0xFFFFu) {
        qWarning() << QString("Board %1: bad fcnt_max=%2, set to 1").arg(i).arg(fcnt_max);
        fcnt_max = 1;
    } else {
        fcnt_max += 1;
    }
    if (prop_reg.bit.fixed != 0x1) {
        qWarning() << QString("Board %1: bad prop_reg=%2, set to 0").arg(i).arg(prop_reg.all);
        prop_reg.all = 0;
    }
    return true;
}

bool AdcmHardwareConfig::validateSlave(int i, const AdcmHardwareConfig &master)
{
    bool result = true;
    if (master.prop_reg.bit.ext_rwind && !prop_reg.bit.ext_rwind) {
        qCritical() << QString("Board %1: inconsistent ext_rwind %2, master has %3")
                       .arg(i)
                       .arg(prop_reg.bit.ext_rwind)
                       .arg(master.prop_reg.bit.ext_rwind);
        result = false;
    }
    if (master.prop_reg.bit.trig_mult && !prop_reg.bit.trig_mult) {
        qCritical() << QString("Board %1: inconsistent trig_mult %2, master has %3")
                       .arg(i)
                       .arg(prop_reg.bit.trig_mult)
                       .arg(master.prop_reg.bit.trig_mult);
        result = false;
    }
    // TODO: allow different master/slave memory size combinations
    if (master.ch_dpm_ks > ch_dpm_ks) {
        qCritical() << QString("Board %1: insufficinet ch_dpm_ks %2, master has %3")
                       .arg(i)
                       .arg(ch_dpm_ks)
                       .arg(master.ch_dpm_ks);
        result = false;
    }
    if (master.flen_max > flen_max) {
        qCritical() << QString("Board %1: insufficinet flen_max %2, master has %3")
                       .arg(i)
                       .arg(flen_max)
                       .arg(master.flen_max);
        result = false;
    }
    if (master.fcnt_max > fcnt_max) {
        qCritical() << QString("Board %1: insufficinet fcnt_max %2, master has %3")
                       .arg(i)
                       .arg(fcnt_max)
                       .arg(master.fcnt_max);
        result = false;
    }

    return result;
}

conf_mem_board_t::conf_mem_board_t(AdcmId typeId, const AdcmHardwareConfig &adcmHardwareConfig) :
    board_type(typeId),
    nch(boardChannelCount(typeId)),
    adcstep_ns(boardAdcStepNs(typeId)),
    hw(adcmHardwareConfig)
{
    qInfo() << descriptionStr();
}

bool conf_mem_board_t::operator==(const conf_mem_board_t &rhs) const
{
    return true
            && board_type == rhs.board_type
            && nch == rhs.nch
            && abs(adcstep_ns - rhs.adcstep_ns) < eps
            && hw == rhs.hw
            ;
}

QString conf_mem_board_t::descriptionStr() const
{
    return QString("%1 %2")
            .arg(adc_board_name_str(board_type))
            .arg(hw.firmwareStr());
}

QString AdcmHardwareConfig::firmwareStr() const
{
    QString fw_tag;
    if (fw_ver != 0 && fw_ver != 0xFFFF) fw_tag = QString("ver %1.%2").arg(fw_ver>>8).arg(fw_ver&0xFF);
    if (fw_rev != 0 && fw_rev != 0xFFFF) fw_tag += QString(".%1").arg(fw_rev);
    if (serial_id != 0 && serial_id != 0xFFFF) fw_tag +=
            QString(", S/N %1")
            .arg(getSerial());
    if (!fw_tag.isEmpty()) {
        fw_tag = QString("%1").arg(fw_tag);
    }
    return fw_tag;
}

//void conf_mem_board_t::setFwProp(quint16 value) {
//    adcm_prop_reg.all = value;
//    qDebug() << fwPropStr();
//}

QString conf_mem_board_t::fwPropStr() const
{
    if (hw.prop_reg.bit.fixed == 1U) {
        return QString("ADCM firmware properties: 0x%1, %2%3%4%5%6%7").arg(hw.prop_reg.all, 4, 16)
                .arg(hw.prop_reg.bit.invert?"inv ":"")
                .arg(hw.prop_reg.bit.hyst?"hyst ":"")
                .arg(hw.prop_reg.bit.trig_mult?"tr_mult ":"")
                .arg(hw.prop_reg.bit.trig_mult?"var_sample ":"")
                .arg(hw.prop_reg.bit.ext_rwind?"ext_rwind ":"")
                .arg(hw.prop_reg.bit.framer_logic?"framer_logic ":"")
                ;
    }
    return QString();
}

QString AdcmHardwareConfig::getSerial() const
{
    return QString("%1-%2")
            .arg(serial_id >> 16, 4, 16, QChar('0'))
            .arg(serial_id & 0xFFFF, 4, 16, QChar('0'))
            .toUpper();
}

int conf_mem_board_t::getAdcHalfRange() const
{
    return 1<<(getAdcBits()-1);
}

int conf_mem_board_t::getAdcMaxValue() const
{
    return (1<<(getAdcBits()-1))-1;
}

int conf_mem_board_t::getAdcMidValue() const
{
    return 0;
}

int conf_mem_board_t::getAdcMinValue() const
{
    return -(1<<(getAdcBits()-1));
}

unsigned int conf_mem_board_t::getAdcBits() const
{
    return boardAdcBits(board_type);
}


conf_mem_t::conf_mem_t(const QMap<int, conf_mem_board_t> &boards) :
    b(boards),
    nch(getNch())
{
}

bool conf_mem_t::operator==(const conf_mem_t &rhs) const
{
    return b == rhs.b && nch == rhs.nch;
}

void conf_mem_t::clear()
{
    *this = conf_mem_t();
}

AdcmId conf_mem_t::master_board_id() const
{
    return master().board_type;
}

adcm_prop_reg_t conf_mem_t::master_firmware_prop() const
{
    return master().hw.prop_reg;
}

bool conf_mem_t::map_ch(int ch, int *bn, int *bc) const
{
    int c=0;
    for (auto i : b.keys()) {
        int i_nch = b[i].nch;
        if (ch >= c && ch < c + i_nch) {
            *bn = i;
            *bc = ch-c;
            return true;
        }
        c += i_nch;
    }
    return false;
}

int conf_mem_t::getNch() const
{
    int c=0;
    for (auto i : b) {
        c += i.nch;
    }
    return c;
}

bool conf_mem_t::hasHyst(int ch) const
{
    int bn, bc;
    if (!map_ch(ch, &bn, &bc)) return false;
    return b[bn].hw.prop_reg.bit.hyst;
}

int conf_mem_t::getAdcHalfRange(int ch) const
{
//    int bn, bc;
//    if (!map_ch(ch, &bn, &bc)) return 0;
    //    return b[bn].getAdcHalfRange();
return master().getAdcHalfRange();
}

int conf_mem_t::getAdcMaxValue(int ch) const
{
    return master().getAdcMaxValue();
}

int conf_mem_t::getAdcMidValue(int ch) const
{
    return master().getAdcMidValue();
}

int conf_mem_t::getAdcMinValue(int ch) const
{
    return master().getAdcMinValue();
}

bool conf_mem_t::hasExtMatchWin() const
{
    return master().hw.prop_reg.bit.ext_rwind;
}

bool conf_mem_t::hasMultiFragment() const
{
    return master().hw.prop_reg.bit.sample_len;
}

bool conf_mem_t::hasTrigMultiplicity() const
{
    return master().hw.prop_reg.bit.trig_mult;
}

bool conf_mem_t::hasGainControl() const
{
    return master().board_type == ADCMG_BOARD_ID;
}


bool conf_mem_t::hasTriggerControl() const
{
    switch (master().board_type) {
    case ADCM_BOARD_ID: return false;
    case ADCMG_BOARD_ID: return true;
    case ADCMLTC_BOARD_ID: return true;
    case ADCM3_BOARD_ID: return false;
    case BPM4100_BOARD_ID: return false;
    case ADC64_BOARD_ID: return true;
    case ADC64S2_BOARD_ID: return true;
    case ADC64VE_BOARD_ID: return true;
    case ADC64ECAL_BOARD_ID: return true;
    }
    return false;
}

int conf_mem_t::adc_get_prewin_min() const
{
    if (hasExtMatchWin()) return -63;
    //  if (firmware_prop.bit.ext_rwind) return -31;
    else return 0;
}
int conf_mem_t::adc_get_prewin_max() const
{
    if (hasExtMatchWin()) return 63;
    return 31;
}
int conf_mem_t::adc_get_matchwin_min() const
{
    return 0;
}
int conf_mem_t::adc_get_matchwin_max() const
{
    if (hasExtMatchWin()) return 1023;
    return 63;
}
int conf_mem_t::adc_get_latency_min() const
{
    return 1;
}
int conf_mem_t::adc_get_latency_max() const //-V524
{
    if (hasExtMatchWin()) return 63;
    return 31;
}

double conf_mem_t::getMinGainDb() const
{
    if (master().board_type == ADCMG_BOARD_ID) return -9.0;
    return 0.0;
}

double conf_mem_t::getMaxGainDb() const
{
    if (master().board_type == ADCMG_BOARD_ID) return 36.0;
    return 0.0;
}

double conf_mem_t::getGainStepDb() const
{
    if (master().board_type == ADCMG_BOARD_ID) return 3.0;
    return 1.0;
}

QStringList conf_mem_t::getDeviceIdStrList() const
{
    QStringList sl;
    for (const auto &i : b) {
        sl.push_back(i.descriptionStr());
        if (boardVirtualSlave(i.board_type)) break;
    }
    return sl;
}

QString conf_mem_t::getMasterTypeStr() const
{
    return adc_board_name_str(master().board_type);
}

QString conf_mem_t::getMasterSerial() const
{
    return master().hw.getSerial();
}

int conf_mem_t::getMasterAdcBits() const
{
    return master().getAdcBits();
}

double conf_mem_t::getMasterAdcStepNs() const
{
    return master().adcstep_ns;
}

int conf_mem_t::getMasterMaxFlen() const
{
    return master().hw.flen_max;
}

int conf_mem_t::getMasterMaxFcnt() const
{
    return master().hw.fcnt_max;
}

int conf_mem_t::getMasterChannelMemorySizeKs() const
{
    return master().hw.ch_dpm_ks;
}

QTextStream &operator<<(QTextStream &s, const adcm_prop_reg_t &b)
{
    s << "invert: " << b.bit.invert << endl;
    s << "hyst: " << b.bit.hyst << endl;
    s << "trig_mult: " << b.bit.trig_mult << endl;
    s << "sample_len: " << b.bit.sample_len << endl;
    s << "ext_rwind: " << b.bit.ext_rwind << endl;
    s << "framer_logic: " << b.bit.framer_logic << endl;
    s << "fixed: " << b.bit.fixed << endl;
    return s;
}

QTextStream &operator<<(QTextStream &s, const AdcmHardwareConfig &b)
{
    s << "ch_dpm_ks: " << b.ch_dpm_ks << endl;
    s << "flen_max: " << b.flen_max << endl;
    s << "fcnt_max: " << b.fcnt_max << endl;
    s << "prop_reg: " << b.prop_reg << endl;
    return s;
}

QTextStream &operator<<(QTextStream &s, const conf_mem_board_t &b)
{
    s << "# --- board ---" << endl;
    s << "# " << b.descriptionStr() << endl;
    s << "board_type: " << b.board_type << endl;
    s << "nch: " << b.nch << endl;
    s << "adcbits: " << b.getAdcBits() << endl;
    s << "adcstep_ns: " << b.adcstep_ns << endl;
    s << "hw: " << b.hw << endl;
    return s;
}

QTextStream &operator<<(QTextStream &s, const conf_mem_t &conf_mem)
{
    s << "# --- conf_mem ---" << endl;
    s << "# " << conf_mem.b.size() << " boards" << endl;
    for (auto b : conf_mem.b)
        s << b;
    return s;
}
