/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ConfigManager.h"

#include <QDebug>
#include <QTimer>
#include <QFileInfo>
#include <QSettings>
#include <QCoreApplication>

#include "combined_ini_io.h"
#include "allconfig.h"
#include "util/TextFormatter.h"

static const double autosave_time = 5.0;

ConfigManager::ConfigManager(QObject *parent)
    : QObject(parent)
{
    qRegisterMetaType<AdcChannelConfig>("AdcChannelConfig");
    QTimer::singleShot(0, this, &ConfigManager::autoLoad);

    auto timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &ConfigManager::autoSave);
    timer->start(autosave_time * 1000);
}

ConfigManager::~ConfigManager()
{
    autoSave();
}

void ConfigManager::setModified(bool state)
{
    modified = state;
    emit configModified(state);
}

void ConfigManager::setCurrentFileName(const QString &s)
{
    currentFileName = s;
    emit fileNameChanged(s);
}

AdcConfig ConfigManager::read_adc_setup_memory() const
{
    AdcConfig adcConfig(allConfig.adc);

    adcConfig.bits = adcConfMem.getMasterAdcBits();
    adcConfig.step = adcConfMem.getMasterAdcStepNs();

    adcConfig.resize(adcConfMem.nch);

    //Make adjusting just if any board recognized
//    if (!adcConfMem.boards().isEmpty())
//        adcConfig.adjust_adc_recwin();

    if (!adcConfMem.hasTriggerControl()) {
        adcConfig.trigbits.bit.coinc_logic = 1;
    }
    return adcConfig;
}

void ConfigManager::addCalZero(const CalZeroConfig &cal)
{
    CalZeroConfig tmp(allConfig.dec.calzero);
    tmp += cal;
    setCalZeroConfig(tmp);
}

void ConfigManager::addCalAvg(const CalAvgConfig &cal)
{
    setCalAvgConfig(allConfig.dec.calavg + cal);
}

QSharedPointer<QSettings> ConfigManager::readSettingsFile(const QString &fileName, ConfigStatus *status)
{
    if (!fileName.isEmpty()) {
        QFileInfo fileInfo(fileName);
        if (!fileInfo.exists()) {
            QString s = QString("File not found: '%1'").arg(fileName);
            qCritical() << s;
            if (status)
                *status << s;
            return nullptr;
        }
        if (!fileInfo.isReadable()) {
            QString s = QString("Permission denied: '%1'").arg(fileName);
            qCritical() << s;
            if (status)
                *status << s;
            return nullptr;
        }
    }
    QSharedPointer<QSettings> settings(new QSettings(fileName, QSettings::IniFormat));
    if (settings->status() != QSettings::NoError) {
        QString s = QString("Error reading config file '%1'").arg(fileName)
                    + ((settings->status() == QSettings::FormatError) ? " (INI format error)" : "");
        qCritical().noquote()<< s;
        if (status)
            *status << s;
        return nullptr;
    }
    return settings;
}

void ConfigManager::newConfig()
{
    qDebug() << __PRETTY_FUNCTION__;
    setAllConfig(AllConfig());
    setCurrentFileName(QString());
    setModified(false);
}

ConfigStatus ConfigManager::loadConfig(const QString &fileName, bool merge)
{
    ConfigStatus status("loadConfig");
    qDebug() << __PRETTY_FUNCTION__ << fileName;
    QSharedPointer<const QSettings> settings = readSettingsFile(fileName, &status);
    if (!settings) {
        status << "";
        return status;
    }
    if (fileName.isEmpty())
        return status;
    AllConfig tmp(allConfig);
    BaseConfigIni<AllConfig>::staticLoad(*settings, &tmp, "", merge);
    status = tmp.isValid();
    if (!status) {
        qCritical() << QString("Invalid configuration data: '%1'\n\n%2").arg(fileName).arg(status.errors.join("\n"));
        return status;
    }
    //    readGuiSettings(&settings);
    //    savedcfg = decltype (savedcfg)(new AllConfig(tmpConfig));
    //    configManager->setAllConfig(tmpConfig);
    //    autoSave();
    setCurrentFileName(fileName);
    setAllConfig(tmp);
    setModified(false);
    qInfo() << QString("Loaded config: %1").arg(settings->fileName());
    return status;
}

bool ConfigManager::saveConfig(const QString &fileName)
{
    QSettings s(fileName, QSettings::IniFormat);
    AllConfigIni::staticSave(allConfig, &s);
    s.sync();
    if (s.status() != QSettings::NoError) {
        qWarning() << "Can't write" << s.fileName();
        return false;
    }
    savedcfg = decltype (savedcfg)(new AllConfig(allConfig));
    setModified(false);
    setCurrentFileName(fileName);
    return true;
}

void ConfigManager::autoLoad()
{
    QSettings s(QSettings::IniFormat,
                QSettings::UserScope,
                QCoreApplication::organizationName(),
                QCoreApplication::applicationName() + ".autosave");
    if (s.status() != QSettings::NoError) {
        qCritical().noquote()
            << QString("Error reading config file '%1'").arg(s.fileName())
            << ((s.status() == QSettings::FormatError) ? " (format error)" : "");
        return;
    }
    AllConfig tmpConfig;
    AllConfigIni::staticLoad(s, &tmpConfig);
    if (!tmpConfig.isValid()) {
        qCritical() << QString("Invalid configuration data: '%1'").arg(s.fileName());
        return;
    }
//    setCurrentFileName(fileName);
    setAllConfig(tmpConfig);
    setModified(false);
    qInfo().noquote() << "Loaded autosave:" << s.fileName();
}

void ConfigManager::autoSave()
{
    QSettings appSettings;
//    if (settings.value("autosavegui").toBool())
//        writeGuiSettings(&settings);

    //    if (!autoSaveLoaded) return;

    if (!appSettings.value("autosave").toBool()) return;
    bool autosaveChanged = true;
    if (lastAutoSaveConfig)
        autosaveChanged = (allConfig != *lastAutoSaveConfig);
    if (!autosaveChanged) return;

    QSettings settings(QSettings::IniFormat,
                       QSettings::UserScope,
                       QCoreApplication::organizationName(),
                       QCoreApplication::applicationName() + ".autosave");
//    settings.clear(); // do not clear, keep old settings
    AllConfigIni::staticSave(allConfig, &settings);
    settings.sync();
    if (settings.status() != QSettings::NoError) {
        qWarning() << "Can't write" << settings.fileName();
    } else {
        lastAutoSaveConfig = decltype (lastAutoSaveConfig)(new AllConfig(allConfig));
        qDebug() << "Autosave to file:" << settings.fileName();
    }
}

void ConfigManager::setAdcConfMem(const conf_mem_t &r) {
    if (r == adcConfMem)
        return;
    adcConfMem = r;
//    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit adcConfMemChanged(adcConfMem);
    const AdcConfig &before = allConfig.adc;
    const AdcConfig &after = read_adc_setup_memory();
    if (before != after) {
        setAdcConfig(after);
    }
}

void ConfigManager::setAllConfig(const AllConfig &r)
{
    qDebug() << __PRETTY_FUNCTION__;
    setAdcConfig(r.adc);
    setDecoderConfig(r.dec);
    setCalAvgConfig(r.dec.calavg);
    setCalZeroConfig(r.dec.calzero);
    if (r.hist != allConfig.hist) {
        allConfig.hist = r.hist;
        setModified(true);
        qDebug() << __PRETTY_FUNCTION__;
        emit histConfigChanged(allConfig.hist);
    }
    if (r.run != allConfig.run) {
        allConfig.run = r.run;
        setModified(true);
        qDebug() << __PRETTY_FUNCTION__;
        emit runConfigChanged(allConfig.run);
    }
    if (r.stor_raw != allConfig.stor_raw) {
        allConfig.stor_raw = r.stor_raw;
        setModified(true);
        qDebug() << __PRETTY_FUNCTION__;
        emit rawStorageConfigChanged(allConfig.stor_raw);
    }
    if (allConfig.stor_dev != r.stor_dev) {
        allConfig.stor_dev = r.stor_dev;
        setModified(true);
        qDebug() << __PRETTY_FUNCTION__;
        emit evStorageConfigChanged(allConfig.stor_dev);
    }
    //        if (r == allConfig)
    //            return;
    //        allConfig = r;
}

void ConfigManager::setAdcConfig(const AdcConfig &r) {
    if (r == allConfig.adc)
        return;
    allConfig.adc = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit adcConfigChanged(allConfig.adc);
}

void ConfigManager::setDecoderConfig(const DecoderConfig &r)
{
    if (r == allConfig.dec)
        return;
    if (allConfig.dec.calzero != r.calzero) {
        allConfig.dec.calzero = r.calzero;
        emit calZeroConfigChanged(allConfig.dec.calzero);
    }
    if (allConfig.dec.calavg != r.calavg) {
        allConfig.dec.calavg = r.calavg;
        emit calAvgConfigChanged(allConfig.dec.calavg);
    }
    allConfig.dec = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit decoderConfigChanged(allConfig.dec);
}

void ConfigManager::setCalAvgConfig(const CalAvgConfig &r) {
    if (r == allConfig.dec.calavg)
        return;
    allConfig.dec.calavg = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit calAvgConfigChanged(allConfig.dec.calavg);
    emit decoderConfigChanged(allConfig.dec);
}

void ConfigManager::setCalZeroConfig(const CalZeroConfig &r) {
    if (r == allConfig.dec.calzero)
        return;
    allConfig.dec.calzero = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit calZeroConfigChanged(allConfig.dec.calzero);
    emit decoderConfigChanged(allConfig.dec);
}

void ConfigManager::setFitProfile(const FitProfile &r) {
    if (r == allConfig.dec.fitprof)
        return;
    allConfig.dec.fitprof = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    //        emit fitProfileChanged(allConfig.dec);
    emit decoderConfigChanged(allConfig.dec);
}

void ConfigManager::setRunConfig(const RunConfig &r)
{
    if (r == allConfig.run)
        return;
    allConfig.run = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit runConfigChanged(allConfig.run);
}

void ConfigManager::setRawStorageConfig(const StorageConfig &r)
{
    if (r == allConfig.stor_raw)
        return;
    allConfig.stor_raw = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit rawStorageConfigChanged(allConfig.stor_raw);
}

void ConfigManager::setEvStorageConfig(const StorageConfig &r)
{
    if (r == allConfig.stor_dev)
        return;
    allConfig.stor_dev = r;
    setModified(true);
    qDebug() << __PRETTY_FUNCTION__;
    emit evStorageConfigChanged(allConfig.stor_dev);
}
