//
//    Copyright 2013 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef DECODERCONFIG_H
#define DECODERCONFIG_H

#include <ostream>
#include <QMap>
#include <QMetaType>

#include "calzeroconfig.h"
#include "calavgconfig.h"
#include "dsplib/fir.h"
#include "adcm16.h"
#include "global_types.h"
#include "adcmlib/adcm_dec.h"
#include "dsplib/WindowFunction.h"
#include "firconfig.h"
#include "ConfigStatus.h"

typedef enum {
    FIT_PROFILE_FAST      = 1,
    FIT_PROFILE_OPTIMAL   = 2,
    FIT_PROFILE_PRECISION = 3,
    FIT_PROFILE_CUSTOM    = 4,
    FIT_PROFILE_SUPERFAST = 5,
    FIT_PROFILE_DEFAULT   = FIT_PROFILE_OPTIMAL
} FitProfileEnum_t;

Q_DECLARE_METATYPE(FitProfileEnum_t)

typedef int prof_id_t;

struct DeconvConfig
{
    ConfigStatus isValid() const;
    bool operator==(const DeconvConfig &other) const;
    bool operator!=(const DeconvConfig &other) const {return !operator==(other);}

    friend QTextStream &operator<<(QTextStream &s, const DeconvConfig &channel);

    double ppos = 0;       //!< ref pulse position, ns
    double pwid = 30;       //!< ref pulse width (2.35*sigma), ns
    double snf = -40;        //!< shape pulse threshold
    double pnf = -40;        //!< ref pulse threshold
    double tadj = 0;       //!< fine time adjust, ns, (-200..200)
    FirConfig fir;      //!< FIR digital filters
    fwinf_t fwinf = WF_RECT;         //!< Filter window function
    //  double  pwmin;       //!< fitted pulse width - minimum
    //  double  pwmax;       //!< fitted pulse width - maximum
    int expfall = 10000;       //!< exponential falloff correction for shape reconstruction
};

Q_DECLARE_METATYPE(DeconvConfig)

class DecoderChannelConfig
{
public:
    DecoderChannelConfig();
    ConfigStatus validate(const QString &title = QString()) const;
    [[deprecated]] ConfigStatus isValid() const;
    bool operator==(const DecoderChannelConfig &other) const;
    bool operator!=(const DecoderChannelConfig &other) const {return !operator==(other);}

    DeconvConfig deconv;

    int blr1;          //!< BLR window start
    int blr2;          //!< BLR window end
    int int1;          //!< integral window start (rel. to threshold position)
    int int2;          //!< integral window end
    int puthr;         //!< Pile-Up threshold, percent (>=100 == disabled)
    int pumindist;     //!< Pile-Up minimum distance, ns
    //    int search_start;
    //    int search_end;
    double baserc;      //!< baseline validation range, rel. to rms noise
};

Q_DECLARE_METATYPE(DecoderChannelConfig)

class FitProfile
{
public:
    FitProfile();
    bool operator==(const FitProfile &other) const;
    bool operator!=(const FitProfile &other) const {return !operator==(other);}

    ConfigStatus isValid() const;
    int get_dec_fititer() const;
    int get_dec_fitalgo() const;
    int get_dec_fitfixpar() const;
    int get_fit_half_np(double pulse_width_steps) const;

    FitProfileEnum_t fprof;     //!< Fitter profile select (superset of fit- iter, side, fixpar, algo)
    int fititer;                //!< Fitter iterations limit
    int fitside;                //!< Fitter number of points (2*N+1 points to fit), 0 - auto
    unsigned int fitfixpar;     //!< bit field, select which parameter to fix
    int fitalgo;                //!< Fitter algorithm select

    friend QTextStream &operator<<(QTextStream &s, const FitProfile &fitprof);
};

Q_DECLARE_METATYPE(FitProfile)

Q_DECLARE_METATYPE(decoder_mode_t)

class DecoderConfig
{
public:
    explicit DecoderConfig(int n = ADC_NCH);
    ConfigStatus isValid() const;
    int nch() const;

    bool operator==(const DecoderConfig &other) const;
    bool operator!=(const DecoderConfig &other) const {return !operator==(other);}

    friend QTextStream &operator<<(QTextStream &s, const DecoderConfig &channel);

    QMap<int, DecoderChannelConfig> channels;
    CalZeroConfig calzero;
    CalAvgConfig calavg;
    decoder_mode_t mode;                   //!< decoder mode
//    unsigned int stages;
    //	QVector<prof_id_t> prof_id; //!< channel profile ID
    FitProfile fitprof;
    double zero_thr;             //!< (mV) Minimal signal threshold for run-time baseline calibration
    double dt_cut_l; //!< ns
    double dt_cut_r; //!< ns
    int dt_cut_en;
    int initialized_nch;
    static constexpr quint32 expected_magic = 0xDEADBEEF;
    quint32 magic = expected_magic;
};

Q_DECLARE_METATYPE(DecoderConfig)

#endif // DECODERCONFIG_H
