/*
**    Copyright 2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ChannelControlWidget.h"

#include <QDebug>
#include <QSpinBox>
#include <QList>
#include <QMap>

#include "adcmlib/ch_label.h"
#include "adcm_df.h"
#include "adcmraw/adcm_mem.h"
#include "config/adcconfig.h"
#include "util/TextFormatter.h"

enum {
    table_cc_col_En=0,
    table_cc_col_Zero=1,
    table_cc_col_Inv=2,
    table_cc_col_M=3,
    table_cc_col_S=4,
    table_cc_col_thr=5,
    table_cc_col_gain=6,
    table_cc_col_rate=7,
    table_cc_col_title=8
};

static const QList<int> checkableCols = { // FIXME
    table_cc_col_En,
    table_cc_col_Inv,
    table_cc_col_Zero,
    table_cc_col_M,
    table_cc_col_S
};


ChannelControlWidget::ChannelControlWidget(QWidget *parent)
    : QTableWidget(parent),
    adcConfMem(new conf_mem_t),
    adcConfig(new AdcConfig)
{
    connect(this, &QTableWidget::cellChanged, this, &ChannelControlWidget::on_cellChanged);

    const QMap<int, QString>
        columnNames = {
            {table_cc_col_En, "En"},
            {table_cc_col_Inv, "Inv"},
            {table_cc_col_Zero, "Z/S"},
            {table_cc_col_M, ustr_ch_label_M},
            {table_cc_col_S, ustr_ch_label_S},
            {table_cc_col_thr, "Thresh."},
            {table_cc_col_gain, "Gain"},
            {table_cc_col_title, "title"},
            {table_cc_col_rate, "Freq, kHz"}
        };

    setColumnCount(columnNames.size());
    for (int i = 0; i< columnCount(); i++)
        setHorizontalHeaderItem(i, new QTableWidgetItem(columnNames[i]));
}

void ChannelControlWidget::setAdcConfMem(const conf_mem_t &r)
{
    if (r == *adcConfMem)
        return;
    *adcConfMem = r;
    init_table_st();
    update();
}

void ChannelControlWidget::setAdcConfig(const AdcConfig &r)
{
    if (r == *adcConfig)
        return;
    *adcConfig = r;
    update();
    ThrRestore();
    GainRestore();
    display_refresh_channel_controls();
}

void ChannelControlWidget::update()
{
    display_update_channel_controls_visibility();
}

void ChannelControlWidget::update_table_cc_size()
{
    //    table_cc->adjustSize();
    //  return;
    //    const QTableWidget *t = table_cc;
    //    int w = t->columnPos(t->columnCount()-1) - t->columnPos(0) + t->columnWidth(t->numCols()-1);
    //    w += t->verticalHeader()->width();
    //    w += 10;
    //  t->resize(QSize(w, t->height()));
    //    QList<int> sizes;
    //    sizes += w;
    //    sizes += table_cc->width();
    //    splitter->setSizes(sizes);
}

void ChannelControlWidget::init_table_st ()
{
    int nch = adcConfMem->nch;
    int nrow = nch;

    setRowCount(nrow);
    suppressUpdates = true;
    for (int row = 0; row < nrow; row++) {
        setVerticalHeaderItem(row, new QTableWidgetItem(QString::number(row)));
        for (int col = 0; col < columnCount(); col++) {
            auto *item = new QTableWidgetItem();
            if (checkableCols.contains(col)) {
                Qt::ItemFlags flags = item->flags();
                flags |= Qt::ItemIsUserCheckable;
                item->setFlags(flags);
                bool checkedState = desiredCheckedState(row, col);
                item->setCheckState(checkedState ? Qt::Checked : Qt::Unchecked);
            }
            setItem(row, col, item);
        }
    }

    for (int i = 0; i < nrow; i++) {
//        item(i, table_cc_col_thr)->setText("");
        //      t->setText (i, table_cc_col_gain, "");
        auto *sb = new QSpinBox(this);
        int min_gain = (int)adcConfMem->getMinGainDb();
        int max_gain = (int)adcConfMem->getMaxGainDb();
        int gain_step = (int)adcConfMem->getGainStepDb();
        sb->setMinimum(min_gain);
        sb->setMaximum(max_gain);
        sb->setSingleStep(gain_step);
        sb->setEnabled(min_gain<max_gain && gain_step > 0);
        sb->setSuffix(" dB");
        sb->setButtonSymbols(QSpinBox::PlusMinus);
        setCellWidget (i, table_cc_col_gain, sb);
        connect(sb, SIGNAL (valueChanged(int)), this, SLOT(gain_changed(int)) );

        QTableWidgetItem *tmp = item(i, table_cc_col_title);
        tmp->setFlags(tmp->flags() ^ Qt::ItemIsEditable);
        tmp->setText("");

        tmp = item(i, table_cc_col_rate);
        tmp->setFlags(tmp->flags() ^ Qt::ItemIsEditable);
        tmp->setText("");
    }

    resizeColumnsToContents();

    if (adcConfMem->hasGainControl())
        showColumn(table_cc_col_gain);
    else
        hideColumn(table_cc_col_gain);

    hideColumn(table_cc_col_title); // hide by default - it is empty
    update_table_cc_size();

    display_refresh_channel_controls();
    ThrRestore();
    GainRestore();
    suppressUpdates = false;
}

bool ChannelControlWidget::desiredCheckedState(int row, int col) const
{
    const auto &flags = adcConfig->channels[row].flags;
    switch (col) {
    case table_cc_col_En: return flags.bit.en;
    case table_cc_col_Inv: return flags.bit.inv;
    case table_cc_col_Zero: return flags.bit.zero;
    case table_cc_col_M: return flags.bit.m;
    case table_cc_col_S: return flags.bit.s;
    default: return false;
    }
}

void ChannelControlWidget::display_refresh_channel_controls ()
{
    for (int row = 0; row < rowCount(); row++) {
        for (auto & col : checkableCols) {
            QTableWidgetItem *currentItem = item (row, col);
            if (!currentItem)
                continue;
            if (! (currentItem->flags() & Qt::ItemIsUserCheckable))
                continue;
            bool checkedState = desiredCheckedState(row, col);
            currentItem->setCheckState(checkedState ? Qt::Checked : Qt::Unchecked);
        }
    }
}

void ChannelControlWidget::GainRestore ()
{
    //    int n = d->cfg.adc.size();
    //    Q_ASSERT(table_cc->numRows() == n);
    //    n = std::min(table_cc->rowCount(), n);
    for (int c = 0; c < rowCount(); c++) {
        auto *sb = static_cast<QSpinBox *>(cellWidget(c, table_cc_col_gain));
        if (sb) {
            //            bool hasConfig = d->cfg.adc.channels.contains(c);
            sb->setValue((int)adcConfig->channels[c].gain);
        }
    }
}

void ChannelControlWidget::ThrRestore ()
{
    //    int n = d->cfg.adc.size();
    //    Q_ASSERT(table_cc->numRows() == n);
    //    n = std::min(table_cc->rowCount(), n);
    for (int c = 0; c < rowCount(); c++) {
        QTableWidgetItem *currentItem = item (c, table_cc_col_thr);
        if (!currentItem)
            continue;
        currentItem->setText (QString("%1").arg(adcConfig->channels[c].thr, 0, 'f', 0));
        //            bool hasConfig = d->cfg.adc.channels.contains(c);
        //    else
        //      qDebug() <<  (2, QString("item(%1, %2) = 0!").arg(c).arg(table_cc_col_thr));
    }
}


void ChannelControlWidget::on_cellChanged( int r, int c)
{
    if (suppressUpdates)
        return;
    if (c == table_cc_col_rate)
        return;

    QTableWidgetItem *currentItem = item(r, c);
    if (!currentItem)
        return;
    const bool checked = currentItem->checkState() == Qt::Checked;

    switch (c) {
    case table_cc_col_thr:
    {
        bool ok = false;
        double value = currentItem->text().toDouble(&ok);

        if (ok) {
            adcConfig->setThr(r, value);
        } else {
            currentItem->setText(QString("%1").arg(adcConfig->channels[c].thr, 0, 'f', 0));
        }
        resizeColumnToContents(table_cc_col_thr);
        break;
    }
        //    case table_cc_col_gain:
        //    {
        //        bool ok = false;
        //        double value = item->text().toDouble(&ok);
        //        if (ok)
        //            cacheAdcConfig.setGain(r, value);
        //        else
        //            item->setText(QString::number(cacheAdcConfig.channels[r].gain));
        //        break;
        //    }
    case table_cc_col_En:
        adcConfig->channels[r].flags.bit.en = checked;
        adcConfig->channels[r].flags.bit.test = checked;
        break;
    case table_cc_col_Zero:
        adcConfig->channels[r].flags.bit.zero = checked;
        break;
    case table_cc_col_Inv:
        adcConfig->channels[r].flags.bit.inv = checked;
        break;
    case table_cc_col_M:
        adcConfig->channels[r].flags.bit.m = checked;
        break;
    case table_cc_col_S:
        adcConfig->channels[r].flags.bit.s = checked;
        break;
    default:
        break;
    }
    //    newAdcConfig.updateTriggerBits();
    emit adcConfigChanged(*adcConfig);
}

void ChannelControlWidget::display_update_channel_controls_visibility()
{
    suppressUpdates = true;
    bool testmode = adcConfig->is_testmode();
    bool hide_MS = testmode || !adcConfig->is_co_trigger(); // Test mode or not internal trigger


    for(int row=0; row < rowCount(); ++row) {
        QTableWidgetItem *currentItem = item(row, table_cc_col_M);
        Qt::ItemFlags flags = currentItem->flags();

        if (hide_MS)
            flags ^= Qt::ItemIsEditable;
        else
            flags |= Qt::ItemIsEditable;
        currentItem->setFlags(flags);

        item(row, table_cc_col_M)->setFlags(flags);
    }

    if (hide_MS) {
        hideColumn(table_cc_col_M);
        hideColumn(table_cc_col_S);
        //    hideColumn(table_cc_col_Zero);
    } else {
        showColumn(table_cc_col_M);
        showColumn(table_cc_col_S);
        //    showColumn(table_cc_col_Zero);
    }
    if (adcConfig->is_ext_trigger())
        hideColumn(table_cc_col_Zero);
    else
        showColumn(table_cc_col_Zero);
    update_table_cc_size();
    suppressUpdates = false;
}

void ChannelControlWidget::onHitCountsChanged(const adcm_counters_t &adcm_counters)
{
    unsigned int nr = std::min(rowCount(), (int)adcm_counters.n);
    for(unsigned int i=0; i<std::min(nr, adcm_counters.n); i++)  {
        double rate = (qint32)adcm_counters.rawhits[i];
        if (adcm_counters.time > 0)
            rate /= adcm_counters.time;
        else
            rate = 0;
        item(i, table_cc_col_rate)->setText(QString("%1").arg(rate/1e3, 0, 'f', 3));
    }
}

void ChannelControlWidget::gain_changed(int value)
{
    bool found = false;
    for (int r=0; r<rowCount(); r++) {
        if (cellWidget(r, table_cc_col_gain) == QObject::sender()) {
            adcConfig->setGain(r, value);
            found = true;
            break;
        }
    }
    if (found) {
        emit adcConfigChanged(*adcConfig);
    }
}

