/*
**    Copyright 2006-2019 Ilja Slepnev
**    Copyright 2014 Ivan Filippov
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AdcmDataBlockDecoder.h"

#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <cerrno>
#include <cassert>

#include <QDebug>

#include "adcm16.h"
#include "adcmlib/adcm_el.h"
#include "adcmlib/adcm_ev.h"
#include "adcmlib/AdcmChbuf.h" // access to adcm_chbuf_t structure
#include "adcm_raw_types.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmraw/adcm_mem.h"

const unsigned int ADCM_MIN_SPW = 2;
const unsigned int ADCM_MAX_SPW = 2;
const unsigned int ADCM_MIN_BIT = 8;
const unsigned int ADCM_MAX_BIT = 16;
const unsigned int ADCM_MIN_NS = ADCM_MIN_N_SAMP;
const unsigned int ADCM_MAX_NS = ADCM_MAX_N_SAMP+1; // LSB is special

//int PRINT_N_FRAME_ERR = 20;
const unsigned int PRINT_N_DATA_ERR = 20;

#ifndef DEBUG_PRINT_DECODER_DATA_ERRORS
static const int DEBUG_PRINT_DECODER_DATA_ERRORS = 0;
#endif

void AdcmDataBlockDecoder::print_decoder_data_err (adcm_decstat_t * decst, const char *fmt, ...)
{
    if (!DEBUG_PRINT_DECODER_DATA_ERRORS) return;
    char buf[1024] = { '\0' };
    size_t len;
    va_list msg;

    //    if (decst) {
    if (decst->data_errors_suppressed)
        return;
    if (decst->n_data_err() > PRINT_N_DATA_ERR) {
        fprintf (stderr, "data error: more errors, suppressed...\n");
        decst->data_errors_suppressed = 1;
        return;
    }
    //    }

    va_start (msg, fmt);
    vsnprintf (buf, sizeof (buf), fmt, msg);
    va_end (msg);

    buf[sizeof (buf) - 1] = '\0';
    len = strlen (buf);
    if ((len > 0) && (buf[len - 1] != '\n'))
        strncat (buf, "\n", sizeof (buf)-1-strlen(buf));

    fprintf (stderr, "data error: %s", buf);
}

#if NO_CONVERT_SIGNEDNESS
// Fast unpacking code
static void unpack_samples_offbin(quint32 bits, const quint32 *iptr, quint32 datablock_ns, AdcmCbufSample *optr)
{
    const quint16 sampmask = (1 << bits) - 1;
    const quint32 n = datablock_ns/2;
    const quint32 *iptrEnd = &iptr[n];
    while(iptr < iptrEnd) {
        // "offset binary"
        *optr++ = static_cast<AdcmCbufSample>((*iptr >> 16) & sampmask);
        *optr++ = static_cast<AdcmCbufSample>((*iptr)     & sampmask);
        iptr++;
    }
}

static void unpack_samples_twoscomp(quint32 bits, const quint32 *iptr, quint32 datablock_ns, AdcmCbufSample *optr)
{
    const quint16 sampmask = (1 << bits) - 1;
    const int16_t offbin = 1 << (bits-1);
    const quint32 n = datablock_ns/2;
    const quint32 *iptrEnd = &iptr[n];
    while(iptr < iptrEnd) {
        // "2's complement"
        *optr++ = (((*iptr >> 16) & sampmask) + offbin) & sampmask;
        *optr++ = (((*iptr)       & sampmask) + offbin) & sampmask;
        iptr++;
    }
}
#else
// unpack n-bit offset-binary to int16_t
void AdcmDataBlockDecoder::unpack_samples_offbin(uint32_t bits, const uint32_t iptr[], uint32_t datablock_ns, int16_t optr[])
{
    const uint16_t sampmask = (1 << bits) - 1;
    const int16_t offbin = 1 << (bits-1);
    const uint32_t n = datablock_ns/2;
    uint32_t j = 0;
    for (uint32_t i=0; i<n; i++) {
        optr[j++] = ((iptr[i] >> 16) & sampmask) - offbin;
        optr[j++] = ((iptr[i]      ) & sampmask) - offbin;
    }
}

// unpack n-bit two's complement to int16_t
void AdcmDataBlockDecoder::unpack_samples_twoscomp(uint32_t bits, const uint32_t iptr[], uint32_t datablock_ns, int16_t optr[], bool fast)
{
    const uint32_t shift = 16 - bits;
    const uint32_t shift_hi = 32 - bits;
    if (fast) {
        const uint32_t *end = iptr + datablock_ns / 2;
        for (; iptr < end; iptr++) {
            *optr++ = static_cast<int32_t>(*iptr << shift) >> shift_hi;
            *optr++ = static_cast<int16_t>(*iptr << shift) >> shift;
        }
    } else {
        const uint32_t n = datablock_ns/2;
        uint32_t j = 0;
        for (uint32_t i=0; i<n; i++) {
            optr[j++] = static_cast<int32_t>(iptr[i] << shift) >> shift_hi;
            optr[j++] = static_cast<int16_t>(iptr[i] << shift) >> shift;
        }
    }
}
#endif

//
// Decode frame data. Input is frame data block without header and trailer.
//

int AdcmDataBlockDecoder::adcm_decode_data_blocks (const unsigned int *f, size_t len, adcm_el_t * el, adcm_decstat_t * st)
{
    assert(f);
    assert(el);
    assert(st);
    if (!f || !el || !st)
        return (-EINVAL);
    if (len == 0)
        return 0;                 // empty

    int err = 0;
    unsigned int n_blk_ok = 0;
    const adcm_data_header_t *h = nullptr;
    //    const adcm_data_trailer_t *t = nullptr;

    const size_t hdrlen = (sizeof (adcm_data_header_t) + 3) / 4;
    const size_t trllen = (sizeof (adcm_data_trailer_t) + 3) / 4;

    const bool var_sample_len = true; //(conf_mem.firmware_prop.bit.sample_len != 0);

    const size_t end = len;
    size_t i = 0;
    while (i < end) {
        // begin data block processing
        st->nblk_all++;

        const size_t remain = end - i;
        if (remain < hdrlen + trllen) {
            // if (st) st->nblk_len++;
            print_decoder_data_err (st, "short data block length:%zd, minimum is %zd words\n", remain, hdrlen + trllen);
            goto badblk;
        }
        // set block header, check block type
        h = reinterpret_cast<const adcm_data_header_t *>(& f[i]);
        switch (h->id) {
        case 0:
        case 1:
        case 2:
            break;
        default:
            print_decoder_data_err (st, "unknown block type: %d\n", h->id);
            goto badblk;
        }

        // calculate block length
        size_t blen = 0;
        const size_t tmp = adcm_block_len (h);
        if (tmp < (hdrlen + trllen)) {
            print_decoder_data_err (st, "bad block len:%d\n", tmp);
            goto badblk;
        }
        if (tmp > remain) {
            print_decoder_data_err (st, "block len:%d > remaining size:%zd words, truncated\n", tmp, remain);
            blen = remain;
            //            goto badblk;
        } else {
            blen = tmp;
        }

        // length is ok. get trailer.
        //        t = reinterpret_cast<const adcm_data_trailer_t *>(& f[i + blen - trllen]);

        // additional checks by individual block type
        if (h->id == 0 || h->id == 2) {
            const adcm_data_header_0_t * const ah = reinterpret_cast<const adcm_data_header_0_t *>(h);

            if ((ah->spw < ADCM_MIN_SPW) || (ah->spw > ADCM_MAX_SPW)) { // spw hardcoded -- data is corrupt
                print_decoder_data_err (st, "dadc, bad spw value:%d, range %d-%d\n", ah->spw, ADCM_MIN_SPW, ADCM_MAX_SPW);
                goto badblk;
            }
            if ((ah->ns < ADCM_MIN_NS) || (ah->ns > ADCM_MAX_NS)) {  // ns hardcoded -- data is corrupt
                print_decoder_data_err (st, "bad ns value:%d, range %d-%d\n", ah->ns, ADCM_MIN_NS, ADCM_MAX_NS);
                goto badblk;
            }
            if ((ah->bits < ADCM_MIN_BIT) || (ah->bits > ADCM_MAX_BIT)) {  // bits hardcoded -- data is corrupt
                print_decoder_data_err (st, "bad bit value:%d, range %d-%d\n", ah->bits, ADCM_MIN_BIT, ADCM_MAX_BIT);
                goto badblk;
            }
            if (ah->ch >= ADCM_N_CH) {
                print_decoder_data_err (st, "bad ch:%d, range %d-%d\n", ah->ch, 0, ADCM_N_CH);
                goto badblk;
            }
        } else if (h->id == 1) {
            const adcm_data_header_1_t * const sh = reinterpret_cast<const adcm_data_header_1_t *>(h);

            if (sh->wpc != 2)
                goto badblk;
            if (sh->nc != 16)
                goto badblk;
        } else {                    // unknown block type
            goto badblk;
        }

        // data block correct. count it.
        switch (h->id) {
        case 0:
            st->nblk_adc_0++;
            break;
        case 1:
            st->nblk_adc_1++;
            //            st->nblk_stat++;
            break;
        case 2:
            st->nblk_adc_2++;
            break;
        }

        // adc data block
        if (h->id == 0 || h->id == 2) {
            int data_good = 1;

            const adcm_data_header_0_t * const ah = reinterpret_cast<const adcm_data_header_0_t *>(h);
            const size_t ahdrlen = (sizeof (adcm_data_header_0_t) + 3) / 4;
            const size_t atrllen = (sizeof (adcm_data_trailer_0_t) + 3) / 4;
            const adcm_data_trailer_0_t * const at = reinterpret_cast<const adcm_data_trailer_0_t *>(& f[i + blen - atrllen]);

            const unsigned int ch = ah->ch;
            //            const unsigned int spw = ah->spw;
            //            const unsigned int nwords = (ah->ns + spw - 1) / spw;
            // TODO: check data block consistency
            if (data_good)
                st->nblk_adc_ok++;

            // Store event
            AdcmEvent *ev = el->getEvent(ah->ec);
            assert(ev);
            // set event header
            ev->format.bits = raw_bits;
            ev->format.encoding = raw_signed ? SIGNED_ENCODING : OFFSET_BINARY_ENCODING;
            ev->format.step = raw_step;
            const quint64 ts = static_cast<quint64>(at->tslo) + (useLowerTimeStampBits ? 0 : (static_cast<quint64>(at->tshi) << 32));

            unsigned int tmpEventNumber = ah->ec;
            // workaround for 16-bit event counter
            {
                if (ah->ec < prev_ah_ec) {
                    ec_offset += 65536;
                }
                tmpEventNumber += ec_offset;
                prev_ah_ec = ah->ec;
            }
            ev->setEventNumber(tmpEventNumber);
            if (h->id == 0)
                ev->setBurstNumber(ah->bc);
            else
                ev->setBurstNumber(0);
            ev->setEventTimestamp(ts);            //FIXME
            if (prev_ec != tmpEventNumber) {
                if (prev_ec+1 != tmpEventNumber && prev_ec!=0) {
                    st->nblk_adc_evnum++;
                    print_decoder_data_err (st,
                                           "event number %u changed to %u (delta=%u)",
                                           prev_ec, tmpEventNumber,
                                           tmpEventNumber - prev_ec);
                }
                prev_ec = tmpEventNumber;
            }

            const quint32 datablock_ns = ah->ns & ~1;
            const bool last_fragm = !var_sample_len || ((ah->ns & 1) != 0);

            //      printf("raw: ev %d, ts %16llX, ch %d, len %d\n", ev->ec, ev->ts, ch, datablock_ns);
            adcm_chbuf_t *evc = &ev->channel[ch];
            if (true) {
                AdcmCbufWaveform &w = evc->w;
                if (evc->flags.bit.has_data == 0) {
                    evc->flags.bit.has_data = 1;
                    evc->flags.bit.finished = 0;
                    evc->ch = ch;
                    w.clear();
                    evc->ts = ts;
                    evc->w_bits = ah->bits;
                }
                if (evc->flags.bit.finished) {
                    print_decoder_data_err (st, "duplicate waveform for channel %d\n", ch);
                    st->nblk_adc_dup++;
                    //          ev->flags |= ADCM_EV_EDUPCH;
                }
                if (last_fragm)
                    evc->flags.bit.finished = 1;
                const size_t start_nsamp = evc->len();
                const size_t newSize = evc->len() + datablock_ns;
                if (newSize <= ADCM_MAX_N_SAMP) {
                    //	      evc->len += ah->ns&~1; // LSB is last-fragment-flag
                    w.reserve(newSize);
                    w.resize(newSize);
                } else {
                    print_decoder_data_err (st, "too many samples for channel %d: %d, limit %d\n", ch, newSize, ADCM_MAX_N_SAMP);
                    err = 1;
                    //          ev->flags |= ADCM_EV_EDUPCH;
                    data_good = 0;
                    break;
                }
                //        printf("ev->ec=%d, ch=%d, evc->len=%d, datablock_ns=%d, start_nsamp=%d\n", ev->ec, ch, evc->len, datablock_ns, start_nsamp);
                //   print_decoder_data_err(st, "ec:%04X, bc:%04X, ts %16llX\n", h->ec, h->bc, ts);
                //   print_decoder_data_err(st, "ADC data, bits:%d, spw:%d, ns:%d, ch:%d\n", h->bits, h->spw, h->ns, h->ch);

                // get samples - Fast unpacking code
                {
                    const quint32 *iptr = &f[i + ahdrlen]; // unsigned???
                    int16_t *optr = &w[start_nsamp];
                    if (raw_signed)
                        unpack_samples_twoscomp(ah->bits, iptr, datablock_ns, optr);
                    else
                        unpack_samples_offbin(ah->bits, iptr, datablock_ns, optr);
                    //                    for (int i=1; i<datablock_ns; i++) {
                    //                        int16_t delta = optr[i] - optr[i-1];
                    //                        if (delta > 8192)
                    //                            optr[i] -= 16384;
                    //                        if (delta < -8192)
                    //                            optr[i] += 16384;
                    //                    }
                }
            } else {                  // (!evc)
                //   print_decoder_data_err(st, "eventlist problem: can't store event #%u\n", h->ec);
            }

        } else if (h->id == 1) {    /* end if (h->id == 0) */
            /*
      adcm_data_header_1_t *ah = (adcm_data_header_1_t *)h;
      adcm_data_trailer_1_t *at;
      size_t ahdrlen = (sizeof(adcm_data_header_1_t)+3)/4;
      size_t atrllen = (sizeof(adcm_data_trailer_1_t)+3)/4;
      at = (adcm_data_trailer_1_t *) & f[i + blen - atrllen];
// stat data block

            */
        } else {
            //            logger.warn(QString("Unknown data block id %1").arg(h->id));
            // unknown data block - do nothing
        }

        // good block
        i += blen;
        // i+=trllen;
        n_blk_ok++;
        continue;

    }
    if (err)
        return (-1);
    else
        return n_blk_ok;

badblk:
    st->nblk_bad++;
    st->pblk_err = i;
    {
        char str[256], tmpstr[256];

        str[255]=0; tmpstr[255]=0;
        //  printf ("Bad data block at %04X", i);
        snprintf (str, 255, "Bad data block at %04lX", i);
        if (h) {
            size_t bblen = adcm_block_len (h);

            snprintf (tmpstr, 255, ", id:%u, len:%lu", h->id, bblen);
            strncat (str, tmpstr, 255-strlen(str));
            /*
    if (h->id == 0) {
          adcm_data_header_0_t *ah = (adcm_data_header_0_t *)h;
      snprintf (tmpstr, 255, ", spw:%d, ns:%d", ah->spw, ah->ns);
      strncat (str, tmpstr, 255);
    }
*/
            //            if ((bblen >= trllen) && (i + bblen - trllen <= len)) {
            //                unsigned int crc;

            //                t = (adcm_data_trailer_t *) & f[i + bblen - trllen];
            //                crc = t->crc;
            //                snprintf (tmpstr, 255, ", crc:%08X %s", crc, crc == ADCM_BLK_CRC_HACK ? "ok" : "bad");
            //                strncat (str, tmpstr, 255-strlen(str));
            //            }

        }
        sprintf (tmpstr, "\n");
        strncat (str, tmpstr, 255-strlen(str));
        print_decoder_data_err (st, str);
        err = 1;
        //  ev->flags |= ADCM_EV_BADBLK;
        return err;
    }
}


void AdcmDataBlockDecoder::reset()
{
    prev_ah_ec=decoderIndex;
    ec_offset=0;
    prev_ec = decoderIndex;
}

void AdcmDataBlockDecoder::setAdcConfMem(const conf_mem_t &conf_mem)
{
    if (conf_mem.boards().isEmpty())
        return;
    const conf_mem_board_t &master = conf_mem.boards()[0];
    useLowerTimeStampBits = master.board_type == ADC64_BOARD_ID && master.hw.fw_rev <= 15256;
    raw_bits = master.getAdcBits();
    raw_step = master.adcstep_ns * 1e-9;
//    raw_signed;
}
