//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <cassert>
#include <cstring> /* memset() */
#include <vector>
#include "Convolver.h"

Convolver::Convolver()
{
}

/// double * double convolution, C++
std::vector<double> Convolver::convolve (const std::vector<double> &v1, const std::vector<double> &v2)
{
    if (v1.empty() || v2.empty()) return std::vector<double>();
    std::vector<double> out(v1.size()+v2.size()-1);
    convolve_double_double (&v1[0], v1.size(), &v2[0], v2.size(), &out[0]);
    return out;
}

std::vector<double> Convolver::convolve (const std::vector<double> &v1, const std::vector<int> &v2)
{
    if (v1.empty() || v2.empty()) return std::vector<double>();
    std::vector<double> out(v1.size()+v2.size()-1);
    convolve_int_double (&v2[0], v2.size(), &v1[0], v1.size(), &out[0]);
    return out;
}

std::vector<double> Convolver::convolve (const std::vector<double> &v1, const std::vector<short> &v2)
{
    if (v1.empty() || v2.empty()) return std::vector<double>();
    std::vector<double> out(v1.size()+v2.size()-1);
    convolve_short_double (&v2[0], v2.size(), &v1[0], v1.size(), &out[0]);
    return out;
}

std::vector<int> Convolver::convolve (const std::vector<int> &v1, const std::vector<int> &v2)
{
    if (v1.empty() || v2.empty()) return std::vector<int>();
    std::vector<int> out(v1.size()+v2.size()-1);
    convolve_int_int (&v1[0], v1.size(), &v2[0], v2.size(), &out[0]);
    return out;
}

//std::vector<int> Convolver::convolve (const std::vector<short> &v1, const std::vector<short> &v2);
std::vector<int> Convolver::convolve (const std::vector<short> &v1, const std::vector<int> &v2)
{
    if (v1.empty() || v2.empty()) return std::vector<int>();
    std::vector<int> out(v1.size()+v2.size()-1);
    convolve_int_short (&v2[0], v2.size(), &v1[0], v1.size(), &out[0]);
    return out;
}



/// int * int convolution, C-style
///
/// \param out destination, required size l1+l2-1
/// \remarks Input Side Algorithm: http://www.dspguide.com/ch6/3.htm
void Convolver::convolve_int_int (const int *x1, int l1, const int *x2, int l2, int *out)
{
    assert(x1);
    assert(x2);
    assert(l1>0);
    assert(l2>0);
    assert(out);

    //    memset(out, 0, sizeof(int)*(l1+l2-1));
    for (int i = 0; i < l1 + l2 - 1; i++) out[i] = 0;

    for (int i = 0; i < l1; i++) {
        for (int j = 0; j < l2; j++)
        {
            out[i + j] += x1[i] * x2[j];
        }
    }
}

/// int * short convolution optimized for speed, C-style
void Convolver::convolve_int_short (const int *x1, int l1, const short int *x2, int l2, int *out)
{
    assert(x1);
    assert(x2);
    assert(l1>0);
    assert(l2>0);
    assert(out);
    const int outlen = l1+l2-1;
    // Important! Zero buffer. Buffer is durty on enter.
    // memset is fastest
    memset(out, 0, sizeof(int)*outlen);
    //      int *optr = &out[0];
    //      for (int i = 0; i < outlen; i++) *optr++ = 0;

    /*
// This is trivial implementation. Demo.

  {
  int i, j;
  for (i = 0; i < l1; i++)
    for (j = 0; j < l2; j++)
    {
      out[i + j] += x1[i] * x2[j];
    }
  }
*/
    /*
  memset(test, 0, sizeof(int)*outlen);
// Good code for continuous filtering with head overlap, assuming missing samples = 0
  {
  const int *coef = x1;
  int nk = l1;
  const short int *data = x2;
  int nd = l2;
  int k, j;
  assert(nk<=nd);

// Compute head overlap
//  test[0] = coef[0] * data[0];
//  test[1] = coef[1] * data[0]+coef[0]*data[1];
//  test[2] = coef[2] * data[0]+coef[1]*data[1]+coef[0]*data[2];
//  test[nk-2] = ;
  for (k = 0; k < nk-1; k++) {
  {
    int tmp=0;
    for(j=0;j<=k;j++) {
      tmp += coef[k-j] * data[j];
    }
    test[k] = tmp;
  }
  }
// Compute main part
  for (j = 0; j < nd; j++) {
    int tmp = 0;
    for (k = 0; k < nk; k++)
    {
      tmp += coef[nk-1-k] * data[j+k];
    }
    test[nk-1+j] = tmp;

  }

  }
*/

    // Welcome to real life. This algo is two times faster.
    // Avoid index operations. Use pointers.
    {
        for (int i = 0; i < l1; i++)
        {
            int *optr = &out[i];
            int x1i = x1[i];
            const short int *x2ptr = x2;
            for (int j = 0; j < l2; j++)
            {
                *optr++ += x1i * *x2ptr++;
            }
        }
    }

    /*
// Another solution. Result is short, cannot use it.
  {
    int y_ptr, c_ptr, x_ptr;
    register int temp;
    y_ptr = 0;
    for (int n = 0; n < l2; n++) {
      c_ptr = 0;
      x_ptr = n;
      temp = 0;
      for (int k = 0; k < l1; k++)
            temp += x1[l1-1-c_ptr++] * x2[x_ptr++];
      test[l1-1+y_ptr++] = temp;
    }
  }
*/
    /*
  {
    int i;
    printf("l1=%2d, l2=%2d, result %d\n", l1, l2, outlen);
    for (i = 0; i < outlen; i++)
//    if (out[i] != test[i])
      printf("[%2d]: int %8d out %8d test %8d\n", i, x2[i], out[i], test[i]);
  }
*/
    /*
// code above optimized
  {
    int *y_ptr;
    const int *c_ptr;
    const short int *x_ptr;
    register int temp;
    y_ptr = &out[0];
    for (int n = 0; n < l2; n++) {
      c_ptr = &x1[0];
      x_ptr = &x2[n];
      temp = *c_ptr++ * *x_ptr++;
      for (int k = 1; k < l1; k++)
            temp = temp + *c_ptr++ * *x_ptr++;
      *y_ptr++ = temp;
    }
  }
*/
}

/// int * double convolution, C-style
void Convolver::convolve_int_double (const int *x1, int l1, const double *x2, int l2, double *out)
{
    assert(x1);
    assert(x2);
    assert(l1>0);
    assert(l2>0);
    assert(out);
    const int outlen = l1+l2-1;

    for (int i = 0; i < outlen; i++)
        out[i] = 0;

    for (int i = 0; i < l1; i++)
        for (int j = 0; j < l2; j++) {
            out[i + j] += x1[i] * x2[j];
        }
}

/// short * double convolution, C-style
void Convolver::convolve_short_double (const short *x1, int l1, const double *x2, int l2, double *out)
{
    assert(x1);
    assert(x2);
    assert(l1>0);
    assert(l2>0);
    assert(out);
    const int outlen = l1+l2-1;

    for (int i = 0; i < outlen; i++)
        out[i] = 0;

    for (int i = 0; i < l1; i++)
        for (int j = 0; j < l2; j++) {
            out[i + j] += x1[i] * x2[j];
        }
}

/// double * double convolution, C-style
void Convolver::convolve_double_double (const double *x1, int l1, const double *x2, int l2, double *out)
{
    assert(x1);
    assert(x2);
    assert(l1>0);
    assert(l2>0);
    assert(out);
    const int outlen = l1+l2-1;

    for (int i = 0; i < outlen; i++)
        out[i] = 0;

    for (int i = 0; i < l1; i++)
        for (int j = 0; j < l2; j++) {
            out[i + j] += x1[i] * x2[j];
        }
}
