//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef DSPLIB_DSPLIB_H
#define DSPLIB_DSPLIB_H

#include <vector>

/** \brief Gaussian function FWHM constant
*
* Full Width at Half Maximum,
* equals to sqrt(2*log(2))
*/
const double GausFWHM = 2.35482004503094933e+00;

inline int BITS2MAXINT(int bits) {return (1<<(bits-1)) - 1;}
inline int BITS2MININT(int bits)  {return -(1<<(bits-1));}
inline unsigned int BITS2MAXUINT(int bits) {return (1<<bits) - 1;}
inline unsigned int BITS2MINUINT(int) {return 0;}

// Find a,b for array of (x,y) of length n
// y = ax + b
void LinearFit (const double *x, const double *y, int n, double *a, double *b, double *c);
void LinearFitInt (const int *y, int n, double *a, double *b, double *c);

/// 2nd derivative
void deriv2 (const int *x, int len, int *out);

/// Return DFT noise floor in dB
double fft_noisefloor (double rmsnoise, int rank);

/// Rectangular to halfcomplex DFT format
std::vector<double> rect2hc(const std::vector<double> &S, const std::vector<double> &C);

/// Halfcomplex to rectangular DFT format
void hc2rect(const std::vector<double> &hc, std::vector<double> *S, std::vector<double> *C);

class BruteFT
{
public:
    // Discrete Fourier Transform, real form, order N
    static void DFT (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C);
    static void DFT_tmp (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C);
    //  void DFT (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C);

    // Inverse Discrete Fourier Transform, real form, order N
    static void IDFT (std::vector<double> *x, const std::vector<double> &S, const std::vector<double> &C);
    static std::vector<double> IDFT (const std::vector<double> &S, const std::vector<double> &C);

    // Calculate power from rectangular FFT
    static void RECT2POW (const std::vector<double> &S, const std::vector<double> &C, std::vector<double> *out);
    static std::vector<double> RECT2POW (const std::vector<double> &S, const std::vector<double> &C);

    // Calculate phase from rectangular FFT
    static std::vector<double> RECT2PHA (const std::vector<double> &S, const std::vector<double> &C);
};

/// Complex Divide operation in rectangular form: (e+if)=(a+ib)/(c+id)
void Compdiv (double a, double b, double c, double d, double *e, double *f); ///< @deprecated

// Normalize to integral=1
void Normalize (std::vector<double> *x); //! deprecated

/// Scale to fit in <bits>-bit integer
std::vector<int> Make_Discrete (const std::vector<double> &x, int bits);
void Make_Discrete (const double *x, int *out, int n, int bits);  ///< @deprecated

/// Arithmetic sum of array elements (int)
int intsum (const std::vector<int> &x);

/// Max/Min value of array elements (int)
double maxdbl (const std::vector<double> &x);
double mindbl (const std::vector<double> &x);
/// DC value
double dcdbl (const std::vector<double> &x);
/// RMS value
double rmsdbl (const std::vector<double> &x, double dc);

/// Decibels to amplitude conversion
double db2val (double db);

/// Amplitude to decibels conversion
double val2db (double val);

/// An unnormalized gaussian function with mean=0 and sigma
///
/// Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed
double gauss (double x, double sigma);

///  Machine independent random number generator.
///
///  Produces uniformly-distributed floating points in [0,1]
///  Identical sequence on all machines of >= 32 bits.
///  Periodicity > 10**14
double rndm ();

/// returns a uniform deviate on the interval ]x1, x2].
double ranuni (double x1, double x2);

/// Return a number distributed following a gaussian with \a sigma
double rangaus (double sigma);

/// \brief Simulate detector pulse
///
/// \param a amplitude
/// \param t time
/// \param rt risetime
/// \param ft falltime
double detpulse (double a, double t, double rt, double ft);

// Convolution
// size of output is l1 + l2 - 1
// Integer+ShortInteger -> Integer
void convolve_i_s_i (const int *x1, const int l1, const short int *x2, const int l2, int *out);

#endif                          /* DSPLIB_DSPLIB_H */
