
#include <cmath>
#include <cassert>
#include <map>
#include <vector>

#include "fftengine.h"

RFFTEngine::~RFFTEngine()
{
#ifdef USE_FFTW
    for(std::map<size_t, fftw_plan>::const_iterator it = r2cCache.begin(); it != r2cCache.end(); ++it)
        fftw_destroy_plan(it->second);
    for(std::map<size_t, fftw_plan>::const_iterator it = c2rCache.begin(); it != c2rCache.end(); ++it)
        fftw_destroy_plan(it->second);
#endif
}

size_t RFFTEngine::getCacheSizeLimit() const
{
    return 100;
}

#ifdef USE_FFTW
fftw_plan RFFTEngine::get_r2c(size_t N)
{
    if (r2cCache.size() > getCacheSizeLimit())
        r2cCache.clear();
    if (r2cCache.find(N) == r2cCache.end()) {
        r2cCache[N] = fftw_plan_dft_r2c_1d(N, 0, 0, FFTW_ESTIMATE|FFTW_UNALIGNED);
    }
    return r2cCache[N];
}

fftw_plan RFFTEngine::get_c2r(size_t N)
{
    if (c2rCache.size() > getCacheSizeLimit())
        c2rCache.clear();
    if (c2rCache.find(N) == c2rCache.end()) {
        c2rCache[N] = fftw_plan_dft_c2r_1d(N, 0, 0, FFTW_ESTIMATE|FFTW_UNALIGNED);
    }
    return c2rCache[N];
}

/**
* Forward transform, real to complex
*
*/
std::vector<std::complex<double> > RFFTEngine::DFT(const std::vector<double> &v)
{
    const size_t N = v.size();
    std::vector<std::complex<double> > out(N);
    if (N == 0) return out;
    const fftw_plan &p = get_r2c(N);
    std::vector<double> in(v);
    fftw_execute_dft_r2c(p, &in[0], reinterpret_cast<fftw_complex*>(&out[0]));
    //    out.resize(N/2+1);
    return out;
}

/**
* Inverse transform, complex to real
*/
std::vector<double> RFFTEngine::IDFT(const std::vector<std::complex<double> > &v)
{
    const size_t N = v.size();
    std::vector<double> out(N);
    if (N == 0) return out;
    const fftw_plan &p = get_c2r(N);
    std::vector<std::complex<double> > in(v);
    fftw_execute_dft_c2r(p, reinterpret_cast<fftw_complex*>(&in[0]), &out[0]);
    return out;
}

#else
std::vector<std::complex<double> > RFFTEngine::DFT(const std::vector<double> &v)
{
    assert(false);
}
std::vector<double> RFFTEngine::IDFT(const std::vector<std::complex<double> > &v)
{
    assert(false);
}
#endif
