/*
**    Copyright 2006-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <vector>
#include <cstdio>
#include <cmath>

#include <QTimer>
#include <QDebug>
#include <QTextStream>

#include "dsplib/th1.h"
#include "histmanager.h"
#include "adcmlib/AdcmChbuf.h"
#include "adcmlib/adcm_ev.h"
#include "adcmlib/adcm_decstat.h"
#include "adcmraw/adcm_mem.h"
//#include "adcmlib/AdcmEventDebug.h"
//#include "util/TextFormatter.h"

static const int sendHistTimerPeriodMs = 1000;

HistTypeList HistList::histTypes() const
{
    QMap<hist_type_t, bool> map;
    for (auto i : keys())
        map[i.t] = true;
    return map.uniqueKeys();
}

HistManager::HistManager (QObject *parent)
    : QObject(parent),
    conf_mem(new conf_mem_t),
    sendHistTimer(new QTimer(this))
{
    QTimer::singleShot(0, this, &HistManager::recreate_histograms);
    sendHistTimer->setInterval(sendHistTimerPeriodMs);
    sendHistTimer->setSingleShot(true);
    connect(sendHistTimer, &QTimer::timeout, this, &HistManager::sendHistograms);
    sendHistTimer->start();
}

void HistManager::setAdcConfMem(const conf_mem_t &value)
{
    if (value == *conf_mem)
        return;
    if (conf_mem->nch != value.nch) {
        *conf_mem = value;
        recreate_histograms();
    }
}

int HistManager::nch() const
{
    return conf_mem->nch;
}

void HistManager::addHist(hist_type_t type, const th1 &h)
{
    hl[hist_index_t(type)] = h;
}

void HistManager::addHist(hist_type_t type, int channel, const th1 &h)
{
    hl[hist_index_t(type, channel)] = h;
}

void HistManager::recreate_histograms()
{
    hl.clear();
    QString s;

    s = hist_title(HIST_TYPE_dT);
    int dT_x_scaler = 4;
    addHist(HIST_TYPE_dT, th1(s, "ns", dT_x_scaler*8000, -100, 100, th1::FIT));

    s = hist_title(HIST_TYPE_dTmax);
    addHist(HIST_TYPE_dTmax, th1(s, "ns", 8000, -1000, 1000, th1::FIT));

    s = hist_title(HIST_TYPE_Tmr);
    addHist(HIST_TYPE_Tmr, th1(s, 200, 0, 1));

    s = hist_title(HIST_TYPE_Tsr);
    addHist(HIST_TYPE_Tsr, th1(s, 200, 0, 1));

    s = hist_title(HIST_TYPE_Ts);
    addHist(HIST_TYPE_Ts, th1(s, "ns", 1000, -300, 500));

    s = hist_title(HIST_TYPE_dT_Ev);
    addHist(HIST_TYPE_dT_Ev, th1(s, "ns", 200, 0, 100000));

    s = hist_title(HIST_TYPE_decodetime);
    addHist(HIST_TYPE_decodetime, th1(s, "us/us", 2000, 0, 200));

    s = hist_title(HIST_TYPE_slideev);
    addHist(HIST_TYPE_slideev, th1(s, ADCM_N_CH, 0, ADCM_N_CH));

    s = hist_title(HIST_TYPE_pulsev);
    addHist(HIST_TYPE_pulsev, th1(s, ADCM_N_CH, 0, ADCM_N_CH));

    s = hist_title(HIST_TYPE_niter);
    addHist(HIST_TYPE_niter, th1(s, 500, 0, 500));

    s = hist_title(HIST_TYPE_wm);
    addHist(HIST_TYPE_wm, th1(s, "ns", 1000, 0, 100, th1::FIT));

    s = hist_title(HIST_TYPE_ws);
    addHist(HIST_TYPE_ws, th1(s, "ns", 1000, 0, 100, th1::FIT));

    for (int i = 0; i < nch(); i++) {
        const int minraw = -32768;
        const int maxraw = 32768;
        s = hist_title(HIST_TYPE_evdist) + QString("[%1]").arg(i);
        addHist(HIST_TYPE_evdist, i, th1(s, "ns", 2000, 0, 20, 0));

        s = hist_title(HIST_TYPE_T) + QString("[%1]").arg(i);
        addHist(HIST_TYPE_T, i, th1(s, "ns", 8000, -1000, 1000, th1::FIT));

        s = hist_title(HIST_TYPE_Int) + QString("[%1]").arg(i);
        addHist(HIST_TYPE_Int, i, th1(s, maxraw * 4, minraw, maxraw, th1::FIT|th1::FIT_RESOL));

        s = hist_title(HIST_TYPE_bl) + QString("[%1]").arg(i);
        addHist(HIST_TYPE_bl, i, th1(s, maxraw * 4, minraw, maxraw, th1::FIT));
    }
    emit histListModified(hl);
}

void HistManager::sendHistograms()
{
    emit histogramsChanged(hl);
    clearAllHists();
    sendHistTimer->start();
}

void HistManager::fill_status (const adcm_decstat_t &st)
{
    double t = st.prof.tl[decst_n_all];
    //  if (t<0) printf("t=%f\n", t);
    //  stime_t tnow = stime();
    //  double t = tnow-tprev;
    //  tprev=tnow;
    if (t > 0) {
        auto &h = hl[hist_index_t(HIST_TYPE_decodetime)];
        t = t * 1e6;
        h.filln(t, (int)(t));
    }
}

void HistManager::fill (const AdcmEvent &ev)
{
    const double step = ev.format.step;
    const double step_ns = step * 1e9;

    std::vector<int> cm;
    std::vector<int> cs;
    std::vector<double> tm;
    std::vector<double> ts;
    std::vector<double> tmr;
    std::vector<double> tsr;
    std::vector<double> wm;
    std::vector<double> ws;
    unsigned int mNumbers = 0;
    unsigned int sNumbers = 0;
    unsigned int pulsesNumbers = 0;
    unsigned int npuls_fitok = 0;

    for (auto c : ev.channel.keys()) {
        const adcm_chbuf_t &cb = ev.channel[c];
        const adcm_channel_flags_t &trig = cb.trig;

        if (cb.flags.bit.has_data) {
            pulsesNumbers++;
            //	  printf("ev %d, ch %d, pulsesNumbers=%d\n", adcm_get_ev_evnum(ev), c, pulsesNumbers);
            hl[hist_index_t(HIST_TYPE_evdist, c)].fill(cb.ts * step * 1e3); // convert to ms
        }
        if (cb.flags.bit.eval_bl)
            hl[hist_index_t(HIST_TYPE_bl, c)].fill (cb.b);

        if (cb.flags.bit.eval_int)
            hl[hist_index_t(HIST_TYPE_Int, c)].fill (cb.i - cb.b);

//        if (0 == (cb.flags & (ADCM_CH_EVAL_TFAST | ADCM_CH_EVAL_TFINE)))
//            continue;
         if (cb.flags.bit.dt_cut)
             continue;

//         qDebug() << cb;
//         npuls_fitok++; // FIXME
         for (auto peak : cb.conv.peaks) {
             hl[hist_index_t(HIST_TYPE_niter)].fill(peak.pniter);

             if (trig.bit.m) {
                 double step = 1.;
                 double rt = peak.pt;
                 rt = rt - trunc(rt/step)*step;
                 if (rt<0) rt+=step;
                 cm.push_back(c);
                 tmr.push_back(rt);
                 tm.push_back(peak.pt);
                 wm.push_back(peak.pw);
                 mNumbers++;
             }
             if (trig.bit.s) {
                 double step = 1.;
                 double rt = peak.pt;
                 rt = rt - trunc(peak.pt/step)*step;
                 if (rt<0) rt+=step;
                 cs.push_back(c);
                 tsr.push_back(rt);
                 ts.push_back(peak.pt);
                 ws.push_back(peak.pw);
                 sNumbers++;
             }
         }
    }
    auto &ihdt = hl[hist_index_t(HIST_TYPE_dT)];
    auto &ihdtmax = hl[hist_index_t(HIST_TYPE_dTmax)];
    auto &ihtmr = hl[hist_index_t(HIST_TYPE_Tmr)];
    auto &ihtsr = hl[hist_index_t(HIST_TYPE_Tsr)];
    auto &ihts = hl[hist_index_t(HIST_TYPE_Ts)];
    auto &ihwm = hl[hist_index_t(HIST_TYPE_wm)];
    auto &ihws = hl[hist_index_t(HIST_TYPE_ws)];

    // fill from cache
    {
        //      if (htm) {
        //        for (unsigned int im = 0; im < mNumbers; im++)
        //          htm->fill (tm[im] * step_ns);
        //      }
        for (const auto & i : tmr)
            ihtmr.fill(i);

        for (const auto & i : wm)
            ihwm.fill(i * step_ns);

        for (const auto & i : tsr)
            ihtsr.fill(i);

        for (const auto & i : ts)
            ihts.fill (i * step_ns);

        for (const auto & i : ws)
            ihws.fill (i * step_ns);


        for (unsigned int im = 0; im < tm.size(); im++) {
            for (unsigned int is = 0; is < ts.size(); is++) {
//                if (cm[im] == cs[is]) // same channel
//                    continue;
                double time_m = tm[im]*step_ns;
                double time_s = ts[is]*step_ns;
                double dt = time_s-time_m;
//                qDebug() << QString("cm[%1]=%2, cs[%3]=%4, time_m=%5, time_s=%6, dt=%7")
//                                .arg(im)
//                                .arg(cm[im]).arg(is).arg(cs[is]).arg(time_m).arg(time_s)
//                                .arg(dt);
                ihdt.fill (dt);
                ihdtmax.fill (dt);
//                hl[hist_index_t(HIST_TYPE_T, cs[is])].fill(dt);
            }
        }

    } // end fill from cache

    hl[hist_index_t(HIST_TYPE_pulsev)].fill(npuls_fitok);
    hl[hist_index_t(HIST_TYPE_slideev)].fill(pulsesNumbers);
}

int HistManager::fillEvDt (const AdcmEvent &ev, quint64 prevEvTs)
{
    double dT = ev.getEventTimestamp() - prevEvTs; // in steps
    dT *= ev.format.step * 1e9; // in ns
//    qDebug() << ev.getEventNumber() << dT;
//    if(dT > 10000)
//        qWarning() << "too big dT=" << dT;

    hl[hist_index_t(HIST_TYPE_dT_Ev)].fill(dT);
    return (0);
}

void HistManager::clearAllHists()
{
    for (auto & h : hl)
        h.clear();
}

void HistManager::clearHist(const hist_index_t &index)
{
    hl[index].clear();
}

QTextStream &operator<<(QTextStream &s, const HistList &r)
{
    QStringList sl;
    for (auto & i : r)
        sl << i.getname();
    s << sl.join(", ");
    return s;
}
