//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <vector>
#include "qxw/QxwPlot.h"
#include "qxw/QxwZoomer.h"
#include <qwt_plot_panner.h>
#include <qwt_plot_grid.h>
#include <qwt_plot_layout.h>
#include <qwt_plot_marker.h>
#include <qwt_legend.h>
//#include <qwt_legend_item.h>

#if QT_VERSION >= 0x040000
QxwPlot::QxwPlot(QWidget* parent) :
    QwtPlot(parent)
  #else
QxwPlot::QxwPlot(QWidget* parent, const char* name) :
    QwtPlot(parent, name)
  #endif
  ,zoomer(nullptr)
  ,panner(nullptr)
{
    setAutoDelete(false);
    // grid
    grid = QwtPlotGridPtr(new QwtPlotGrid);
    grid->setMajorPen(QPen(Qt::darkGray, 0, Qt::DotLine));
    //    grid->setMinorPen(QPen(Qt::gray, 0 , Qt::DotLine));
    grid->attach(this);
    //  insertLegend(new QwtLegend(), QwtPlot::RightLegend);
    //    legend = new QwtLegend();
    setupZoomer();
    enableLegend(true);
    connect(this, SIGNAL(legendChecked(QwtPlotItem*,bool)), SLOT(onLegendChecked(QwtPlotItem*,bool)));
    setAutoReplot(true);
}

QxwPlot::~QxwPlot()
{
    //    delete grid;
    //    delete zoomer;
    //    delete panner;
}

/// LeftButton for the zooming
/// MidButton for the panning
/// RightButton: zoom out by 1
/// Ctrl+RighButton: zoom out to full size
void QxwPlot::setupZoomer()
{
    plotLayout()->setAlignCanvasToScales(true);
    //    replot();

    delete zoomer;
    zoomer = new QxwZoomer(canvas());
    zoomer->setMousePattern(QwtEventPattern::MouseSelect2,
                            Qt::RightButton, Qt::ControlModifier);
    zoomer->setMousePattern(QwtEventPattern::MouseSelect3,
                            Qt::RightButton);

    delete panner;
    panner = new QwtPlotPanner(canvas());
    panner->setAxisEnabled(QwtPlot::yRight, false);
    panner->setMouseButton(Qt::MidButton);
}

QRectF QxwPlot::getDataBoundRect() const
{
    QRectF r;
    //  double xmin=MAXDOUBLE, xmax=-MAXDOUBLE, ymin=MAXDOUBLE, ymax=-MAXDOUBLE;
    for (const auto & curve : curves) {
        if (!curve || !curve->plot()) continue;
        const QRectF &curRect = curve->boundingRect();
        if (!curRect.isValid()) continue;
        r = r.united(curRect);
    }
    return r;
}

void QxwPlot::clear()
{
    clearCurves();
    clearXMarkers();
    clearRefMarkers();
    //    QwtPlot::detachItems();
}

QwtPlotCurvePtr QxwPlot::getCurve(int id)
{
    Q_ASSERT(curves.contains(id));
    return curves[id];
}

void QxwPlot::clearCurves()
{
    for (auto & i : curves)
        if (i->plot())
            i->detach();
}

void QxwPlot::clearXMarkers()
{
    for (auto & marker : xMarkers) {
        if (marker && marker->plot()) {
            marker->detach();
        }
    }
}

void QxwPlot::clearRefMarkers()
{
    for (auto & marker : yMarkers) {
        if (marker && marker->plot()) {
            marker->detach();
        }
    }
}

bool QxwPlot::curveAttached(int id) const
{
    if (!curves.contains(id)) return false;
    return curves[id]->plot();
}

bool QxwPlot::curveExists(int id) const
{
    return curves.contains(id);
}

bool QxwPlot::xmarkerExists(int id) const
{
    return xMarkers.contains(id);
}

bool QxwPlot::markrefExists(int id) const
{
    return yMarkers.contains(id);
}

void QxwPlot::setCurveBrush(int id, const QBrush &brush)
{
    if (curveExists(id)) {
        getCurve(id)->setBrush(brush);
    }
}

void QxwPlot::setCurvePen(int id, const QPen &pen)
{
    if (curveExists(id)) {
        getCurve(id)->setPen(pen);
    }
}

void QxwPlot::setCurveStyle(int id, QwtPlotCurve::CurveStyle style)
{
    if (curveExists(id)) {
        getCurve(id)->setStyle(style);
    }
}

//void QxwPlot::setCurveSymbol(int id, const QwtSymbol &symbol)
//{
//    if (curveExists(id)) {
//        getCurve(id)->setSymbol(symbol);
//    }
//}

void QxwPlot::setBrush(const QBrush &brush)
{
    for (auto & curve : curves)
        curve->setBrush(brush);
}

void QxwPlot::setPen(const QPen &pen)
{
    for (auto & curve : curves)
        curve->setPen(pen);
}

void QxwPlot::setStyle(QwtPlotCurve::CurveStyle style)
{
    for (auto & curve : curves)
        curve->setStyle(style);
}

//void QxwPlot::setSymbol(const QwtSymbol &symbol)
//{
//    Q_FOREACH(QwtPlotCurvePtr curve, curves)
//        curve->setSymbol(symbol);
//}

void QxwPlot::setYMarkers(const std::vector<double> &v)
{

    for (auto & marker : yMarkers) {
        marker->detach();
    }
    yMarkers.clear();
    int id = 0;
    for(std::vector<double>::const_iterator it = v.begin(); it != v.end(); ++it) {
        double value = *it;
        QwtPlotMarkerPtr mark(new QwtPlotMarker());
        mark->setLabelAlignment(Qt::AlignRight | Qt::AlignTop);
        mark->setLineStyle(QwtPlotMarker::HLine);
        mark->setLinePen(QPen(QColor(Qt::black), 1, Qt::DotLine));
        mark->setLabel(QString("%1").arg(value));
        mark->setXValue(value);
        mark->attach(this);
        yMarkers[id] = mark;
        id++;
    }
}

void QxwPlot::setXMarkers(const std::vector<double> &v)
{

    for (auto & marker : xMarkers) {
        marker->detach();
    }
    xMarkers.clear();
    int id = 0;
    for(std::vector<double>::const_iterator it=v.begin(); it != v.end(); ++it) {
        double value = *it;
        QwtPlotMarkerPtr mark(new QwtPlotMarker());
        mark->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        mark->setLineStyle(QwtPlotMarker::VLine);
        mark->setLinePen(QPen(QColor(Qt::black), 1, Qt::DotLine));
        mark->setLabel(QString("%1").arg(value));
        mark->setXValue(value);
        mark->attach(this);
        xMarkers[id] = mark;
        id++;
    }
}

void QxwPlot::onLegendChecked(QwtPlotItem *plotItem, bool on)
{
    plotItem->setVisible(on);
    //    qDebug() << "onLegendChecked(" << plotItem->title().text() <<  ", " << on << ")";
    //    qDebug() << "autoReplot " << (autoReplot()?"on":"off");
    if (autoReplot()) replot();
}

void QxwPlot::updateLegendItems()
{
    /*
    if (!legend()) return;
    foreach(int id, curves.keys()) {
        QxwCurvePtr curve = curves.value(id);
        if (!curve || !curve->plot()) continue;
        qDebug() << "curve " << curve->title().text() << (curve->isVisible()?" shown":" hidden");
        if (curve->isVisible()) {
            curve->setItemAttribute(QwtPlotItem::Legend, true);
            curve->updateLegend(legend());
            QwtLegendItem *legendItem = static_cast<QwtLegendItem*>(curve->legendItem());
            qDebug() << "legendItem: " << legendItem->text().text();
            legendItem->setChecked(true);

        }
    }
*/
}

void QxwPlot::enableLegend(bool enable)
{
    if (enable) {
        if (! legend()) {
            QwtLegend *legend = new QwtLegend(this);
//            legend->setItemMode(QwtLegend::CheckableItem); // FIXME: port to qwt6
            insertLegend(legend, QwtPlot::RightLegend);
        }
    } else {
        if (legend())
            delete legend();
    }
}

void QxwPlot::disableLegend(bool disable)
{
    enableLegend(!disable);
}
