
#define _USE_MATH_DEFINES
#include <cmath>

#include "WaveformFilter.h"
#include "dsplib/fir_types.h"
#include "dsplib/fir.h"
#include "dsplib/dsplib.h"
#include "dsplib/WindowFunction.h"
#include "dsplib/Convolver.h"
#include "dsplib/FftSpectrum.h"

//! Vector of absolute values of v elements
std::vector<double> getAbsDb(const std::vector<std::complex<double> > &v)
{
    std::vector<double> r(v.size());
    for(size_t i=0; i<v.size(); i++)
        r[i] = val2db(std::abs(v[i]));
    return r;
}

WaveformFilterFIR::WaveformFilterFIR()
{
    size_t  numtaps = 128;
    //  BandSpec bs;
    //  bs.push_back(BandSpecItem(0, 0.2, 1, 1));
    //  bs.push_back(BandSpecItem(0.3, 0.5, 0, 1));
    //  std::vector<double> fircoef = RemezApproximationFIR(numtaps, bs, BANDPASS);

    firkernel.resize(numtaps);
    //  FILE *F = fopen("fir.txt", "w");
    const double cutoff = 0.45;
    for(size_t i=0; i<numtaps; i++) {
        double wf = windowingFunction(WF_NUTTALL, i, numtaps);
        double f = 1;
        double x = M_PI*((double)i/numtaps - 0.5) * (numtaps*cutoff);
        if (x!=0) f = sin(x)/(x);
        firkernel[i] = 0.5*wf*f;
        if (fabs(firkernel[i]) < 1e-14) firkernel[i] = 0;
        //    fprintf(F, "%e %e\n", x/(2*M_PI), fircoef[i]);
    }
    //   fclose(F);
}

WaveformData WaveformFilterFIR::filter(const WaveformData &wd) const
{
    const size_t len = wd.p.size();
    if (len<2) return WaveformData();
    WaveformData result;
    const double x0 = wd.p[0].x;
    const double xstep = wd.p[1].x - wd.p[0].x;
    std::vector<double> adcSamples(len);
    for(uint i=0;i<len;i++) {
        adcSamples[i] = wd.p[i].y;
    }

    std::vector<double> convout = Convolver::convolve(adcSamples, firkernel);
    const size_t convlen = convout.size();
    result.c = wd.c;
    /*
  const size_t shift = (convlen-len)/2;
  result.p.resize(len);
  for(uint i=0;i<len;i++) {
        result.p[i].y = convout[i+shift];
        result.p[i].x = x0 + xstep*i;
  }
*/
    /*
  result.p.resize(convlen);
  for(uint i=0;i<convlen;i++) {
        result.p[i].y = convout[i];
        result.p[i].x = x0 + xstep*i;
  }
*/
    result.p.resize(convlen/2);
    for(uint i=0;i<convlen/2;i++) {
        result.p[i].y = convout[i*2];
        result.p[i].x = x0 + xstep*i*2;
    }
    return result;
}

WaveformDataAll WaveformFilterBase::filterAll(const WaveformDataAll &wda) const
{
    WaveformDataAll result;
    for (WaveformDataAll::const_iterator it = wda.begin(); it != wda.end(); ++it) {
        int id = it->first;
        const WaveformData &wd = it->second;
        result[id] = filter(wd);
    }
    return result;
}


WaveformData WaveformFilterFFT::filter(const WaveformData &wdinp) const
{
    WaveformData wd(wdinp);
    //  if (wd.c == 0) wd = filterFIR(wd);
    WaveformData result;
    result.c = wd.c;
    const size_t len = wd.p.size();
    if (len < 2) return result; // empty;
    const double x_step = wd.p[1].x - wd.p[0].x;
    //  printf("c=%d, x_step=%e\n", wd.c, x_step);

    // FFT start
    const std::vector<double> &adcSamples = wd.getPointsY();
    /*
// subtract DC
    double dc = 0;
        for(uint i=0;i<adcSamples.size();i++) {
          dc += adcSamples[i];
        }
        for(uint i=0;i<adcSamples.size();i++) {
          adcSamples[i] -= dc/adcSamples.size();
        }
*/
    const std::vector<std::complex<double> > &cfAdcSamples = fft.DFT(adcSamples);
    const size_t dftSize = cfAdcSamples.size() / 2 + 1;

    // compute magnitude (linear)
    const double freqStep = 1.0 / x_step / len;
    result.p.clear();
    // discard DC component [0]
    for(size_t i=1; i<dftSize; i++) {
        const point_t p(i * freqStep, std::abs(cfAdcSamples[i]) * 1./len);
        result.p.push_back(p);
    }
    /*
        TimeDomainData samples(adcSamples, 0, 1.0/x_step); // freq, MHz
//	samples.print();
        if (fft.size() > 100) fft.clear(); // purge
        const FFTSpectrum spectrum = fft[samples.size()].getFFTSpectrum(samples);
        const size_t spectrumSize = spectrum.size();
        result.p.resize(spectrumSize);
    for(uint i=0;i<spectrumSize;i++) {
          result.p[i].x = spectrum[i].freq;
          result.p[i].y = spectrum[i].power;
//  	  if (i<10)
//  		printf("c %d, fft[%d] = %e : %e\n", result.c, i, result.p[i].x, result.p[i].y);
    }
    */
    // FFT finish
    return result;
}
