#include "FileWriter.h"
#include "metric-sender/MetricSender.h"

#include <QDir>
#include <QDateTime>
#include <QTimer>


namespace {
const int WRITER_TIMER_PERIOD = 100; //ms
const int METRICS_SEND_PERIOD_MS = 1000;
}
FileWriter::FileWriter(RingBuffer<char> * const writerRB, QObject *parent) :
    QObject(parent),
    ringBuf(writerRB),
    stats(new TcpServerStats()),
    writeTimer(new QTimer(this)),
    statusTimer(new QTimer(this))
{
    qRegisterMetaType<FileWriterStats>();
    qInfo() << QString("Predefined storage is:").append(predefStorage);
    connect(writeTimer, &QTimer::timeout, this, &FileWriter::writeData);
    writeTimer->setSingleShot(false);
    writeTimer->start(WRITER_TIMER_PERIOD);
    defineFileName("", 0, 0);
    connect(statusTimer, &QTimer::timeout, this, &FileWriter::sendStatus);
    statusTimer->start(500);

    QTimer * sendMetricsTimer = new QTimer(this);
    connect(sendMetricsTimer, &QTimer::timeout, this, &FileWriter::sendMetrics);
    sendMetricsTimer->start(METRICS_SEND_PERIOD_MS);
}

FileWriter::~FileWriter()
{
    writeTimer->stop();
    statusTimer->stop();
    if(ringBuf->getDataSize()){
        qWarning() << "Some data haven't been written to file";
    }
    if(dataFile.isOpen())
        closeFile();
}

void FileWriter::openFileToWrite(QString runIndex, int runNumber, int writerIndex)
{
    QString &fileName = totalStats.fileName;
    defineFileName(runIndex, runNumber, writerIndex);

    QDir dir(QFileInfo(fileName).dir());
    if(!dir.exists())
        dir.mkpath(".");

    if(dataFile.isOpen())
        dataFile.close();
    dataFile.setFileName(fileName);
    if(dataFile.open(QFile::WriteOnly)) {
//        totalStats.missedFileSize = 0;
//        totalStats.writtenFileSize = 0;
        bool ok = dataFile.setPermissions(QFile::WriteOwner);
        if(ok) {
            qInfo() << QFileInfo(dataFile).absoluteFilePath().prepend("Start writting to file:");
            emit fileOpened(runIndex, runNumber);
        } else {
            const QString errStr = QFileInfo(dataFile).absoluteFilePath().prepend("Can't close permissions to file:");
            qCritical() << errStr;
            emit error(-1, errStr);
        }
    } else {
        QString errStr = dataFile.errorString().prepend("Error on file open:");
        emit error(-1, errStr);
        fileName="";
        qWarning() << errStr;
    }
    sendStatus();
}

void FileWriter::writeData()
{
    writeLatencyElapsedTimer.restart();

    int size = ringBuf->getContinuosDataSize(0);
    bool doubleWrite = false;
    const int MaxWritePart = ringBuf->getAllSize()/4;
    while(size) {
        doubleWrite = ringBuf->getDataSize()!=size;
        size = qMin(MaxWritePart, size);
        qint64 written = dataFile.write(ringBuf->getReadPointer(), size);
        if(written == -1){
            const QString errStr = QString("Error on file write:").append(dataFile.errorString());
            qWarning() << errStr;
            emit error(-1, errStr);
            return;
        }
//        qInfo() << "written:" << written << "size:" << size;
        ringBuf->trashData(written);
        int s = ringBuf->getContinuosTrashSize();
        ringBuf->freeData(s);
        s = ringBuf->getContinuosTrashSize();
        if(s)
            ringBuf->freeData(s);
        stats->bytesSent += written;
        totalStats.writtenFileSize += written;

        if(written != size){
            if(written == -1){
                totalStats.missedFileSize += size;
                const QString &errStr = QString("Error on file write: %1")
                        .arg(dataFile.errorString());
                qWarning() << errStr;
                emit error(-1, errStr);
            } else {
                totalStats.missedFileSize += size-written;
                dataFile.error();
                const QString &errStr = QString("Incomplete file write: wrote %1 bytes of %2")
                        .arg(written).arg(size).arg(dataFile.errorString() );
                qWarning() << errStr;
                emit error(-1, errStr);
            }
        }
        size = ringBuf->getContinuosDataSize(0);
    }

    if(doubleWrite)
        writeTimer->start(0);
    else
        writeTimer->start(WRITER_TIMER_PERIOD);

    qint64 elapsed = writeLatencyElapsedTimer.nsecsElapsed();
    stats->addLatencyValue(elapsed);
}

void FileWriter::closeFile()
{
//    ringBuf->clear();
    if(ringBuf->getFreeSize() != ringBuf->getAllSize()) {
        writeData();
        writeData();
        if(ringBuf->getFreeSize() != ringBuf->getAllSize())
            qInfo() << "closeFile data size:" << ringBuf->getDataSize();
    }
    if(!dataFile.flush())
        qWarning() << dataFile.errorString().prepend("Error on file flush:");
    dataFile.close();
    bool ok = dataFile.setPermissions(QFile::ReadOwner | QFile::ReadUser |
                                      QFile::ReadOther | QFile::ReadGroup);
    if(ok) {
        qInfo() << "File closed";
    } else {
        const QString errStr = "Failed set read permissions on file";
        qCritical() << errStr;
        emit error(-1, errStr);
    }
}

void FileWriter::clear()
{
    totalStats.clear();
    totalStats.predefStorage = predefStorage;
    ringBuf->clear();
    sendMetrics();
    stats->clear();
}

void FileWriter::storageChanged(QString dir)
{
    predefStorage = dir;
    qInfo() << QString("Change storage folder to:").append(predefStorage);
    totalStats.predefStorage = predefStorage;
    sendStatus();
}

void FileWriter::sendStatus()
{
    totalStats.timeStamp.start();
    emit statusUpdated(totalStats);
}

void FileWriter::sendMetrics()
{
    MetricSender::send("out.file.bytes", stats->bytesSent);
//    MetricSender::send("out.file.buffer.used", 1.0*ringBuf->getDataSize() / ringBuf->getAllSize());
    MetricSender::send("out.file.latency.mean", stats->getMeanLatency());
}

void FileWriter::defineFileName(QString runIndex, int runNumber, int writerIndex)
{
    QString &fileName = totalStats.fileName;
    if(runNumber) {
        QString prefix = "/data/mpd-evb/";
        if(!predefStorage.isEmpty())
            prefix = predefStorage + QDir::separator();

        if(writerIndex)
            fileName = QString("%1mpd_run_%2_%3_%4.data")
                    .arg(prefix)
                    .arg(runIndex)
                    .arg(runNumber, 3, 10, QChar('0'))
                    .arg(writerIndex);
        else
            fileName = QString("%1mpd_run_%2_%3.data")
                    .arg(prefix)
                    .arg(runIndex)
                    .arg(runNumber, 3, 10, QChar('0'));
    } else {
        fileName = QString("mpd_%1.data").arg(QDateTime::currentDateTime().toString("yyyyMMdd_hhmmss"));
        if(!predefStorage.isEmpty())
            fileName.prepend(predefStorage+QDir::separator());
    }
    QDir dir(QFileInfo(fileName).dir());
    if(!dir.exists())
        dir.mkpath(".");
    if(!dir.exists()){
        QString errStr = QString("Output folder '%1' doesn't exist").arg(dir.path());
        qCritical() << errStr;
        emit error(-1, errStr);
    } else {
        qInfo() << QString("Output folder set to '%1'").arg(dir.path());
    }
}

