﻿//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "ClientManager.h"
#include "Combiner.h"
#include "TcpClient.h"
#include "StatisticDialog.h"
#include "add-pnp-dialog/AddPnpDialog.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/RunControlCommands.h"
#include "base/SyslogMessageSender.h"
#include "metric-sender/MetricSender.h"
#include "base/threadQuit.h"
#include "configuration-manager/ConfigurationManager.h"
#include "pnp-server/PNPServer.h"
#include "base/AboutBox.h"

#include <QActionGroup>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostAddress>
#include <QInputDialog>
#include <QMessageBox>
#include <QMetaType>
#include <QThread>
#include <QTimer>
#include <QSemaphore>
#include <QSettings>
#include <time.h>
#include <QTime>
#include <iostream>

namespace {
enum COLUMNS{ COL_ID, COL_READOUT_CB, COL_CONNECTION_STATE, COL_PROG_TYPE, COL_PROG_INDEX, COL_PROG_HOST, COL_STATUS, COL_EV_NUM, COL_TRASHED_EVENTS, COL_ALL_EVENTS, COL_FLUSHED_SIZE,
              COL_TIME_DISTR, COL_TCP_BUFF, COL_EV_BUFF, COL_TOTAL};

QString getTcpName(int tcpId) {
    return tcpId ? QString("Tcp #%1").arg(tcpId-1) : QString("Mon");
}
}

using namespace std;

Q_DECLARE_METATYPE(Qt::GlobalColor)

MainWindow::MainWindow(StartupOptions &start_opt, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    config(start_opt.program_type, start_opt.program_index),
    default_program_index(start_opt.program_index),
    clientLastId(0),
    combiner(new Combiner()),
    combinerThread(new QThread(this)),
    remoteControl(new RemoteControlServer(0)),
    remoteControlThread(new QThread(this)),
    pnpServer(new PNPServer()),
    pnpServerThread(new QThread(this)),
    run(false),
    loc_runNumber(0),
    lastEvNum(0),
    labelFileErrSize(new QLabel(this)),
    clientDistrTime(new QTime()),
    curStateIndex(0),
    statisticDialog(new StatisticDialog(this)),
    backPreasure(false)
{
    if(!start_opt.configuration_name.isEmpty()){
        config.configuration_name = start_opt.configuration_name;
    }

    config.read_config();

    QThread::currentThread()->setObjectName("GUI thread");
    combinerThread->setObjectName("combinerThread");
    remoteControlThread->setObjectName("remoteControlThread");
    pnpServerThread->setObjectName("pnpServerThread");
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<AnalyzerStat>();
    ui->setupUi(this);
    ui->mainToolBar->hide();
    QActionGroup *dataOutActionGroup = new QActionGroup(this);
    dataOutActionGroup->addAction(ui->actionOutToFile);
    dataOutActionGroup->addAction(ui->actionOutToTcp);
    dataOutActionGroup->addAction(ui->actionCombinerIdle);
    connect(dataOutActionGroup, &QActionGroup::triggered, this, &MainWindow::data_out_changed);

    connect(ui->actionQuit, &QAction::triggered, qApp, &QCoreApplication::quit);

    connect(this, &MainWindow::closeProgram, pnpServer, &PNPServer::closeProgram);
    connect(this, &MainWindow::changeProgramDescription, pnpServer, &PNPServer::changeProgramDescription);
    connect(this, &MainWindow::changeProgramDescription, remoteControl, &RemoteControlServer::changeProgramDescription);
    connect(pnpServer, &PNPServer::programDiscovered, this, &MainWindow::gotProgramDescription);
    connect(pnpServer, &PNPServer::programClosed, this, &MainWindow::gotClosedProgram);
    connect(this, &MainWindow::destroyAll, pnpServer, &PNPServer::deleteLater);
    connect(pnpServer, &PNPServer::destroyed, pnpServerThread, &QThread::quit);
    QStringList searchTargs;
    searchTargs << DaqConfig::getTypeName(DaqConfigEventBuilder);
    searchTargs << DaqConfig::getTypeName(DaqConfigGenerator);
    searchTargs << DaqConfig::getTypeName(DaqConfigMStream);
    searchTargs << DaqConfig::getTypeName(DaqConfigVmeDaq);
    searchTargs << DaqConfig::getTypeName(DaqConfigLand);
    pnpServer->startSearch(searchTargs, false);
    pnpServer->moveToThread(pnpServerThread);
    pnpServerThread->start();

    connect(ui->widgetRunControl, &RunControl::stateChanged, this, &MainWindow::runControlStateUpdated);

    remoteControl->moveToThread(remoteControlThread);
    remoteControlThread->start();
    connect(remoteControl, &RemoteControlServer::destroyed, remoteControlThread, &QThread::quit);
    connect(remoteControl, &RemoteControlServer::startRequest, this, &MainWindow::remoteStartRequest);
    connect(remoteControl, &RemoteControlServer::stopRequest, this, &MainWindow::remoteStopRequest);
    connect(this, &MainWindow::destroyAll, remoteControl, &RemoteControlServer::deleteLater);
    connect(this, &MainWindow::remoteCompletedStart, remoteControl, &RemoteControlServer::completedStart);
    connect(this, &MainWindow::remoteCompletedStop, remoteControl, &RemoteControlServer::completedStop);
    connect(this, &MainWindow::sendRemoteState, remoteControl, &RemoteControlServer::sendState);
    connect(this, &MainWindow::sendRunError, remoteControl,
        QOverload<QString, int, QString>::of(&RemoteControlServer::onRunError));

    connect(this, &MainWindow::sendRunWarning, remoteControl, &RemoteControlServer::onRunWarning);
    connect(this, &MainWindow::sendBackPreasure, remoteControl, &RemoteControlServer::onBackPressure);

    connect(remoteControl, &RemoteControlServer::connectionStatusUpdated, this, &MainWindow::onRemoteStatusUpdated);
    connect(this, &MainWindow::allowRemoteControl, remoteControl, &RemoteControlServer::startListen);
    connect(this, &MainWindow::config_transfer_to_rc, remoteControl, &RemoteControlServer::config_transfer);

    configuration_manager = new ConfigurationManager(this, start_opt.program_type, config.default_root);
    connect(configuration_manager, &ConfigurationManager::switch_to_config, this, &MainWindow::switch_to_config);

    ui->actionAllow_Remote_Control->blockSignals(true);
    ui->actionAllow_Remote_Control->setChecked(config.isRemoteEnabled);
    ui->actionAllow_Remote_Control->blockSignals(false);

    initCombiner();
    init();
}

void MainWindow::init()
{
    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.evb." + shortHostName + "." + config.program_index);

    setupTable();
    setup_clients();
    initProgramDescription();
    update_combiner_to_config();
    emit allowRemoteControl(config.isRemoteEnabled);
    statusBar()->addPermanentWidget(labelFileErrSize);
    statusBar()->setToolTip("reg:\tNumber of regular events\n"
                            "skip:\tNumber of skipped events from clients\n"
                            "eob:\tNumber of 'end of burst' events\n"
                            "clients:\tClient's ids from last combined event\n"
                            "nextEv:\tEvent number of next expected event\n"
                            "monitor:\tNumber of sent events to monitor");
    labelFileErrSize->hide();
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.program_index, config.configuration_name);
}

void MainWindow::switch_to_config(QString program_index, QString configuration_name)
{
    configuration_manager->hide();
    config.program_index = program_index;
    config.configuration_name = configuration_name;
    for(int row =0; row < ui->tableWidget->rowCount(); ++row){
        int clientId = getClientIdByRow(row);
        QTableWidget *table = ui->tableWidget;
        emit destroyClient(clientId);
        clients.remove(clientId);
        config.clients.remove(clientId);
        QWidget *chButton = table->cellWidget(row, COL_READOUT_CB);
        QWidget *stateLabels = table->cellWidget(row, COL_CONNECTION_STATE);
        Q_ASSERT(chButton);
        Q_ASSERT(stateLabels);
        delete chButton;
        delete stateLabels;
        recivedStates.remove(clientId);
        lastConnactionState.remove(clientId);
    }
    ui->tableWidget->setRowCount(0);
    updateAnalizerList();

    config.read_config();

    emit allowRemoteControl(!config.isRemoteEnabled);
    init();
}

MainWindow::~MainWindow()
{
    QElapsedTimer timer;
    timer.start();

    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    config.write_config();

    RunControl * const rc = ui->widgetRunControl;
    bool ret;
    if(rc->getState()==RC_STARTING || rc->getState()==RC_STARTED){
//        qInfo()<<"[MainWindow] "<<timer.elapsed()<<"waitState(RC_STARTED";
        ret = rc->waitState(RC_STARTED, 2000);
//        Q_ASSERT(ret);
        if(rc->getState() == RC_STARTED) {
            rc->setState(RC_STOPPING);
            qInfo()<<"[MainWindow] "<<timer.elapsed()<<"setState(RC_STOPPING";
        }
    }
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"waitState(RC_STOPPED";
    ret = rc->waitState(RC_STOPPED, 2000);
    if(!ret){
        qWarning() << "Failed to stop system";
    }
    emit closeProgram(progDesc);

//    Q_ASSERT(ret);

//    config.writeSettings(); // save all setting right after changes

    emit destroyClient(0); // all
    emit destroyCombiner(0);
    emit destroyAll();
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"destroyClient(0);";
    //todo: wait all clients are destroyed


    timer.start();
    //Give time to proceed destroyed signal from RemoteControl
    if(!threadQuiting(remoteControlThread, 5000, &timer)){
        qWarning() << "RemoteControl's thread isn't stopped";
    }

    if(!threadQuiting(pnpServerThread, 5000, &timer)){
        qWarning() << "PnpServer's thread isn't stopped";
    }

    //Give time to proceed destroyed signal from Combiner
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"threadQuiting(combinerThread";
    if(!threadQuiting(combinerThread, 5000, &timer)){
        qWarning() << "Combiner's thread isn't stopped";
//        qInfo()<<"[MainWindow] "<<timer.elapsed()<<"Combiner's thread isn't stopped";
    }

    delete ui;
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::on_pushButtonAdd_clicked()
{
    QVector<ProgramDescription> freeProgs = getFreeProgs();
    if(!freeProgs.size())
        return;

    AddPnpDialog dialog(freeProgs, true, this);
    connect(pnpServer, &PNPServer::programClosed, &dialog, &AddPnpDialog::gotClosedProgram);
    connect(pnpServer, &PNPServer::programDiscovered, &dialog, &AddPnpDialog::gotProgramDescription);

    dialog.set_function_filter(std::bind(&MainWindow::isFreeProg, this, std::placeholders::_1));

    if(dialog.exec() == QDialog::Accepted) {
        for(auto pd : dialog.getProgDescr()) {
            ClientConfig cc;
            cc.descr = pd;
            cc.id = addClient(cc);
            if(cc.id != -1){
                config.clients[cc.id] = cc;
                updateAnalizerList();
                config.write_config();
            }
        }
    }
}

void MainWindow::on_pushButtonRemove_clicked()
{
    QString progType, progIndex;
    QTableWidget *table = ui->tableWidget;

    int row = getSelectedRow();
    int clientId=-1;
    if(row >= 0) {
        clientId = getClientIdByRow(row);
        progType = table->item(row, COL_PROG_TYPE)->text();
        progIndex = table->item(row, COL_PROG_INDEX)->text();
    }

    if(clientId<0 || !clients.contains(clientId)) {
        qWarning() << "Don't know what client was selected; removing canceled";
        return;
    }

    QString message = QString("Are you shure to remove client %1 #%2?").arg(progType,progIndex);
    int ret = QMessageBox::warning(this, tr("Confirm removing"), message, QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
    if(ret == QMessageBox::No) {
        qDebug() << "Client's removing canceled";
        return;
    }

    emit destroyClient(clientId);
    clients.remove(clientId);
    config.clients.remove(clientId);
    config.write_config();
    QWidget *chButton = table->cellWidget(row, COL_READOUT_CB);
    QWidget *stateLabels = table->cellWidget(row, COL_CONNECTION_STATE);
    table->removeRow(row);
    Q_ASSERT(chButton);
    Q_ASSERT(stateLabels);
    delete chButton;
    delete stateLabels;

    qInfo() << QString("Client deleted %1 %2").arg(progType, progIndex);
    recivedStates.remove(clientId);
    updateAnalizerList();
}

void MainWindow::setupTable()
{
    QTableWidget *table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);
#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    table->setHorizontalHeaderItem(COL_ID, new QTableWidgetItem("id"));
    table->setHorizontalHeaderItem(COL_READOUT_CB, new QTableWidgetItem("readout"));
    table->setHorizontalHeaderItem(COL_CONNECTION_STATE, new QTableWidgetItem("Online"));
    table->setHorizontalHeaderItem(COL_PROG_TYPE, new QTableWidgetItem("type"));
    table->setHorizontalHeaderItem(COL_PROG_INDEX, new QTableWidgetItem("index"));
    table->setHorizontalHeaderItem(COL_PROG_HOST, new QTableWidgetItem("host"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("state"));
    table->setHorizontalHeaderItem(COL_EV_NUM, new QTableWidgetItem("evNum"));
    table->setHorizontalHeaderItem(COL_TRASHED_EVENTS, new QTableWidgetItem("trashed"));
    table->setHorizontalHeaderItem(COL_ALL_EVENTS, new QTableWidgetItem("all"));
    table->setHorizontalHeaderItem(COL_FLUSHED_SIZE, new QTableWidgetItem("flushedSize"));

    table->setHorizontalHeaderItem(COL_TIME_DISTR, new QTableWidgetItem("Time distr"));
    table->setHorizontalHeaderItem(COL_TCP_BUFF, new QTableWidgetItem("Raw data buff"));
    table->setHorizontalHeaderItem(COL_EV_BUFF, new QTableWidgetItem("Ev. meta-info"));

    on_actionEnable_client_time_distribution_toggled(config.timeDistrEn);
    on_actionAdvanced_counters_toggled(config.advancedCounters);
}

void MainWindow::on_tableWidget_itemSelectionChanged()
{
    updatePushButtonState();
    updateTablePalette();
}

void MainWindow::updateClientStatus(int index, bool isConnected, QString status)
{
    if(!config.clients.contains(index))
        return;
    QTableWidget *table = ui->tableWidget;
    bool connectionChanged = !lastConnactionState.contains(index)
            || lastConnactionState[index] != isConnected;
    lastConnactionState[index] = isConnected;
    if(connectionChanged)
        updateTablePalette();
    const int row = getTableRowByClientIndex(index);
    if(row != -1) {
        table->item(row, COL_STATUS)->setText(status);
        QLabel *label = qobject_cast<QLabel*>(table->cellWidget(row, COL_CONNECTION_STATE));
        QPixmap p;
        if (isConnected){
            label->setToolTip("Device sends discover packets");
            p = QApplication::style()->standardPixmap(QStyle::SP_DialogApplyButton);
        } else {
            if(run && config.clients.contains(index) && config.clients[index].enabled){
                QString errStr = QString("%1 (id:%2) disconnected")
                        .arg(getClientNameByIndex(index)).arg(index);
                qWarning() << errStr;
                // todo handle mstream restart period
                if(lastCombinerStats.hasPendingEvent && lastCombinerStats.regularEvents)
                    deviceError(index, errStr);
            }
            label->setToolTip("Device doesn't send discover packets");
            p = QApplication::style()->standardPixmap(QStyle::SP_DialogCloseButton);
        }
        if(connectionChanged)
            updateTableTextColor(row, index);

        label->setPixmap(p);
//        qDebug()<<QString("[%1:%2] status: %3").arg(table->item(row, COL_ADDRESS)->text(), table->item(row, COL_PORT)->text(), status);
        updateTableSize();
        return;
    }
//        qDebug()<<QString("Got status (%1) for unknown client index=%2").arg(status).arg(index);
}

void MainWindow::updateTcpStatistic(int index, const TcpStatistic &statistic)
{
    QTableWidget *table = ui->tableWidget;
    const int row = getTableRowByClientIndex(index);
    if(row != -1) {
        // Row found
        QString text = QString("%1/%2/%3")
                .arg( statistic.freePart, 4, 'f', 1)
                .arg( statistic.dataPart, 4, 'f', 1)
                .arg(statistic.trashPart, 4, 'f', 1);
        table->item(row, COL_TCP_BUFF)->setText(text);
        updateTableSize();
        return;
    }
    // Row not found
}

void MainWindow::setup_clients()
{
    Q_ASSERT(clients.isEmpty());
    qDebug() << QString("Start evNumber=%1; evNumberSpet=%2; evNumberIndex=%3")
                .arg(config.startEvNumber)
                .arg(config.evNumberStep)
                .arg(config.getEvNumIndex());
    setOutTcpServerNumber(config.combinerOutTcpNum);

    emit setEvNumberStep(config.startEvNumber, config.evNumberStep);

    for(auto cc : config.clients.values()) {
        int id = addClient(cc);
        Q_ASSERT(cc.id==id);
    }
}


int MainWindow::addClient(const ClientConfig &cc)
{
    // Client id=0 for Combiner; id of TcpClients > 0
    int clientId = cc.id;
    Q_ASSERT(ui->widgetRunControl->getState()==RC_STOPPED);

    if(clientId == 0) {
        qWarning()<<"Trying to add client with Combiner's clientId="<<clientId;
        clientId = -1;
    } else {
        for(ClientManager *manager : clients.values()) {
            if(manager->getProgDescr().sameProgram(cc.descr)){
                qWarning() << QString("Client %1 %2 duplicated with clientId=%3")
                              .arg(cc.descr.type).arg(cc.descr.index).arg(manager->getIndex());
                return -1;
            }
            if(clientId != -1 && manager->getIndex() == clientId){
                qWarning()<<"Trying to add client with existing clientId="<<clientId;
                clientId = -1;
            }
        }
    }

    if(clientId==-1){
        clientId = clientLastId+1;
    }
    clientLastId = qMax(clientLastId, clientId);

    int preferedBufSize = cc.inputBufSize ? cc.inputBufSize : DATA_RING_SIZE;
    if(cc.inputBufSize/EVENT_RING_SIZE < EVENT_SIZE_MIN)
        preferedBufSize = EVENT_SIZE_MIN*EVENT_RING_SIZE;
    ClientManager *clientManager = new ClientManager(clientId, config.getEvNumIndex(), cc.enabled, preferedBufSize, this);
    connect(clientManager, &ClientManager::sendStatus, this, &MainWindow::updateClientStatus);
    connect(clientManager, &ClientManager::sendStatistic, this, &MainWindow::updateTcpStatistic);
    connect(clientManager, &ClientManager::sendStatistic, statisticDialog, &StatisticDialog::updateTcpStatistic);
    connect(this, &MainWindow::reset, clientManager, &ClientManager::resetClient);
    connect(this, &MainWindow::starting, clientManager, &ClientManager::startClient);
    connect(this, &MainWindow::stoping, clientManager, &ClientManager::stopClient);
    connect(clientManager, &ClientManager::started, this, &MainWindow::clientStarted);
    connect(clientManager, &ClientManager::stopped, this,  &MainWindow::clientStopped);
    connect(clientManager, &ClientManager::resetted, this, &MainWindow::deviceResetted);
    connect(this, &MainWindow::destroyClient, clientManager, &ClientManager::destroyClient);
    connect(this, &MainWindow::updateClientReadout, clientManager, &ClientManager::updateClientReadout);
    connect(clientManager, &ClientManager::statUpdated, statisticDialog, &StatisticDialog::analyzerStatUpdated);

    connect(clientManager, &ClientManager::inputBufferSizeUpdated, this, &MainWindow::inputBufferSizeUpdated);
    clients[clientId] = clientManager;

    QTableWidget *table = ui->tableWidget;
    int row = table->rowCount();
    table->setRowCount(row+1);
    QCheckBox *cb = new QCheckBox(table);
    cb->setChecked(cc.enabled);
    connect(cb, &QCheckBox::toggled, this, &MainWindow::setClientReadOut);
    QLabel *label = new QLabel(table);
    label->setAlignment(Qt::AlignCenter);

    table->setItem(row, COL_ID, new QTableWidgetItem(QString::number(clientId)));
    table->setCellWidget(row, COL_READOUT_CB, cb);
    table->setCellWidget(row, COL_CONNECTION_STATE, label);
    table->setItem(row, COL_PROG_TYPE, new QTableWidgetItem(cc.descr.type));
    table->setItem(row, COL_PROG_INDEX, new QTableWidgetItem(cc.descr.index));
    table->setItem(row, COL_PROG_HOST, new QTableWidgetItem(cc.descr.hostName.split('.').first()));
    table->setItem(row, COL_STATUS, new QTableWidgetItem("creating..."));
    table->setItem(row, COL_EV_NUM, new QTableWidgetItem());
    table->setItem(row, COL_TRASHED_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_ALL_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_FLUSHED_SIZE, new QTableWidgetItem());
    table->setItem(row, COL_TIME_DISTR, new QTableWidgetItem());
    table->setItem(row, COL_TCP_BUFF, new QTableWidgetItem());
    table->setItem(row, COL_EV_BUFF, new QTableWidgetItem());

    table->item(row, COL_TIME_DISTR)->setToolTip("Aver. time diff for earliest/\nAver. time full, ms");
    table->item(row, COL_TCP_BUFF)->setToolTip("free/data/trash");
    table->item(row, COL_EV_BUFF)->setToolTip("free/prepared/flushed");

    clientManager->setProgDescription(cc.descr);
    recivedStates[clientId] = false;
    if(cc.enabled)
        emit updateClientReadout(clientId, true);
    updateTableSize();
    qDebug() << QString("Added client %1 %2").arg(cc.descr.type).arg(cc.descr.index);
    return clientId;
}

void MainWindow::updatePushButtonState()
{
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    bool isStopped = ui->widgetRunControl->getState()==RC_STOPPED;
    if (ranges.empty() || !isStopped) {
        ui->pushButtonRemove->setEnabled(false);
    } else {
        Q_ASSERT(ranges.size()==1);
        ui->pushButtonRemove->setEnabled(true);
    }

    ui->pushButtonAdd->setEnabled(isStopped && !getFreeProgs().isEmpty());
    ui->pushButtonReset->setEnabled(isStopped);
    ui->tableWidget->setEnabled(isStopped);
    ui->actionSet_EvB_index->setEnabled(isStopped);
    ui->actionOutToFile->setEnabled(isStopped);
    ui->actionOutToTcp->setEnabled(isStopped);
    ui->actionCombinerIdle->setEnabled(isStopped);
    ui->actionChange_storage_dir->setEnabled(isStopped);
    ui->actionReset_to_defaults->setEnabled(isStopped);
}

void MainWindow::updateTableSize()
{
    int height=ui->tableWidget->horizontalHeader()->height();
    for(int i=0;i<ui->tableWidget->rowCount();++i){
        height += ui->tableWidget->rowHeight(i);
    }
    height += 5;
    ui->tableWidget->setMinimumHeight(height);

    int width=ui->tableWidget->verticalHeader()->width();
    for(int i=0;i<ui->tableWidget->columnCount();++i){
        width += ui->tableWidget->columnWidth(i);
    }
    width += 5;
    ui->tableWidget->setMinimumWidth(width);

//    QRect geo = ui->tableWidget->geometry();
//    geo.setHeight(height);
//    geo.setWidth(width);
//    ui->tableWidget->setGeometry(geo);
}

void MainWindow::updateAnalizerList()
{
    QVector<Analizer *> analizers;
    QMap<int, QString> clientNames;
    for(ClientManager *manager : clients) {
        if(manager->isReadoutEnabled()) {
            analizers.push_back(manager->getAnalizer());
            clientNames[manager->getIndex()] = manager->getName();
        }
    }
    enabledClientNum = analizers.size();
    emit analizerListUpdated(analizers, clientNames);
    updateRemoteStatus();
    updateProgramDescription();
}

void MainWindow::updateRemoteStatus()
{
    QString stateStr;
    int clientsEnabled=0;
    for(const ClientManager *mgr : clients.values()){
        if(mgr->isReadoutEnabled())
            ++clientsEnabled;
    }
    if(clientsEnabled)
        stateStr = QString::number(clientsEnabled).append(" clients");
    else
        stateStr = "no clients";
    emit sendRemoteState(stateStr, true);
}

void MainWindow::initProgramDescription()
{
    progDesc.type = config.program_type;
    progDesc.index = config.program_index;
    progDesc.name = progDesc.getName();
}

void MainWindow::updateProgramDescription()
{
    bool hasRcInterface = progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl);
    ProgramInterface rcInterface;
    if(hasRcInterface)
        rcInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    progDesc.clear();
    if(hasRcInterface)
        progDesc.interfaces.push_back(rcInterface);

    for (int id : lastCombinerStats.tcpStatus.keys()) {
        const OutputTcpStatus &tcpStatus = lastCombinerStats.tcpStatus[id];
        if(!tcpStatus.listening)
            continue;
        ProgramInterface tcpInterface;
        tcpInterface.setType(id ? ProgramInterfaceDataFlow : ProgramInterfaceMonitorData);
        tcpInterface.enabled = true;
        tcpInterface.id = id ? id-1 : 0;
        tcpInterface.port = tcpStatus.port;
        tcpInterface.isFree = tcpStatus.isFree;
        if(!tcpStatus.isFree) {
            tcpInterface.peers << ProgramInterfacePeer(tcpStatus.peer_host, tcpStatus.peer_port);
        }
        progDesc.interfaces.push_back(tcpInterface);
    }

    for (ClientManager *manager : clients) {
        if(manager->isReadoutEnabled()) {
            progDesc.options[QString("Client-%1").arg(manager->getIndex())] =
                    manager->getProgDescr().getName();
        }
    }

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif
    progDesc.options["run"] = QVariant(run).toString();
    progDesc.options["runIndex"] =loc_runIndex;
    progDesc.options["runNumber"] = QString::number(loc_runNumber);
    progDesc.options["output"] = config.getOutputStr();

    emit changeProgramDescription(progDesc);
}

void MainWindow::updateTcpPortLabels()
{
    for (OutputTcpStatus st : lastCombinerStats.tcpStatus.values()) {
        QLabel *l = getTcpLabel(st.index);
        Q_ASSERT(l);
        QString text = QString("%1: %2")
                .arg(getTcpName(st.index))
                .arg(st.port);
//        if(st.dataPart>0.1){
            text.append(QString("(%1%)").arg((int)(100*st.dataPart)));
//        }
        l->setText(text);
        l->setStyleSheet(QString("QLabel { color : %1; }")
                         .arg(st.isFree ? "red" : "green"));
    }
}

void MainWindow::setOutTcpServerNumber(int tcpNum)
{
    if(tcpNum < 1)
        tcpNum = 1;
    if(config.combinerOutTcpNum != tcpNum){
        config.combinerOutTcpNum = tcpNum;
        config.write_config();
    }

    QLayout *layout = ui->horizontalLayoutOutTcp;
    for(int tcpId=0; tcpId <= tcpNum; ++tcpId){
        if(getTcpLabel(tcpId)) continue;
        QLabel *label = new QLabel(QString("%1: creating...").arg(getTcpName(tcpId)), ui->widgetOutTcp);
//        if(tcpId == 0) label->setVisible(false);
        label->setProperty("id", tcpId);
        layout->addWidget(label);
        label->setAlignment(Qt::AlignHCenter);
    }

    for(int index=0; index < layout->count(); ) {
        bool ok;
        QWidget *w = layout->itemAt(index)->widget();
        if(w->property("id").toInt(&ok) > tcpNum && ok) {
            layout->removeWidget(w);
            delete w;
        } else {
            ++index;
        }
    }

    emit setOutput(config.combinerOut, config.combinerOutTcpNum);
}

QVector<ProgramDescription> MainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for (const ProgramDescription &descr : progDescrs) {
        if(isFreeProg(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool MainWindow::isFreeProg(const ProgramDescription &pd) const
{
    for (auto client : clients.values()) {
        if(pd.sameProgram(client->getProgDescr())){
            return false;
        }
    }

    int interface_id=config.startEvNumber % config.evNumberStep;
    for (const ProgramInterface &i : pd.interfaces) {
        if(i.type == ProgramInterface::getTypeName(ProgramInterfaceDataFlow)
                && i.enabled && i.isFree && i.id == interface_id) {
            return true;
        }
    }

    return false;
}

int MainWindow::getTableRowByClientIndex(int index) const
{
    for(int row=0; row<ui->tableWidget->rowCount(); ++row)
        if(ui->tableWidget->item(row, COL_ID)->text().toInt() == index)
            return row;
    return -1;
}

QString MainWindow::getClientNameByIndex(int index) const
{
    return config.clients.contains(index) ?
                config.clients[index].descr.getName() :
                QString("Client #%1").arg(index);
}

int MainWindow::getSelectedRow() const
{
    int row=-1;
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    if(!ranges.isEmpty()){
        QTableWidgetSelectionRange firstRange = ranges[0];
        row = firstRange.topRow();
    }
    return row;
}

int MainWindow::getClientIdByRow(int row) const
{
    if(row<0 || row>=ui->tableWidget->rowCount()){
        return -1;
    } else {
        return ui->tableWidget->item(row, COL_ID)->text().toInt();
    }
}

void MainWindow::updateTableTextColor(int row, int clientId) const
{
    if(clientId==-1)
        clientId = getClientIdByRow(row);
    if(clientId==-1)
        return;

    const bool &isOk = !config.clients[clientId].enabled || lastConnactionState[clientId];
    QColor color = isOk ? QColor() : Qt::red;
    QTableWidgetItem *i = ui->tableWidget->item(row, COL_STATUS);
    if(i->textColor() != color)
        i->setTextColor(color);
}

void MainWindow::combinerOutUpdated()
{
    if(config.isOutTcp() && config.combinerOutTcpNum==0){
        setOutTcpServerNumber(1);
        config.write_config();
    }

    ui->actionSet_number_of_outputTcpServers->setEnabled(config.isOutTcp());
    ui->labelDataFileName->setVisible(config.isOutFile());
    ui->labelCombinerIdle->setVisible(config.isOutIdle());
    ui->widgetOutTcp->setVisible(config.isOutTcp());
    emit setOutput(config.combinerOut, config.combinerOutTcpNum);
}

QLabel *MainWindow::getTcpLabel(int id) const
{
    bool ok;
    for(int i=0; i<ui->horizontalLayoutOutTcp->count(); ++i) {
        QLayoutItem *item = ui->horizontalLayoutOutTcp->itemAt(i);
        if(item) {
            QWidget *w = item->widget();
            if(w) {
                QLabel *l = qobject_cast<QLabel *>(w);
                if(l) {
                    if(l->property("id").toInt(&ok)==id && ok)
                        return l;
                }
            }
        }
    }
    return nullptr;
}


void MainWindow::initCombiner()
{
    combiner->initOutputBuffers(config.outputBufSize, config.combinerOut, config.combinerOutTcpNum);

    combiner->moveToThread(combinerThread);
    combinerThread->start();
    connect(this, &MainWindow::reset, combiner, &Combiner::reset);
    connect(this, &MainWindow::starting, combiner, &Combiner::start);
    connect(this, &MainWindow::stoping, combiner, &Combiner::stop);
    connect(this, &MainWindow::destroyClient, combiner, &Combiner::destroyClient);
    connect(this, &MainWindow::destroyCombiner, combiner, &Combiner::destroyClient);
    connect(this, &MainWindow::storageChanged, combiner, &Combiner::storageChanged);
    connect(this, &MainWindow::setOutput, combiner, &Combiner::setOutput);
    connect(this, &MainWindow::analizerListUpdated, combiner, &Combiner::updateAnalizerList);
    connect(this, &MainWindow::analizerListUpdated, statisticDialog, &StatisticDialog::analizerListUpdated);
    connect(this, &MainWindow::enableTimeDistr, combiner, &Combiner::enableTimeDistr);
    connect(this, &MainWindow::setEvNumberStep, combiner, &Combiner::setStartEvNumber);
    connect(combiner, &Combiner::started, this, &MainWindow::clientStarted);
    connect(combiner, &Combiner::stopped, this, &MainWindow::clientStopped);
    connect(combiner, &Combiner::resetted, this, &MainWindow::deviceResetted);
    connect(combiner, &Combiner::error, this, &MainWindow::deviceError);

    connect(combiner, &Combiner::destroyed, combinerThread, &QThread::quit);
    connect(combiner, &Combiner::status, this, &MainWindow::updateStatus);
    connect(combiner, &Combiner::statusRaw, this, &MainWindow::updateCombinerStats);
    connect(combiner, &Combiner::statusRaw, statisticDialog, &StatisticDialog::updateCombinerStats);
    connect(combiner, &Combiner::fileWriterStatsUpdated, this, &MainWindow::fileWriterStatsUpdated);
    connect(combiner, &Combiner::outputBufferSizeUpdated, this, &MainWindow::outputBufferSizeUpdated);
    connect(this, &MainWindow::setMonitorPortNum, combiner, &Combiner::setMonitorPortNum);

    recivedStates[0] = false; // Combiner's id = 0

    update_combiner_to_config();
}

void MainWindow::update_combiner_to_config()
{
    if(!config.storageDir.isEmpty()){
        emit storageChanged(config.storageDir);
    }
    updateAnalizerList();

    ui->actionOutToFile->setChecked(config.isOutFile());
    ui->actionOutToTcp->setChecked(config.isOutTcp());
    ui->actionCombinerIdle->setChecked(config.isOutIdle());
    combinerOutUpdated();

    ui->actionEnable_client_time_distribution->setChecked(config.timeDistrEn);
    emit setEvNumberStep(config.startEvNumber, config.evNumberStep);
    if(config.monitorPort)
        emit setMonitorPortNum(config.monitorPort);
}

void MainWindow::runControlStateUpdated(const RC_STATE &state, uint stateIndex)
{
    curStateIndex = stateIndex;
    if(state==RC_STARTING)
        emit starting(curStateIndex);
    else if(state==RC_STOPPING) {
        emit stoping(curStateIndex);
    } else if(state==RC_RESETTING) {
        emit reset(loc_runIndex, loc_runNumber, curStateIndex);
        backPreasure = false;
    }
    updatePushButtonState();
}

void MainWindow::clientStarted(int index, bool ok,uint stateIndex)
{
    if(curStateIndex != stateIndex){
        qWarning()<<QString("%1 (id:%2) send wrong stateIndex(%3; curIndex=%4) on clientStarted")
                    .arg(getClientNameByIndex(index)).arg(index)
                    .arg(stateIndex).arg(curStateIndex);
        return;
    }

    Q_ASSERT(recivedStates.contains(index));
    recivedStates[index] = true;
    allOk &= ok;
    for(bool st : recivedStates) {
        if(st==false)
            return;
    }

    for(int key : recivedStates.keys())
        recivedStates[key] = false;

    ui->widgetRunControl->setState(RC_STARTED);
}

void MainWindow::clientStopped(int index, uint stateIndex)
{
    if(curStateIndex != stateIndex) {
        qWarning()<<QString("%1 (id:%2) send wrong stateIndex(%3; curIndex=%4) on clientStopped")
                    .arg(getClientNameByIndex(index)).arg(index)
                    .arg(stateIndex).arg(curStateIndex);
        return;
    }
    Q_ASSERT(recivedStates.contains(index));
    recivedStates[index] = true;
    for(bool st : recivedStates) {
        if(st==false)
            return;
    }

    for(int key : recivedStates.keys())
        recivedStates[key] = false;

    ui->widgetRunControl->setState(RC_STOPPED);
}

void MainWindow::deviceResetted(int index, uint stateIndex)
{
    if(curStateIndex != stateIndex){
        qWarning()<<QString("%1 (id:%2) send wrong stateIndex(%3; curIndex=%4) on deviceResetted")
                    .arg(getClientNameByIndex(index)).arg(index)
                    .arg(stateIndex).arg(curStateIndex);
        return;
    }
    Q_ASSERT(recivedStates.contains(index));
    recivedStates[index] = true;
    for(bool st : recivedStates) {
        if(st==false)
            return;
    }
    qInfo() << "All parts were resetted";

    for(int key : recivedStates.keys())
        recivedStates[key] = false;

    ui->widgetRunControl->setState(RC_STOPPED);
}

void MainWindow::deviceError(int index, QString errStr)
{
    QString msg = index ? getClientNameByIndex(index) : "Combiner";
    msg.prepend("Got error from ");
    qCritical() << msg << errStr;
    if(run)
        emit sendRunError(loc_runIndex, loc_runNumber, errStr);
}

void MainWindow::updateStatus(const QString &status)
{
    statusBar()->showMessage(status);
    emit sendRemoteState(status, true);
}

void MainWindow::updateCombinerStats(const CombinerStats &ss)
{
    if(backPreasure){
        if(ss.fwDataPart > 0.9) {
            backPreasure = false;
            QString mess = "Enable backPressure";
            qWarning() << mess;
            emit sendRunWarning(loc_runIndex, loc_runNumber, mess);
            emit sendBackPreasure(loc_runIndex, loc_runNumber, true);
        }
    } else {
        if(ss.fwDataPart < 0.5) {
            backPreasure = true;
            qInfo() << "BackPreasure disabled";
            emit sendBackPreasure(loc_runIndex, loc_runNumber, false);
        }
    }
    bool outDataInterfaceChanged = false;
    bool outDataPartsChanged = false;
    for (int id : ss.tcpStatus.keys()) {
        if(!lastCombinerStats.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
        const OutputTcpStatus &prevS = lastCombinerStats.tcpStatus[id];
        const OutputTcpStatus &nextS = ss.tcpStatus[id];
        outDataInterfaceChanged |= prevS.isFree!=nextS.isFree || prevS.port!=nextS.port;
        outDataPartsChanged |= !qFuzzyCompare(prevS.dataPart, nextS.dataPart);
    }
    for (int id : lastCombinerStats.tcpStatus.keys()) {
        if(!ss.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
    }
    lastCombinerStats = ss;
    QTableWidget * table = ui->tableWidget;
    const bool tableUpdatesEnabled = table->updatesEnabled();
    table->setUpdatesEnabled(false);
    if(outDataInterfaceChanged)
        updateProgramDescription();
    if(outDataInterfaceChanged || outDataPartsChanged)
        updateTcpPortLabels();

    if(ss.lastEventNum.isEmpty()) {
        table->setUpdatesEnabled(tableUpdatesEnabled);
        return;
    }
    for(int row=0;row<table->rowCount();++row){
        bool ok;
        int clientId = table->item(row, COL_ID)->text().toInt(&ok);
        if(!ok)
            continue;
        QString str;
        if(ss.lastEventNum.contains(clientId))
            str = QString::number(ss.lastEventNum[clientId]);
        table->item(row, COL_EV_NUM)->setText(str);
        table->item(row, COL_TRASHED_EVENTS)->setText(QString::number(ss.trashedEvents[clientId]));
        table->item(row, COL_ALL_EVENTS)->setText(QString::number(ss.dataEvents[clientId]));
        const quint64 flushedSize = ss.flushedSize[clientId];
        table->item(row, COL_FLUSHED_SIZE)->setText(humanReadableSize(flushedSize));
        table->item(row, COL_FLUSHED_SIZE)->setToolTip(QString::number(flushedSize));
        if(config.timeDistrEn){
            if(ss.timeDistrCnt && ss.timeDistr.contains(clientId) && ss.timeDistrFull.contains(clientId))
                table->item(row, COL_TIME_DISTR)->
                        setText(QString("%1/%2")
                                .arg(ss.timeDistr[clientId]/ss.timeDistrCnt)
                                .arg(ss.timeDistrFull[clientId]/ss.timeDistrCnt));
            else
                table->item(row, COL_TIME_DISTR)->setText("-");
        }

        if(ss.freeEvPart.contains(clientId)) {
            QString evBuff = QString("%1/%2/%3")
                    .arg(ss.freeEvPart[clientId])
                    .arg(ss.preparedEvPart[clientId])
                    .arg(ss.flushedEvPart[clientId]);
            table->item(row, COL_EV_BUFF)->setText(evBuff);
        }

        // metrics
        QString clientName = QString("%1").arg(clientId, 8, 16, QChar('0'));
        if (clients.contains(clientId) && clients[clientId]) {
            const ClientManager * cm = clients[clientId];
            const ProgramDescription &pd = cm->getProgDescr();
            clientName = pd.index;
        }
        QString metricPrefix = QString("in.%1.").arg(clientName);
        MetricSender::send(metricPrefix+"subevents.drop", ss.trashedEvents[clientId]);
        MetricSender::send(metricPrefix+"subevents.pass", ss.dataEvents[clientId] - ss.trashedEvents[clientId]);
    }

    {
        QString mess;
//        if(!enabled)
//            mess.append("stopped; ");
        QString clientsStr;
        QList<int> clientsList = ss.lastEventClients.toList();
        qSort(clientsList);
        int first=-1;
        int last=-1;
        for(int i : clientsList) {
            if(first==-1) {
                first = i;
            } else if(last+1!=i) {
                if(first == last)
                    clientsStr.append(QString("%1 ").arg(first));
                else
                    clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
                first = i;
            }
            last = i;
        }

        if(first!=-1){
            if(first == last)
                clientsStr.append(QString("%1 ").arg(first));
            else
                clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
        }

        mess.append(QString("reg:%1;skip:%2;eob:%3;clients:%4;nextEv:%5  monitor:%6")
                    .arg(ss.regularEvents)
                    .arg(ss.skipped)
                    .arg(ss.eob)
                    .arg(clientsStr)
                    .arg(ss.expectedEvNum)
                    .arg(ss.monitorEvCount)
                    );
//        for(int i=0;i<events.size(); ++i){
//            const EventMetaInfo* ev = events[i];
//            if(ev == nullptr) continue;
//            mess.append(QString("; #%1=%2").arg(i).arg(events[i]->evNum));
//        }
        updateStatus(mess);
    }

    labelFileErrSize->hide();
    labelFileErrSize->setText(QString("fileErr:%1B").arg(ss.missedFileSize));
    if(ss.missedFileSize){
        labelFileErrSize->show();
        labelFileErrSize->setStyleSheet("QLabel { color : red;}");
    }
    labelFileErrSize->setStyleSheet(QString("QLabel { color : %1;}")
                                    .arg(ss.missedFileSize?"red" : "black"));
    updateTableSize();
    // Periodic print of client's number distribution
//    if(!clientDistrTime->isValid() || clientDistrTime->elapsed()>1000){
//        QString str("clientNumInEvents\t");
//        quint64 sum=0;
//        for(int n : ss.clientNumDistr.keys()) {
//            str.append(QString("%1:%2; ").arg(n).arg(ss.clientNumDistr[n]));
//            sum += ss.clientNumDistr[n]*n;
//        }
//        qInfo()<<str;
//        qInfo()<<"Sum:"<<sum<<"; skipped:"<<ss.skipped<<";Sum+skipped:"<<sum+ss.skipped;
//        qInfo()<<"Exp:"<<ss.expectedEvNum<<"; clients:"<<enabledClients<<"Exp*clients:"<<ss.expectedEvNum*enabledClients;
//        clientDistrTime->start();
//    }
    table->setUpdatesEnabled(tableUpdatesEnabled);
}

void MainWindow::fileWriterStatsUpdated(const FileWriterStats &fwStat)
{
    QString dataFileNameText;
    if(!fwStat.fileName.isEmpty()){
        dataFileNameText = QString("%1 (%2)")
                .arg(fwStat.fileName)
                .arg(humanReadableSize(fwStat.writtenFileSize));
        if(fwStat.missedFileSize)
            dataFileNameText.append(QString(" (missed: %1b)").arg(fwStat.missedFileSize));
    } else {
        dataFileNameText = QString(fwStat.predefStorage).append("...");
    }
    ui->labelDataFileName->setText(dataFileNameText);
}

void MainWindow::inputBufferSizeUpdated(int id, quint64 size)
{
    qInfo() << "Save in cgf input resize:" << id << size;
    if(size<=0) {
        qWarning() << "Got inputBufferSizeUpdated with incorrect size:"<<size;
        return;
    }
    if(!config.clients.contains(id)) {
        qWarning() << "Got inputBufferSizeUpdated for missing clientId:"<<id;
        return;
    }

    config.clients[id].inputBufSize = size;
    config.write_config();
}

void MainWindow::outputBufferSizeUpdated(quint64 size)
{
    qInfo() << "Save in cgf output resize:" << size;
    if(size<=0) {
        qWarning() << "Got outputBufferSizeUpdated with incorrect size:"<<size;
        return;
    }
    config.outputBufSize = size;
    config.write_config();
}

void MainWindow::on_pushButtonReset_clicked()
{
    loc_runNumber=0;
    if(!ui->widgetRunControl->setState(RC_RESETTING)) {
        qWarning() << "Resetting denied";
    }
}

void MainWindow::remoteStartRequest(QString runIndex, int runNumber, quint64 startEvNumber)
{
    bool ok=true;
    ui->plainTextEdit->appendPlainText("");
    qInfo() << QString("Starting remote run: runIndex=%1 runNumber=%2")
               .arg(runIndex).arg(runNumber);
    config.run_number = runNumber;
    config.write_config();
    Q_UNUSED(startEvNumber);
    if(loc_runIndex == runIndex && loc_runNumber > runNumber){
        qWarning()<<QString("Got wrong stop Run args: last run (%1_%2) is greater than remote run (%3_%4)")
                    .arg(loc_runIndex).arg(loc_runNumber)
                    .arg(runIndex).arg(runNumber);
        emit remoteCompletedStop(false, runIndex, runNumber);
        return;
    }

    RunControl *rc = ui->widgetRunControl;
    RC_STATE curState = rc->getState();

    if((loc_runIndex != runIndex || loc_runNumber != runNumber) &&
            (curState == RC_STARTED || curState == RC_STARTING || curState == RC_RESETTING)) {
        ok = tryToStop();
        if(ok) {
            run=false;
            qInfo()<<"Wrong Run has been stopped";
        } else
            qWarning()<<"Failed to stop wrong Run";
    }

    if(!ok){
        emit remoteCompletedStart(ok, runIndex, runNumber);
        return;
    }
    curState = rc->getState();

    loc_runIndex = runIndex;
    loc_runNumber = runNumber;
    run=true;
    switch (curState) {
    case RC_STOPPING:
        ok = rc->waitState(RC_STOPPED);
        if(!ok){
            qWarning()<<"Failed to wait RC_STOPPED state from RC_STOPPING state";
            break;
        }
    case RC_STOPPED:
        rc->setState(RC_RESETTING);
        ok = rc->waitState(RC_STOPPED, 2000);
        if(!ok){
            qWarning()<<"Failed to wait RC_STOPPED state from RC_RESETTING state";
            break;
        }
        allOk = true;
        rc->setState(RC_STARTING);
    case RC_STARTING:
        ok = rc->waitState(RC_STARTED, 2000);
        if(!ok){
            qWarning()<<"Failed to wait RC_STARTED state from RC_STARTING state";
            break;
        }
        if(!allOk){
            ok = false;
            qWarning()<<"Some client failed to stat (allOk=false)";
            break;
        }
    }
    if(!ok){
        run=false;
    }

    emit remoteCompletedStart(ok, runIndex, runNumber);
    if(ok){
        emit config_transfer_to_rc(config.root.get_json().toBinaryData());
    }
}

void MainWindow::remoteStopRequest(QString runIndex, int runNumber, quint64 stopEvNumber)
{
    bool ok=true;
    RunControl *rc = ui->widgetRunControl;
    if(loc_runIndex != runIndex || loc_runNumber != runNumber){
        qWarning()<<QString("Got wrong stop Run args: last run (%1_%2), remote run (%3_%4)")
                    .arg(loc_runIndex).arg(loc_runNumber)
                    .arg(runIndex).arg(runNumber);
        emit remoteCompletedStop(false, runIndex, runNumber);
        return;
    }
    lastEvNum = stopEvNumber;
    if(lastEvNum) {
        QElapsedTimer timer;
        timer.start();
        while( (lastCombinerStats.expectedEvNum != lastEvNum+1)
               && !timer.hasExpired(2000)){
            QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 50);
        }
        if(lastCombinerStats.expectedEvNum != lastEvNum+1){
            qWarning() << "Failed to stop evb";
        }
    }

    RC_STATE curState = rc->getState();
    switch (curState) {
    case RC_STARTING:
        ok = rc->waitState(RC_STARTED);
        if(!ok){
            qWarning()<<"Failed to wait RC_STARTED state from RC_STARTING state";
            break;
        }
    case RC_STARTED:
         rc->setState(RC_STOPPING);
         ok = rc->waitState(RC_STOPPED, 2000);
         if(!ok){
             qWarning()<<"Failed to wait RC_STOPPED state from RC_STOPPING state";
             break;
         }
    case RC_STOPPING:
        ok = rc->waitState(RC_STOPPED, 2000);
        if(!ok){
            qWarning()<<"Failed to wait RC_STOPPED state from RC_STOPPING state";
            break;
        }
    case RC_STOPPED:
        break;
    }

    run = false;
    emit remoteCompletedStop(ok, runIndex, runNumber);
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat)
{
    if(!progDesc.containsProgramInterface(ProgramInterfaceRemoteControl,0)){
        ProgramInterface remoteControlInterface;
        remoteControlInterface.setType(ProgramInterfaceRemoteControl);
        remoteControlInterface.id = 0;
        progDesc.interfaces.push_back(remoteControlInterface);
    }

    ProgramInterface &remoteInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    remoteInterface.port = rcSerStat.tcpPort;
    remoteInterface.enabled = rcSerStat.listening;
    remoteInterface.isFree = !rcSerStat.portInUse;

    if(rcSerStat.portInUse) {
        remoteInterface.setPeer(rcSerStat.peerHost, rcSerStat.peerPort);
        updateRemoteStatus();
        qInfo() << "Remote control connected.";
    } else {
        qWarning() << "Remote control disconnected.";
        loc_runIndex.clear();
        loc_runNumber = 0;

        if(ui->widgetRunControl->getState() != RC_STOPPED) {
            qCritical() << "Stopping current run";
            tryToStop();
        }
    }

    if(ui->widgetRunControl->isEnabled() == rcSerStat.portInUse){
        ui->widgetRunControl->setEnabled(!rcSerStat.portInUse);
    }
    updateProgramDescription();
}

void MainWindow::addLog(QString log, Qt::GlobalColor color)
{
//    qDebug()<<QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);
    int repeat=0;
    int now = time(nullptr);
    bool found=false;
    for(int i=0;i<logHistory.size();++i) {
        LogHistorEntry &e = logHistory[i];
        if(log == e.mess){
            found=true;
            repeat = e.repeatCnt++;
            if(now - e.time > 5){
                e.time = now;
                logHistory.move(i,logHistory.size()-1);
            } else
                return;
        }
    }
    if(!found){
        if(logHistory.size() >= (3*clients.size()+3))
            logHistory.removeAt(0);
        LogHistorEntry e;
        e.mess = log;
        e.repeatCnt = 0;
        e.time = now;
        logHistory.append(e);
    }

    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(color));

    SyslogMessageSender::log(color==Qt::red ? LOG_PRI_WARNING : LOG_PRI_NOTICE, log.toStdString().c_str());
    if(repeat)
        log.append(QString(" (repeat: %1 times)").arg(repeat));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(QString("[%1] %2").arg(QTime::currentTime().toString(), log));
}

void MainWindow::addLog(QtMsgType severity, QString log, QString thrName)
{
    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

    Qt::GlobalColor color;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    default:
        color = Qt::black;
    }
    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(QString("[%1] [%2] %3")
                                          .arg(QTime::currentTime().toString(), thrName, log));
}

void MainWindow::on_actionChange_storage_dir_triggered()
{
//    QFileDialog::QFileDialog ( QWidget * parent = 0, const QString & caption = QString(), const QString & directory = QString(), const QString & filter = QString() )
//    QFileDialog dialog(this, "Set predefined storage for mpd data file", "", "*.data");
//    dialog.setOption(QFileDialog::ShowDirsOnly);
//    dialog.exec();
//    qDebug() << dialog.selectedFiles();

    QString dir = QFileDialog::getExistingDirectory( this, "Choose storage directory...",
                                                     config.storageDir, QFileDialog::ShowDirsOnly );
    if(!dir.isEmpty() && config.storageDir != dir){
        config.storageDir = dir;
        config.write_config();
        emit storageChanged(config.storageDir);
    }
}

void MainWindow::on_actionReset_to_defaults_triggered()
{
    config.storageDir = "";
    config.isRemoteEnabled = true;
    config.reset_to_defaults();
    config.write_config();

    emit storageChanged(config.storageDir);
//    emit allowRemoteControl(config.isRemoteEnabled);
}

void MainWindow::on_actionAdvanced_counters_toggled(bool showColumns)
{
    if(config.advancedCounters != showColumns) {
        config.advancedCounters = showColumns;
        config.write_config();
    }

    QTableWidget *table = ui->tableWidget;
//    table->setColumnHidden(COL_EV_NUM, !showColumns);
//    table->setColumnHidden(COL_ID, !showColumns);
    table->setColumnHidden(COL_TCP_BUFF, !showColumns);
    table->setColumnHidden(COL_EV_BUFF, !showColumns);
    updateTableSize();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    if(config.isRemoteEnabled != checked) {
        config.isRemoteEnabled = checked;
        config.write_config();
    }
    emit allowRemoteControl(config.isRemoteEnabled);
}

void MainWindow::data_out_changed(QAction *a)
{
    if(a == ui->actionCombinerIdle) config.combinerOut = COMBINER_IDLE;
    else if(a == ui->actionOutToTcp) config.combinerOut = COMBINER_TCP;
    else if(a == ui->actionOutToFile) config.combinerOut = COMBINER_FILE;
    else {
        qCritical() << "Data out was switch to unknown state.";
    }

    combinerOutUpdated();
    config.write_config();
    updateProgramDescription();
}

void MainWindow::on_actionSet_number_of_outputTcpServers_triggered()
{
    bool ok;
    const int tcpNum = QInputDialog::getInt(this, tr("TCP servers for output"),
                                 tr("Set number of TCP servers for output:"), config.combinerOutTcpNum, 1, 20, 1, &ok);
    if (ok && config.combinerOutTcpNum != tcpNum)
        setOutTcpServerNumber(tcpNum);
}

void MainWindow::on_actionSet_EvB_index_triggered()
{
    bool ok;
    quint32 startNumber = QInputDialog::getInt(this, tr("Event builder index"),
                                 tr("Set 1st event number (EvB index):"), config.startEvNumber, 0, 19, 1, &ok);
    if(!ok)
        startNumber = config.startEvNumber;
    quint32 stepNumber = QInputDialog::getInt(this, tr("EvB event numbet step"),
                                 tr("Set event number step:"), config.evNumberStep, 1, 20, 1, &ok);
    if(!ok)
        startNumber = config.evNumberStep;

    if (config.startEvNumber != startNumber || config.evNumberStep != stepNumber) {
        int evbIndexOld = config.getEvNumIndex();
        config.startEvNumber = startNumber;
        config.evNumberStep = stepNumber;
        emit setEvNumberStep(config.startEvNumber, config.evNumberStep);
        config.write_config();
        if(evbIndexOld != config.getEvNumIndex()){
            QMessageBox msgBox;
            msgBox.setText("Event Builder index was changed, configuration was saved.\n"
                           "Needed manual restart\n"
                           "Closing...");
            msgBox.exec();
            close();
        }
    }
}

void MainWindow::on_actionShow_statistics_triggered()
{
    statisticDialog->show();
}

void MainWindow::on_actionEnable_client_time_distribution_toggled(bool en)
{
    if(config.timeDistrEn != en) {
        config.timeDistrEn = en;
        config.write_config();
    }

    emit enableTimeDistr(config.timeDistrEn);
    QTableWidget *table = ui->tableWidget;
    table->setColumnHidden(COL_TIME_DISTR, !config.timeDistrEn);
    updateTableSize();
}

void MainWindow::on_actionSet_monitor_port_number_triggered()
{
    bool ok;
    const int MIN_PORT_NUMBER = 0;
    const int MAX_PORT_NUMBER = 32999;
    const int portNum = QInputDialog::getInt(this, tr("Set monitor port number"),
                                             QString("Set monitor port number (in range %1-%2):")
                                             .arg(MIN_PORT_NUMBER).arg(MAX_PORT_NUMBER),
                                             config.monitorPort ? config.monitorPort : MAX_PORT_NUMBER, MIN_PORT_NUMBER, MAX_PORT_NUMBER, 1, &ok);
    if (ok && config.monitorPort != portNum) {
        config.monitorPort = portNum;
        config.write_config();
        emit setMonitorPortNum(static_cast<quint16>(portNum));
//        QMessageBox msgBox;
//        msgBox.setText("Monitor port number was changed, configuration was saved.\n"
//                       "Needed manual restart\n"
//                       "Closing...");
//        msgBox.exec();
//        close();
    }
}

void MainWindow::setClientReadOut(bool readout)
{
    QTableWidget *table = ui->tableWidget;
    for(int row=0; row<table->rowCount(); ++row){
        if(table->cellWidget(row, COL_READOUT_CB)==reinterpret_cast<QWidget *>(sender())){
            bool ok;
            int clientId = table->item(row, COL_ID)->text().toInt(&ok);
            Q_ASSERT(ok);
//            qInfo() << QString("client at row=%1 clientId=%2 was %3abled")
//                       .arg(row).arg(clientId).arg(readout?"en":"dis");
            Q_ASSERT(config.clients.contains(clientId));
            if(config.clients[clientId].enabled != readout) {
                config.clients[clientId].enabled = readout;
                config.write_config();
                Q_ASSERT(clients.contains(clientId));
                updateTableTextColor(row, clientId);
                updateTablePalette();
            }
            emit updateClientReadout(clientId, readout);
            updateAnalizerList();
            config.write_config();
            recivedStates[clientId] = false;
            return;
        }
    }
}

void MainWindow::gotProgramDescription(const ProgramDescription &newDescr)
{
    if(newDescr.type != DaqConfig::getTypeName(DaqConfigMStream) &&
       newDescr.type != DaqConfig::getTypeName(DaqConfigEventBuilder) &&
       newDescr.type != DaqConfig::getTypeName(DaqConfigGenerator) &&
       newDescr.type != DaqConfig::getTypeName(DaqConfigVmeDaq) &&
       newDescr.type != DaqConfig::getTypeName(DaqConfigLand))
       return;

    bool missing = true;

    for (int i=0; i < progDescrs.size(); ++i) {
        ProgramDescription &descr = progDescrs[i];
        if(newDescr.sameProgram(descr)) {
            missing = false;
            if(descr == newDescr)
                return; // Got dublicate
            descr = newDescr;
            break;
        }
    }

    if(missing)
        progDescrs.push_back(newDescr);

    for(int index : clients.keys()) {
        ClientManager *client = clients[index];
        if(newDescr.sameProgram(client->getProgDescr())) {
            client->setProgDescription(newDescr);
            const int row = getTableRowByClientIndex(client->getIndex());

            if(row != -1) {
                const QString h = newDescr.hostName.split('.').first();
                if(ui->tableWidget->item(row, COL_PROG_HOST)->text() != h)
                    ui->tableWidget->item(row, COL_PROG_HOST)->setText(h);
            }

            break;
        }
    }
    updatePushButtonState();
}

void MainWindow::gotClosedProgram(const ProgramDescription &newDescr)
{
    for(int i=0; i< progDescrs.size(); ++i) {
        if(progDescrs[i].uuid == newDescr.uuid){
            progDescrs.remove(i);
            break;
        }
    }
    updatePushButtonState();
}

void MainWindow::updateTablePalette()
{
    int row = getSelectedRow();
    if(row == -1)
        return;
    int clientId = getClientIdByRow(row);
    if(clientId == -1)
        return;
    const bool &isConn = lastConnactionState.contains(clientId) ? lastConnactionState[clientId] : false;
    const bool &isOk = isConn || !config.clients[clientId].enabled;

    QPalette p = ui->tableWidget->palette();
    QColor color = isOk ? QColor() : Qt::red;
    if(p.color(QPalette::HighlightedText) != color) {
        p.setColor(QPalette::HighlightedText, color);
        ui->tableWidget->setPalette(p);
    }
}

QString MainWindow::humanReadableSize(double size)
{
    QStringList list;
    list << "KB" << "MB" << "GB" << "TB";

    QStringListIterator i(list);
    QString unit("bytes");

    while(size >= 1024.0 && i.hasNext())
    {
        unit = i.next();
        size /= 1024.0;
    }
    return QString().setNum(size,'f',2)+" "+unit;
}

bool MainWindow::tryToStop()
{
    qWarning()<<"Trying to stop wrong current Run";

    bool ok= true;
    RunControl *rc = ui->widgetRunControl;
    RC_STATE curState = rc->getState();
    switch (curState) {
    case RC_RESETTING:
    case RC_STARTING:
        ok = rc->waitState(RC_STARTED, 2000);
        if(!ok){
            qWarning()<<"Failed to wait RC_STARTED state from RC_STARTING state";
            break;
        }
    case RC_STARTED:
        rc->setState(RC_STOPPING);
        ok = rc->waitState(RC_STOPPED, 2000);
        if(!ok){
            qWarning()<<"Failed to wait RC_STOPPED state from RC_STOPPING state";
            break;
        }
        break;
    default:
        break;
    }
    return ok;
}

