#include "OutputTcpServer.h"

#include <QCoreApplication>
#include <QHostInfo>
#include <QTcpServer>
#include <QTcpSocket>
#include <QTimer>
#include <QElapsedTimer>
#include <stdexcept>

#include "params.h"
#include "metric-sender/MetricSender.h"
#include "TcpServerStats.h"

namespace {
const int MINIMAL_PACKET_SIZE = 100e3;
const int TIMEOUT_SEND = 10;
const int METRICS_SEND_PERIOD_MS = 1000;
const int FLUSH_TIMEOUT_MS = 1000;
}

OutputTcpServer::OutputTcpServer(int _index, quint16 port, QObject *parent) :
    QObject(parent),
    index(_index),
    preferedPortNum(port),
    tcpPort(0),
    qTcpServer(new QTcpServer(this)),
    stats(new TcpServerStats()),
    tcpSocket(nullptr),
    writeTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    slowConnection(false),
    ringBuf(new RingBuffer<char>(TCP_SERVER_BUFFER_SIZE)),
    finishId(0),
    status(index),
    runEnabled(false)
{
    connect(qTcpServer, &QTcpServer::newConnection, this, &OutputTcpServer::updateConnections);
    connect(writeTimer, &QTimer::timeout, this, &OutputTcpServer::writeData);
    connect(statusTimer, &QTimer::timeout, this, &OutputTcpServer::sendStatus);
    statusTimer->start(500);
    writeTimer->setSingleShot(false);

    QTimer * sendMetricsTimer = new QTimer(this);
    connect(sendMetricsTimer, &QTimer::timeout, this, &OutputTcpServer::sendMetrics);
    sendMetricsTimer->start(METRICS_SEND_PERIOD_MS);
}

OutputTcpServer::~OutputTcpServer()
{
    qDebug() << "deletting...";
    if(tcpSocket) {
        tcpSocket->deleteLater();
        tcpSocket=nullptr;
    }
    qTcpServer->close();
    delete stats;
}

void OutputTcpServer::startListen(int i, bool listen)
{
    if(index != i)
        return;
    if(listen) {
        while(!qTcpServer->listen(QHostAddress::Any, preferedPortNum)) {
            QAbstractSocket::SocketError errNum = qTcpServer->serverError();
            if(errNum==QAbstractSocket::AddressInUseError){
                qWarning() << tr("Unable to start on port %1: %2, restart in 5 s")
                              .arg(preferedPortNum).arg(qTcpServer->errorString());
                exit(-1);
            }
            QString s = tr("Unable to start on port %1: %2").arg(preferedPortNum).arg(qTcpServer->errorString());
            qWarning() << s;
            throw std::runtime_error(s.toStdString());
        }
        tcpPort = qTcpServer->serverPort();
        Q_ASSERT(!preferedPortNum || (tcpPort == preferedPortNum));
        sendStatus();
        qInfo() << QString("Started on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    } else {
        qTcpServer->close();
        tcpPort = 0;
        sendStatus();
        qInfo() << "Stopped";
    }
}

void OutputTcpServer::deleteTcp(int i)
{
    if(i==index || i==0) {
        deleteLater();
    }
}

void OutputTcpServer::writeData()
{
    ringBuf->waitDataSpace(MINIMAL_PACKET_SIZE, TIMEOUT_SEND);
    int size = ringBuf->getContinuosDataSize(0);
    if(size==0)
        return;
    //Check connetion
    if(!tcpSocket || !tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState) {
        writeTimer->stop();
        return;
    }

    if(tcpSocket->bytesToWrite() > TCP_SERVER_BUFFER_SIZE) {
        if(!slowConnection && index){
            qWarning() << QString("Connection with %1:%2 is slow. Wait all data will be read.")
                          .arg(peerHost).arg(tcpSocket->peerPort());
            slowConnection = true;
        }
        writeTimer->start(TIMEOUT_SEND);
        return;
    }
    writeTimer->setInterval(0);
    if(slowConnection) {
        qInfo() << QString("Connection with %1:%2 normolized.")
                   .arg(peerHost).arg(tcpSocket->peerPort());
        slowConnection = false;
    }

    writeLatencyElapsedTimer.restart();
    qint64 sentSize = tcpSocket->write(ringBuf->getReadPointer(), size);
    qint64 elapsed = writeLatencyElapsedTimer.nsecsElapsed();
    stats->addLatencyValue(elapsed);
    if(sentSize == -1) {
        qWarning() << tcpSocket->errorString().prepend("Error:");
    } else {
        stats->bytesSent += sentSize;
        ringBuf->trashData(sentSize);
        ringBuf->freeData(sentSize);
        if(sentSize != size)
            qWarning() << QString("Incomplete send %1 of %2 bytes:%3")
                          .arg(sentSize).arg(data.size())
                          .arg(tcpSocket->errorString());
    }
}


void OutputTcpServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)) {
            if(tcpSocket->bytesToWrite())
                qWarning() << QString("Aborting tcp socket while there is still %1 bytesToWrite")
                              .arg(tcpSocket->bytesToWrite());
            tcpSocket->abort();
            tcpSocket->deleteLater();
            tcpSocket = nullptr;
            peerHost.clear();
        }
        if(tcpSocket==nullptr) {
            tcpSocket = pendingConnection;
            peerHost = QHostInfo::fromName(tcpSocket->peerAddress().toString()).hostName();
            connect(tcpSocket, &QTcpSocket::disconnected, this, &OutputTcpServer::socketDisconnected);
            sendStatus();
        } else {
            pendingConnection->abort();
            pendingConnection->deleteLater();
            return;
        }
    }
    if(tcpSocket==nullptr) {
        qDebug()<<"There is no connections.";
    } else {
        qInfo() << QString("Current connection: %1:%2").arg(peerHost).arg(tcpSocket->peerPort());
        writeTimer->start(0);
    }
}

void OutputTcpServer::socketDisconnected()
{
    if(index && tcpSocket->bytesToWrite())
        qWarning() << QString("socket disconnected while there is still %1 bytesToWrite")
                      .arg(tcpSocket->bytesToWrite());
    tcpSocket->deleteLater();
    tcpSocket=nullptr;
    peerHost.clear();
    const QString errStr = "socketDisconnected";
    (index ? qWarning() : qInfo()) << errStr;
    if(index) {
        // Send error only in Run state and if tcp out selected
        if(runEnabled && qTcpServer->isListening())
            emit error(index, errStr);
    }
    sendStatus();
}

void OutputTcpServer::clear()
{
    ringBuf->clear();
    sendMetrics();
    stats->clear();
}

void OutputTcpServer::sendMetrics()
{
    MetricSender::send(QString("out.tcp.%1.bytes").arg(index), stats->bytesSent);
//    MetricSender::send("out.tcp.buffer.used", 1.0*ringBuf->getDataSize() / ringBuf->getAllSize());
    MetricSender::send(QString("out.tcp.%1.latency.mean").arg(index), stats->getMeanLatency());
}

QString OutputTcpServer::getServerName() const
{
    return index ? QString("Output TcpServer #%1").arg(index-1) : "Monitor TcpServer";
}

void OutputTcpServer::sendStatus() const
{
    status.port = tcpPort;
    status.listening = qTcpServer->isListening();
    status.isFree = isFree();
    status.dataPart = 1.*ringBuf->getDataSize()/ringBuf->getAllSize();
    if(status.isFree) {
        status.peer_host.clear();
        status.peer_port = 0;
    } else {
        status.peer_host = tcpSocket->peerAddress();
        status.peer_port = tcpSocket->peerPort();
    }
    emit statusUpdated(status);
}

void OutputTcpServer::startWork()
{
    finishId = 0;
    runEnabled = true;
}

void OutputTcpServer::finishWork(uint fid)
{
    runEnabled = false;
    if(fid==0){
        qWarning() << QString("Got signal to finishWork with incorrect fid=%1").arg(fid);
        emit workFinished(index, fid);
        return;
    }
    if(finishId)
        qWarning() << QString("Got signal to finishWork with fid=%1, but i'm already in finishing state with finishId=%2")
                      .arg(fid).arg(finishId);
    finishId = fid;

    QElapsedTimer timer;
    timer.start();

    bool done=false;
    while(timer.elapsed() > FLUSH_TIMEOUT_MS){
        if(ringBuf->getDataSize()==0 &&
                (!tcpSocket || tcpSocket->bytesToWrite()==0) ) {
            done=true;
            break;
        }
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
    }
    if(done)
        emit workFinished(index, finishId);
}

void OutputTcpServer::setMonitorPortNum(quint16 portNum)
{
    preferedPortNum = portNum;
    if(qTcpServer->isListening()) {
        startListen(index, false);
        startListen(index, true);
    }
    Q_ASSERT(!preferedPortNum || (tcpPort == preferedPortNum));
    if(!index)
        qInfo() << "Monitor portNum was set to:" << tcpPort;
}
