#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "Config.h"
#include "DeviceWidget.h"
#include "DeviceConfig.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "mldiscover/DeviceDescription.h"
#include "mregdev/qtmregdevice.h"
#include "base/AboutBox.h"

#include "QFileDialog"
#include "QDebug"
#include <QInputDialog>
#include "QLabel"
#include "QMessageBox"
#include "QTextCharFormat"
#include "QThread"
#include "QTime"
#include "QPlainTextEdit"
#include "QSplitter"
#include "QSpacerItem"
#include "QShortcut"
#include "QTableWidgetItem"
#include "QScrollBar"

const QString PROGRAM_VERSION = "1.0";
const int DataRoleId = Qt::UserRole;
const int DataRoleSerial = Qt::UserRole+1;

enum { COL_SER_TXT, COL_FW, COL_TEMP, COL_STATUS, COL_TOTAL};

MainWindow::MainWindow(StartupOptions &start_opt, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    config(start_opt),
    discover_dialog(new DiscoverDialog(this)),
    pnp_server(new PNPServer(this))
{
    qRegisterMetaType<DeviceStatus>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DeviceConfig>();
    ui->setupUi(this);
    config.read_config();

    init();
}

MainWindow::~MainWindow()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();

    emit delete_all();
    emit close_program(program_description);
    delete ui;
}

void MainWindow::init()
{
    init_gui();

    configuration_manager = new ConfigurationManager(this, config.program_type, config.default_root);
    connect(configuration_manager, SIGNAL(switch_to_config(QString, QString)), this, SLOT(switch_to(QString, QString)));

    connect(this, SIGNAL(change_program_description(ProgramDescription)), pnp_server, SLOT(changeProgramDescription(ProgramDescription)));
    connect(this, SIGNAL(close_program(ProgramDescription)), pnp_server, SLOT(closeProgram(ProgramDescription)));
    connect(this, SIGNAL(delete_all()), pnp_server, SLOT(deleteLater()));
    ui->actionFind_device->setShortcut(Qt::CTRL + Qt::Key_F);

    discover_dialog->setSingleMode(false);
//    discover_dialog->addDeviceIdFilter(QtMregDevice::DEVICE_ID_U40VE_RC);
    discover_dialog->addDeviceIdFilter(QtMregDevice::DEVICE_ID_MSC16VE_TM);

    update_program_description();

    QVector<DeviceDescription> device_description_list;
    for(DeviceIndex device_index  : config.device_configs.keys()){
        DeviceDescription device_description(device_index);
        device_description_list.append(device_description);
    }
    discover_dialog->setSelected(device_description_list);
}

void MainWindow::init_gui()
{
    set_window_title();

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.endGroup();

    ui->tableWidgetFindDevice->horizontalHeader()->setStretchLastSection(false);

    //--- log
    QPalette Pal(palette());
    Pal.setColor(QPalette::Background, Qt::gray);
    ui->widgetLogButton->setAutoFillBackground(true);
    ui->widgetLogButton->setPalette(Pal);
    ui->pushButtonLog->setChecked(config.show_log);
    ui->pushButtonLog->setText(config.show_log ? "Hide log" : "Show log");
    ui->plainTextEditLog->setHidden(!config.show_log);


    //--- splitter
    QSplitter *splitter = ui->split_horizontal;

    int index_find_device = splitter->indexOf(ui->tableWidgetFindDevice);
    int index_main = splitter->indexOf(ui->widgetMain);
    int index_log = splitter->indexOf(ui->plainTextEditLog);
    int index_log_button = splitter->indexOf(ui->widgetLogButton);

    splitter->setCollapsible(index_find_device, true);
    splitter->setCollapsible(index_log_button, true);

    splitter->setStretchFactor(index_find_device, 1);
    splitter->setStretchFactor(index_main, 30);
    splitter->setStretchFactor(index_log, 1);
    splitter->setStretchFactor(index_log_button, 1);

    for(DeviceIndex device_index : config.device_configs.keys()){
        if(config.device_configs[device_index].enabled){
            create_device_widget(config.device_configs[device_index]);
        }
    }

    populate_devices_table();
}

void MainWindow::create_device_widget(DeviceConfig device_config)
{
    DeviceWidget *device_widget = new DeviceWidget(device_config, this);
    connect(device_widget, SIGNAL(update_device_config(DeviceConfig)), this, SLOT(update_device_config(DeviceConfig)));
    connect(device_widget, SIGNAL(update_status(DeviceStatus)), this, SLOT(update_status(DeviceStatus)));

    connect(ui->actionFind_Baseline, SIGNAL(triggered()), device_widget, SIGNAL(find_base_line()));
    connect(ui->actionSoft_clear, SIGNAL(triggered()), device_widget, SIGNAL(soft_clear()));
    connect(discover_dialog, SIGNAL(deviceDiscovered(DeviceDescription)), device_widget, SIGNAL(device_discovered(DeviceDescription)));
    device_widgets[device_config.device_index] = device_widget;
    ui->layout_main->addWidget(device_widget);
}

void MainWindow::populate_devices_table()
{
    QTableWidget *table = ui->tableWidgetFindDevice;

    table->setColumnCount(COL_TOTAL);

    table->setHorizontalHeaderItem(COL_SER_TXT, new QTableWidgetItem("Serial"));
    table->setHorizontalHeaderItem(COL_FW, new QTableWidgetItem("f/w"));
    table->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("Status"));

    table->setRowCount(0);

    int row_count=0;
    for(DeviceIndex device_index : device_widgets.keys()){
        if(config.device_configs[device_index].enabled){
            ++row_count;
        }
    }

    table->setRowCount(row_count);
    int row_numder = 0;
    for(DeviceIndex device_index : device_widgets.keys()){
        DeviceConfig device_config = device_widgets[device_index]->get_device_config();

        QTableWidgetItem *serial_item = new QTableWidgetItem(device_config.device_index.getSerialStr());
        serial_item->setData(DataRoleId,     device_config.device_index.getDevId());
        serial_item->setData(DataRoleSerial, device_config.device_index.getSerial());
        table->setItem(row_numder, COL_SER_TXT, serial_item);
        table->setItem(row_numder, COL_FW, new QTableWidgetItem());
        table->setItem(row_numder, COL_TEMP, new QTableWidgetItem());
        table->setItem(row_numder, COL_STATUS, new QTableWidgetItem("disconnected"));

        table->item(row_numder, COL_STATUS)->setTextColor(Qt::red);
        table->item(row_numder, COL_SER_TXT)->setTextColor(Qt::red);
        ++row_numder;
    }
    table->resizeColumnsToContents();
    table->resizeRowsToContents();

    update_table_size(table);
}

void MainWindow::set_selected_device()
{
    QVector<DeviceDescription> device_description_list = discover_dialog->getSelectedDevices();

    // remove old config
    for(DeviceIndex device_index : config.device_configs.keys()){
        bool found = false;
        for(int i=0; i<device_description_list.size(); ++i){
            if(device_description_list[i].getIndex() == device_index){
                found = true;
            }
        }
        if(!found) {
            config.device_configs.remove(device_index);
        }
    }

    QMap<DeviceIndex, DeviceConfig> new_device_configs;
    // add new config
    for(int i=0; i<device_description_list.size(); ++i){
        DeviceDescription device_description = device_description_list[i];
        DeviceIndex device_index = device_description.getIndex();
        DeviceConfig device_config;
        device_config.device_index = device_index;
        device_config.enabled = device_description.enabled;

        bool found = false;
        for(DeviceIndex device_index_saved : config.device_configs.keys()){
            if(device_index == device_index_saved){
                found = true;
            }
        }

        if(found){
            config.device_configs[device_index].enabled = device_description.enabled;
            continue;
        }
        config.device_configs[device_index] = device_config;
    }

    // add widget config
    for(DeviceDescription device_description : device_description_list){
        DeviceIndex device_index = device_description.getIndex();

        if(!device_widgets.contains(device_index) && device_description.enabled){
            create_device_widget(config.device_configs[device_index]);
        }
    }

    // remove widget config
    for(DeviceIndex device_index : device_widgets.keys()) {
        if(!config.device_configs.contains(device_index) || !config.device_configs[device_index].enabled) {
            delete  device_widgets[device_index];
            device_widgets.remove(device_index);
        }
    }

    populate_devices_table();
    config.write_config();
}

void MainWindow::update_device_config(DeviceConfig device_config)
{
    config.device_configs[device_config.device_index] = device_config;
    config.write_config();
}

void MainWindow::update_status(DeviceStatus device_status)
{
    QTableWidget *table = ui->tableWidgetFindDevice;
    int row = get_device_row(device_status.index);
    if(row == -1){
        return;
    }

    table->item(row, COL_FW)->setText(device_status.fwStr);
    table->item(row, COL_STATUS)->setText(device_status.status);
    table->item(row, COL_STATUS)->setTextColor(device_status.connected ? QColor() : Qt::red);
    table->item(row, COL_SER_TXT)->setTextColor(device_status.connected ? Qt::darkGreen : Qt::red);
    table->item(row, COL_TEMP)->setText(QString::number(device_status.temp, 'f', 1)
                                        .append(QChar(0x02da)).append('C'));
    table->resizeColumnsToContents();
    update_table_size(table);
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.program_index, config.configuration_name);
}

void MainWindow::switch_to(QString program_index, QString configuration_name)
{
    configuration_manager->hide();

    for(DeviceIndex device_index : device_widgets.keys()){
        delete  device_widgets[device_index];
        device_widgets.remove(device_index);
    }

    config.program_index = program_index;
    config.configuration_name = configuration_name;
    config.read_config();
    init_gui();
}

void MainWindow::on_actionFind_device_triggered()
{
    if(discover_dialog->exec() == QDialog::Accepted){
        set_selected_device();
    }
}

void MainWindow::on_actionSet_hit_width_extension_triggered()
{
    bool ok;
    int val = QInputDialog::getInt(this, "Set hit width", "Hit extension in clocks:", config.hit_extension, 0, 16, 1, &ok);
    if(ok) {
        config.hit_extension = val;
        emit config_updated();
    }
}

void MainWindow::on_actionLut_c_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile("lut.html");
    QString html;

    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
    } else
        html = "<b>There is no lut.c example.</b>";

    QMessageBox::about (this, "lut.c", html);
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::on_pushButtonLog_clicked(bool checked)
{
    ui->plainTextEditLog->setHidden(!checked);
    config.show_log = checked;
    if(checked){
        ui->pushButtonLog->setText("Hide log");
    }
    else{
        ui->pushButtonLog->setText("Show log");
    }
    ui->plainTextEditLog->setHidden(!config.show_log);
    config.write_config();
}

int MainWindow::get_device_row(const DeviceIndex &index)
{
    for(int row=0; row<ui->tableWidgetFindDevice->rowCount(); ++row){
        auto i = get_index_by_row(row);
        if(index == i){
            return row;
        }
    }
    return -1;
}

DeviceIndex MainWindow::get_index_by_row(int row)
{
    auto item = ui->tableWidgetFindDevice->item(row, COL_SER_TXT);
    return DeviceIndex(static_cast<quint16>(item->data(DataRoleId).toUInt()),
                       item->data(DataRoleSerial).toULongLong());
}

void MainWindow::update_program_description()
{
    program_description.options.clear();
    program_description.type = config.program_type;
    program_description.index = config.program_index;
    emit change_program_description(program_description);
}

void MainWindow::add_log(QtMsgType severity, const QString &log, const QString &thrName)
{
    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<<"";
    }
    QTextCharFormat tf = ui->plainTextEditLog->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEditLog->setCurrentCharFormat(tf);
    QString str = log;
    if(!thrName.isEmpty())
        str.prepend(QString("[%1] ").arg(thrName));
    str.prepend(QString("[%1] ").arg(QTime::currentTime().toString()));
    ui->plainTextEditLog->appendPlainText(str);
}

void MainWindow::set_window_title()
{
    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(QCoreApplication::applicationName()).arg(config.program_index).arg(config.configuration_name));
}

void MainWindow::update_table_size(QTableWidget *t)
{
    int width=t->verticalHeader()->width();
    for(int i=0;i<t->columnCount();++i){
        if(t->isColumnHidden(i)) continue;
        width += t->columnWidth(i);
    }
    width += 5;
    if(t->verticalScrollBar()->isVisible())
        width += t->verticalScrollBar()->width();
    t->setMinimumWidth(width);
    t->updateGeometry();

    int height=t->horizontalHeader()->height();
    for(int i=0;i<t->rowCount();++i){
        if(t->isRowHidden(i)) continue;
        height += t->rowHeight(i);
    }
    height += 5;
    if(t->horizontalScrollBar()->isVisible())
        height += t->horizontalScrollBar()->height();
    t->setMinimumHeight(height);
}
