#include "PNPDiscoverDialog.h"
#include "ui_PNPDiscoverDialog.h"
#include "pnp-server/PNPServer.h"
#include "base/DnsLookup.h"

#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QSettings>
#include <QTimer>
#include <QUuid>

namespace {

enum {COL_ID, COL_UUID, COL_NAME, COL_TYPE, COL_INDEX,
      COL_INTERFACES, COL_IP, COL_LEN, COL_SEQ, COL_REFRESH_TIME, COL_TOTAL};
const int UPDATE_TIME_MS = 1000;
}

PNPDiscoverDialog::PNPDiscoverDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::PNPDiscoverDialog),
    pnpServer(new PNPServer(this)),
    dnsLookup(new DnsLookup(this)),
    hostFilter(0),
    updateTimer(new QTimer(this))
{
    ui->setupUi(this);
    connect(pnpServer, SIGNAL(programClosed(ProgramDescription)),
            SLOT(programClosed(ProgramDescription)));
    connect(pnpServer, SIGNAL(programDiscovered(ProgramDescription,int)),
            SLOT(programDiscovered(ProgramDescription,int)));

    connect(dnsLookup, &DnsLookup::reverseLookupComplete,
            this, &PNPDiscoverDialog::lookedUp);

    connect(updateTimer, SIGNAL(timeout()), SLOT(updateTable()));
    updateTimer->setSingleShot(false);
    updateTimer->start(UPDATE_TIME_MS);

    QSettings s;
    s.beginGroup("Main");
    typeFilter = s.value("typeFilter").toString();
    hostFilter = s.value("hostFilter", hostFilter).toUInt();
    s.endGroup();
    doGroupLogic = true;
    groups_count_changed_flag = true;

    ui->comboBoxType->blockSignals(true);
    ui->comboBoxType->addItem("All", "");
    ui->comboBoxType->insertSeparator(1);
    if(!typeFilter.isEmpty()) {
        ui->comboBoxType->insertItem(2, typeFilter, typeFilter);
        ui->comboBoxType->setCurrentIndex(2);
    }
    ui->comboBoxType->blockSignals(false);

    ui->comboBoxHost->blockSignals(true);
    curHostList[0] = "All";
    ui->comboBoxHost->addItem("All", 0);
    ui->comboBoxHost->insertSeparator(1);
    if(hostFilter) {
        curHostList[hostFilter] = QHostAddress(hostFilter).toString();
        ui->comboBoxHost->addItem(curHostList[hostFilter], hostFilter);
        ui->comboBoxHost->setCurrentIndex(2);
    }
    ui->comboBoxHost->blockSignals(false);
    s.beginGroup("MainWindow");
    restoreGeometry(s.value("geometry").toByteArray());
    s.endGroup();
}

void PNPDiscoverDialog::done(int r)
{
    QSettings s;
    s.beginGroup("MainWindow");
    s.setValue("geometry", saveGeometry());
    s.endGroup();

    QDialog::done(r);
}

PNPDiscoverDialog::~PNPDiscoverDialog()
{
    QSettings s;
    s.beginGroup("MainWindow");
    s.setValue("geometry", saveGeometry());
    s.endGroup();

    delete ui;
}

void PNPDiscoverDialog::init(QStringList args)
{
    if(args.size() > 1){
        searchTargets = args;
        searchTargets.removeAt(0);
        QString targStr = searchTargets.join(" ");
        if(targStr.length()>20)
            targStr = targStr.left(17).append("...");

        setWindowTitle(QString("%1 (%2)")
                       .arg(QCoreApplication::applicationName())
                       .arg(targStr));
    }

    setupTable();
    pnpServer->startSearch(searchTargets, true);
}

void PNPDiscoverDialog::programClosed(const ProgramDescription &descr)
{
    if(programs.contains(descr.uuid)) {
        int row = getRowNumber(descr);
        if(row == -1){
            qWarning()<<"row == -1";
        } else {
            ui->tableWidget->removeRow(row);
        }
        programs.remove(descr.uuid);
        updateTableSize();
        doGroupLogic = true;
    } else {
        qWarning()<<"Reached closing signal from unknown program:"<<descr.toString();
    }
}

void PNPDiscoverDialog::programDiscovered(const ProgramDescription &descr, int packetSize)
{
    if(!programs.contains(descr.uuid)) {
        foreach (const ProgramDescription old, programs.values()) {
            if(old.sameProgram(descr)) {
                qWarning()<<"Old program with same type/index:";
                qWarning()<<"Old:"<<old.toString();
                qWarning()<<"New:"<<descr.toString();
                programClosed(old);
                break;
            }
        }
    }

    progPacketSize[descr.uuid] = packetSize;
    if(!programs.contains(descr.uuid) || programs[descr.uuid].options != descr.options){
        doGroupLogic = true;
    }
    programs[descr.uuid] = descr;
    const quint32 hostIp = descr.getHostIp();
    const QString hostName = descr.getHostName();
    if(fitTypeFilter(descr.type)) {
        int hostItem = ui->comboBoxHost->findData(hostIp);
        if(hostItem==-1) {
            curHostList[hostIp] = hostName;
            ui->comboBoxHost->insertItem(getIndexOfNewHost(hostName), hostName, hostIp);
        } else if(hostName != curHostList[hostIp]){
            curHostList[hostIp] = hostName;
            ui->comboBoxHost->setItemText(hostItem, hostName);
        }
    }

    pendingToUpdate << descr.uuid;
}

void PNPDiscoverDialog::setupTable()
{
    QTableWidget *table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);

    table->setHorizontalHeaderItem(COL_ID, new QTableWidgetItem("id"));
    table->setHorizontalHeaderItem(COL_UUID, new QTableWidgetItem("UUID"));
    table->setHorizontalHeaderItem(COL_NAME, new QTableWidgetItem("Name"));
    table->setHorizontalHeaderItem(COL_TYPE, new QTableWidgetItem("Type"));
    table->setHorizontalHeaderItem(COL_INDEX, new QTableWidgetItem("Index"));
    table->setHorizontalHeaderItem(COL_INTERFACES, new QTableWidgetItem("Interfaces"));
    table->setHorizontalHeaderItem(COL_IP, new QTableWidgetItem("Ip"));
    table->setHorizontalHeaderItem(COL_LEN, new QTableWidgetItem("Len"));
    table->setHorizontalHeaderItem(COL_SEQ, new QTableWidgetItem("Seq"));
    table->setHorizontalHeaderItem(COL_REFRESH_TIME, new QTableWidgetItem("Refresh time"));
    hideColumns();
}

int PNPDiscoverDialog::getRowNumber(const ProgramDescription &descr)
{
    QTableWidget *table = ui->tableWidget;
    int rows= table->rowCount();
    QString uuidStr = descr.uuid.toString();
    for(int row=0; row<rows; ++row){
        if(uuidStr == table->item(row, COL_UUID)->text())
            return row;
    }
    return -1;
}

void PNPDiscoverDialog::updateInfo(int row)
{
    QString txtStr;
    if(row!=-1) {
        const QString uuidStr = ui->tableWidget->item(row, COL_UUID)->text();
        const QUuid uuid(uuidStr);
        foreach (const ProgramDescription &descr, programs) {
            if(descr.uuid == uuid){
                QStringList lines;
                lines << QString("Name:%1").arg(descr.name);
                if(!descr.ver_hash.isEmpty())
                    lines << QString("Version hash: %1").arg(descr.ver_hash);
                if(!descr.ver_date.isEmpty())
                    lines << QString("Version date: %1").arg(descr.ver_date);
                if(descr.options.size()){
                    lines << QString("Options:");
                    foreach (QString key, descr.options.keys()) {
                        lines << QString("    %1: %2").arg(key)
                                 .arg(descr.options[key]);
                    }
                }
                if(descr.interfaces.size()){
                    lines << QString("Interfaces:");
                    foreach (const ProgramInterface &i, descr.interfaces) {
                        QStringList peersStrList;
                        foreach (const ProgramInterfacePeer &peer, i.peers) {
                            dnsLookup->reverseLookup(peer.host.toString());
                            const quint32 connIp = peer.host.toIPv4Address();
                            peersStrList << (lookUpMap.contains(connIp) ?
                                                 QString("%1:%2").arg(lookUpMap[connIp]).arg(peer.port)
                                               : peer.toString());
                        }
                        lines << QString("    %1 #%2 %3:%4 (%5)")
                                 .arg(i.type).arg(i.id).arg(i.host.toString())
                                 .arg(i.port)
                                 .arg(i.isFree?"free": peersStrList.join(";"));
                    }
                }
                txtStr = lines.join('\n');
            }
        }
        Q_ASSERT(!txtStr.isEmpty());
    }
    if(ui->textEdit->toPlainText() != txtStr) {
        ui->textEdit->clear();
        ui->textEdit->setText(txtStr);
    }
}

void PNPDiscoverDialog::updateTableItem(const ProgramDescription &descr, int packetSize, const QDateTime &currentDateTime)
{
    QTableWidget *table = ui->tableWidget;
    if(table->updatesEnabled()) {
        table->setSortingEnabled(false);
    }
    int row=getRowNumber(descr);

    if(row!=-1) {
        if(table->item(row,COL_NAME)->text() != descr.name)
            table->item(row,COL_NAME)->setText(descr.name);
        if(table->item(row,COL_TYPE)->text() != descr.type)
        table->item(row,COL_TYPE)->setText(descr.type);
        if(table->item(row,COL_INDEX)->text() != descr.index)
        table->item(row,COL_INDEX)->setText(descr.index);

//        if(descr.interfaces.size()){
//            QStringList iStr;
//            foreach (const ProgramInterface &i, descr.interfaces) {
//                iStr << QString("%1 #%2 %3:%4 %5")
//                        .arg(i.type).arg(i.id).arg(i.host.toString())
//                        .arg(i.port).arg(i.isFree?"free":"busy");
//            }
//            table->item(row, COL_INTERFACES)->setText(iStr.join('\n'));
//        } else {
//            table->item(row, COL_INTERFACES)->setText("");
//        }
        if(table->item(row,COL_IP)->text() != descr.getHostName())
        table->item(row, COL_IP)->setText(descr.getHostName());
        if(table->item(row,COL_IP)->toolTip() != descr.host.toString())
        table->item(row, COL_IP)->setToolTip(descr.host.toString());
        if(table->item(row,COL_LEN)->text() != QString::number(packetSize))
        table->item(row, COL_LEN)->setText(QString::number(packetSize));
        if(table->item(row,COL_SEQ)->text() != QString::number(descr.seq))
        table->item(row, COL_SEQ)->setText(QString::number(descr.seq));
        if(table->item(row,COL_REFRESH_TIME)->text() != currentDateTime.toString())
        table->item(row, COL_REFRESH_TIME)->setText(currentDateTime.toString());
    } else {
        populateTypeFilter(descr);
        row=ui->tableWidget->rowCount();
        table->setRowCount(row+1);
        table->setItem(row, COL_ID, new QTableWidgetItem(QString::number(row)));
        table->setItem(row, COL_UUID, new QTableWidgetItem(descr.uuid.toString()));
        table->setItem(row, COL_NAME, new QTableWidgetItem(descr.name));
        table->setItem(row, COL_TYPE, new QTableWidgetItem(descr.type));
        table->setItem(row, COL_INDEX, new QTableWidgetItem(descr.index));
        table->setItem(row, COL_INTERFACES, new QTableWidgetItem());
        table->setItem(row, COL_IP, new QTableWidgetItem(descr.getHostName()));
        table->item(row, COL_IP)->setData(Qt::UserRole, descr.getHostIp());
        table->setItem(row, COL_LEN, new QTableWidgetItem(QString::number(packetSize)));
        table->setItem(row, COL_SEQ, new QTableWidgetItem(QString::number(descr.seq)));
        if(descr.interfaces.size()){
            QStringList iStr;
            foreach (const ProgramInterface &i, descr.interfaces) {
                iStr << QString("%1 #%2 %3:%4 %5")
                        .arg(i.type).arg(i.id).arg(i.host.toString())
                        .arg(i.port).arg(i.isFree?"free":"busy");
            }
            table->item(row, COL_INTERFACES)->setText(iStr.join('\n'));
        }
        table->setItem(row, COL_REFRESH_TIME,
                       new QTableWidgetItem(currentDateTime.toString()));
        table->setRowHidden(row, !fitFilter(descr));
    }

    int curRow = table->currentRow();
    if(curRow<0 && fitFilter(descr)) {
        table->setCurrentCell(row, COL_ID);
    } else if(curRow == row) {
        updateInfo(row);
    }
    if(table->updatesEnabled()) {
        updateTableSize();
        table->setSortingEnabled(true);
    }
}

void PNPDiscoverDialog::updateTableSize()
{
//    hideColumns();
    QTableWidget *t = ui->tableWidget;
    t->updateGeometry();
    int width=t->verticalHeader()->width()+2;
    for(int col=0; col<t->columnCount(); ++col){
        if(!t->isColumnHidden(col))
            width += t->columnWidth(col);
    }
    int height=t->horizontalHeader()->height()+2+40;
    for(int row=0; row<t->rowCount(); ++row){
        if(!t->isRowHidden(row))
            height += t->rowHeight(row);
    }
    height = qMin(height, 600);
    t->setMinimumSize(width, height);
}

bool PNPDiscoverDialog::fitFilter(const ProgramDescription &pd)
{
    return  fitTypeFilter(pd.type) && fitHostFilter(pd.getHostIp())  && fitGroupFilter(pd.name)  && fitMyGroupFilter(pd.name);
}

bool PNPDiscoverDialog::fitFilter(int row)
{
    QTableWidget *t = ui->tableWidget;
    return fitTypeFilter(t->item(row, COL_TYPE)->text()) && fitHostFilter(t->item(row, COL_IP)->data(Qt::UserRole).toUInt()) &&
           fitGroupFilter(t->item(row, COL_NAME)->text())  && fitMyGroupFilter(t->item(row, COL_NAME)->text());
}

bool PNPDiscoverDialog::fitTypeFilter(const QString &type)
{
    return typeFilter.isEmpty() || type.contains(typeFilter, Qt::CaseInsensitive);
}

bool PNPDiscoverDialog::fitHostFilter(const quint32 &hostIp)
{
    return hostFilter==0 || hostIp==hostFilter;
}

void PNPDiscoverDialog::updateAvailableCnt()
{
    QTableWidget *t = ui->tableWidget;
    int cnt=0;
    for(int r=0; r<t->rowCount(); ++r)
        if(!t->isRowHidden(r))
            ++cnt;
    ui->labelAvailable->setText(QString::number(cnt));
}

int PNPDiscoverDialog::getIndexOfNewHost(const QString &h)
{
    QStringList list = curHostList.values();
    list << h;
    qSort(list);
    return list.indexOf(h)+1; // +1 - for separator
}

void PNPDiscoverDialog::populateTypeFilter(const ProgramDescription &pd)
{
    if(ui->comboBoxType->findData(pd.type) != -1)
        return;

    QStringList typeList;
    typeList << pd.type;
    for(int i=ui->comboBoxType->count()-1; i>1; --i) {
        typeList << ui->comboBoxType->itemData(i).toString();
    }
    qSort(typeList);

    int index = typeList.indexOf(pd.type)+2;
    ui->comboBoxType->insertItem(index, pd.type , pd.type);
}

void PNPDiscoverDialog::hideColumns()
{
    QTableWidget *t = ui->tableWidget;
    t->hideColumn(COL_ID);
    t->hideColumn(COL_UUID);
    t->hideColumn(COL_INTERFACES);
}

void PNPDiscoverDialog::on_pushButtonRefresh_clicked()
{
    while(ui->tableWidget->rowCount())
        ui->tableWidget->removeRow(0);
    programs.clear();
    groups_count_changed_flag = true;
    pnpServer->startSearch(searchTargets, true);
    updateTableSize();
}

void PNPDiscoverDialog::on_pushButtonClear_clicked()
{
    while(ui->tableWidget->rowCount())
        ui->tableWidget->removeRow(0);
    programs.clear();
    updateTableSize();
}

void PNPDiscoverDialog::on_comboBoxType_currentIndexChanged(int index)
{
    typeFilter = ui->comboBoxType->itemData(index).toString();
    QSettings s;
    s.beginGroup("Main");
    s.setValue("typeFilter", typeFilter);
    s.endGroup();

    QTableWidget *table = ui->tableWidget;
    QMap<quint32, QString> hostList;
    hostList[0] = curHostList[0];
    foreach (const ProgramDescription &pd, programs) {
        if(fitTypeFilter(pd.type))
            hostList[pd.getHostIp()] = pd.getHostName();
    }

    if(curHostList != hostList) {
        // remove exess items
        foreach (quint32 ip, curHostList.keys()) {
            if(!hostList.contains(ip)) {
                int index = ui->comboBoxHost->findData(ip);
                if(index != -1) {
                    ui->comboBoxHost->removeItem(index);
                    curHostList.remove(ip);
                }
            }
        }

        foreach (quint32 ip, hostList.keys()) {
            const QString &newHostName = hostList[ip];
            if(curHostList.contains(ip)) {
                if(curHostList[ip] != newHostName){
                    int index = ui->comboBoxHost->findData(ip);
                    if(index != -1) {
                        ui->comboBoxHost->setItemText(index, newHostName);
                        curHostList[ip] = newHostName;
                    }
                }
            } else {
                ui->comboBoxHost->insertItem(getIndexOfNewHost(newHostName), newHostName, ip);
                curHostList[ip] = hostList[ip];
            }
        }
    }


    for(int row = ui->tableWidget->rowCount()-1; row>=0; --row) {
        if(fitFilter(row)) {
            table->setRowHidden(row, false);
            ui->tableWidget->resizeRowToContents(row);
            ui->tableWidget->resizeColumnsToContents();
        } else
            table->setRowHidden(row, true);
     }

     set_first_cell_current();
     updateTableSize();
     table->updateGeometry();
}

void PNPDiscoverDialog::on_comboBoxHost_currentIndexChanged(int index)
{
    hostFilter = ui->comboBoxHost->itemData(index).toUInt();
    QSettings s;
    s.beginGroup("Main");
    s.setValue("hostFilter", hostFilter);
    s.endGroup();

    QTableWidget *table = ui->tableWidget;
    for(int row = ui->tableWidget->rowCount()-1; row>=0; --row) {
        table->setRowHidden(row, !fitFilter(row));
    }

    set_first_cell_current();
    updateTableSize();
}

void PNPDiscoverDialog::set_first_cell_current()
{
    int first_row =0;
    for(int row=0; row <= ui->tableWidget->rowCount()-1; ++row){
        if(fitFilter(row)){
            first_row = row;
            break;
        }
    }
    ui->tableWidget->setCurrentCell(first_row,0);
}

void PNPDiscoverDialog::on_pushButtonPrintCurrent_clicked()
{
    QTableWidget *table = ui->tableWidget;
    int row = table->currentRow();
    if(row>=0) {
        const QString uuidStr = ui->tableWidget->item(row, COL_UUID)->text();
        const QUuid uuid(uuidStr);
        foreach (const ProgramDescription &descr, programs) {
            if(descr.uuid == uuid) {
                foreach (QString s, descr.getAnnounceStr().split("\n")) {
                    qInfo().noquote() << s;
                }
                return;
            }
        }
    }
}

void PNPDiscoverDialog::on_tableWidget_currentCellChanged(int currentRow, int, int, int)
{
    updateInfo(currentRow);
    currentRowNumber = currentRow;
}

void PNPDiscoverDialog::updateTable()
{
    if(!pendingToUpdate.isEmpty()) {
        QTableWidget *table = ui->tableWidget;
        QDateTime currentDateTime = QDateTime::currentDateTime();
        table->setUpdatesEnabled(false);
        table->setSortingEnabled(false);
        foreach (const QUuid &id, pendingToUpdate) {
            if(programs.contains(id))
                updateTableItem(programs[id], progPacketSize[id], currentDateTime);
        }
        updateTableSize();
        table->setSortingEnabled(true);
        table->setUpdatesEnabled(true);
        pendingToUpdate.clear();
        updateAvailableCnt();
        ui->tableWidget->resizeColumnsToContents();
    }
    //---Do graph logic, if new some program discovered or closed
    if(doGroupLogic) {
        graphMatrixUpdate();
        graphAdjacencyListUpdate();
        graphAdjacencyListComponents();
        if (groups_count_changed_flag){
            populateGroupFilter();
            groups_count_changed_flag = false;
        }
        doGroupLogic = true;
    }
}

void PNPDiscoverDialog::lookedUp(const QHostInfo &i)
{
    foreach (const QHostAddress &addr, i.addresses()) {
        const quint32 ip = addr.toIPv4Address();
        if(!lookUpMap.contains(ip) || lookUpMap[ip] != i.hostName()) {
            lookUpMap[ip] = i.hostName();
            updateInfo(ui->tableWidget->currentRow());
        }
    }
}

//---Fill the programms map and the oriented graph matching matrix---
void PNPDiscoverDialog::graphMatrixUpdate()
{
    graphMap.clear();
    component.clear();
    vector_of_graphMatrix.clear();
    graphMatrix.clear();

    //---programms map---
    int i=0;
    foreach (const ProgramDescription &descr, programs) {
        QString name =  descr.name;
        name.remove(QRegExp("[#]"));
        graphMap.insert(i,name);
        ++i;
    }
    //---oriented graph matching matrix---
    QVector <bool> tempVector;
    int maxN = graphMap.size();
    foreach (const ProgramDescription &descr, programs) {
        QString name =  descr.name;
        name.remove(QRegExp("[#]"));
        for(int i=0; i<maxN ; ++i) tempVector.append(false);
        if(descr.options.size()){
            foreach (QString key, descr.options.keys()){
                if (key.contains("client-", Qt::CaseInsensitive)){
                    QString client = descr.options[key].section(",",0,0);
                    client= client.simplified();
                    client= client.remove(QRegExp("[#]"));
                    for(int i=0; i< maxN; ++i){
                        if (client == graphMap.value(i)){
                            tempVector[i] = true;
                        }
                    }
                }
            }
        }
        graphMatrix.append(tempVector);
        tempVector.clear();
    }
//     matrix_debug_printing(maxN, graphMatrix);// ----- do it, if you want to see the result matrix
}

//-----Fill the non-oriented graph matching matrix and form the adjacency list------------------
void PNPDiscoverDialog::graphAdjacencyListUpdate()
{
    //-----non-oriented graph matching matrix-------
    QVector <QVector<bool>> graphMatrix_no_orient;
    QVector <bool> tempVector_no_orient;
    for(int i=0; i<graphMap.size(); ++i){
        for(int j=0; j<graphMap.size(); ++j){
                tempVector_no_orient.append(false);
        }
        graphMatrix_no_orient.append(tempVector_no_orient);
        tempVector_no_orient.clear();
    }


    for(int i=0; i<graphMap.size(); ++i){
        for(int j=0; j<graphMap.size(); ++j){
            if (graphMatrix[i][j] == true){
                graphMatrix_no_orient[i][j] = true;
                graphMatrix_no_orient[j][i] = true;
            }
        }
    }
    //   matrix_debug_printing(maxN, &graphMatrix_no_orient);----- do it, if you want to see the matrix

    //----form the adjacency list----------------
//    vector_of_graphMatrix.clear();
    QVector <int> tempVector2;
    for(int i=0; i<graphMap.size(); ++i){
        for (int j=0; j<graphMap.size(); ++j){
            if (graphMatrix_no_orient[i][j] == true){
                tempVector2.append(j);
            }
        }
        vector_of_graphMatrix.append(tempVector2);
        if (tempVector2.size()){
            tempVector2.clear();
        }
    }
}

//---Break the list of contiguity into components---------
void PNPDiscoverDialog::graphAdjacencyListComponents()
{
    int n_max = graphMap.size();
    int n_comp=0;
    QVector <int> component_temp;
    component.clear();
    QVector <bool> used;
    for(int i=0; i<n_max; ++i)
        used.append(false);
    for(int i=0; i<n_max; ++i){
        if(!used[i]){
            component_temp.clear();
            traversing_graph_in_depth(i, &used, &component_temp);
            ++n_comp;
            component.append(component_temp);
        }
    }
    int groups_count_current=0;
    for(int i=0; i<component.size(); ++i){
         if (component[i].size()>1){
             ++groups_count_current;
         }
    }
    if (groups_count_current != groups_count){
        groups_count_changed_flag = true;
        groups_count = groups_count_current;
    }



}

//----Algorithm for traversing the graph in depth----------------------
void PNPDiscoverDialog::traversing_graph_in_depth(int v, QVector <bool> *used, QVector <int> *component_temp)
{
    used->operator[](v) = true;
    component_temp->append(v);
    for(int i=0; i<vector_of_graphMatrix[v].size(); ++i){
        int to=vector_of_graphMatrix[v][i];
        if(!used->at(to))
            traversing_graph_in_depth(to, used, component_temp);
    }
}

//----We will populate group filter if the groups_count_changed_flag is true----
void PNPDiscoverDialog::populateGroupFilter()
{
    groups.clear();
    groups_test.clear();
    ui->comboBoxGroup->clear();
    ui->comboBoxGroup->addItem("All", "");
    ui->comboBoxGroup->insertSeparator(1);
    int index=2;
    for(int i=0; i<component.size(); ++i){
        if (component[i].size()>1){
            QString group_name = "Group " + QString::number(index-1);
            ui->comboBoxGroup->insertItem(index, group_name);
            groups.insert(group_name , i);

            for(int j=0; j<component[i].size(); j++){
                int graphMapKey = component[i][j];
                QString tmp = graphMap.value(graphMapKey);
                groups_test[group_name].insert(j, tmp);
            }
            ++index;
        }
        else{
            QString group_name = "Programs without group";
            groups_test[group_name].append(graphMap.value(component[i][0]));
        }
    }

    ui->comboBoxGroup->insertSeparator(ui->comboBoxGroup->count());
    ui->comboBoxGroup->addItem("Programs without group", "");
}

bool PNPDiscoverDialog::fitMyGroupFilter(const QString &name)
{
    QString tmp = name;
    tmp.remove(QRegExp("[#]"));
    return myGroupFilter.isEmpty() || myGroupFilter.contains(tmp, Qt::CaseInsensitive);
}

//----To show only my group clicked logic-------------------
void PNPDiscoverDialog::on_checkBox_clicked(bool checked)
{
    myGroupFilter.clear();

    if (checked){
        QString current_prg_name = ui->tableWidget->item(currentRowNumber, COL_NAME)->text();
        current_prg_name.remove(QRegExp("[#]"));
        QString current_group_name;
        for(QString str:groups_test.keys()){
            for(int i=0; i<groups_test.value(str).size(); ++i){
                if(current_prg_name == groups_test.value(str)[i]){
                    current_group_name = str;
                }
            }
        }
        for(int i=0; i<groups_test[current_group_name].size(); ++i){
           myGroupFilter.append(groups_test[current_group_name][i]);
        }

        for (int index = 0; index < ui->comboBoxGroup->count(); index++){
            if(current_group_name == ui->comboBoxGroup->itemText(index)){
                ui->comboBoxGroup->setCurrentIndex(index);
            }
        }
    }

    QTableWidget *table = ui->tableWidget;
    for(int row = ui->tableWidget->rowCount()-1; row>=0; --row) {
        if(fitFilter(row)) {
            table->setRowHidden(row, false);
            ui->tableWidget->resizeRowToContents(row);
            ui->tableWidget->resizeColumnsToContents();
        } else
            table->setRowHidden(row, true);
     }
     set_first_cell_current();
     updateTableSize();
     table->updateGeometry();
}

bool PNPDiscoverDialog::fitGroupFilter(const QString &name)
{
    QString tmp = name;
    tmp.remove(QRegExp("[#]"));
    return groupFilter.isEmpty() || groupFilter.contains(tmp, Qt::CaseInsensitive);
}

//---Groups combo box logic---
void PNPDiscoverDialog::on_comboBoxGroup_currentIndexChanged(int index)
{
    groupFilter.clear();
    myGroupFilter.clear();
    if(index == 0){
        ui->checkBox->setChecked(false);
    }
    if (index != -1 && index !=0){
        if (index == (ui->comboBoxGroup->count()-1)){
            for(int i=0; i< component.size(); ++i){
                if (component[i].size() == 1){
                    int graphMapKey = component[i][0];
                    QString tmp = graphMap.value(graphMapKey);
                    groupFilter.append(tmp);
                }
            }
        }
        else {
            QString key = "Group " +  QString::number(index-1);
            for(int i=0; i< component[groups.value(key)].size(); ++i){
                int graphMapKey = component[groups.value(key)][i];
                QString tmp = graphMap.value(graphMapKey);
                groupFilter.append(tmp);
            }
        }
    }

    QTableWidget *table = ui->tableWidget;
    for(int row = ui->tableWidget->rowCount()-1; row>=0; --row) {
        if(fitFilter(row)) {
            table->setRowHidden(row, false);
            ui->tableWidget->resizeRowToContents(row);
            ui->tableWidget->resizeColumnsToContents();
        } else
            table->setRowHidden(row, true);
     }
     set_first_cell_current();
     updateTableSize();
     table->updateGeometry();
}

//-----debug matrix printing, useful only in debugging mode------------
void PNPDiscoverDialog::matrix_debug_printing(int maxN, QVector<QVector<bool>> matrix_to_print)
{
    QString str_tmp = "      ";
    for(int i=0; i< maxN ; ++i){
        if (i<=9) str_tmp = str_tmp  + QString::number(i) + "  ";
        else str_tmp = str_tmp  + QString::number(i) + " ";
    }
    QString str_i;
    for(int i=0; i<maxN ; ++i){
        for(int j=0; j<maxN ; ++j)
            str_i = str_i + "  " + QString("%1").arg(matrix_to_print.at(i).at(j));
        if (i<=9) qInfo() << i << "  " << qPrintable(str_i);
        else qInfo() << i << " " << qPrintable(str_i);
        str_i = "";
    }
    qInfo()<< "==========================================";
}


