//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "Config.h"
#include "pnp-server/PNPMessage.h"
#include "mldiscover/DeviceDescription.h"
#include "mongo/StartupOptions.h"
#include "mongo/RootConfig.h"
#include "configuration-manager/ConfigurationManager.h"
#include "TrigDevice.h"

#include <QMainWindow>
#include <QMap>
#include <QTime>

class QHostAddress;
class QLabel;
class QTimer;
class QThread;
class RootConfig;
class DiscoverDialog;
class PNPServer;
class TrigDevice;
class TrigStatus;
class SlowControlServer;
class RemoteControlServer;
class RemoteControlServerStatus;
class EvNumChecker;

namespace Ui {
class MainWindow;
}


class MainWindow : public QMainWindow
{
    Q_OBJECT
public:
    enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_VME_TRIGS=2, TAB_LOGS=3};

    explicit MainWindow(StartupOptions &startup_options, QWidget *parent = nullptr);
    ~MainWindow();

public Q_SLOTS:
    void addLog(QtMsgType severity, const QString &log, const QString &thrName);

private Q_SLOTS:
    void init();
    void on_actionQuit_triggered();
    void on_actionAbout_triggered();

    void trigStatusUpdated(const TrigStatus &status);
    void setTrigStatus(QString);

    void onRunError(QString runIndex, int runNumber, QString errStr);
    void onRunWarning(QString runIndex, int runNumber, QString warnStr);
    void onBackPressure(QString runIndex, int runNumber, bool backPressure);
    void onVmeFlagUpdated();
    void on_pushButtonRestartAll_clicked();
    void on_pushButtonStopAll_clicked();
    void on_actionFindDevice_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_actionEnableStatusDump_toggled(bool val);
    void on_actionSpillPlot100_ms_triggered();
    void on_actionSpillPlot10_ms_triggered();
    void on_actionSpillPlot1_ms_triggered();
    void on_actionSpillPlot350_us_triggered();
    void on_actionSpillPlot100_us_triggered();

    void on_actionTimeHist_1_ms_triggered();
    void on_actionTimeHist_100_us_triggered();
    void on_actionTimeHist_10_us_triggered();
    void on_actionTimeHist_1_us_triggered();
    void on_actionTimeHist_100_ns_triggered();
    void on_actionSetStatusDumpDirectory_triggered();

    void on_tabWidget_currentChanged(int index);

    void findDevice();
    void onRemoteStartRequest(QString globIndex, int runNumber, quint64 startEvNumber=0);
    void onRemoteStopRequest(QString globIndex, int runNumber, quint64 stopEvNumber=0);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);
    void onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse);
    void setClientTabState(bool isOk);
    void setTabState(bool isOk, int tabIndex);
    void blockControl(bool block);

    void on_actionConfiguration_manager_triggered();
    void switch_to(QString program_index, QString configuration_name);

    void config_transfer_from_clients(int id, RootConfig root);
    void config_timeout();
    void client_config_receiver(QMap<int, ClientConfig> client_config_);


signals:
    void runStart(QString runIndex, int runNumber) const;
    void runStop() const;
    void deleteHardware(const DeviceIndex &) const;
    void destroyAll() const;
    void changeProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void slowControlStateUpdated(QString, QString) const;
    void allowSlowControl(bool) const;
    void allowRemoteControl(bool) const;
    void sendRemoteState(QString,bool) const;
    void sendRunError(QString, int, QString) const;
    void sendRunWarn(QString, int, QString) const;
    void sendBackPressure(QString, int, bool) const;
    void sendVmeFlag(bool) const;
    void deleteAll() const;
    void deviceDiscovered(DeviceDescription) const;
    void remoteStartCompleted(bool success, QString runIndex, int runNumber);
    void remoteStopCompleted(bool success, QString runIndex, int runNumber);
    void spillPlotStepUpdated() const;
    void sendControlEnable(bool block) const;
    void setStatusDumpDirName(const QString &);
    void setStatusDumpEnable(bool);

    void config_transfer_to_rc(QByteArray data);

protected:
    virtual void closeEvent(QCloseEvent *event);

private:
    bool perfomeStart(QString runIndex, int runNumber);
    bool perfomeStop(bool isError);
    void updateDeviceLabel();
    int addClient(QHostAddress host, int port, int trig, int clientIndex=-1);
    void loadSettings();
    void connectToTrig(DeviceDescription dd=DeviceDescription());
    void initProgramDescription();
    void updateProgramDescription();
    void updateStatStopButton();
    bool isReadForRun() const;
    void updateSpillPlotStateMenu();
    void updateTimeHistogramStateMenu();
    void updateGroupBoxWarn();
    void updateRemoteBlocking();

    void run_root_prepare();
    void clearRunWarning();

    Ui::MainWindow *ui;
    ConfigurationManager *configuration_manager;
    Config config;
    QLabel * const labelState;
    QLabel * const labelSlowControl;
    QLabel * const labelDevice;
    QLabel * const labelFirmware;
    QLabel * const labelTemp;
    DiscoverDialog * const discoverDialog;
    PNPServer * const pnpServer;
    QTime curIpLastDiscover;
    TrigDevice * trigDevice;
    QThread * const trigDevThread;
    SlowControlServer * const slowControl;
    QThread * const slowControlThread;
    RemoteControlServer * const remoteControl;
    QThread * const remoteControlThread;
    ProgramDescription progDesc;
    bool remoteConnected;
    QMap<int, bool> tabOk;
    EvNumChecker * const evNumChecker;
    bool hasRunError;
    TrigStatus lastTrigStatus;

    QTimer * config_timer;
    QMap<int, RootConfig> clients_config;
    bool megamaster_flg;
};

#endif // MAINWINDOW_H
