#include "SlowControlServer.h"

#include <QTcpServer>
#include <QtNetwork>
#include <QJsonObject>
#include <QJsonDocument>

#include "TrigDevice.h"
#include "QTimer"

namespace {
const quint16 SLOW_CONTROL_PORT = 33306;
}

SlowControlServer::SlowControlServer() :
    QObject(nullptr),
    listening(false),
    tcpPort(SLOW_CONTROL_PORT),
    portInUse(false),
    qTcpServer(new QTcpServer(this)),
    tcpSocket(nullptr),
    daqState("INIT"),
    runNumber(0),
    spillCount(0),
    evNumber(0)
{
    responce();

#ifdef QT_DEBUG
    QTimer *periodicResponceTimer = new QTimer(this);
    connect(periodicResponceTimer, &QTimer::timeout,
            this, &SlowControlServer::responce);
    periodicResponceTimer->start(1000);
#endif
}

SlowControlServer::~SlowControlServer()
{
    blockSignals(true);
    startListen(false);
    delete qTcpServer;
    blockSignals(false);
}

void SlowControlServer::startListen(bool state)
{
    listening = state;
    if(listening){
        init();
    } else {
        qTcpServer->close();
        if(tcpSocket){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        qDebug() << QString("TcpServer closed on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    }
    updateStatus();
}

void SlowControlServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        if(!tcpSocket){
            tcpSocket = pendingConnection;
            portInUse=true;
            connect(tcpSocket, &QAbstractSocket::readyRead,
                    this, &SlowControlServer::readCommandRequest);
            connect(tcpSocket, &QAbstractSocket::disconnected,
                    this, &SlowControlServer::onDisconnected);
            QHostInfo info = QHostInfo::fromName(tcpSocket->peerAddress().toString());
            QString msg = QString("Accepted connection from %1:%2")
                    .arg(info.hostName())
                    .arg(tcpSocket->peerPort());
            qInfo() << msg;
            updateStatus();
#ifdef QT_DEBUG
responce();
#endif
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(!tcpSocket){
        qDebug()<<"There are no connections.";
    }
}

void SlowControlServer::readCommandRequest()
{
    if(!tcpSocket)
        return;
    while(tcpSocket->bytesAvailable()){
        QByteArray lineData = tcpSocket->readLine();
        QString line = QString::fromLatin1(lineData);
        if(line.startsWith("a")){
            responce();
        }
    }
}

void SlowControlServer::onDisconnected()
{
    portInUse=false;
    updateStatus();
    qInfo() << "Disconnected";
}

void SlowControlServer::slowControlStateUpdated(QString _state, QString _status)
{
    daqState = _state;
    daqStatus = _status;
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::trigStatusUpdated(const TrigStatus &status)
{
    rcState = (status.configValid && status.liveMagikOk) ? SC_STATE_ON : SC_STATE_FAULT;
    rcDescr = status.ident;
    spillCount = status.spillCount;
    evNumber = status.eventNumber();
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::spillPlotUpdated(const SpillPlotData &spillPlot)
{
    lastSpillPlot = spillPlot;
}

void SlowControlServer::triggerHistogramUpdated(const TriggerHistogramData &trigData)
{
    lastTrigData = trigData.histograms.isEmpty() ?
                TriggerHistogramChannelData() :
                trigData.histograms.last(); // fbp.multihist ? [15] : [0]
}

void SlowControlServer::onRunStart(QString index, int number)
{
    runIndex = index;
    runNumber = number;
}

void SlowControlServer::init()
{
    if(!qTcpServer->isListening()) {
        if(!qTcpServer->listen(QHostAddress::Any, tcpPort)) {
            QString s = tr("Unable to start on port %1: %2").arg(tcpPort).arg(qTcpServer->errorString());
            qCritical() << s;
            throw std::runtime_error(s.toStdString());
        }
    }
    tcpPort = qTcpServer->serverPort();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &SlowControlServer::updateConnections);
    qInfo() << QString("Listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void SlowControlServer::updateStatus() const
{
    emit connectionStatusUpdated(tcpPort, listening, portInUse);
}

void SlowControlServer::responce() const
{
    if(!tcpSocket)
        return;

    QJsonObject json;
    QJsonObject daqJson = getDaq();
    json["DAQ"] = daqJson;

    QJsonDocument jDoc(json);
    QString jsonLine = jDoc.toJson(QJsonDocument::Compact);
    jsonLine.append("\r\n");
    tcpSocket->write(jsonLine.toLatin1());
}

QString SlowControlServer::getRcBlock() const
{
    return QString("\"RunControlDevice\":{\"state\":\"%1\",\"descr\":\"%2\"}").arg(rcState, rcDescr);
}

QJsonObject SlowControlServer::getDaq() const
{
    QJsonObject json;
    json["state"] = daqState;
    json["status"] = daqStatus;
    json["runIndex"] = runIndex;
    json["runNumber"] = runNumber;
    if(!lastSpillPlot.data.isEmpty())
        json["spillPlot"] = getSpillJson();
    if(!lastTrigData.data.isEmpty())
        json["trigPlot"] = getTrigJson();
    QJsonObject rcJson = getRcDevice();

    if(!rcJson.isEmpty())
        json["RunControlDevice"] = rcJson;
    return json;
}

QJsonObject SlowControlServer::getRcDevice() const
{
    QJsonObject json;
    json["state"] = rcState;
    json["descr"] = rcDescr;
    json["spillCount"] = QString::number(spillCount);
    json["evNumber"] = QString::number(evNumber);
    return json;
}

QJsonObject SlowControlServer::getSpillJson() const
{
    QJsonObject json;
    json["spillCount"] = QJsonValue::fromVariant(lastSpillPlot.spillCount);
    json["spillCurrentTimeMs"] = QJsonValue::fromVariant(lastSpillPlot.spillCurrentTimeMs);
    json["stepSec"] = lastSpillPlot.stepSec;
    json["startOffset"] = lastSpillPlot.startOffset;
    QJsonArray plotArray;
    foreach (const quint32 &v, lastSpillPlot.data)
        plotArray.push_back(QJsonValue::fromVariant(v));
    json["data"] = plotArray;
    return json;
}

QJsonObject SlowControlServer::getTrigJson() const
{
    QJsonObject json;
    json["spillCount"] = QJsonValue::fromVariant(lastTrigData.spillCount);
    json["spillCurrentTimeMs"] = QJsonValue::fromVariant(lastTrigData.spillCurrentTimeMs);
//    json["stepSec"] = lastTrigData.stepSec;
    json["overflowCount"] = QJsonValue::fromVariant(lastTrigData.overflowCount);
    // px
    QJsonArray px;
    foreach (const double &v, lastTrigData.px)
        px.push_back(QJsonValue::fromVariant(v));
    json["px"] = px;
    // py
    QJsonArray plotArray;
    foreach (const quint32 &v, lastTrigData.data)
        plotArray.push_back(QJsonValue::fromVariant(v));
    json["data"] = plotArray;
    return json;
}
