//
//    Copyright 2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "SpillPlot.h"
#include "ui_SpillPlot.h"
#include "TrigDevice.h"
#include "qxw/WaveformPlot.h"
#include "qwt_scale_engine.h"

#include <QDebug>
#include <QLayout>
#include <QTimer>

static const int updatePeriodMs = 100;

SpillPlot::SpillPlot(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::SpillPlot)
{
    ui->setupUi(this);
    plot = new WaveformPlot();
    QLayout *layout = new QVBoxLayout(this);
    layout->addWidget(plot);
    auto scaleEngine = plot->axisScaleEngine(QwtPlot::yLeft);
    scaleEngine->setAttributes(QwtScaleEngine::IncludeReference);
    scaleEngine->setMargins(0, 1);
    plot->setAxisTitle(QwtPlot::xBottom, "Spill time, ms");
    plot->enableLegend(false);
    QTimer::singleShot(0, this, &SpillPlot::checkForReplot);
}

SpillPlot::~SpillPlot()
{
    delete ui;
}

QVector<quint32> trimRight(const QVector<quint32> &data)
{
    if (data.size() < 2)
        return data;
    if (data.last() != 0)
        return data;
    for (int i=data.size() - 2; i>0; i--)
        if (data[i] > 0) {
            QVector<quint32> result(data);
            result.resize(i+2);
            return result;
        }
    return data;
}

void SpillPlot::scheduleReplot()
{
    replot = true;
}

void SpillPlot::checkForReplot()
{
    if (replot)
        performReplot();
    QTimer::singleShot(updatePeriodMs, this, &SpillPlot::checkForReplot);
}

void SpillPlot::performReplot()
{
    quint32 py_max = 0;
    WaveformDataAll wda;
    // Z-order: first draw previous line, then new line
    if (history.size() > 1) {
        const SpillPlotData &previous = history.at(history.size()-2);
        const double rateToHz = previous.stepSec ? 1./previous.stepSec : 1;
        WaveformData wd1;
        const QVector<quint32> &py = trimRight(previous.data);
        for (int i=0; i<py.size(); i++) {
            double x = (previous.startOffset + i * previous.stepSec) * 1.0e3;
            if (py[i] > py_max)
                py_max = py[i];
            double y = rateToHz*py[i];
            wd1.p.push_back(point_t(x, y));
        }
        wd1.c = 1;
        wda.push_back(wd1);
    }
    if (! history.isEmpty()) {
        const SpillPlotData &shd = history.last();
        WaveformData wd0;
        const double rateToHz = shd.stepSec ? 1./shd.stepSec : 1;
        const QVector<quint32> &py = trimRight(shd.data);
        for (int i=0; i<py.size(); i++) {
            double x = (shd.startOffset + i * shd.stepSec) * 1.0e3;
            if (py[i] > py_max)
                py_max = py[i];
            double y = rateToHz*py[i];
            wd0.p.push_back(point_t(x, y));
        }
        wd0.c = 0;
        wda.push_back(wd0);
    }
    bool py_empty = (py_max == 0);
    auto scaleEngine = plot->axisScaleEngine(QwtPlot::yLeft);
    scaleEngine->setAttributes(QwtScaleEngine::IncludeReference);
    scaleEngine->setMargins(0, py_empty ? 1 : 0);

    bool autoReplot = plot->autoReplot();
    plot->setAutoReplot(false);
    plot->setAllData(wda);
    plot->setCurvePen(0, QPen(Qt::blue, 2, Qt::SolidLine));
    plot->setCurvePen(1, QPen(Qt::gray, 1, Qt::SolidLine));
    //    plot->setCurveStyle(0, QwtPlotCurve::Sticks);
    //    plot->setCurveStyle(1, QwtPlotCurve::Steps);
    plot->setAutoReplot(autoReplot);
}

void SpillPlot::updateSpillPlot(const SpillPlotData &shd)
{
    if (history.isEmpty()) {
        history.push_back(SpillPlotData()); // initialize
    }
    if (history.last().spillCount != shd.spillCount) {
        history.push_back(shd);
    }
    history.last() = shd;
    while (history.size() > 2) {
        history.pop_front();
    }
    scheduleReplot();
}
