//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "TrigDevice.h"

#include <cmath>

#include <QDateTime>
#include <QDebug>
#include <QFile>
#include <QHostAddress>
#include <QTimer>
#include <QTextStream>

#include "Config.h"

#include "RunControlRegs.h"
#include "WrEpState.h"
#include "regio/regop.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

namespace {
const int TRIG_STATUS_POLL_TIME_MS_MIN = 20;
const int TRIG_STATUS_POLL_TIME_MS_DEFAULT = 200;
const bool DEBUG_INFO = false;
constexpr fw_version_t REV_XOFF_COUNTERS(1, 0, 23162);
constexpr fw_version_t REV_HAMMING_CODE(1, 0, 22655);
constexpr fw_version_t REV_BLOCK_COUNTERS(1, 0, 22706);
constexpr fw_version_t REV_GLOB_EV_NUM(1, 0, 22712);
constexpr fw_version_t REV_GLOB_EV_NUM_NO_STAT(1, 0, 28815);
constexpr fw_version_t REV_PLOT_SOURCE(1, 0, 22778);
}

QString SpillStateBits::toString() const
{
    return QString("%1 %2 sel=%3 phase=%4 (0x%5)")
            .arg(reg.bit.ttl_ok?"TTL_NCU_Ok":"TTL_NCU_FAIL")
            .arg(reg.bit.wr_ok?"WR_NCU_Ok":"WR_NCU_FAIL")
            .arg(reg.bit.sel)
            .arg(reg.bit.phase)
            .arg(reg.all, 8, 16, QChar('0'))
            ;
}

bool TriggerHistogramChannelData::isEmpty() const
{
    if (data.isEmpty()) return true;
    for (int i=0; i<data.size(); i++) {
        if (data[i] > 0) {
            return false;
        }
    }
    return true;
}

int TriggerHistogramChannelData::getFirstNotemptyBin() const
{
    if (data.isEmpty()) return 0;
    for (int i=0; i<data.size(); i++) {
        if (data[i] > 0) {
            return i;
        }
    }
    return 0;
}

int TriggerHistogramChannelData::getLastNotemptyBin() const
{
    if (data.isEmpty()) return 0;
    // find last non-zero bin
    for (int i = data.size() - 1; i>0; i--) {
        if (data[i] > 0) {
            return i;
        }
    }
    return data.size() - 1;
}

TrigStatus::TrigStatus()
    : fwRev(),
      nch(0),
      connected(false),
      connectStChange(true),
      configValid(false),
      supportLvdsParity(false),
      liveMagikOk(false)
{
    clear();
}

void TrigStatus::clear()
{
    curTrigState = 0;
    curXOffState = 0;
    curXOffTimeoutState = 0;
    parityErrorCnt = 0;
    isL0Module = false;
    spillPlotAddr = 0;
    spillState = 0;
    spillCount = 0;
    spillCurrentTimeMs = 0;
    spillTtlLastPeriodMs = 0;
    spillWrLastPeriodMs = 0;
    ncuTai = WrTaiCode(0);
    wrUptime = 0;
    wrOtherErrorCount = 0;
    wrCodeErrorCount = 0;
    wrLinkLostCount = 0;
    wrSyncLostCount = 0;

    temp = 0;
    fwStr.clear();
    trigSrc = 0;
//    cnt_d00_base.clear();
    globEvNum = 0;
    globEvNum_cash = 0;
}

double TrigStatus::acceptRatio() const
{
    if (trigOnRunOnSpill() > 0)
        return static_cast<double>(trigAccepted()) / trigOnRunOnSpill();
    else
        return 0.;
}

double TrigStatus::missRatio() const
{
    if (trigOnRunOnSpill() > 0)
        return static_cast<double>(trigMissed()) / trigOnRunOnSpill();
    else
        return 0.;
}

double TrigStatus::deadTimeRatio() const
{
    if (timeReadout() > 0)
        return 1.0 * timeXoff() / timeReadout();
    else
        return 0.;
}

TrigDevice::TrigDevice(RcTrigConfig conf, quint32 ip, QObject *parent) :
    QtMregDevice(conf.rcDeviceId, conf.rcSerial, parent),
    lastConfig(conf),
    trigStatusTimer(new QTimer(this)),
    statusPollTimeMs(TRIG_STATUS_POLL_TIME_MS_DEFAULT)
{
    regCtrl.all = 0;
    connect(trigStatusTimer, &QTimer::timeout,
            this, &TrigDevice::trigStatusPoll);
    trigStatusTimer->setSingleShot(true);
    if(ip) {
        QHostAddress host(ip);
        setHostAddress(host);
    }
    connect(this, &TrigDevice::setStatusDumpEnable,
            &tsw, &TrigStatisticsWriter::setStatusDumpEnable);
    connect(this, &TrigDevice::setStatusDumpDirName,
            &tsw, &TrigStatisticsWriter::setStatusDumpDirName);
}

FirmwareBuildParams TrigDevice::readFirmwareBuildParams()
{
    if (!isConnected())
        return fbp;

    RegOpVector v;
    v << RegRead32(REG32_RC_BUILD_PARAM_0);
    v << RegRead32(REG32_RC_BUILD_PARAM_1);
    v << RegRead32(REG32_RC_BUILD_PARAM_2);
    v << RegRead32(REG32_RC_BUILD_PARAM_3);
    RegOpVector result = regOpExec(v);
    if (result.size() != v.size()) {
        return fbp;
    }
    quint32 fbp_0 = get32(result,0);
    quint32 fbp_1 = get32(result,2);
    quint32 fbp_2 = get32(result,4);
    quint32 fbp_3 = get32(result,6);
    fbp.ver = fbp_0 & 0xFF;
    if (fbp.ver == 0) { // before r.25460
        fbp.edge_rollover = false;
        fbp.two_level_trig = isL0Module();
        fbp.tih_log = false;
        fbp.multihist = false;
        fbp.nch = static_cast<quint8>(regRead(REG_RC_CHANNEL_CNT));
        fbp.ntrig = getTrigSrcCount();
        fbp.clk_hz = 62500000;
        fbp.spill_plot_ram_addr_bits = 13;
        fbp.trig_hist_ram_addr_bits = 13;
        fbp.level_cnt_bits = 48;
        fbp.edge_cntr_bits = 32;
    }
    if (fbp.ver == 0x81) { // v 1.0
        fbp.edge_rollover = fbp_0 & 0x100;
        fbp.two_level_trig = fbp_0 & 0x200;
        fbp.tih_log = fbp_0 & 0x400;
        fbp.multihist = fbp_0 & 0x800;
        fbp.nch = fbp_1 & 0xFF;
        fbp.ntrig = (fbp_1 >> 8) & 0xFF;
        fbp.spill_plot_ram_addr_bits = fbp_2 & 0xFF;
        fbp.trig_hist_ram_addr_bits = (fbp_2 >> 8) & 0xFF;
        fbp.level_cnt_bits = (fbp_2 >> 16) & 0xFF;
        fbp.edge_cntr_bits = (fbp_2 >> 24) & 0xFF;
        fbp.clk_hz = fbp_3;
        if (fbp.tih_log)
            if (fbp.trig_hist_ram_addr_bits > 10)
                fbp.trig_hist_ram_addr_bits = 10;
    }
    if (! fbp.isValid()) {
        fbp = FirmwareBuildParams();
    }
    return fbp;
}

void TrigDevice::clearStats(bool onStartRun)
{
    if (!isConnected()) return;

    RegOpVector v;
    const quint16 csrReg = (lastConfig.checkLvdsParity && isSupportParityCheck()) ? 0x4 : 0;
    v << RegWrite(REG_CSR_CTRL, csrReg | 0x2);
    regCtrl.bit.clear = 1;
    v << RegWrite(REG_RC_CTRL, regCtrl.all);
    regCtrl.bit.clear = 0;
    v << RegWrite(REG_RC_CTRL, regCtrl.all);

    for(int i=0; i<= 0x3f; ++i){
        v << RegRead(REG64_CNT_XOFF_EVENTS_BASE+i);
    }
    v << RegRead64(globEvNumRegister());

    v << RegWrite(REG_CSR_CTRL, csrReg);
    auto res = regOpExec(v);
    if(res.size() != v.size()){
        qCritical() << __func__ << " regOpExec error";
        return;
    }

    QVector<mlink::RegOp>::const_iterator regIt = res.begin();
    bool ok;
    while(regIt != res.end()) {
        ok = true;
        if(onStartRun && regIt->addr == globEvNumRegister()) {
            status.globEvNum_cash = get64(regIt, &ok);
            if(ok) regIt -= 4;
        }
        if(regIt->addr>=REG64_CNT_XOFF_EVENTS_BASE && regIt->addr<=(REG64_CNT_XOFF_EVENTS_BASE+0x3f)) {
            status.xOffEventCnt_base[(regIt->addr-REG64_CNT_XOFF_EVENTS_BASE)/4] = get64(regIt, &ok);
            if(ok) continue; else break;
        }
        ++regIt;
    }
}

void TrigDevice::resetAll()
{
    if (!isConnected()) return;
    const quint16 val = (lastConfig.checkLvdsParity && isSupportParityCheck()) ? 0x4 : 0;

    RegOpVector v;
    v << RegWrite(REG_CSR_CTRL, val | 1);
    v << RegWrite(REG_CSR_CTRL, val);
    regOpExec(v);

    writeSetup(lastConfig);
}

void TrigDevice::onRunStart()
{
    evNumReset();
    clearStats(true);
}

RegOpVector TrigDevice::setTrigCmd(const TrigConfig &trig)
{
    RegOpVector v;
    if(status.configValid && lastConfig.trig == trig)
        return v;

    v << setTrigTimerFreqCmd(trig.trigTimerFreq);
    v << setTrigRandomFreqCmd(trig.trigRandomFreq);
    v << setTrigRandomDeadTimeCmd(trig.trigRandomDeadTimeUs);
    v << RegWrite(REG_RC_TRIG_SRC, trig.trigSrc);
    v << RegWrite(REG_RC_XOFF_MASK, trig.xOffMask);

    return v;
}

bool TrigDevice::checkLiveMagic()
{
    bool ok = QtMregDevice::checkLiveMagic();
    status.configValid &= ok;
    return ok;
}

bool TrigDevice::isSupportParityCheck() const
{
    return (getDeviceFwVersion() >= REV_HAMMING_CODE)
            && (getDeviceId() == DEVICE_ID_U40VE_RC);
}

bool TrigDevice::isL0Module() const
{
    return (lastConfig.chNum == 1)
            && (getDeviceId() == DEVICE_ID_U40VE_RC);
}
bool TrigDevice::isSupportedXoffCounters() const
{
    return (getDeviceFwVersion() >= REV_XOFF_COUNTERS);
}

void TrigDevice::writeSetup(RcTrigConfig cfg, int setupSeqNum)
{
    if (!isConnected()) return;

    bool ok = writeLiveMagic();
    if(ok) {
        RegOpVector v;
        v << setTrigCmd(cfg.trig);
        v << setSpillCmd(cfg.spill);
        if(isL0Module()) {
            v << RegWrite(REG_RC_LVDS_TRIG_MASK, cfg.lvdsMask);
            v << RegWrite(REG_RC_TRIG_WORD_MASK, cfg.trigWordMask);
            quint16 rcTrigProtectCtrl = 0;
            if(cfg.checkBeforeProtection) rcTrigProtectCtrl |= 1;
            if(cfg.checkAfterProtection) rcTrigProtectCtrl |= 2;
            if(cfg.bapBlock) rcTrigProtectCtrl |= 4;
            v << RegWrite(REG_RC_TRIG_BEFORE_AFTER_PROT_EN, rcTrigProtectCtrl);
            v << RegWrite(REG_RC_TRIG_REJECT_DELAY, cfg.trigRejectDelay);
            v << RegWrite(REG_RC_TRIG_BEFORE_WIDTH, cfg.beforeProtWidth);
            v << RegWrite(REG_RC_TRIG_AFTER_WIDTH, cfg.afterProtWidth);
        }
        if(getDeviceFwVersion() >= REV_PLOT_SOURCE)
            if(!status.configValid || lastConfig.plotSrc != cfg.plotSrc)
                v << RegWrite(REG_RC_SPILLPLOT_SOURCE, cfg.plotSrc);
        if(!status.configValid ||
                lastConfig.trig.trigOpen == cfg.trig.trigOpen||
                lastConfig.spill.softSpill == cfg.spill.softSpill)
            v << setControlStateCmd(cfg.trig.trigOpen, cfg.spill.softSpill);
        const RegOpVector &result =  regOpExec(v);
        ok = (result.size() == v.size());
    }
    status.configValid = ok && isConnected() && checkLiveMagic();
    lastConfig = cfg;

    if(status.configValid) {
        emit setupWritten(setupSeqNum);
        emit updateTabState(true);
    } else {
        qWarning()<<"Failed to write setup in run control device";
        emit updateTabState(false);
    }
}

RegOpVector TrigDevice::setControlStateCmd(bool trigOpen, bool softSpill)
{
    RegOpVector v;
    if (!isConnected()) return v;
    regCtrl.bit.run = trigOpen ? 1 : 0;
    regCtrl.bit.softspill = softSpill ? 1 : 0;
    v << RegWrite(REG_RC_CTRL, regCtrl.all);
    return v;
}

RegOpVector TrigDevice::setTrigTimerFreqCmd(double freq)
{
    double periodInClocks = TrigDevice::getClockFreqHz() / freq;
    if (periodInClocks < 1) periodInClocks = 1;
    if (periodInClocks > quint32(-1)) periodInClocks = quint32(-1);
    quint32 regValue = periodInClocks;
    return RegWrite32(REG32_RC_TRIG_PERIODICAL, regValue);
}

RegOpVector TrigDevice::setTrigRandomFreqCmd(double freq)
{
    double period = 1.0 * freq * (quint32)(-1) / TrigDevice::getClockFreqHz();
    if (period < 1) period = 1;
    if (period > quint32(-1)) period = quint32(-1);
    quint32 regValue = period;
    return RegWrite32(REG32_RC_TRIG_RANDOM, regValue);
}

RegOpVector TrigDevice::setTrigRandomDeadTimeCmd(double microseconds)
{
    double ticks = 1e-6 * microseconds * TrigDevice::getClockFreqHz();
    if (ticks < 0) ticks = 0;
    if (ticks > quint32(-1)) ticks = quint32(-1);
    quint32 regValue = ticks;
    return RegWrite32(REG32_RC_TRIG_PDEAD, regValue);
}

RegOpVector TrigDevice::setSpillCmd(const SpillConfig &spill)
{
    RegOpVector v;
    if(status.configValid && lastConfig.spill == spill)
        return v;

    // disable spill
    v << RegWrite32(REG32_SPILL_CTRL, 0);
    // write settings
    v << RegWrite32(REG32_SPILL_DUR_PHASE_1_MS, spill.durPhase1);
    v << RegWrite32(REG32_SPILL_DUR_PHASE_2_MS, spill.durPhase2);
    v << RegWrite32(REG32_SPILL_DUR_PHASE_3_MS, spill.durPhase3);
    v << RegWrite32(REG32_SPILL_DUR_PHASE_4_MS, spill.durPhase4);
//    v << RegWrite32(REG32_SPILL_GATE_EXTEND_MS, spill.durPhase4);
    v << RegWrite32(REG32_SPILL_DUR_PHASE_5_MS, spill.durPhase5);
    v << RegWrite32(REG32_SPILL_SOFTSPILL_PERIOD_MS, spill.timerPeriod);
    v << RegWrite32(REG32_RC_TRIGTIMEPLOT_STEP, spill.trigTimeStep);
    v << RegWrite32(REG32_RC_SPILLPLOT_STEP, spill.spillPlotStep);
    // enable spill
    v << RegWrite32(REG32_SPILL_CTRL, (spill.source & 0xF) << 1);

    return v;
}

void TrigDevice::trigStatusPoll()
{
    if (!isConnected())
        return;
    status.clear();
    try{
        status.fwRev = getDeviceFwVersion();
        status.nch = lastConfig.chNum;
        status.ident = getIdent();
        status.fwStr = getFirmwareVersionStr();
        status.supportLvdsParity = isSupportParityCheck();
        status.isL0Module = isL0Module();
        status.temp = getTemperature();

        RegOpVector v;
        const quint16 mLinkCsrCtrl = (lastConfig.checkLvdsParity && isSupportParityCheck()) ? 0x4 : 0;
        if(getDeviceFwVersion() >= REV_BLOCK_COUNTERS)
            v << RegWrite(REG_CSR_CTRL, mLinkCsrCtrl + 0x2);
        v << RegRead(REG_WR_STATUS);
        v << RegRead64(REG64_WR_TAI_CODE);
        v << RegRead32(REG32_WR_LINK_LOSS_COUNT);
        v << RegRead32(REG32_WR_SYNC_LOSS_COUNT);
        v << RegRead32(REG32_WR_CODE_ERROR_COUNT);

        v << RegRead(REG_RC_TRIG_SRC);
        v << RegRead(REG_RC_TRIG_STATE);
        v << RegRead(REG_RC_XOFF_STATE);
        v << RegRead(REG_RC_XT_STATE);
        v << RegRead32(REG32_RC_SPILLPLOT_ADDR);

        v << RegRead32(REG32_SPILL_COUNT);
        v << RegRead32(REG32_SPILL_CURRENT_TIME_MS);
        v << RegRead32(REG32_SPILL_TTL_LAST_PERIOD_MS);
        v << RegRead32(REG32_SPILL_WR_LAST_PERIOD_MS);
        v << RegRead32(REG32_SPILL_CTRL);
        v << RegRead32(REG32_SPILL_STATE);
        v << RegRead64(REG64_SPILL_NCU_TAI_TIMESTAMP);
        if(getDeviceFwVersion() >= REV_GLOB_EV_NUM_NO_STAT)
            v << RegRead64(REG64_RC_GLOB_EV_NUM);

        if (false) {
            // Full readout
            for (quint16 i=0; i<REG64_CNT_TOTAL_NUM; i++)
                v << RegRead64(REG64_CNT_BASE + 4 * i);
        } else {
            /// Selective readout
            QVector<quint16> timeCntIndex;
            for (quint16 j=0; j<=std::min(13, lastConfig.chNum); j++)
                timeCntIndex << j;
            timeCntIndex << 14 << 15;
            for(quint16 j : timeCntIndex) {
                for (quint16 i=0; i<4; i++) {
                    quint16 index = i*2*REG64_CNT_BLOCK_SIZE + j;
                    v << RegRead64(REG64_CNT_BASE + 4*index);
                }
            }
            QVector<quint16> edgeCntIndex;
            edgeCntIndex << 0 << 14 << 15;
            for(quint16 i : edgeCntIndex) {
                for (quint16 j=0; j<4; j++) {
                    quint16 index = (j*2+1)*REG64_CNT_BLOCK_SIZE + i;
                    v << RegRead64(REG64_CNT_BASE + 4*index);
                }
            }
            for(quint16 ch=0; ch<lastConfig.chNum; ++ch) {
                v << RegRead64(REG64_CNT_XOFF_HI_BASE + 4*ch);
                v << RegRead64(REG64_CNT_XOFF_LOW_BASE + 4*ch);
                v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*ch);
            }
            if(status.isL0Module) {
                // Read-out first 4 64-bit counters
                for(quint16 ch=1; ch<5; ++ch)
                    v << RegRead64(REG64_CNT_LVDS_BASE + 4*ch);
            }
            if(isSupportParityCheck() && lastConfig.checkLvdsParity)
                v << RegRead64(REG64_CNT_LVDS_BASE);

            v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*14);
            v << RegRead64(REG64_CNT_XOFF_EVENTS_BASE + 4*15);
        }
        if(getDeviceFwVersion() >= REV_BLOCK_COUNTERS)
            v << RegWrite(REG_CSR_CTRL, mLinkCsrCtrl);

        status.cnt.clear();
        status.xOffHiTimeoutCnt.clear();
        status.xOffLowTimeoutCnt.clear();
        status.xOffEventCnt = status.xOffEventCnt_base;
        status.l0RegCnt.clear();

        quint64 ncuTaiTimestamp = 0;
        reg_wr_tai_code_t curTaiTimestamp; curTaiTimestamp.all = 0;
        RegOpVector result = regOpExec(v);
        if(v.size() != result.size()) {
            qWarning() << "bad result of regOpExec";
            return;
        }

        quint32 vPlot = regRead32(REG32_RC_SPILLPLOT_STEP);
        if(vPlot != lastConfig.spill.spillPlotStep){
            qWarning() << "REG32_RC_SPILLPLOT_STEP:" << hex << vPlot;
            regWrite32(REG32_RC_SPILLPLOT_STEP, lastConfig.spill.spillPlotStep);
        }
        int vSrc = regRead(REG_RC_SPILLPLOT_SOURCE);
        if(vSrc != lastConfig.plotSrc){
            qWarning() << "REG_RC_SPILLPLOT_SOURCE:" << hex << vSrc;
            regWrite(REG_RC_SPILLPLOT_SOURCE, lastConfig.plotSrc);
        }

        bool stateChanged=false;

        bool ok;
        QVector<mlink::RegOp>::const_iterator regOpIt = result.begin();
        while(regOpIt != result.end()) {
            ok = true;
            const mlink::RegOp &regOp = *regOpIt;
            switch (regOp.addr) {
            case REG_WR_STATUS:
                status.wrStatus.code.all = regOp.data;
                break;
            case REG32_WR_UPTIME:
            case REG32_WR_UPTIME+1: {
                int byte = regOp.addr % 4;
                status.wrUptime |= quint64(regOp.data) << 16*byte;
                break;
            }
//            case REG64_WR_TAI_CODE:
//            case REG64_WR_TAI_CODE+1:
//            case REG64_WR_TAI_CODE+2:
//            case REG64_WR_TAI_CODE+3: {
//                int byte = regOp.addr % 4;
//                curTaiTimestamp.all |= quint64(regOp.data) << 16*byte;
//                break;
//            }
            case REG64_WR_TAI_CODE: {
                curTaiTimestamp.all = get64(regOpIt, &ok);
                if(ok) continue; else break;
            }
            case REG32_WR_LINK_LOSS_COUNT:
            case REG32_WR_LINK_LOSS_COUNT+1: {
                int byte = regOp.addr % 4;
                status.wrLinkLostCount |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG32_WR_SYNC_LOSS_COUNT:
            case REG32_WR_SYNC_LOSS_COUNT+1: {
                int byte = regOp.addr % 4;
                status.wrSyncLostCount |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG32_WR_CODE_ERROR_COUNT:
            case REG32_WR_CODE_ERROR_COUNT+1: {
                int byte = regOp.addr % 4;
                status.wrCodeErrorCount |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG32_WR_OTHER_ERROR_COUNT:
            case REG32_WR_OTHER_ERROR_COUNT+1: {
                int byte = regOp.addr % 4;
                status.wrOtherErrorCount |= quint64(regOp.data) << 16*byte;
                break;
            }

            case REG_RC_TRIG_SRC:
                if(status.trigSrc != regOp.data) {
                    status.trigSrc = regOp.data;
                    stateChanged = true;
                }
                break;
            case REG_RC_TRIG_STATE:
                if(status.curTrigState != regOp.data) {
                    status.curTrigState = regOp.data;
                    stateChanged = true;
                }
                break;
            case REG_RC_XOFF_STATE:
                if(status.curXOffState != regOp.data) {
                    status.curXOffState = regOp.data;
                    stateChanged = true;
                }
                break;
            case REG_RC_XT_STATE:
                if(status.curXOffTimeoutState != regOp.data) {
                    status.curXOffTimeoutState = regOp.data;
                    stateChanged = true;
                }
                break;
            case REG32_RC_SPILLPLOT_ADDR: {
                int byte = regOp.addr % 4;
                status.spillPlotAddr |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG32_SPILL_COUNT:
                status.spillCount = regOp.data;
                break;
            case REG32_SPILL_COUNT+1:
                status.spillCount |= (quint32)regOp.data << 16;
                break;
            case REG32_SPILL_CURRENT_TIME_MS:
                status.spillCurrentTimeMs = regOp.data;
                break;
            case REG32_SPILL_CURRENT_TIME_MS+1:
                status.spillCurrentTimeMs |= (quint32)regOp.data << 16;
                break;
            case REG32_SPILL_TTL_LAST_PERIOD_MS:
            case REG32_SPILL_TTL_LAST_PERIOD_MS+1: {
                int byte = regOp.addr % 2;
                status.spillTtlLastPeriodMs |= (quint32)regOp.data << 16*byte;
                break;
            }
            case REG32_SPILL_WR_LAST_PERIOD_MS:
            case REG32_SPILL_WR_LAST_PERIOD_MS+1: {
                int byte = regOp.addr % 2;
                status.spillWrLastPeriodMs |= (quint32)regOp.data << 16*byte;
                break;
            }
            case REG32_SPILL_CTRL:
//                qDebug() << QString("SPILL CTRL: %1").arg(regOp.data, 4, 16, QChar('0'));
                break;
            case REG32_SPILL_STATE:
                status.spillState = SpillStateBits(regOp.data);
//                reg_spill_status_t r; r.all = regOp.data;
                break;
            case REG64_SPILL_NCU_TAI_TIMESTAMP:
            case REG64_SPILL_NCU_TAI_TIMESTAMP+1:
            case REG64_SPILL_NCU_TAI_TIMESTAMP+2:
            case REG64_SPILL_NCU_TAI_TIMESTAMP+3: {
                int byte = regOp.addr % 4;
                ncuTaiTimestamp |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG64_CNT_LVDS_BASE:
            case REG64_CNT_LVDS_BASE+1:
            case REG64_CNT_LVDS_BASE+2:
            case REG64_CNT_LVDS_BASE+3: {
                int byte = regOp.addr % 4;
                status.parityErrorCnt |= quint64(regOp.data) << 16*byte;
                break;
            }
            case REG64_RC_GLOB_EV_NUM:
                status.globEvNum = get64(regOpIt, &ok);
                if(ok) continue; else break;
            default:
                if (regOp.addr >= REG64_CNT_BASE && regOp.addr < REG64_CNT_BASE + 4 * REG64_CNT_TOTAL_NUM) {
                    if(((regOp.addr-REG64_CNT_BASE) % 4) == 0) {
                        const int index =  (regOp.addr - REG64_CNT_BASE) / 4; // index of 64-bit register
                        const int i = index / REG64_CNT_BLOCK_SIZE; // ch-Number shift
                        const int j = index % REG64_CNT_BLOCK_SIZE; // counter type
                        status.cnt[i][j] |= get64(regOpIt, &ok);
                        if(ok) continue; else break;
                    }
                }
//                if (regOp.addr >= REG64_CNT_BASE && regOp.addr < REG64_CNT_BASE + 4 * REG64_CNT_TOTAL_NUM) {
//                    int byte = regOp.addr % 4;
//                    int index =  (regOp.addr - REG64_CNT_BASE) / 4; // index of 64-bit register
//                    int i = index / REG64_CNT_BLOCK_SIZE; // ch-Number shift
//                    int j = index % REG64_CNT_BLOCK_SIZE; // counter type
//                    status.cnt[i][j] |= ((quint64)regOp.data) << 16*byte;
//                }
                if (regOp.addr >= REG64_CNT_XOFF_HI_BASE && regOp.addr < REG64_CNT_XOFF_HI_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                    int byte = regOp.addr % 4;
                    int ch =  (regOp.addr - REG64_CNT_XOFF_HI_BASE) / 4;
                    status.xOffHiTimeoutCnt[ch] |= (quint64(regOp.data)) << 16*byte;
                }
                if (regOp.addr >= REG64_CNT_XOFF_LOW_BASE && regOp.addr < REG64_CNT_XOFF_LOW_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                    int byte = regOp.addr % 4;
                    int ch =  (regOp.addr - REG64_CNT_XOFF_LOW_BASE) / 4;
                    status.xOffLowTimeoutCnt[ch] |= (quint64(regOp.data)) << 16*byte;
                }
                if (regOp.addr >= REG64_CNT_LVDS_BASE && regOp.addr < REG64_CNT_LVDS_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                    int byte = regOp.addr % 4;
                    int ch =  (regOp.addr - REG64_CNT_LVDS_BASE) / 4;
                    status.l0RegCnt[ch] |= (quint64(regOp.data)) << 16*byte;
                }
                if (regOp.addr >= REG64_CNT_XOFF_EVENTS_BASE && regOp.addr < REG64_CNT_XOFF_EVENTS_BASE + 4 * REG64_CNT_BLOCK_SIZE) {
                    int byte = regOp.addr % 4;
                    int ch =  (regOp.addr - REG64_CNT_XOFF_EVENTS_BASE) / 4;
                    status.xOffEventCnt[ch] |= (quint64(regOp.data)) << 16*byte;
                }
                break;
            }
            if(!ok)
                qWarning() << "";
            ++regOpIt;
        }

        status.ncuTai = WrTaiCode(ncuTaiTimestamp);

        if (false) {
            qDebug() << "--- trigStatusPoll ---";
            qDebug() << "WR state: " <<  status.wrStatus.toString();
            qDebug() << "WR Uptime: " <<  status.wrUptime;
            qDebug() << "WR Link Loss  Count: " <<  status.wrLinkLostCount;
            qDebug() << "WR Sync Loss  Count: " <<  status.wrSyncLostCount;
            qDebug() << "WR Code Error Count: " <<  status.wrCodeErrorCount;
            qDebug() << "WR Other Error Count: " <<  status.wrOtherErrorCount;
            WrTaiCode curTai(curTaiTimestamp.all);
            qDebug() << "now: " + curTai.toString();
            qDebug() << "NCU: " + status.ncuTai.toString();
            qDebug() << "Spill state: " + status.spillState.toString();
            qDebug() << "Spill count: " << status.spillCount;
            qDebug() << "Spill current time: " << status.spillCurrentTimeMs;
            qDebug() << "Spill TTL last period: " << status.spillTtlLastPeriodMs;
            qDebug() << "Spill WR last period: " << status.spillWrLastPeriodMs;
        }
        if (false) {
            QMap<int, QString> blockName;
            blockName[0] = "Time tot";
            blockName[1] = "Edge tot";
            blockName[2] = "Time run";
            blockName[3] = "Edge run";
            blockName[4] = "Time spl";
            blockName[5] = "Edge spl";
            blockName[6] = "Time r&s";
            blockName[7] = "Edge r&s";
            qInfo() << "--- counters ---";
            for (int i : status.cnt.keys()) {
                QString str = blockName[i];
                for (int j : status.cnt[i].keys())
//                    if (status.cnt[i][j] != 0)
//                    str += QString(" %1").arg(1.0 * status.cnt[i][j], 7, 'g', 2);
                str += QString(" %1").arg(status.cnt[i][j], 16, 16);
                qInfo() << str;
            }
        }
        if (false) {
            for (int i=0; i<16; i++) {
                qInfo() << hex << i << status.xOffEventCnt[i];
            }
        }

        {
            const SpillPlotData &shd = readSpillPlotRam();
            const TriggerHistogramData &thd = readTriggerTimeHistogramRam();
            tsw.writeStatisticsDump(status, shd, thd);
        }
        status.liveMagikOk = checkLiveMagic();

        emit updateStatus(status);
        if(status.connectStChange)
            status.connectStChange = false;
    } catch (std::exception &e){
        qWarning() << "try/catch error updateStatus";
    }
    trigStatusTimer->start(statusPollTimeMs);
}

void TrigStatisticsWriter::writeFile()
{
    if (!enable)
        return;
    // write file
    QString fileName = QString("rc_%1_%2.txt")
            .arg(QDateTime::currentDateTime().toString("yyyyMMdd_hhmmss"))
            .arg(status.spillCount);

    QSettings settings(dirName + "/" + fileName, QSettings::IniFormat);
    settings.beginGroup("main");
    settings.setValue("ident", status.ident);
    settings.setValue("spillCount", status.spillCount);
    settings.setValue("SpillTimeMs", status.spillCurrentTimeMs);
    settings.endGroup();
    settings.beginGroup("spill_plot");
    if (!spd.data.isEmpty()) {
        settings.setValue("stepSec", spd.stepSec);
        settings.setValue("startOffset", spd.startOffset);
        settings.beginGroup("0");
        QStringList sl;
        Q_FOREACH(quint32 val, spd.data)
            sl += QString("%1").arg(val);
        settings.setValue("data", sl.join(","));
        settings.endGroup();
    }
    settings.endGroup();
    settings.beginGroup("histograms");
    Q_FOREACH(const TriggerHistogramChannelData &thcd, thd.histograms) {
        if (thcd.isEmpty())
            continue;
        settings.beginGroup(QString("%1").arg(thcd.channel));
        QStringList sl;
        Q_FOREACH(quint32 val, thcd.data)
            sl += QString("%1").arg(val);
        settings.setValue("data", sl.join(","));
        settings.setValue("overflowCount", QString("%1").arg(thcd.overflowCount));
        settings.endGroup();
    }
    settings.endGroup();
    settings.sync();
    if (settings.status() != QSettings::NoError) {
        qWarning() << QString("Error writing status file '%1': %2").arg(settings.fileName()).arg(settings.status());
    }
}
void TrigStatisticsWriter::setStatusDumpEnable(bool val)
{
    enable = val;
}

void TrigStatisticsWriter::setStatusDumpDirName(const QString &str)
{
    dirName = str;
}

void TrigStatisticsWriter::writeStatisticsDump(const TrigStatus &newStatus, const SpillPlotData &newSpd, const TriggerHistogramData &newThd)
{
    bool endOfReadout = false;
    // graphs are cleared on beginning of phase 2
    if (status.spillState.reg.bit.phase == 4) {
        if (newStatus.spillState.reg.bit.phase < 2 || newStatus.spillState.reg.bit.phase == 5) {
            // flush existing graphs
            spd = newSpd;
            thd = newThd;
            endOfReadout = true;
        }
        if (newStatus.spillState.reg.bit.phase == 4) {
            if (status.spillCount != newStatus.spillCount) {
                endOfReadout = true; // spill incremented
            }
        } else {
            endOfReadout = true;
        }
    }
    if (endOfReadout)
        writeFile();
    status = newStatus;
    spd = newSpd;
    thd = newThd;
}

quint64 TrigStatus::eventNumber() const
{
    if(fwRev >= REV_GLOB_EV_NUM_NO_STAT)
        return globEvNum;

    if(fwRev >= REV_XOFF_COUNTERS)
        return xOffEventCnt[15]-xOffEventCnt_base[15];

    if(fwRev >= REV_GLOB_EV_NUM)
        return xOffEventCnt[0]-xOffEventCnt_base[0];

    return trigAccepted();
}

quint16 TrigDevice::globEvNumRegister() const
{
    if (getDeviceFwVersion() >= REV_GLOB_EV_NUM_NO_STAT)
        return REG64_RC_GLOB_EV_NUM;
    if (getDeviceFwVersion() >= REV_XOFF_COUNTERS)
        return REG64_CNT_XOFF_EVENTS_BASE + 4*15;
    if (getDeviceFwVersion() >= REV_GLOB_EV_NUM)
        return REG64_CNT_XOFF_EVENTS_BASE;
    return
        REG64_CNT_BASE + (7*REG64_CNT_BLOCK_SIZE+15)*4;
}

void TrigDevice::evNumReset()
{
    if(getDeviceFwVersion() < REV_GLOB_EV_NUM_NO_STAT) {
        qWarning() << QString("[%1] Device doesn't support 'evNumReset' function.").arg(getIdent());
        return;
    }
    regWrite64(REG64_RC_GLOB_EV_NUM_LOAD, 0);
}

void TrigDevice::evNumRequest(int checkIndex, bool local)
{
    if (!isConnected())
        return;
    try {
        // We needed only Run&Spill Edge counters (it is 8th counter block)
        // And in this block we read "Accepted trigger count" (the last one)
        RegOpVector v;
        const quint16 mLinkCsrCtrl = (lastConfig.checkLvdsParity && isSupportParityCheck()) ? 0x4 : 0;
        if(getDeviceFwVersion() >= REV_BLOCK_COUNTERS)
            v << RegWrite(REG_CSR_CTRL, mLinkCsrCtrl + 0x2);
        v << RegRead64(globEvNumRegister());
        if(getDeviceFwVersion() >= REV_BLOCK_COUNTERS)
            v << RegWrite(REG_CSR_CTRL, mLinkCsrCtrl);

        RegOpVector result = regOpExec(v);
        quint64 globEvNum=0;
        for (auto regOp : result) {
            if(regOp.addr >= globEvNumRegister() && regOp.addr < globEvNumRegister()+4) {
                const int byte = regOp.addr % 4;
                globEvNum |= quint64(regOp.data) << 16*byte;
            }
        }
        globEvNum -= status.globEvNum_cash;
//        const quint64 evNum = regRead64(REG64_CNT_BASE + (7*REG64_CNT_BLOCK_SIZE+15)*4);
        if(DEBUG_INFO)
            qInfo()<<"TrigDevice::evNumRequest:"<<checkIndex<<globEvNum;

        if(local)
            emit trigEvNumResponse(checkIndex, globEvNum);
        else
            emit devEvNumResponse(checkIndex, getDeviceSerialNumber(), globEvNum);
    } catch (std::exception &e){
        qWarning() << "try/catch error evNumRequest";
    }
}

SpillPlotData TrigDevice::readSpillPlotRam()
{
    if (!isConnected()) return SpillPlotData();
//    qDebug() << regRead32(REG32_RC_SPILLPLOT_ADDR);
    SpillPlotData shd;
    shd.spillCount = status.spillCount;
    shd.spillCurrentTimeMs = status.spillCurrentTimeMs;
    shd.startOffset = 1.0e-3 * (lastConfig.spill.durPhase1);
    shd.stepSec = 1.0 * lastConfig.spill.spillPlotStep / getClockFreqHz();
    quint32 readSize = std::min(status.spillPlotAddr, (quint32)spillPlotSize());
    const QVector<quint32> &ram = QVector<quint32>::fromStdVector(memReadBlk(0, readSize));
    shd.data = ram;
    emit spillPlotUpdated(shd);
    return shd;
}

TriggerHistogramData TrigDevice::readTriggerTimeHistogramRam()
{
    if (!isConnected()) return TriggerHistogramData();
    TriggerHistogramData thd;
    thd.spillCount = status.spillCount;
    thd.spillCurrentTimeMs = status.spillCurrentTimeMs;
    quint32 readSize = triggerHistogramSize();
    quint32 hist_channels = fbp.multihist ? 16 : 1;
    for (int ch=0; ch<hist_channels; ch++) {
        if (ch >= fbp.nch && ch < 15) // unused channels, except last
            continue;
        bool xOff = lastConfig.trig.xOffMask&(1<<ch);
        if (ch < fbp.nch && !xOff) // disabled channel
            continue;
        TriggerHistogramChannelData thcd;
        const QVector<quint32> &ram = QVector<quint32>::fromStdVector(memReadBlk(0x10000 + ch * readSize, readSize));
        if (!ram.isEmpty()) {
            thcd.spillCount = thd.spillCount; // alias
            thcd.spillCurrentTimeMs = thd.spillCurrentTimeMs; // alias
            thcd.channel = ch;
            thcd.overflowCount = ram.last();
            thcd.data = ram;
            thcd.data.pop_back();
            thcd.px.resize(thcd.data.size());
            if (fbp.tih_log) {
                const double tickSec = 1.0 / getClockFreqHz();
                for (int i=0; i<thcd.data.size(); i++) {
                    const double scaler_2 = 54.0; // FIXME
                    double x = exp(i/scaler_2) * tickSec * 1e6; // μs
                    thcd.px[i] = x;// i*stepSec;
                }
            } else {
                const double stepSec = 1.0 * lastConfig.spill.trigTimeStep / getClockFreqHz();
                for (int i=0; i<thcd.data.size(); i++) {
                    double x = i * stepSec * 1e6; // μs
                    thcd.px[i] = x;
                }
                thcd.px[0] = thcd.px[1]; // log scale workaround
            }
            thd.histograms[ch] = thcd;
        }
    }
    emit triggerHistogramUpdated(thd);
    return thd;
}

quint8 TrigDevice::getTrigSrcCount() const
{
    quint8 trigSrcCnt = 1;
    if (getDeviceFwVersion() >= REV_PLOT_SOURCE) {
        switch (getDeviceId()) {
        case DEVICE_ID_U40VE_RC:
            trigSrcCnt = 1;
            break;
        case DEVICE_ID_UT24VE:
            trigSrcCnt = 4;
            break;
        default:
            trigSrcCnt = 1;
        }
    }
    return trigSrcCnt;
}

void TrigDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
    trigStatusTimer->start(0);
    status.connected = true;
    status.connectStChange = true;
    status.hostStr = getAddress().toString();

    int chNum = regRead(REG_RC_CHANNEL_CNT);
    emit chNumberUpdated(chNum);
    readFirmwareBuildParams();
    emit trigSrcCntUpdated(getTrigSrcCount());
    if(!status.configValid)
        resetAll();
}

void TrigDevice::on_deviceDisconnected()
{
    QtMregDevice::on_deviceDisconnected();
    status.configValid = false;
    status.connected = false;
    status.connectStChange = true;
    emit updateStatus(status);
    emit updateTabState(false);
}

void TrigDevice::setStatusPollTimeMs(int ms)
{
    if (ms < TRIG_STATUS_POLL_TIME_MS_MIN)
        ms = TRIG_STATUS_POLL_TIME_MS_MIN;
    statusPollTimeMs = ms;
}
