#ifndef TRIGDEVICE_H
#define TRIGDEVICE_H

#include "mregdev/qtmregdevice.h"
#include "Config.h"
#include "RunControlRegs.h"
#include "WrEpState.h"

#include <QVector>
#include <QMap>
#include <QMetaType>

class SpillConfig;
class TrigConfig;
namespace mlink {class DeviceAttr;}

struct FirmwareBuildParams {
    FirmwareBuildParams()
        : ver(0), edge_rollover(false), two_level_trig(false), tih_log(false), multihist(false),
          nch(1), ntrig(1),
          spill_plot_ram_addr_bits(13), trig_hist_ram_addr_bits(13), level_cnt_bits(48), edge_cntr_bits(32),
          clk_hz(62500000) { }
    bool isValid() const {
        return true
                && ((ver == 0) || (ver == 0x81))
                && ((nch > 0) && (nch <= 16))
                && ((ntrig > 0) && (ntrig <= 16))
                && ((spill_plot_ram_addr_bits > 0) && (spill_plot_ram_addr_bits <= 24))
                && ((trig_hist_ram_addr_bits > 0) && (trig_hist_ram_addr_bits <= 24))
                && ((level_cnt_bits >= 16) && (level_cnt_bits <= 64))
                && ((edge_cntr_bits >= 16) && (edge_cntr_bits <= 64))
                && ((clk_hz > 0) && (clk_hz <= 1000000000))
                ;
    }
    quint8 ver;
    bool edge_rollover;
    bool two_level_trig;
    bool tih_log;
    bool multihist;
    quint8 nch;
    quint8 ntrig;
    quint8 spill_plot_ram_addr_bits;
    quint8 trig_hist_ram_addr_bits;
    quint8 level_cnt_bits;
    quint8 edge_cntr_bits;
    quint32 clk_hz;
};

struct SpillStateBits
{
    SpillStateBits(quint16 init = 0) {reg.all = init;}
    QString toString() const;
    reg_spill_status_t reg;
};

class SpillPlotData
{
public:
    SpillPlotData() : spillCount(0), spillCurrentTimeMs(0), stepSec(1.0), startOffset(0.0) {}
    QVector<quint32> data;
    quint32 spillCount;
    quint32 spillCurrentTimeMs;
    double stepSec;
    double startOffset; // in sec
};
Q_DECLARE_METATYPE(SpillPlotData)

class TriggerHistogramChannelData
{
public:
    TriggerHistogramChannelData() : overflowCount(0), spillCount(0), spillCurrentTimeMs(0), channel(0) {}
    bool isEmpty() const;
    int getFirstNotemptyBin() const;
    int getLastNotemptyBin() const;
    QVector<double> px;
    QVector<quint32> data;
    quint64 overflowCount;
    quint32 spillCount; // duplicated TriggerHistogramData::spillCount
    quint32 spillCurrentTimeMs;// duplicated TriggerHistogramData::spillCurrentTimeMs
    quint32 channel;
};

Q_DECLARE_METATYPE(TriggerHistogramChannelData)

class TriggerHistogramData
{
public:
    TriggerHistogramData() : spillCount(0), spillCurrentTimeMs(0) {}
    QMap<int, TriggerHistogramChannelData> histograms;
    quint32 spillCount;
    quint32 spillCurrentTimeMs;
};
Q_DECLARE_METATYPE(TriggerHistogramData)

class TrigStatus
{
public:
    TrigStatus();
    void clear();
    double acceptRatio() const;
    double missRatio() const;
    double deadTimeRatio() const;

    /// First counter index: [0,1] - Total, [2,3] - Run, [4,5] - Spill, [6,7] - Run&Spill
    /// Even - time, odd - edges
    /// cntLevel[*][15] - Elapsed Time
    /// cntLevel[*][14] - XOFF Time

    qint64 trigAll() const {return cnt[1][14];}
    qint64 trigOnRun() const {return cnt[3][14];}
    qint64 trigOnSpill() const {return cnt[5][14];}
    qint64 trigOnRunOnSpill() const {return cnt[7][14];}
    qint64 trigAccepted() const {return cnt[7][15];}
    qint64 trigMissed() const {return trigOnRunOnSpill() - trigAccepted();}
    qint64 timeXoff() const { return cnt[6][14];}

    // Elapsed Time
    qint64 timeClock() const { return cnt[0][15];}
    qint64 timeRun() const { return cnt[2][15];}
    qint64 timeSpill() const { return cnt[4][15];}
    qint64 timeReadout() const { return cnt[6][15];}
    qint64 timeStandby() const {qint64 v = timeClock() - timeRun(); return (v>0) ? v : 0;}
    quint64 eventNumber() const;

    // trig
    quint16 curTrigState;
    quint16 curXOffState;
    quint16 curXOffTimeoutState;
    quint64 parityErrorCnt;
    // spill
    quint32 spillPlotAddr;
    SpillStateBits spillState;
    quint32 spillCount;
    quint32 spillCurrentTimeMs;
    quint32 spillTtlLastPeriodMs;
    quint32 spillWrLastPeriodMs;
    // WR
    WrStatusBits wrStatus;
    WrTaiCode ncuTai;
    quint32 wrUptime;
    quint32 wrOtherErrorCount;
    quint32 wrCodeErrorCount;
    quint32 wrLinkLostCount;
    quint32 wrSyncLostCount;
    QMap<int, QMap<int, qint64> > cnt; // <cntType, <ch, cntVal> >
    QMap<int, quint64> xOffHiTimeoutCnt; // <ch, cntVal>
    QMap<int, quint64> xOffLowTimeoutCnt; // <ch, cntVal>
    QMap<int, quint64> xOffEventCnt; // <ch, cntVal>
    QMap<int, quint64> l0RegCnt; // <ch, cntVal>
    // other
    fw_version_t fwRev;
    int nch;
    bool connected;
    bool connectStChange;
    bool configValid;
    bool supportLvdsParity;
    bool isL0Module;
    bool liveMagikOk;
    QString ident;
    double temp;
    QString fwStr;
    QString hostStr;
    quint16 trigSrc;
    QString hostAddress;

    QMap<int, quint64> xOffEventCnt_base;
    quint64 globEvNum; // since f/w 28815
    quint64 globEvNum_cash; // no need since f/w 28815
};

class TrigStatisticsWriter : public QObject
{
    Q_OBJECT
public:
    explicit TrigStatisticsWriter(QObject *parent = nullptr)
        : QObject(parent), enable(false) {}
    void writeStatisticsDump(const TrigStatus &newStatus, const SpillPlotData &newSpd, const TriggerHistogramData &newThd);
public slots:
    void setStatusDumpEnable(bool val);
    void setStatusDumpDirName(const QString &str);

private:
    void writeFile();
    bool enable;
    QString dirName;
    TrigStatus status;
    SpillPlotData spd;
    TriggerHistogramData thd;
};

class TrigDevice : public QtMregDevice
{
    Q_OBJECT
public:
    explicit TrigDevice(RcTrigConfig conf, quint32 ip=0, QObject *parent = nullptr);
    quint8 getTrigSrcCount() const;
    FirmwareBuildParams readFirmwareBuildParams();
    int spillPlotSize() const { return 1 << fbp.spill_plot_ram_addr_bits;}
    int triggerHistogramSize() const {return 1 << fbp.trig_hist_ram_addr_bits;}
    mlink::RegOpVector setControlStateCmd(bool trigOpen, bool softSpill);
    mlink::RegOpVector setTrigTimerFreqCmd(double freq);
    mlink::RegOpVector setTrigRandomFreqCmd(double freq);
    mlink::RegOpVector setTrigRandomDeadTimeCmd(double microseconds);
    mlink::RegOpVector setSpillCmd(const SpillConfig &spill);
    mlink::RegOpVector setTrigCmd(const TrigConfig &trig);
    static double getClockFreqHz() { return 62500000; }
    static double getTrigTimerMinFreqHz() { return getClockFreqHz() / (quint32)(-1); }
    static double getTrigTimerMaxFreqHz() { return getClockFreqHz() / 2; }
    virtual bool checkLiveMagic();
    bool isSupportParityCheck() const;
    bool isL0Module() const;
    bool isSupportedXoffCounters() const;
    void setStatusPollTimeMs(int ms);

signals:
    void updateStatus(const TrigStatus &status);
    void spillPlotUpdated(const SpillPlotData &);
    void triggerHistogramUpdated(const TriggerHistogramData &);
    void chNumberUpdated(int chNum);
    void trigSrcCntUpdated(quint8 trigScrCnt);
    void setupWritten(int setupSeqNum);
    void updateTabState(bool isOk, int tabIndex=0);
    void trigEvNumResponse(int checkIndex, quint64 evNum);
    void devEvNumResponse(int checkIndex, quint64 devSer, quint64 evNum);
    void setStatusDumpDirName(const QString &);
    void setStatusDumpEnable(bool);

public slots:
    void writeSetup(RcTrigConfig cfg, int setupSeqNum=0);
    void resetAll();
    void onRunStart();
    void clearStats(bool onStartRun);
    void evNumRequest(int checkIndex, bool local);

protected slots:
    void trigStatusPoll();
    virtual void on_deviceConnected();
    virtual void on_deviceDisconnected();

protected:
    quint16 globEvNumRegister() const;

private:
    void evNumReset();
    SpillPlotData readSpillPlotRam();
    TriggerHistogramData readTriggerTimeHistogramRam();

    TrigStatisticsWriter tsw;
    RcTrigConfig lastConfig;
    QTimer * const trigStatusTimer;
    FirmwareBuildParams fbp;
    TrigStatus status;
    reg_spill_ctrl_t regCtrl;
    int statusPollTimeMs;
};

#endif // TRIGDEVICE_H
