//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "VmeTrigsWidget.h"
#include "ui_VmeTrigsWidget.h"
#include "Config.h"
#include "VmeTrigDevice.h"
#include "daq-config/DaqConfig.h"
#include "base/threadQuit.h"
#include "device-discover/DiscoverDialog.h"

#include <QCheckBox>
#include <QDebug>
#include <QElapsedTimer>
#include <QInputDialog>
#include <QHostAddress>
#include <QMessageBox>
#include <QThread>
#include <QTimer>

Q_DECLARE_METATYPE(VmeTrigStatus)

namespace {
enum COL_VME_TRIGS {COL_VME_ID, COL_VME_SERIAL, COL_VME_READOUT, COL_VME_HOST,
                    COL_VME_TEMP, COL_VME_EV_NUM, COL_VME_STATUS,
                    COL_VME_CHECK_WR, COL_VME_WR_TIME_VALID, COL_VME_WR_SYNC_LOST, COL_VME_WR_LINK_ERROR,
                    COL_VME_BUSY,COL_VME_IDLE,COL_VME_OCCUPANCY,
                    COL_VME_TOTAL  };
}

VmeTrigsWidget::VmeTrigsWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::VmeTrigsWidget),
    config(NULL),
    last_got_run_number(0),
    lastRunState(RS_INIT),
    runEnable(false),
    checkOutOfDateTimer(new QTimer(this)),
    controlEnabled(true)
{
    ui->setupUi(this);

    QTableWidget *table = ui->tableWidgetVmeTrig;

#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    QStringList columnHeaders;
    columnHeaders << "ID" << "Serial" << "Readout" << "Host" << "Temp";
    columnHeaders << "evNum" << "Status";
    columnHeaders << "Check WR" << "WR Time" << "WR Sync Lost" << "WR Link Error";
    columnHeaders << "Busy" << "Idle" << "Ocupancy";
    Q_ASSERT(columnHeaders.size() == COL_VME_TOTAL);
    table->setColumnCount(columnHeaders.size());
    table->setHorizontalHeaderLabels(columnHeaders);
    table->hideColumn(COL_VME_ID);

    updateEnableAddButton();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<VmeTrigStatus>();
    connect(checkOutOfDateTimer, &QTimer::timeout,
            this, &VmeTrigsWidget::checkOutOfDate);
    checkOutOfDateTimer->setSingleShot(true);
}

VmeTrigsWidget::~VmeTrigsWidget()
{
    emit deleteTrig(0);
    foreach (int serial, trigThreads.keys()) {
        QThread *thread = trigThreads[serial];
        if(!threadQuiting(thread,1000)){
            qInfo()<<"terminating thread of vme trig; serial:0x"<<hex<<serial;
            thread->terminate();
        }
    }
    delete ui;
}

void VmeTrigsWidget::init(Config *conf)
{
    config = conf;
    foreach (VmeTrigConfig vc, config->vmeTrigs.values()) {
        addTrig(vc);
    }
    emit updateTabState(getSummaryState());
}

void VmeTrigsWidget::updateDevicesList(const QVector<DeviceDescription> devices)
{
    allDevices.clear();
    foreach (const DeviceDescription &dd, devices) {
        if(trigs.keys().contains(dd.serial_id))
            continue;
        allDevices.push_back(dd);
    }
    checkOutOfDate();
}

void VmeTrigsWidget::insertAvailableDevice(const DeviceDescription &dd)
{
    if(dd.device_id != QtMregDevice::DEVICE_ID_VME_TRIG)
        return;

    if(!checkOutOfDateTimer->isActive())
        checkOutOfDateTimer->start(dd.ttl*1000);
    for(int i=0; i<allDevices.size(); ++i) {
        DeviceDescription &oldDd = allDevices[i];
        if(oldDd.getIndex() == dd.getIndex()){
            oldDd = dd;
            updateEnableAddButton();
            return;
        }
    }

    allDevices.push_back(dd);
    updateEnableAddButton();
}

void VmeTrigsWidget::on_pushButtonAddVmeTrig_clicked()
{
    bool ok;
    QStringList items;
    QMap<QString, quint64> map;
    for(int i=0; i<allDevices.size(); ) {
        const DeviceDescription &dd = allDevices[i];
        if(dd.isOutOfDate()){
            allDevices.remove(i);
            updateEnableAddButton();
            continue;
        }
        ++i;
        quint64 serial = dd.serial_id;
        if(trigs.keys().contains(serial))
            continue;
        QString item = dd.getSerialIdStr();
        if(trigHostMap.contains(dd.serial_id)){
            item.append(QString(" (vmedaq host: %1)").arg(trigHostMap[dd.serial_id]));
        }
        map[item] = serial;
        items << item;
    }

    if(items.isEmpty()){
        QMessageBox msgBox;
         msgBox.setText("There is no free FVME2TM module.");
         msgBox.exec();
         return;
    }

    QString item = QInputDialog::getItem(this, "Select VME trigger", "Select FVME2TMWR",
                          items, 0, false, &ok); //, Qt::WindowFlags flags = 0, Qt::InputMethodHints inputMethodHints = Qt::ImhNone
    if(!ok)
        return;
    Q_ASSERT(map.contains(item));
    VmeTrigConfig vc(map[item]);
    addTrig(vc);
    Q_ASSERT(!config->vmeTrigs.contains(vc.serial));
    config->vmeTrigs[vc.serial] = vc;
    config->write_config();
}

void VmeTrigsWidget::on_pushButtonRemoveVmeTrig_clicked()
{
    int rowToRemove = getSelectedRow();
    if(rowToRemove == -1){
        ui->pushButtonRemoveVmeTrig->setEnabled(false);
        return;
    }

    quint64 serial = getSerialByRow(rowToRemove);
    qInfo()<<"Removing tableWidgetVmeTrig row="<<rowToRemove<<"; serial=0x"<<hex<<serial;

    emit deleteTrig(serial);
    QThread *thread = trigThreads.take(serial);
    if(!threadQuiting(thread,1000)){
        qInfo()<<"terminating thread of vme trig; serial:0x"<<hex<<serial;
        thread->terminate();
    }

    trigs.remove(serial);
    trigThreads.remove(serial);
    trigStates.remove(serial);
    trigStatus.remove(serial);
    ui->tableWidgetVmeTrig->removeRow(rowToRemove);
    Q_ASSERT(config->vmeTrigs.contains(serial));
    config->vmeTrigs.remove(serial);
    config->write_config();
    Q_ASSERT(thread);
    delete thread;

    qInfo()<<"vme trig device removed; serial:0x"<<hex<<serial;
}

void VmeTrigsWidget::on_tableWidgetVmeTrig_itemSelectionChanged()
{
    const bool en = !runEnable && (getSelectedRow()!=-1);
    ui->pushButtonRemoveVmeTrig->setEnabled(en);
    updateTablePalette();
}

void VmeTrigsWidget::trigStatusUpdated(const VmeTrigStatus &status)
{
    int row = -1;
    quint64 serial=0;
    foreach (serial, trigs.keys()) {
        if(trigs[serial] == sender()){
            row = gerRowBySerial(serial);
            break;
        }
    }
    if(row==-1) {
        qDebug()<<"Can't find row to trigStatusUpdated:"<<status.temp<<" serial:"<<serial;
        return;
    }
    ui->tableWidgetVmeTrig->item(row, COL_VME_TEMP)->setText(QString::number(status.temp, 'f', 1).append(QChar(0x02da)).append('C'));
    ui->tableWidgetVmeTrig->item(row, COL_VME_EV_NUM)->setText(QString::number(status.evNum));
    ui->tableWidgetVmeTrig->item(row, COL_VME_WR_TIME_VALID)->setText(status.isWrTimeValid() ? "Ok" : "Invalid");
    ui->tableWidgetVmeTrig->item(row, COL_VME_WR_TIME_VALID)->setTextColor(status.isWrTimeValid() ? QColor() : Qt::red);
    ui->tableWidgetVmeTrig->item(row, COL_VME_STATUS)->setTextColor(status.connected ? QColor() : Qt::red);
    ui->tableWidgetVmeTrig->item(row, COL_VME_WR_SYNC_LOST)->setText(QString::number(status.wrSyncFail));
    ui->tableWidgetVmeTrig->item(row, COL_VME_WR_LINK_ERROR)->setText(QString::number(status.wrLinkError));
    ui->tableWidgetVmeTrig->item(row, COL_VME_BUSY)->setText(QString::number(status.busyTimer));
    ui->tableWidgetVmeTrig->item(row, COL_VME_IDLE)->setText(QString::number(status.idleTimer));
    ui->tableWidgetVmeTrig->item(row, COL_VME_OCCUPANCY)->setText(QString::number(100.*status.busyTimer/(status.busyTimer+status.idleTimer)));
    bool prevOk = getSummaryState();
    bool updatePalette = !trigStatus.contains(serial) ||
            (trigStatus[serial].isWrTimeValid() != status.isWrTimeValid()) ||
            (trigStatus[serial].connected != status.connected);

    bool stopRunCondition = false;
    QString errStr;

    // Detect changes in status
    const VmeTrigStatus &oldStatus = trigStatus[serial];
    const QString &oldPrefix = oldStatus.ident + " ";
    const QString &prefix = trigStatus[serial].ident + " ";

    qint32 wrLinkErrorCount = status.wrLinkError - oldStatus.wrLinkError;
    if (wrLinkErrorCount > 0) {
        qWarning() << prefix + QString("WR link errors: %1").arg(wrLinkErrorCount);
    }
    qint32 wrSyncLostCount = status.wrSyncFail - oldStatus.wrSyncFail;
    if (wrSyncLostCount > 0) {
        qWarning() << prefix + QString("WR sync lost errors: %1").arg(wrSyncLostCount);
    }
    if(config->vmeTrigs[serial].checkWr) {
        if (!oldStatus.isWrTimeValid() && status.isWrTimeValid()) {
            qInfo() << prefix + "WR time valid";
        }
        if (oldStatus.isWrTimeValid() && !status.isWrTimeValid()) {
            errStr = prefix + "WR time invalid";
            (runEnable?qWarning():qInfo()) << errStr;
            stopRunCondition = true;
        }
    }
    if (!oldStatus.connected && status.connected) {
        qInfo() << prefix + "connected";
    }
    if (oldStatus.connected && !status.connected) {
        errStr = oldPrefix + "disconnected";
        (runEnable?qWarning():qInfo()) << errStr;
        stopRunCondition = true;
    }
    if (runEnable && config->vmeTrigs[serial].readout && stopRunCondition) {
        emit runError(last_got_run_index, last_got_run_number, errStr);
    }

    trigStatus[serial] = status;
    if(updatePalette)
        updateTablePalette();
    bool ok = getSummaryState();
    if(prevOk != ok)
        emit updateTabState(ok);
}

void VmeTrigsWidget::setStatus(const QString &status)
{
    int row = -1;
    quint64 serial=0;
    foreach (serial, trigs.keys()) {
        if(trigs[serial] == sender())
            row = gerRowBySerial(serial);
    }
    if(row==-1){
        qDebug()<<"Can't find row to setStatus:"<<status<<" serial:"<<serial;
        return;
    }
    ui->tableWidgetVmeTrig->item(row, COL_VME_STATUS)->setText(status);
}

void VmeTrigsWidget::trigRunStateUpdated(quint64 serial, RUN_STATES state, QString run_index, int run_number)
{
    if(run_index != last_got_run_index){
        qWarning()<<QString("Got wrong response(run_index=%1; currun_index=%2) from VmeTrig module 0x%3")
                  .arg(run_index).arg(last_got_run_index)
                  .arg(QtMregDevice::getSerialIdStr(serial));
        return;
    }

    if(run_number<last_got_run_number){
        qWarning()<<QString("Got old response(run=%1) from vme trig module %2")
                  .arg(run_number).arg(QtMregDevice::getSerialIdStr(serial));
        return;
    }
    Q_ASSERT(last_got_run_number==run_number);
    if (trigStates.contains(serial)) {
        if(state==RS_ERROR)
            qWarning()<<"vme trig module is offline; serial=0x"<<hex<<serial;
        trigStates[serial] = state;
    } else {
        qWarning()<<"Got trigRunStateUpdated from unknown serial=0x"<<hex<<serial;
    }
}

void VmeTrigsWidget::onRunStart(QString run_index, int run_number)
{
    runEnable=true;
    updateControlEnable();
    qDebug()<<"Start Vme Trig resetting...";
    lastRunState = RS_STARTING;
    last_got_run_index = run_index;
    last_got_run_number = run_number;

    bool noEnabledModules=true;
    foreach (const VmeTrigConfig &vc, config->vmeTrigs.values()) {
        if(vc.readout) {
            noEnabledModules=false;
            break;
        }
    }
    if(noEnabledModules) {
        lastRunState = RS_STARTED;
        sendVmeMap();
        return;
    }

    foreach (quint64 serial, trigStates.keys()) {
        if(config->vmeTrigs[serial].readout)
            trigStates[serial] = RS_STARTING;
        else
            trigStates[serial] = RS_STARTED;
    }
    emit resetEvNum(run_index, run_number);
    QElapsedTimer timer;
    timer.start();
    bool looping=true;
    // wait all ended
    while(looping && !timer.hasExpired(5000)){
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        looping=false;
        foreach (quint64 serial, trigStates.keys()) {
            if(!config->vmeTrigs[serial].readout)
                continue;
            if(trigStates[serial] == RS_STARTING){
                looping=true;
                break;
            }
        }
    }

    bool isOk = !looping; // break previous while on looping flag
    foreach (quint64 serial, trigStates.keys()) {
        if(!config->vmeTrigs[serial].readout)
            continue;
        if(trigStates[serial] != RS_STARTED){
            isOk=false;
            break;
        }
    }

    qDebug()<<"All Vme Trig resetting finished; "<< (isOk?"success":"error");
    lastRunState = isOk ? RS_STARTED : RS_ERROR;
    sendVmeMap();
}

void VmeTrigsWidget::onRunStop()
{
    runEnable=false;
    updateControlEnable();
}

void VmeTrigsWidget::vmeReadoutToggled(bool readout)
{
    QTableWidget *table = ui->tableWidgetVmeTrig;
    for(int row=0; row<table->rowCount(); ++row){
        if(table->cellWidget(row, COL_VME_READOUT) == reinterpret_cast<QWidget *>(sender())){
            quint64 vmeSerial = getSerialByRow(row);
            Q_ASSERT(vmeSerial);

            Q_ASSERT(config->vmeTrigs.contains(vmeSerial));
            bool prevOk = getSummaryState();
            config->vmeTrigs[vmeSerial].readout = readout;
            config->write_config();
            qInfo() << QString("vmeTrig at row=%1(%2) was %3abled")
                       .arg(row).arg(QtMregDevice::getSerialIdStr(vmeSerial))
                       .arg(readout?"en":"dis");
            emit vmeReadoutUpdated(vmeSerial, readout);
            bool curOk = getSummaryState();
            if(prevOk != curOk)
                emit updateTabState(curOk);

            const bool wrCheckEn = !runEnable && readout;
            ui->tableWidgetVmeTrig->cellWidget(row, COL_VME_CHECK_WR)->setEnabled(wrCheckEn);
            return;
        }
    }
}

void VmeTrigsWidget::wrCheckToggled(bool check)
{
    QTableWidget *table = ui->tableWidgetVmeTrig;
    for(int row=0; row<table->rowCount(); ++row){
        if(table->cellWidget(row, COL_VME_CHECK_WR) == reinterpret_cast<QWidget *>(sender())){
            quint64 vmeSerial = getSerialByRow(row);
            Q_ASSERT(vmeSerial);

            Q_ASSERT(config->vmeTrigs.contains(vmeSerial));
            bool prevOk = getSummaryState();
            config->vmeTrigs[vmeSerial].checkWr = check;
            config->write_config();
            qInfo() << QString("checking WR-time on vmeTrig at row=%1(%2) was %3abled")
                       .arg(row).arg(QtMregDevice::getSerialIdStr(vmeSerial))
                       .arg(check?"en":"dis");
            bool curOk = getSummaryState();
            if(prevOk != curOk)
                emit updateTabState(curOk);
            return;
        }
    }
}

void VmeTrigsWidget::checkOutOfDate()
{
    for(int i=0; i<allDevices.size(); ) {
        const DeviceDescription &dd = allDevices[i];
        if(dd.isOutOfDate()){
            allDevices.remove(i);
            updateEnableAddButton();
            continue;
        }
        ++i;
    }
    if(!allDevices.isEmpty())
        checkOutOfDateTimer->start();
}

void VmeTrigsWidget::gotProgramDescription(const ProgramDescription &pd)
{
    if(pd.type != DaqConfig::getTypeName(DaqConfigVmeDaq))
        return;

    if(!programMap.contains(pd.uuid)){
        // Remove possible dublicates with another uuid
        foreach (QUuid uid, programMap.keys()) {
            const ProgramDescription &oldPD = programMap[uid];
            if(oldPD.sameProgram(pd)){
                programMap.remove(uid);
                break;
            }
        }
    }
    bool ok;
    for(int i=0; i<24; ++i) {
        const QString optKeyId = QString("slot%1_id").arg(i, 2, 10, QChar('0'));
        const QString optKeySerial = QString("slot%1_serial").arg(i, 2, 10, QChar('0'));
        if(!pd.options.contains(optKeyId))
            continue;
        quint16 id = pd.options[optKeyId].toUShort(&ok, 16);
        if(!ok || id != 0xa)
            continue;
        if(!pd.options.contains(optKeySerial))
            continue;
        quint64 ser = pd.options[optKeySerial].toULongLong(&ok, 16);
        if(!ok)
            continue;
        if(!pd.options.contains("interfaceIndex"))
            continue;
        quint16 interfaceIndex = pd.options["interfaceIndex"].toUShort(&ok, 10);
        if(!ok)
            continue;
        trigHostMap[ser] = QString("%1:%2").arg(pd.hostName).arg(interfaceIndex);
        int row = gerRowBySerial(ser);
        if(row!=-1){
            ui->tableWidgetVmeTrig->item(row, COL_VME_HOST)->setText(trigHostMap[ser]);
        }
        break;
    }
    programMap[pd.uuid] = pd;
}

void VmeTrigsWidget::gotClosedProgram(const ProgramDescription &pd)
{
    if(programMap.contains(pd.uuid))
        programMap.remove(pd.uuid);
}

void VmeTrigsWidget::setControlEnable(bool enabled)
{
    controlEnabled = enabled;
    setEnabled(controlEnabled);
}

void VmeTrigsWidget::addTrig(const VmeTrigConfig &vc)
{
    const quint64 &serial = vc.serial;
    quint32 ip=0;
    foreach (const DeviceDescription &dd, allDevices) {
        if(dd.serial_id == serial){
            ip = dd.ip_addr;
            break;
        }
    }
    VmeTrigDevice *dev = new VmeTrigDevice(vc.serial);
    QThread *thread = new QThread(this);
    thread->setObjectName(QString("VmeTrigThr_%1").arg(dev->getSerialIdStr()));
    dev->moveToThread(thread);
    thread->start();
    trigs[serial] = dev;
    trigThreads[serial] = thread;
    trigStates[serial] = RS_INIT;

//    connect(dev, &VmeTrigDevice::deviceConnected(, this, &VmeTrigsWidget::on_deviceConnected()));
    connect(this, &VmeTrigsWidget::deviceDiscovered, dev, &VmeTrigDevice::deviceUpdated);
    connect(this, &VmeTrigsWidget::resetEvNum, dev, &VmeTrigDevice::resetEvNum);
    connect(this, &VmeTrigsWidget::vmeReadoutUpdated, dev, &VmeTrigDevice::vmeReadoutUpdated);
    connect(dev, &VmeTrigDevice::setStatus, this, &VmeTrigsWidget::setStatus);
    connect(dev, &VmeTrigDevice::updateStatus, this, &VmeTrigsWidget::trigStatusUpdated);
    connect(dev, &VmeTrigDevice::runStateUpdated, this, &VmeTrigsWidget::trigRunStateUpdated);
    connect(this, &VmeTrigsWidget::deleteTrig, dev, &VmeTrigDevice::deleteTriger);
    connect(dev, &VmeTrigDevice::destroyed, thread, &QThread::quit);
    connect(this, &VmeTrigsWidget::evNumRequest, dev, &VmeTrigDevice::evNumRequest);
    connect(dev, &VmeTrigDevice::evNumResponse, this, &VmeTrigsWidget::evNumResponse);

    ui->tableWidgetVmeTrig->setRowCount(trigs.size());
    int row = trigs.size()-1;
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_ID, new QTableWidgetItem(QString::number(serial)));
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_SERIAL, new QTableWidgetItem(dev->getSerialIdStr()));
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_HOST, new QTableWidgetItem());
    if(trigHostMap.contains(serial))
        ui->tableWidgetVmeTrig->item(row, COL_VME_HOST)->setText(trigHostMap[serial]);
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_TEMP, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_EV_NUM, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_STATUS, new QTableWidgetItem("creating..."));
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_CHECK_WR, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_WR_TIME_VALID, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_WR_SYNC_LOST, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_WR_LINK_ERROR, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_BUSY, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_IDLE, new QTableWidgetItem());
    ui->tableWidgetVmeTrig->setItem(row, COL_VME_OCCUPANCY, new QTableWidgetItem());
    QCheckBox *cb = new QCheckBox(ui->tableWidgetVmeTrig);
    cb->setChecked(true);
    ui->tableWidgetVmeTrig->setCellWidget(row, COL_VME_READOUT, cb);
    connect(cb, &QCheckBox::toggled,
            this, &VmeTrigsWidget::vmeReadoutToggled);


    QCheckBox *cbWr = new QCheckBox(ui->tableWidgetVmeTrig);
    cbWr->setChecked(true);
    ui->tableWidgetVmeTrig->setCellWidget(row, COL_VME_CHECK_WR, cbWr);
    connect(cbWr, &QCheckBox::toggled, this, &VmeTrigsWidget::wrCheckToggled);

    updateEnableAddButton();
    if(!vc.readout)
        cb->setChecked(vc.readout);
    if(!vc.checkWr)
        cbWr->setChecked(vc.checkWr);
}

void VmeTrigsWidget::updateEnableAddButton()
{
    if(!runEnable) {
        foreach (const DeviceDescription &dd, allDevices) {
            if(!trigs.contains(dd.serial_id)) {
                ui->pushButtonAddVmeTrig->setEnabled(true);
                return;
            }
        }
    }
    ui->pushButtonAddVmeTrig->setEnabled(false);
}

quint64 VmeTrigsWidget::getSerialByRow(int row) const
{
    bool ok;
    quint64 s = ui->tableWidgetVmeTrig->item(row, COL_VME_ID)->text().toULongLong(&ok);
    return ok ? s : 0;
}

int VmeTrigsWidget::gerRowBySerial(quint64 serial) const
{
    for(int row=0; row<ui->tableWidgetVmeTrig->rowCount();++row){
        if(getSerialByRow(row) == serial)
            return row;
    }
    return -1;
}

void VmeTrigsWidget::updateControlEnable()
{
    updateEnableAddButton();
    const bool en = !runEnable;
    ui->pushButtonRemoveVmeTrig->setEnabled(en && (getSelectedRow() != -1));
    for(int row=0; row<ui->tableWidgetVmeTrig->rowCount(); ++row) {
        const quint64 vmeSerial = getSerialByRow(row);
        Q_ASSERT(vmeSerial);
        Q_ASSERT(config->vmeTrigs.contains(vmeSerial));
        const bool wrCheckEn = en && config->vmeTrigs[vmeSerial].readout;

        ui->tableWidgetVmeTrig->cellWidget(row, COL_VME_READOUT)->setEnabled(en);
        ui->tableWidgetVmeTrig->cellWidget(row, COL_VME_CHECK_WR)->setEnabled(wrCheckEn);
    }
}

int VmeTrigsWidget::getSelectedRow() const
{
    int row=-1;
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidgetVmeTrig->selectedRanges();
    if(!ranges.isEmpty()){
        QTableWidgetSelectionRange firstRange = ranges[0];
        row = firstRange.topRow();
    }
    return row;
}

bool VmeTrigsWidget::getSummaryState() const
{
    foreach (const VmeTrigConfig &c, config->vmeTrigs.values()) {
        if(!c.readout)
            continue;
        if(!trigStatus.contains(c.serial) || !trigStatus[c.serial].connected)
            return false;
        if(c.checkWr && !trigStatus[c.serial].isWrTimeValid())
            return false;
    }
    return true;
}

void VmeTrigsWidget::updateTablePalette()
{
    int row = getSelectedRow();
    if(row != -1) {
        quint64 s = getSerialByRow(row);
        bool isOk = trigStatus[s].connected && trigStatus[s].isWrTimeValid();
        isOk = trigStatus[s].connected;
        QPalette p = ui->tableWidgetVmeTrig->palette();
        p.setColor(QPalette::HighlightedText, isOk ? QColor() : Qt::red);
        ui->tableWidgetVmeTrig->setPalette(p);
    }
}

void VmeTrigsWidget::sendVmeMap() const
{
    QMap<quint64, QString> map;
    foreach (const VmeTrigConfig &c, config->vmeTrigs.values()) {
        if(c.readout)
            map[c.serial] = QtMregDevice::getSerialIdStr(c.serial);
    }
    emit  vmeMapUpdated(map);
}
