#include "InlCheckDialog.h"
#include "ui_InlCheckDialog.h"

#include <cmath>

#include <QComboBox>
#include <QDateTime>
#include <QDirIterator>
#include <QFile>
#include <QFileDialog>
#include <QMessageBox>
#include <QPushButton>
#include <QTimer>
#include <QSettings>

namespace {
enum tree_inlCheck_col {
    tree_inlCheck_name,
    tree_inlCheck_cnt,
    tree_inlCheck_mean,
    tree_inlCheck_rms,
    tree_inlCheck_num_cols
};
const double RMS_WARN = 27;
const double RMS_ERROR = 30;
}
InlCheckDialog::InlCheckDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::InlCheckDialog),
    rootTreeItem(new QTreeWidgetItem(QStringList("All")))
{
    ui->setupUi(this);

    QTreeWidget *tree = ui->treeWidget;
    tree->setColumnCount(tree_inlCheck_num_cols);
    QStringList headerLabels;
    headerLabels << "name"
                 << "hit cnt"
                 << "Mean, ns"
                 << "Rms, ps";
    tree->setHeaderLabels(headerLabels);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    tree->resizeColumnToContents(tree_inlCheck_name);
    ui->buttonBox->button(QDialogButtonBox::Reset)->setText("Refresh");
    ui->treeWidget->setAlternatingRowColors(true);
}

InlCheckDialog::~InlCheckDialog()
{
    delete ui;
}

void InlCheckDialog::on_treeWidget_itemSelectionChanged()
{
    auto tree = ui->treeWidget;
    QColor col = tree->currentItem()->background(tree_inlCheck_rms).color();
    if(col.isValid()) {
        tree->setStyleSheet(QString("QTreeWidget::item:selected {selection-color: rgb(%1, %2, %3);}")
                            .arg(col.red()).arg(col.green()).arg(col.blue()));
    } else {
        tree->setStyleSheet(QString());
    }
}

void InlCheckDialog::on_pushButtonUpdateHrbCal_clicked()
{
    QSettings s;
    s.beginGroup("INL_check");
    QString dir = s.value("hrbFolder").toString();
    dir = QFileDialog::getExistingDirectory(this, "Set folder with HRB calibrations...", dir);
    if(dir.isEmpty())
        return;
    s.setValue("hrbFolder", dir);
    s.endGroup();

    uploadHrbCal(dir);

    for(auto index : modTreeItems.keys()) {
        auto modItem = modTreeItems[index];
        auto cb = qobject_cast<QComboBox*>(ui->treeWidget->itemWidget(modItem, tree_inlCheck_mean));
        auto prevPreset = cb->currentText();
        ui->treeWidget->removeItemWidget(modItem, tree_inlCheck_mean);
        delete cb;
        cb = createHrbComboBox(index);
        cb->setCurrentText(prevPreset);
    }
}

void InlCheckDialog::on_buttonBox_clicked(QAbstractButton *button)
{
    QPushButton *b = qobject_cast<QPushButton *>(button);
    if(ui->buttonBox->button(QDialogButtonBox::Reset) == b) {
        emit retry();
    } else if(ui->buttonBox->button(QDialogButtonBox::Save) == b) {
        save_results();
    }
}

void InlCheckDialog::resultUpdated(const InlCheckResults &res)
{
    if(hrbCalib.cal.isEmpty())
        uploadHrbCal();

    inlResults = res;
    int devCnt = 0;
    ui->treeWidget->blockSignals(true);

    // Remove missing results
    for(auto index : modTreeItems.keys()) {
        if(inlResults.devices.contains(index)) {
            const InlCheckDevice &deviceResults = inlResults.devices[index];
            auto modItem = modTreeItems[index];
            for(int childId=0; childId<modItem->childCount(); ++childId) {
                auto chItem = modItem->child(childId);
                int ch = chItem->data(tree_inlCheck_name, Qt::UserRole).toInt();
                if(deviceResults.channels.contains(ch)) {
                   ++childId;
                } else {
                    modItem->removeChild(chItem);
                }
            }
        } else {
            rootTreeItem->removeChild(modTreeItems.take(index));
//            deviceOffsets.remove(index);
        }
    }

    // Add new one
    for(DeviceIndex index: inlResults.devices.keys()) {
        QStringList list;
        list << QString("%1: %2").arg(++devCnt).arg(index.getIdent());

        auto &modItem = modTreeItems[index];
        if(!modItem) {
            modItem = new QTreeWidgetItem(rootTreeItem, list);
            modItem->setExpanded(true);
        }

        createHrbComboBox(index);
        auto chOffsets = deviceOffsets.value(index);

        quint64 modCnt=0;
        double modMin = 0;
        double modMax = 0;

        for(auto chRes: inlResults.devices[index].channels.values()) {
            auto &chDist = chRes.dist;

            QTreeWidgetItem *chItem = nullptr;
            for(int i=0; i<modItem->childCount(); ++i) {
                bool ok;
                if(modItem->child(i)->data(tree_inlCheck_name, Qt::UserRole).toInt(&ok)==chRes.ch && ok) {
                    chItem = modItem->child(i);
                    break;
                }
            }

            if(chDist.cnt) {
                if(!chItem) {
                    QStringList colTextList;
                    colTextList << QString("%1").arg(chRes.ch,2,10,QChar('0')) << QString::number(chDist.cnt,'f',0) << "-" << "-";
                    chItem = new QTreeWidgetItem(modItem, colTextList);
                    chItem->setData(tree_inlCheck_name, Qt::UserRole, chRes.ch);
                }

                if(modCnt) {
                    modMin = qMin(modMin, chDist.rms);
                    modMax = qMax(modMax, chDist.rms);
                } else {
                    modMin = modMax = chDist.rms;
                }
                modCnt += chDist.cnt;
                chItem->setText(tree_inlCheck_cnt, QString("%1").arg(chDist.cnt));
                chItem->setText(tree_inlCheck_mean, QString("%1").arg(chDist.mean-chOffsets[chRes.ch], 0, 'f', 3));
                chItem->setText(tree_inlCheck_rms, QString("%1").arg(std::floor(chDist.rms*1000), 0, 'f', 0));
                QColor bgCol;
                if(chDist.rms*1000 > RMS_ERROR) {
                    bgCol = Qt::red;
                } else if(chDist.rms*1000 > RMS_WARN) {
                    bgCol = Qt::yellow;
                } else {
                    bgCol = Qt::transparent;
                }
                auto chBrush = chItem->background(tree_inlCheck_rms);
                chBrush.setColor(bgCol);
                chBrush.setStyle(Qt::SolidPattern);
                chItem->setBackground(tree_inlCheck_rms, chBrush);
            } else {
                if(chItem) {
                    modItem->removeChild(chItem);
                    delete chItem;
                }
            }
        }
        if(modCnt)
            modItem->setText(tree_inlCheck_rms, QString("%1-%2")
                             .arg(modMin*1000, 0, 'f', 0).arg(modMax*1000, 0, 'f', 0));
        modItem->sortChildren(tree_inlCheck_name, Qt::AscendingOrder);
    }

    for(auto modItem : modTreeItems.values()) {
        modItem->setHidden(modItem->childCount()==0);
    }

    ui->treeWidget->blockSignals(false);
    ui->treeWidget->resizeColumnToContents(tree_inlCheck_name);
}

void InlCheckDialog::hrbPresetChanged(QString preset)
{
    QComboBox *cb = qobject_cast<QComboBox*>(sender());
    Q_ASSERT(cb);
    DeviceIndex devIndex;
    for(auto i : modTreeItems.keys())
        if(ui->treeWidget->itemWidget(modTreeItems[i], tree_inlCheck_mean) == cb){
            devIndex = i;
            break;
        }
    if(devIndex.isNull()) return;

    qInfo() << "Dev:" << devIndex.getIdent() << " use " << preset;
    QSettings s;
    s.beginGroup("INL_check");
    s.setValue(devIndex.getIdent().replace(' ','_'), preset);
    s.endGroup();

    deviceOffsets[devIndex] = hrbCalib.cal.value(preset);
    refreshModResults(devIndex);
}

void InlCheckDialog::showEvent(QShowEvent *ev)
{
    QDialog::showEvent(ev);
    emit retry();
}

QComboBox *InlCheckDialog::createHrbComboBox(DeviceIndex index)
{
    QSettings s;
    s.beginGroup("INL_check");
    QString presetName = s.value(index.getIdent()).toString();
    s.endGroup();

    auto cb = new QComboBox(ui->treeWidget);
    for(auto name : hrbCalib.cal.keys())
        cb->addItem(name);
    cb->setCurrentText(presetName);
    deviceOffsets[index] = hrbCalib.cal.value(presetName);
    connect(cb, &QComboBox::currentTextChanged,
            this, &InlCheckDialog::hrbPresetChanged);
    ui->treeWidget->setItemWidget(modTreeItems[index], tree_inlCheck_mean, cb);
    refreshModResults(index);
    return cb;
}

void InlCheckDialog::refreshModResults(DeviceIndex devIndex)
{
    const auto &hrbOffsets = deviceOffsets[devIndex];
    auto &devRes = inlResults.devices[devIndex];
    auto modItem = modTreeItems[devIndex];
    for(int childId=0; childId<modItem->childCount(); ++childId) {
        auto chItem = modItem->child(childId);
        int ch = chItem->data(tree_inlCheck_name, Qt::UserRole).toInt();
        const auto &chRes = devRes.channels[ch];
        if(!chRes.dist.cnt) continue;
        chItem->setText(tree_inlCheck_mean, QString("%1").arg(chRes.dist.mean-hrbOffsets.value(ch, 0), 0, 'f', 3));
    }
}

void InlCheckDialog::uploadHrbCal(QString dir)
{
    hrbCalib.clear();
    if(dir.isEmpty()) {
        QSettings s;
        s.beginGroup("INL_check");
        dir = s.value("hrbFolder").toString();
        s.endGroup();
        if(dir.isEmpty()) {
            hrbCalib.load();
            return;
        }
    }

    QDirIterator it(dir, QDir::Files | QDir::NoDotDot, QDirIterator::Subdirectories);
    while(it.hasNext()) {
        auto fn = it.next();
        if(!hrbCalib.load(fn)) {
            auto b = QMessageBox::warning(this, "HRB preset was overwritten",
                                          "Some HRB preset was overwritten while uploading file:\n"+fn,
                                          QMessageBox::Abort|QMessageBox::Ok,QMessageBox::Ok);
            if(b == QMessageBox::Abort) {
                qWarning() << "Uploading HRB calibrations was aborted";
                break;
            }
        }
    }
}

DeviceIndex InlCheckDialog::get_selected_device() const
{
    DeviceIndex index;
    auto selectedItems = ui->treeWidget->selectedItems();
    for(auto item : selectedItems) {
        DeviceIndex item_index = modTreeItems.key(item);
        if(item_index.isNull()) {
            item_index = modTreeItems.key(item->parent());
            if(item_index.isNull())
                return DeviceIndex();
        }
        if(index.isNull())
            index = item_index;
        else {
            if(index != item_index)
                return DeviceIndex();
        }
    }
    return  index;
}

void InlCheckDialog::save_results()
{
    DeviceIndex selected_index = get_selected_device();
    QString selected_file_name = QString("%1.txt").arg(selected_index.getSerialStr());
    QSettings s;
    s.beginGroup("INL_check");
    QString filePath = s.value("fileWithInlRes").toString();
    if(!selected_index.isNull()) {
        if(filePath.isEmpty()) {
            filePath = selected_file_name;
        } else {
            QFileInfo fi(filePath);
            filePath = fi.path()+QDir::separator()+selected_file_name;
        }
    }
    s.endGroup();
    filePath = QFileDialog::getSaveFileName(this, "Save INL check results to file...", filePath);
    if(!filePath.isEmpty()) {
        s.beginGroup("INL_check");
        s.setValue("fileWithInlRes", filePath);
        s.endGroup();
        QFile file;
        file.setFileName(filePath);
        if(!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qCritical() << "Failed to open file for INL-check result:"
                        << filePath << file.errorString();
            return;
        }
        QTextStream textStream(&file);
        QStringList line;
        textStream << "Date\t" << QDateTime::currentDateTime().toString() << "\n";
        textStream << "StopDev\t" << inlResults.stopDev.getIdent() << "\n";
        textStream << "StopCh\t" << inlResults.stopCh << "\n";
        for(auto devIndex : inlResults.devices.keys()) {
            if(!selected_index.isNull() && devIndex != selected_index)
                continue;
            textStream << "\n";
            auto chOffsets = deviceOffsets.value(devIndex);
            textStream << "Dev\t" << devIndex.getIdent() << "\n";
            auto preset = qobject_cast<QComboBox*>(ui->treeWidget->itemWidget(modTreeItems[devIndex], tree_inlCheck_mean))->currentText();
            textStream << "HRB preset\t" << preset << "\n";
            auto &devRes = inlResults.devices[devIndex];
            bool printHeader = true;
            for(auto ch : devRes.channels.keys()) {
                DistributionParams chDist = devRes.channels[ch].dist;
                if(!chDist.cnt) continue;
                if(printHeader) {
                    printHeader = false;
                    line << "ch"
                         << "cnt"
                         << "mean ns"
                         << "mean-hrbOffset ns"
                         << "RMS ps";
                    textStream << line.join('\t') << '\n';
                    line.clear();
                }

                line << QString::number(ch)
                     << QString("%1").arg(chDist.cnt)
                     << QString("%1").arg(chDist.mean, 0, 'f', 3)
                     << QString("%1").arg(chDist.mean-chOffsets[ch], 0, 'f', 3)
                     << QString("%1").arg(std::floor(chDist.rms*1000), 0, 'f', 0);
                textStream << line.join('\t') << '\n';
                line.clear();
            }
            if(printHeader) {
                textStream << "no hits" << '\n';
            }
        }
    }
}
