#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <QFileDialog>
#include <QHostAddress>
#include <QSemaphore>
#include <QSettings>
#include "QShortcut"
#include <QSpinBox>
#include <QStandardPaths>
#include <QThread>
#include <QLabel>

#include "HptdcVoltageTest.h"
#include "CalibrationDialog.h"
#include "ConfigParams.h"
#include "SaveInlDialog.h"
#include "file-reader-dialog/FileReaderDialog.h"
#include "TDCAnalyzer.h"
#include "device-discover/DiscoverDialog.h"
#include "mregdev/qtmregdevice.h"
#include "daq-config/DaqConfig.h"
#include "base/threadQuit.h"
#include "add-pnp-dialog/AddPnpDialog.h"
//#include "RootWriter.h"
#include "TcpClient.h"
#include "LogDialog.h"
#include "pnp-server/PNPServer.h"
#include "DeviceHelper.h"
#include "InlCheckDialog.h"

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qxw/WaveformPlot.h>
#include "version.h"
#include <QPlainTextEdit>
#include <QBoxLayout>
#include <QTableWidgetItem>
#include <QTreeWidgetItem>
#include "HistWorker.h"
#include "TdcTtvxsMapDialog.h"
#include "StatusDialog.h"
#include "base/AboutBox.h"

namespace {
enum tree_tdc_col {
    tree_tdc_name,
    tree_tdc_ch,
    tree_tdc_inv,
    tree_tdc_inl,
    tree_tdc_ttvxs,
    tree_tdc_raw_cnt,
    tree_tdc_tdc_hits,
    tree_tdc_num_cols
};

const QStringList SOURCE_TYPES = {DaqConfig::getTypeName(DaqConfigEventBuilder), DaqConfig::getTypeName(DaqConfigMStream)};
void setLabelColor(QLabel * const lab, QColor col) {
    QPalette p = lab->palette();
    p.setColor(lab->foregroundRole(), col);
    lab->setPalette(p);
}

QString inlMark(int st) {
    switch (st) {
    case 1:
        return QChar(0x2714);
    case -1:
        return QChar(0x2716);
    }
    return QChar('?');
}
QColor inlCol(int st) {
    switch (st) {
    case 1:
        return Qt::darkGreen;
    case -1:
        return Qt::red;
    }
    return Qt::black;
}
void updateItemInl(QTreeWidgetItem * item, int st){
    item->setText(tree_tdc_inl, inlMark(st));
    auto brush = item->foreground(tree_tdc_inl);
    brush.setColor(inlCol(st));
    item->setForeground(tree_tdc_inl, brush);
}
void updateItemTtvxs(QTreeWidgetItem * item, int st){
    item->setText(tree_tdc_ttvxs, inlMark(st));
    auto brush = item->foreground(tree_tdc_ttvxs);
    brush.setColor(inlCol(st));
    item->setForeground(tree_tdc_ttvxs, brush);
}
}

MainWindow::MainWindow(TDC2ViewerStartupOptions &startup_options, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    analyzerSem(new QSemaphore(DATA_QUEUE_SIZE)),
    histSem(new QSemaphore(DATA_QUEUE_SIZE)),
    writerSem(new QSemaphore(DATA_QUEUE_SIZE)),

    config(startup_options),
    prev_invert_map(config.invertMap),
    fileReaderDialog(new FileReaderDialog(&config.file_reader_config, analyzerSem, this)),
    pnpServer(new PNPServer(this)),

    dataSourceLabel(new QLabel(this)),
    dataRateLabel(new QLabel(this)),
    analyzerStatusLabel(new QLabel(this)),
    semStatusLabel(new QLabel(this)),
    vxsClockLabel(new QLabel(this)),
    analyzerLabel(new QLabel(this)),
    logDialog(new LogDialog(this)),
    calibrationDialog(new CalibrationDialog(this)),
    inlCheckDialog(startup_options.inl ? new InlCheckDialog(this) : nullptr),
    tcpClient(new TcpClient(analyzerSem)),
    tcpClientThread(new QThread(this)),
    analyzer(new TDCAnalyzer(analyzerSem, histSem, writerSem)),
    analyzerThread(new QThread(this)),
    worker(new HistWorker(histSem)),
    worker_thread(new QThread(this)),
    rootTreeItem(new QTreeWidgetItem(QStringList("All"))),
    rawDataPeriodicTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    rawDataReqDone(true),
    rawEventCnt(0),
    tcpConnOk(false),
    statusDialog(new StatusDialog(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<InvertMapType>();
    qRegisterMetaType<TdcTtvxsMapType>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<DeviceCalibration>();
    qRegisterMetaType<InlResults>();
    qRegisterMetaType<QMap<int,double> >();
    qRegisterMetaType<QMap<int,int> >();
    qRegisterMetaType<DecodedData>();
    qRegisterMetaType<AnalyzerStats>();
    qRegisterMetaType<InlCheckResults>();
    qRegisterMetaType<HistContainer>();
    qRegisterMetaType<HistMngrCfg>();
    ui->setupUi(this);

    HistMngrCfg histCfg = ui->widgetHistMngr->init(startup_options.inl);
    worker->init(histCfg);
    if(!startup_options.inl) {
        ui->menuHPTDC_INL->menuAction()->setVisible(false);
        ui->actionHPTDC_vol_check->setVisible(startup_options.inl);
    }
    ui->plainTextEditRawData->setFont(QFont("monospace"));

    statusBar()->addWidget(dataSourceLabel,10);
    statusBar()->addWidget(dataRateLabel,3);
    statusBar()->addWidget(analyzerStatusLabel,10);
    statusBar()->addWidget(vxsClockLabel,5);
    statusBar()->addWidget(semStatusLabel,4);
    statusBar()->addWidget(analyzerLabel,3);
    analyzerStatusLabel->setToolTip("Analyzer status");
    semStatusLabel->setToolTip("Queue's semaphores");

    setWindowTitle(QCoreApplication::applicationName());

    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &QWidget::show);
    connect(this, &MainWindow::addLog, logDialog, &LogDialog::addLog);

    connect(pnpServer, &PNPServer::programDiscovered,
            this, &MainWindow::programDiscovered);
    connect(ui->widgetHistMngr, &HistMngrWidget::hist_cfg_updated,
            worker, &HistWorker::hist_cfg_updated);
    connect(pnpServer, &PNPServer::programClosed,
            this, &MainWindow::programClosed);
    pnpServer->startSearch(SOURCE_TYPES);

    analyzer->moveToThread(analyzerThread);
    tcpClient->moveToThread(tcpClientThread);
    worker->moveToThread(worker_thread);

    analyzerThread->setObjectName("Analizer");
    tcpClientThread->setObjectName("TcpClient");
    worker_thread->setObjectName("Worker");

    QObject::connect(analyzer, &QObject::destroyed, analyzerThread, &QThread::quit);
    QObject::connect(tcpClient, &QObject::destroyed, tcpClientThread, &QThread::quit);
    QObject::connect(worker, &QObject::destroyed, worker_thread, &QThread::quit);
    connect(analyzer, &TDCAnalyzer::deviceDataUpdated, worker, &HistWorker::dataUpdated);
    connect(analyzer, &TDCAnalyzer::deviceStatsUpdated, this, &MainWindow::updateHitStats);
    connect(analyzer, &TDCAnalyzer::deviceStatsUpdated, statusDialog, &StatusDialog::updateHitStats);
    connect(analyzer, &TDCAnalyzer::rawDataUpdated, this, &MainWindow::rawDataUpdated);
    connect(this, &MainWindow::rawDataEveryEvUpdated, analyzer, &TDCAnalyzer::setRawDataEveryEv);
    connect(this, &MainWindow::requestRawData, analyzer, &TDCAnalyzer::requestRawData);
    connect(this, &MainWindow::resetAll, analyzer, &TDCAnalyzer::resetStats);
    connect(this, &MainWindow::resetAll, worker, &HistWorker::reset);
    connect(this, &MainWindow::resetAll, statusDialog, &StatusDialog::reset);
    connect(this, &MainWindow::set_can_skip, analyzer, &TDCAnalyzer::set_can_skip);
    connect(this, &MainWindow::set_show_skip_info, analyzer, &TDCAnalyzer::set_show_skip_info);
    connect(worker, &HistWorker::resetHistMngr, ui->widgetHistMngr, &HistMngrWidget::resetHistMngr);

    connect(ui->widgetHistMngr, &HistMngrWidget::recreateTaiHist, worker, &HistWorker::recreateTaiHist);
    connect(ui->widgetHistMngr, &HistMngrWidget::clearHist, worker, &HistWorker::clearHist);
    connect(this, &MainWindow::inlConfigUpdated, worker, &HistWorker::inlConfigUpdated);
    connect(worker, &HistWorker::devListUpdated, ui->widgetHistMngr, &HistMngrWidget::devListUpdated);
    connect(worker, &HistWorker::hist_updated, ui->widgetHistMngr, &HistMngrWidget::refreshCurrentPlotB);

    connect(fileReaderDialog, &FileReaderDialog::writeSettings, this, &MainWindow::writeSettings);
    connect(fileReaderDialog, &FileReaderDialog::dataAccuired, analyzer, &TDCAnalyzer::dataAccuired);
    connect(fileReaderDialog, &FileReaderDialog::fileNameUpdated, this, &MainWindow::sourceFileNameUpdated);
    connect(this, &MainWindow::closeFile, fileReaderDialog, &FileReaderDialog::onCloseFile);
    connect(this, &MainWindow::closeAll, tcpClient, &QObject::deleteLater);
    connect(this, &MainWindow::closeAll, analyzer, &QObject::deleteLater);
    connect(this, &MainWindow::closeAll, worker, &QObject::deleteLater);
    connect(this, &MainWindow::tcpAddressChanged, tcpClient, &TcpClient::setAddress);
    connect(tcpClient, &TcpClient::connectedStateChanged, this, &MainWindow::tcpConnected);
    connect(tcpClient, &TcpClient::dataAccuired, analyzer, &TDCAnalyzer::dataAccuired);
    connect(tcpClient, &TcpClient::statusUpdated, this, &MainWindow::tcpStatusUpdated);
    analyzerThread->start();
    tcpClientThread->start();
    worker_thread->start();

    if(startup_options.inl) {
        connect(inlCheckDialog, &InlCheckDialog::retry,
                ui->widgetHistMngr, &HistMngrWidget::inlCheckRequest);
        connect(ui->widgetHistMngr, &HistMngrWidget::inlCheckUpdated,
                inlCheckDialog, &InlCheckDialog::resultUpdated);
    }

    connect(rawDataPeriodicTimer, &QTimer::timeout, this, &MainWindow::rawDataPeriodicRequest);
    rawDataPeriodicTimer->setSingleShot(true);

    connect(this, &MainWindow::tdcTtvxMapUpdated, worker, &HistWorker::tdcTtvxMapUpdated);

    connect(this, &MainWindow::requestInlResults, worker, &HistWorker::requestInlResults);
    connect(worker, &HistWorker::inlResultsUpdated, this, &MainWindow::inl_corrections_updated);

    connect(this, &MainWindow::invertUpdated, analyzer, &TDCAnalyzer::invertUpdated);
    connect(statusTimer, &QTimer::timeout, this, &MainWindow::updateStatus);
    statusTimer->setSingleShot(false);
    statusTimer->start(500);
    rawEventCntTimer.start();
    initTree();
    read_config();
    updateSourceLabel();

    configuration_manager = new ConfigurationManager(this, startup_options.program_type, config.default_root);
    connect(configuration_manager, &ConfigurationManager::switch_to_config,
            this, &MainWindow::switch_to_config);
}

MainWindow::~MainWindow()
{
    writeSettings();

    emit closeAll();
    QElapsedTimer closeTimer;
    closeTimer.start();
    if(!threadQuiting(analyzerThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop analyzer thread";
    }
    if(!threadQuiting(tcpClientThread, 2000, &closeTimer)) {
        qWarning() << "Can't stop TCP-readout thread";
    }
    if(!threadQuiting(worker_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop worker thread";
    }

    delete ui;
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
    configuration_manager->reset(config.program_index, config.configuration_name);
}

void MainWindow::switch_to_config(QString program_index, QString configuration_name)
{
    configuration_manager->hide();
    config.program_index = program_index;
    config.configuration_name = configuration_name;

    read_config();
    updateSourceLabel();
}

void MainWindow::on_actionSet_source_file_triggered()
{
    ui->actionDisconnect_from_program->trigger();
    fileReaderDialog->show();
}

void MainWindow::on_actionConnect_to_source_program_triggered()
{
    auto freeProgs = getFreeProgs();
    if(freeProgs.isEmpty()) {
        ui->actionConnect_to_source_program->setEnabled(false);
        return;
    }

    AddPnpDialog dialog(freeProgs, false, this);
    dialog.set_function_filter(std::bind(&MainWindow::isFreeProg, this, std::placeholders::_1));
    connect(pnpServer, &PNPServer::programDiscovered,
            &dialog, &AddPnpDialog::gotProgramDescription);
    connect(pnpServer, &PNPServer::programClosed,
            &dialog, &AddPnpDialog::gotClosedProgram);

    if(dialog.exec() != QDialog::Accepted)
        return;

    auto pd_list = dialog.getProgDescr();
    Q_ASSERT(!pd_list.isEmpty());
    if(pd_list.isEmpty()) return;
    ProgramDescription pd = pd_list.first();
    if(pd.type != selectedProgram.type || pd.index != selectedProgram.index || selectedInterface.isEmpty()) {
        ProgramInterface i = getFreeProgramInterface(pd);
        if(!i.isEmpty()) {
            sendLog(QString("[MW] change source from %1 to %2 (%3)")
                    .arg(selectedProgram.getName()).arg(pd.getName()).arg(i.type));
            disconnectFromProg();
            selectedProgram = pd;
            selectedInterface.type = i.type;
            selectedInterface.id = i.id;
            config.sourse_program_type = selectedProgram.type;
            config.sourse_program_index = selectedProgram.index;
            config.progInterfaceType = selectedInterface.type;
            changeSourceState(SOURCE_STREAM);
            config.write_config();
            emit closeFile();
            emit rawDataEveryEvUpdated(config.rawDataEveryEv);
        }
        updateTcpInterface(pd);
        writeSettings();
    } else {
        qWarning()<<"Selected program has no free interface";
    }
}

void MainWindow::on_actionDisconnect_from_program_triggered()
{
    disconnectFromProg();
    changeSourceState(SOURCE_IDLE);
}

void MainWindow::on_pushButtonResetAll_clicked()
{
    emit resetAll();
}

void MainWindow::tcpStatusUpdated(bool connOk, double dataRate)
{
    tcpConnOk = connOk;
    dataRateLabel->setText(QString("%1 kB/s").arg(dataRate/1024, 0, 'f', 2));
    if(config.source == SOURCE_STREAM)
        setSourceColor(tcpConnOk?"darkGreen":"red");
//    qInfo() << QString("Sem status: anal=%1; writer=%2; hist=%3")
//               .arg(analyzerSem->available())
//               .arg(writerSem->available())
//               .arg(histSem->available());
}

void MainWindow::programDiscovered(const ProgramDescription &newDescr)
{
    if(!SOURCE_TYPES.contains(newDescr.type))
        return;

    if(isFreeProg(newDescr))
        progDescrsMap[newDescr.uuid] = newDescr;
    else
        progDescrsMap.remove(newDescr.uuid);

    updateTcpInterface(newDescr);
    ui->actionConnect_to_source_program->setEnabled(hasFreeProgs());
}

void MainWindow::programClosed(const ProgramDescription &newDescr)
{
    if(progDescrsMap.contains(newDescr.uuid))
        progDescrsMap.remove(newDescr.uuid);
    ui->actionConnect_to_source_program->setEnabled(hasFreeProgs());
}

void MainWindow::updateHitStats(const AnalyzerStats &stats)
{
    bool doResize = lastAnalStat.tdcHits.keys() != stats.tdcHits.keys();
    if(doResize)
        updateModelList(stats);
//    qInfo() << "Decoded events:" << stats.eventDecoded << stats.lastEvNum;
    updateTreeWithStats();

    QString analStr = QString("Ev:%1").arg(stats.eventDecoded);
    if(lastAnalStat.eventDecoded+1 != stats.eventDecoded) {
        analStr += QString("(lastEv=%2)").arg(lastAnalStat.eventDecoded);
    }
    {
        qint64 dt = stats.decodeTimeNs-lastAnalStat.decodeTimeNs;
        qint64 ev = stats.eventDecoded-lastAnalStat.eventDecoded;
        qint64 size = stats.sizeDecoded-lastAnalStat.sizeDecoded;
//        qInfo() << "dt:" << dt;
        analStr.append(QString("rate(%1ev/s; %2MB/s)").arg(ev? (1e9*ev/dt) : 0.,0,'f',0)
                       .arg(size? (1e3*size/dt) : 0.,0,'f',2));
    }
//    if(stats.eventDecoded) {
//        analStr += QString("; Time per ev:%1us; Event: %2B; Size per sec: %3KB")
//                .arg(1e-3*stats.decodeTimeNs/stats.eventDecoded)
//                .arg(stats.sizeDecoded/stats.eventDecoded)
//                .arg(1e6*stats.sizeDecoded/stats.decodeTimeNs);
//    }

    analyzerStatusLabel->setText(analStr);
    lastAnalStat = stats;
}

void MainWindow::rawDataUpdated(const QStringList &data)
{
    ++rawEventCnt;
    ui->plainTextEditRawData->setPlainText(data.join('\n'));
    rawDataReqDone = true;
    ui->plainTextEditRawData->setEnabled(rawDataReqDone);
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);

    if(!config.rawDataEveryEv && config.rawDataPeriodic) {
        rawDataPeriodicTimer->start();
    } else {
        rawDataPeriodicTimer->stop();
    }
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::on_actionCalibrationData_triggered()
{
    calibrationDialog->show();
}

void MainWindow::on_actionReset_all_triggered()
{
    emit resetAll();
}

void MainWindow::inl_corrections_updated(InlResults res, bool allRes)
{
    if(allRes) {
        // actionHPTDC_vol_check
        HptdcVoltageTest test;
        HistContainer hc = ui->widgetHistMngr->getHists();
        HistMngrCfg histCfg = ui->widgetHistMngr->getCfg();
        test.print(fileReaderDialog->getCurrentFile(), lastAnalStat, hc, histCfg, res);
        return;
    }
    SaveInlDialog dialog(this);
    dialog.setInlResults(res);
    if(dialog.exec()) {
        if(dialog.result() == QDialog::Accepted) {
            auto newInl = dialog.getInlResults();
            auto path = dialog.getPath();
            config.calibrationUpdated(newInl, path);
        }
    }
}

void MainWindow::on_actionSave_INL_corrections_triggered()
{
    emit requestInlResults(false);
}

void MainWindow::on_actionShow_INL_check_triggered()
{
    if(inlCheckDialog)
        inlCheckDialog->show();
}

void MainWindow::on_actionSet_INL_separate_location_toggled(bool checked)
{
    config.useSeparateInl = checked;
    if(config.useSeparateInl) {
        auto dir = QFileDialog::getExistingDirectory(this, "Set folder with separate INL calibrations...", config.separateInlFolder);
        if(!dir.isEmpty()) {
            config.separateInlFolder = dir;
        } else {
            ui->actionSet_INL_separate_location->setChecked(false);
        }
    }
    config.reloadCalibration();
    config.write_config();
    ui->widgetHistMngr->updateInlCorr(config.allCal);
    emit inlConfigUpdated(config.allCal);
    updateInlMarks();
}

void MainWindow::on_actionHPTDC_vol_check_triggered()
{
    emit requestInlResults(true);
}

void MainWindow::on_actionSet_TDC_TTVXS_map_triggered()
{
    TdcTtvxsMapDialog dialog(lastAnalStat, config.tdc_ttvxs_map, this);
    if(dialog.exec()){
        config.tdc_ttvxs_map = dialog.getDevMap();
        config.write_config();
        emit tdcTtvxMapUpdated(config.tdc_ttvxs_map);
        updateTtvxsMarks();
    }
}

void MainWindow::on_actionShow_skip_info_toggled(bool show)
{
    emit set_show_skip_info(show);
}

void MainWindow::on_actionSatistic_dialog_triggered()
{
    statusDialog->show();
}

void MainWindow::on_actionHPTDC_errors_triggered()
{
    QDialog *d = new QDialog(this, Qt::Dialog);
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(d);
    //    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d->setLayout(mainLayout);
    d->show();

    //    QMessageBox::about (this, "Raw data format", html);
}

//void MainWindow::on_actionTxt_data_format_triggered()
//{
//    QString s;
//    s += "<h2>TDC2 Txt Data format</h2>";
//    s += "<p>File constructs with following lines:</p>";

//    s += "<ul>";
//    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
//    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
//    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
//    s += "</ul>";
//    s += "<p>where:</p>";
//    s += "<dl>";
//    s += "<dt><b>TAI time</b></dt>";
//    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
//    s += "<dt><b>ch</b></dt>";
//    s += "<dd><i>Channel number in range 0-15</i></dd>";
//    s += "<dt><b>tdcLeadTime</b></dt>";
//    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
//    s += "<dt><b>adcTime</b></dt>";
//    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
//    s += "<dt><b>sample</b></dt>";
//    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
//    s += "</dl>";

//    QMessageBox::about (this, "Txt data format", s);
//}

//void MainWindow::on_actionRaw_data_format_triggered()
//{
//    QDialog d;
//    QPlainTextEdit scrollArea;
//    QLabel htmlLabel;

//    QFile htmlFile(":/rawDataFormat.html");
//    QString html;
//    if(htmlFile.open(QIODevice::ReadOnly)){
//        html = QString(htmlFile.readAll());
//        htmlFile.close();
//        if(html.isEmpty())
//            html = "<b>Channel map file is empty.</b>";
//    } else
//        html = "<b>There is no channel map file.</b>";

//    scrollArea.appendHtml(html);
//    htmlLabel.setText(html);

//    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
////    mainLayout->addWidget(&scrollArea);
//    mainLayout->addWidget(&htmlLabel);
//    d.setLayout(mainLayout);
//    d.exec();

////    QMessageBox::about (this, "Raw data format", html);
//}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::on_radioButtonRawEvery_toggled(bool checked)
{
    config.rawDataEveryEv = checked;
    config.write_config();
    emit rawDataEveryEvUpdated(config.rawDataEveryEv);

    updateRawDataControl();
}

void MainWindow::on_pushButtonRequestRaw_clicked()
{
    rawDataReqDone = false;
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);
    ui->plainTextEditRawData->setEnabled(rawDataReqDone);
    emit requestRawData(config.rawDataLimit);
}

void MainWindow::on_spinBoxRawEvents_valueChanged(int limit)
{
    config.rawDataLimit = limit;
    config.write_config();
}

void MainWindow::on_checkBoxRawPeriodical_toggled(bool checked)
{
    config.rawDataPeriodic = checked;
    config.write_config();
    updateRawDataControl();
    if(!config.rawDataEveryEv && config.rawDataPeriodic)
        rawDataPeriodicTimer->start();
    else
        rawDataPeriodicTimer->stop();
}

void MainWindow::on_spinBoxRawTimer_valueChanged(int val)
{
    if(config.rawDataPeriod != val) {
        config.rawDataPeriod = val;
        config.write_config();
    }
    rawDataPeriodicTimer->setInterval(config.rawDataPeriod*1000);
}

void MainWindow::sourceFileNameUpdated(QString fn)
{
    if(!fn.isEmpty())
        changeSourceState(SOURCE_FILE);
    updateSourceLabel();
}

void MainWindow::tcpConnected(bool connected)
{
    tcpConnOk = connected;
    if(config.source == SOURCE_STREAM)
        setSourceColor(tcpConnOk?"darkGreen":"red");
}

void MainWindow::rawDataPeriodicRequest()
{
    if(!config.rawDataEveryEv && rawDataReqDone && config.rawDataPeriodic)
        on_pushButtonRequestRaw_clicked();
}

void MainWindow::updateStatus()
{
    double rate = 1e3*rawEventCnt/rawEventCntTimer.restart();
    rawEventCnt=0;
    if(rate>100 && config.rawDataEveryEv) {
        qWarning() << QString("Cancel rawData on every event; rate:%1Hz").arg(rate);
        ui->radioButtonRawRequest->setChecked(true);
    }

    semStatusLabel->setText(QString("a:%1;h:%2;w:%3")
                            .arg(analyzerSem->available())
                            .arg(histSem->available())
                            .arg(writerSem->available()));
}

void MainWindow::writeSettings()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.setValue("splitterState", ui->splitter->saveState());
    settings.setValue("tabIndex", ui->tabWidget->currentIndex());
    settings.endGroup();

    config.write_config();
}

void MainWindow::read_config()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    ui->splitter->restoreState(settings.value("splitterState").toByteArray());
    int tabCurrentIndex = settings.value("tabIndex").toInt();
    settings.endGroup();

    config.read_config();

    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
        .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    updateRawDataControl();
    if(config.useSeparateInl) {
        ui->actionSet_INL_separate_location->blockSignals(true);
        ui->actionSet_INL_separate_location->setChecked(config.useSeparateInl);
        ui->actionSet_INL_separate_location->blockSignals(false);
    }
    (config.rawDataEveryEv ? ui->radioButtonRawEvery : ui->radioButtonRawRequest)
            ->setChecked(true);
    ui->tabWidget->setCurrentIndex(tabCurrentIndex);
    ui->spinBoxRawEvents->setValue(config.rawDataLimit);
    ui->checkBoxRawPeriodical->setChecked(config.rawDataPeriodic);
    ui->spinBoxRawTimer->setValue(config.rawDataPeriod);
    rawDataPeriodicTimer->setInterval(config.rawDataPeriod*1000);
    if(config.rawDataPeriodic) rawDataPeriodicTimer->start();

    fileReaderDialog->configUpdated();
    fileReaderDialog->setPackSizeLimit(PACK_SIZE_LIMIT);

    selectedProgram.clear();
    selectedProgram.type = config.sourse_program_type;
    selectedProgram.index = config.sourse_program_index;
    selectedInterface.type = config.progInterfaceType;

    emit invertUpdated(config.invertMap);
    emit tdcTtvxMapUpdated(config.tdc_ttvxs_map);
    prev_invert_map = config.invertMap;
    emit inlConfigUpdated(config.allCal);
    emit set_can_skip(config.source==SOURCE_STREAM);
    updateInlMarks();
    updateTtvxsMarks();
}

void MainWindow::initTree()
{
    QTreeWidget *tree = ui->treeWidgetChCfg;
    tree->setColumnCount(tree_tdc_num_cols);
    QStringList headerLabels;
    headerLabels << "name"
                 << "ch"
                 << "Inv"
                 << "INL"
                 << "TTVXS"
                 << "Raw Count"
                 << "TDC Hits";
    tree->setHeaderLabels(headerLabels);
    tree->hideColumn(tree_tdc_ch);
    tree->hideColumn(tree_tdc_raw_cnt);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    rootTreeItem->setFlags(rootTreeItem->flags()|Qt::ItemIsUserCheckable);
    rootTreeItem->setCheckState(tree_tdc_inv, Qt::PartiallyChecked);
    tree->resizeColumnToContents(tree_tdc_name);
    tree->resizeColumnToContents(tree_tdc_inv);
    tree->resizeColumnToContents(tree_tdc_inl);
    tree->resizeColumnToContents(tree_tdc_ttvxs);
}

void MainWindow::updateModelList(const AnalyzerStats &analStat)
{
    QTreeWidget *tree = ui->treeWidgetChCfg;

    tree->blockSignals(true);
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        rootTreeItem->removeChild(i);
        delete i;
    }
    modTreeItem.clear();

    int modChOffest = 0;
    int devCnt = 0;
    for (const DeviceIndex &index: analStat.tdcHits.keys()) {
        QStringList list;
        list << QString("%1: %2").arg(++devCnt).arg(index.getIdent())
             << index.getSerialStr()
             << "" << "-" << "0" << "0";
        QTreeWidgetItem * const modItem = new QTreeWidgetItem(rootTreeItem, list);
        modItem->setFlags(modItem->flags()|Qt::ItemIsUserCheckable);
        modItem->setCheckState(tree_tdc_inv, Qt::Unchecked);
        modItem->setExpanded(false);
        modTreeItem[index] = modItem;
        ui->widgetHistMngr->updateInlCorr(index, config.allCal.value(index).inl);
//        emit inlConfigUpdated(config.allCal);

        const int modNch = getDeviceChNum(index);
        for(int ch=0; ch<modNch; ++ch){
            QStringList list;
            list << QString::number(ch)
                 << QString::number(modChOffest+ch)
                 << "" << "-" << "0";
            auto treeItem = new QTreeWidgetItem(modItem, list);
            treeItem->setData(tree_tdc_name, Qt::UserRole, ch);
            treeItem->setFlags(treeItem->flags()|Qt::ItemIsUserCheckable);
            treeItem->setCheckState(tree_tdc_inv, config.invertMap[index][ch] ? Qt::Checked :Qt::Unchecked);
        }

        modChOffest += modNch;
    }
    tree->blockSignals(false);
    tree->resizeColumnToContents(tree_tdc_name);
    tree->resizeColumnToContents(tree_tdc_ch);
    updateTreeItemStates();
    updateInlMarks();
    updateTtvxsMarks();
}

void MainWindow::on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column)
{
    if(column != tree_tdc_inv)
        return;
    //todo
//    bakConfig = config;
    Qt::CheckState state = item->checkState(tree_tdc_inv);
    if(state == Qt::PartiallyChecked)
        return;
    bool checked = (state == Qt::Checked);
    if(item==rootTreeItem) {
        for (const DeviceIndex &index: modTreeItem.keys()) {
            if(modTreeItem[index]->checkState(tree_tdc_inv) != state)
                modTreeItem[index]->setCheckState(tree_tdc_inv, state);
        }
    } else {
        QTreeWidgetItem *p = item->parent();
        bool found=false;
        for (const DeviceIndex &index: modTreeItem.keys()){
            if(item==modTreeItem[index]){
                // Module item
                found = true;
                auto &chInvertMap = config.invertMap[index];
                for (int ch: chInvertMap.keys())
                    chInvertMap[ch] = checked;
                ui->treeWidgetChCfg->blockSignals(true);
                for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
                    item->child(childIndex)->setCheckState(tree_tdc_inv, state);
                }
                ui->treeWidgetChCfg->blockSignals(false);
            } else if(p==modTreeItem[index]) {
                // Channel item
                found = true;
                bool ok;
                int modCh = item->data(tree_tdc_name, Qt::UserRole).toInt(&ok);
                auto &chInvertMap = config.invertMap[index];
                Q_ASSERT(ok);
//                Q_ASSERT(modCh<=ds.getNCh());
                chInvertMap[modCh] = checked;
            }
            if(found) {
                updateTreeItemStates();
                break;
            }
        }
        Q_ASSERT(found);
    }
    if(prev_invert_map != config.invertMap) {
        emit invertUpdated(config.invertMap);
        config.write_config();
        prev_invert_map = config.invertMap;
    }

//    updateGuiVal();
}

void MainWindow::updateTreeWithStats()
{
    ui->treeWidgetChCfg->setUpdatesEnabled(false);
//    const th1Ptr h_tdc_input_ch = hmgr->getHist("tdc_input_ch");
    qint64 totRaw=0,totTdc=0;
    int modChOffset=0;

    for (const DeviceIndex &index: lastAnalStat.tdcHits.keys()){
        const TdcHits &s = lastAnalStat.tdcHits[index];
        qint64 modRaw=0,modTdc=0;
        const QMap<int, quint32> &chCntMap = s.hits;
        if(!modTreeItem.contains(index))
            continue; // Reset was invoked recently
        const int modNch = static_cast<int>(getDeviceChNum(index));
        QTreeWidgetItem *modItem = modTreeItem[index];
        for(int ch=0; ch<modNch; ++ch){
            QTreeWidgetItem *chItem = modItem->child(ch);
            quint32 n = chCntMap[ch];
            chItem->setText(tree_tdc_tdc_hits, QString::number(n));
            modTdc += n;
        }
        totRaw += modRaw;
        totTdc += modTdc;
        modItem->setText(tree_tdc_raw_cnt, QString::number(modRaw));
        modItem->setText(tree_tdc_tdc_hits, QString::number(modTdc));
        modChOffset += modNch;
    }
    rootTreeItem->setText(tree_tdc_raw_cnt, QString::number(totRaw));
    rootTreeItem->setText(tree_tdc_tdc_hits, QString::number(totTdc));
    ui->treeWidgetChCfg->setUpdatesEnabled(true);
}

void MainWindow::updateTreeItemStates()
{
    ui->treeWidgetChCfg->blockSignals(true);
    Qt::CheckState allSt = rootTreeItem->checkState(tree_tdc_inv);
    Qt::CheckState st = Qt::PartiallyChecked;
    bool firstMod=true;
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        for(int chIndex=0; chIndex<i->childCount(); ++chIndex){
            Qt::CheckState childState = i->child(chIndex)->checkState(tree_tdc_inv);
            if(chIndex==0){
                st = childState;
                continue;
            }
            if(st!=childState){
                st = Qt::PartiallyChecked;
                break;
            }
        }
        i->setCheckState(tree_tdc_inv, st);
        if(firstMod) {
            allSt = st;
            firstMod = false;
        } else if(allSt != st)
            allSt = Qt::PartiallyChecked;
    }

    rootTreeItem->setCheckState(tree_tdc_inv, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void MainWindow::updateInlMarks()
{
    ui->treeWidgetChCfg->blockSignals(true);
    rootTreeItem->data(tree_tdc_inl, Qt::UserRole).toInt();
    int allSt = 0;
    bool firstMod=true;
    for (DeviceIndex index: modTreeItem.keys()) {
        int modSt = 0;
        QTreeWidgetItem * const modItem = modTreeItem[index];
        int nch = modItem->childCount();
        auto devInl = config.allCal.value(index).inl;
        for(int chIndex=0; chIndex<nch; ++chIndex) {
            auto chItem = modItem->child(chIndex);
            bool chHasInl = devInl.value(chIndex).size() == 1024;
            int chSt = chHasInl ? 1 : -1;
            updateItemInl(chItem, chSt);
            if(chIndex==0){
                modSt = chSt;
            }
            if(modSt!=chSt){
                modSt = 0;
            }

        }
        updateItemInl(modItem, modSt);
        if(firstMod) {
            allSt = modSt;
            firstMod = false;
        } else if(allSt != modSt)
            allSt = 0;
    }

    updateItemInl(rootTreeItem, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void MainWindow::updateTtvxsMarks()
{
    ui->treeWidgetChCfg->blockSignals(true);
    rootTreeItem->data(tree_tdc_ttvxs, Qt::UserRole).toInt();
    int allSt = 0;
    bool firstMod=true;
    for (DeviceIndex index: modTreeItem.keys()) {
        int modSt = 0;
        QTreeWidgetItem * const modItem = modTreeItem[index];
        int nch = modItem->childCount();
        auto devHasTtvxs = !config.tdc_ttvxs_map.value(index).isNull();
        for(int chIndex=0; chIndex<nch; ++chIndex) {
            auto chItem = modItem->child(chIndex);
            bool chHasTtvxs = devHasTtvxs;
            int chSt = chHasTtvxs ? 1 : -1;
            updateItemTtvxs(chItem, chSt);
            if(chIndex==0) {
                modSt = chSt;
            }
            if(modSt!=chSt){
                modSt = 0;
            }
        }
        updateItemTtvxs(modItem, modSt);
        if(firstMod) {
            allSt = modSt;
            firstMod = false;
        } else if(allSt != modSt)
            allSt = 0;
    }

    updateItemTtvxs(rootTreeItem, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void MainWindow::sendLog(QString l)
{
    emit addLog(l);
}

void MainWindow::updateSourceLabel()
{
    QString msg, colorStr;
    switch (config.source) {
    case SOURCE_IDLE:
        colorStr = "gray";
        msg = "Idle";
        break;
    case SOURCE_FILE:
        colorStr = "black";
        msg = QString("file:%1").arg(fileReaderDialog->getCurrentFile());
        break;
    case SOURCE_STREAM:
        colorStr = tcpConnOk?"darkGreen":"red";
        msg = QString("stream:%1").arg(selectedProgram.getName());
        msg.append(selectedInterface.isEmpty() ? " (no interface)":
            QString(" (%1 #%2)").arg(selectedInterface.type).arg(selectedInterface.id));
        break;
    }
    dataSourceLabel->setText(msg);
    setSourceColor(colorStr);
    ui->actionDisconnect_from_program->setEnabled(!selectedProgram.isEmpty());
    dataRateLabel->setVisible(config.source == SOURCE_STREAM);
}

void MainWindow::changeSourceState(SOURCE_DATA st)
{
    if(st == config.source)
        return;

    switch (config.source) {
    case SOURCE_FILE:
        fileReaderDialog->hide();
        break;
    case SOURCE_STREAM:
        disconnectFromProg();
        break;
    default:
        break;
    }
    emit set_can_skip(st==SOURCE_STREAM);
    config.source = st;

    updateSourceLabel();
}

QVector<ProgramDescription> MainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for(const ProgramDescription &descr: progDescrsMap.values()) {
        if(isFreeProg(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool MainWindow::hasFreeProgs() const
{
    for(const ProgramDescription &descr: progDescrsMap.values()) {
        if(isFreeProg(descr))
            return true;
    }
    return false;
}

bool MainWindow::isFreeProg(const ProgramDescription &pd) const
{
    if(pd.index == "0x07a9fcfe")
        return !getFreeProgramInterface(pd).isEmpty();
    return !getFreeProgramInterface(pd).isEmpty();
}

ProgramInterface MainWindow::getFreeProgramInterface(const ProgramDescription &pd) const
{
    ProgramInterface i = pd.getProgramInterface(ProgramInterfaceMonitorData);
    if(!i.isEmpty() && i.isFree && i.enabled)
        return i;

    if(pd.type == DaqConfig::getTypeName(DaqConfigMStream)){
        i = pd.getProgramInterface(ProgramInterfaceDataFlow);
        if(!i.isEmpty() && i.isFree && i.enabled)
            return i;
    }

#ifdef ACCEPT_MAIN_DATA_FLOW
    i = pd.getProgramInterface(ProgramInterfaceDataFlow);
    if(!i.isEmpty() && i.isFree && i.enabled)
        return i;
#endif

    return ProgramInterface();
}

void MainWindow::updateTcpInterface(const ProgramDescription &descr)
{
    if(descr.type == selectedProgram.type && descr.index == selectedProgram.index) {
        ProgramInterface i = descr.getProgramInterface(selectedInterface.type);
        if(i.isEmpty())
            return;
        if(i.host != selectedInterface.host
                || i.port != selectedInterface.port) {
            sendLog(QString("[MW] change interface from %1:%2 to %3:%4")
                  .arg(selectedInterface.host.toString()).arg(selectedInterface.port)
                  .arg(i.host.toString()).arg(i.port));

            selectedInterface = i;
            emit tcpAddressChanged(i.host, i.port);
            updateSourceLabel();
        }
    }
}

void MainWindow::updateRawDataControl()
{
    ui->pushButtonRequestRaw->setEnabled(!config.rawDataEveryEv && rawDataReqDone);
    ui->spinBoxRawEvents->setEnabled(!config.rawDataEveryEv);
    ui->checkBoxRawPeriodical->setEnabled(!config.rawDataEveryEv);
    ui->spinBoxRawTimer->setEnabled(!config.rawDataEveryEv && config.rawDataPeriodic);
}

void MainWindow::disconnectFromProg()
{
    selectedProgram = ProgramDescription();
    selectedInterface = ProgramInterface();
    config.sourse_program_index.clear();
    config.sourse_program_type.clear();
    config.progInterfaceType.clear();
    config.write_config();

    sendLog(QString("[MW] disconnect from %1:%2")
          .arg(selectedInterface.host.toString()).arg(selectedInterface.port));

    emit tcpAddressChanged(QHostAddress(), 0);
    emit resetAll();
}

void MainWindow::setSourceColor(QString col)
{
    setLabelColor(dataSourceLabel, col);
    setLabelColor(dataRateLabel, col);
}

void MainWindow::addLogFromSender(QtMsgType severity, const QString &log, const QString &thrName)
{
    // todo: should we implement it?
    Q_UNUSED(severity)
    Q_UNUSED(log)
    Q_UNUSED(thrName)
//    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

//    Qt::GlobalColor color;
//    switch (severity) {
//    case QtCriticalMsg:
//        color = Qt::magenta;
//        break;
//    case QtWarningMsg:
//        color = Qt::red;
//        break;
//    case QtInfoMsg:
//        color = Qt::blue;
//        break;
//    case QtDebugMsg:
//#ifndef QT_DEBUG
//        return;
//#endif
//    default:
//        color = Qt::black;
//    }
}
