#include "SaveInlDialog.h"
#include "ui_SaveInlDialog.h"

#include <QPushButton>
#include <QFileDialog>
#include <QFileInfo>
#include <QSettings>

#include "HptdcErrors.h"

namespace {
const int CH_PER_HPTDC = 8;
enum tree_inl_col {
    tree_inl_name,
    tree_inl_raw_cnt,
    tree_inl_min_max,
    tree_inl_temp,
    tree_inl_errors,
    tree_inl_num_cols
};

double getTempMcuMin(const DeviceCalibration &data, bool skipOneWire=true)
{
    double tempMcu=0;
    for (int id: data.minTemp.keys()) {
        if(skipOneWire && id==0) continue;
        if(qFuzzyIsNull(tempMcu)) {
            tempMcu = data.minTemp[id];
        } else {
            tempMcu = qMin(tempMcu, data.minTemp[id]);
        }
    }
    return tempMcu;
}

double getTempMcuMax(const DeviceCalibration &data, bool skipOneWire=true)
{
    double tempMcu=0;
    for (int id: data.maxTemp.keys()) {
        if(skipOneWire && id==0) continue;
        if(qFuzzyIsNull(tempMcu)) {
            tempMcu = data.maxTemp[id];
        } else {
            tempMcu = qMax(tempMcu, data.maxTemp[id]);
        }
    }
    return tempMcu;
}

}
SaveInlDialog::SaveInlDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::SaveInlDialog),
    rootTreeItem(new QTreeWidgetItem(QStringList("All")))
{
    ui->setupUi(this);

    QTreeWidget *tree = ui->treeWidget;
    tree->setColumnCount(tree_inl_num_cols);
    QStringList headerLabels;
    headerLabels << "name"
                 << "Raw Count"
                 << "Min/Max"
                 << "Temp"
                 << "Errors";
    tree->setHeaderLabels(headerLabels);
//    tree->hideColumn(tree_inl_ch);
//    tree->hideColumn(tree_inl_raw_cnt);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    tree->resizeColumnToContents(tree_inl_name);

    QSettings s;
    s.beginGroup("InlDialog");
    baseDir = s.value("outDir", "").toString();
    separateFile = s.value("separateFile", false).toBool();
    if(separateFile)
        ui->radioButtonSeparate->setChecked(separateFile);
    if(!baseDir.isEmpty())
        ui->lineEditPath->setText(baseDir);
    s.endGroup();
    ui->labelCommonFile->setText(s.fileName());
}

void SaveInlDialog::setInlResults(const InlResults &res)
{
    inlResults = res;
    int devCnt = 0;
    ui->treeWidget->blockSignals(true);
    for(DeviceIndex index: inlResults.keys()) {
        QStringList list;
        list << QString("%1: %2").arg(++devCnt).arg(index.getIdent());

        QTreeWidgetItem * const modItem = new QTreeWidgetItem(rootTreeItem, list);
        modItem->setFlags(modItem->flags()|Qt::ItemIsUserCheckable);
        modItem->setCheckState(tree_inl_name, Qt::Unchecked);

        modTreeItem[index] = modItem;
        double modCnt=0;
        double modMin = 0;
        double modMax = 0;

        for(int ch: inlResults[index].inl.keys()) {
            const auto &codeDens = inlResults[index].inl[ch];
            auto minmax = std::minmax_element(codeDens.begin(), codeDens.end());
            double min = *minmax.first;
            double max = *minmax.second;
            double cnt = std::accumulate(codeDens.begin(), codeDens.end(), 0);
            inlResults[index].hitCnt[ch] = static_cast<uint>(cnt);
            if(!qFuzzyIsNull(cnt)) {
                if(!qFuzzyIsNull(modCnt)) {
                    modMin = qMin(modMin, min);
                    modMax = qMax(modMax, max);
                } else {
                    modMin = min;
                    modMax = max;
                }
                modCnt += cnt;
            }

            QStringList colTextList;
            colTextList << QString::number(ch) << QString::number(cnt,'f',0) << "-" << "-";
            auto treeItem = new QTreeWidgetItem(modItem, colTextList);

            double tempMcuMin = getTempMcuMin(inlResults[index], false);
            double tempMcuMax = getTempMcuMax(inlResults[index], false);

            if(!qFuzzyIsNull(cnt)) {
                treeItem->setText(tree_inl_min_max, QString("%1-%2").arg(min, 0, 'f', 0).arg(max, 0, 'f', 0));
                if(!qFuzzyIsNull(tempMcuMin) && !qFuzzyIsNull(tempMcuMin)) {
                    treeItem->setText(tree_inl_temp, QString("%1%3-%2%3")
                                      .arg(tempMcuMin, 0, 'f', 1)
                                      .arg(tempMcuMax, 0, 'f', 1)
                                      .arg(QChar(0x02da)));
                }
            }
            const uint hptdc_id = static_cast<uint>(ch/CH_PER_HPTDC);
            quint16 errMask = inlResults[index].errMask.value(hptdc_id, 0);
            if(errMask) {
                QStringList errList;
                for(int bit=0; errMask; ++bit){
                    if(errMask & 1) {
                        errList << HPTDC_ERR_STR[bit];
                    }
                    errMask >>= 1;
                }
                treeItem->setText(tree_inl_errors, QString("0x%1").arg(errMask,4, 16, QChar('0')));
                auto brush = treeItem->foreground(tree_inl_errors);
                brush.setColor(Qt::red);
                treeItem->setForeground(tree_inl_errors, brush);
                treeItem->setToolTip(tree_inl_errors, errList.join('\n'));
            }
            treeItem->setData(tree_inl_name, Qt::UserRole, ch);
            treeItem->setFlags(treeItem->flags()|Qt::ItemIsUserCheckable);
            bool isChOk = cnt>1e6;
            if(!qFuzzyIsNull(cnt) && min < 10) {
                isChOk &= false;
                auto brush = treeItem->foreground(tree_inl_min_max);
                brush.setColor(Qt::red);
                treeItem->setForeground(tree_inl_min_max, brush);
            }
            treeItem->setCheckState(tree_inl_name, isChOk ? Qt::Checked :Qt::Unchecked);
        }
        modItem->setExpanded(!qFuzzyIsNull(modCnt));
        modItem->setText(tree_inl_raw_cnt, QString::number(modCnt,'f',0));
        modItem->setText(tree_inl_min_max, QString("%1-%2")
                         .arg(modMin, 0, 'f', 0).arg(modMax, 0, 'f', 0));

    }

    updateButtonEnable();
    ui->treeWidget->blockSignals(false);
    updateTreeItemStates();
    ui->treeWidget->resizeColumnToContents(tree_inl_name);
}

InlResults SaveInlDialog::getInlResults() const
{
    InlResults results;
    for(auto index : modTreeItem.keys()) {
        auto item = modTreeItem[index];
        DeviceCalibration devCal;
        bool ok;

        for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
            auto childItem = item->child(childIndex);
            if(childItem->checkState(tree_inl_name) == Qt::Checked) {
                int ch=childItem->text(tree_inl_name).toInt(&ok);
                Q_ASSERT(ok);
                if(inlResults[index].hitCnt[ch])
                    devCal.inl[ch] = inlResults[index].inl[ch];
            }
        }
        if(!devCal.inl.isEmpty()) {
            devCal.index = index;
            devCal.minTemp = inlResults[index].minTemp;
            devCal.maxTemp = inlResults[index].maxTemp;
            devCal.fw_versions = inlResults[index].fw_versions;
            devCal.bmc_versions = inlResults[index].bmc_versions;
            results[index] = devCal;
        }
    }
    return results;
}

QString SaveInlDialog::getPath() const
{
    return separateFile ? baseDir : QString();
}

SaveInlDialog::~SaveInlDialog()
{
    delete ui;
}

void SaveInlDialog::on_treeWidget_itemChanged(QTreeWidgetItem *item, int column)
{
    if(column != tree_inl_name) return;

    Qt::CheckState state = item->checkState(tree_inl_name);
    if(state == Qt::PartiallyChecked)
        return;
    if(item==rootTreeItem) {
        for (const DeviceIndex &index: modTreeItem.keys()) {
            if(modTreeItem[index]->checkState(tree_inl_name) != state)
                modTreeItem[index]->setCheckState(tree_inl_name, state);
        }
    } else {
        QTreeWidgetItem *p = item->parent();
        bool found=false;
        for (const DeviceIndex &index: modTreeItem.keys()){
            if(item==modTreeItem[index]) {
                // Module item
                found = true;
                ui->treeWidget->blockSignals(true);
                for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
                    item->child(childIndex)->setCheckState(tree_inl_name, state);
                }
                ui->treeWidget->blockSignals(false);
            } else if(p==modTreeItem[index]) {
                // Channel item
                found = true;
            }
            if(found) {
                updateTreeItemStates();
                break;
            }
        }
        Q_ASSERT(found);
    }
}

void SaveInlDialog::on_radioButtonSeparate_toggled(bool checked)
{
    separateFile = checked;
    QSettings s;
    s.beginGroup("InlDialog");
    s.setValue("separateFile", separateFile);
    s.endGroup();
    ui->labelCommonFile->setEnabled(!separateFile);
    ui->labelOutPath->setEnabled(separateFile);
    ui->lineEditPath->setEnabled(separateFile);
    ui->pushButtonChangeOutputFolder->setEnabled(separateFile);
}

void SaveInlDialog::on_pushButtonChangeOutputFolder_clicked()
{
    QString dirStr = QFileDialog::getExistingDirectory(this, "Select output directory", baseDir, QFileDialog::ShowDirsOnly);
    if(dirStr.isEmpty())
        return;

    QFileInfo dirInfo(dirStr);
    if(dirInfo.exists() && dirInfo.isWritable()) {
        baseDir = dirStr;
        ui->lineEditPath->setText(dirStr);
        QSettings s;
        s.beginGroup("InlDialog");
        s.setValue("outDir", dirStr);
        s.endGroup();
    }
}

void SaveInlDialog::on_buttonBox_accepted()
{

}

void SaveInlDialog::updateTreeItemStates()
{
    const int col = tree_inl_name;

    ui->treeWidget->blockSignals(true);
    Qt::CheckState allSt = rootTreeItem->checkState(col);
    Qt::CheckState st = Qt::PartiallyChecked;
    bool firstMod=true;
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        for(int chIndex=0; chIndex<i->childCount(); ++chIndex){
            Qt::CheckState childState = i->child(chIndex)->checkState(col);
            if(chIndex==0){
                st = childState;
                continue;
            }
            if(st!=childState){
                st = Qt::PartiallyChecked;
                break;
            }
        }
        i->setCheckState(col, st);
        if(firstMod) {
            allSt = st;
            firstMod = false;
        } else if(allSt != st)
            allSt = Qt::PartiallyChecked;
    }

    rootTreeItem->setCheckState(col, allSt);
    ui->treeWidget->blockSignals(false);
    updateButtonEnable();
}

void SaveInlDialog::updateButtonEnable()
{
    QPushButton *pb = ui->buttonBox->button(QDialogButtonBox::Ok);
    Q_ASSERT(pb);

    pb->setEnabled(rootTreeItem->checkState(tree_inl_name) != Qt::Unchecked);
}
