#include "TDCAnalyzer.h"

#include <iostream>

#include <QByteArray>
#include <QDebug>
#include <QSemaphore>
#include <QTimer>

#include "ConfigParams.h"
#include "DeviceHelper.h"
#include "HptdcErrors.h"
#include "mregdev/qtmregdevice.h"
#include "../tdc2/TdcRegs.h"

namespace  {
// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}
QString get_raw_data_prefix(const quint32 *word,const void *firstWord)
{
    const long ev_hdr_offset = word - reinterpret_cast<const quint32*>(firstWord);
    return QString("[%1] %2").arg(ev_hdr_offset,3,10,QChar('0')).arg(*word, 8, 16, QChar('0'));
}
}

TDCAnalyzer::TDCAnalyzer(QSemaphore *_analyzerSem, QSemaphore *_histSem, QSemaphore *_writerSem, QObject *parent) :
    QObject(parent),
    analyzerSem(_analyzerSem),
    histSem(_histSem),
    writerSem(_writerSem),
    statTimer(new QTimer(this)),
    readTimer(new QTimer(this)),
    prevEventNum(0),
    enabled(true),
    signalInverted(false),
    rawDataEvents(0),
    printBadEvent(true),
    rawDataEveryEv(false),
    clear_err_stack(0),
    can_skip(true),
    show_skip_info(false)
{
    connect(statTimer, &QTimer::timeout, this, &TDCAnalyzer::sendStat);
    connect(readTimer, &QTimer::timeout, this, &TDCAnalyzer::read);
    statTimer->start(500);
}

void TDCAnalyzer::dataAccuired(QByteArray newData)
{
    if(data.size())
        data.append(newData);
    else
        data = newData;
    if(enabled){
        read();
    }

    analyzerSem->release();
}


void TDCAnalyzer::read()
{
    curPos = 0;
    curEventStartPos = 0;
//    qInfo()<<"read begin:"<<rdata.size();

    QElapsedTimer decodeTime;
    decodeTime.start();
    quint32 eventDecoded=0;
    while(true) {
        try{
            if(rawDataEveryEv)
                rawDataEvents = 1;
            if(!readEvent()) {
//                qInfo()<<"End of data";
                break;
            }
            ++eventDecoded;
            if(decodedData.event_cnt > DECODED_EVENTS_LIMIT)
                flushData();
        } catch(const char* message) {
            ++analyzerStats.formatErrorCnt;

            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(mess);
            } else{
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        } catch(const QString message) {
            ++analyzerStats.formatErrorCnt;
            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(message);
            } else {
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        }
    }

    analyzerStats.decodeTimeNs += decodeTime.nsecsElapsed();
    analyzerStats.eventDecoded += eventDecoded;
    analyzerStats.sizeDecoded += curEventStartPos;
    ++analyzerStats.decodeLoop;

    if(curEventStartPos) {
        if(curEventStartPos==static_cast<uint>(data.size()))
            data.clear();
        else
            data = data.right(data.size()-static_cast<int>(curEventStartPos));
    }

    flushData();
}

void TDCAnalyzer::resetStats()
{
    analyzerStats.clear();
}

void TDCAnalyzer::sendStat()
{
    if((++clear_err_stack)%10==0)
        lastErrStr.clear();
    printBadEvent = true;
    analyzerStats.timer.start();
    emit deviceStatsUpdated(analyzerStats);
}

void TDCAnalyzer::requestRawData(int events)
{
    rawDataTxt.clear();
    rawDataEvents = events;
}

void TDCAnalyzer::setRawDataEveryEv(bool everyEv)
{
    this->rawDataEveryEv = everyEv;
}

void TDCAnalyzer::invertUpdated(InvertMapType _invertMap)
{
    invertMap = _invertMap;
}

void TDCAnalyzer::set_can_skip(bool skip)
{
    can_skip = skip;
}

void TDCAnalyzer::set_show_skip_info(bool show)
{
    show_skip_info = show;
}

bool TDCAnalyzer::findNextEventHeader()
{
    uint shift = 0;
    const uint dataSize = static_cast<uint>(data.size());
    const char * dataStart = data.data();
    uint pos;
    while(true) {
        shift += 2;
        pos = curEventStartPos+shift;
        if(pos+sizeof(MpdEventHdr) > dataSize) {
            break;
        }
        mpdEventHdr = reinterpret_cast<const MpdEventHdr *>(dataStart+pos);
        if(mpdEventHdr->sync == MPD_SYNC_NORM || mpdEventHdr->sync == MPD_SYNC_EOB) {
            curPos = curEventStartPos + shift;

            QString mess = QString("Syncword was found at 0x%1; Shift:0x%2")
                    .arg(curEventStartPos,0,16).arg(shift,0,16).toStdString().c_str();
            qInfo()<<mess;

            return true;
        }
    }
    curEventStartPos = dataSize;
    return false;
}

bool TDCAnalyzer::readEvent()
{
    curEventStartPos = curPos;
    quint32 bRead;
    quint32 leftSize;
    if(static_cast<uint>(data.size())<curPos+sizeof(MpdEventHdr))
        return false;

    mpdEventHdr = reinterpret_cast<MpdEventHdr *>(data.data()+curPos);

    if(mpdEventHdr->sync != MPD_SYNC_NORM) {
        if(mpdEventHdr->sync == MPD_SYNC_EOB ||
                mpdEventHdr->sync == MPD_SYNC_RUN_START ||
                mpdEventHdr->sync == MPD_SYNC_RUN_STOP){
            curPos += mpdEventHdr->length + sizeof(MpdEventHdr);
            return true;
        }
        QString mess = QString("Format error: wrong MPD_SYNC_WORD: %1; at file pos:%2")
                .arg(mpdEventHdr->sync, 0, 16).arg(curPos);
        throw(mess.toStdString().c_str());
    }

    if(static_cast<uint>(data.size()) < curPos+mpdEventHdr->length+sizeof(MpdEventHdr))
        return false;

//    skip_data = mpdEventHdr->evNum%10 == 9;
    if(can_skip && (rawDataEveryEv || rawDataEvents==0) && skip_work(analyzerSem)) {
        ++analyzerStats.skip;
        curPos += mpdEventHdr->length + sizeof(MpdEventHdr);
        return true;
    } else {
        ++analyzerStats.pass;
        previousProceed=true;
    }

    if(rawDataEvents) {
//        printRawTxt(sizeof(MpdEventHdr), "MpdEventHdr");
        rawDataTxt << QString("\t%1:").arg("MpdEventHdr");
        printRawWord(&mpdEventHdr->sync,":\t\tSyncWord");
        printRawWord(&mpdEventHdr->length, QString("\t\tlen:%1")
                     .arg(mpdEventHdr->length));
        printRawWord(&mpdEventHdr->evNum, QString("\t\tevNum:%1")
                     .arg(mpdEventHdr->evNum));
    }

    curPos += sizeof(MpdEventHdr);
    leftSize = mpdEventHdr->length;

//    qInfo()<<"Event number is "<<mpdEventHdr->evNum;
    curCompleteEv.clear();
    curCompleteEv.evNum = mpdEventHdr->evNum;
    while(leftSize) {
        bRead = readDevice(leftSize);
        if(leftSize < bRead)
            throw("Format error: Sum of MpdDevice blocks is lager then declared in MpdEvent header");
        leftSize -= bRead;
    }
    decodedData.events[curCompleteEv.evNum] = curCompleteEv;
    ++decodedData.event_cnt;
    curCompleteEv.clear();

    prevEventNum = mpdEventHdr->evNum;
    lastEventNumAtPrevFrame = prevEventNum;
    analyzerStats.lastEvNum = prevEventNum;

    if(rawDataEvents) {
        if(--rawDataEvents == 0) {
            emit rawDataUpdated(rawDataTxt);
            rawDataTxt.clear();
        } else
            rawDataTxt << "==============\n";
    }

    return true;
}

void TDCAnalyzer::flushData()
{
    if(decodedData.events.empty())
        return;

    int retryCnt=0;
    while (!histSem->tryAcquire(1, 100)) {
        if((retryCnt++)%10==0)
            qInfo()<<"There is no free space for histogram queue";
    }
    emit deviceDataUpdated(decodedData);
    decodedData.clear();
}

void TDCAnalyzer::decodeTdcBlock(const quint32 *payloadPtr)
{
    quint16 tdcWord = 0;
    int tdcWordCnt=0;
    quint16 tdcTS = 0;
    QString wordStr;
    bool hasHptdcError = false;
    QMap<quint8, QVector<TdcEdge> > chEdgeMap;
    DeviceIndex index = DeviceIndex(mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);
    if(rawDataEvents) {
        printRawTxt(0, "DataBlockHeader (HPTDC)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%2; len:%3")
                     .arg(dataBlockHdr->data_type).arg(dataBlockHdr->data_len));
        if(dataBlockHdr->specific_bits & 0x1)
            rawDataTxt.last().append(" (event FIFO overflow)");
    }
    while(dataBlockHdr->data_len > tdcWord*sizeof(quint32)){
        const quint32 &word = *(payloadPtr+tdcWord);
        ++tdcWordCnt;
        const int dt = word >> 28;
//        qInfo() << tdcWord<< ":  " << hex << word << dec;
        wordStr = QString(":\t\t");
        switch (dt) {
        case 2: // tdc ev hdr
        {
            tdcTS = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF; // evNum in TdcHeader
            if(rawDataEvents)
                wordStr.append(QString("ev hdr evNum:%1").arg(tdcEv));
//            qInfo() << "tdc ev hdr: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
        }
            break;
        case 3: // tdc ev trl
        {
            quint16 wordCnt = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF;
            if(rawDataEvents)
                wordStr.append(QString("ev trl evNum:%1; words:%2").arg(tdcEv).arg(wordCnt));
//            qInfo() << "tdc ev trl: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
            Q_ASSERT(tdcWordCnt == wordCnt);
            tdcWordCnt=0;
        }
            break;
        case 4: // tdc leading edge
//            qInfo() << tdcWord<< ":  " << hex << word << dec;
        case 5: // tdc trl edge
        {
            const quint8 ch = (word>>21)&0x7F;
            const quint32 time = word&0x1FFFFF;
            const bool risening = (dt==5)==signalInverted;
            if(rawDataEvents)
                wordStr.append(QString("%1 ch:%2; t(ps):%L3")
                               .arg(risening?"leading":"trailing").arg(ch,2).arg(24000*time/1024));
            chEdgeMap[ch].push_back(TdcEdge(tdcTS,time, risening));
            break;
        }
        case 6:
        {
            hasHptdcError = true;
            curTdcEv.errors.push_back(word);
            const int tdcId = (word>>24)&0xF;
            QStringList errList;
            for(int errBit=0; errBit< 15; ++errBit){
                if((word & 0x7FFF) & 1<<errBit) {
                    errList << QString("%1\t\t%2").arg(1<<errBit, 14, 16).arg(HPTDC_ERR_STR[errBit]);
                }
            }
            QString errStr = QString("Got HPTDC error (0x%1 tdcId=%2):\n%3")
                    .arg(mpdDeviceHdr->deviceSerial, 8, 16, QChar('0'))
                    .arg(tdcId).arg(errList.join("\n"));
            if(rawDataEvents)
                wordStr.append(errStr);
            if(!lastErrStr.contains(errStr)) {
                qWarning().noquote() << errStr;
                lastErrStr << errStr;
                if(lastErrStr.size() > 10)
                    lastErrStr.removeFirst();
            }

            break;
        }
        default:
            break;
        }
        if(rawDataEvents)
            printRawWord(&word, wordStr);
        ++tdcWord;
    }
    if(hasHptdcError) {
        qWarning() << "Event header:";
        quint16 wordCnt = 0;
        const quint32 *evPtr = reinterpret_cast<const quint32 *>(mpdEventHdr);
        while(wordCnt*sizeof(quint32) < sizeof(MpdEventHdr)){
            const quint32 *wordPtr = evPtr+wordCnt;
            qWarning()  << get_raw_data_prefix(wordPtr, evPtr);
            ++wordCnt;
        }

        qWarning() << "Device event payload:";
        wordCnt = 0;
        const quint32 *devPtr = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
        while(wordCnt*sizeof(quint32) < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length){
            const quint32 *wordPtr = devPtr+wordCnt;
            qWarning()  << get_raw_data_prefix(wordPtr, devPtr);
            ++wordCnt;
        }
    }

    {
        auto &errors = analyzerStats.tdcErrors[index].errors;
        quint32 evNum = mpdEventHdr->evNum;
        for(quint32 word : curTdcEv.errors) {
            auto &errEvList = errors[word];
            if(errEvList.isEmpty() || errEvList.last().second+1 != evNum) {
                errEvList<< QPair<quint32, quint32>(evNum,evNum);
            } else {
                ++errEvList.last().second;
            }
        }
    }

    // Combine TdcEdge into TdcChEvent and fill into
//    bool signal=false;
    bool hasTdcChHit = false;
    QMap<int, bool> &devInvMap = invertMap[index];
    if(devInvMap.isEmpty()) {
        for(int i=0; i<getDeviceChNum(index); ++i)
            devInvMap[i] = false;
    }
    auto &tdcHits = analyzerStats.tdcHits[index].hits;
    for (quint8 ch: chEdgeMap.keys()) {
        for (const TdcEdge &edge: chEdgeMap[ch]) {
            if(edge.rising != devInvMap[ch]) {
                if(hasTdcChHit) {
                    curTdcEv.tdcHits[ch].push_back(curTdcHit);
                    ++tdcHits[ch];
                    curTdcHit.clear();
                } else {
                    hasTdcChHit = true;
                }
                curTdcHit.ch = ch;
                curTdcHit.leadingTime = edge.time;
                curTdcHit.tdcTs = edge.tdcTs;
            } else {
                curTdcHit.trailingTime = edge.time;
                if(!hasTdcChHit)
                    curTdcHit.tdcTs = edge.tdcTs;
                curTdcEv.tdcHits[ch].push_back(curTdcHit);
                ++tdcHits[ch];
                curTdcHit.clear();
                hasTdcChHit = false;
            }
        }
        if(hasTdcChHit) {
            curTdcEv.tdcHits[curTdcHit.ch].push_back(curTdcHit);
            ++tdcHits[curTdcHit.ch];
            curTdcHit.clear();
            hasTdcChHit = false;
        }
    }
}

void TDCAnalyzer::decodeTtvxsBlock(const quint32 *payloadPtr)
{
    if(dataBlockHdr->data_len < sizeof (curTtvxsEv.trig_src))
        throw(QString("Format error: TTVXS Trigger Data Block length (%1) should contains at lest 'Trigger word'")
            .arg(dataBlockHdr->data_len));

    curTtvxsEv.trig_src.all = *payloadPtr;
    if(!analyzerStats.ttvxsList.contains(curTtvxsEv.index))
        analyzerStats.ttvxsList << curTtvxsEv.index;
    if(rawDataEvents) {
        QStringList trigList;
        if(curTtvxsEv.trig_src.bit.pulser) trigList << "pulser";
        if(curTtvxsEv.trig_src.bit.random_pulser) trigList << "random_pulser";
        if(curTtvxsEv.trig_src.bit.external_ttl) trigList << "external_ttl";
        if(curTtvxsEv.trig_src.bit.digital_sfp) trigList << "digital_sfp";
        printRawWord(payloadPtr, QString("\t\ttrigSrc: %1").arg(trigList.join(',')));
        for(uint word=1; word<dataBlockHdr->data_len/sizeof (quint32); ++word) {
            QString suffix;
            switch (word) {
            case 1:
                suffix = QString("\t\ttrigCandid: %L1").arg(*(payloadPtr+word));
                break;
            case 2:
                suffix = QString("\t\ttrigAccept: %L1").arg(*(payloadPtr+word));
                break;
            }
            printRawWord(payloadPtr+word, suffix);
        }
    }
}

void TDCAnalyzer::decodeTdcStat(const quint32 *payloadPtr)
{
    curTdcStat.index = DeviceIndex(mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);
    curTdcStat.regIoTimeout = dataBlockHdr->specific_bits & (1<<16);
    curTdcStat.regIoError = dataBlockHdr->specific_bits & (1<<17);
    if(rawDataEvents) {
        printRawTxt(0, "DataBlockHeader (Stats)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%2; len:%3")
                     .arg(dataBlockHdr->data_type).arg(dataBlockHdr->data_len));

        if(curTdcStat.regIoTimeout)
            rawDataTxt.last().append(" (has RegIO timeout)");
        if(curTdcStat.regIoError)
            rawDataTxt.last().append(" (has RegIO error)");
    }

    int parsed_len = 0;
    while(parsed_len < dataBlockHdr->data_len) {
        const quint32 &word32 = *payloadPtr;
        quint16 addr = word32>>16;
        quint16 val = word32 & 0xFFFF;
        curTdcStat.stats[addr] = val;
        if(rawDataEvents) {
            printRawWord(payloadPtr, QString(":\t\tAddr:%2\tval:%3")
                          .arg(addr, 4, 16, QChar('0'))
                          .arg(val, 4, 16, QChar('0')));
        }
        switch (addr) {
        case REG_TEMPERATURE:
        case REG_BMC_TEMP_1:
        case REG_BMC_TEMP_2:
        case REG_BMC_TEMP_3:
        case REG_BMC_TEMP_4: {
            double temp = addr==REG_TEMPERATURE ? mlink::MregDevice::convertOneWireTemp(val)
                                                : decodeTemp_v1(val);
            int temp_id = addr==REG_TEMPERATURE ? 0
                                                : addr-REG_BMC_TEMP_1+1;
            if(val)
                curTdcEv.temp[temp_id] = temp;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" temp[%1]:%2").arg(temp_id).arg(temp));
            }
            break;
        case REG_FW_REV:
            if(val)
                curTdcEv.fw_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" fw_rev:%1").arg(val));
            break;
        case REG_BMC_REVERSION:
            if(val)
                curTdcEv.bmc_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" bmc_rev:%1").arg(val));
            break;
        }
        ++payloadPtr;
        parsed_len += sizeof(quint32);
    }
}

void TDCAnalyzer::printRawTxt(uint size, QString header, int shift)
{
    if(!rawDataEvents) return;
    if(!header.isEmpty())
        rawDataTxt << QString("\t%1:").arg(header);
    quint32 *word = reinterpret_cast<quint32*>(data.data()+curPos+shift);
    for(uint w=0; w<(size/sizeof(quint32)); ++w){
        printRawWord(word+w);
    }
}

void TDCAnalyzer::printRawWord(const void *word, QString suffix)
{
    rawDataTxt << get_raw_data_prefix(reinterpret_cast<const quint32*>(word),
                                      mpdEventHdr).append(suffix);
}

quint32 TDCAnalyzer::readDevice(uint upperLimitToDevLength)
{
    quint32 byteRead = 0;

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr))
        throw("Format error: There is no data for MpdDeviceHdr");

    mpdDeviceHdr = reinterpret_cast<MpdDeviceHdr *>(data.data()+curPos);
    if(rawDataEvents) {
        printRawTxt(0, "MpdDeviceHdr");
        quint32 *word = reinterpret_cast<quint32 *>(data.data()+curPos);
        printRawWord(word, ":\t\tSerial");
        printRawWord(word+1, QString(":\t\tDevId:0x%1 Len:%2")
                     .arg(mpdDeviceHdr->deviceId, 2, 16, QChar('0'))
                     .arg(mpdDeviceHdr->length));
    }

//    qInfo()<<" Device:"<<hex<<mpdDeviceHdr->deviceSerial<<dec;
    curPos += sizeof(MpdDeviceHdr);
    byteRead += sizeof(MpdDeviceHdr);

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length)
        throw("Format error: There is no data for MpdDevice payload");

    if(mpdDeviceHdr->deviceId != QtMregDevice::DEVICE_ID_TDC72VXS &&
            mpdDeviceHdr->deviceId != QtMregDevice::DEVICE_ID_TDC64VHLE &&
            mpdDeviceHdr->deviceId != QtMregDevice::DEVICE_ID_TQDC16VS_E &&
            mpdDeviceHdr->deviceId != QtMregDevice::DEVICE_ID_TTVXS
            ) {
        if(rawDataEvents) {
            printRawTxt(mpdDeviceHdr->length, "Skipped device block");
        }

        curPos += mpdDeviceHdr->length;
        byteRead += mpdDeviceHdr->length;
        return byteRead;
    }
//    qInfo()<<"mpdDevice: length:"<<mpdDeviceHdr->length<<" serial:"<<hex<<mpdDeviceHdr->deviceSerial;

    quint32 devPayloadLen=0;
    curTdcEv.clear();
    curTtvxsEv.clear();
    curTdcStat.clear();
    curTtvxsEv.evNum = curTdcEv.evNum = mpdEventHdr->evNum;
    curTtvxsEv.index = curTdcEv.index = DeviceIndex(mpdDeviceHdr->deviceId, mpdDeviceHdr->deviceSerial);
    while(mpdDeviceHdr->length > devPayloadLen) {
        if(devPayloadLen+sizeof(MpdMStreamHdr)>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream header(%1):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(sizeof(MpdMStreamHdr)).arg(mpdDeviceHdr->length).arg(devPayloadLen));
        }
        mstreamHdr = reinterpret_cast<MpdMStreamHdr *>(data.data()+curPos);


//        qInfo()<<"mstreamHdr: length:"<<mstreamHdr->words32b*sizeof(quint32);
        if(rawDataEvents && mstreamHdr->subtype !=0xF) {
            printRawTxt(0, "MStreamHdr");
            const quint32 *word = reinterpret_cast<const quint32 *>(mstreamHdr);
            printRawWord(word, QString(":\t\tsubtype:%1; words:%2")
                         .arg(mstreamHdr->subtype).arg(mstreamHdr->words32b));
        }

        curPos += sizeof(MpdMStreamHdr);
        devPayloadLen += sizeof(MpdMStreamHdr);
        const quint32 mStreamWholeSize = mstreamHdr->words32b*sizeof(quint32);
        if(devPayloadLen+mStreamWholeSize>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream payload(%1 32-bit words):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(mstreamHdr->words32b).arg(mpdDeviceHdr->length).arg(devPayloadLen));
        }
        switch (mstreamHdr->subtype) {
        case 0:
        {
//            if(mstreamHdr->words32b*sizeof(quint32) != sizeof(AdcDataType1))
//                throw( QString("Format error: Invalid size(%1 Bytes) of MStream subtype=0 block; should be %2 Bytes.")
//                    .arg(mstreamHdr->words32b*sizeof(quint32)).arg(sizeof(AdcDataType0)).toStdString().c_str());
            tsTAI = reinterpret_cast<AdcDataType1 *>(data.data()+curPos);
            quint32 blockLen = sizeof(AdcDataType1);
            if(rawDataEvents) {
                printRawTxt(0, "TAI");
                QString suffix = QString(":\t\t%2")
                        .arg(QDateTime::fromTime_t(tsTAI->wfTsHi).toString());
                printRawWord(&tsTAI->wfTsHi, suffix);
                suffix = QString(":\t\tnsec:%L1")
                        .arg(tsTAI->wfTsLow,9,10,QChar('0'));
                printRawWord(&tsTAI->wfTsHi+1, suffix);
            }

            while(mStreamWholeSize > blockLen) {
                dataBlockHdr = reinterpret_cast<DataBlockHeader *>(data.data()+curPos+blockLen);
                if(blockLen+sizeof(DataBlockHeader)+dataBlockHdr->data_len > mStreamWholeSize)
                    throw(QString("Format error: TDC Data Block length (%1) is not equal to MStream length:"
                                  "MStream-tsTAI-DataBlockHeader=%2-%3-%4")
                        .arg(dataBlockHdr->data_len).arg(mStreamWholeSize).arg(blockLen).arg(sizeof(DataBlockHeader)));
                blockLen += sizeof(DataBlockHeader);

                const quint32* payloadPtr = reinterpret_cast<const quint32*>(data.data()+curPos+blockLen);
                switch(dataBlockHdr->data_type) {
                case 0:
                    curTdcEv.tai = TAI(tsTAI);
                    decodeTdcBlock(payloadPtr);
                    break;
                case 0xA:
                    curTtvxsEv.tai = TAI(tsTAI);
                    decodeTtvxsBlock(payloadPtr);
                    break;
                case 0xF:
                    decodeTdcStat(payloadPtr);
                    break;
                }
                blockLen += dataBlockHdr->data_len;
            }
            break;
        }
        default:
            if(rawDataEvents) {
                printRawTxt(mstreamHdr->words32b*sizeof(quint32), "MStreamPayload");
            }
            throw( QString("Format error: Invalid subtype number in MStream header: %1.")
                .arg(mstreamHdr->subtype).toStdString().c_str());
        }

        curPos += mstreamHdr->words32b*sizeof(quint32);
//        qInfo()<<"MS block:ev="<<mpdEventHdr->evNum<<"; subtype="<<mstreamHdr->subtype<<"; curpos="<<curpos;
        devPayloadLen += mstreamHdr->words32b*sizeof(quint32);
    }
    if(mpdDeviceHdr->length != devPayloadLen)
        throw(QString("Format error: Sum of sizes of MStream blocks(%1) isn't equal to one declared in MpdDevice header(%2)")
            .arg(devPayloadLen).arg(mpdDeviceHdr->length).toStdString().c_str());

    switch(mpdDeviceHdr->deviceId) {
    case QtMregDevice::DEVICE_ID_TDC72VXS:
    case QtMregDevice::DEVICE_ID_TDC64VHLE:
    case QtMregDevice::DEVICE_ID_TQDC16VS_E:
        curCompleteEv.tdcEv[curTdcEv.index] = curTdcEv;
        curTdcEv.clear();
        break;
    case QtMregDevice::DEVICE_ID_TTVXS:
        curCompleteEv.ttvxsEv[curTtvxsEv.index] = curTtvxsEv;
        curTtvxsEv.clear();
        break;
    }

    if(!curTdcStat.index.isNull()) {
        curCompleteEv.tdcStats[curTdcStat.index] = curTdcStat;
        curTdcStat.clear();
    }

    if(rawDataEvents) {
        rawDataTxt << "";
    }

    byteRead += devPayloadLen;
    return byteRead;
}

void TDCAnalyzer::sendLog(QString s) const
{
    emit addLog(s);
}
