#include "Config.h"
#include "mregdev/qtmregdevice.h"

#include <QSettings>

namespace {
const QString DEV_CONF_PREFIX = "DevConf_";
const QString DEV_CONF_DEFAULT = "DevConfDefault";
}

Config::Config(StartupOptions &startup_options):
    program_type(startup_options.program_type),
    program_version(startup_options.program_version),
    program_index(startup_options.program_index),
    configuration_name(startup_options.configuration_name),
    debug_mode(startup_options.debug_mode),
    mstreamShowTable(true),
    mstreamShowLog(true),
    global_mode(true),
    trace_device_info(false),
    show_power_current(false),
    configFileName(""),
    preset("default"),
    allow_remote_control(true),
    sort_column(-1),
    sort_asc(true),
    show_cfg_widget(true),
    run(false),
    curDev(DeviceIndex())
{
    update_root();
    default_root = root;
    default_root.set_data("configuration_name", QString("default"));
}

Config::~Config()
{
}

bool Config::read_config()
{
    root = BaseConfig::read_config(program_type, program_index, configuration_name);
    if(root.isEmpty()){
        write_config();
    }
    program_index = root.get_data<QString>("program_index");
    configuration_name = root.get_data<QString>("configuration_name");
    run_index = root.get_data<QString>("run_index");
    run_number = root.get_data<int>("run_number");
    run = false;

    mstreamShowTable= root.get_data<bool>("mstreamShowTable");
    mstreamShowLog= root.get_data<bool>("mstreamShowLog");
    global_mode= root.get_data<bool>("global_mode");
    trace_device_info= root.get_data<bool>("trace_device_info");
    show_power_current= root.get_data<bool>("show_power_current");
    configFileName= root.get_data<QString>("configFileName");
    allow_remote_control= root.get_data<bool>("allow_remote_control");
    sort_column = root.get_data<int>("sort_column", sort_column);
    sort_asc = root.get_data<bool>("sort_asc", sort_asc);
    show_cfg_widget = root.get_data<bool>("show_cfg_widget", show_cfg_widget);

    knownSetups.clear();
    for(QString child_name : root.children().keys()){
        if(child_name.split("_").at(0) + "_" + child_name.split("_").at(1) == "current_device"){
            RootConfig device_index_root = root.child(child_name);
            curDev.first = device_index_root.get_data<quint16>("device_id");
            curDev.second = device_index_root.get_data<quint32>("device_serial");
        }

        if(child_name == "known_setups"){
            RootConfig known_setups_root = root.child(child_name);
            for(QString grandson_name : known_setups_root.children().keys()){
                RootConfig device_root = known_setups_root.child(grandson_name);
                DeviceIndex device_index_tmp;
                device_index_tmp.first = device_root.get_data<quint16>("device_id");
                device_index_tmp.second = device_root.get_data<quint32>("device_serial");
                if(device_index_tmp.isNull()) continue;
                DeviceSetup device_setup;
                device_setup.root_to_variables(device_root);
                knownSetups[device_index_tmp] = device_setup;
            }
        }
        if(child_name == "default_setup"){
            RootConfig device_setup_root = root.child(child_name);
            default_setup.root_to_variables(device_setup_root);
        }
    }
    return true;
}

bool Config::write_config()
{
    update_root();
    return BaseConfig::write_config(root);
}

void Config::update_root(){
    root.remove_all_data();
    root.set_data("program_type", program_type);
    root.set_data("program_version", program_version);
    root.set_data("program_index", program_index);
    root.set_data("configuration_name", configuration_name);
    root.set_data("debug_mode", debug_mode);
    root.set_data("run_index", run_index);
    root.set_data("run_number", run_number);
    root.set_data("run", run);

    root.set_data("mstreamShowTable", mstreamShowTable);
    root.set_data("mstreamShowLog", mstreamShowLog);
    root.set_data("global_mode", global_mode);
    root.set_data("trace_device_info", trace_device_info);
    root.set_data("show_power_current", show_power_current);
    root.set_data("configFileName", configFileName);
    root.set_data("allow_remote_control", allow_remote_control);
    root.set_data("sort_column", sort_column);
    root.set_data("sort_asc", sort_asc);
    root.set_data("show_cfg_widget", show_cfg_widget);

    if(!curDev.isNull()){
        RootConfig root_cur_dev;
        root_cur_dev.set_data("device_id", curDev.first);
        root_cur_dev.set_data("device_serial", curDev.second);
        root_cur_dev.item_name = QString("current_device_%1").arg(curDev.getSerialStr());
        root.append_child(root_cur_dev);
    }

    if(!knownSetups.isEmpty()){
        RootConfig root_known_setups;
            for(DeviceIndex device_index : knownSetups.keys()){
                RootConfig root_device;
                root_device = knownSetups[device_index].variables_to_root();
                for(QString name : root_device.children().keys()){
                    if(name.split("_").at(0) == "device"){
                        root_device.remove_child(name);
                    }
                }
                root_device.set_data("device_id", device_index.first);
                root_device.set_data("device_serial", device_index.second);
                root_device.item_name = QString("device_%1").arg(device_index.getSerialStr());
                root_known_setups.append_child(root_device);
            }

        root_known_setups.item_name = "known_setups";
        root.append_child(root_known_setups);
    }

    RootConfig root_default_setup;
    root_default_setup = default_setup.variables_to_root();
    root_default_setup.item_name = "default_setup";
    root.append_child(root_default_setup);
}

DeviceSetup &Config::getConfig(const DeviceIndex &index)
{
    Q_ASSERT(!index.isNull());
    if(!knownSetups.contains(index)) {
        DeviceSetup &ds = knownSetups[index];
        ds.index = index;
        ds.copy_config(default_setup);
    }
    return knownSetups[index];
}

DeviceSetup &Config::getCurrentConfig()
{
    return getConfig(curDev);
}

bool Config::isRun()
{
    for(auto s : knownSetups.values())
        if(s.enabled && s.run) return true;
    return false;
}

bool Config::has_enabled_device() const
{
    for(auto s : knownSetups.values())
        if(s.enabled) return true;
    return false;
}

DeviceIndex Config::get_device_index_by_key(QString key)
{
    if(key.startsWith(DEV_CONF_PREFIX)) {
        const auto var = key.split("_");
        if(var.size() >= 3) {
            bool ok1,ok2;
            const auto id = var[1].toUShort(&ok1, 16);
            const auto ser = var[2].toULongLong(&ok2, 16);
            if(ok1 && ok2)
                return DeviceIndex(id, ser);
        }
    }
    return DeviceIndex();
}


// --- Garbage
