#include "TDCDevice.h"
#include "mregdev/qtmregdevice.h"

#include <QByteArray>
#include <QDateTime>
#include <QFile>
#include <QTextCodec>
#include <QTextStream>
#include <QVector>
#include <math.h>

#include "mregdev/mregdevice.h"
#include "regio/regop.h"
#include <TdcRegs.h>
#include "DeviceFeatures.h"

using mlink::RegOpVector;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;

namespace {
int hptdcSetupIndex = 0;

const qint64 STANDBY_TIMEOUT = 200; // ~100ms for pll reset

const QString HPTDC_ST_STR[16] = {
    "Vernier Error",
    "Coarse Error",
    "Channel Select Error",
    "L1 Buffer Parity Error",
    "Trigger FIFO Parity Error",
    "Trigger Matching State Error",
    "Readout FIFO Parity Error",
    "Readout State Error",
    "Setup Parity Error",
    "Control Parity Error",
    "JTAG Instruction Error",
    "TDC Have Readout Token",
    "Readout FIFO Full",
    "Readout FIFO Empty",
    "Trigger FIFO Full",
    "Trigger FIFO Empty"
};
const quint16 HPTDC_ERR_MASK = 0x7FF;

// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}
double decodeTemp_v2(quint16 val)
{
    double v = static_cast<qint16>(val);
    return v / 32.0;
}

QMap<quint16, quint32> getChMask(const DeviceSetup &setup) {
    QMap<quint16, quint32> chMask;
    const quint16 nch = static_cast<quint16>(setup.getNCh());
    for(quint16 ch=0; ch<nch; ch++) {
        quint16 oct = (nch-1-ch)/8;
        if(!chMask.contains(oct)) chMask[oct] = 0;
        if(setup.chTdcEn.value(ch, false)) {
            const quint32 v = 0xF<<(((nch-1-ch)%8)*4);
            chMask[oct] |= v;
        }
    }
    return chMask;
}
}


TDCDevice::TDCDevice(quint16 devId, quint64 devSerial, bool isDebug):
    QtMregDevice(devId, devSerial),
    debug(isDebug),
    setupCacheValid(false),
    hitCntTime(0),
    isBlock(false)
{
    qRegisterMetaType<DeviceSetup>();
    status.deviceId = devId;
    status.serialId = devSerial;
}

void TDCDevice::writeSetup()
{
    if (!isConnected()) return;
    try {
        bool allOk = true;
        StopRun();
        if(!waitEndOfRun()) {
            setupCacheValid = false;
            if(isConnected()) {
                qWarning("Failed wait device pause (run bit disabled; ev queue isn't empty). Writing setup is failed.");
            } else {
                return;
            }
        }

        RegOpVector regOp;
        if (!setupCacheValid || setupCache.trigsetup.all != setup.trigsetup.all)
            regOp << WriteTrigSetup(setup.trigsetup);

        if (!setupCacheValid || setupCache.trigPeriod != setup.trigPeriod) {
            regOp << RegWrite32(REG32_TRIG_TIMER_PERIOD, setup.trigPeriod);
        }
        if (!setupCacheValid || setupCache.pll_bypass != setup.pll_bypass) {
            quint16 reg_val = 1;
            if(debug && !setup.pll_bypass)
                reg_val &= ~1;
            regOp << RegWrite(REG_ON_BOARD_CLK_CSR, reg_val);
        }

        regOp << writeStatCfg();

        RegOpVector regOpHptdc;
        auto chMask = getChMask(setup), chMaskPrev = getChMask(setupCache);
        if(!setupCacheValid || chMask != chMaskPrev) {
            for(auto oct : chMask.keys()) {
                regOpHptdc << RegWrite32(REG32_TDC_CH_EN_1+oct*2, chMask[oct]);
            }
        }

        if (!setupCacheValid || setupCache.latency != setup.latency
                || setupCache.matchWin != setup.matchWin) {
            regOpHptdc << RegWrite32(REG32_TDC_TRIG_WIN_SETUP,
                                (setup.matchWin&0xFFF)<<12 | (setup.latency & 0xFFF));
        }
        if (!setupCacheValid || setupCache.hptdcHdr != setup.hptdcHdr
                || setupCache.hptdcTrl != setup.hptdcTrl
                || (debug &&
                    setupCache.adjust_set.readout_fifo_size !=
                    setup.adjust_set.readout_fifo_size)) {
            reg_read_ctrl readCtr;
            readCtr.all = regRead32(REG32_TDC_READ_CTRL);

            readCtr.bit.enable_local_header = setup.hptdcHdr;
            readCtr.bit.enable_local_trailer = setup.hptdcTrl;
            if(debug)
                readCtr.bit.readout_fifo_size = setup.adjust_set.readout_fifo_size;

            regOpHptdc << RegWrite32(REG32_TDC_READ_CTRL, readCtr.all);
        }

        if (!setupCacheValid ||
            setupCache.leadEdge != setup.leadEdge ||
            setupCache.trailEdge != setup.trailEdge ||
            (debug && setupCache.adjust_set != setup.adjust_set)) {
            reg_measurement_ctrl measurmentCtr;
            measurmentCtr.all = regRead32(REG32_TDC_MEASURE_CTRL);
            measurmentCtr.bit.enable_leading = setup.leadEdge;
            measurmentCtr.bit.enable_trailing = setup.trailEdge;

            if(debug) {
                measurmentCtr.bit.leading_resolution = setup.adjust_set.leading_resolution;
                measurmentCtr.bit.DLL_Mode = setup.adjust_set.dll_mode;
                measurmentCtr.bit.enable_pair = setup.adjust_set.enable_pair;
                measurmentCtr.bit.core_clock_source = setup.adjust_set.core_clock_source;
                measurmentCtr.bit.dll_clock_source = setup.adjust_set.dll_clock;
                measurmentCtr.bit.mode_rc_compression = setup.adjust_set.mode_rc_compression;
            }

            regOpHptdc << RegWrite32(REG32_TDC_MEASURE_CTRL, measurmentCtr.all);
        }

        if(!regOpHptdc.isEmpty()) {
            regOp << RegWrite32(REG32_TDC_GEN_CTRL, REG32_TDC_GEN_CTRL_BIT_STANDBY);
            if(getDeviceFwVersion() >= REV_HPTDC_STANDBY) {
                regOpExec(regOp);
                regOp.clear();
                if(!waitHptdcInit()) {
                    allOk = false;
                }
            }
            if(allOk) {
                regOp << regOpHptdc;
                regOp << RegWrite32(REG32_TDC_GEN_CTRL, 0);
                standByTime.start();
            }
        }

        if(allOk) {
            regOpExec(regOp);

            if(setup.run) {
                if(!setupCacheValid || !setupCache.run)
                    resetEvNumAndStats();
                StartRun();
            }

//            readoutHptdcSetup();

            writeLiveMagic();
            setupCache = setup;
            setupCacheValid = true;
        }
    }
    catch (std::runtime_error &e) {
        setupCacheValid = false;
        qCritical() << QString("Communication error: %1").arg(e.what());
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }

    writeLiveMagic();
    if(setupCacheValid)
        emit deviceSetupWritten(getDeviceIndex(), setup.setup_index);
    clearStatistics();
}

void TDCDevice::setSetup(const DeviceSetup &s)
{
    if(getDeviceIndex() != s.index) return;
    setup = s;
    writeSetup();
}

void TDCDevice::writeSetupForce()
{
    StopRun();
    setupCacheValid = false;
    on_deviceConnected();
//    writeSetup();
}

void TDCDevice::softReset()
{
    if(!isConnected())
        return;
    setupCacheValid = false;
    RegOpVector op;
    op << RegWrite(REG_DEVICE_CTRL, 1);
    op << RegWrite(REG_DEVICE_CTRL, 0);
    regOpExec(op);
}

void TDCDevice::blockDeviceRun(bool _isBlock)
{
    isBlock = _isBlock;
    if(isBlock) {
        if(isConnected())
            StopRun();
    } else
        writeSetup();
}

void TDCDevice::reset_error_counters()
{
    update_err_cnt();
}

void TDCDevice::statusPoll()
{
    QtMregDevice::statusPoll();
    if(isConnected() && !setupCacheValid)
        writeSetup();

    if(standByTime.isValid()) {
        if(standByTime.elapsed() < STANDBY_TIMEOUT){
            return;
        }
        standByTime.invalidate();
        standByDoneTime.start();
    }

    status.temp.clear();
    status.temp[0] = getTemperature();
    status.connected = isConnected();

    const bool &fwWithNewCnt = status.statistic.fwWithNewCnt = getDeviceFwVersion()>=REV_RUN_COUNTERS;
    quint16 csrCtrlVal = regRead(REG_DEVICE_CTRL);
    RegOpVector v;
    int regPos=0;
    if(fwWithNewCnt) {
        v << RegWrite(REG_DEVICE_CTRL, csrCtrlVal | 2); regPos+=1;
    } else {
        v << RegWrite(REG_TRIG_CSR, REG_TRIG_CSR_BIT_CNT_LOCK); regPos+=1;
    }
    int trigQueuePos  = regPos;
    v << RegRead(fwWithNewCnt ?
                     REG_TRIG_QUEUE_NEW : REG_TRIG_QUEUE); regPos+=1;
    int runStatusPos = regPos;
    v << RegRead(REG_DEVICE_CTRL); regPos+=1;
    const int busyStatusPos = regPos;
    v << RegRead(fwWithNewCnt ?
                     REG_TRIG_BUSY_STATUS_NEW : REG_TRIG_BUSY_STATUS); regPos+=1;
    int curEvNumPos = -1;
    if(fwWithNewCnt) {
        curEvNumPos = regPos;
        v << RegRead64(REG64_TRIG_CUR_EV_NUM); regPos+=4;
    }
    const int generalStatusPos = regPos;
    v << RegRead(REG32_TDC_GENERAL_STATUS); regPos+=1;
    const int clockStatusPos = regPos;
    v << RegRead(REG_CLOCK_STATUS); regPos+=1;
    const int clockSourcePos = regPos;
    v << RegRead(REG_SYS_CLK_SOURCE); regPos+=1;
    int evCandPos = -1;
    int evAcceptedPos = -1;
    if(!fwWithNewCnt) {
        evCandPos = regPos;
        v << RegRead64(REG64_TRIG_CNT_CANDID); regPos+=4;
        evAcceptedPos = regPos;
        v << RegRead64(REG64_TRIG_CNT_ACCEPTED); regPos+=4;
    }

    const int statPos = regPos;
    if(fwWithNewCnt) {
        for(int i=0; i<REG_RUN_STAT_CNT; ++i) {
            v << RegRead64(REG_RUN_STAT_BASE+i*4); regPos+=4;
        }
    }

    if(fwWithNewCnt) {
        v << RegWrite(REG_DEVICE_CTRL, csrCtrlVal & ~0x2); regPos+=1;
    } else {
        v << RegWrite(REG_TRIG_CSR, 0); regPos+=1;
    }
    const int bmcOnboardPllStatusPos = regPos;
    v << RegRead(REG_BMC_ONBOARD_PLL_STATUS); regPos+=1;
    const int bmcPllUnlockPos = regPos;
    v << RegRead(REG_BMC_PLL_UNLOCK_CNT); regPos+=1;
//    const int taiPos = regPos;
//    v << RegRead32(REG_TAI_0); regPos+=2;

    const int bmcVerPos = regPos;
    v << RegRead(REG_BMC_VERSION); regPos+=1;
    v << RegRead(REG_BMC_REVERSION); regPos+=1;
    const int bmcTempPos = regPos;
    v << RegRead(REG_BMC_TEMP_1); regPos+=1;
    v << RegRead(REG_BMC_TEMP_2); regPos+=1;
    v << RegRead(REG_BMC_TEMP_3); regPos+=1;
    v << RegRead(REG_BMC_TEMP_4); regPos+=1;
    const int bmcStatusPos = regPos;
    v << RegRead(REG_BMC_SYSTEM_STATUS); regPos+=1;
    v << RegRead(REG_BMC_POWER_STATUS); regPos+=1;
    v << RegRead(REG_BMC_PLL_STATUS); regPos+=1;
    const int bmcDevIdPos = regPos;
    v << RegRead(REG_BMC_DEV_ID); regPos+=1;
    const int bmcPowerPos = regPos;
    const int power_reg_start = REG_BMC_VOL_TDC_A;
    const int power_reg_end = getDeviceId()==DEVICE_ID_TDC64VHLE ? REG_BMC_CUR_TDC_D : REG_BMC_CUR_BP_33;
    Q_ASSERT(power_reg_start<power_reg_end &&
             (power_reg_end-power_reg_start) < 16);
    for(quint16 r=power_reg_start; r<=power_reg_end; ++r) {
        v << RegRead(r); regPos+=1;
    }

    const int errCntPos = regPos;
    RegOpVector errCntOp = getErrCntRegOp();
    v << errCntOp;
    regPos += errCntOp.size();

    RegOpVector regRes = regOpExec(v);
    if(regRes.size() != v.size()) {
        status.connected = isConnected();
        emit statusUpdated(status);
        return;
    }

    v.clear();
    regPos = 0;

    status.trigQueue = regRes.at(trigQueuePos).data & 0x1f;
    status.runFinished = (regRes.at(runStatusPos).data & REG_DEVICE_CTRL_BIT_RUN) == 0;
    status.runFinished &= status.trigQueue==0;

    status.hptdcChipErrorFlags = (regRes.at(generalStatusPos).data>>16) & 0x1FF;
    status.clockStatus = regRes.at(clockStatusPos).data;
    status.sysClockSource = regRes.at(clockSourcePos).data;
    status.busyStatus = regRes.at(busyStatusPos).data;

    if(curEvNumPos != -1)
        status.curEvNum = get64(regRes, curEvNumPos);

    if(fwWithNewCnt) {
        status.statistic.run_time        = get64(regRes, statPos+REG64_RUN_STAT_RUN_TIME);
        status.statistic.evNumCandidateRun = get64(regRes, statPos+REG64_RUN_STAT_TRIG_CANDID);
        status.statistic.evNumAccepted  = get64(regRes, statPos+REG64_RUN_STAT_ACCEPT);
        status.statistic.evNumBlocked   = get64(regRes, statPos+REG64_RUN_STAT_BLOCKED);
        status.statistic.busy_total     = get64(regRes, statPos+REG64_RUN_STAT_BUSY_TOTAL);
        status.statistic.busy_tdc       = get64(regRes, statPos+REG64_RUN_STAT_BUSY_TDC);
        status.statistic.busy_readout   = get64(regRes, statPos+REG64_RUN_STAT_BUSY_READOUT);
    } else {
        status.statistic.evNumCandidateRun = get64(regRes, evCandPos);
        status.statistic.evNumAccepted = get64(regRes, evAcceptedPos);
    }

    BmcStatus &bmc_status = status.bmc_status;
    bmc_status.dev_id = regRes[bmcDevIdPos].data;

    bmc_status.onboard_pll_status.raw = regRes[bmcOnboardPllStatusPos].data & 0xF;
    bmc_status.pllUnlockCnt = regRes[bmcPllUnlockPos].data & 0xFFF;

    bmc_status.overall_status = regRes[bmcStatusPos].data & 0x3;
    bmc_status.power_status = regRes[bmcStatusPos+1].data & 0x3;
    bmc_status.pll_status = regRes[bmcStatusPos+2].data & 0x3;

    bmc_status.mcuVer.raw = regRes[bmcVerPos].data;
    bmc_status.mcuRev = regRes[bmcVerPos+1].data;
    if(bmc_status.mcuVer.raw) {
        const int MAX_TEMP_CNT =
        getDeviceId() == QtMregDevice::DEVICE_ID_TDC64VHLE ? 2 : 4;
        for(int i=0; i<MAX_TEMP_CNT; ++i) {
            quint16 &rawTemp = regRes[bmcTempPos+i].data;
            if(rawTemp != 0x8000 && rawTemp != 0x7FFF)
                status.temp[1+i] = bmc_status.mcuVer.b.major >= 2 ?
                            decodeTemp_v2(rawTemp) : decodeTemp_v1(rawTemp);
        }
    }

    read_bmc_power(regRes, bmcPowerPos);
    read_err_cnt(regRes, errCntPos, status.statistic.errCnt);

//    status.statistic.timeValidLoss = regRead32(REG_WR_SYNC_LOST_COUNTER);
//    status.statistic.rxCodeError = regRead32(REG_WR_LINK_ERROR_COUNTER);
//    status.statistic.missExtTrig = regRead32(REG_TRIG_ON_XOFF_ERROR_COUNTER);

    RegOpVector htpdcRegOp;
    const int hptdcCnt = getNumberOfChannels()/8;
    for (int hptdcId = 0; hptdcId < hptdcCnt; ++hptdcId) {
        for(int reg=0; reg<4; ++reg) {
            htpdcRegOp << RegRead(REG_HPTDC1_STATUS_1+(hptdcId*4)+reg);
        }
    }

    RegOpVector hptdcRes = regOpExec(htpdcRegOp);

    if(hptdcRes.size() == htpdcRegOp.size()) {
        for (int hptdcId = 0; hptdcId < hptdcCnt; ++hptdcId) {
            for(int reg=0; reg<4; ++reg) {
                status.hptdcStatus[hptdcId].regValues[reg] = hptdcRes[hptdcId*4+reg].data;
            }
        }
    } else {
        qWarning() << "Error while recieving RegOpVector hptdcRes = regOpExec(htpdcRegOp);";
    }

    emit statusUpdated(status);
}

void TDCDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
    status.fwStr = getFirmwareVersionStr();
    // Before this rev. there was no default value for MLink CSR registers
    if(getDeviceFwVersion() > REV_MSTREAM_CFG) {
        status.mstreamMultiAck = hasMultiAckSupport();
        status.mstreamHwBufSize = getHwBufSize();
    }

    setupCacheValid = false;
    writeTime();

    regWrite32(REG32_TDC_GEN_CTRL, REG32_TDC_GEN_CTRL_BIT_STANDBY);
    quint32 measureCtrl = regRead32(REG32_TDC_MEASURE_CTRL);

    {
        measureCtrl &= ~(static_cast<quint32>(0xF)<<3);
        measureCtrl |= 0xA<<3; // width_select
    }
    measureCtrl |= 3<<8; // mode_rc mode_rc_compression
    measureCtrl &= ~(static_cast<quint32>(1)<<13); // disable trailing
//    measureCtrl |= 1<<9; // mode_rc
    regWrite32(REG32_TDC_MEASURE_CTRL, measureCtrl);
    regWrite32(REG32_TDC_GEN_CTRL, 0);

    writeSetup();
    update_err_cnt();

    emit deviceInfoUpdated(getDeviceIndex(), getNumberOfChannels());
}

void TDCDevice::on_deviceDisconnected()
{
    QtMregDevice::on_deviceDisconnected();
    setupCacheValid = false;
}

void TDCDevice::evNumRequest(int checkIndex)
{
    if (!isConnected())
        return;

    const bool &fwWithNewCnt = getDeviceFwVersion()>=REV_RUN_COUNTERS;
    if(!fwWithNewCnt){
        return;
    }

    quint16 csrCtrlVal = setupCache.run ? 0x8000 : 0;
    RegOpVector v;
    v << RegWrite(REG_DEVICE_CTRL, csrCtrlVal | 2);
    v << RegRead64(REG64_TRIG_CUR_EV_NUM);
    v << RegWrite(REG_DEVICE_CTRL, csrCtrlVal);
    auto regRes = regOpExec(v);
    quint64 globEvNum = get64(regRes, 1);
    if(getDeviceFwVersion()<=REV_EVNUM_INCREMENT)
        ++globEvNum;
    emit evNumResponse(checkIndex, getDeviceSerialNumber(), globEvNum);
}

RegOpVector TDCDevice::WriteTrigSetup(reg_trig_ctrl trigsetup)
{
    RegOpVector regOp;
    regOp << RegWrite(REG_TRIG_SRC, trigsetup.all);
    return regOp;
}

void TDCDevice::StartRun()
{
    if(isBlock)
        return;
    RegOpVector v;
//    v << RegWrite(REG_DEVICE_CTRL, 0); // clear RUN
    v << RegWrite(REG_DEVICE_CTRL, REG_DEVICE_CTRL_BIT_RUN); // set RUN
    regOpExec(v);
}

void TDCDevice::StopRun()
{
    regWrite(REG_DEVICE_CTRL, 0); // clear RUN
}

int TDCDevice::getNumberOfChannels() const
{
    int nch;
    switch (getDeviceId()) {
    case DEVICE_ID_TDC72VXS: nch = 72; break;
    case DEVICE_ID_TDC64VHLE: nch = 64; break;
    default: nch = 1; break;
    }
    return nch;
}

void TDCDevice::writeTime()
{
    uint sec = QDateTime::currentDateTime().toTime_t();
    RegOpVector regOp;
    regOp << RegWrite32(REG_TAI_0, sec);
    regOp << RegWrite(REG_TAI_UPDATE, 0);
    regOpExec(regOp);
}

void TDCDevice::readoutHptdcSetup()
{
    const int HPTDC_SETUP_LEN = 41;
    RegOpVector regOp;
    for(int i=0; i<HPTDC_SETUP_LEN*2; ++i)
        regOp << RegRead(0x1000 + i);
    RegOpVector regRes = regOpExec(regOp);
    if(regRes.size() != regOp.size())
        return;

    for(int tdcId=0; tdcId<2; ++tdcId){
        QFile file(QString("hptdc_eth_%1_%2.dat").arg(tdcId?"B":"A").arg(hptdcSetupIndex));
//        QFile file(QString("hptdc_%1_%2.dat").arg("eth").arg(hptdcSetupIndex));
        file.open(QIODevice::WriteOnly);
        QTextStream out(&file);
        for(int i=0; i<HPTDC_SETUP_LEN; ++i) {
            mlink::RegOp op = regRes[HPTDC_SETUP_LEN*tdcId+i];
            out << QString("%1 %2: ")
                   .arg(i*2, 3, 10, QChar('0'))
                   .arg(i*2*8, 3, 10, QChar('0'));
            for(int p=7; p>=0; --p)
                out << ((op.data & (1<<p)) ? "1":"0");
            out << QString(" %1\n").arg(op.data&0xFF, 0, 16);

            out << QString("%1 %2: ")
                   .arg(i*2+1, 3, 10, QChar('0'))
                   .arg(i*2*8+8, 3, 10, QChar('0'));
            for(int p=15; p>=8; --p)
                out << ((op.data & (1<<p)) ? "1":"0");
            out << QString(" %1\n").arg(op.data>>8, 0, 16);
            out.flush();
        }

        file.close();
    }
    ++hptdcSetupIndex;
}

bool TDCDevice::waitEndOfRun()
{
    QElapsedTimer timer;
    timer.start();
    const quint16 regNum = getDeviceFwVersion()>=REV_RUN_COUNTERS ?
                REG_TRIG_QUEUE_NEW : REG_TRIG_QUEUE;
    while(regRead(regNum) & 0xF){
        if(timer.elapsed() > 2000){
            qWarning() << "regRead(REG_TRIG_QUEUE):"<< hex << regRead(regNum)<<dec;
            return false;
        }
    }
    return true;
}

void TDCDevice::resetEvNumAndStats()
{
    RegOpVector regOp;
    if(getDeviceFwVersion()>=REV_RUN_COUNTERS) {
        regOp << RegWrite64(REG64_TRIG_EVENT_NUM_LOAD, 0);
        regOp << RegWrite(REG_TRIG_CSR, REG_TRIG_CSR_BIT_CNT_RESET); // reset stats
        regOp << RegWrite(REG_TRIG_CSR, 0);
    } else {
        regOp << RegWrite32(REG32_TRIG_EVENT_NUM_LOAD, 0);
        regOp << RegWrite(REG_TRIG_CSR, REG_TRIG_CSR_BIT_CNT_RESET); // reset stats
        regOp << RegWrite(REG_TRIG_CSR, 0);
    }
    regOpExec(regOp);
}

RegOpVector TDCDevice::writeStatCfg()
{
    quint16 reg_base = getDeviceFwVersion() >= REV_STAT_PRESET2 ? REG_STAT_BASE_NEW : REG_STAT_BASE;
    RegOpVector regOp;
    if (!setupCacheValid || setupCache.stat_readout != setup.stat_readout) {
        regOp << RegWrite(reg_base+REG_STAT_CTRL, setup.stat_readout ? 1 : 0);
    }
    if (!setupCacheValid || setupCache.stat_period != setup.stat_period) {
        regOp << RegWrite(reg_base+REG_STAT_PERIOD, setup.stat_period);
    }

    if(getDeviceFwVersion() >= REV_STAT_PRESET && getDeviceFwVersion() < REV_60000) {
        if (!setupCacheValid || setupCache.stat_preset.regs != setup.stat_preset.regs) {
            QVector<quint16> regs = setup.stat_preset.regs;
            if(regs.isEmpty()) {
                qCritical() << "Statistic preset is empty; init it with one address equal to 0";
                regs << 0;
            }
            regOp << RegWrite(reg_base+REG_STAT_BLOCK_SIZE,
                              static_cast<quint16>(regs.size()-1));
            for(quint16 i=0; i<regs.size(); ++i) {
                regOp << RegWrite(reg_base+REG_STAT_RAM_ADDR, i);
                regOp << RegWrite(reg_base+REG_STAT_RAM_DATA, regs[i]);
            }
        }
    }
    return regOp;
}

mlink::RegOpVector TDCDevice::getErrCntRegOp() const
{
    RegOpVector regOp;
    regOp << RegRead64(REG64_HWIP_LINK_DOWN);
    regOp << RegRead64(REG64_REGIO_ERR);
    regOp << RegRead64(REG64_SFP_ERR);
    regOp << RegRead64(REG64_HWIP_ERR);
    regOp << RegRead64(REG64_TX_ERR);
    regOp << RegRead64(REG64_RX_ERR);

    return regOp;
}

void TDCDevice::read_err_cnt(const mlink::RegOpVector &regRes, int errCntPos, TdcErrCounters &errCnt)
{
    errCnt.hwipLinkDown = get64(regRes, errCntPos);
    errCnt.regIoErr=get64(regRes, errCntPos+4);
    errCnt.sfpErr= get64(regRes, errCntPos+8);
    errCnt.hwIpErr=get64(regRes, errCntPos+12);
    errCnt.txErr=get64(regRes, errCntPos+16);
    errCnt.rxErr=get64(regRes, errCntPos+20);
}

void TDCDevice::read_bmc_power(const RegOpVector &regRes, int pos)
{
    BmcPower &p = status.bmc_status.power;
    p.clear();
    p.setupInProgress = standByTime.isValid() || (standByDoneTime.isValid() && standByDoneTime.elapsed()<1000);
    if(status.bmc_status.mcuRev >= 9 &&
            regRes.size() < pos+REG_BMC_VOL_BP_33-REG_BMC_VOL_TDC_A)
        return;

    for(mlink::RegOp op : regRes) {
        if(op.addr>=REG_BMC_VOL_TDC_A && op.addr <= REG_BMC_CUR_TDC_D) {
            quint16 shift = op.addr-REG_BMC_VOL_TDC_A;
            bool cur = shift%2;
            quint16 id = shift/2;
            auto &map = cur ? p.cur : p.vol;
            map[id] = op.data;
        }
    }
}

bool TDCDevice::update_err_cnt()
{
    auto regOp = getErrCntRegOp();
    auto regRes = regOpExec(regOp);
    if(regRes.size() != regOp.size()) {
        return false;
    }

    read_err_cnt(regRes, 0, status.statistic.powerOnErrCnt);
    return true;
}

bool TDCDevice::waitHptdcInit()
{
    QElapsedTimer timer;
    timer.start();
    while((regRead(REG32_TDC_GENERAL_STATUS) & 0x1) == 0){
        if(timer.elapsed() > STANDBY_TIMEOUT){
            qWarning() << "Failed to wait INIT state at device:" << getIdent()
                       << "TDC_GEN_CTRL:" << hex << regRead(REG32_TDC_GEN_CTRL);
            return false;
        }
    }
    return true;
}

void TDCDevice::clearStatistics()
{
    status.statistic.clear();
}

bool TDCDevice::checkLiveMagic()
{
    bool liveMagicCorrect = QtMregDevice::checkLiveMagic();
    if(!liveMagicCorrect) {
        setupCacheValid = false;
    }
    return liveMagicCorrect;
}

bool HptdcStatus::isOk() const
{
    return (regValues[0] & HPTDC_ERR_MASK) && isDllLock();
}

bool HptdcStatus::isDllLock() const
{
    return regValues[1] & (1<<12);
}

bool HptdcStatus::hasError() const
{

    return !isDllLock() || (regValues[0] & 0x7F);
}

QString HptdcStatus::getSummary() const
{
    QStringList sList;
    for(int i=0; i<16; ++i)
        if(regValues[0] & (1<<i) & HPTDC_ERR_MASK)
            sList << HPTDC_ST_STR[i];

//    quint8 readOutFifoOccup = regValues[1] & 0xFF;
//    quint8 trigFifoOccup = (regValues[1]>>12) & 0xFF;

//    quint8 occupL1Buffer[4];
//    occupL1Buffer[0] = regValues[2] & 0xFF;
//    occupL1Buffer[1] = (regValues[2]>>12) & 0xFF;
//    occupL1Buffer[2] = regValues[3] & 0xFF;
//    occupL1Buffer[3] = (regValues[4]>>12) & 0xFF;
    if(!isDllLock())
        sList << QString("Dll %1lock").arg(isDllLock()?"":"un");

    return sList.join(' ');
}

QString BmcPower::getIdStr(quint16 addr_or_id) {
    quint16 id = addr_or_id >= REG_BMC_VOL_TDC_A ?
                (addr_or_id-REG_BMC_VOL_TDC_A)/2 : addr_or_id;
    switch (id) {
    case 0: return "TDC_A";
    case 1: return "TDC_B";
    case 2: return "TDC_C";
    case 3: return "Backplane_5V";
    case 4: return "Backplane_3.3V";
    case 5: return "TDC_D";
    }
    return QString();
}

QString BmcPower::getSummary(bool withCurrent) const
{
    QStringList list;
    for(quint16 id : vol.keys()) {
        if(id > 2 && id < 5) continue;
        QString s = QString("%1: %2mV")
                .arg(getIdStr(id)).arg(vol.value(id,0));
        if(withCurrent)
            s += QString(" %3mA").arg(cur.value(id, 0));
        list << s;
    }
    return list.join('\n');
}

bool BmcPower::isCriticalCurrent() const
{
    if(setupInProgress) return false;
    for(quint16 id : cur.keys()) {
        if(id > 2 && id < 5) continue;
        quint16 val = cur[id];
        if(val>1500)
            return true;
    }
    return false;
}

QString TdcStatus::getClockStatus() const
{
    if((sysClockSource & clockStatus) != sysClockSource)
        return QString("Wrong clock logic\nstatus:0x%1\nsource:0x%2").arg(clockStatus,0,16).arg(sysClockSource,0,16);

    QStringList st;
    for(int i=2; i>=0; --i) {
        quint16 val = 1;
        val <<=i;
        if(val & clockStatus) {
            QString str;
            switch (i) {
            case 0:
                str = "osc";
                break;
            case 1:
                str = "TTC";
                break;
            case 2:
                str = "VXS";
                break;
            }
            if(val & sysClockSource)
//                str = QString("<body bgcolor='green'>%1</body>").arg(str);
                str = QString("<b>%1</b>").arg(str);
            st << str;
        }
    }

    return st.join(' ');
}
