#include "CalConfig.h"

#include <QCoreApplication>
#include <QSettings>
#include "mregdev/qtmregdevice.h"
#include "mldiscover/DeviceIndex.h"

namespace {
const QString INL_PREFIX = "inl_corr";
const QString BASE_CAL_PREFIX = "base_cal";


inline QString getSectionStr(const DeviceIndex &devIndex, const QString &sectionType)
{
    return QString("%1-%2-%3")
            .arg(QtMregDevice::getDeviceTypeName(devIndex.getDevId()))
            .arg(devIndex.getSerial(), 8, 16, QChar('0'))
            .arg(sectionType);
}

inline QString getKeyName(QString type, int ch)
{
    return QString("%1_%2").arg(type).arg(ch);
}

double getDouble(QSettings &s, QString key, bool *ok)
{
    if(s.contains(key)) {
        return s.value(key).toDouble(ok);
    } else {
        *ok = false;
        return 0;
    }
}
int getInt(QSettings &s, QString key, bool *ok)
{
    if(s.contains(key)) {
        return s.value(key).toInt(ok);
    } else {
        *ok = false;
        return 1920;
    }
}

void print_INL_ch_range(QString settingFileName, DeviceIndex devIndex, QList<int> read_ch) {
    qSort(read_ch);
    QString chMaskStr;
    QPair<int,int> range;
    range.first = -1;
    for(int ch : read_ch) {
        if(range.first == -1) {
            //first range
            range.first=ch;
            range.second=ch;
        } else if(range.second+1 == ch) {
            // continuation of range
            range.second=ch;
        } else {
            // new range
            if(range.first == range.second)
                chMaskStr.append(QString(" %1").arg(range.first));
            else
                chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
            range.first=ch;
            range.second=ch;
        }
    }
    // Flush last range if exists
    if(range.first != -1) {
        if(range.first == range.second)
            chMaskStr.append(" ").append(QString::number(range.first));
        else
            chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
    }
    qInfo() << settingFileName << "Got inl" << devIndex.getIdent() << chMaskStr;
}
}

void CalConfig::reloadCalibration()
{
    QSettings s(QSettings::IniFormat, QSettings::UserScope, QCoreApplication::organizationName(), "tqdc2_cal");
    reloadCalibration(s);
}

//void CalConfig::saveCalibration(QSettings &s, const DeviceIndex devIndex)
//{
//    QString group = getCalSectionStr(devIndex);
//    s.remove(group);
//    s.beginGroup(group);
//    const DeviceCalibration &devCal = allCal[devIndex];
//    for(int ch=0; ch<16; ++ch) {
//        s.setValue(getKeyName("x1", ch), devCal.x1[ch]);
//        s.setValue(getKeyName("x4", ch), devCal.x4[ch]);
//        s.setValue(getKeyName("thrOffset", ch), devCal.thrOffset[ch]);
//    }
//    s.endGroup();
//}

void CalConfig::reloadCalibration(QSettings &s)
{
    bool ok;
    double valD;
    int valI;
    for(const QString &group : s.childGroups()) {
        QString groupType;
        DeviceIndex devIndex = getDevIndexByGroup(group, groupType);
        if(devIndex.isNull())
            continue;

        DeviceCalibration &devCal = allCal[devIndex];
        s.beginGroup(group);

        if(group.endsWith(BASE_CAL_PREFIX)) {
            const int nch = s.value("nch", -1).toInt(&ok);
            if(ok) {
                for(int ch=0; ch<nch; ++ch) {
                    valD = getDouble(s, getKeyStr_X1(ch), &ok);
                    if(ok)
                        devCal.x1[ch] = valD;
                    valD = getDouble(s, getKeyStr_X4(ch), &ok);
                    if(ok)
                        devCal.x4[ch] = valD;
                    valI = getInt(s, getKeyStr_ThrOffset(ch), &ok);
                    if(ok)
                        devCal.thrOffset[ch] = valI;
                }
            }
        } else if(group.endsWith(INL_PREFIX)) {
            const int nch = s.value("nch", 16).toInt(&ok);
            if(ok && nch) {
                int ch;
                QVector<double> corr;
                QList<int> read_ch;

                for(const QString &key: s.childKeys()) {
                    corr.reserve(1024);
                    ch = key.toInt(&ok);
                    if(ok) {
                        const QList<QVariant> &corrList = s.value(key).toList();
                        for (QVariant c : corrList) {
                            corr << c.toDouble(&ok);
                            if(!ok) {
                                qWarning()<< "Bad inl corr format in config";
                                break;
                            }
                        }
                        if(ok && corr.size()==1024) {
                            devCal.inl[ch] = corr;
                            read_ch << ch;
                        }

                        corr.clear();
                    }
                }

                print_INL_ch_range(s.fileName(), devIndex, read_ch);
            }
        }

        s.endGroup();
    }
}

QString CalConfig::getInlSectionStr(const DeviceIndex &devIndex)
{
    return getSectionStr(devIndex, INL_PREFIX);
}

QString CalConfig::getCalSectionStr(const DeviceIndex &devIndex)
{
    return getSectionStr(devIndex, BASE_CAL_PREFIX);
}

QString CalConfig::getKeyStr_X1(int ch)
{
    return getKeyName("x1", ch);
}

QString CalConfig::getKeyStr_X4(int ch)
{
    return getKeyName("x4", ch);
}

QString CalConfig::getKeyStr_ThrOffset(int ch)
{
    return getKeyName("thrOffset", ch);
}

DeviceIndex CalConfig::getDevIndexByGroup(QString groupStr, QString &type)
{
    DeviceIndex index;
    if(groupStr.endsWith(BASE_CAL_PREFIX)){
        type = BASE_CAL_PREFIX;
    } else if(groupStr.endsWith(INL_PREFIX)){
        type = INL_PREFIX;
    } else {
        return index;
    }
    QString devIndexStr = groupStr.left(groupStr.length()-type.length()-1); // "TQDC16VS_E-046F0C75"
    QString devType = devIndexStr.left(devIndexStr.length()-9); // "-046F0C75"
    QString devSerial = devIndexStr.right(8); // "046F0C75"
    bool ok;
    quint32 serial = devSerial.toUInt(&ok, 16);
    if(ok) {
        quint8 id=1;
        for(; id!=0; ++id) {
            if(devType == QtMregDevice::getDeviceTypeName(id)) {
                break;
            }
        }
        if(id)
            return DeviceIndex(id, serial);
    }
    return DeviceIndex();
}
