#include "CalibrationDialog.h"
#include "ui_CalibrationDialog.h"

#include "Config.h"
#include "TQDCDevice.h"
#include "MainWindow.h"

namespace {
enum {COL_CAL_CH, COL_CAL_BASELINE, COL_CAL_BASELINE_RMS,
      COL_CAL_BASELINE4, COL_CAL_BASELINE4_RMS,
      COL_CAL_THR, COL_CAL_INL, COL_CAL_TOTAL};
const int RMS_LIMIT = 20; // in adc64 100
}

CalibrationDialog::CalibrationDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::CalibrationDialog),
    chNum(0)
{
    ui->setupUi(this);
    initTable();
}

CalibrationDialog::~CalibrationDialog()
{
    delete ui;
}

void CalibrationDialog::updateBaseline(const QMap<int, double> &baseline,
                                       const QMap<int, double> &baseline4,
                                       const QMap<int, double> rms,
                                       const QMap<int, double> rms4)
{
    ui->tableWidget->setUpdatesEnabled(false);
    for(int ch=0; ch<chNum; ++ch) {
        setItemValue(ch, COL_CAL_BASELINE, baseline, false);
        setItemValue(ch, COL_CAL_BASELINE4, baseline4, false);
    }

    for(int ch=0; ch<chNum; ++ch) {
        setItemValue(ch, COL_CAL_BASELINE_RMS, rms, true);
        setItemValue(ch, COL_CAL_BASELINE4_RMS, rms4, true);
    }
    chBaseline = baseline;
    chBaselineRms = rms;
    chBaseline4 = baseline4;
    chBaseline4Rms = rms4;
    ui->tableWidget->setUpdatesEnabled(true);
    MainWindow::updateTableSize(ui->tableWidget);
}

void CalibrationDialog::updateDacOffset(const QMap<int, int> &thr)
{
    ui->tableWidget->setUpdatesEnabled(false);
    for(int ch=0; ch<chNum; ++ch){
        QTableWidgetItem *item = ui->tableWidget->item(ch, COL_CAL_THR);
        if(!item)
            continue;
        if(thr.contains(ch)){
            if(thr[ch] != dacOffset[ch])
                item->setText(getThrText(thr[ch]));
        } else {
            item->setText("");
        }
    }
    dacOffset = thr;
    ui->tableWidget->setUpdatesEnabled(true);
    MainWindow::updateTableSize(ui->tableWidget);
}

void CalibrationDialog::setChNumber(int num)
{
    if(chNum == num)
        return;
    QTableWidget *table = ui->tableWidget;
    chNum = num;
    table->setRowCount(chNum);
    QString txt;
    for(int ch=0; ch<chNum; ++ch) {
        txt = QString::number(ch+1);
        table->setItem(ch, COL_CAL_CH, new QTableWidgetItem(txt));

        if(chBaseline.contains(ch))
            txt = QString::number(chBaseline[ch], 'f', 2);
        else
            txt = "";
        table->setItem(ch, COL_CAL_BASELINE, new QTableWidgetItem(txt));

        if(chBaselineRms.contains(ch))
            txt = QString::number(chBaselineRms[ch], 'f', 2);
        else
            txt = "";
        table->setItem(ch, COL_CAL_BASELINE_RMS, new QTableWidgetItem(txt));

        if(chBaseline4.contains(ch))
            txt = QString::number(chBaseline4[ch], 'f', 2);
        else
            txt = "";
        table->setItem(ch, COL_CAL_BASELINE4, new QTableWidgetItem(txt));

        if(chBaseline4Rms.contains(ch))
            txt = QString::number(chBaseline4Rms[ch], 'f', 2);
        else
            txt = "";
        table->setItem(ch, COL_CAL_BASELINE4_RMS, new QTableWidgetItem(txt));

        if(dacOffset.contains(ch))
            txt = getThrText(dacOffset[ch]);
        else
            txt = "";
        table->setItem(ch, COL_CAL_THR, new QTableWidgetItem(txt));

        table->setItem(ch, COL_CAL_INL, new QTableWidgetItem());
    }
    MainWindow::updateTableSize(ui->tableWidget);
}

void CalibrationDialog::init(const DeviceCalibration &devCal)
{
    updateBaseline(devCal.x1, devCal.x4, QMap<int, double>(), QMap<int, double>());
    updateDacOffset(devCal.thrOffset);
    setChNumber(devCal.nch);
    for(int ch=0; ch<chNum; ++ch) {
        inlPresent[ch] = false;
        setTableInl(ch, false);
    }
    updateInl(devCal.inl);
}

void CalibrationDialog::updateInl(const QMap<int, QVector<double> > &inlCorr)
{
    for (int ch : inlCorr.keys()) {
        inlPresent[ch] = inlCorr.value(ch).size()==1024;
        setTableInl(ch, inlPresent[ch]);
    }
}

void CalibrationDialog::enableControl(bool enabled)
{
    ui->pushButtonCalibrate->setEnabled(enabled);
    ui->pushButtonClear->setEnabled(enabled);
}

QString CalibrationDialog::getThrText(int dacValue) const
{
    return QString("%1 mV (%2)")
            .arg(TQDCDevice::dac_mV(dacValue), 0, 'f', 2)
            .arg(dacValue);
}

void CalibrationDialog::initTable()
{
    QTableWidget *table = ui->tableWidget;
    table->verticalHeader()->hide();
    table->setColumnCount(COL_CAL_TOTAL);
    table->setHorizontalHeaderItem(COL_CAL_CH, new QTableWidgetItem("Ch #"));
    table->setHorizontalHeaderItem(COL_CAL_BASELINE, new QTableWidgetItem("Baseline"));
    table->setHorizontalHeaderItem(COL_CAL_BASELINE_RMS, new QTableWidgetItem("Baseline rms"));
    table->setHorizontalHeaderItem(COL_CAL_BASELINE4, new QTableWidgetItem("Baseline x4"));
    table->setHorizontalHeaderItem(COL_CAL_BASELINE4_RMS, new QTableWidgetItem("Baseline x4 rms"));
    table->setHorizontalHeaderItem(COL_CAL_THR, new QTableWidgetItem("Thr. offset, mV (dacValue)"));
    table->setHorizontalHeaderItem(COL_CAL_INL, new QTableWidgetItem("INL corr."));

    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    table->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
//    updateTableSize(table);
//    setTableRowsCount(16);
}

void CalibrationDialog::setTableInl(int ch, bool inlOk)
{
    QTableWidgetItem *i = ui->tableWidget->item(ch, COL_CAL_INL);
    i->setText(inlOk ? "Yes":"No");
    i->setTextColor(inlOk ? Qt::darkGreen : Qt::red);
}

void CalibrationDialog::setItemValue(int ch, int col, const QMap<int, double> &chData, bool isRms)
{
    QTableWidgetItem *item = ui->tableWidget->item(ch, col);
    if(!item)
        return;
    if(chData.contains(ch)) {
        const double &val = chData[ch];
        item->setText(QString::number(val, 'f', 2));
        if(isRms)
            item->setTextColor(val > RMS_LIMIT ? Qt::red : Qt::black);
    } else {
        item->setText("");
    }
}

void CalibrationDialog::on_pushButtonCalibrate_clicked()
{
    emit calibrationRequest();
}

void CalibrationDialog::on_pushButtonClear_clicked()
{
    emit clearCalibration();
}

void CalibrationDialog::on_pushButtonClose_clicked()
{
    close();
}
