#include "TqdcFileWriter.h"

#include <QDebug>
#include <QString>
#include <QTextStream>
#include <QSemaphore>
#include <QElapsedTimer>
#include <QTimer>

#include "DeviceData.h"

TqdcFileWriter::TqdcFileWriter(QSemaphore *_writerSem):
    QObject(nullptr),
    writerSem(_writerSem),
    writeEnabled(false),
    rawDataWriter(true),
    lookForStopEv(false),
    lastWrittenEvNum(0),
    statusTimer(new QTimer(this))
{
    connect(statusTimer, &QTimer::timeout, this, &TqdcFileWriter::sendStat);
    statusTimer->setInterval(1e3);
}

//void FileWriter::write(const QMap<quint32, QVector<double> > &counters)
//{
//    dataFile.open( QIODevice::WriteOnly | QIODevice::Text );
//    QTextStream out(&dataFile);

//    for(quint32 devId : counters.keys()){
//        for(int ch=0; ch < counters[devId].size(); ch++){
//            out << ch <<' '<<counters[devId][ch] << '\n';
//            if(counters[devId][ch])
//                qDebug() << ch << '\t' << counters[devId][ch];
//        }
//    }
//    dataFile.close();
//    qDebug()<< "counters writen to "<<dataFile.fileName();
//}

void TqdcFileWriter::startWrite(QString fileName, bool rawData)
{
    if(dataFile.isOpen()) {
        qFatal("File writer unsyncron work of start/stop");
    }
    dataFile.setFileName(fileName);
    if(!dataFile.open(QIODevice::WriteOnly)) {
        qWarning() << QString("Can't opent file '%1' to write")
                      .arg(fileName);
    }
    writeEnabled = true;
    statusTimer->start();
    statistic.clear();
    rawDataWriter = rawData;
    qInfo() << QString("FileWriter start write %1 data to:%2")
               .arg((rawData?"raw":"decoded"), fileName);
    lookForStopEv = false;
    stopEvNum = 0;
    lastWrittenEvNum = 0;
}

void TqdcFileWriter::stopWrite()
{
    writeEnabled = false;
    lookForStopEv = false;
    stopEvNum = 0;
    dataFile.close();
    sendStat();
    statusTimer->stop();
    emit stopped(dataFile.fileName());
    qInfo() << "FileWriter stoped";
}

void TqdcFileWriter::writeRawData(const QByteArray &data, bool write)
{
    if(!write)
        return;
    if(!writeEnabled || !rawDataWriter) {
        qWarning() << "Possible loss of writer sem. in writeRawData";
        return;
    }
    QElapsedTimer writeTime;
    writeTime.start();

    writerSem->release();

    if(!dataFile.isOpen()) {
        qWarning() << "Data file not opened on writting.";
    } else {
        qint64 val = dataFile.write(data);
        if(val == -1) {
            qWarning() << "Got error on file write, ret==-1";
        } else if(val != data.size()) {
            qWarning() << QString("Got error on file write, wroteSize=%1; dataSize=%2")
                          .arg(val).arg(data.size());
        }
    }
    statistic.writeTimeNs = writeTime.nsecsElapsed();
}

void TqdcFileWriter::writeDecodedData(const DeviceData &data, bool write)
{
    if(!write)
        return;
    if(!writeEnabled || rawDataWriter) {
        qWarning() << "Possible loss of writer sem. in writeDecodedData";
        return;
    }
    writerSem->release();
    bool isFileEmpty = dataFile.size()==0;
    QElapsedTimer writeTime;
    writeTime.start();
    int eventWritten=0;

    if(!dataFile.isOpen()) {
        qWarning() << "Data file not opened on writting.";
    } else {
        QTextStream textStream(&dataFile);
        for (const CompleteEvent &ev : data.events) {
            if(isFileEmpty) {
                if(ev.evNum!= 0)
                    qWarning() << "First event in file has evNum="<<ev.evNum;
                isFileEmpty = false;
            }
            lastWrittenEvNum = ev.evNum;
            if(lookForStopEv && ev.evNum==stopEvNum){
                qInfo() << "FileWriter::gotStopEvernt "<<stopEvNum;
                emit gotStopEvernt(stopEvNum);
            }
            textStream << QString("Ev: %1 %2.%3\n")
                          .arg(ev.evNum)
                          .arg(ev.taiSec).arg(ev.taiNSec);
            std::map<int, QVector<TqdcEvent> >::const_iterator itTdc=ev.tqdcEvents.begin();
            while(itTdc!=ev.tqdcEvents.end()) {
                const QVector<TqdcEvent> &tdcEvs = itTdc->second;
                for (const TqdcEvent &tdcEv : tdcEvs) {
                    textStream << QString("Tdc %1: %2\n").arg(tdcEv.ch).arg(tdcEv.leadingTime);
                }
                ++itTdc;
            }


            std::map<int, QVector<AdcEvent> >::const_iterator itAdc=ev.adcEvents.begin();
            while(itAdc!=ev.adcEvents.end()) {
                const QVector<AdcEvent> &adcEvs = itAdc->second;
                const int &ch = itAdc->first;
                for (const AdcEvent adcEv : adcEvs) {
                    QString adcStr;
                    adcStr = QString("Adc %1: %2;")
                            .arg(ch).arg(adcEv.adcTs);

                    for (const double &sampe : adcEv.wf) {
                        adcStr += " " + QString::number(sampe, 'f', 0);
                    }
                    textStream << adcStr << '\n';
//                    textStream << QString("Adc %1: %2;")
//                                  .arg(ch).arg(adcEv.adcTs);

//                    for (const double &sampe : adcEv.wf) {
//                        textStream << " " << sampe;
//                    }
//                    textStream << '\n';
                }
                ++itAdc;
            }
            ++eventWritten;
        }
    }
    statistic.writeTimeNs += writeTime.nsecsElapsed();
    statistic.evWritten += eventWritten;
    if(lookForStopEv)
        sendStat();
}

void TqdcFileWriter::prepareForStop(quint32 evNum)
{
    if(!writeEnabled || rawDataWriter) {
        if(writeEnabled)
            emit gotStopEvernt(evNum);
        return;
    }

    qInfo() << "FileWriter::prepareForStop "<<evNum<<" lastWrittenEvNum:"<<lastWrittenEvNum;
    if(evNum == lastWrittenEvNum) {
        qInfo() << "Writer is already ready for stop";
        emit gotStopEvernt(evNum);
    } else {
        lookForStopEv = true;
        stopEvNum = evNum;
    }
}

void TqdcFileWriter::sendStat()
{
    if(lookForStopEv) {
        qInfo() << "FileWriter lastWrittenEvNum:"<<lastWrittenEvNum;
    }
    statistic.fileSize = dataFile.size();
    emit fileStatisticUpdated(statistic);
}
