#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "SnmpDialog.h"
#include "daq-config/DaqConfig.h"
#include "device-discover/DiscoverDialog.h"
#include "lib-common/UIStyles.h"
#include "metric-sender/MetricSender.h"
#include "mregdev/qtmregdevice.h"
#include "pnp-server/PNPServer.h"
#include "base/threadQuit.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "base/AboutBox.h"

#ifdef __linux__
#include <signal.h>
#endif

#include <QDateTime>
#include <QDebug>
#include <QHostAddress>
#include <QHostInfo>
#include <QMessageBox>
#include <QProcess>
#include <QScrollBar>
#include <QShortcut>
#include <QStandardPaths>
#include <QStyle>
#include <QThread>
#include <QTextCharFormat>
#include <QTextFormat>
#include <QTimer>

namespace {

const int MAX_SLOT_NUMBER = 21;
const int CRATE_SLOT_ZERO = 10; // slot number for ch:0
const int CONFIG_TIMEOUT = 4000;

enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_MSTREAM=2, TAB_LOGS=3};

enum {
    ROW_NAME,ROW_STATUS,
    ROW_DEV_FW, ROW_DEV_MCU_VER, ROW_DEV_TEMP,
    ROW_DEV_STREAM, ROW_DEV_MS_ERR, ROW_WR_TIME,
    ROW_SPACER_0,
    ROW_UP_TIME, ROW_CUR_EV_NUM, ROW_LEMO_CNT, ROW_LEMO_RATE, ROW_SFP_CNT, ROW_SFP_RATE,
    ROW_SPACER_1,
    ROW_RUN_TIME, ROW_RUN_TRIG_CANDID, ROW_RUN_TRIG_ACCEPTED, ROW_RUN_TRIG_RATE,
    ROW_SPACER_2,
    ROW_CH_SLOT_1, ROW_CH_SLOT_2, ROW_CH_SLOT_3, ROW_CH_SLOT_4, ROW_CH_SLOT_5, ROW_CH_SLOT_6, ROW_CH_SLOT_7,
    ROW_CH_SLOT_8, ROW_CH_SLOT_9, ROW_CH_SLOT_10,ROW_CH_SLOT_11,ROW_CH_SLOT_12,ROW_CH_SLOT_13,ROW_CH_SLOT_14,
    ROW_CH_SLOT_15,ROW_CH_SLOT_16,ROW_CH_SLOT_17,ROW_CH_SLOT_18,ROW_CH_SLOT_19,ROW_CH_SLOT_20,ROW_CH_SLOT_21,
    ROW_CH_CNT
};
const QSet<int> SPACER_ROWS = {ROW_SPACER_0,ROW_SPACER_1,ROW_SPACER_2};

int chToSlot(int ch)
{
    bool odd = ch&1;
    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;

    int slot = (ch&1) ? (slotBase+(ch>>1)) : (slotBase-(ch>>1));
    if(slot<0 || slot>MAX_SLOT_NUMBER) return -1;

    return slot;
}

int slotToCh(int slot)
{
    if(slot>MAX_SLOT_NUMBER) return -1;
    if(slot>CRATE_SLOT_ZERO && slot<CRATE_SLOT_ZERO+3) return -1;
    bool odd = slot>CRATE_SLOT_ZERO;

    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;
    int ch = qAbs(slotBase-slot);
    ch <<=1;
    if(odd)
        ch |= 1;
    return ch;
}
QString getDevMetricPrefix(DeviceIndex i) {
    return QString("%1.").arg(i.getSerialStr());
}
}

MainWindow::MainWindow(StartupOptions &start_opt, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    config(start_opt),
    discover_dialog(new DiscoverDialog(this)),
    pnpServer(new PNPServer(this)),
    remoteControl(new RemoteControlServer(0)),
    remoteControlThread(new QThread(this)),
    remoteConnected(false),
    evNumChecker(new EvNumChecker(this)),
    hasRunError(false),
    blockCfgSave(0),
    writeConfigTimer(new QTimer(this)),
    writeRunHistoryTimer(new QTimer(this)),
    megamaster_flg(false)
{
    qRegisterMetaType<TtvxsStatus>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<TtvxsConfig>();
    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<RootConfig>();

    ui->setupUi(this);

    QThread::currentThread()->setObjectName("GUI thread");
    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));

    init_GUI_style();
    init_GUI_default_settings();
    initTableDev();
    initStatPresets();

    connect(writeConfigTimer, &QTimer::timeout, this, &MainWindow::delayed_write_config);
    writeConfigTimer->setSingleShot(true);
    writeConfigTimer->start(500);
    connect(writeRunHistoryTimer, &QTimer::timeout, this, &MainWindow::write_run_history);
    writeRunHistoryTimer->setSingleShot(true);
    writeRunHistoryTimer->setInterval(CONFIG_TIMEOUT);

    connect(discover_dialog, &DiscoverDialog::deviceDiscovered, this, &MainWindow::device_updated);
    discover_dialog->setSingleMode(false);
    discover_dialog->setDeviceIdFilter(QtMregDevice::DEVICE_ID_TTVXS);

    connect(ui->clientManagerWidget, &ClientManagerWidget::client_config_changed, this, &MainWindow::client_config_receiver);

//    QTimer::singleShot(500, this, &MainWindow::on_actionSNMP_dialog_triggered()));

    connect(ui->clientManagerWidget, &ClientManagerWidget::runWarn, this, &MainWindow::onRunWarning);
    connect(ui->clientManagerWidget, &ClientManagerWidget::runError, this, &MainWindow::onRunError);
    connect(ui->clientManagerWidget, &ClientManagerWidget::backPressure, this, &MainWindow::onBackPressure);
    connect(ui->clientManagerWidget, &ClientManagerWidget::updateTabState, this, &MainWindow::setClientTabState);
    connect(ui->clientManagerWidget, &ClientManagerWidget::config_transfer, this, &MainWindow::config_transfer_from_clients);

    connect(this, &MainWindow::restartAll, ui->clientManagerWidget, &ClientManagerWidget::onRestartAll);
    connect(this, &MainWindow::stopAll, ui->clientManagerWidget, &ClientManagerWidget::onStopAll);
    connect(this, &MainWindow::sendControlEnable, ui->clientManagerWidget, &ClientManagerWidget::setControlEnable);

    connect(pnpServer, &PNPServer::programDiscovered, ui->clientManagerWidget, &ClientManagerWidget::programDiscovered);
    connect(pnpServer, &PNPServer::programClosed, ui->clientManagerWidget, &ClientManagerWidget::programClosed);
    connect(this, &MainWindow::changeProgramDescription, pnpServer, &PNPServer::changeProgramDescription);
    connect(this, &MainWindow::closeProgram, pnpServer, &PNPServer::closeProgram);
    connect(this, &MainWindow::deleteAll, pnpServer, &PNPServer::deleteLater);

    QStringList searchTargs;
    searchTargs << DaqConfig::getTypeName(DaqConfigAdc64System);
    searchTargs << DaqConfig::getTypeName(DaqConfigEventBuilder);
    searchTargs << DaqConfig::getTypeName(DaqConfigHrbSystem);
    searchTargs << DaqConfig::getTypeName(DaqConfigRunControl);
    searchTargs << DaqConfig::getTypeName(DaqConfigVmeDaq);
    searchTargs << DaqConfig::getTypeName(DaqConfigLand);
    searchTargs << DaqConfig::getTypeName(DaqConfigGenerator);
    searchTargs << DaqConfig::getTypeName(DaqConfigTdcControl);
    searchTargs << DaqConfig::getTypeName(DaqConfigTtvxs);
    pnpServer->startSearch(searchTargs);

    connect(evNumChecker, &EvNumChecker::evNumRequest,
            ui->clientManagerWidget, &ClientManagerWidget::evNumRequest);
    connect(evNumChecker, &EvNumChecker::checkFail, this, &MainWindow::onEvNumCheckFail);
    connect(ui->clientManagerWidget, &ClientManagerWidget::evNumResponse,
            evNumChecker, &EvNumChecker::gotClientEvNum);
    connect(ui->clientManagerWidget, &ClientManagerWidget::clientMapUpdated,
            evNumChecker, &EvNumChecker::updateClientMap);

    connect(this, &MainWindow::deleteAll, remoteControl, &RemoteControlServer::deleteLater);
    connect(remoteControl, &RemoteControlServer::destroyed, remoteControlThread, &QThread::quit);

    connect(remoteControl, &RemoteControlServer::startRequest,
            this, &MainWindow::onRemoteStartRequest);
    connect(remoteControl, &RemoteControlServer::stopRequest, this, &MainWindow::onRemoteStopRequest);
    connect(this, &MainWindow::remoteStartCompleted, remoteControl, &RemoteControlServer::completedStart);
    connect(this, &MainWindow::remoteStopCompleted, remoteControl, &RemoteControlServer::completedStop);
    connect(this, &MainWindow::sendRemoteState, remoteControl, &RemoteControlServer::sendState);
    connect(this, &MainWindow::sendRunError, remoteControl, &RemoteControlServer::onRunError);
    connect(this, &MainWindow::sendRunWarn, remoteControl, &RemoteControlServer::onRunWarning);
    connect(this, &MainWindow::sendBackPressure, remoteControl, &RemoteControlServer::onBackPressure);

    connect(remoteControl, &RemoteControlServer::connectionStatusUpdated,
            this, &MainWindow::onRemoteStatusUpdated);
    connect(this, &MainWindow::allowRemoteControl,
            remoteControl, &RemoteControlServer::startListen);
    connect(this, &MainWindow::changeProgramDescription,
            remoteControl, &RemoteControlServer::changeProgramDescription);

    connect(remoteControl, &RemoteControlServer::evNumCheckRequest,
            evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
    connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
            remoteControl, &RemoteControlServer::evNumCheckResponse);

    connect(this, &MainWindow::config_transfer_to_rc,
            remoteControl, &RemoteControlServer::config_transfer);

    remoteControl->moveToThread(remoteControlThread);
    remoteControlThread->setObjectName("RemoteCtrl");
    remoteControlThread->start();

    readSettings();

    configuration_manager = new ConfigurationManager(this, config.program_type, config.default_root);
    connect(configuration_manager, &ConfigurationManager::switch_to_config,
            this, &MainWindow::switch_to);
    connect(ui->mstream_output, &MStreamOutput::mstream_stat_updated,
            this, &MainWindow::mstream_status_updated);
}

MainWindow::~MainWindow()
{
    saveSettings(true);
    emit closeAll();
    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stateRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    emit deleteAll();
    disconnectFromDev(DeviceIndex());

    if(!threadQuiting(remoteControlThread, 5000)) {
        qWarning() << "Failed to stop remoteControlThread";
    }

    delete ui->clientManagerWidget;
    delete ui;
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->reset(config.program_index, config.configuration_name);
    configuration_manager->show();
}

void MainWindow::switch_to(QString program_index, QString configuration_name)
{
    config.program_index = program_index;
    config.configuration_name = configuration_name;

    disconnectFromDev(DeviceIndex());

    readSettings();
    setWindowTitle(QString("%1   |   Program index: %2   |   Configuration name:  %3")
                   .arg(config.program_type).arg(config.program_index).arg(config.configuration_name));
}

void MainWindow::init_GUI_style()
{
    QString styleLargeFont = "font-size: 18pt";
    ui->pushButton_start->setStyleSheet(styleLargeFont);
    ui->pushButton_stop->setStyleSheet(styleLargeFont);
    ui->labelRunState->setStyleSheet(styleLargeFont);
    ui->pushButton_start->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButton_stop->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

    ui->groupBox_trigger_source->setStyleSheet(gbox_fill_blue);
    ui->groupBox_trigger_state->setStyleSheet(gbox_fill_blue);
    ui->groupBoxStat->setStyleSheet(gbox_fill_blue);
    ui->plainTextEdit_log->setReadOnly(true);

    ui->spinBox_timer->setMaximum(10000000);
    ui->spinBox_random->setMaximum(1000000);
}

void MainWindow::init_GUI_default_settings()
{
    ui->pushButton_stop->setEnabled(false);

    labelWarnTtvxs = new QLabel();
    labelWarnClients = new QLabel();
    labelWarnOther = new QLabel();

    ui->statusBar->addWidget(labelWarnTtvxs);
    ui->statusBar->addWidget(labelWarnClients);
    ui->statusBar->addWidget(labelWarnOther);

    labelWarnTtvxs->hide();
    labelWarnClients->hide();
    labelWarnOther->hide();
}

void MainWindow::initTableDev()
{
    auto t = ui->tableWidgetDev;
    t->setRowCount(ROW_CH_CNT);

    t->setVerticalHeaderItem(ROW_NAME, new QTableWidgetItem("Name"));
    t->setVerticalHeaderItem(ROW_STATUS, new QTableWidgetItem("Status"));

    t->setVerticalHeaderItem(ROW_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    t->setVerticalHeaderItem(ROW_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));
    t->setVerticalHeaderItem(ROW_DEV_TEMP, new QTableWidgetItem("Temp"));
    t->setVerticalHeaderItem(ROW_DEV_STREAM, new QTableWidgetItem("Data Stream, kB/s"));
    t->setVerticalHeaderItem(ROW_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    t->setVerticalHeaderItem(ROW_WR_TIME, new QTableWidgetItem("WR time"));

    t->setVerticalHeaderItem(ROW_UP_TIME, new QTableWidgetItem("Up time"));
    t->setVerticalHeaderItem(ROW_CUR_EV_NUM, new QTableWidgetItem("Cur evNum"));
    t->setVerticalHeaderItem(ROW_LEMO_CNT, new QTableWidgetItem("TTL pulses"));
    t->setVerticalHeaderItem(ROW_LEMO_RATE, new QTableWidgetItem("TTL rate"));
    t->setVerticalHeaderItem(ROW_SFP_CNT, new QTableWidgetItem("SFP pulses"));
    t->setVerticalHeaderItem(ROW_SFP_RATE, new QTableWidgetItem("SFP rate"));

    t->setVerticalHeaderItem(ROW_RUN_TIME, new QTableWidgetItem("Run time"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_CANDID, new QTableWidgetItem("Trig run cand."));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_ACCEPTED, new QTableWidgetItem("Trig run accepted"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_RATE, new QTableWidgetItem("Run trig rate"));

    for(int row : SPACER_ROWS) {
        t->setVerticalHeaderItem(row, new QTableWidgetItem(" "));
    }

    for(int row=ROW_CH_SLOT_1; row<=ROW_CH_SLOT_21; ++row) {
        auto item = new QTableWidgetItem(QString("Slot %2").arg(row-ROW_CH_SLOT_1+1, 2, 10, QChar('0')));
        if(isSlotBlocked(row)) {
            item->setTextColor(Qt::darkCyan);
        }
        t->setVerticalHeaderItem(row, item);
    }

    t->resizeRowsToContents();
    t->setWordWrap(false);
    t->setStyleSheet(":disabled {background-color:lightGray}");

    int spacer_height = 8;
    auto header = t->verticalHeader();
    header->setSectionResizeMode(QHeaderView::Fixed);
    header->setMinimumSectionSize(spacer_height);

    for(int row : SPACER_ROWS) {
        header->resizeSection(row, spacer_height);
        auto spItem = t->verticalHeaderItem(row);
        spItem->setBackgroundColor(Qt::lightGray);
    }
}

void MainWindow::initStatPresets()
{
    ui->comboBoxPreset->blockSignals(true);
    for(auto preset : StatPresetFabric::getAllPresets()){
        ui->comboBoxPreset->addItem(preset.label, preset.type);
    }
    ui->comboBoxPreset->blockSignals(false);
}

void MainWindow::writeSetup(DeviceIndex index)
{
    saveSettings();
    for(auto i : devGlob.keys()) {
        if(!index.isNull() && i!=index) continue;
        emit setupChanged(config.devicesCfg[i]);
    }
}

void MainWindow::addDevice(DeviceIndex index)
{
    addDevice(DeviceDescription(index));
}

void MainWindow::addDevice(DeviceDescription dd)
{
    const auto &index = dd.getIndex();
    if(devGlob.contains(index)) return;

    DeviceGlob * dg;
    devGlob.insert(index, dg = new DeviceGlob(index));

    //--------
    ui->mstream_output->add_device(dd.getSerialIdStr());
    connect(dg->mstreamProc, SIGNAL(readyReadStandardError()),
            SLOT(mstream_output_read()));
    connect(dg->mstreamProc, SIGNAL(readyReadStandardOutput()),
            SLOT(mstream_output_read()));
    connect(dg->mstreamProc, SIGNAL(error(QProcess::ProcessError)),
            SLOT(process_error(QProcess::ProcessError)));
    connect(dg->mstreamProc, SIGNAL(finished(int,QProcess::ExitStatus)),
            SLOT(process_finished(int,QProcess::ExitStatus)));
    set_mstream_args(index);
    start_process(index);
    //--------

    int colDev = -1;
    auto tableDev = ui->tableWidgetDev;

    if(colDev == -1){
        colDev = tableDev->columnCount();
        tableDev->setColumnCount(colDev+1);
        for(int row=0; row<tableDev->rowCount(); ++row) {
            auto item = new QTableWidgetItem();
            tableDev->setItem(row, colDev, item);
            if(row == ROW_NAME) {
                item->setFlags(item->flags() | Qt::ItemIsEditable);
                item->setText(config.devicesCfg[index].custom_name);
            } else {
                item->setFlags(item->flags() & ~Qt::ItemIsEditable);
            }
            if(SPACER_ROWS.contains(row)) {
                item->setBackgroundColor(Qt::lightGray);
                continue;
            }
            if(row < ROW_CH_SLOT_1 || row > ROW_CH_SLOT_21) continue;
            if(isSlotBlocked(row)){
                item->setBackgroundColor(Qt::darkCyan);
            } else {
                item->setCheckState(Qt::Unchecked);
            }
        }
    }

    const quint32 &vxsChMask = config.devicesCfg[index].vxsChMask;
    Q_ASSERT(colDev != -1);
    for(int row=ROW_CH_SLOT_1; row<=ROW_CH_SLOT_21; ++row) {
        const int slot = row-ROW_CH_SLOT_1+1;
        const int ch = slotToCh(slot);
        if(!isSlotBlocked(row)){
            tableDev->item(row, colDev)->setCheckState(vxsChMask&(1<<ch)?Qt::Checked:Qt::Unchecked);
        }
    }

    auto header_item = new QTableWidgetItem(index.getSerialStr());
    header_item->setData(Qt::UserRole, QList<QVariant>() << index.getDevId() << index.getSerial());
    header_item->setTextColor(Qt::red);
    tableDev->setHorizontalHeaderItem(colDev, header_item);

    try{
        auto &dg = devGlob[index];
        TtvxsDevice *dev = dg->dev;
        connect(this, &MainWindow::closeAll, dev, &TtvxsDevice::deleteLater);
        connect(this, &MainWindow::deleteDevice, dg, &DeviceGlob::deleteHardware);

        connect(this, &MainWindow::setupChanged, dev, &TtvxsDevice::setSetup);
        connect(this, &MainWindow::softClear, dev, &TtvxsDevice::softClear);
        connect(this, &MainWindow::clearStat, dev, &TtvxsDevice::clearStat);
        connect(this, &MainWindow::resetTimestamp, dev, &TtvxsDevice::sendResetTimestamp);
        connect(this, &MainWindow::evNumReset, dev, &TtvxsDevice::evNumReset);

        connect(this, &MainWindow::setDevAddr, dev, &TtvxsDevice::deviceUpdated);
        connect(discover_dialog, &DiscoverDialog::deviceDiscovered, dev, &TtvxsDevice::deviceUpdated);

        connect(dev, &TtvxsDevice::statusUpdated, this, &MainWindow::deviceStatusUpdated);
        connect(this, &MainWindow::restartAll, dev, &TtvxsDevice::onRestartAll);

        connect(evNumChecker, &EvNumChecker::getTrigEvNum,
                dev, &TtvxsDevice::evNumRequest);
        connect(dev, &TtvxsDevice::trigEvNumResponse,
                evNumChecker, &EvNumChecker::gotTopLevelEvNum);
        connect(dev, &TtvxsDevice::devEvNumResponse,
                evNumChecker, &EvNumChecker::gotDevEvNum);

        dg->startThread();

        // Connecting
        QHostAddress hostAddr;
        if(dd.ip_addr) {
            emit setDevAddr(dd);
        }

        writeSetup(index);
    } catch (std::exception &e){
        sendLog(QString("Connect to device %1").arg(e.what()));
    }
    qInfo() << QString("%1 was added").arg(dd.getSerialIdStr());

}

//------mstream slots
void MainWindow::set_mstream_args(const DeviceIndex &index)
{
    QString mStreamPath = QStandardPaths::findExecutable("afi-mstream");
    if(mStreamPath.isEmpty()) {
        mStreamPath = QStandardPaths::findExecutable("mstream");
        if(mStreamPath.isEmpty()) {
            qCritical()<<"Can't find mstream executable";
            mStreamPath="mstream";
        }
    }
    QStringList args;
    args << QString("0x%1").arg(index.getSerial(), 8, 16, QChar('0'));
    args << "-v" << "2.2";
    args << "--hwBufSize" << QString::number(devGlob[index]->lastDevStatus.mstreamHwBufSize);
    args << "--ackSize" << (devGlob[index]->lastDevStatus.mstreamMultiAck ? "0" : "1");
    devGlob[index]->mstreamProc->setProgram(mStreamPath);
    devGlob[index]->mstreamProc->setArguments(args);
}


void MainWindow::start_process(const DeviceIndex &index)
{
    QProcess * p = devGlob[index]->mstreamProc;
    if(p->state() != QProcess::NotRunning) {
        stop_process(index);
    }

    p->start();

    qInfo()<<"mstreamProc->arguments():"<<p->arguments();
    if(p->waitForStarted(1000)){
        sendLog(index.getSerialStr().prepend("MStream started to "));
    }
    else{
        sendLog(index.getSerialStr().prepend("Failed to start MStream to "));
    }
}

void MainWindow::restart_process(const DeviceIndex &index)
{
    QProcess * p = devGlob[index]->mstreamProc;
    if(p->state() != QProcess::NotRunning){
        stop_process(index);
        ui->mstream_output->mstream_log_output(index.getSerialStr(), "\n\n########### restartProcess ###########");
    }
    set_mstream_args(index);
    start_process(index);
}

void MainWindow::stop_process(const DeviceIndex &index)
{
    QProcess * p = devGlob[index]->mstreamProc;
    if(p->state() == QProcess::NotRunning)
        return;

    p->blockSignals(true);
#ifdef __linux__
    kill(static_cast<__pid_t>(p->pid()), SIGTERM);
#else
    p->close();
#endif
    if(!p->waitForFinished(1000) && p->state() != QProcess::NotRunning) {
        qInfo()<<"!mstreamProc->waitForFinished()";
        qInfo()<<"mstreamProc->terminate();";
        p->terminate();
    }
    p->blockSignals(false);
    sendLog("Close MStream.");
}

void MainWindow::mstream_output_read()
{
    DeviceIndex index;
    for (const DeviceIndex &i: devGlob.keys()) {
        if(devGlob[i]->mstreamProc == sender()) {
            index = i;
            break;
        }
    }
    if(index.isNull())
        return;
    QProcess * mstreamProc = devGlob[index]->mstreamProc;

    QByteArray out;
    out.append(mstreamProc->readAllStandardOutput());
    out.append(mstreamProc->readAllStandardError());
    QString outStrFull(out);

    ui->mstream_output->mstream_table_output(index.getSerialStr(), outStrFull);
    ui->mstream_output->mstream_log_output(index.getSerialStr(), outStrFull.trimmed());

    if(outStrFull.indexOf("CloseClose")!=-1){
#ifdef QT_DEBUG
//        if(!isDevBlocked)
//            on_pushButtonBlockRun_clicked();
#endif
        qInfo() << "CloseClose";
    }
}

void MainWindow::process_error(QProcess::ProcessError err)
{
    DeviceIndex index;
    for(const auto &i: devGlob.keys()) {
        if(devGlob[i]->mstreamProc == sender()) {
            index = i;
            break;
        }
    }
    qWarning()<<QString("[%1] MStream process had an error: %2")
                .arg(index.getIdent()).arg(err);
}

void MainWindow::process_finished(int exitCode, QProcess::ExitStatus exitStatus)
{
    DeviceIndex index;
    for(const auto &i: devGlob.keys()) {
        if(devGlob[i]->mstreamProc == sender()) {
            index = i;
            break;
        }
    }

    qWarning()<<QString("MStream process %1 exit; exitCode=%2; dev=%3")
                .arg(exitStatus==QProcess::NormalExit ? "normal" : "crash")
                .arg(exitCode)
                .arg(index.isNull() ? "unknown" : index.getIdent());
    if(!index.isNull()) {
        start_process(index);
    }
}

void MainWindow::on_actionReset_MStream_triggered()
{
    for (DeviceIndex i: devGlob.keys()) {
        restart_process(i);
    }
}

void MainWindow::mstream_status_updated(QString serial, const MStreamStatus &status)
{
    int devCol = get_col_by_serial(serial);
    ui->tableWidgetDev->item(ROW_DEV_STREAM, devCol)->setText(status.recv_speed);
    ui->tableWidgetDev->item(ROW_DEV_MS_ERR, devCol)->setText(status.bad_format);
    bool hide_row = true;
    for(int col=0; col<ui->tableWidgetDev->columnCount(); ++col) {
        if(!ui->tableWidgetDev->item(ROW_DEV_MS_ERR, col)->text().isEmpty()) {
            hide_row = false;
            break;
        }
    }

    if(hide_row != ui->tableWidgetDev->isRowHidden(ROW_DEV_MS_ERR))
        ui->tableWidgetDev->setRowHidden(ROW_DEV_MS_ERR, hide_row);
    update_table_size(ui->tableWidgetDev);
}

int MainWindow::get_col_by_serial(QString serial)
{
    QTableWidget *table = ui->tableWidgetDev;
    for(int col=table->columnCount()-1; col>=0; --col) {
        if(getDevByTableItem(table->horizontalHeaderItem(col)).getSerialStr() == serial)
            return col;
    }
    return -1;
}

void MainWindow::update_table_size(QTableWidget * const table, bool widthOnly)
{
//    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

void MainWindow::delayed_write_config()
{
    config.write_config();
}

void MainWindow::write_run_history()
{
    RootConfig my_config = config.root;
    my_config.item_name = config.program_type + "_" + config.program_index + "_" + QString::number(config.run_number);

    for(int id_tmp : clients_config.keys()){
        RootConfig root_client = clients_config[id_tmp];
        QString program_type = root_client.get_data<QString>("program_type");
        QString program_index = root_client.get_data<QString>("program_index");
        QString run_number = QString::number(root_client.get_data<int>("run_number"));
        QString item_name = program_type + "_" + program_index + "_" + run_number;
        root_client.item_name = item_name;

        RootConfig my_child;
        QString child_name;
        for(int i=1; i<=clients_config.size(); ++i){
            child_name = QString("Client_%1").arg(i);
            my_child = my_config.child(child_name);
            QString program_type_tmp = my_child.get_data<QString>("program_type");
            QString program_index_tmp = my_child.get_data<QString>("program_index");
            if((program_type == program_type_tmp)&&(program_index == program_index_tmp)){
                my_child.append_child(root_client);
                my_config.remove_child(child_name);
                my_config.append_child(my_child);
                break;
            }
        }
    }

    if(megamaster_flg) {
        RootConfig run_root = RootConfig();
        run_root.set_data("program_type", config.program_type);
        run_root.set_data("run_index", config.run_index);
        run_root.set_data("run_number", config.run_number);
        QDateTime current_dt = QDateTime::currentDateTime();
        run_root.set_data("date_of_creation", current_dt);

        run_root.append_child(my_config);
        config.write_history_config(run_root);
    } else {
        emit config_transfer_to_rc(my_config.get_json_data());
    }
    writeRunHistoryTimer->stop();
}

void MainWindow::onRunError(QString run_index, int run_number, QString errStr)
{
    if(config.runEnable && config.run_index == run_index && config.run_number == run_number) {
        if(remoteConnected){
            QString mess = QString("Retransmit runError to the remoteControl in Run #%1_%2")
                    .arg(config.run_index).arg(config.run_number);
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning() << mess;
            emit sendRunError(run_index, run_number, errStr);
        } else {
            QString mess = QString("Errors in Run #%1_%2").arg(config.run_index).arg(config.run_number);
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning()<< mess;

            perfomeStop(true);
        }
//    } else {
//        qWarning() << QString("Got RunError from %1 run").arg(config.runEnable?"previous":"stopped");
    }
}

void MainWindow::onRunWarning(QString run_index, int run_number, QString warnStr)
{
    if(!warnStr.isEmpty() &&
            (!config.runEnable ||
             config.run_index != run_index ||
             config.run_number != run_number) )
        return;

    QLabel *label=nullptr;
    if(sender() == ui->clientManagerWidget){
        label = labelWarnClients;
        warnStr.prepend("Clients: ");
    }
    else
        if(devGlob.values().contains(qobject_cast<DeviceGlob*>(sender()))){
            label = labelWarnTtvxs;
            warnStr.prepend("Ttvxs: ");
        }
        else{
            label = labelWarnOther;
            warnStr.prepend("Other: ");
        }
    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    label->setText(warnStr);
    label->setHidden(false);

    updateGroupBoxWarn();
}

void MainWindow::onBackPressure(QString run_index, int run_number, bool backPressure)
{
    if(config.runEnable && config.run_index == run_index && config.run_number == run_number) {
        if(remoteConnected) {
            QString mess = QString("Retransmit backPressure(%3) to the remoteControl in Run #%1_%2")
                    .arg(config.run_index).arg(config.run_number).arg(backPressure);
            if(backPressure)
                qWarning() << mess;
            else
                qInfo() << mess;
            emit sendBackPressure(run_index, run_number, backPressure);
        } else {
            if(backPressure) {
                ui->radioButton_close->setChecked(true);
                qWarning()<< "BackPressure enabled";
            } else {
                ui->radioButton_open->setChecked(true);
                qInfo()<< "BackPressure disabled";

            }
        }
    }
}

void MainWindow::onEvNumCheckFail(QString mess)
{
    onRunWarning(config.run_index, config.run_number, mess);
}

void MainWindow::client_config_receiver(QMap<int, ClientConfig> client_config_)
{
    config.clients = client_config_;
    config.write_config();
}

void MainWindow::setClientTabState(bool isOk)
{
    setTabState(isOk, TAB_CLIENTS);
}

void MainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    QColor c = bar->tabTextColor(tabIndex);
    if(c.isValid() == isOk) {
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
    }
    bool prevOk = isReadForRun();
    tabOk[tabIndex] = isOk;
    bool newOk = isReadForRun();
    if(newOk != prevOk)
        updateProgramDescription();
}

void MainWindow::config_transfer_from_clients(int id, RootConfig root)
{
    for(int id_tmp : config.clients.keys()) {
        if(config.clients[id_tmp].id == id) {
            clients_config[id] = root;
        }
    }

    bool all_clients_config = true;
    for(int id_tmp : config.clients.keys()) {
        if(!clients_config.contains(id_tmp)) {
            all_clients_config = false;
        }
    }

    if(all_clients_config) {
        write_run_history();
    }
}
//-----------------------------------------------------------------


void MainWindow::disconnectFromDev(DeviceIndex index)
{
    if(!index.isNull() && !devGlob.contains(index)) return;
    emit deleteDevice(index);
    QElapsedTimer t;
    t.start();
    for(auto i : devGlob.keys()) {
        if(!index.isNull() && i!=index) continue;

        auto &dg = devGlob[i];
        dg->disconnectFromDev(&t);
        delete dg;
        devGlob.remove(i);
        int col = getChCfgCol(i);
        if(col!=-1) {
            ui->tableWidgetDev->removeColumn(col);
        }
        ui->mstream_output->remove_device(index.getSerialStr());

        if(!index.isNull()) return;
    }
}

void MainWindow::sendLog(QString log)
{
    qInfo() << log;
}

void MainWindow::readSettings()
{
    config.read_config();
    config.unify_config();
    initProgramDescription();

    ui->clientManagerWidget->init(config.clients);

    ++blockCfgSave;
    QVector<DeviceDescription> devicesAll;
    for(auto index : config.devicesCfg.keys()) {
        const auto &devCfg = config.devicesCfg[index];
        DeviceDescription dd(devCfg.device_index);
        dd.enabled = devCfg.enabled;
        devicesAll << dd;
        if(dd.enabled)
            addDevice(dd);
    }

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("state").toByteArray());
    settings.endGroup();

    const auto &cfg = config.default_cfg;

    ui->checkBox_external_ttl->setChecked(cfg.trigSrc.bit.ttl);
    ui->checkBox_timer->setChecked(cfg.trigSrc.bit.fixed_pulser);
    ui->checkBox_random->setChecked(cfg.trigSrc.bit.random_pulser);
    ui->spinBox_timer->setValue(cfg.trigTimerFreq);
    ui->spinBox_random->setValue(cfg.trigRandomFreq);
    ui->doubleSpinBox_dead_time->setValue(cfg.trigRandomDeadTimeUs);

    ui->checkBox_aux_cnt_for_trig->setChecked(cfg.aux_cnt_for_trig);

    ui->groupBoxStat->setChecked(cfg.stat_readout);
    ui->spinBoxStatPeriod->setValue(cfg.stat_period);
    ui->comboBoxPreset->setCurrentText(cfg.stat_preset.label);

    ui->actionAllow_Remote_Control->setChecked(config.allowRemoteControl);

    --blockCfgSave;
    discover_dialog->setSelected(devicesAll);

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.ttvxs-rc." + shortHostName + "." + config.program_index);
}

void MainWindow::saveSettings(bool force)
{
    if(blockCfgSave) return;
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("state", saveState());
    settings.endGroup();

    if(force) {
        config.write_config();
        writeConfigTimer->stop();
    } else {
        writeConfigTimer->start();
    }
}

int MainWindow::getChCfgCol(DeviceIndex i) const
{
    auto t = ui->tableWidgetDev;
    if(i.isNull()) return -1;
    for(int col=0; col<t->columnCount(); ++col) {
        if(getDevByTableItem(t->horizontalHeaderItem(col)) == i)
            return col;
    }
    return -1;
}

DeviceIndex MainWindow::getDevByTableItem(const QTableWidgetItem *item) const
{
    if(!item) return DeviceIndex();

    bool ok1,ok2;
    QList<QVariant> list = item->data(Qt::UserRole).toList();
    if(list.size()>=2) {
        quint16 id = static_cast<quint16>(list.first().toUInt(&ok1));
        quint64 ser = list.at(1).toULongLong(&ok2);
        if(ok1 && ok2)
            return DeviceIndex(id, ser);
    }
    return DeviceIndex();
}

DeviceIndex MainWindow::getDevByCol(int col) const
{
    auto item = ui->tableWidgetDev->horizontalHeaderItem(col);
    return getDevByTableItem(item);
}

bool MainWindow::isSlotBlocked(int row)
{
    return row==ROW_CH_SLOT_1 || row==ROW_CH_SLOT_11 || row==ROW_CH_SLOT_12;
}

bool MainWindow::perfomeStart(QString run_index, int run_number)
{
    blockControl(true);
    ui->pushButton_start->setEnabled(false);
    bool isOk = isReadForRun();
    QElapsedTimer timer;
    timer.start();
    if(!isOk) {
        QString msg = "Can't start run. There are some errors. Check ";
        if(!tabOk.contains(TAB_TRIG_DEV) || !tabOk[TAB_TRIG_DEV])
            msg.append("TTVXS device(s);");
        if(!tabOk.contains(TAB_CLIENTS) || !tabOk[TAB_CLIENTS])
            msg.append("Clients;");
        qCritical() << msg;
    }

    hasRunError = !isOk;
    if(isOk) {
        ui->pushButton_start->setEnabled(false);
        config.run_index = run_index;
        config.run_number = run_number;
        config.runEnable = true;
        config.write_config();
        ui->clientManagerWidget->set_run_enable(config.runEnable);
        ui->labelRunState->setText(QString("Starting Run #%1_%2...").arg(config.run_index).arg(config.run_number));
        ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
        ui->plainTextEdit_log->appendPlainText("");
        qInfo() << QString("!!! Preparing Run (run_index=%1;run_number=%2)... !!! ")
                   .arg(config.run_index)
                   .arg(config.run_number);
//        ui->trigMainForm->closeTrig();
        ui->radioButton_close->setChecked(true);
        emit clearStat();
        emit restartAll(config.run_index, config.run_number);

        // Get started state from Tcp Clients (sub-programms)
        while(!timer.hasExpired(10000)){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
            if(ui->clientManagerWidget->get_run_index()==config.run_index &&
                    ui->clientManagerWidget->get_run_number()==config.run_number &&
                    ui->clientManagerWidget->getRunState() != RS_STARTING )
                // Break on error or started
                break;
        }
        auto sttt = ui->clientManagerWidget->getRunState();
        if(sttt != RS_STARTED){
            qDebug() << "Some client failed to start";
            isOk = false;
        }
    }

    if(isOk) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 200);
//        ui->trigMainForm->openTrig();
        ui->radioButton_open->setChecked(true);
        QString txt = QString("Run #%1_%2 started").arg(config.run_index).arg(config.run_number);
        ui->labelRunState->setText(txt);
        qInfo() << "!!! Run started !!!";
        ui->pushButton_stop->setEnabled(true);
        QVector<quint64> devSerials;
        if(!megamaster_flg) {
            for(auto cfg : config.devicesCfg) {
                if(cfg.enabled)
                    devSerials << cfg.device_index.getSerial();
            }
        }
        evNumChecker->set_dev_serials(devSerials);
        if(megamaster_flg)
            evNumChecker->startCheck();
    } else {
        hasRunError = true;
        qCritical() << "!!! Run failed !!!";
        QString txt = QString("Run #%1_%2 failed").arg(config.run_index).arg(config.run_number);
        ui->labelRunState->setText(txt);
        if(config.runEnable)
            perfomeStop(true);
    }
    ui->labelRunState->setStyleSheet(isOk ? "QLabel { color : green;}" : "QLabel { color : red;}");
    updateStatStopButton();
    blockControl(false);

    if(isOk) {
        writeConfigTimer->start();
        writeRunHistoryTimer->start();
    }
    return isOk;
}

bool MainWindow::perfomeStop(bool isError)
{
    blockControl(true);
    bool isOk=true;
    ui->pushButton_stop->setEnabled(false);
    ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
    if(!hasRunError)
        ui->labelRunState->setText(QString("Stopping Run #%1_%2...").arg(config.run_index).arg(config.run_number));
    QString msg = QString("!!! Stopping Run (run_number=%1)... !!! ").arg(config.run_number);
    update();
    if(isError)
        qWarning() << msg;
    else
        qInfo() << msg;
    config.runEnable = false;
    config.write_config();
    ui->clientManagerWidget->set_run_enable(config.runEnable);

    evNumChecker->stopCheck();
//    ui->trigMainForm->closeTrig();
    ui->radioButton_close->setChecked(true);
    // TODO: Smart wait that all clients finished it's work
    if(remoteConnected) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 1000);
    } else {
        QTime t;
        t.start();
        int elapsedTime;
        while((elapsedTime=t.elapsed())<1000){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 1000-elapsedTime);
            QThread::msleep(100);
        }
    }
    emit stopAll();

    QElapsedTimer timer;
    timer.start();
    // Get stoped state from Tcp Clients (sub-programms)
    while(!timer.hasExpired(10000)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if(ui->clientManagerWidget->get_run_number()==config.run_number && ui->clientManagerWidget->getRunState() != RS_STOPPING )
            break;
    }
    if(ui->clientManagerWidget->getRunState() != RS_STOPPED){
        qDebug() << "Some client failed to stop";
        isOk = false;
    }


    if(isOk) {
        if(!hasRunError){
            ui->labelRunState->setText(QString("Run #%1_%2 %3")
                                       .arg(config.run_index).arg(config.run_number)
                                       .arg(isError ? "crashed" : "stopped"));
        }
        if(isError)
            qWarning() << "!!! Run stoped !!!";
        else
            qInfo() << "!!! Run stoped !!!";
    } else {
        if(!hasRunError){
            hasRunError = true;
            ui->labelRunState->setText(QString("Run #%1_%2 stop failed").arg(config.run_index).arg(config.run_number));
        }
        qCritical() << "!!! Run stop failed !!!";
        emit stopAll();
    }
    ui->labelRunState->setStyleSheet((isOk && !isError) ? "QLabel { color : green;}" : "QLabel { color : red;}");
    updateStatStopButton();

    blockControl(false);
    return isOk;
}

bool MainWindow::isReadForRun() const
{
    bool allOk=true;
    for(bool ok: tabOk.values())
        allOk &= ok;
    return allOk;
}

void MainWindow::initProgramDescription()
{
    progDesc.type = config.program_type;
    progDesc.index = config.program_index;
    progDesc.name = progDesc.getName();
}

void MainWindow::updateProgramDescription()
{
    QString remoteState;

    progDesc.options.clear();
    progDesc.options["runEnable"] = QVariant(config.runEnable).toString();
    progDesc.options["run_index"] = config.run_index;
    progDesc.options["run_number"] = QVariant(config.run_number).toString();

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    bool readyForRun = isReadForRun();
    for(const ClientConfig &c: config.clients.values()) {
        remoteState += QString("%1 #%2; ").arg(c.program_type).arg(c.program_index);
        /*
        progDesc.options[QString("client_%1_progType").arg(c.id)] = c.progType;
        progDesc.options[QString("adc_%1_progIndex").arg(c.id)] = c.progIndex;
        progDesc.options[QString("adc_%1_readout").arg(c.id)] = QVariant(c.readout).toString();
        */
        progDesc.options[QString("Client-%1").arg(c.id)] = c.program_type +" "+ c.program_index +" , "+ QVariant(c.readout).toString();
    }
    emit changeProgramDescription(progDesc);
    emit sendRemoteState(remoteState, readyForRun);
}

void MainWindow::blockControl(bool block)
{
    const bool &isRun = config.runEnable;
    const bool enabled_main = !block && (!isRun || config.debug_mode);
    const bool enabled_menu_and_trig = !block && (!isRun || config.debug_mode || megamaster_flg);
    emit sendControlEnable(enabled_main);
    ui->menuOptions->setEnabled(enabled_menu_and_trig);

    auto table = ui->tableWidgetDev;
    for(int col=0; col<table->columnCount(); ++col){
        for(int row=ROW_CH_SLOT_1; row<=ROW_CH_SLOT_21; ++row) {
            if(isSlotBlocked(row))
                continue;
            auto i = table->item(row, col);
            if(enabled_main) {
                i->setFlags(i->flags() | Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
            } else {
                i->setFlags(i->flags() & (~Qt::ItemIsUserCheckable) & (~Qt::ItemIsEnabled));
            }
        }
    }



    ui->checkBox_aux_cnt_for_trig->setEnabled(enabled_main);
    ui->groupBoxStat->setEnabled(enabled_main);
    ui->groupBox_trigger_source->setEnabled(enabled_main);
    ui->groupBox_trigger_state->setEnabled(enabled_menu_and_trig);
}

void MainWindow::updateStatStopButton()
{
    ui->pushButton_start->setEnabled(!config.runEnable && !remoteConnected);
    ui->pushButton_stop->setEnabled(config.runEnable && !remoteConnected);
}

void MainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!labelWarnTtvxs->isHidden()) {
        hidden = false;
        warnStrList << labelWarnTtvxs->text();
    }
    if(!labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << labelWarnClients->text();
    }
    if(!labelWarnOther->isHidden()) {
        hidden = false;
        warnStrList << labelWarnOther->text();
    }

    if(!hidden)
        emit sendRunWarn(config.run_index, config.run_number, warnStrList.join("; "));
}

void MainWindow::addLog(QtMsgType severity, QString log, QString thrName)
{
    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<< "";
    }
    QTextCharFormat tf = ui->plainTextEdit_log->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEdit_log->setCurrentCharFormat(tf);
    ui->plainTextEdit_log->appendPlainText(QString("[%1] [%2] %3")
                                           .arg(QTime::currentTime().toString(), thrName, log));
}

void MainWindow::on_actionChange_devices_triggered()
{
    if(discover_dialog->exec() == QDialog::Accepted) {
        const auto devSelected = discover_dialog->getSelectedDevices();
        for(auto index : config.devicesCfg.keys()) {
            bool found = false;
            for(auto dd : devSelected) {
                if(dd.getIndex()==index) {
                    found = true;
                    break;
                }
            }
            if(!found) {
                config.devicesCfg.remove(index);
                disconnectFromDev(index);
            }
        }

        for(auto dd : devSelected) {
            auto &ds = config.devicesCfg[dd.getIndex()];
            ds.device_index = dd.getIndex();
            ds.enabled = dd.enabled;
            ds.unify_config(config.default_cfg);

            if(dd.enabled) {
                addDevice(dd);
            } else {
                disconnectFromDev(dd.getIndex());
            }
        }
        saveSettings();
    }
}

void MainWindow::on_actionReset_timestamp_triggered()
{
    emit resetTimestamp();
}

void MainWindow::on_actionSoftclear_triggered()
{
    emit softClear();
}

void MainWindow::on_actionClear_stats_triggered()
{
    emit clearStat();
}

void MainWindow::on_actionSNMP_dialog_triggered()
{
    SnmpDialog dialog(this);
    dialog.exec();
    close();
}

void MainWindow::on_actionEvent_number_reset_triggered()
{
    emit evNumReset();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.allowRemoteControl = val;
    config.write_config();
    emit allowRemoteControl(config.allowRemoteControl);
    updateProgramDescription();
}

void MainWindow::on_actionAbout_triggered()
{
    AboutBox::about();
}

void MainWindow::on_actionQuit_triggered()
{
    close();
}

void MainWindow::on_pushButton_start_clicked()
{
    megamaster_flg = true;
    clients_config.clear();

    bool success = perfomeStart(config.program_index, ++config.localRunNumber);

    if(success && config.clients.isEmpty()){
        RootConfig run_root = RootConfig();
        run_root.set_data("program_type", config.program_type);
        run_root.set_data("run_index", config.run_index);
        run_root.set_data("run_number", config.run_number);
        QDateTime current_dt = QDateTime::currentDateTime();
        run_root.set_data("date_of_creation", current_dt);

        RootConfig my_root = config.root;
        my_root.item_name = config.program_type + "_" + config.program_index + "_" + QString::number(config.run_number);

        run_root.append_child(my_root);
        config.write_history_config(run_root);
    }
    config.write_config();
}

void MainWindow::on_pushButton_stop_clicked()
{
    perfomeStop(false);
}

void MainWindow::on_radioButton_open_toggled(bool checked)
{
    config.default_cfg.trig_open = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_checkBox_external_ttl_toggled(bool checked)
{
    config.default_cfg.trigSrc.bit.ttl = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_checkBox_timer_toggled(bool checked)
{
    ui->spinBox_timer->setEnabled(checked);
    config.default_cfg.trigSrc.bit.fixed_pulser = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_checkBox_random_toggled(bool checked)
{
    ui->spinBox_random->setEnabled(checked);
    ui->doubleSpinBox_dead_time->setEnabled(checked);
    ui->labelDeadTime->setEnabled(checked);
    config.default_cfg.trigSrc.bit.random_pulser = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::onRemoteStartRequest(QString globIndex, int run_number, quint64 startEvNumber)
{
    Q_UNUSED(startEvNumber)
    megamaster_flg = false;
    bool success = perfomeStart(globIndex, run_number);
    clients_config.clear();
    emit remoteStartCompleted(success, globIndex, run_number);
    if(success && config.clients.isEmpty()) {
        emit config_transfer_to_rc(config.root.get_json_data());
    }

    updateProgramDescription();
}

void MainWindow::onRemoteStopRequest(QString globIndex, int run_number, quint64 stopEvNumber)
{
    Q_UNUSED(stopEvNumber)
    bool success = perfomeStop(false);

    emit remoteStopCompleted(success, globIndex, run_number);
    updateProgramDescription();
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl) == 0) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        progDesc.interfaces.push_back(interface);
    }

    ProgramInterface &i = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !status.portInUse;

    if(status.portInUse)
        i.setPeer(status.peerHost, status.peerPort);
    else
        i.peers.clear();

    updateProgramDescription();
    bool remoteDisconnectedOnRun = config.runEnable
            && remoteConnected && !status.portInUse;
    remoteConnected = status.portInUse;
    if(remoteDisconnectedOnRun) {
        qWarning() << "Stop run due to remoteControl disconnect";
        perfomeStop(true);
    }
    updateStatStopButton();
}

void MainWindow::on_spinBox_timer_valueChanged(int val)
{
    config.default_cfg.trigTimerFreq = val;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_spinBox_random_valueChanged(int val)
{
    config.default_cfg.trigRandomFreq = val;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_doubleSpinBox_dead_time_valueChanged(double val)
{
    config.default_cfg.trigRandomDeadTimeUs = val;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_checkBox_aux_cnt_for_trig_toggled(bool checked)
{
    config.default_cfg.aux_cnt_for_trig = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_groupBoxStat_toggled(bool checked)
{
    config.default_cfg.stat_readout = checked;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_spinBoxStatPeriod_valueChanged(int period_ms)
{
    config.default_cfg.stat_period = static_cast<quint16>(period_ms);
    config.unify_config();
    writeSetup();
}

void MainWindow::on_comboBoxPreset_currentIndexChanged(int index)
{
    StatPreset stat_preset;
    int preset_type = ui->comboBoxPreset->itemData(index).toInt();
    stat_preset = StatPresetFabric::getPreset(preset_type);

    config.default_cfg.stat_preset = stat_preset;
    config.unify_config();
    writeSetup();
}

void MainWindow::on_tableWidgetDev_cellChanged(int row, int column)
{
    if(row == ROW_NAME) {
        auto table = ui->tableWidgetDev;
        DeviceIndex index = getDevByCol(column);
        if(index.isNull())
            return;
        if(!config.devicesCfg.contains(index))
            return;
        config.devicesCfg[index].custom_name = table->item(row, column)->text();
        saveSettings();
        return;
    }
    if(row < ROW_CH_SLOT_1 || row > ROW_CH_SLOT_21) return;
    DeviceIndex index = getDevByCol(column);
    if(index.isNull()) return;
    const int slot = row-ROW_CH_SLOT_1+1;
    if(isSlotBlocked(row)) return;
    const int ch = slotToCh(slot);

    Q_ASSERT(config.devicesCfg.contains(index));
    auto &cfg = config.devicesCfg[index];
    bool en = ui->tableWidgetDev->item(row, column)->checkState() == Qt::Checked;
    if(((cfg.vxsChMask & (1<<ch)) != 0) != en) {
        if(en)
            cfg.vxsChMask |= 1<<ch;
        else
            cfg.vxsChMask &= ~static_cast<quint32>(1<<ch);

        writeSetup(index);
    }
}

void MainWindow::deviceStatusUpdated(TtvxsStatus st)
{
    const auto &index = st.devIndex;
    if(!devGlob.contains(index))
        return;

    auto &lastDevStatus = devGlob[index]->lastDevStatus;
    auto prevDevStatus = lastDevStatus;
    auto tableDev = ui->tableWidgetDev;
    const int colCh = getChCfgCol(index);

    Q_ASSERT(colCh!=-1);
    if(colCh == -1) return;

    const auto &aux_cnt = st.aux_cnt;
    const auto &devCfg = config.devicesCfg[index];
    const auto devMetricPrefix = getDevMetricPrefix(index);
    auto vxsChMask = devCfg.vxsChMask;
    for(int ch=0; ch<18; ++ch) {
        bool inputSt = st.inputState & (1<<ch);
        bool timeoutSt = st.timeoutState & (1<<ch);
        const int slot = chToSlot(ch);
        if(!prevDevStatus.devIndex.isNull()) {
            if(timeoutSt != ((prevDevStatus.timeoutState & (1<<ch)) != 0)) {
                QString baseStr = QString("[%3] %1 %4 timeout on slot=%2")
                        .arg(index.getIdent()).arg(slot)
                        .arg(QDateTime::currentDateTime().toString());
                if(timeoutSt) {
                    qWarning() << baseStr.arg("got");
                } else {
                    qInfo() << baseStr.arg("remove");
                }
            }
        }
        const int row = ROW_CH_SLOT_1+slot-1;
        auto chItem = tableDev->item(row, colCh);
        chItem->setBackgroundColor(inputSt ? Qt::green : Qt::transparent);
        if(timeoutSt != (chItem->textColor() == Qt::red)) {
            chItem->setTextColor(timeoutSt ? Qt::red : QColor());
            chItem->setText(timeoutSt ? "timeout" : "");
        }
        bool ch_en = vxsChMask&(1<<ch);

        tableDev->item(row, colCh)->setText(ch_en ? aux_cnt.getBusyTime(ch) : "");
        if(ch_en)
            MetricSender::send(QString(devMetricPrefix+"time.busy.slot.%1").arg(slot), aux_cnt.getBusyTimeUs(ch));
    }
    tableDev->item(ROW_UP_TIME, colCh)->setText(aux_cnt.getUpTime());
    MetricSender::send(devMetricPrefix+"time.uptime", aux_cnt.getUpTimeSec());
    tableDev->item(ROW_LEMO_CNT, colCh)->setText(aux_cnt.getTtlCnt());
    tableDev->item(ROW_SFP_CNT, colCh)->setText(aux_cnt.getSfpCnt());
    tableDev->item(ROW_CUR_EV_NUM, colCh)->setText(QString("%L1").arg(st.curEvNum));
    MetricSender::send(devMetricPrefix+"curEvNum", st.curEvNum);

    if(devCfg.slot){
        tableDev->item(ROW_CH_SLOT_1-1+devCfg.slot, colCh)->setText(QString("ttvxs %1").arg(aux_cnt.getResultingBusyTime()));
        MetricSender::send(devMetricPrefix+"time.busy.total", aux_cnt.getResultingBusyTimeUs());
    }
    tableDev->item(ROW_RUN_TIME, colCh)->setText(aux_cnt.getRunTime());
    MetricSender::send(devMetricPrefix+"time.runtime", aux_cnt.getRunTimeSec());
    tableDev->item(ROW_RUN_TRIG_CANDID, colCh)->setText(aux_cnt.getTrigRunCandCnt());
    tableDev->item(ROW_RUN_TRIG_ACCEPTED, colCh)->setText(aux_cnt.getTrigRunAcceptCnt());

    tableDev->item(ROW_DEV_FW, colCh)->setText(st.fwStr);
    tableDev->item(ROW_DEV_TEMP, colCh)->setText(QString("%1 %2C").arg(st.temp, 0, 'f', 1).arg(QChar(0x2da)));

    auto wr_item = tableDev->item(ROW_WR_TIME, colCh);
    QDateTime wrTime = QDateTime::fromMSecsSinceEpoch(st.wrEpStatus.mSecsSinceEpoch());
    bool wrTimeValid = st.wrEpStatus.timeValid();
    wr_item->setText(st.wrEpStatus.timeValid() ? wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss") : "invalid"); // add ".zzz" for ms
    wr_item->setBackgroundColor(wrTimeValid ? Qt::green : Qt::lightGray);

    tableDev->horizontalHeaderItem(colCh)->setTextColor(st.isConnected ? QColor() : Qt::gray);
    tableDev->item(ROW_STATUS, colCh)->setText(st.isConnected ? "Connected" : "Disconnected");
    tableDev->item(ROW_STATUS, colCh)->setTextColor(st.isConnected ? Qt::darkGreen : Qt::red);
    auto mcu_ver_item = tableDev->item(ROW_DEV_MCU_VER, colCh);
    mcu_ver_item->setText(st.bmc.getVerStr());
    mcu_ver_item->setToolTip(QString("overall - %1\npower - %2\nPLL- %3")
                             .arg(st.bmc.get_overall_status_str())
                             .arg(st.bmc.get_power_status_str())
                             .arg(st.bmc.get_pll_status_str()));
//    QBrush mcu_ver_brush = mcu_ver_item->background();
//    mcu_ver_brush.setColor(QColor(255,0,0));
//    mcu_ver_item->setBackground(mcu_ver_brush);
    mcu_ver_item->setBackgroundColor(st.bmc.get_overall_status_col());

    lastDevStatus = st;
    const auto &aux1 = st.aux_cnt;
    const auto &aux2 = prevDevStatus.aux_cnt;
    tableDev->item(ROW_RUN_TRIG_RATE, colCh)->setText(aux1.getRunTrigRate(aux2));
    tableDev->item(ROW_LEMO_RATE, colCh)->setText(aux1.getTtlRate(aux2));
    tableDev->item(ROW_SFP_RATE, colCh)->setText(aux1.getSfpRate(aux2));
    if(st.mstreamMultiAck != prevDevStatus.mstreamMultiAck ||
            st.mstreamHwBufSize != prevDevStatus.mstreamHwBufSize) {
        restart_process(index);
    }

    if(!st.isConnected && config.runEnable) {
        onRunError(config.run_index, config.run_number, QString("TTVXS #%1 s/n:%2 offline")
                   .arg(config.program_index).arg(st.devIndex.getSerialStr()));
    }
}

void MainWindow::device_updated(DeviceDescription dd)
{
    if(!config.devicesCfg.contains(dd.getIndex())) return;
    auto &cfg = config.devicesCfg[dd.getIndex()];
    if(!cfg.enabled) return;

    if(cfg.slot != dd.chassis_slot) {
        cfg.slot = dd.chassis_slot;

        const int col = getChCfgCol(dd.getIndex());
        if(col == -1) return;
        auto ttvxsItem = ui->tableWidgetDev->item(ROW_CH_SLOT_1+dd.chassis_slot-1, col);
        if(ttvxsItem->text() != "ttvxs") {
            ttvxsItem->setText("ttvxs");
            ui->tableWidgetDev->resizeColumnsToContents();
        }
    }
}
