#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QProcess>
#include <QTableWidget>
#include <QLabel>

#include "TtvxsConfig.h"
#include "DeviceGlob.h"
#include "Config.h"
#include "mongo/StartupOptions.h"
#include "mldiscover/DeviceDescription.h"
#include "mldiscover/DeviceIndex.h"
#include "device-discover/DiscoverDialog.h"
#include "mregdev/qtmregdevice.h"
#include "mongo/BaseConfig.h"
#include "configuration-manager/ConfigurationManager.h"
#include "m-stream-output/MStreamOutput.h"
#include "pnp-server/PNPMessage.h"

class EvNumChecker;

namespace Ui {
class MainWindow;
}

class PNPServer;
class RemoteControlServer;
class RemoteControlServerStatus;

class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(StartupOptions &start_opt, QWidget *parent = nullptr);
    ~MainWindow();

signals:
    void setupChanged(TtvxsConfig cfg);
    void setDevAddr(DeviceDescription dd);
    void deleteDevice(DeviceIndex);
    void closeAll();
    void resetTimestamp(DeviceIndex=DeviceIndex());
    void evNumReset(DeviceIndex=DeviceIndex());
    void softClear(DeviceIndex=DeviceIndex());
    void clearStat(DeviceIndex=DeviceIndex());
    void changeProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void deleteAll() const;
    void allowRemoteControl(bool) const;
    void sendRemoteState(QString,bool) const;
    void sendRunError(QString, int, QString) const;
    void sendRunWarn(QString, int, QString) const;
    void sendBackPressure(QString, int, bool) const;

    void config_transfer_to_rc(QByteArray data);

    void sendControlEnable(bool block) const;
    void stopAll() const;
    void restartAll(QString runIndex, int runNumber) const;

    void remoteStartCompleted(bool, QString, int) const;
    void remoteStopCompleted(bool, QString, int) const;

private slots:
    void addLog(QtMsgType severity, QString log, QString thrName);
    void on_actionChange_devices_triggered();
    void on_actionReset_timestamp_triggered();
    void on_actionSoftclear_triggered();
    void on_actionClear_stats_triggered();
    void on_actionSNMP_dialog_triggered();
    void on_actionEvent_number_reset_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_actionAbout_triggered();
    void on_actionQuit_triggered();
    void on_pushButton_start_clicked();
    void on_pushButton_stop_clicked();
    void on_radioButton_open_toggled(bool checked);
    void on_checkBox_external_ttl_toggled(bool checked);
    void on_checkBox_timer_toggled(bool checked);
    void on_checkBox_random_toggled(bool checked);

    void onRemoteStartRequest(QString globIndex, int run_number, quint64 startEvNumber);
    void onRemoteStopRequest(QString globIndex, int run_number, quint64 stopEvNumber);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);

    void on_spinBox_timer_valueChanged(int val);
    void on_spinBox_random_valueChanged(int val);
    void on_doubleSpinBox_dead_time_valueChanged(double val);

    void on_checkBox_aux_cnt_for_trig_toggled(bool checked);

    void on_groupBoxStat_toggled(bool checked);
    void on_spinBoxStatPeriod_valueChanged(int period_ms);
    void on_comboBoxPreset_currentIndexChanged(int index);

    void on_tableWidgetDev_cellChanged(int row, int column);

    void deviceStatusUpdated(TtvxsStatus st);
    void device_updated(DeviceDescription dd);

    //---configuration manager
    void on_actionConfiguration_manager_triggered();
    void switch_to(QString program_index, QString configuration_name);

    //---mstream slots
    void set_mstream_args(const DeviceIndex &index);
    void start_process(const DeviceIndex &index);
    void restart_process(const DeviceIndex &index);
    void stop_process(const DeviceIndex &index);
    void mstream_output_read();
    void process_error(QProcess::ProcessError err);
    void process_finished(int exitCode, QProcess::ExitStatus exitStatus);
    void on_actionReset_MStream_triggered();
    void mstream_status_updated(QString serial, const MStreamStatus &status);
    int get_col_by_serial(QString serial);
    static void update_table_size(QTableWidget * const table, bool widthOnly=false);

    void delayed_write_config();
    void write_run_history();

    void onRunError(QString runIndex, int run_number, QString errStr);
    void onRunWarning(QString runIndex, int runNumber, QString warnStr);
    void onBackPressure(QString runIndex, int run_number, bool backPressure);
    void onEvNumCheckFail(QString mess);

    void client_config_receiver(QMap<int, ClientConfig> client_config_);
    void setClientTabState(bool isOk);
    void setTabState(bool isOk, int tabIndex);
    void config_transfer_from_clients(int id, RootConfig root);

private:
    void init_GUI_style();
    void init_GUI_default_settings();
    void initTableDev();
    void initStatPresets();
    void writeSetup(DeviceIndex index=DeviceIndex());
    void addDevice(DeviceIndex index);
    void addDevice(DeviceDescription dd);
    void disconnectFromDev(DeviceIndex index);
    void sendLog(QString log);
    void readSettings();
    void saveSettings(bool force=false);
    int getChCfgCol(DeviceIndex i) const;
    DeviceIndex getDevByTableItem(const QTableWidgetItem *item) const;
    DeviceIndex getDevByCol(int col) const;
    bool isSlotBlocked(int row);
    bool perfomeStart(QString run_index, int run_number);
    bool perfomeStop(bool isError);
    bool isReadForRun() const;
    void initProgramDescription();
    void updateProgramDescription();
    void blockControl(bool block);
    void updateStatStopButton();
    void updateGroupBoxWarn();

    Ui::MainWindow *ui;
    ConfigurationManager *configuration_manager;
    QString program_index;
    Config config;
    QMap<int, RootConfig> clients_config;
    DiscoverDialog * const discover_dialog;
    PNPServer * const pnpServer;
    ProgramDescription progDesc;
    RemoteControlServer * const remoteControl;
    QThread * const remoteControlThread;
    bool remoteConnected;
    EvNumChecker * const evNumChecker;
    bool hasRunError;
    int blockCfgSave;
    QMap<DeviceIndex, DeviceGlob *> devGlob;
    QTimer * const writeConfigTimer;
    QTimer * const writeRunHistoryTimer;
    QMap<int, bool> tabOk;
    bool megamaster_flg;

    QLabel *labelWarnTtvxs;
    QLabel *labelWarnClients;
    QLabel *labelWarnOther;
};

#endif // MAINWINDOW_H
