#include "TtvxsDevice.h"

#include "TtvxsRegs.h"

#include "regio/regop.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

namespace {
constexpr fw_version_t REV_TTVXS_EVNUM_SET(1, 0, 28380);
constexpr fw_version_t REV_TTVXS_RUN_CSR(1, 0, 28390);
constexpr fw_version_t REV_TTVXS_AUX_CNT(1, 0, 28412);
constexpr fw_version_t REV_TTVXS_STAT_READOUT(1, 0, 28365);
constexpr fw_version_t REV_TTVXS_AUX_FOR_TRIG(1, 0, 28381);
constexpr fw_version_t REV_EVNUM_INCREMENT(1, 0, 28510);

quint16 getMcuData(const RegOpVector &regOp, int base, int reg){
    int v = base+reg-REG_BMC_BASE;
    return regOp[v].data;
}
}

TtvxsDevice::TtvxsDevice(DeviceIndex devIndex, QObject *parent) :
    QtMregDevice(devIndex.getDevId(), devIndex.getSerial(), parent),
    configValid(false)
{
    status.devIndex = devIndex;
    connect(this, SIGNAL(TempChanged(double)), SLOT(onTempChanged(double)));
}

void TtvxsDevice::writeSetup()
{
    setSetup(configCashe);
}

void TtvxsDevice::setSetup(const TtvxsConfig &cfg)
{
    if(cfg.device_index != getDeviceIndex())
        return;
    if(!isConnected()) {
        configCashe = cfg;
        return;
    }
    writeLiveMagic();
    RegOpVector regOp;

    if(!configValid || configCashe.trig_open != cfg.trig_open
            || configCashe.aux_cnt_for_trig != cfg.aux_cnt_for_trig) {
        reg_rc_ctrl rcCtrl;
        rcCtrl.all = regRead(REG_RC_CTRL);

        if(getDeviceFwVersion() >= REV_TTVXS_RUN_CSR) {
            reg_base_ctrl base_crtl;
            base_crtl.bit.run_csr = cfg.trig_open;
            regOp << RegWrite(REG_DEVICE_CTRL, base_crtl.all);
        } else {
            rcCtrl.bit.run = cfg.trig_open;
        }
        if(getDeviceFwVersion() > REV_TTVXS_AUX_FOR_TRIG)
            rcCtrl.bit.en_aux_cnt = cfg.aux_cnt_for_trig;
        regOp << RegWrite(REG_RC_CTRL, rcCtrl.all);
    }
    if(!configValid || configCashe.trigSrc.all != cfg.trigSrc.all) {
        regOp << RegWrite(REG_RC_TRIG_SRC, cfg.trigSrc.all);
//        qInfo() << QString("[%1] trigSrc:   0x%2").arg(getIdent())
//                   .arg(cfg.trigSrc.all, 8, 16, QChar('0'));
    }
    if(!configValid || configCashe.vxsChMask != cfg.vxsChMask) {
        regOp << RegWrite32(REG32_RC_BUSY_MASK, cfg.vxsChMask);
//        qInfo() << QString("[%1] vxsChMask: 0x%2").arg(getIdent())
//                   .arg(cfg.vxsChMask, 8, 16, QChar('0'));
    }
    if(!configValid || configCashe.trigTimerFreq != cfg.trigTimerFreq) {
        quint32 period = static_cast<quint32>(125e6/cfg.trigTimerFreq);
        regOp << RegWrite32(REG32_FIXED_FREQ_PULSER_PERIOD, period);
//        qInfo() << QString("[%1] fixed period:  %2 (%3 Hz)").arg(getIdent()).arg(period)
//                   .arg(cfg.trigTimerFreq);
    }
    if(!configValid || configCashe.trigRandomFreq != cfg.trigRandomFreq) {
        double period = 1.0 * cfg.trigRandomFreq * (quint32)(-1) / 125e6;
        if (period < 1) period = 1;
        if (period > quint32(-1)) period = quint32(-1);
        quint32 regValue = static_cast<quint32>(period);
        regOp << RegWrite32(REG32_RANDOM_PULSER_PROBABILITY, regValue);
//        qInfo() << QString("[%1] rnd period:    %2 (%3 Hz)").arg(getIdent()).arg(period)
//                   .arg(cfg.trigRandomFreq);
    }
    if(!configValid || !qFuzzyCompare(configCashe.trigRandomDeadTimeUs, cfg.trigRandomDeadTimeUs)) {
        quint32 period = static_cast<quint32>(cfg.trigRandomDeadTimeUs/8e-3);
        regOp << RegWrite32(REG32_RANDOM_PULSER_DEAD_TIME, period);
//        qInfo() << QString("[%1] rnd dead time: %2 (%3us)").arg(getIdent()).arg(period)
//                   .arg(cfg.trigRandomDeadTimeUs, 0, 'f', 2);
    }

    regOp << writeStatCfg(cfg);


    auto regRes = regOpExec(regOp);

    configCashe = cfg;
    if(regRes.size() != regOp.size() || !isConnected()) {
        qWarning() << "Faild to write configuration to device";
        return;
    }

    configValid = checkLiveMagic();
}

RegOpVector TtvxsDevice::writeStatCfg(const TtvxsConfig &cfg)
{
    RegOpVector regOp;
    if(getDeviceFwVersion()< REV_TTVXS_STAT_READOUT)
        return regOp;

    if(!configValid || configCashe.stat_readout != cfg.stat_readout) {
        regOp << RegWrite(REG_STAT_CTRL, cfg.stat_readout ? 1 : 0);
    }
    if(!configValid || configCashe.stat_period != cfg.stat_period) {
        regOp << RegWrite(REG_STAT_PERIOD, cfg.stat_period);
    }

    if (!configValid || configCashe.stat_preset.regs != cfg.stat_preset.regs) {
        auto regs = cfg.stat_preset.regs;
        if(regs.isEmpty()) {
            qCritical() << "Statistic preset is empty; init it with one address equal to 0";
            regs << 0;
        }
        regOp << RegWrite(REG_STAT_BLOCK_SIZE,
                          static_cast<quint16>(regs.size()-1));
        for(quint16 i=0; i<regs.size(); ++i) {
            regOp << RegWrite(REG_STAT_RAM_ADDR, i);
            regOp << RegWrite(REG_STAT_RAM_DATA, regs[i]);
        }
    }
    return regOp;
}

void TtvxsDevice::evNumReset(DeviceIndex index)
{
    if(!index.isNull() && index!=getDeviceIndex())
        return;
    if(getDeviceFwVersion() < REV_TTVXS_EVNUM_SET) {
        qWarning() << QString("[%1] Device doesn't support 'evNumReset' function.").arg(getIdent());
        return;
    }
    RegOpVector regOp;
    regOp << RegWrite64(REG64_EVENT_NUMBER_LOAD, 0);
    regOpExec(regOp);
}

void TtvxsDevice::onTempChanged(double temp)
{
    status.temp = temp;
}

void TtvxsDevice::onRestartAll(QString, int)
{
    softClear(DeviceIndex());
    evNumReset();
    sendResetTimestamp();
    clearStat(DeviceIndex());
}

void TtvxsDevice::statusPoll()
{
    QtMregDevice::statusPoll();
    if (!isConnected())
        return;

    RegOpVector v;

    int pos = 0;
//    quint16 ctrlRegVal = configCashe.run ? (1<<15) : 0;
    // Lock counters
    reg_base_ctrl base_crtl;
    base_crtl.bit.run_csr = configCashe.trig_open;
    base_crtl.bit.cntLock = 1;
    v << RegWrite(REG_DEVICE_CTRL, base_crtl.all); ++pos;
    const int trigStatePos = pos;
    v << RegRead(REG_RC_TRIG_STATE); ++pos;
    const int busyInputStatePos = pos;
    v << RegRead32(REG32_BUSY_INPUT_STATE); pos+=2;
    const int busyTimeoutPos = pos;
    v << RegRead32(REG32_BUSY_TIMEOUT_STATE); pos+=2;
    const int buildParamPos = pos;
    v << RegRead64(REG128_BUILD_PARAMS); pos+=4;
    v << RegRead64(REG128_BUILD_PARAMS+4); pos+=4;
    const int hwipLinkDOwnPos = pos;
    v << RegRead64(REG64_HWIP_Link_Down); pos+=4;
    const int regIoErrPos = pos;
    v << RegRead64(REG64_REG_IO_ERROR); pos+=4;
    const int sfpErrEthPos = pos;
    v << RegRead64(REG64_SFP_ERROR_ETH); pos+=4;
    const int hwipErrPos = pos;
    v << RegRead64(REG64_HWIP_ERROR); pos+=4;
    const int txErrEthPos = pos;
    v << RegRead64(REG64_TX_ERROR_ETH); pos+=4;
    const int rxErrEthPos = pos;
    v << RegRead64(REG64_RX_ERROR_ETH); pos+=4;
    const int auxCntPos = getDeviceFwVersion() >= REV_TTVXS_AUX_CNT ? pos : -1;
    if(auxCntPos >= 0) {
        for(quint16 i=0; i<TtvxsAuxCounters::CNT_NUM; ++i) {
            v << RegRead64(REG_AUX_CNT_BASE+i*4); pos+=4;
        }
    }
    const int curEvNumPos = pos;
    v << RegRead64(REG64_CUR_EVENT_NUMBER); pos+=4;
    const int trigQueuePos = pos;
    v << RegRead(REG_TRIG_QUEUE); pos+=1;


    const int mcuPos = pos;
    for(quint16 reg=REG_BMC_BASE; reg<=REG_BMC_CUR_BP_33; ++reg) {
        v << RegRead(reg); ++pos;
    }

    // Unlock counters
    base_crtl.bit.cntLock = 0;
    v << RegWrite(REG_DEVICE_CTRL, base_crtl.all); ++pos;

    RegOpVector regRes = regOpExec(v);

    status.trigState = regRes[trigStatePos].data;
    status.inputState = get32(regRes, busyInputStatePos);
    status.timeoutState = get32(regRes, busyTimeoutPos);
    for(int i=0; i<4; ++i)
        status.buildParams[i] = get32(regRes, buildParamPos+4*i);
    status.curEvNum = get64(regRes, curEvNumPos);
    status.trigQueue = regRes[trigQueuePos].data;

    // todo show HWIP errors
    status.hwipErrors.hwipLinkDown= get64(regRes, hwipLinkDOwnPos);
    status.hwipErrors.regIoError  = get64(regRes, regIoErrPos);
    status.hwipErrors.sfpErrorEth = get64(regRes, sfpErrEthPos);
    status.hwipErrors.hwipError   = get64(regRes, hwipErrPos);
    status.hwipErrors.txErrorEth  = get64(regRes, txErrEthPos);
    status.hwipErrors.rxErrorEth  = get64(regRes, rxErrEthPos);

    if(auxCntPos >= 0) {
        auto &aux_cnt = status.aux_cnt;
        for(int i=0; i<aux_cnt.CNT_NUM; ++i) {
            aux_cnt.data[i] = get64(regRes, auxCntPos+i*4);
        }
    }
    if(mcuPos) {
        auto &bmc = status.bmc;
        bmc.clear();
        bmc.dev_id = getMcuData(regRes, mcuPos, REG_BMC_DEV_ID);
        bmc.onboard_pll_status.raw = getMcuData(regRes, mcuPos, REG_BMC_ONBOARD_PLL_STATUS);
        bmc.mcuVer.raw = getMcuData(regRes, mcuPos, REG_BMC_VERSION);
        bmc.mcuRev = getMcuData(regRes, mcuPos, REG_BMC_REVERSION);
        bmc.overall_status = getMcuData(regRes, mcuPos, REG_BMC_SYSTEM_STATUS) & 0x3;
        bmc.power_status = getMcuData(regRes, mcuPos, REG_BMC_POWER_STATUS) & 0x3;
        bmc.pll_status = getMcuData(regRes, mcuPos, REG_BMC_PLL_STATUS) & 0x3;

        BmcPower &p = bmc.power;

        for(int reg=REG_BMC_VOL_BP_5; reg<=REG_BMC_CUR_BP_33; ++reg) {
            const auto &op = regRes[mcuPos+reg-REG_BMC_BASE];
            if(op.addr>=REG_BMC_VOL_BP_5 && op.addr <= REG_BMC_CUR_BP_33) {
                quint16 shift = op.addr-REG_BMC_VOL_BP_5;
                bool cur = shift%2;
                quint16 id = shift/2;
                auto &map = cur ? p.cur : p.vol;
                map[id] = op.data;
            }
        }

        auto hwVer = getHwVer();
        if(hwVer.first==1 && hwVer.second==0) {
            for(int i=0; i<1; ++i) {
                qint16 t = static_cast<qint16>(getMcuData(regRes, mcuPos, REG_BMC_TEMP_1));
//                status.temp[1+i] = decodeTemp_v1(rawTemp);
                t <<= 2;
                bmc.temp.append(1.*t/128);
            }
        }
    }
    WrEpStatus &wrEpStatus = status.wrEpStatus;
    wrEpStatus.hasWrEpStatus = hasWrEpStatus();
    if (wrEpStatus.hasWrEpStatus) {
        wrEpStatus = readWrEpStatus();
    }

    status.liveMagikOk = checkLiveMagic();
    emit statusUpdated(status);
}

void TtvxsDevice::on_deviceConnected()
{
    QtMregDevice::on_deviceConnected();
    status.isConnected = true;
    status.fwStr = getFirmwareVersionStr();
    // Before this rev. there was no default value for MLink CSR registers

    status.mstreamMultiAck = hasMultiAckSupport();
    status.mstreamHwBufSize = getHwBufSize();

    configValid = false;

    sendResetTimestamp();
    evNumReset();
    writeSetup();
    emit statusUpdated(status);
}

void TtvxsDevice::on_deviceDisconnected()
{
    QtMregDevice::on_deviceDisconnected();
    configValid = false;
    status.isConnected = false;
    emit statusUpdated(status);
}

void TtvxsDevice::sendResetTimestamp(DeviceIndex i)
{
    if(!i.isNull() && i != getDeviceIndex()) return;

    reg_rc_ctrl regVal;
    regVal.all = regRead(REG_RC_CTRL);
    RegOpVector regOp;
    regVal.bit.sendTimeReset = 0;
    regOp << RegWrite(REG_RC_CTRL, regVal.all);
    regVal.bit.sendTimeReset = 1;
    regOp << RegWrite(REG_RC_CTRL, regVal.all);
    regOpExec(regOp);
}

void TtvxsDevice::softClear(DeviceIndex i)
{
    if(!i.isNull() && i != getDeviceIndex()) return;

    RegOpVector regOp;
    reg_base_ctrl regVal;
    regVal.bit.run_csr = configCashe.trig_open;
    regVal.bit.softClear = 1;
    regOp << RegWrite(REG_DEVICE_CTRL, regVal.all);
    regVal.bit.softClear = 0;
    regOp << RegWrite(REG_DEVICE_CTRL, regVal.all);
    regOpExec(regOp);

    configValid = false;
    writeSetup();
}

void TtvxsDevice::clearStat(DeviceIndex i)
{
    // TODO remember up-time;
    if(getDeviceFwVersion() < REV_TTVXS_AUX_CNT) return;
    if(!i.isNull() && i != getDeviceIndex()) return;

    reg_rc_ctrl rcCtrl;
    rcCtrl.all = regRead(REG_RC_CTRL);

    RegOpVector regOp;

    rcCtrl.bit.clear = 1;
    regOp << RegWrite(REG_RC_CTRL, rcCtrl.all);

    rcCtrl.bit.clear = 0;
    regOp << RegWrite(REG_RC_CTRL, rcCtrl.all);

    regOpExec(regOp);
}

void TtvxsDevice::evNumRequest(int check_index, bool top_level)
{
    if (!isConnected())
        return;
    RegOpVector v;
    reg_base_ctrl regVal;
    regVal.bit.run_csr = configCashe.trig_open;
    regVal.bit.cntLock = 1;
    v << RegWrite(REG_DEVICE_CTRL, regVal.all);
    v << RegRead64(REG64_CUR_EVENT_NUMBER);
    regVal.bit.cntLock = 0;
    v << RegWrite(REG_DEVICE_CTRL, regVal.all);

    auto regRes = regOpExec(v);
    if(regRes.size()!=v.size())
        return;

    quint64 evNum = get64(regRes, 1);
    if(getDeviceFwVersion()<=REV_EVNUM_INCREMENT)
        ++evNum;
    if(top_level)
        emit trigEvNumResponse(check_index, evNum);
    else
        emit devEvNumResponse(check_index, getDeviceSerialNumber(), evNum);
}
