#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <QComboBox>
#include <QDir>
#include <QFile>
#include <QHostAddress>
#include <QProcess>
#include <QSettings>
#include <QTextStream>
#include <QTimer>

#include "device-discover/DiscoverDialog.h"

namespace  {
enum {COL_SERIAL, COL_ID, COL_SLOT, COL_START, COL_ATTEMPTS, COL_ONLINE, COL_RATIO, COL_IP_CHANGE, COL_PING, COL_STATUS, COL_TOTAL};
const QString timeFormatMs = "dd/MM/yyyy hh:mm:ss.zzz";
//const QString VME_HOST = "afi-vme-r3-1"; // 225-right
const QString VME_HOST = "afi-vme-r4-1"; // 209
//const QString VME_HOST = "afi-vme-r5-1"; //214-upper
//const QString VME_HOST = "afi-vme-r5-3"; //214-bottom

const QString SNMP_CONTROL_DIR = QDir::homePath() + QDir::separator() +"SnmpControl"+ QDir::separator();

QPair<qint64, qint64> getMinMaxTime(const DevStat &stat, bool &empty) {
    qint64 min=0,max=0;
    for(auto loop : stat.list.values()) {
        if(loop.powerOnTime.isNull() || loop.start.isNull()) continue;
        qint64 dt = loop.powerOnTime.msecsTo(loop.start);
        if(empty) {
            min = dt;
            max = dt;
            empty = false;
        } else {
            min = qMin(min, dt);
            max = qMax(max, dt);
        }
    }
    return QPair<qint64,qint64>(min,max);
}
}

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    discoverDialog(new DiscoverDialog(this)),
    powerTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    crateStatusTimer(new QTimer(this)),
    powerOn(false),
    loopId(0),
    loopGoodCnt(0),
    loopBadCnt(0),
    run(false),
    mainStatus(new QLabel(this)),
    crateStatus(new QLabel(this))
{
    ui->setupUi(this);

    addLog(" ");
    addLog("====================");
    addLog(" ");
    addLog("Program started");

    connect(powerTimer, SIGNAL(timeout()), SLOT(switchPower()));
    powerTimer->setSingleShot(true);
    powerTimer->start(0);

    connect(statusTimer, SIGNAL(timeout()), SLOT(updateStatus()));
    statusTimer->setSingleShot(false);
    statusTimer->start(200);

    connect(crateStatusTimer, SIGNAL(timeout()), SLOT(getCrateState()));
    crateStatusTimer->setSingleShot(false);
    crateStatusTimer->start(100);

    QString winTitle = QCoreApplication::applicationName();
    if(!VME_HOST.isEmpty())
        winTitle += " #"+VME_HOST;
    setWindowTitle(winTitle);

    initTable();
    readCfg();
    ui->tableWidget->sortByColumn(COL_SLOT, Qt::AscendingOrder);
    ui->tableWidget->resizeColumnsToContents();

//    addDev(DeviceIndex(0xd4, 0x080C3B59));
//    addDev(DeviceIndex(0xcc, 0x076CED09));

    connect(discoverDialog, SIGNAL(deviceDiscovered(DeviceDescription)),
            SLOT(deviceUpdated(DeviceDescription)));
    connect(discoverDialog, SIGNAL(deviceExpired(DeviceIndex)),
            SLOT(deviceExpired(DeviceIndex)));
    statusBar()->addPermanentWidget(mainStatus, 10);
    statusBar()->addPermanentWidget(crateStatus, 1);
}

MainWindow::~MainWindow()
{
    addLog("Program closed");
    saveCfg();

    delete ui;
}

void MainWindow::on_pushButtonAdd_clicked()
{
    auto indexList = ui->comboBox->currentData().toList();
    if(indexList.size() != 2) return;
    DeviceIndex index = getComboBoxDevIndex(ui->comboBox->currentIndex());
    if(!index.isNull())
        addDev(index);
    saveCfg();
}

void MainWindow::on_pushButtonRemove_clicked()
{
    QTableWidget *t = ui->tableWidget;
    int row = t->currentRow();
    if(row == -1) return;
    DeviceIndex index = getDevIndex(row);
    if(index.isNull()) return;
    t->removeRow(row);
    devStats.remove(index);
    saveCfg();
}

void MainWindow::on_pushButtonPower_clicked()
{
    switchPower(true);
}

void MainWindow::on_spinBoxTimeout_valueChanged(int)
{
    saveCfg();
}

void MainWindow::on_radioButtonRun_toggled(bool checked)
{
    run = checked;
    if(run && !powerOn)
        switchPower();
}

void MainWindow::on_pushButtonReset_clicked()
{
    if(powerOn)
        switchPower();
    QElapsedTimer timer;
    timer.start();
    bool powerTimerIsActive = powerTimer->isActive();
    if(powerTimerIsActive)
        powerTimer->stop();
    while(!isAllDevInState() && timer.elapsed()<10000) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
    }

    for(auto index: devStats.keys()) {
        devStats[index].clear();
        updateDeviceInfo(index);
    }
    loopGoodCnt = 0;
    loopBadCnt = 0;
    if(powerTimerIsActive)
        powerTimer->start();
    qInfo() << "\nReset was performed!\n";
}

void MainWindow::on_pushButtonInfo_clicked()
{
    qInfo() << "\n\t====Start/stop list:";
    auto devStatList = devStats.values();
    qSort(devStatList);
    for(auto stat : devStatList) {
        qInfo() << QString("%1 attempts:%2 online:%3")
                   .arg(stat.index.getIdent())
                   .arg(stat.attemptsCnt).arg(stat.onlineCnt);
        for(auto loop : stat.list) {
            qInfo().noquote() << loop.toString();
        }
    }

    qInfo() << "\n\t====Ip change:";
    for(auto stat : devStatList) {
        qInfo().noquote() << QString("%1 ipChanges:%2").arg(stat.index.getIdent()).arg(stat.ipList.size());
        for(auto change : stat.ipList)
            qInfo().noquote() << "\t" << change.time.toString() << change.getIpStr();
    }

    qInfo() << "\n\t====Summary:";
    for(auto stat : devStatList) {
        bool empty = true;
        auto minMax = getMinMaxTime(stat, empty);

        qInfo().noquote()
                << QString("| %1| %2| 0x%3| %4| %5| %6| %7| %8| %9|")
                   .arg(stat.slot, 2)
                   .arg(stat.index.getSerialStr())
                   .arg(stat.index.getDevId(),2,16,QChar('0'))
                   .arg(stat.lastDescr.getFirmWareStr(), 12)
                   .arg(stat.ipList.size())
                   .arg(QString("%1-%2").arg(1e-3*minMax.first, 6, 'f', 3).arg(1e-3*minMax.second, 6, 'f', 3))
                   .arg(stat.attemptsCnt, 4).arg(stat.onlineCnt, 4)
                   .arg(stat.attemptsCnt ? stat.onlineCnt*100/stat.attemptsCnt : 0, 3);
    }
}

void MainWindow::on_tableWidget_cellChanged(int row, int column)
{
    if(column != COL_SLOT) return;
    DeviceIndex index = getDevIndex(row);
    if(index.isNull()) return;
    if(!devStats.contains(index)) {
        Q_ASSERT(devStats.contains(index));
        return;
    }
    auto &stat = devStats[index];
    uint slot = ui->tableWidget->item(row, column)->text().toUInt();
    if(stat.slot != slot) {
        stat.slot = slot;
        saveCfg();
        ui->tableWidget->sortByColumn(COL_SLOT, Qt::AscendingOrder);
    }
}

void MainWindow::on_tableWidget_currentCellChanged(int currentRow, int currentColumn,
                                                   int previousRow, int previousColumn)
{
    Q_UNUSED(previousRow);
    Q_UNUSED(previousColumn);
    ui->pushButtonRemove->setEnabled(currentRow != -1 && currentColumn != -1);
}

void MainWindow::deviceUpdated(const DeviceDescription &dd)
{
    DeviceIndex index = dd.getIndex();
//    qInfo() << index.getIdent();
    if(devStats.keys().contains(index)) {

        int row = getDevRow(index);
        Q_ASSERT(row != -1);
        DevStat &stat = devStats[index];
        stat.lastDescr = dd;
        if(dd.ip_addr && (stat.ipList.isEmpty() || dd.ip_addr!=stat.ipList.last().ip)) {
            QDateTime time = QDateTime::currentDateTime();
            stat.ipList << IpChange(dd.ip_addr, time);
            addLog(QString("newIp 0x%1 %2 %3")
                   .arg(index.getSerial(),8,16,QChar('0'))
                   .arg(QHostAddress(dd.ip_addr).toString())
                   .arg(time.toString(timeFormatMs)));
            auto item = ui->tableWidget->item(row, COL_IP_CHANGE);
            item->setText(QString::number(stat.ipList.size()));
            QStringList ttList;
            for(auto change : stat.ipList) {
                ttList << QString("%1 - %2").arg(change.time.toString("hh:mm:ss")).arg(change.getIpStr());
            }
            item->setToolTip(ttList.join('\n'));
        }
        if(!stat.online) {
            QDateTime time = QDateTime::currentDateTime();
            addLog(QString("online 0x%1 %3").arg(index.getSerial(),8,16,QChar('0')).arg(time.toString(timeFormatMs)));
            if(stat.list.contains(loopId)) {
                Q_ASSERT(!stat.list[loopId].powerOnTime.isNull());
                ++stat.onlineCnt;
                stat.list[loopId].start = time;
            }
            stat.online = true;
            ui->tableWidget->item(row, COL_STATUS)->setText("online");
            ui->tableWidget->item(row, COL_STATUS)->setTextColor(Qt::darkGreen);
        }
        if(dd.chassis_slot && dd.chassis_slot != stat.slot) {
            stat.slot = dd.chassis_slot;
            saveCfg();
            ui->tableWidget->setSortingEnabled(false);
            QString slotStr = QString("%1").arg(dd.chassis_slot, 2, 10, QChar('0'));
            if(ui->tableWidget->item(row, COL_SLOT)->text() != slotStr) {
                ui->tableWidget->item(row, COL_SLOT)->setText(slotStr);
                ui->tableWidget->sortByColumn(COL_SLOT, Qt::AscendingOrder);
                ui->tableWidget->resizeColumnToContents(COL_SLOT);
            }
            ui->tableWidget->setSortingEnabled(true);
        }

        if(powerOn && isAllDevInState()) {
            qInfo() << "force stop" << index.getIdent();
            switchPower();
        }
        updateDeviceInfo(index);
    } else {
        QComboBox *cb = ui->comboBox;
        bool missing = true;
        for(int i=0; i < cb->count(); ++i) {
            DeviceIndex cbIndex = getComboBoxDevIndex(i);
            if(cbIndex.isNull()) continue;
            if(cbIndex == index) {
                missing = false;
                break;
            }
        }
        if(missing){
            QList<QVariant> indexVar;
            indexVar << index.getDevId() << index.getSerial();
            cb->addItem(index.getIdent().append(QString(" slot:%1").arg(dd.chassis_slot)), indexVar);
            cb->adjustSize();
            ui->pushButtonAdd->setEnabled(true);
        }
    }
}

void MainWindow::deviceExpired(const DeviceIndex &index)
{
    if(devStats.keys().contains(index)) {
        QDateTime time = QDateTime::currentDateTime();
        const QString format = "dd/MM/yyyy hh:mm:ss.zzz";
        addLog(QString("expire 0x%1 %3").arg(index.getSerial(),8,16,QChar('0')).arg(time.toString(format)));
        DevStat &stat = devStats[index];
        if(stat.online) {
            stat.online = false;
            if(stat.list.contains(loopId))
                stat.list[loopId].end = time;
            int row = getDevRow(index);
            Q_ASSERT(row != -1);
            ui->tableWidget->item(row, COL_STATUS)->setText("offline");
            ui->tableWidget->item(row, COL_STATUS)->setTextColor(Qt::red);
        }
        if(!powerOn && run && isAllDevInState()) {
            qInfo() << "force start";
            switchPower();
        }
    }
}

void MainWindow::initTable()
{
    QTableWidget * const table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);
    table->setHorizontalHeaderItem(COL_ID, new QTableWidgetItem("id"));
    table->setHorizontalHeaderItem(COL_SERIAL, new QTableWidgetItem("serial"));
    table->setHorizontalHeaderItem(COL_SLOT, new QTableWidgetItem("Slot"));
    table->setHorizontalHeaderItem(COL_PING, new QTableWidgetItem("Ping"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("status"));
    table->setHorizontalHeaderItem(COL_ATTEMPTS, new QTableWidgetItem("attempts"));
    table->setHorizontalHeaderItem(COL_START, new QTableWidgetItem("startTime"));
    table->setHorizontalHeaderItem(COL_ONLINE, new QTableWidgetItem("onlines"));
    table->setHorizontalHeaderItem(COL_IP_CHANGE, new QTableWidgetItem("IpChange"));
    table->setHorizontalHeaderItem(COL_RATIO, new QTableWidgetItem("Ratio"));
}

void MainWindow::readCfg()
{
    QSettings s;
    s.beginGroup("Main");
    ui->spinBoxTimeout->blockSignals(true);
    ui->spinBoxTimeout->setValue(s.value("timeout", 40).toInt());
    ui->spinBoxTimeout->blockSignals(false);
    s.endGroup();

    int cnt = s.beginReadArray("list");
    bool ok1, ok2;
    for(int i=0; i<cnt; ++i) {
        s.setArrayIndex(i);
        quint16 id = static_cast<quint16>(s.value("id").toString().toUInt(&ok1, 16));
        quint64 ser = s.value("ser").toString().toULongLong(&ok2, 16);
        DeviceIndex index(id,ser);
        if(!ok1 || !ok2 || index.isNull()) continue;
        uint slot = s.value("slot", 0).toUInt();
        addDev(index, slot);
    }
    s.endArray();

    s.beginGroup("MainWindow");
    restoreGeometry(s.value("geometry").toByteArray());
    restoreState(s.value("windowState").toByteArray());
    s.endGroup();
}

void MainWindow::saveCfg()
{
    QSettings s;
    s.beginGroup("Main");
    s.setValue("timeout", ui->spinBoxTimeout->value());
    s.endGroup();
    s.remove("list");
    s.beginWriteArray("list");
    int i=0;
    for(auto stat : devStats.values()) {
        s.setArrayIndex(i++);
        s.setValue("id", QString("0x%1").arg(stat.index.getDevId(),0,16));
        s.setValue("ser", QString("0x%1").arg(stat.index.getSerial(),0,16));
        s.setValue("slot", stat.slot);
    }
    s.endArray();

    s.beginGroup("MainWindow");
    s.setValue("geometry", saveGeometry());
    s.setValue("windowState", saveState());
//    s.setValue("splitterState", ui->splitter->saveState());
//    s.setValue("tabIndex", ui->tabWidget->currentIndex());
    s.endGroup();

}

void MainWindow::addDev(const DeviceIndex &index, uint slot)
{
    QTableWidget *t = ui->tableWidget;
    int row = t->rowCount();
    t->setSortingEnabled(false);
    t->blockSignals(true);
    t->setRowCount(row+1);
    for(int col=0; col<COL_TOTAL; ++col)
        t->setItem(row, col, new QTableWidgetItem());

    t->item(row, COL_ID)->setText(QString::number(index.getDevId(), 16));
    t->item(row, COL_SERIAL)->setText(index.getSerialStr());

    t->item(row, COL_ID)->setData(Qt::UserRole, index.getDevId());
    t->item(row, COL_SERIAL)->setData(Qt::UserRole, index.getSerial());

    t->item(row,COL_PING)->setTextAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
    t->item(row,COL_PING)->setTextColor(Qt::darkYellow);

    devStats[index] = DevStat(index);
    if(slot) {
        QString slotStr = QString("%1").arg(slot, 2, 10, QChar('0'));
        t->item(row, COL_SLOT)->setText(slotStr);
        devStats[index].slot = slot;
    }
    t->blockSignals(false);
    t->resizeColumnsToContents();

    for(int i=0; i<ui->comboBox->count(); ++i) {
        if(getComboBoxDevIndex(i) == index) {
            ui->comboBox->removeItem(i);
            break;
        }
    }
    t->setSortingEnabled(true);
}

DeviceIndex MainWindow::getComboBoxDevIndex(int index) const
{
    auto indexVar = ui->comboBox->itemData(index).toList();
    if(indexVar.size()!=2) return DeviceIndex();

    return DeviceIndex(static_cast<quint16>(indexVar.first().toUInt()),
                       indexVar.last().toULongLong());
}

DeviceIndex MainWindow::getDevIndex(int row) const
{
    DeviceIndex index;
    QTableWidget *t = ui->tableWidget;
    if(row < t->rowCount() && row != -1) {
        index = DeviceIndex(static_cast<quint16>(t->item(row, COL_ID)->data(Qt::UserRole).toUInt()),
                      t->item(row, COL_SERIAL)->data(Qt::UserRole).toULongLong());
    }
    return  index;
}

int MainWindow::getDevRow(const DeviceIndex &index) const
{
    QTableWidget *t = ui->tableWidget;
    int rowCnt = t->rowCount();
    for(int row=0; row<rowCnt; ++row){
        if(index == getDevIndex(row)) return row;
    }
    return -1;
}

bool MainWindow::isAllDevInState()
{
    for(auto stat : devStats.values()) {
        if(stat.list.contains(loopId) && stat.online != powerOn)
            return false;
    }
    return true;
}

void MainWindow::updateDeviceInfo(const DeviceIndex &index)
{
    if(!devStats.contains(index)) return;
    int row = getDevRow(index);
    if(row == -1) return;
    QTableWidgetItem *item = ui->tableWidget->item(row, COL_START);
    auto stat = devStats[index];
    bool empty = true;
    auto minMax = getMinMaxTime(stat, empty);
    item->setText(empty ? "-" :
                          QString("%1-%2")
                          .arg(1e-3*minMax.first, 0, 'f', 3)
                          .arg(1e-3*minMax.second, 0, 'f', 3));
    ui->tableWidget->item(row, COL_ATTEMPTS)->setText(QString::number(stat.attemptsCnt));
    ui->tableWidget->item(row, COL_ONLINE)->setText(QString::number(stat.onlineCnt));
    auto ratioStr = !stat.attemptsCnt ? QString("-") : QString::number(stat.onlineCnt*100/stat.attemptsCnt);
    ui->tableWidget->item(row, COL_RATIO)->setText(ratioStr);
    QColor col;
    if(stat.attemptsCnt) {
        if(stat.onlineCnt==stat.attemptsCnt)
            col = Qt::darkGreen;
        else if(1.*stat.onlineCnt/stat.attemptsCnt < 0.9)
            col = Qt::red;
        else
            col = Qt::darkYellow;
    } else
        col = Qt::darkGray;
    ui->tableWidget->item(row, COL_RATIO)->setTextColor(col);
    ui->tableWidget->resizeColumnToContents(COL_START);
}

void MainWindow::updateStatusBar()
{
    QString str = QString("loopId:%1 (%2/%3) %4")
            .arg(loopId)
            .arg(loopGoodCnt).arg(loopGoodCnt+loopBadCnt)
            .arg(powerOn ? "starting" : "stopping");
    if(switchTime.isValid())
        str += QString(" %1sec...").arg(switchTime.elapsed()/1000);
    mainStatus->setText(str);
}

void MainWindow::addLog(QString log)
{
    QFile file;
    file.setFileName("transceiver_test.log");
    if(!file.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
        qCritical() << "Failed to open file";
        return;
    }
    QTextStream textStream(&file);
    textStream << log << "\n";
}

bool MainWindow::switchPowerScript(bool on)
{
    qInfo() << "switchPower:" << (on?"on":"off");
    QProcess p;
    p.start(SNMP_CONTROL_DIR+"on-off.sh", QStringList() << VME_HOST << (on ? "1" :"0"));
    if(!p.waitForFinished(500))
        qFatal("Can't wait end of execution SNMP script (switchPower)");
    if(p.exitCode() != 0)
        qFatal("Return code (exitCode=%d) is wrong for SNMP script (switchPower",
               p.exitCode());
    QString outputStr = p.readAllStandardOutput();
    auto lines = outputStr.split('\n');
    for(auto line : lines) {
        if(line.contains("WIENER-CRATE-MIB::sysMainSwitch.0")) {
            if(line.contains("on(1)")) {
                return true;
            }
            if(line.contains("off(0)")) {
                return false;
            }
        }
    }
    qFatal("Can't parse output of SNMP script (crate status)");
}

bool MainWindow::getCrateState()
{
    QProcess p;
    p.start(SNMP_CONTROL_DIR+"getStatus.sh", QStringList() << VME_HOST);
    if(!p.waitForFinished(100)) {
        qCritical() << p.readAllStandardError();
        qCritical() << p.readAllStandardOutput();
        qCritical() << p.state();
        qFatal("Can't wait end of execution SNMP script (getStatus)");
    }
    if(p.exitCode() != 0) {
        qFatal("Return code (exitCode=%d) is wrong for SNMP script (getStatus)",
               p.exitCode());
    }
    QString outputStr = p.readAllStandardOutput();
    auto lines = outputStr.split('\n');
    for(auto line : lines) {
        if(line.contains("WIENER-CRATE-MIB::sysMainSwitch.0")) {
            if(line.contains("on(1)")) {
                crateStatus->setText("power: ON");
                crateStatus->setStyleSheet("background-color:green;");
                return true;
            }
            if(line.contains("off(0)")) {
                crateStatus->setText("power: OFF");
                crateStatus->setStyleSheet("background-color:red;");
                return false;
            }
        }
    }
    qFatal("Can't parse output of SNMP script (crate status)");
}

void MainWindow::switchPower(bool force)
{
    if(powerOn) {
        bool hasPrevLoop = true; // all devices contains prev LoopInfo
        bool prevLoopIsGood = true;
        powerOn = false;
        updateStatusBar();
        QDateTime time = QDateTime::currentDateTime();
        const QString format = "dd/MM/yyyy hh:mm:ss.zzz";
        addLog(QString("endLoop %1 %2").arg(loopId, 5).arg(time.toString(format)));

        switchPowerScript(powerOn);
        switchTime.start();
        QDateTime endTime = QDateTime::currentDateTime();
        for(auto index : devStats.keys()) {
            auto &stat = devStats[index];
            if(stat.list.contains(loopId)) {
                const auto &loopInfo = stat.list[loopId];
                if(loopInfo.start.isNull()) {
                    qWarning() << index.getIdent() << "didn't start";
                    prevLoopIsGood = false;
                }
            } else {
                hasPrevLoop = false;
            }
            stat.list[loopId].powerOffTime = endTime;
        }
        if(hasPrevLoop) {
            if(prevLoopIsGood) {
                ++loopGoodCnt;

            } else {
                ++loopBadCnt;
            }
        }
        qInfo() << "End"<<(hasPrevLoop ? (prevLoopIsGood?"good":"bad") : "") <<"loop:" << loopId;
        qInfo() << "";
        if(run)
            powerTimer->start(10000); // plan next loop start-up
    } else if(run || force) {
        ++loopId;
        powerOn = true;
        updateStatusBar();
        QDateTime time = QDateTime::currentDateTime();
        const QString format = "dd/MM/yyyy hh:mm:ss.zzz";
        addLog(QString("startLoop %1 %2").arg(loopId, 5).arg(time.toString(format)));

        qInfo() << "Start loop:" << loopId;
        switchPowerScript(powerOn);
        switchTime.start();

        for(auto index : devStats.keys()) {
            auto &stat = devStats[index];
            if(stat.list.contains(loopId-1)) {
                const auto &loopInfo = stat.list[loopId-1];
                if(!loopInfo.start.isNull() && loopInfo.end.isNull()) {
                    qWarning() << index.getIdent() << "didn't stop on previous loop";
                }
            }
            if(!stat.list.contains(loopId)) {
                stat.list[loopId] = LoopInfo(loopId);
                ++stat.attemptsCnt;
            }
            updateDeviceInfo(index);
        }
        updateStatusBar();
        powerTimer->start(ui->spinBoxTimeout->value()*1000); // plan this loop shutdown
    }
}

void MainWindow::updateStatus()
{
    updateStatusBar();
    bool hasChanges = false;
    qint64 dt = pingTime.restart();
    auto table =ui->tableWidget;
    for(auto stat : devStats.values()) {
        int row = getDevRow(stat.index);
        auto item = table->item(row, COL_PING);
        if(row<0) continue;

        bool isPingActive = stat.lastDescr.discoverTimer.msecsTo(pingTime) <= dt;
        if(isPingActive == item->text().isEmpty()){
            if(!hasChanges) {
                hasChanges = true;
                ui->tableWidget->setUpdatesEnabled(false);
            }
            item->setText(isPingActive ? QString(QChar(0x2600 + (qrand()&0xFF))) : QString());
        }
    }

    if(hasChanges)
        ui->tableWidget->setUpdatesEnabled(true);
}
