/*
**    Reverse DNS lookup (get PTR record)
**
**    Copyright 2016-2020 Ivan Filippov
**    Copyright 2016-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DnsLookup.h"

#include <QHostInfo>
#include <QThread>

DnsLookup::DnsLookup(QObject *parent)
    : QObject(parent),
      mainThreadId(thread()->currentThreadId())
{
    connect(this, &DnsLookup::reverseLookupComplete,
            this, &DnsLookup::gotHostInfo,
            Qt::QueuedConnection);
}

void DnsLookup::reverseLookup(const QString &name)
{
    QHostInfo::lookupHost(name, this, &DnsLookup::reverseLookupCallback);
}

void DnsLookup::reverseLookupCallback(const QHostInfo &h) // called from other thread
{
    emit reverseLookupComplete(h);
}

void DnsLookup::gotHostInfo(QHostInfo h)
{
    Q_ASSERT(mainThreadId == thread()->currentThreadId());
    Q_ASSERT(h.error() == QHostInfo::NoError);
    if(h.error() != QHostInfo::NoError)
        return;
    QString hostName = h.hostName();
    Q_ASSERT(! hostName.isEmpty());
    if (hostName.isEmpty())
        return;
    bool is_ip_addr = ! QHostAddress(hostName).isNull();
    if (is_ip_addr) // skip: asked PTR, got IP
        return;

    const QString &tail = "." + QHostInfo::localDomainName();
    bool isChanged = false;
    DnsReverseEntry ent;
    for (const auto &addr : h.addresses()) {
        hostName.remove(tail);
        bool ok = false;
        const auto ip = addr.toIPv4Address(&ok);
        Q_ASSERT(ok);
        auto it = hostMap.find(ip);
        if (it == hostMap.end()) {
            // new entry
            isChanged = true;
            ent.first = ip;
            ent.second = hostName;
            hostMap.insert(ip, hostName);
            continue;
        } else {
            // update
            if (it.value() != hostName) {
                isChanged = true;
                ent.first = ip;
                ent.second = hostName;
                *it = hostName;
            }
        }
    }
    if (isChanged) {
        emit reverseEntryUpdated(ent);
        emit reverseMapUpdated(hostMap);
    }
}
