#include "DeviceDescription.h"

DeviceDescription::DeviceDescription(DeviceIndex devIndex)
{
    device_id=devIndex.getDevId();
    serial_id=devIndex.getSerial();
    clear();
}


DeviceDescription::DeviceDescription(quint16 devId, quint64 devSerial, quint32 ip)
{
    device_id=devId;
    serial_id=devSerial;

    clear();
    ip_addr=ip;
}

void DeviceDescription::clear()
{
    if(!isNull())
        discoverTimer.start();

    haddr=0;
    ip_addr=0;
    ttl=0;

    chassis_slot = 0;
    master_mac=0;
    master_ip=0;
    master_udp_port=0;
    mstream_mac=0;
    mstream_ip=0;
    mstream_udp_port=0;
    mstream_locked=0; master_locked=0;

    enabled=true;
}

bool DeviceDescription::isOutOfDate() const
{
    return !discoverTimer.isValid() || discoverTimer.elapsed() > (int)ttl*1000;
}

QString DeviceDescription::getIdent() const
{
    return QString("%1 %2")
            .arg(modelname_str.trimmed())
            .arg(getSerialIdStr());
}

QString DeviceDescription::getSerialIdStr() const
{
    QString s = sn_str.trimmed();
    s = s.replace("0000-", "");
    if (s.isEmpty()) {
        quint16 s3 = (serial_id >> 32) & 0xFFFF;
        quint16 s2 = (serial_id >> 16) & 0xFFFF;
        quint16 s1 = serial_id & 0xFFFF;
        QString str = QString("%1-%2")
                .arg(s2, 4, 16, QChar('0'))
                .arg(s1, 4, 16, QChar('0'));
        if (s3)
            str = QString("%1-%2")
                .arg(s3, 4, 16, QChar('0'))
                    .arg(str);
        return str.toUpper();
    }
    return s;
}

QString DeviceDescription::getModelStr() const
{
    QString s = modelname_str.trimmed();
    if(s.isEmpty())
        s = QString("0x%1").arg(device_id, 2, 16, QChar('0'));
    return s;
}

QString DeviceDescription::getModelVerStr() const
{
    QString s = getModelStr();
    if (s.isEmpty())
        return s;
    if (!hw_str.isEmpty())
        s += "-" + hw_str;
    return s;
}

QString DeviceDescription::getFirmWareStr() const
{
    return fw_str.trimmed();
}

int DeviceDescription::getHwMajVer() const
{
    return DeviceDescription::getHwMajVer(hw_str);
}

int DeviceDescription::getHwMajVer(const QString &str)
{
    if(str.isEmpty()) return 0;
    bool ok;
    int hwMajVer = str.left(2).toInt(&ok);
    return ok ? hwMajVer : 0;
}

QPair<int, int> DeviceDescription::getHwVer() const
{
    return getHwVer(hw_str);
}

QPair<int,int> DeviceDescription::getHwVer(const QString &str)
{
    QPair<int,int> ver(0,0);
    QStringList list = str.split('.');
    if(list.size()==2) {
        int a,b;
        bool ok1,ok2;
        a = list.first().toInt(&ok1);
        b = list.last().toInt(&ok2);
        if(ok1 && ok2) {
            ver.first = a;
            ver.second = b;
        }
    }
    return ver;
}

bool DeviceDescription::deviceInfoAreSame(const DeviceDescription &dd) const
{
    // All exept enabled & discoverTime
    return true &&
            haddr == dd.haddr &&
            ip_addr == dd.ip_addr &&
            ttl == dd.ttl &&
            portID == dd.portID &&
            hw_str == dd.hw_str &&
            fw_str == dd.fw_str &&
            sn_str == dd.sn_str &&
            mn_str == dd.mn_str &&
            modelname_str == dd.modelname_str &&
            device_id == dd.device_id &&
            serial_id == dd.serial_id &&
            chassis_slot == dd.chassis_slot &&
            master_mac == dd.master_mac &&
            master_ip == dd.master_ip &&
            master_udp_port == dd.master_udp_port &&
            mstream_mac == dd.mstream_mac &&
            mstream_ip == dd.mstream_ip &&
            mstream_udp_port == dd.mstream_udp_port &&
            mstream_locked == dd.mstream_locked &&
            master_locked == dd.master_locked;
}

bool DeviceDescription::operator ==(const DeviceDescription &dd) const
{
    return deviceInfoAreSame(dd) &&
            enabled == dd.enabled &&
            discoverTime == dd.discoverTime &&
            discoverTimer == dd.discoverTimer;
}
