#include "FileReader.h"
#include <QCoreApplication>
#include <QDebug>
#include <QDir>
#include <QTime>
#include <QTimer>
#include <QSemaphore>

#include "lib-common/MpdRawTypes.h"

namespace {
const qint64 PROGRESS_INCREMENT_SIZE = 1e6; // in Bytes
}

FileReader::FileReader(QSemaphore *sem):
    freeRawDataSem(sem),
    readFileTimer(new QTimer(this)),
    doCloseAll(false),
    pack_size_limit(20e6)
{
    connect(readFileTimer, &QTimer::timeout, this, &FileReader::readFile);
    readFileTimer->setSingleShot(false);
    readFileTimer->setInterval(1000);
//    readFileTimer->start();
}

void FileReader::readFile()
{
    if(!dataFile.exists() || !dataFile.isReadable())
        return;
    while(!dataFile.atEnd()) {
        qint64 sizeToRead = qMin(dataFile.size()-dataFile.pos(), pack_size_limit);
        if(sizeToRead==0) {
            break;
        }

        if(!freeRawDataSem->tryAcquire()){
            qInfo() << QString("[%1] file read delayed (analyzer is slow)")
                     .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
            break;
        }

        emit dataAccuired(dataFile.read(sizeToRead));
    }
    if(dataFile.atEnd())
        qInfo() << "File was read.";
    else
        readFileTimer->start(10);
}

void FileReader::readRecord(int recordNum, int count)
{
    if(!dataFile.isOpen()) return;
    if(recordNum > records.size() || recordNum==0) {
        qWarning() << "Record number is out of range";
        return;
    }
    const FileReaderRecord &first_record = records[recordNum-1];
    const quint32 &evNum = first_record.evNum;
    quint32 evNumLast;
    const qint64 &pos = first_record.pos;
    if(dataFile.pos() != pos)
        dataFile.seek(pos);


    qint64 fullSize;
    if(count==1) {
        fullSize = first_record.length;
        evNumLast = evNum;
    } else {
        const FileReaderRecord *last_record;
        int last_record_index = recordNum-1+count-1;
        if(records.size()>(last_record_index)) {
            last_record = &records[last_record_index];
        } else {
            last_record = &records[records.size()-1];
        }
        evNumLast = last_record->evNum;
        const qint64 &end_pos = last_record->pos+last_record->length;
        fullSize = end_pos-pos;
    }
    if(pos+fullSize>dataFile.size()) {
        qFatal("[FileReader] Incomplete event: %d", evNum);
    }
    QByteArray evData = dataFile.read(fullSize);

    if(!freeRawDataSem->tryAcquire(1, 1000)) {
        qInfo() << QString("[%1] file read delayed (analyzer is slow)")
                 .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
        freeRawDataSem->acquire();
    }

    emit eventRead(evNum, evNumLast, fullSize);
    emit dataAccuired(evData);
}

void FileReader::closeFile()
{
    if(dataFile.isOpen())
        dataFile.close();
    records.clear();
    emit fileMapUpdated(0, "", records.size(), 1);
}

void FileReader::closeAll()
{
    doCloseAll = true;
    deleteLater();
}

void FileReader::setPackSizeLimit(qint64 limit)
{
    pack_size_limit = limit;
}

void FileReader::updateFileMap()
{
    qInfo() << "Start creating file map:";
    records.clear();
    const qint64 fileSize = dataFile.size();
    dataFile.seek(0);
    qint64 pos = dataFile.pos();
    qint64 progressPos = pos;
    qint64 read_size;
    qint64 record_pos = pos;
    qint64 record_size = 0;
    MpdEventHdr evHdr;
    bool ok=true;
    QTime progressTime;
    progressTime.start();
    while(pos+static_cast<qint64>(sizeof(MpdEventHdr)) < fileSize) {
        if(pos>=progressPos) {
            QCoreApplication::processEvents(QEventLoop::AllEvents, 100); // to read event befor end of mapping
            if(doCloseAll) {
                qInfo() << "Creation of file map was canceled (closeAll signal).";
                return;
            }
            if(!dataFile.isOpen()) {
                qWarning() << "Creation of file map was canceled (file was closed).";
                return;
            }
            if(progressTime.elapsed() > 3000) {
                qInfo() << QString("Completed %1% %2 ms").arg(100.*pos/fileSize, 0, 'f', 1).arg(progressTime.restart());
                emit fileMapUpdated(fileSize, QFileInfo(dataFile).lastModified().toString(), records.size(), 1.*pos/fileSize);
            }
            progressPos += PROGRESS_INCREMENT_SIZE;
        }
        dataFile.seek(pos);
        read_size = dataFile.read(reinterpret_cast<char*>(&evHdr), sizeof(MpdEventHdr));
        if(read_size == sizeof(MpdEventHdr)) {
            switch (evHdr.sync) {
            case MPD_SYNC_NORM:
            case MPD_SYNC_EOB:
            case MPD_SYNC_RUN_START:
            case MPD_SYNC_RUN_STOP:
                if(evHdr.sync == MPD_SYNC_NORM || evHdr.sync == MPD_SYNC_EOB) {
                    record_size += sizeof(MpdEventHdr);
                } else {
                    record_size += sizeof(MpdHdr);
                    evHdr.evNum = 0;
                }
                record_size += evHdr.length;

                if((pos+record_size) <= fileSize) {
                    records.push_back(FileReaderRecord(record_pos, record_size, evHdr.sync, evHdr.evNum));
                    pos += record_size;
                    record_pos = pos;
                    record_size = 0;
                } else {
                    ok=false;
                    qWarning() << QString("File ended with incompleted event. Mpd header sync-word: 0x%1; len: ")
                                  .arg(evHdr.sync, 0, 16).arg(evHdr.length);
                    break;
                }
                break;
            default:
                ok=false;
                qWarning() << "Unknown Mpd header sync-word: 0x" << hex << evHdr.sync;
                break;
            }
        } else {
            ok=false;
            qWarning() << "Can't read MpdEventHdr in FileReader::updateFileMap(): "<< dataFile.errorString();
            break;
        }

        if(!ok)
            break;
    }

    qInfo() << QString("updateFileMap completed: parsed %1/%2 bytes; %3 records")
               .arg(pos).arg(fileSize).arg(records.size());
    dataFile.seek(0);

    emit fileMapUpdated(fileSize, QFileInfo(dataFile).lastModified().toString(), records.size(), 1.*pos/fileSize);
}

void FileReader::setFileName(QString fileName)
{
    closeFile();
    dataFile.setFileName(fileName);
    if(dataFile.open(QIODevice::ReadOnly))
        qInfo() << QString("Source file '%1' was opened").arg(dataFile.fileName());
    else
        qWarning() << QString("Failed to open file '%1': %2")
                      .arg(dataFile.fileName(), dataFile.errorString());
    updateFileMap();
}
