#include "HistMngrWidget.h"
#include "ui_HistMngrWidget.h"

#include <cmath>

#include "qwt_plot.h"
#include "qwt_interval_data.h"
#include "qwt_plot_curve.h"
#include <qwt_scale_engine.h>

#include <QTimer>
#include <QSettings>
#include <QSemaphore>

#include "DeviceData.h"
#include "DeviceHelper.h"

namespace  {
const int TDC_MAX_LAT_NS = 100;
const int SHRINK_MAX_SIZE = 1000;

void shrinkIntervals(QwtArray<QwtDoubleInterval> &intervals, QwtArray<double> &y_values){
    const int oldSize = intervals.size();
    Q_ASSERT(y_values.size()==oldSize);

    int group_size = oldSize/SHRINK_MAX_SIZE;
    if(group_size<=1)
        return;

    if(oldSize%SHRINK_MAX_SIZE)
        ++group_size;
    int newSize = oldSize/group_size;
    if(oldSize%group_size)
        ++newSize;
    QwtArray<QwtDoubleInterval> xVals(newSize);
    QwtArray<double> yVals(newSize);
    QwtDoubleInterval curInterval;
    double curVal=0;
    int step=0;
    int newBin=0;
    for(int oldBin=0; oldBin<oldSize; ++oldBin) {
        if(step==0) {
            curInterval = intervals[oldBin];
            curVal = y_values[oldBin];
        } else {
            curVal += y_values[oldBin];
        }
        if(++step == group_size){
            curInterval.setMaxValue(intervals[oldBin].maxValue());
            xVals[newBin] = curInterval;
            yVals[newBin] = curVal;
            ++newBin;
            curVal=0;
            step=0;
        }
    }
    if(oldSize%group_size) {
        curInterval.setMaxValue(intervals.last().maxValue());
        xVals[newBin] = curInterval;
        yVals[newBin] = curVal;
    }
//#ifdef QT_DEBUG
//    //To Check if logic is correct
//    double sum1=0;
//    double sum2=0;
//    for(auto v : y_values)
//        sum1 += v;
//    for(auto v : yVals)
//        sum2 += v;
//    Q_ASSERT(qFuzzyCompare(sum1, sum2));
//    Q_ASSERT(qFuzzyCompare(xVals.first().minValue(), intervals.first().minValue()));
//    Q_ASSERT(qFuzzyCompare(xVals.last().maxValue(), intervals.last().maxValue()));
//#endif
    intervals = xVals;
    y_values = yVals;
}
}

HistMngrWidget::HistMngrWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::HistMngrWidget),
    refreshTimer(new QTimer(this)),
    hide_zero_bin(false)
{
    connect(refreshTimer, &QTimer::timeout, this, &HistMngrWidget::refreshCurrentPlotC);
    refreshTimer->setInterval(500);
    ui->setupUi(this);
}

HistMngrWidget::~HistMngrWidget()
{
    delete ui;
}


void HistMngrWidget::refreshCurrentPlotB(const HistContainer &hc, bool force)
{
    histContainer = hc;
    refreshCurrentPlotA(force);
}

void HistMngrWidget::refreshCurrentPlotA(bool force)
{
    if(!force && lastRefreshTime.isValid() && lastRefreshTime.elapsed()<500) {
        return;
    }
    Histogramm *curHist = getCurHist();
    auto histCfg = getCurHistCfg();
    if(!curHist)
        return;
    Q_ASSERT(curHist);

    const int nBins = curHist->nBins;

    QwtArray<QwtDoubleInterval> intervals(nBins);
    QwtArray<double> y_values(curHist->bins);

    QwtArray<double> x_values(nBins);
    int leftEdge=-1,rightEdge=-1;
    qint64 sum = 0;
    int binMin=0, binMax=0;
    bool isInlCurve = histCfg.groupId == HistId::inl;

    for (int bin = 0; bin < nBins; ++bin)
    {
        if(isInlCurve){
            x_values[bin] = curHist->getBinCenter(bin);
        } else {
            auto &i = intervals[bin];
            i.setMinValue(curHist->getBinLower(bin));
            i.setMaxValue(curHist->getBinUpper(bin));
        }
        if(cfg.curHistId == HistId::tdcHitEv && hide_zero_bin && qFuzzyIsNull(curHist->getBinLower(bin)))
            y_values[bin]=0;

        int gbin = static_cast<int>(y_values[bin]);
        if(sum) {
            binMin = qMin(binMin, gbin);
            binMax = qMax(binMax, gbin);
        } else {
            binMin = gbin;
            binMax = gbin;
        }
        sum += gbin;
        if(gbin) {
            if(leftEdge==-1)
                leftEdge=bin;
            rightEdge=bin;
        }
    }
    if(leftEdge != -1){
        intervals = intervals.mid(leftEdge, rightEdge-leftEdge+1);
        y_values = y_values.mid(leftEdge, rightEdge-leftEdge+1);
    }
    if(isInlCurve) {
        if(y_values.size()==1024 && !cfg.inlHistCodeDens)
            y_values = makeDNLCorrectionTable(y_values);

        QwtPlotCurve *inlCurv = inlCurvs[getCurDev()][getCurCh()];
        Q_ASSERT(inlCurv);
        inlCurv->setData(x_values, y_values);
        inlCurv->itemChanged();
    } else {
        shrinkIntervals(intervals, y_values);

        HistogramItem *qwtHist = histCfg.isBaseHist ? histQwtItems[cfg.curHistId] : devChHists[getCurDev()][getCurCh()];
        Q_ASSERT(qwtHist);
        qwtHist->setData(QwtIntervalData(intervals, y_values));
        updateFit(curHist);
        qwtHist->itemChanged();
    }
    ui->qwtPlotHists->setAxisAutoScale(QwtPlot::xBottom);

    QwtScaleEngine *scale_engine = ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom);
    scale_engine->setAttribute(QwtScaleEngine::Floating, true);
//    ui->qwtPlotHists->setAxisScale(QwtPlot::xBottom, intervals.first().minValue(), intervals.last().maxValue());
    ui->qwtPlotHists->replot();

    QStringList histInfo;
    if(!isInlCurve) {
        histInfo << QString("under:%1; over:%2")
                    .arg(curHist->getUnderFlow())
                    .arg(curHist->getOverFlow());
        if(!intervals.isEmpty())
            histInfo << QString("range: [%1; %2]")
                        .arg(intervals.first().minValue())
                        .arg(intervals.last().maxValue());
    } else {
        histInfo << QString("min/max: %1-%2").arg(binMin).arg(binMax);
    }
    histInfo << QString("count: %1").arg(curHist->getEntries());
    lastRefreshTime.start();

    ui->labelHistInfo->setText(histInfo.join('\n'));
}

void HistMngrWidget::on_comboBoxGroup_currentIndexChanged(int)
{
    onCurHistChanged();
}

void HistMngrWidget::on_checkBoxHistGroupEn_toggled(bool checked)
{
    HistCfg& histGroupCfg = getCurHistCfg();
    histGroupCfg.histGroupEn = checked;
    ui->comboBoxGroup->setItemText(ui->comboBoxGroup->currentIndex(),
                                   histGroupCfg.getGroupTitle());
    cfg.save();
    emit hist_cfg_updated(cfg);
}

void HistMngrWidget::on_comboBoxDev_currentIndexChanged(int)
{
    updateSpinBoxMaxCh(false);
    onCurHistChanged();
}

void HistMngrWidget::on_spinBoxCh_valueChanged(int)
{
    onCurHistChanged();
}

void HistMngrWidget::on_groupBoxStopCh_toggled(bool val)
{
    cfg.useStopCh = val;
    cfg.save();
    emit hist_cfg_updated(cfg);
}

void HistMngrWidget::on_comboBoxStopDev_currentIndexChanged(int)
{
    updateSpinBoxMaxCh(true);
    cfg.stopDevIndex = getCurDev(true);
    cfg.save();
    emit hist_cfg_updated(cfg);
}

void HistMngrWidget::on_spinBoxStopCh_valueChanged(int ch)
{
    cfg.stopCh = static_cast<quint16>(ch);
    cfg.save();
    emit hist_cfg_updated(cfg);
}

void HistMngrWidget::on_spinBoxMwMin_valueChanged(int min)
{
    cfg.matchWinUpdated(min, cfg.mwMax);
    cfg.save();
    emit hist_cfg_updated(cfg);
    ui->spinBoxMwMax->setMinimum(min);

    auto hist = getHistById(HistId::tdcTime);
    if(hist) {
        auto hCfg = cfg.histCfg[HistId::tdcTime];
        hist->resize(hCfg.nBins, hCfg.min, hCfg.max);
    }
    if(cfg.curHistId == HistId::tdcTime)
        refreshCurrentPlotA(true);
}

void HistMngrWidget::on_spinBoxMwMax_valueChanged(int max)
{
    cfg.matchWinUpdated(cfg.mwMin, max);
    cfg.save();
    emit hist_cfg_updated(cfg);
    ui->spinBoxMwMin->setMaximum(max);

    auto hist = getHistById(HistId::tdcTime);
    if(hist) {
        auto hCfg = cfg.histCfg[HistId::tdcTime];
        hist->resize(hCfg.nBins, hCfg.min, hCfg.max);
    }
    if(cfg.curHistId == HistId::tdcTime)
        refreshCurrentPlotA(true);
}

void HistMngrWidget::on_radioButtonInl_toggled(bool checked)
{
    cfg.inlHistCodeDens = !checked;
    cfg.save();
    if(getCurHistCfg().groupId == HistId::inl) {
        refreshCurrentPlotA(true);
    }
}

void HistMngrWidget::on_pushButtonClear_clicked()
{
    HistCfg hCfg = getCurHistCfg();
    if(hCfg.isBaseHist) {
        if(cfg.curHistId == HistId::tdcTaiTime)
            emit recreateTaiHist();
        else
            emit clearHist(cfg.curHistId);
    } else {
        emit clearHist(cfg.curHistId, getCurDev(), getCurCh());
    }
}

void HistMngrWidget::on_pushButtonClearAll_clicked()
{
    emit clearHist();
}

void HistMngrWidget::on_checkBoxGlobalEnable_toggled(bool checked)
{
    cfg.globalEnabled = checked;
    cfg.save();
    emit hist_cfg_updated(cfg);
    if(cfg.globalEnabled)
        refreshTimer->start();
    else {
        refreshTimer->stop();
        refreshCurrentPlotA(true);
    }
}

void HistMngrWidget::on_groupBoxFit_toggled(bool checked)
{
/*
    if(checked) {
        Hist *curHist = getCurHist();
        int bin;
        qInfo() << "bin[-1.5]="<<(bin=curHist->findBin(-1.5))
                << "bin[   0]="<<curHist->findBin(0)
                << "bin[ 1.5]="<<curHist->findBin(1.5)
                << "binStep:"<<curHist->binStep << "nBins:" << curHist->nBins;
        qInfo() << curHist->getBinLower(bin) << curHist->getBinCenter(bin) << curHist->getBinUpper(bin);
        if(curHist->nEntries==1) {
            curHist->fill(-1.5, 10);
            curHist->fill(1.5, 10);
        }
        if(curHist->nEntries==0) {
            for(int i=0; i<1000; ++i) {
                double rnd = QRandomGenerator::global()->generateDouble();
                double v = rnd*(curHist->max-curHist->min)+curHist->min;

                int bin = curHist->findBin(v);
                double lower = curHist->getBinLower(bin);
                double upper = curHist->getBinUpper(bin);
                Q_ASSERT(v>=lower && v<upper);
            }
        }
        updateFit(curHist, true);
    }
*/
    auto &histCfg = getCurHistCfg();
    histCfg.showFit = checked;
    cfg.save();
    refreshCurrentPlotA(true);
}

void HistMngrWidget::on_checkBoxLogY_toggled(bool checked)
{
    auto &histCfg = getCurHistCfg();
    histCfg.showFit = checked;
    cfg.save();
    refreshCurrentPlotA(true);
}

void HistMngrWidget::matchWinUpdated(int matchWin)
{
    if(HistId::tdcDoublePulseTime==-1) return;

    const auto &hc = cfg.histCfg[HistId::tdcDoublePulseTime];

    histContainer.baseHist[HistId::tdcDoublePulseTime] =
            Histogramm(hc.title, 2048*matchWin, -1*matchWin, matchWin);
}

void HistMngrWidget::inlCheckRequest()
{
    if(HistId::inlCheck == -1) return;

    InlCheckResults res;
    res.stopDev = cfg.stopDevIndex;
    res.stopCh = cfg.stopCh;
    for(DeviceIndex index: histContainer.inlHist.keys()){
        DeviceCalibration devCal(index);
        for(int ch: histContainer.inlHist[index].keys()){
            InlCheckCh &chRes = res.devices[index].channels[ch];

            auto &cutHist = histContainer.inlCheckHist[index][ch];

            chRes.ch = ch;
            chRes.dist = cutHist.getDistribution();
        }
    }
    emit inlCheckUpdated(res);
}

HistMngrCfg HistMngrWidget::init(bool isInlEnabled)
{
    cfg.init(isInlEnabled);
    cfg.load();
    ui->spinBoxCh->blockSignals(true);
    ui->spinBoxCh->setMaximum(HistId::nch-1);
    ui->spinBoxCh->blockSignals(false);
    ui->radioButtonInl->blockSignals(true);
    ui->radioButtonInl->setChecked(!cfg.inlHistCodeDens);
    ui->radioButtonInl->blockSignals(false);

    ui->spinBoxMwMin->blockSignals(true);
    ui->spinBoxMwMax->blockSignals(true);
    ui->spinBoxMwMin->setMaximum(cfg.mwMax);
    ui->spinBoxMwMin->setMinimum(0);
    ui->spinBoxMwMax->setMinimum(cfg.mwMin);
    ui->spinBoxMwMin->setValue(cfg.mwMin);
    ui->spinBoxMwMax->setValue(cfg.mwMax);
    ui->spinBoxMwMin->blockSignals(false);
    ui->spinBoxMwMax->blockSignals(false);
    ui->checkBoxLogY->setChecked(cfg.log_y_axis);
    ui->checkBoxLogY->hide();

    ui->comboBoxGroup->blockSignals(true);
    for (const HistCfg &c: cfg.histCfg) {
        const int &id = c.groupId;
        if(c.isBaseHist) {
            histContainer.baseHist[id] = Histogramm(c.title, c.nBins, c.min, c.max);

            histQwtItems[id] = new HistogramItem(c.getGroupTitle());
            histQwtItems[id]->setColor(Qt::darkCyan);
        }

        if(ui->comboBoxGroup->findData(c.groupId) == -1)
            ui->comboBoxGroup->insertItem(c.groupId, c.getGroupTitle(), c.groupId);
        if(cfg.curHistTitle == c.title) {
            if(c.devCh != -1) {
                ui->spinBoxCh->blockSignals(true);
                if(ui->spinBoxCh->maximum()<c.devCh)
                    ui->spinBoxCh->setMaximum(c.devCh);
                ui->spinBoxCh->setValue(c.devCh);
                ui->spinBoxCh->blockSignals(false);
            }
            ui->comboBoxGroup->setCurrentIndex(ui->comboBoxGroup->findData(c.groupId));
        }
    }
    ui->comboBoxGroup->blockSignals(false);

    onCurHistChanged();

    ui->checkBoxGlobalEnable->setChecked(cfg.globalEnabled);
    if(cfg.globalEnabled)
        refreshTimer->start();

    ui->qwtPlotHists->replot();
    ui->groupBoxStopCh->setChecked(cfg.useStopCh);
    if(!cfg.stopDevIndex.isNull())
        updateCbStopDev(QList<DeviceIndex>() << cfg.stopDevIndex);
    ui->spinBoxStopCh->setValue(cfg.stopCh);

    if(HistId::tdcDoublePulseTime != -1)
        matchWinUpdated(TDC_MAX_LAT_NS);
    return cfg;
}

QVector<double> HistMngrWidget::makeDNLCorrectionTable(const QVector<double> code_density)
{
    if (code_density.size() != 1024) {
        qWarning() << "makeDNLCorrectionTable: invalid input vector size";
        return QVector<double>();
    }
    double sumv = 0;
    for(int i=0; i<1024; i++) {
        double v = code_density[i];
        sumv += v;
    }
    QVector<double> inl(1024, 0.0);
    double avgv = sumv / 1024.;
    double sumrelv = 0;
    double inlsum = 0;
    for(int i=0; i<1024; i++) {
        double relv = (code_density[i] - avgv) / avgv;
        sumrelv += relv;
        inl[i] = sumrelv;
        inlsum += sumrelv;
    }
//    qInfo() << "avgv = " << avgv; // Ok
//    qInfo() << "sumrelv = " << sumrelv;
//    qInfo() << "inlsum = " << inlsum;
    QVector<double> corr(1024, 0.0);
    for(int i=0; i<1024; i++) {
        corr[i] = inl[i] - inlsum/1024.;
        corr[i] = round(corr[i]*100.0) / 100.0; // quantize
    }
//    qInfo() << "TDC DNL for channel " << ch << corr;
    return corr;
}

void HistMngrWidget::updateInlCorr(QMap<DeviceIndex, DeviceCalibration> allCal)
{
    inlFromConfig = allCal;
}

void HistMngrWidget::updateInlCorr(const DeviceIndex &index, const QMap<int, QVector<double> > &corrMap)
{
    for(auto ch: corrMap.keys())
        inlFromConfig[index].inl[ch] = corrMap[ch];
}

void HistMngrWidget::updateFit(const Histogramm *curHist)
{
    HistCfg curHistCfg = getCurHistCfg();

    if(curHist && curHistCfg.showFit && curHist->getEntries() && curHist->nBins>3){
        const auto hist_dist = curHist->getDistribution();
        ui->labelFitInfo->setText(QString("Mean:  %1\nSigma: %2")
                                  .arg(hist_dist.mean, 0, 'f', 3)
                                  .arg(hist_dist.rms, 0, 'f', 3));
    } else {
        ui->labelFitInfo->setText(curHistCfg.showFit?"No fit data":"");
    }
}

DeviceIndex HistMngrWidget::getDevIndex(bool getStopDev, int idInComboBox) const
{
    auto cb = getStopDev ? ui->comboBoxStopDev : ui->comboBoxDev;
    QList<QVariant> indexList;
    if(idInComboBox == -1)
        indexList = cb->currentData().toList();
    else
        indexList = cb->itemData(idInComboBox).toList();
    if(indexList.isEmpty()) return DeviceIndex();

    bool ok;
    const quint16 id = static_cast<quint16>(indexList.first().toUInt(&ok));
    if(!ok) qFatal("Wrong filling of ui->comboBoxDev->count()");
    const quint64 ser = indexList.last().toULongLong(&ok);
    if(!ok) qFatal("Wrong filling of ui->comboBoxDev->count()");
    DeviceIndex dev(id, ser);
    if(dev.isNull()) qFatal("Wrong filling of ui->comboBoxDev->count()");
    return  dev;
}

DeviceIndex HistMngrWidget::getCurDev(bool getStopDev) const
{
    return getDevIndex(getStopDev);
}

int HistMngrWidget::getCurCh() const
{
    return ui->spinBoxCh->value();
}

HistCfg &HistMngrWidget::getCurHistCfg()
{
    bool ok;
    int groupId = ui->comboBoxGroup->currentData().toInt(&ok);
    Q_ASSERT(ok && groupId!=-1);
    Q_ASSERT(cfg.histCfg.size() > groupId);
    return cfg.histCfg[groupId];
}

Histogramm *HistMngrWidget::getCurHist()
{
    auto &hCfg = getCurHistCfg();
    if(hCfg.isBaseHist) {
        return getHistById(cfg.curHistId);
    } else {
        DeviceIndex index = getCurDev();
        return getHistByIdIndexCh(hCfg.groupId, index, getCurCh());
    }
}

HistCfg HistMngrWidget::onCurHistChanged()
{
    HistCfg &curHistCfg = getCurHistCfg();
    DeviceIndex curDevIndex = getCurDev();
    uint curDevCh = getCurCh();

    if(cfg.curHistId != -1) {
        ui->qwtPlotHists->detachItems(QwtPlotItem::Rtti_PlotItem, false);
    }
    cfg.curHistId = curHistCfg.groupId;
    curHistCfg.devIndex = curDevIndex;
    curHistCfg.devCh = static_cast<int>(curDevCh);
    cfg.curHistTitle = curHistCfg.title;

    QwtPlotItem *newPlotItem = nullptr;
    if(curHistCfg.isBaseHist) {
        newPlotItem = histQwtItems.value(curHistCfg.groupId, nullptr);
        Q_ASSERT(newPlotItem);
    } else {
        if(curHistCfg.groupId == HistId::inl)
            newPlotItem = inlCurvs.value(curDevIndex).value(curDevCh, nullptr);
        else if(curHistCfg.groupId == HistId::inlCheck)
            newPlotItem = devChHists.value(curDevIndex).value(curDevCh, nullptr);
    }

    ui->groupBoxInl->setVisible(curHistCfg.groupId == HistId::inl);
    ui->groupBoxMatchWin->setVisible(curHistCfg.groupId == HistId::tdcTime && HistId::tdcTime!=-1);
    ui->checkBoxZeroBin->setVisible(curHistCfg.groupId == HistId::tdcHitEv);
    if(newPlotItem) {
        newPlotItem->attach(ui->qwtPlotHists);
    }

    ui->qwtPlotHists->setTitle(curHistCfg.getGroupTitle());
    // bring fit curv to foreground, curve will be drawn on next updateFit(...)
    refreshCurrentPlotA(true);
    cfg.save();

    ui->qwtPlotHists->axisScaleEngine(QwtPlot::xBottom)->setMargins(1, 1);
//    ui->qwtPlotHists->setAxisAutoScale(QwtPlot::xBottom);
//    qInfo()<<"axisAutoScale:"<<ui->qwtPlotHists->axisAutoScale(QwtPlot::xBottom);

    // Update GUI for this hist
    const int groupIndex = ui->comboBoxGroup->findData(curHistCfg.groupId);
    if(groupIndex!=-1) {
        ui->comboBoxGroup->setItemText(groupIndex, curHistCfg.getGroupTitle());
    }
    updateHistToolTip();

    ui->checkBoxHistGroupEn->blockSignals(true);
    ui->checkBoxHistGroupEn->setChecked(curHistCfg.histGroupEn);
    ui->checkBoxHistGroupEn->blockSignals(false);
    ui->groupBoxDevChSelect->setVisible(!curHistCfg.isBaseHist);
    ui->groupBoxStopCh->setVisible(curHistCfg.hasStopCh());
    ui->groupBoxFit->setVisible(curHistCfg.fittable);
    ui->groupBoxFit->setChecked(curHistCfg.showFit);

    return curHistCfg;
}

void HistMngrWidget::devListUpdated(const HistDevType &devChOffset, const HistContainer &hc)
{
    // Update ComboBox StopCh
    cfg.devChOffset = devChOffset;
    const int maxChNum = cfg.getMaxChNum();
    auto &tdcChCfg = cfg.histCfg[HistId::tdcCh];
    tdcChCfg.max = maxChNum;
    tdcChCfg.nBins = maxChNum;
    auto &tdcHitEvCfg = cfg.histCfg[HistId::tdcHitEv];
    int maxHit = maxChNum ? (maxChNum*10) : 100;
    tdcHitEvCfg.max = maxHit;
    tdcHitEvCfg.nBins = maxHit;
    histContainer = hc;
    updateCbStopDev(cfg.devChOffset.keys());
    QMap<int, QString> map_tool_tip;

    // Update hist 'Hits by dev/ch'
    QMap<DeviceIndex, QMap<int, int> > devChCnt;
    Histogramm &chHist = histContainer.baseHist[HistId::tdcCh];
    for (const DeviceIndex &i: devChOffset.keys()) {
        addDevice(i);
        int chOffset = devChOffset[i];
        map_tool_tip[chOffset] = QString("\n%1-%2: %3").arg(chOffset)
                .arg(chOffset+getDeviceChNum(i)).arg(i.getIdent());
    }

    QString &tool_tip = cfg.toolTipSuffix[HistId::tdcCh];
    tool_tip.clear();
    for(QString s : map_tool_tip.values()) {
        tool_tip.append(s);
    }
    updateHistToolTip();

    int totChNum = getTotalChNum(cfg.devChOffset);
    chHist.resize((totChNum ? totChNum : 1), 0, totChNum);
    int hits_max = (totChNum ? (totChNum*3/2) : 100);
    histContainer.baseHist[HistId::tdcHitEv].resize(hits_max, 0, hits_max);

    for (const DeviceIndex &i: devChCnt.keys()) {
        if(!cfg.devChOffset.contains(i)) continue;
        int chOffset = cfg.devChOffset[i];
        auto chCnt = devChCnt[i];
        for (int ch: chCnt.keys()) {
            chHist.fill(chOffset+ch, chCnt[ch]);
        }
    }
}

void HistMngrWidget::resetHistMngr()
{
    QComboBox *curDrvCb = ui->comboBoxDev;
    curDrvCb->blockSignals(true);
    while(curDrvCb->count())
        curDrvCb->removeItem(0);
    curDrvCb->blockSignals(false);
    for(auto i : inlCurvs.keys()) {
        for(uint ch : inlCurvs[i].keys()) {
            QwtPlotCurve *curve = inlCurvs[i][ch];
            curve->detach();
            delete curve;
        }
    }
    inlCurvs.clear();
    for(auto i : devChHists.keys()) {
        for(uint ch : devChHists[i].keys()) {
            HistogramItem *curve = devChHists[i][ch];
            curve->detach();
            delete curve;
        }
    }
    devChHists.clear();
    updateCbStopDev(QList<DeviceIndex>());
    knownDevices.clear();
}

int HistMngrWidget::getTotalChNum(const QMap<DeviceIndex, int> &devChOffset)
{
    int chNum=0;
    for (const DeviceIndex &i: devChOffset.keys()) {
        chNum += getDeviceChNum(i);
    }
    return chNum;
}

inline bool HistMngrWidget::hasHistById(int id) const
{
    return histContainer.baseHist.contains(id);
}

inline Histogramm *HistMngrWidget::getHistById(int id)
{
    return hasHistById(id) ? &histContainer.baseHist[id] : nullptr;
}

Histogramm *HistMngrWidget::getHistByIdIndexCh(int id, const DeviceIndex &index, int ch)
{
    if(id == HistId::inl &&
            histContainer.inlHist.contains(index) &&
            histContainer.inlHist[index].contains(ch)) {
        return &histContainer.inlHist[index][ch];
    }
    if(id == HistId::inlCheck &&
            histContainer.inlCheckHist.contains(index) &&
            histContainer.inlCheckHist[index].contains(ch)) {
        return &histContainer.inlCheckHist[index][ch];
    }
    return nullptr;
}

void HistMngrWidget::updateCbStopDev(QList<DeviceIndex> list)
{
    int stopIndex=-1;
    QComboBox *box = ui->comboBoxStopDev;
    box->blockSignals(true);
    while(box->count()) {
        box->removeItem(0);
    }
    if(!cfg.stopDevIndex.isNull() && !list.contains(cfg.stopDevIndex)) {
        list << cfg.stopDevIndex;
    }
    for (const DeviceIndex &index: list) {
        QList<QVariant> userData;
        userData << index.getDevId() << index.getSerial();
        box->addItem(index.getIdent(), userData);
        if(index == cfg.stopDevIndex) {
            stopIndex = box->findText(index.getIdent());
            box->setCurrentIndex(stopIndex);
        }
    }
    box->blockSignals(false);


    if(cfg.stopDevIndex.isNull() && !list.isEmpty()) {
        if(box->currentIndex() != 0) {
            box->setCurrentIndex(0);
        } else {
            on_comboBoxStopDev_currentIndexChanged(0);
        }
    }
    updateSpinBoxMaxCh(true);
}

QVector<double> HistMngrWidget::getInl(const TdcEvent &ev, const TdcChHit &tdcEv) const
{
    return getInl(ev.index, tdcEv.ch);
}

QVector<double> HistMngrWidget::getInl(const DeviceIndex &index, int ch) const
{
    return inlFromConfig.value(index).inl.value(ch, QVector<double>());
}

void HistMngrWidget::addBaseHist(HistCfg hCfg, int nbins, int min, int max)
{
    cfg.histCfg.push_back(hCfg);
    histContainer.baseHist[hCfg.groupId] = Histogramm(hCfg.title, nbins, min, max);

    histQwtItems[hCfg.groupId] = new HistogramItem(hCfg.getGroupTitle());
    histQwtItems[hCfg.groupId]->setColor(Qt::darkCyan);

}

void HistMngrWidget::updateHistToolTip()
{
    int curHistId = getCurHistCfg().groupId;
    QString toolTip = cfg.getGroupToolTip(curHistId, true);

    ui->comboBoxGroup->setToolTip(cfg.getGroupToolTip(curHistId));
    ui->labelHistDescr->setText(toolTip);
    ui->labelHistDescr->setVisible(!toolTip.isEmpty());
}

void HistMngrWidget::updateSpinBoxMaxCh(bool stopDev)
{
    auto devIndex = getDevIndex(stopDev);
    uint devChNum = getDeviceChNum(devIndex);
    auto sb = stopDev ? ui->spinBoxStopCh : ui->spinBoxCh;
    sb->setMaximum(static_cast<int>(devChNum-1));
}

void HistMngrWidget::addDevice(const DeviceIndex &index)
{
    if(knownDevices.contains(index)) return;

    const uint nch = getDeviceChNum(index);
    for(uint ch=0; ch<nch; ++ch) {
        if(HistId::inl!=-1 && !inlCurvs[index].contains(ch)) {
            inlCurvs[index][ch] = new QwtPlotCurve(cfg.histCfg[HistId::inl].getGroupTitle());
        }

        if(HistId::inlCheck!=-1 && !devChHists.value(index).contains(ch)) {
            devChHists[index][ch] = new HistogramItem(cfg.histCfg[HistId::inlCheck].getGroupTitle());
            devChHists[index][ch]->setColor(Qt::darkCyan);
        }
    }

    int devCnt = ui->comboBoxDev->count();
    bool found = false;
    for(int i=0; i<devCnt; ++i) {
        DeviceIndex dev = getDevIndex(false, i);
        if(dev.isNull()) qFatal("Wrong filling of ui->comboBoxDev->count()");
        if(index == dev) {
            found = true;
            break;
        }
    }
    if(!found) {
        QList<QVariant> indexList;
        indexList << index.getDevId() << index.getSerial();
        ui->comboBoxDev->addItem(index.getIdent(), indexList);
    }

    knownDevices << index;
}

void HistMngrWidget::on_checkBoxZeroBin_clicked(bool checked)
{
    hide_zero_bin = checked;
}
