//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Client.h"

#include "mongo/RootConfig.h"
#include "remote-control-server/RunControlCommands.h"

#include <QTcpSocket>
#include <QJsonDocument>
#include <QDebug>
#include <QTimer>

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

namespace {
const qint32 EXPIRE_TIMEOUT = 1000;

#pragma pack(push, 1)
struct TypeLengthHeader {
    quint32 type;
    quint32 len;
};
#pragma pack(pop)
}

Client::Client(int i, QString _progType, QString _progIndex) :
    QObject(nullptr),
    id(i),
    progType(_progType),
    progIndex(_progIndex),
    enabled(false),
    tcpSocket(new QTcpSocket(this)),
    hostPort(0),
    reconnectTimer(new QTimer(this)),
    connectingAttempts(0),
    run(false),
    runNumber(0),
    block_size(0),
    dataTimer(new QTimer(this))
{
    Q_ASSERT(id); // client id can't be equal to 0
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<CLIENT_PROTO_ID>();

    connect(tcpSocket, &QTcpSocket::readyRead,
            this, &Client::readCommandResponce, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::disconnected,
            this, &Client::disconnected, Qt::QueuedConnection);
    qRegisterMetaType<QAbstractSocket::SocketError>();
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)),
            SLOT(displayError(QAbstractSocket::SocketError)), Qt::QueuedConnection);
    connect(reconnectTimer, &QTimer::timeout,
            this, &Client::reconnectToServer, Qt::QueuedConnection);
    reconnectTimer->setSingleShot(true);

    connect(dataTimer, &QTimer::timeout,
            this, &Client::checkData);
    dataTimer->setSingleShot(true);
    dataTimer->setInterval(EXPIRE_TIMEOUT);
}

void Client::readCommandResponce()
{
    while (tcpSocket->bytesAvailable()) {
        const auto d = tcpSocket->readAll();
        recivedData += d;
    }

    processIncomingMsg();
}

void Client::reconnectToServer()
{
    if(!enabled) return;

    try{
        if(tcpSocket->isValid() &&
                (tcpSocket->state() == QTcpSocket::ConnectedState ||
                 tcpSocket->state() == QTcpSocket::ConnectingState) ) {
            return;
        }
        if(tcpSocket->state() != QTcpSocket::UnconnectedState)
            tcpSocket->abort();
        if(hostAddr.isNull() || hostPort==0) {
            reconnectTimer->start(5000);
            return;
        }

        tcpSocket->connectToHost(hostAddr, hostPort);
        ++connectingAttempts;
        if(tcpSocket->waitForConnected(1000)){
            qDebug() << getIdentStr() << "TcpClient connected";
            emit statusUpdated(id, "Connected");
        } else {
            if(connectingAttempts<=1) {
                QString mess = QString("TcpClient can't connected.Err:%2")
                        .arg(tcpSocket->errorString());
                qWarning() << getIdentStr() << mess;
//                sendLog(mess);
                emit statusUpdated(id, mess);
                reconnectTimer->start(1000);
            } else
                reconnectTimer->start(5000);
        }
    } catch(...) {
        qWarning() << getIdentStr() << "Exception in TcpClient::requestNewData()";
        reconnectTimer->start(5000);
    }
}

void Client::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qInfo() << getIdentStr() << "Remote hoste closed connection.";
        emit statusUpdated(id, "Remote hoste closed connection");
        break;
    case QAbstractSocket::HostNotFoundError:
        qInfo() << getIdentStr() << "The host was not found." 
                << "Please check the host name and port settings.";
        emit statusUpdated(id, "host was not found");
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
            qInfo() << getIdentStr()
                    << tr("The connection was refused by the peer. "
                   "Make sure the fortune server is running, "
                   "and check that the host name and port "
                   "settings are correct.");
            emit statusUpdated(id, "The connection was refused by the peer");
        }
        break;
    default:
        qInfo() << getIdentStr()
                << tr("The following error occurred: %1.")
                .arg(tcpSocket->errorString());
        emit statusUpdated(id, tr("The following error occurred: %1.")
                        .arg(tcpSocket->errorString()));
    }
}

void Client::sendRunState(RUN_STATES state, QString _runIndex, int _runNumber)
{
    if(!tcpSocket->isValid() ||
            (tcpSocket->state() != QTcpSocket::ConnectedState &&
             tcpSocket->state() != QTcpSocket::ConnectingState) ){
        emit runStateUpdated(id, RS_ERROR, _runIndex, _runNumber);
        return;
    }

    switch (state) {
    case RS_STARTING:
        run = true;
        runIndex = _runIndex;
        runNumber = _runNumber;
        break;
    case RS_STOPPING:
        run = false;
        break;
    default:
        qCritical() << getIdentStr() << __func__ << "Invalid RUN_STATES:" << state;
        return;
    }

    QString mess = QString("%1 %2 %3 %4")
            .arg(run ? RC_CMD_START : RC_CMD_STOP)
            .arg(_runIndex)
            .arg(_runNumber)
            .arg(RC_CMD_END);
    sendCommand(mess);
}

void Client::startReadOut(int i)
{
    if(i && id!=i) {
        return;
    }

    if(enabled) {
        qWarning() << getIdentStr() << "Trying to start started client";
        return;
    }
    reconnectToServer();
    enabled = true;
    reconnectTimer->start(0);
    emit statusUpdated(id, "connecting to server...");
}


void Client::stopReadOut(int i)
{
    if(i && id!=i) {
        return;
    }
    if(!enabled) {
        qWarning() << getIdentStr() <<"Trying to stop stopped client";
        return;
    }
    enabled = false;
    protocolId = CLIENT_PROTO_UNKNOWN;
    emit protocolChanged(id, protocolId);
    if(tcpSocket->state() != QTcpSocket::UnconnectedState)
        tcpSocket->disconnectFromHost();
    reconnectTimer->stop();
    emit statusUpdated(id, "tcp: stoped readout");
}

void Client::deleteClient(int i)
{
    if(i && i!=id)
        return;
    deleteLater();
}

void Client::setAddress(int i, QHostAddress addr, quint16 port)
{
    if(id != i)
        return;
    if(hostAddr == addr && hostPort == port)
        return;
    // Different addr/port can send program from other host
    if(isConnectionAlive())
        return;
    bool enabledBak = enabled;
    if(enabled){
        blockSignals(true);
        stopReadOut(id);
        blockSignals(false);
    }
    hostAddr = addr;
    hostPort = port;

    if(enabledBak)
        startReadOut(id);
}

void Client::setEmptyAddress(int i)
{
    if(id != i)
        return;
    hostAddr.clear();
    hostPort = 0;
}

void Client::evNumRequest(int checkIndex)
{
    if(!enabled)
        return;
    if(!tcpSocket->isValid() ||
            (tcpSocket->state() != QTcpSocket::ConnectedState &&
             tcpSocket->state() != QTcpSocket::ConnectingState) ){
        emit runStateUpdated(id, RS_ERROR, runIndex, runNumber);
        return;
    }
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_EV_NUM_CHECK)
            .arg(checkIndex)
            .arg(RC_CMD_END);
    sendCommand(mess);
}

void Client::disconnected()
{
    recivedData.clear();
    if(enabled) {
        QString s = "TcpClient was disconnected, trying to reconnect";
        (run?qWarning():qInfo()) << getIdentStr() << s;
        emit statusUpdated(id, s);
        emit ready4Run(id, false);
        if(run)
            emit runError(runIndex, runNumber, QString("Client was disconnected"));
        connectingAttempts=0;
        reconnectTimer->start(1000);
    } else {
        qInfo() << getIdentStr() << "TcpClient disconnected.";
    }
}

void Client::checkData()
{
    if(!recivedData.isEmpty()) {
        qWarning() << getIdentStr() << "has timeout on incompleted data" << "recivedData.size()="<<recivedData.size();
        dataTimerTimeout=true;
        recivedData.remove(0,1);
        ++curruptedSize;
        processIncomingMsg();
    }
    qWarning() << getIdentStr() << "curruptedSize " << curruptedSize;
}

void Client::sendCommand(const QString &mess)
{
    switch (protocolId) {
    case CLIENT_PROTO_V1:
        sendString(mess);
        break;
    case CLIENT_PROTO_V2:
        send_byte_array(RC_MSG_CMD_OLD, mess.toUtf8());
        break;
    case CLIENT_PROTO_V3:
        send_byte_array(RC_MSG_CMD, mess.toUtf8());
        break;
    default:
        qWarning() << getIdentStr() << "Can't send command in case of FORMAT_ID_UNKNOWN";
    }
}

void Client::sendString(const QString &mess)
{
//--- FORMAT_ID_V1
    QByteArray messData = mess.toUtf8();
    if(tcpSocket==nullptr)
        return;
    int written = tcpSocket->write(messData);
    if(written != messData.size()) {
        qWarning() << getIdentStr()
                   << "Failed to send string to run control server:"
                   << tcpSocket->errorString();

    } else {
        qDebug() << getIdentStr()
                 << "send to run control server:"
                 << mess.trimmed();
    }
}

void Client::send_byte_array(quint32 type, QByteArray data_ba)
{
    if (!isConnectionAlive()) {
        return;
    }
    QByteArray data = QByteArray(sizeof(TypeLengthHeader), Qt::Uninitialized);
    data.append(data_ba);
    quint16 check = 0;

    TypeLengthHeader *hdr = reinterpret_cast<TypeLengthHeader *>(data.data());
    hdr->type = type;
    hdr->len = static_cast<quint32>(data.size()) + sizeof (check);

    check = qChecksum(data.data(), static_cast<uint>(data.size()), Qt::ChecksumIso3309);
    data.append(reinterpret_cast<char*>(&check), sizeof (check));

    int written = static_cast<int>(tcpSocket->write(data));

    if(written == data.size()) {
        if(type == RC_MSG_CONFIG || type == RC_MSG_CONFIG_OLD) {
            qInfo() << getIdentStr() << "Sent configuration to client";
        }
    } else {
        qWarning() << getIdentStr() << "Failed to send data to run control client";
    }
}

bool Client::isConnectionAlive() const
{
    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return true;
    return false;
}

bool Client::processIncomingMsg()
{
    while (!recivedData.isEmpty()) {
        bool ok = true;
        if(protocolId == CLIENT_PROTO_V2 || protocolId == CLIENT_PROTO_V3 || protocolId == CLIENT_PROTO_UNKNOWN) {
            ok = processIncomingMsgV23();
        }

        if(protocolId == CLIENT_PROTO_V1 || protocolId == CLIENT_PROTO_UNKNOWN) {
            ok = processIncomingMsgV1();
        }

        if(ok) {
            dataTimer->stop();
            dataTimerTimeout = false;
        } else {
            if(protocolId != CLIENT_PROTO_UNKNOWN) {
                protocolId = CLIENT_PROTO_UNKNOWN;
//                qWarning() << getIdentStr() <<  "communication sync. fail or big incoming data block";
                emit protocolChanged(id, protocolId);
                continue; // try again both of them
            }
            if(dataTimerTimeout) {
                recivedData.remove(0,1);
                ++curruptedSize;
            } else {
                break;
            }
        }
    }

    if(recivedData.isEmpty()) {
        dataTimer->stop();
        dataTimerTimeout = false;
    } else {
        if(!dataTimer->isActive()) {
            dataTimer->start();
        }
    }
    return true;
}

bool Client::processIncomingMsgV1()
{
    bool ok = false;
    int completedLen = recivedData.indexOf(RC_CMD_END);
    if(completedLen != -1) {
        completedLen += RC_CMD_END.size();
        QByteArray completedData = recivedData.left(completedLen);
        Q_ASSERT(completedData.size() == completedLen);

        ok = command_line_parse(completedData);
        if(ok) {
            if(protocolId != CLIENT_PROTO_V1) {
                protocolId = CLIENT_PROTO_V1;
                qInfo() << getIdentStr() << "set communication protocol v1";
                emit protocolChanged(id, protocolId);
            }
            recivedData.remove(0, completedData.size());
        }
    }
    return ok;
}

bool Client::processIncomingMsgV23()
{
    TypeLengthHeader *hdr = nullptr;
    hdr = reinterpret_cast<TypeLengthHeader*>(recivedData.data());
    const auto type = hdr->type;

    bool ok = (type == RC_MSG_CONFIG) || (type == RC_MSG_CMD)
            || (type == RC_MSG_CONFIG_OLD) || (type == RC_MSG_CMD_OLD);
    if(!ok)
        return ok;

    if(hdr->len > static_cast<quint32>(recivedData.size())) {
        if(!recivedData.isEmpty()) {
            if(!dataTimer->isActive()) {
                dataTimer->start();
            }
        }
        return false;
    }

    // crc-check
    quint16 check_calc = qChecksum(recivedData, hdr->len-sizeof (quint16));
    quint16 check = *reinterpret_cast<quint16*>(recivedData.data()+hdr->len-sizeof (quint16));
    if(check != check_calc) {
        qWarning() << getIdentStr() << "communication error: checksum is incorrect.";
        return false;
    }

    QByteArray data_ba = recivedData.mid(sizeof (TypeLengthHeader), hdr->len- sizeof (TypeLengthHeader)-sizeof (quint16));
    if(type == RC_MSG_CONFIG || type == RC_MSG_CONFIG_OLD) {
        qDebug() << hex << type << dec << "data" << data_ba.size();
        config_parse(data_ba);
    } else if(type == RC_MSG_CMD || type == RC_MSG_CMD_OLD) {
        qDebug() << hex << type << dec << "data" << QString::fromUtf8(data_ba);
        command_line_parse(data_ba);
    } else {
        qWarning() << getIdentStr() << QString("Unknown command type 0x%1").arg(type);
    }

    if(ok) {
        recivedData.remove(0, static_cast<int>(hdr->len));
        if(protocolId != CLIENT_PROTO_V2 && protocolId != CLIENT_PROTO_V3) {
            switch (type) {
            case RC_MSG_CMD_OLD:
            case RC_MSG_CONFIG_OLD:
                protocolId = CLIENT_PROTO_V2;
                break;
            case RC_MSG_CMD:
            case RC_MSG_CONFIG:
                protocolId = CLIENT_PROTO_V3;
                break;
            }
            qInfo() << getIdentStr() << "set communication protocol" << getProtocolIdStr(protocolId);
            emit protocolChanged(id, protocolId);
        }
    }

    return ok;
}

bool Client::command_line_parse(QByteArray data)
{
    bool allOk=true;
    QString message = QString::fromUtf8(data);

    QStringList lines = message.split(RC_CMD_END, QString::SkipEmptyParts);
    if(lines.isEmpty())
        return false;
    for(QString line: lines) {
        bool lineParsed=true;
        line = line.trimmed();
        QStringList lineParts = line.split(" ", QString::SkipEmptyParts);
        if(lineParts.isEmpty())
            continue;
        qDebug() << getIdentStr() <<"got from clien: "<<line;
        if(lineParts[0] == RC_CMD_DESCR) {
            descrStr = line.right(line.size()-RC_CMD_DESCR.size()).trimmed();
            emit descrUpdated(id, descrStr);
        } else if(lineParts[0] == RC_CMD_STATE || lineParts[0] == RC_CMD_STATUS) {
            statusStr = line.right(line.size()-RC_CMD_STATE.size()).trimmed();
            emit statusUpdated(id, statusStr);
        } else if(lineParts[0] == RC_CMD_READY4RUN) {
            bool ready = QVariant(line.right(line.size()-RC_CMD_READY4RUN.size()).trimmed()).toBool();
            emit ready4Run(id, ready);
        } else if(lineParts[0] == RC_CMD_EV_NUM_CHECK) {
            lineParsed = lineParts.size()>1;
            int checkIndex = 0;
            EvNumSet evNums;
            quint64 evNum;
            if(lineParsed)
                checkIndex = lineParts[1].toInt(&lineParsed);
            if(lineParsed) {
                for(int i=2; i<lineParts.size(); ++i){
                    evNum = lineParts[i].toULongLong(&lineParsed);
                    if(!lineParsed)
                        break;
                    evNums.insert(evNum);
                }
            }
            if(lineParsed)
                emit evNumResponse(checkIndex, id, evNums);
        } else if(lineParts[0] == RC_CMD_ERROR) {
            lineParsed = (lineParts.size()>=3);
            int runNum = 0;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                QString errStr = line;
                for(int i=0;i<3;++i)
                    errStr.replace(errStr.indexOf(lineParts[i]), lineParts[i].size(),"");
                emit runError(lineParts[1], runNum, errStr.trimmed());
            }
        } else if(lineParts[0] == RC_CMD_WARN) {
            lineParsed = (lineParts.size()>=3);
            int runNum = 0;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                QString warnStr = line;
                for(int i=0;i<3;++i)
                    warnStr.replace(warnStr.indexOf(lineParts[i]), lineParts[i].size(),"");
                emit runWarning(lineParts[1], runNum, warnStr.trimmed());
            }
        } else if(lineParts[0] == RC_CMD_BACK_PRESSURE) {
            lineParsed = (lineParts.size()==4);
            int runNum = 0;
            int boolVal;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                boolVal = lineParts[3].toInt(&lineParsed);
                emit backPressure(lineParts[1], runNum, boolVal);
            }
        } else if(lineParts[0] == RC_CMD_VME_FLAG) {
            lineParsed = (lineParts.size()==2);
            int hasVme;
            if(lineParsed) {
                hasVme = lineParts[1].toInt(&lineParsed);
                emit vmeFlag(id, hasVme);
            }
        } else {
            // Other possible command is success/fail
            bool success = false;
            RUN_STATES rs = RS_ERROR;
            QStringList lineParts = line.split(" ", QString::SkipEmptyParts);
            lineParsed = (lineParts.size()==4);

            if(lineParsed){
                if(lineParts[0] == RC_CMD_SUCCESS_PREFIX) {
                    success=true;
                } else if(lineParts[0] == RC_CMD_FAILED_PREFIX) {
                    success=false;
                } else {
                    // unknown command; skipp it
                    lineParsed=false;
                }
            }

            if(lineParsed) {
                QString clRunIndex = lineParts[2];
                int clRunNumber = QString(lineParts[3]).toInt(&lineParsed);

                if(lineParts[1] == RC_CMD_START)
                    rs = RS_STARTED;
                else if(lineParts[1] == RC_CMD_STOP)
                    rs = RS_STOPPED;
                else
                    lineParsed = false;

                if(lineParsed)
                    emit runStateUpdated(id, success ? rs : RS_ERROR,
                                         clRunIndex, clRunNumber);
            }
        }
        if(!lineParsed){
            allOk = false;
            if(protocolId != CLIENT_PROTO_UNKNOWN)
                qWarning() << "Can't parse line from client "
                           << getIdentStr() <<": "<<line;
        }
    }
    return allOk;
}

void Client::config_parse(QByteArray data)
{
    RootConfig root;
    root.set_root_from_json(QJsonDocument::fromBinaryData(data));
    emit config_transfer(id, root);
    qInfo() << QString("%1 got RootConfig (size=%L2)")
                .arg(getIdentStr()).arg(data.size());
}
