/*
**    Copyright 2006-2019 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MlinkDeviceInterfaceBase.h"

#include <QDebug>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/ioctl.h>

#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <array>

#ifndef _MSC_VER
#include <zlib.h>
#endif

#include "mlink.h"
#include "mlink_ioctls.h"
#include "mlink/frame_types.h"
#include "mlink/frameset.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/frame_hw_debug.h"
#include "regio/regiopacket.h"

namespace mlink {

static constexpr int TX_CRC = 1;

#ifndef MLINK_CHECK_FRAME_CRC
#define MLINK_CHECK_FRAME_CRC 0
#endif

#if ML_RAW_DUMP
#include "colorlog/log.h"

using colorlog::debug_level_ge;

#endif

#ifndef DEBUG_DUMP_RX_RAW_FRAMES
#define DEBUG_DUMP_RX_RAW_FRAMES 0
#endif

#ifndef DEBUG_DUMP_TX_RAW_FRAMES
#define DEBUG_DUMP_TX_RAW_FRAMES 0
#endif

#ifndef DEBUG_DUMP_DECODED_FRAMES
#define DEBUG_DUMP_DECODED_FRAMES 0
#endif

const char *MlinkDeviceInterfaceBase::ml_strerror(int errnum) const
{
    switch (errnum) {
    case ERR_TIMERECV: return "Receive timeout";
    case ERR_TIMESEND: return "Send timeout";
    case ERR_BADCRC: return "Bad CRC";
    case ERR_REMOTEIO: return "Remote I/O error";
    default: return strerror(errnum);
    }
}

//--------------------------------------------------------------------------
//
// Low-level device functions
//

int MlinkDeviceInterfaceBase::ml_open (const char *devname)
{
    if (!devname || strlen(devname)==0) devname = "/dev/mlink0";

    if (ml_fd != -1)
        ml_close ();
    ml_fd = open (devname, O_RDWR);
    if (ml_fd < 0) {
        ml_fd = -1;
        return (-errno);
    }
    return 0;
}

bool MlinkDeviceInterfaceBase::isOpen() const
{
    return ml_fd != -1;
}

int MlinkDeviceInterfaceBase::ml_close ()
{
    if (ml_fd != -1) {
        close (ml_fd);
        ml_fd = -1;
    }
    return 0;
}

int MlinkDeviceInterfaceBase::ml_get_caps (mlink_caps_t *caps)
{
    Q_ASSERT(caps);

    if (ml_fd == -1)
        return (-EINVAL);
    memset(caps, 0, sizeof(*caps));
    if ((ioctl (ml_fd, ML_IOC_GET_CAPS, caps))) {
        qWarning() << QString("ML_IOC_GET_CAPS failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

bool MlinkDeviceInterfaceBase::is_link_up() const
{
    mlink_status_t st;
    int ret = ml_get_driver_status(&st);
    if (ret != 0)
        return false;
    return ! st.bit.link_down;
}

int MlinkDeviceInterfaceBase::ml_get_driver_status (mlink_status_t *st) const
{
    Q_ASSERT(st);

    if (ml_fd == -1)
        return (-EINVAL);
    if (!st)
        return (-EINVAL);
    memset(st, 0, sizeof(*st));
    if ((ioctl (ml_fd, ML_IOC_GET_STATUS, st))) {
        qWarning() << QString("ML_IOC_GET_STATUS failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int MlinkDeviceInterfaceBase::ml_get_driver_version(unsigned long *n)
{
    Q_ASSERT(n);

    if (ml_fd == -1)
        return (-EINVAL);
    *n = 0;
    if ((ioctl (ml_fd, ML_IOC_GET_DRIVER_VERSION, n))) {
        qWarning() << QString("ML_IOC_GET_DRIVER_VERSION failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int MlinkDeviceInterfaceBase::ml_get_board_serial_id(uint64_t *n)
{
    Q_ASSERT(n);

    if (ml_fd == -1)
        return (-EINVAL);
    *n = 0;
    if ((ioctl (ml_fd, ML_IOC_GET_SERIAL_ID, n))) {
        qWarning() << QString("ML_IOC_GET_SERIAL_ID failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int MlinkDeviceInterfaceBase::ml_reset ()
{
    //  ssp_omb1_t omb1;
    if (ml_fd == -1)
        return (-EINVAL);
    if ((ioctl (ml_fd, ML_IOC_RESET))) {
        qWarning() << QString("ML_IOC_RESET failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    //    qDebug() << "M-Link interface reset";
    return 0;
}

int MlinkDeviceInterfaceBase::ml_dma_read (uint32_t buf[], size_t nwords, size_t * words_received, int timeout)
{
    ml_dmaop_t dmaop;
    memset(&dmaop, 0, sizeof(dmaop));
    if (!buf)
        return (-EINVAL);

    dmaop.size = nwords * 4;
    dmaop.buf = reinterpret_cast<unsigned char *>(buf);
    dmaop.timeout = timeout;
    //  fprintf(stderr, "dmaop.buf = %p\n", dmaop.buf);
    if ((ioctl (ml_fd, ML_IOC_DMA_RECV, &dmaop))) {
        //    int prio = LOG_ERR;
        //    if (errno == ETIME)
        //      prio = LOG_WARNING;
        if (errno != ETIME)
            qDebug() << QString("frame receive failed: %1").arg(ml_strerror (errno));
        if (errno == ETIME) errno = ERR_TIMERECV;
        return (-errno);
    }
    size_t bytes_recv = dmaop.size;
    rxb += bytes_recv;
    rxf++;
    //    qDebug() << QString("received frame with %1 bytes").arg(nrecv);

    if (words_received)
        *words_received = bytes_recv / 4;
    return 0;
}

int MlinkDeviceInterfaceBase::ml_dma_write (const uint32_t buf[], size_t nwords, int timeout)
{
    ml_dmaop_ro_t dmaop;
    memset(&dmaop, 0, sizeof(dmaop));
    if (!buf)
        return (-EINVAL);

    dmaop.size = nwords * 4;
    dmaop.buf = reinterpret_cast<const unsigned char *>(buf);
    dmaop.timeout = timeout;
    if ((ioctl (ml_fd, ML_IOC_DMA_SEND, &dmaop))) {
        qWarning() << QString("frame send failed: %1").arg(ml_strerror (errno));
        if (errno == ETIME) errno = ERR_TIMESEND;
        return (-errno);
    }
    txb += dmaop.size;
    txf++;
    //    qDebug() << QString("sent frame with %1 bytes").arg(nbytes);
    return 0;
}

//--------------------------------------------------------------------------
//
// Frame operations
//

static int ml_make_crc (const void *buf, int len)
{
    uLong crc = crc32(0L, Z_NULL, 0);
    crc = crc32 (crc, (Bytef *)buf, len);
    return crc;
}

int MlinkDeviceInterfaceBase::ml_send_packet (const RegIoPacket &packet, int timeout)
{
    const size_t data_size = packet.data.size();
    Q_ASSERT(data_size <= ML_MAX_FRAME_DATA_WORDS);
    if (data_size > ML_MAX_FRAME_DATA_WORDS) {
        return (-EINVAL);
    }
    const int word_size = sizeof(PacketRawDataType);
    size_t frame_size = data_size + ML_FRAME_HEADER_WORDS + ML_FRAME_TRAILER_WORDS;
    // fill in header
    auto fh = reinterpret_cast<frame_hdr_t *>(&tx_buffer[0]);
    fh->sync = ML_FRAME_SYNC;
    fh->type = packet.type;
    fh->len = frame_size;
    fh->seq = packet.seq;
    fh->dst = packet.dst;
    fh->src = packet.src;
    // copy data
    if (data_size)
        memcpy (&tx_buffer[ML_FRAME_HEADER_WORDS], &packet.data[0], data_size * word_size);
    // fill in crc
    quint32 crc = 0;
    if (TX_CRC)
        crc = ml_make_crc (&tx_buffer[0], (data_size + 3) * word_size);
    tx_buffer[ML_FRAME_HEADER_WORDS + data_size] = crc;
    int ret = ml_dma_write (&tx_buffer[0], frame_size, timeout);
    return ret;
}

int MlinkDeviceInterfaceBase::ml_recv_packet (RegIoPacket *packet, int timeout)
{
    struct frameset_t frameset;
    int ret = ml_recv_data_frame_multi(&rx_buffer[0], sizeof(rx_buffer), &frameset, timeout);
    // copy first frame data
    for (size_t i=0; i<frameset.count; i++) {
        const struct framedesc_t *fdesc = &frameset.fdesc[i];
        if (fdesc->len >= ML_FRAME_HEADER_WORDS) {
            if (packet) {
                auto fh = reinterpret_cast<const frame_hdr_t *>(fdesc->rawdata);
                packet->type = fh->type;
                packet->src = fh->src;
                packet->dst = fh->dst;
                packet->seq = fh->seq;
                size_t datawords = fdesc->len - ML_FRAME_HEADER_WORDS;
                if (datawords >= ML_FRAME_TRAILER_WORDS) datawords -= ML_FRAME_TRAILER_WORDS;
                packet->data.resize(datawords);
                const quint32 *frame_data = &fdesc->rawdata[ML_FRAME_HEADER_WORDS];
                memcpy(&packet->data[0], frame_data, datawords*sizeof(PacketRawDataType));
            }
        }
        // copy only first frame
        break;
    }
    return ret;
}

// fdata is raw buffer with size fsize words. Pointers to frames are returned in framedesc.
int MlinkDeviceInterfaceBase::ml_recv_data_frame_multi(uint32_t rawbuf[], size_t rawbufsize, struct frameset_t *frameset, int timeout)
{
    frameset->count = 0;
    size_t received_rawwords = 0;
    int ret = ml_dma_read (rawbuf, rawbufsize, &received_rawwords, timeout);

    if (ret == -ERR_TIMERECV) return ret;

    if (received_rawwords == 0) {
        print_link_error(ret, "no received frame available");
        ret = -ERR_REMOTEIO;
        return (ret);
    }

    const size_t received_rawbytes = received_rawwords * 4;

    //    qDebug() << QString("Received frameset of %1 words").arg(received_rawwords);

    size_t p = 0;
    size_t frame_count = 0;
    // int frame_warning_count = 0;
    while (p < received_rawwords) {
        const uint32_t *raw_frame = &rawbuf[p];
        auto fh = reinterpret_cast<const frame_hdr_t *>(raw_frame);
        const size_t rawwords = received_rawwords - p;
        int frame_error = 0;
        int frame_warning = 0;

        int syncval = fh->sync;
        //        int type = fh->type;
        if (syncval != ML_FRAME_SYNC) {
            print_frame_error(ret, QString("Invalid frame #%1: bad sync 0x%2, expected 0x%3")
                                       .arg(frame_count).arg(syncval, 4, 16, QChar('0')).arg(ML_FRAME_SYNC, 4, 16, QChar('0')));
            frame_error = 1;
            ml_print_raw_buffer (rawbuf, received_rawbytes);
            break;
        }

        if (rawwords < ML_MIN_RAW_FRAME_WORDS) {
            print_frame_error(ret, QString("Invalid frame #%1 size: %2, minimum is %3")
                                       .arg(frame_count).arg(rawwords).arg(ML_MIN_RAW_FRAME_WORDS));
            ml_print_raw_buffer (raw_frame, rawwords*4);
            frame_error = 1;
            break;
        }


        size_t hdrlen = fh->len;
        //        int seq = fh->seq;
        //        int src = fh->src;
        //        int dst = fh->dst;

        size_t rawlen = hdrlen;
        if (rawlen > rawwords) {
            ret = -ERR_REMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: header len %2, remaining %3 words, frame truncated")
                                       .arg(frame_count).arg(hdrlen).arg(rawwords));
            frame_error = 1;
            rawlen = rawwords;
        }

        if (rawlen < ML_MIN_RAW_FRAME_WORDS) {
            //    ret = -ERR_REMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: invalid size %2").arg(frame_count).arg(rawlen));
            //    rawlen = ML_MAX_RAW_FRAME_WORDS;
            frame_error = 1;
            rawlen = ML_MIN_RAW_FRAME_WORDS;
        }

        if (rawlen > ML_MAX_RAW_FRAME_WORDS) {
            //    ret = -ERR_REMOTEIO;
            print_frame_error(ret, QString("Invalid frame #%1: size too large: %2, maximum is %3")
                                       .arg(frame_count).arg(rawlen).arg(ML_MAX_RAW_FRAME_WORDS));
            //    rawlen = ML_MAX_RAW_FRAME_WORDS;
            frame_warning++;
        }

#if MLINK_CHECK_FRAME_CRC
        size_t crcoff = rawlen - 1;

        unsigned int crc = raw_frame[crcoff];
        unsigned int calcrc = ml_make_crc(raw_frame, crcoff*4);

        //  if (crc != calcrc && crc != ML_CRC_HACK) {
        if (crc != calcrc) {
            ret = -ERR_BADCRC; /*-ERR_REMOTEIO;*/
            print_frame_error(ret, QString("Invalid frame #%1: bad CRC 0x%2, calculated 0x%3")
                                       .arg(frame_count).arg(crc, 8, 16, QChar('0')).arg(calcrc, 8, 16, QChar('0')));
            frame_warning++;
            //    return (ret);
        }
#else
        //        unsigned int crc = 0;
        //        unsigned int calcrc = 0;
#endif

        if (frame_error) {
            qCritical() << QString("Received invalid frame #%1:").arg(frame_count) << *fh;
        } else if (frame_warning) {
            qWarning() << QString("Received possibly incorrect frame #%1:").arg(frame_count) << *fh;
        } else {
            //            qDebug() << QString("Received frame #%1:").arg(frame_count) << *fh;
        }
#if ML_RAW_DUMP
        int need_raw_dump = 0;
        if ((frame_error||frame_warning) && debug_level_ge(1))
            need_raw_dump = 1;
        if (DEBUG_DUMP_RX_RAW_FRAMES) {
            //            if (debug_level_ge(5))
            need_raw_dump = 1;
        }

        if (need_raw_dump) {
            //	debug(0, "Received raw frame #%zu:", frame_count);
            //  	ml_print_raw_buffer (raw_frame, rawwords*4);
            qDebug() << "Received raw data:";
            ml_print_raw_buffer (rawbuf, received_rawbytes);
        }

#if DEBUG_DUMP_DECODED_FRAMES
        if (debug_level_ge(3)) {
            frame_print (raw_frame, rawwords, 1);
        }
#endif
#endif
        // fill frame descriptor, if available
        if (frameset->count < ML_FRAMESET_NUMDESC/*frameset->maxdesc*/) {
            frameset->fdesc[frameset->count].rawdata = raw_frame;
            frameset->fdesc[frameset->count].len = rawlen;
            frameset->count++;
        } else {
            qWarning() << "Not enough frame descriptors";
            break;
        }

        // frame_warning_count += frame_warning;
        frame_count++;
        p += hdrlen;
    } // while()
    return 0;
}

int MlinkDeviceInterfaceBase::ml_get_rxb ()
{
    return (rxb);
}

int MlinkDeviceInterfaceBase::ml_get_txb ()
{
    return (txb);
}

int MlinkDeviceInterfaceBase::ml_get_rxf ()
{
    return (rxf);
}

int MlinkDeviceInterfaceBase::ml_get_txf ()
{
    return (txf);
}

void MlinkDeviceInterfaceBase::ml_clear_rxtx_counters ()
{
    rxb = 0;
    txb = 0;
    rxf = 0;
    txf = 0;
}

//------------------------------------------------------------------------------

int MlinkDeviceInterfaceBase::ml_get_event (mlink_event_t *ev, int timeout_ms)
{
    if (ml_fd == -1)
        return (-EINVAL);
    ml_poll_ev_op_t op;
    memset(&op, 0, sizeof(op));
    if (ml_fd == -1)
        return (-EINVAL);
    op.timeout_ms = timeout_ms;
    op.mask=0;
    if ((ioctl (ml_fd, ML_IOC_GET_EVENT, &op))) {
        qWarning() << QString("ML_IOC_GET_EVENT failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    if (ev) *ev = op.mask;
    return 0;
}


int MlinkDeviceInterfaceBase::ml_read_port(int port, unsigned int offset, unsigned int *data)
{
    ml_port_op_t op;
    op.port = port;
    op.offset = offset;
    op.data = 0;

    if ((ioctl (ml_fd, ML_IOC_READ_PORT, &op))) {
        qWarning() << QString("ML_IOC_READ_PORT failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    if (data)
        *data = op.data;
    return 0;
}

int MlinkDeviceInterfaceBase::ml_write_port(int port, unsigned int offset, unsigned int data)
{
    ml_port_op_t op;
    op.port = port;
    op.offset = offset;
    op.data = data;

    if ((ioctl (ml_fd, ML_IOC_WRITE_PORT, &op))) {
        qWarning() << QString("ML_IOC_WRITE_PORT failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}


int MlinkDeviceInterfaceBase::ml_read_flasher(unsigned int offset, unsigned int *data)
{
    ml_flasher_op_t op;
    op.offset = offset;
    op.data = 0;

    if ((ioctl (ml_fd, ML_IOC_READ_FLASHER, &op))) {
        qWarning() << QString("ML_IOC_READ_FLASHER failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    if (data)
        *data = op.data;
    return 0;
}

int MlinkDeviceInterfaceBase::ml_write_flasher(unsigned int offset, unsigned int data)
{
    ml_flasher_op_t op;
    op.offset = offset;
    op.data = data;

    if ((ioctl (ml_fd, ML_IOC_WRITE_FLASHER, &op))) {
        qWarning() << QString("ML_IOC_WRITE_FLASHER failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

int MlinkDeviceInterfaceBase::ml_read_board_info(board_info_t *data)
{
    if(!data) return 0;

    memset(data, 0, sizeof (board_info_t));
    if ((ioctl (ml_fd, ML_IOC_READ_BOARD_INFO, data))) {
        qWarning() << QString("ML_IOC_READ_BOARD_INFO failed: %1").arg(ml_strerror (errno));
        return (-errno);
    }
    return 0;
}

} // namespace mlink
