
#include "regiomlink.h"

#include <QDebug>
#include <string>
#include <sstream>
#include <iomanip>
#include <cstring>
#include <cerrno>

#include "mlink.h"
#include "mlink/frame_types.h"
#include "mlink/frameset.h"
#include "mlink/mlink_frame_header.h"
#include "mlink/MlinkDeviceInterfaceBase.h"
#include "mlink/MlinkDeviceInterfaceDebug.h"
#include "exc.h"

namespace mlink {

    static const int send_timeout = 1000; // use -1 to disable wait, return immediately
    static const int default_buffer_size = 10000; // maximum number of packets

    RegIOMLink::RegIOMLink(const std::string &name)
        : dev(new MlinkDeviceInterfaceDebug()),
        txPacketSequenceNumber(0)
    {
        bufSizeLimit = default_buffer_size;

        if (name.empty())
            setDevName(defaultDevName());
        else
            setDevName(name);
//        try {
//            open();
//        }
//        catch (exc_hwif &) {}
    }

    RegIOMLink::~RegIOMLink()
    {
        close();
    }

    int RegIOMLink::open()
    {
        Q_ASSERT(dev != nullptr);
        if (dev == nullptr)
            return -ENOENT;
        int rc = dev->ml_open(devName.c_str());
        if (rc)
            return rc;
        unsigned long data;
        rc = dev->ml_get_driver_version(&data);
        if (rc)
            return -ENODEV;
        driverVersion.major = data >> 16;
        driverVersion.minor = data & 0xFFFF;
        return 0;
    }

    int RegIOMLink::close()
    {
        Q_ASSERT(dev != nullptr);
        if (dev == nullptr)
            return -ENOENT;
        return dev->ml_close();
    }

    bool RegIOMLink::isOpen() const
    {
        Q_ASSERT(dev != nullptr);
        if (dev == nullptr)
            return false;
        return dev->isOpen();
    }

    std::string RegIOMLink::defaultDevName()
    {
        return std::string("/dev/")+std::string(ML_DEVICE_NAME)+std::string("0");
    }

    void RegIOMLink::reset()
    {
        RegIOBase::reset();
        Q_ASSERT(dev != nullptr);
        if (dev == nullptr)
            return;
        txPacketSequenceNumber = 0;
        int rc = 0;
        {
            rc = dev->ml_reset();
        }
        if (rc) {
            throw exc_hwif(std::string("Device reset failed: ") + dev->ml_strerror(-rc));
        }
    }

    void RegIOMLink::rawSend(const RegIoPacket &packet)
    {
        Q_ASSERT(dev);
        if (!dev)
            return;
        int rc = dev->ml_send_packet (packet, send_timeout);
        if (rc) {
            QString str = QString("Failed frame send: %1").arg(dev->ml_strerror(-rc));
            throw exc_hwif(str.toStdString());
        }
    }

    RawPacketList RegIOMLink::getRxPackets(double timeout)
    {
        Q_ASSERT(dev != nullptr);
        if (dev == nullptr)
            return RawPacketList();
        rawRxBuf.clear();
        rawRxBuf.resize(ML_RX_BUFFER_WORDS, 0);
        const size_t bufSize = sizeof(rawRxBuf[0]) * rawRxBuf.size();
        //  memset(&rawRxBuf[0], 0, bufSize);
        // initialize frameset
        frameset_t frameset;// frameset.maxdesc = ML_FRAMESET_NUMDESC;
        frameset.count = 0;
        int ret = dev->ml_recv_data_frame_multi(&rawRxBuf[0], bufSize, &frameset, int(timeout*1000));
//        qDebug() << "ml_recv_data_frame_multi returned " << ret;
//        std::cerr.flush();
        if (ret != 0)
            return RawPacketList();
        // copy frame data
        RawPacketList rxBuf;
        rxBuf.reserve(frameset.count);
        for(size_t i=0; i<frameset.count; i++) {
            const struct framedesc_t *fdesc = &frameset.fdesc[i];
            const size_t datawords = fdesc->len;// - (ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS);
            const QByteArray rawPacket(reinterpret_cast<const char *>(fdesc->rawdata), 4*datawords);
//            qDebug() << "rawPacket size " << rawPacket.size();
            //            const int *frame_data = (const int *) &fdesc->rawdata[ML_FRAME_HEADER_WORDS];
            //            const frame_hdr_t *fh = (const frame_hdr_t *) fdesc->rawdata;
            //            PacketData data;
            //            data.assign(&frame_data[0], &frame_data[datawords]);
            //            RegIoPacket pkt(fh->type, fh->seq, fh->src, fh->dst, data);
            //            rxBuf.push_back(pkt);
            rxBuf.push_back(rawPacket);
        }
        return rxBuf;
    }

    double RegIOMLink::getDeviceThroughput() const
    {
        return 160e6;
    }

    unsigned long RegIOMLink::getMaxDataSize()
    {
        return ML_MAX_FRAME_DATA_WORDS;
    }
    bool RegIOMLink::isConnected()
    {
        if (!dev)
            return false;
        return dev->is_link_up();
    }

} // namespace mlink
