
#include "frame_util.h"

#include <cstdio>
#include <iostream>
#include <stdexcept>
#include <cassert>

#ifndef _MSC_VER
#include <zlib.h>
#endif

#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"

namespace mlink {

    MlinkFrame::MlinkFrame(const quint32 *frameData, size_t frameLength)
        : f(frameData), len(frameLength)
    {
    }

    bool MlinkFrame::isValid() const
    {
        return ML_FRAME_VALIDATE_OK == checkStructure();
    }

    quint32 MlinkFrame::ml_make_crc (const void *buf, int size)
    {
#if defined _MSC_VER || defined WIN32
        return 0; // FIXME
#else
        uLong crc = crc32(0L, Z_NULL, 0);
        crc = crc32 (crc, reinterpret_cast<const Bytef *>(buf), size);
        return crc;
#endif
    }

    void MlinkFrame::swap_endian (quint32 *frameData, size_t frameLength)
    {
        /*
  for(i=0;i<len;i++) buf[i] =
  ((buf[i] & 0xFF) << 24 ) |
  (((buf[i]>>8) & 0xFF) << 16 ) |
  (((buf[i]>>16) & 0xFF) << 8 ) |
  (((buf[i]>>24)));
*/
        for (size_t i = 0; i < frameLength; i++)
            frameData[i] = ((frameData[i] & 0xFFFF) << 16) | (((frameData[i] >> 16) & 0xFFFF));
    }

    int MlinkFrame::checkStructure() const
    {
        return checkStructure(f, len);
    }

    int MlinkFrame::checkStructure(const quint32 *frameData, size_t frameLength)
    {
        int result = ML_FRAME_VALIDATE_OK;
        const auto *fh = reinterpret_cast<const frame_hdr_t *>(frameData);

        if (fh->sync != ML_FRAME_SYNC)
            result |= ML_FRAME_VALIDATE_SYNC;
        if (fh->len != frameLength) {
            std::cerr << "fh->len " << fh->len << " != size " << frameLength;
            result |= ML_FRAME_VALIDATE_LEN;
        }
        if (frameLength > 1) {
            unsigned int mycrc;
            unsigned int crc = frameData[frameLength - 1];
            mycrc = ml_make_crc (frameData, (frameLength-1)*4);
            if (crc != mycrc)
                result |= ML_FRAME_VALIDATE_CRC;
        }
        return result;
    }


    // Print RAW frame buffer
    void MlinkFrame::frame_print (bool rx_not_tx) const
    {
        assert(f);

        const auto *hdr = reinterpret_cast<const frame_hdr_t *>(f);
        int i;
        int newlen = len;
        int datalen = len-4;
        int err;

        if (newlen > ML_MAX_FRAME_DATA_WORDS) newlen = ML_MAX_FRAME_DATA_WORDS;
        if (!f)
            return;

        //  printf("---------- Frame dump ----------\n");
        /*
  printf("----------- RAW ----------------\n");
  for (i=0;i<len;i++)
        printf("%08x ", f[i]);
  printf("\n");
  printf("----------- Decoded ------------\n");
*/
//        unsigned int crc = f[newlen-1];
        err=0;
        if (hdr->sync!=ML_FRAME_SYNC) err=1;
        if (hdr->src == 2) err=1;

        if (err)
            printf("%s Sync %04X %s, type %04X, len %4d%s, seq %04X, %04X %s %04X, crc %04X\n",
                   rx_not_tx?"< ":" >",
                   hdr->sync, (hdr->sync==ML_FRAME_SYNC)?"Ok":"XX",
                   hdr->type, hdr->len, (hdr->len==newlen)?"":" BAD", hdr->seq, hdr->src, rx_not_tx?"<-":"->", hdr->dst, f[newlen-1]);

        /*
  if (hdr->type == FRAME_TYPE_LOOPTEST) {
        for (i=0;i<datalen;i++) {
          unsigned int d = f[3+i];
          if (i%4==0) printf("%s [%4d]: ", rx_not_tx?"< ":" >", i);
          int j;
          for(j=0;j<2;j++) {
                unsigned int h;
                char *c = "";
                char *c0 = ANSI_NORM;
                h = (j%2)?(d&0xFFFF):(d>>16);
                if (h == ML_FRAME_SYNC)
                  c0 = ANSI_GREEN;
                printf("%s%04X%s ", c, h, c0);
          }
          printf(" ");
          if (i%4==3) printf("\n");
        }
        if ((datalen % 4) != 3)
          printf("\n");

  } else
*/
        for (i=0;i<datalen;i++) {
            quint32 d=f[3+i];
            printf("%s [%4d]: %08x", rx_not_tx?"< ":" >", i, d);
            if (hdr->type == FRAME_TYPE_CTRL_REQ)
                printf(" -- ctrl req -- %s addr %04X, data %04X", (d&0x80000000)?"read":"write", (d>>16)&0x7FFF, d&0xFFFF);
            if (hdr->type == FRAME_TYPE_CTRL_ACK)
                printf(" -- ctrl ack -- ");
            if (hdr->type == FRAME_TYPE_STAT)
                printf(" -- status -- ");
            if (hdr->type == FRAME_TYPE_FVME_REQ) {
                if (d & 0x40000000) { // handshake
                    printf(" -- fvme bus transaction --");
                } else {
                    printf(" -- fvme req -- %s addr %04X, data %04X", (d&0x80000000)?"read":"write", (d>>16)&0x7FFF, d&0xFFFF);
                }
            }
            printf("\n");
        }

        /*
  if (datalen>10) {
    for (i=0;i<5;i++)
                printf("%s [%4d]: %08x\n", rx_not_tx?"< ":" >",i, f[3+i]);
        printf("     ...     \n");
    for (i=datalen-5;i<datalen;i++)
                printf("%s [%4d]: %08x\n", rx_not_tx?"< ":" >",i, f[3+i]);
  } else {
    for (i=0;i<datalen;i++)
      printf("%s [%4d]: %08x\n", rx_not_tx?"< ":" >",i, f[3+i]);
  }
*/
        /*
  if (len>10) {
    for (i=0;i<5;i++)
                printf("[%4d]: %08x\n", i, f[i]);
        printf("     ...     \n");
    for (i=len-5;i<len;i++)
                printf("[%4d]: %08x\n", i, f[i]);
  } else {
    for (i=0;i<len;i++)
      printf("[%4d]: %08x\n", i, f[i]);
  }
*/
        //  printf("--------------------------------\n");
    }

    void MlinkFrame::print_raw_frame (const quint32 *frameData, size_t frameLength)
    {
        int ferr = checkStructure(frameData, frameLength);
        /*
  for (i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
    unsigned int d = f[i];
    printf ("[%4d]:%08x  ", i, d);
//    printf ("[%4d]: %08x", i, d);
//    printf ("\n");
  }
  printf ("\n");
*/
        fprintf (stderr, "--------------- Raw frame dump --------[%s]-\n", ferr ? " BAD " : " Ok  ");
        for (size_t i = 0; i < frameLength; i++) {
            quint32 d = frameData[i];

            int c = i % 4;
            if (c == 0)
                fprintf (stderr, "[%04X]:  ", (unsigned int)i);
            fprintf (stderr, "%08x  ", d);
            //    fprintf (stderr, "[%4d]: %08x", i, d);
            //    fprintf (stderr, "\n");
            if (c == 3)
                fprintf (stderr, "\n");
        }
        fprintf (stderr, "-----------------------------------------------\n");
    }

    void MlinkFrame::print_ctrlreq_frame () const
    {

        for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
            quint32 d = f[i];

            printf ("[%4u]: %08x", (unsigned int)i, d);
            printf (" -- ctrl req -- %s addr %04X, data %04X", (d & 0x80000000) ? "read" : "write", (d >> 16) & 0x7FFF, d & 0xFFFF);
            printf ("\n");
        }
    }

    void MlinkFrame::print_ctrlack_frame () const
    {
        for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
            quint32 d = f[i];

            printf ("[%4u]: %08x", (unsigned int)i, d);
            printf (" -- ctrl ack -- ");
            printf ("\n");
        }
    }

    void MlinkFrame::print_status_frame () const
    {

        for (size_t i = ML_FRAME_HEADER_WORDS; i < len - ML_FRAME_TRAILER_WORDS; i++) {
            quint32 d = f[i];

            printf ("[%4u]: %08x", (unsigned int)i, d);
            printf (" -- status -- ");
            printf ("\n");
        }
    }

} // namespace mlink
