//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include "mlinkdevice.h"

#include <QDebug>
#include <sstream>
#include <iomanip>
#include <cassert>
#include <stdexcept>
#include <sstream>

#include "mlink/regiomlink.h"
#include "regiopacket.h"
#include "mlinkpacketprotocol.h"
#include "mlink/frame_types.h"
#include "TextFormatter.h"

namespace mlink {

typedef std::runtime_error exc_dev;

const int MlinkHostAddress = 1;

MlinkDevice::MlinkDevice(QSharedPointer<RegIOMLink> p, quint16 address)
    : link(p),
    packetProtocol(new MlinkPacketProtocol()),
    myAddress(address)
{
    setDefaults();
}

MlinkDevice::~MlinkDevice()
{
    delete packetProtocol;
}

//--------------------------------------------------------------
// read 16-bit register
quint16 MlinkDevice::regRead(int regNumber)
{
    std::vector<quint16> rx = regReadBlk(regNumber, 1);
    assert(rx.size() == 1);
    if (rx.size() != 1)
        return 0;
    return rx[0];
}
// read two 16-bit registers to 32-bit word
quint32 MlinkDevice::regRead32(int regNumber)
{
    const std::vector<quint16> &rd = regReadBlk(regNumber, 2);
    assert(rd.size() == 2);
    if (rd.size() != 2)
        return 0;
    quint32 lo = rd[0];
    quint32 hi = rd[1];
    return (hi << 16) | lo;
}
// read four 16-bit registers to 64-bit word
quint64 MlinkDevice::regRead64(int regNumber)
{
    const std::vector<quint16> &rd = regReadBlk(regNumber, 4);
    assert(rd.size() == 4);
    if (rd.size() != 4)
        return 0;
    quint64 b0 = rd[0];
    quint64 b1 = rd[1];
    quint64 b2 = rd[2];
    quint64 b3 = rd[3];
    return (b3 << 48) | (b2 << 32) | (b1 << 16) | b0;
}
// write 16-bit register
void MlinkDevice::regWrite(int regNumber, quint16 data)
{
    std::vector<quint16> v(1, data);
    regWriteBlk(regNumber, v);
}
// write 32-bit word to two 16-bit registers, low word even, high word odd
void MlinkDevice::regWrite32(int regNumber, quint32 data)
{
    std::vector<quint16> v(2);
    v[0] = data & 0xFFFF;
    v[1] = data >> 16;
    regWriteBlk(regNumber, v);
}
// write 64-bit word to four 16-bit registers, low word even, high word odd
void MlinkDevice::regWrite64(int regNumber, quint64 data)
{
    std::vector<quint16> v(4);
    v[0] = data & 0xFFFF;
    v[1] = (data >> 16) & 0xFFFF;
    v[2] = (data >> 32) & 0xFFFF;
    v[3] = (data >> 48) & 0xFFFF;
    regWriteBlk(regNumber, v);
}
//--------------------------------------------------------------
// read regCount 16-bit registers starting from startReg
std::vector<quint16> MlinkDevice::regReadBlk(int startReg, size_t regCount)
{
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_CTRL_REQ;
    std::vector<quint16> result;
    result.reserve(regCount);
    size_t totalRead = 0;
    while(totalRead < regCount) {
        size_t toRead = std::min(regCount - totalRead, getMaxRegRequestSize());
        tx.data.resize(toRead);
        for(size_t i=0; i<toRead; i++)
            tx.data[i] = 0x80000000 | (((startReg+totalRead+i)&0x7FFF)<<16);
        const RegIoPacket &rx = ctrlExchangeSingle(tx);
        //Q_ASSERT(rx.data.size() == toRead);
        if (rx.data.size() != toRead)
            throw std::runtime_error("Protocol error");
        for(size_t i=0; i<toRead; i++)
            result.push_back(rx.data[i] & 0xFFFF);
        totalRead += toRead;
    }
    return result;
}
// read 16-bit registers at given addresses
std::vector<quint16> MlinkDevice::regReadBlk(std::vector<int> registers)
{
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_CTRL_REQ;
    std::vector<quint16> result;
    const size_t regCount = registers.size();
    result.reserve(regCount);
    size_t totalRead = 0;
    while(totalRead < regCount) {
        size_t toRead = std::min(regCount - totalRead, getMaxRegRequestSize());
        tx.data.resize(toRead);
        for(size_t i=0; i<toRead; i++)
            tx.data[i] = 0x80000000 | (((registers[totalRead+i])&0x7FFF)<<16);
        const RegIoPacket &rx = ctrlExchangeSingle(tx);
        for(size_t i=0; i<toRead; i++)
            result.push_back(rx.data[i] & 0xFFFF);
        totalRead += toRead;
    }
    return result;
}
// write 16-bit registers starting from startReg
void MlinkDevice::regWriteBlk(int startReg, const std::vector<quint16> &data)
{
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_CTRL_REQ;

    // limit request size
    size_t ds = std::min(data.size(), getMaxRegRequestSize());
    if (ds != data.size()) {
        // warning - request size limit
        std::ostringstream os;
        os << "Failed register read: request size "
           << data.size()
           << " limited to "
           << ds;
        throw exc_dev(os.str());
    }
    for(size_t i=0; i<data.size(); i++)
        tx.data.push_back( 0x00000000 | (((startReg+i)&0x7FFF)<<16) | (data[i] & 0xFFFF));

    ctrlExchangeSingle(tx);
}

/// evaluate register operations
RegOpVector MlinkDevice::regOpExec(const RegOpVector &data)
{
    RegOpVector result(data);
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_CTRL_REQ;

    // limit request size
    int ds = std::min(data.size(), (int)getMaxRegRequestSize());
    if (ds != data.size()) {
        // warning - request size limit
        std::ostringstream os;
        os << "Failed register read: request size "
           << data.size()
           << " limited to "
           << ds;
        throw exc_dev(os.str());
    }
    for(int i=0; i<data.size(); i++) {
        bool w = data[i].write;
        quint16 a = data[i].addr;
        quint16 d = data[i].data;
        tx.data.push_back(
                    (w?0x00000000:0x80000000)
                    | ((a & 0x7FFF)<<16)
                    | (d & 0xFFFF));
    }

    const RegIoPacket &rx = ctrlExchangeSingle(tx);
    Q_ASSERT((int)rx.data.size() == result.size());
    for(int i=0; i<std::min((int)rx.data.size(), result.size()); i++)
        result[i].data = rx.data[i] & 0x0000FFFF;
    return result;
}
//--------------------------------------------------------------
// read 32-bit word from EEPROM at word offset wordNumber
quint32 MlinkDevice::eepromRead(int wordNumber)
{
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_EEPROM_REQ;

    // request size = 1
    tx.data.push_back( 0x80000000 |(1<<22)| (wordNumber&0x7FFF) ); // up to 1Mbit (32k words)

    const RegIoPacket &rx = ctrlExchangeSingle(tx);
    bool ackOk = validateIoAck(tx, rx);
    if (!ackOk || (rx.data.size() != 2)) {
        std::ostringstream ost;
        ost << "Failed eeprom read: received " << rx.data.size() << " words, expected 2";
        throw exc_dev(ost.str());
    }
    assert(rx.data.size() == 2);
    return rx.data[1];
}
// write 32-bit word to EEPROM at word offset wordNumber
void MlinkDevice::eepromWrite(int wordNumber, quint32 data)
{
    //  std::vector<quint32> v(1, data);
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_EEPROM_REQ;

    // request size = 1
    tx.data.push_back( 0x00000000 |(1<<22)| (wordNumber&0x7FFF) ); // up to 1Mbit (32k words)
    tx.data.push_back( data );

    const RegIoPacket &rx = ctrlExchangeSingle(tx);
    if (rx.data.size() != 2) {
        std::ostringstream ost;
        ost << "Failed eeprom write: received " << rx.data.size() << " words, expected 2";
        throw exc_dev(ost.str());
    }
}

// read 32-bit word from RAM at word offset wordNumber
quint32 MlinkDevice::memRead(int wordNumber)
{
    const std::vector<quint32> &v = memReadBlk(wordNumber, 1);
    assert(v.size() == 1);
    return v[0];
}
// write 32-bit word to RAM at word offset wordNumber
void MlinkDevice::memWrite(int wordNumber, quint32 data)
{
    std::vector<quint32> v(1, data);
    memWriteBlk(wordNumber, v);
}
// read wordCount 32-bit words from RAM starting from startWord offset
std::vector<quint32> MlinkDevice::memReadBlk(int startWord, size_t wordCount)
{
    std::vector<quint32> result;
    result.reserve(wordCount);
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_RAM_REQ;
    const size_t txSz = 1;
    tx.data.resize(txSz);
    const size_t rxHdrSz = 1;

    size_t totalRead = 0;
    while(totalRead < wordCount) {
        size_t blockSize = std::min(wordCount - totalRead, getMaxRamRequestSize());
        tx.data[0] = ( 0x80000000 |((blockSize&0x1FF)<<22)| ((startWord+totalRead)&0x3FFFFF) );
        const size_t rxSz = rxHdrSz+blockSize;

        const RegIoPacket &rx = ctrlExchangeSingle(tx);
        if (rx.data.size() != rxSz) {
            std::ostringstream ost;
            ost << "Failed mem read: received " << rx.data.size() << " words, expected " << rxSz;
            throw exc_dev(ost.str());
        }
        for(size_t i=0; i<blockSize; i++)
            result.push_back(rx.data[rxHdrSz+i]);
        totalRead += blockSize;
    }
    return result;
}
// write 32-bit words to RAM starting from startWord offset
void MlinkDevice::memWriteBlk(int startWord, const std::vector<quint32> &data)
{
    RegIoPacket tx;
    tx.src = MlinkHostAddress;
    tx.dst = myAddress;
    tx.type = FRAME_TYPE_RAM_REQ;
    const size_t txHdrSz = 1;
    const size_t wordCount = data.size();

    size_t totalWrite = 0;
    while(totalWrite < wordCount) {
        size_t blockSize = std::min(wordCount - totalWrite, getMaxRamRequestSize());
        const size_t txSz = txHdrSz+blockSize;
        const size_t rxSz = txSz;
        tx.data.resize(txSz);
        tx.data[0] = ( 0x00000000 |((blockSize&0x1FF)<<22)| ((startWord+totalWrite)&0x3FFFFF) );

        for(size_t i=0; i<blockSize; i++)
            tx.data[txHdrSz+i] = data[totalWrite+i];

        RegIoPacket const &rx = ctrlExchangeSingle(tx);
        if (rx.data.size() != rxSz) {
            std::ostringstream ost;
            ost << "Failed mem write: received " << rx.data.size() << " words, expected " << rxSz;
            throw exc_dev(ost.str());
        }
        totalWrite += blockSize;
        //	debug(0, "w1: wordCount=%d, blockSize=%d, totalWrite=%d", wordCount, blockSize, totalWrite);
    }
}
// maximum number of data words for single RAM read/write operation
size_t MlinkDevice::getMaxRamRequestSize() const
{
    // limit request size
    return std::min(RegIOMLink::getMaxDataSize()-1, 0x1FFul); // 2^9 words maximum
}

// maximum number of 16-bit registers for single register read/write operation
size_t MlinkDevice::getMaxRegRequestSize() const
{
    return RegIOMLink::getMaxDataSize(); // one command word for one register
}

void MlinkDevice::detect()
{
    onlineState = false;
    if (!enableState) return;
    qDebug() << QString("MlinkDevice: Trying address %1").arg(getAddress(), 4, 16, QChar('0'));
    try {
        onlineState = true;
        // Implement me
        //    bool state = readId();
        //    onlineState = state;
        //    log_info("Found BPM device at 0x%04X, serial %08X", getAddress(), getDeviceSerial());
    }

    catch (std::exception &e) {
        qDebug() << QString("MlinkDevice::detect %1").arg(e.what());
    }
}

void MlinkDevice::setDefaults()
{
    onlineState = false;
    enableState = true;
    masterState = false;
}

bool MlinkDevice::validateIoAck(const RegIoPacket &tx, const RegIoPacket &rx, const std::string &description) const
{
    std::string preamble(description);
    if (preamble.empty())
        preamble = "I/O operation";
    size_t exp_rx_size = MlinkPacketProtocol::ackSize(tx);
    unsigned int exp_rx_type = MlinkPacketProtocol::ackType(tx);
    unsigned int exp_rx_src = tx.dst;
    unsigned int exp_rx_dst = tx.src;
    bool ok = true;
    std::ostringstream ost;
    ost << "Communication error" << std::endl;
    if (rx.type != exp_rx_type) {
        ost << preamble << ": " << "Bad acknowledge type, received 0x" << std::hex << rx.type << ", expected 0x" << exp_rx_type << std::dec << std::endl;
        ok = false;
    }
    // exclude broadcast
    if (tx.dst != 0xFEFE && rx.src != exp_rx_src) {
        ost << preamble << ": " << "Bad acknowledge src addr, received 0x" << std::hex << rx.src << ", expected 0x" << exp_rx_src << std::dec << std::endl;
        ok = false;
    }
    if (rx.dst != exp_rx_dst) {
        ost << preamble << ": " << "Bad acknowledge dst addr, received 0x" << std::hex << rx.dst << ", expected 0x" << exp_rx_dst << std::dec << std::endl;
        ok = false;
    }
    // validate size for known frame types
    if ( (tx.type == FRAME_TYPE_LOOPTEST)
         || (tx.type == FRAME_TYPE_CTRL_REQ)
         || (tx.type == FRAME_TYPE_EEPROM_REQ)
         || (tx.type == FRAME_TYPE_RAM_REQ)
         ) {
        if (rx.data.size() != exp_rx_size) {
            ost << preamble << ": " << "Bad acknowledge size, received " << rx.data.size() << " words, expected " << exp_rx_size << std::endl;
            if (rx.data.size() < exp_rx_size)
                ok = false;
        }
        // [ temporarily disabled
        //        if (rx.seq != tx.seq) {
        //            ost << preamble << ": " << "Bad sequence number, received " << rx.seq << ", expected " << tx.seq << std::endl;
        //            ok = false;
        //        }
        // temporarily disabled ]

    }
    // validate CTRL request contents
    if (ok && (tx.type == FRAME_TYPE_CTRL_REQ)) {
        for(size_t i=0; i<tx.data.size(); i++) {
            const PacketRawDataType txData = tx.data[i];
            const PacketRawDataType rxData = rx.data[i];
            const bool isRead = 0x80000000u & txData;
            unsigned int mask = isRead ? 0xFFFF0000u : 0xFFFFFFFFu;
            if ((rxData & mask) != (txData & mask)) {
                ost << preamble << ": " << "Bad acknowledge data, sent "
                    << std::hex << txData << " received " << rxData << std::endl;
                ok = false;
            }
        }
    }
    if (!ok) {
        //        tx.dump();
        //        rx.dump();
        ost << "TX: " << tx << std::endl;
        ost << "RX: " << rx << std::endl;
        ost << std::endl;
        qWarning() << "validateIoAck: " << QString::fromStdString(ost.str());
        //            throw exc_dev(ost.str());
    }
    return ok;
}

PacketList MlinkDevice::ctrlExchange(const RegIoPacket &tx)
{
    PacketList pl;
    pl.push_back(tx);
    return ctrlExchange(pl);
}

const int retry_attempts = 0; // ******************* FIXME ******************* //

PacketList MlinkDevice::ctrlExchange(const PacketList &tx)
{
    Q_ASSERT(link);
    if (!link)
        throw std::runtime_error("MlinkDevice::ctrlExchange(): not connected");
    int retry_count = 0;
    while (true) {
        try {
            link->sendPackets(tx);
//            unsigned int expectRxSize = 0;
            int expRxPacketCount = 0;
            for(PacketList::const_iterator it=tx.begin(); it!=tx.end(); ++it) {
//                expectRxSize += packetProtocol->ackSize(*it);
                expRxPacketCount++;
            }
            const PacketList &rx = link->getCtrlPackets(expRxPacketCount, MlinkPacketProtocol::getTimeout());
            return rx;
        }
        catch (std::exception &e) {
            retry_count++;
            if (!link->isConnected()) {
                throw std::runtime_error("MlinkDevice::ctrlExchangeSingle(): link down");
            }
            qWarning() << QString("ctrlExchange failed (%1)").arg(e.what());
            qDebug() << "TX was: " << tx;
            if (retry_count > retry_attempts) {
                qWarning() << QString("giving up after %1 attempts").arg(retry_count);
                throw;
            }
            qDebug() << QString("Retry #%1").arg(retry_count);
        }
    }
    return PacketList();
}

RegIoPacket MlinkDevice::ctrlExchangeSingle(const RegIoPacket &tx_in)
{
    Q_ASSERT(link);
    if (!link)
        throw std::runtime_error("MlinkDevice::ctrlExchangeSingle(): not initialized");

    int retry_count = 0;
    int seq = nextTxPacketSequenceNumber();
    while (true) {
        RegIoPacket tx(tx_in);
        tx.seq = seq;
        try {
            link->rawSend(tx);
            double random_delay = 1.0/*MlinkPacketProtocol::random_delay(retry_count)*/;
            //                const PacketList &rxList = link.getCtrlPackets(1, MlinkPacketProtocol::getTimeout() * random_delay);
            const PacketList &rxList = link->getCtrlPacketBySeq(tx.seq, MlinkPacketProtocol::getTimeout() * random_delay);
            // FIXME: sequence number check is disabled
            /*
                bool found = false;
                for(PacketList::const_iterator it = rxList.begin(); it != rxList.end(); ++it) {
                    const RegIoPacket &rx = *it;
                    if (rx.seq == tx.seq) {
                        found = true;
                        bool ackOk = validateIoAck(tx, rx);
                        //                    qDebug() << "seq " << rx.seq << "Ok";
                        return rx;
                    }
                }
*/
            if (rxList.isEmpty()) {
                throw std::runtime_error("MlinkDevice::ctrlExchangeSingle(): Receive timeout");
            }
            const RegIoPacket &rx = rxList.at(0);
            bool ackOk = validateIoAck(tx, rx);
            if (!ackOk) {
                throw std::runtime_error("MlinkDevice::ctrlExchangeSingle(): Communication error");
            }
            return rx;
        }
        catch (std::runtime_error &e) {
            retry_count++;
            if (!link->isConnected()) {
                throw std::runtime_error("MlinkDevice::ctrlExchangeSingle(): link down");
            }
            {
                if (retry_count > MlinkPacketProtocol::getMaxRetryCount()) {
                    qWarning() << QString("ctrlExchangeSingle failed after %1 attempts: %2").arg(retry_count).arg(e.what());
                    throw;
                }
            }
//            qDebug() << QString("Retry #%1").arg(retry_count);
        }
    }
    return RegIoPacket();
}

} // namespace mlink
