//
//    Copyright 2010-2019 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef REGIO_MLINKDEVICE_H
#define REGIO_MLINKDEVICE_H

#include <cstddef>
#include <vector>
#include <string>

#include <QSharedPointer>
#include <QVector>

namespace mlink {
class RegOp
{
public:
#if (QT_VERSION < QT_VERSION_CHECK(5, 13, 0))
    RegOp() = default; // required for QVector
#endif
    RegOp(bool in_write, quint16 in_addr, quint16 in_data)
        : write(in_write), addr(in_addr), data(in_data) {}
    bool write = false;
    quint16 addr = 0;
    quint16 data = 0;
};

class RegRead : public RegOp
{
public:
    explicit RegRead(quint16 in_addr) : RegOp(false, in_addr, 0) {}
};

class RegWrite : public RegOp
{
public:
    RegWrite(quint16 in_addr, quint16 in_data) : RegOp(true, in_addr, in_data) {}
};

class RegOpVector: public QVector<RegOp>
{
public:
};

class RegIoPacket;
class PacketList;
class MlinkPacketProtocol;
class RegIOMLink;

class MlinkDevice {
public:
    MlinkDevice(const MlinkDevice &) = delete; // no copy
    MlinkDevice &operator=(const MlinkDevice &) = delete; // no copy
    MlinkDevice(QSharedPointer<RegIOMLink> p, quint16 address);

    virtual ~MlinkDevice();
    /// read 16-bit register
    quint16 regRead(int regNumber);

    /// read two 16-bit registers to 32-bit word
    quint32 regRead32(int regNumber);

    /// read four 16-bit registers to 64-bit word
    quint64 regRead64(int regNumber);

    /// write 16-bit register
    void regWrite(int regNumber, quint16 data);

    /// evaluate register operations
    RegOpVector regOpExec(const RegOpVector &data);

    /// write 32-bit word to two 16-bit registers, low word even, high word odd
    void regWrite32(int regNumber, quint32 data);

    /// write 64-bit word to four 16-bit registers, low word even, high word odd
    void regWrite64(int regNumber, quint64 data);

    /// read regCount 16-bit registers starting from startReg
    std::vector<quint16> regReadBlk(int startReg, size_t regCount);

    /// read 16-bit registers at given addresses
    std::vector<quint16> regReadBlk(std::vector<int> registers);

    /// write 16-bit registers starting from startReg
    void regWriteBlk(int startReg, const std::vector<quint16> &data);

    /// read 32-bit word from EEPROM at word offset wordNumber
    quint32 eepromRead(int wordNumber);

    /// write 32-bit word to EEPROM at word offset wordNumber
    void eepromWrite(int wordNumber, quint32 data);

    /// read 32-bit word from RAM at word offset wordNumber
    quint32 memRead(int wordNumber);

    /// write 32-bit word to RAM at word offset wordNumber
    void memWrite(int wordNumber, quint32 data);

    /// read wordCount 32-bit words from RAM starting from startWord offset
    std::vector<quint32> memReadBlk(int startWord, size_t wordCount);

    /// write 32-bit words to RAM starting from startWord offset
    void memWriteBlk(int startWord, const std::vector<quint32> &data);

    /// maximum number of data words for single RAM read/write operation
    size_t getMaxRamRequestSize() const;

    /// maximum number of 16-bit registers for single register read/write operation
    size_t getMaxRegRequestSize() const;

    /// device address
    int getAddress() const { return myAddress; }
    void setAddress(int address) { myAddress = address; }
    void enable() { enableState = true; }
    void disable() { enableState = false; }
    void setMaster(bool state = true) { masterState = state; }
    bool isEnabled() const { return enableState; }
    bool isOnline() const { return onlineState; }
    bool isMaster() const { return masterState; }
    virtual void detect();
    virtual void setDefaults();
    virtual bool validateIoAck(const RegIoPacket &regReadBlk_txPacket, const RegIoPacket &rx, const std::string &description = std::string()) const;

    //    int64_t readData(char *data, int64_t maxSize);

protected:
    virtual RegIoPacket ctrlExchangeSingle(const RegIoPacket &); ///< Send, receive and validate control packet
    PacketList ctrlExchange(const RegIoPacket &);  // ctrlExchange(PacketList) wrapper
    virtual PacketList ctrlExchange(const PacketList &); ///< Send, receive and validate control packets
    bool hardwareAccessAllowed() const { return enableState && onlineState; }
    QSharedPointer<RegIOMLink> link;
    MlinkPacketProtocol * const packetProtocol;
    bool onlineState = false;
    bool enableState = false;
    bool masterState = false;

private:
    int nextTxPacketSequenceNumber() { return txPacketSequenceNumber++; }
    quint16 myAddress = 0;
    int txPacketSequenceNumber = 0;
};

} // namespace mlink

#endif /* REGIO_MLINKDEVICE_H */
