
#include "regiobase.h"

#include <QDebug>
#include <stdexcept>
#include <sstream>
#include <iostream>
#include <algorithm>

#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "regio/regiopacket.h"
#include "util/gettimeofday.h"
#include "mlink/exc.h"
#include "frame_util.h"
#include "TextFormatter.h"

static const bool disable_seq_check = true;

namespace mlink {

    static double getCurrentTimestamp()
    {
        struct timeval tv;
        gettimeofday(&tv, nullptr);
        return tv.tv_sec + 1e-6*tv.tv_usec;
    }

    RegIOBase::RegIOBase(const std::string &name)
        : devName(name),
        bufSizeLimit(0) // no limit
    {
    }

    RegIOBase::~RegIOBase()
    {
    }

    std::string RegIOBase::getDevName() const {
        return devName;
    }

    DeviceDriverVersion RegIOBase::getDriverVersion() const
    {
        return driverVersion;
    }

    bool RegIOBase::packetTypeIsData(int type)
    {
        if (type == FRAME_TYPE_ADC_DATA) return true;
        if (type == FRAME_TYPE_STREAM) return true;
        if (type == FRAME_TYPE_STAT) return true;
        return false;
    }

    void RegIOBase::sendPackets(const PacketList &pl)
    {
        for(PacketList::const_iterator it = pl.begin(); it != pl.end(); ++it) {
            rawSend(*it);
        }
    }

    void RegIOBase::reset()
    {
        rxBufCtrl.clear();
        rxBufData.clear();
    }

    qint64 RegIOBase::getRxBufSizeLimit() const
    {
        return 1024*1024*10; // 10 MB
    }

    double RegIOBase::getDeviceThroughput() const
    {
        return 1e6; // default value
    }

    double RegIOBase::getPollInterval() const
    {
        return getRxBufSizeLimit() / getDeviceThroughput();
    }

    void RegIOBase::receive()
    {
//        qDebug() << "Receive...";
        const int rxBufCtrlStartSize = 0; //rxBufCtrl.size();
        const double timeLimit = 1.1 * getPollInterval();
        double timeBegin = getCurrentTimestamp();
        while(true) {
            //        double timeRxBegin = getCurrentTimestamp();
            const RawPacketList &rawPacketList = getRxPackets(timeLimit);
            demuxRxBuf(rawPacketList);
            if (rxBufCtrl.size() > rxBufCtrlStartSize) {
//                qDebug("Control packet received, interrupting receive()");
//                            std::cerr.flush();
                break;
            }
            //        if (rxBufData.size() + ML_MAX_FRAME_DATA_WORDS > rxBufDataSizeLimit) break;
            //        if (rxBufData.size() > getRxBufSizeLimit()) {
            //            std::cerr << "RegIOBase::receive(): rxBufDataSizeLimit" << std::endl;
            //            std::cerr.flush();
            //            break;
            //        }
            double timeNow = getCurrentTimestamp();
            if (timeNow - timeBegin > timeLimit) {
                //            std::cerr << "RegIOBase::receive(): timeLimit" << std::endl;
                //            std::cerr.flush();
                break;
            }
            //        if (timeNow - timeRxBegin < 0.5*timeLimit) {
            //            // wait some time
            //            usleep(int(1e6*(timeLimit - (timeNow - timeRxBegin))));
            //        }
        }
//        qDebug() << "Receive complete";
    }

    void RegIOBase::demuxRxBuf(const RawPacketList &rxBuf)
    {
        // sort incoming packets by data and control
        bool ctrlOverflow = false;
        bool dataOverflow = false;
        for (auto & rawpkt : rxBuf) {
            RegIoPacket pkt(rawpkt);
//            qDebug() << pkt;
            bool data = packetTypeIsData(pkt.type);
            if (data) {
                if ((bufSizeLimit > 0) && (rxBufData.size() > bufSizeLimit)) {
                    if (!dataOverflow) {
//                        std::ostringstream os;
//                        os << "rxBufData overflow: size " << rxBufData.size() << ", limit " << bufSizeLimit;
//                        throw exc(os.str());
                        dataOverflow = true;
                    }
                } else {
                    rxBufData.push_back(rawpkt);
                    //                std::cerr << pkt << std::endl;
//                    // re-assemble packet // FIXME
//                    rxBufData.push_back((quint32(ML_FRAME_SYNC)<< 16) | pkt.type);
//                    rxBufData.push_back((quint32(pkt.data.size()+4)<< 16) | pkt.seq);
//                    rxBufData.push_back((quint32(pkt.dst)<< 16) | pkt.src);
//                    rxBufData.insert(rxBufData.end(), pkt.data.begin(), pkt.data.end());
//                    quint32 crc = 0; // MlinkFrame::ml_make_crc(&pkt.data[0], (pkt.data.size()+3)*sizeof(PacketRawDataType));
//                    rxBufData.push_back(crc);
//                    qDebug() << "demuxRxBuf: data packet received";
                    //                std::cerr << " Added data packet 0x" << std::hex << pkt.type
                    //                        << " len " << std::dec << 4+pkt.data.size()
                    //                        << ", total size " << std::dec << rxBufData.size()
                    //                        << " 32-bit words" << std::endl;
                    //                std::cerr.flush();
                }
            } else {
                if (bufSizeLimit && (rxBufCtrl.size() > bufSizeLimit)) {
                    if (!ctrlOverflow) {
                        ctrlOverflow = true;
                        throw exc("rxBufCtrl overflow");
                    }
                } else {
                    rxBufCtrl.push_back(pkt);
                    //                std::ostringstream ost;
                    //                ost << "RegIOBase::demuxRxBuf(): received " << pkt;
                    //                log_info(ost.str());
                }
            }
        }
        if (dataOverflow) {
            std::ostringstream os;
            os << "rxBufData overflow: limit " << bufSizeLimit;
            throw exc(os.str());
        }
    }

    size_t RegIOBase::getCtrlPacketCount() const
    {
        return rxBufCtrl.size();
    }

    size_t RegIOBase::getDataSize() const
    {
        return rxBufData.size();
    }

    /// get control packets with specified sequence number, drop other packets
    PacketList RegIOBase::getCtrlPacketBySeq(int seq, double timeout)
    {
//        bool found = false;
        double tv1 = getCurrentTimestamp();
        PacketList pl;
        int receivedFrameCount = 0;
        int tryCount = 0;
        while (true)
        {
            try {
                receive();
                tryCount++;
            }
            catch (std::exception &e) {
                qWarning() << QString("Exception in receive(): %1").arg(e.what());
            }
            if (rxBufCtrl.isEmpty()) {
                if (!isConnected()) {
                    qWarning() << QString("RegIOBase::getCtrlPacketBySeq(): link down");
                    break;
                }
            }
            //        qDebug() << rxBufCtrl.size();
            for (const auto & pkt : rxBufCtrl) {
                receivedFrameCount++;
                //            qDebug() << "seq = " << it->seq << ", expected " << seq;

                if (disable_seq_check || (pkt.seq == seq)) {
                    //                qDebug() << "Ok: rxBufCtrl=" << rxBufCtrl;
//                    found = true;
//                    break;
                    pl.push_back(pkt);
                } else {
                    qWarning() << QString("Unexpected control ack #%1: received seq %2, expected %3").arg(receivedFrameCount).arg(pkt.seq).arg(seq);
                    qDebug() << pkt;
                }
            }
//            pl = rxBufCtrl;
            rxBufCtrl.clear();
            if (!pl.isEmpty()) break;
            double tv2 = getCurrentTimestamp();
            double dt = tv2 - tv1;
            if (dt > timeout) {
                qWarning().noquote()
                    << "getCtrlPacketBySeq:"
                    << QString("Protocol timeout, try %1, time %2, seq %3")
                           .arg(tryCount).arg(dt).arg(seq)
                    << (receivedFrameCount ? QString(", skipped %1 frames").arg(receivedFrameCount) : QString());
                //            qDebug() << "rxBufCtrl=" << rxBufCtrl;
                break;
            }
            //        qApp->processEvents();
            if (receivedFrameCount > 10) {
            qWarning() << QString("No received frame available with seq = %1").arg(seq);
            break;
            }
        }
        return pl;
    }

    PacketList RegIOBase::getCtrlPackets(size_t packetCount, double timeout)
    {
        double tv1 = getCurrentTimestamp();
        while(getCtrlPacketCount() < packetCount)
        {
            try {
                receive();
            }
            catch (std::exception &e) {
                qWarning() << QString("Exception in receive(): %1").arg(e.what());
            }
            if (getCtrlPacketCount() >= packetCount) break;
            double tv2 = getCurrentTimestamp();
            double dt = tv2 - tv1;
            if (dt > timeout) {
                qWarning() << "Protocol timeout, dt=" << dt << " s (getControlPackets(" << packetCount << ", " << timeout << "))";
                qDebug() << "rxBufCtrl=" << rxBufCtrl;
                qDebug() << "rxBufData.size=" << rxBufData.size();
                throw std::runtime_error("Timeout");
            }
        }
        const PacketList pl(rxBufCtrl);
        rxBufCtrl.clear();
        return pl;
    }

} // namespace mlink
