//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include "regiopacket.h"

#include <QTextStream>
#include <QDebug>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include <iterator>

#include "mlink/mlink_frame_header.h"

namespace mlink {

    void RegIoPacket::dump(const std::string &description) const
    {
        std::string s;
        if (! description.empty())
            s = " (" + description + ")";
        std::cerr << "--- Packet dump begin" << s << " ---" << std::endl;
        std::cerr << (*this);
        std::cerr << "--- Packet dump end  " << s << " ---" << std::endl;
    }

    std::ostream &operator<< (std::ostream & os, const PacketRawData &p)
    {
        std::ostringstream sstr;
        sstr << "PacketRawData[" << p.size() << "] = {";
        sstr << std::hex;
        //    std::copy(p.begin(), p.end(), std::ostream_iterator<PacketRawDataType>(sstr, " "));
        size_t n = 0;
        for (auto it : p) {
            if (n%8==0) sstr << std::endl << "  ";
            sstr << std::setw(8) << std::setfill('0') << it << " ";
            n++;
        }
        if (n%8!=0) sstr << std::endl;
        sstr << "};";
        os << sstr.str();
        return os;
    }

    std::ostream &operator<< (std::ostream & os, const RegIoPacket &p)
    {
        std::ostringstream sstr;
        sstr << std::hex;
        sstr << "type 0x" << std::setw(4) << std::setfill('0') << p.type
                << ", seq 0x" << std::setw(4) << std::setfill('0') << p.seq << ", ";
        if (p.src < p.dst) {
            sstr << std::setw(4) << std::setfill('0') << p.src
                    << "->" << std::setw(4) << std::setfill('0') << p.dst;
        } else {
            sstr << std::setw(4) << std::setfill('0') << p.dst
                    << "<-" << std::setw(4) << std::setfill('0') << p.src;
        }
        sstr << ", " << p.data;
        os << sstr.str();
        return os;
    }

    QTextStream &operator<< (QTextStream & os, const RegIoPacket &p)
    {
        std::ostringstream sstr;
        sstr << p;
        const std::string &stdstr = sstr.str();
        const QString &qs = QString::fromStdString(stdstr);
        os << qs;
        return os;
    }


    RegIoPacket::RegIoPacket(const QByteArray &buf)
    {
        const int *frame_data = reinterpret_cast<const int *>(&buf.constData()[4*ML_FRAME_HEADER_WORDS]);
        const auto *fh = reinterpret_cast<const frame_hdr_t *>(buf.constData());
        const int datawords = buf.size() / 4 - ML_FRAME_HEADER_WORDS;
        type = fh->type;
        seq = fh->seq;
        src = fh->src;
        dst = fh->dst;
        if (datawords > 0)
            data.assign(&frame_data[0], &frame_data[datawords-1]);
//        qDebug() << "RegIoPacket: " << *this;
    }

    QByteArray RegIoPacket::toByteArray() const
    {
        QByteArray buf;
        frame_hdr_t fh;
        fh.type = type;
        fh.sync = ML_FRAME_SYNC;
        fh.len = data.size();
        fh.seq = seq;
        fh.src = src;
        fh.dst = dst;
        QByteArray rawHdr(reinterpret_cast<const char *>(&fh), sizeof(fh));
        buf.append(rawHdr);
        QByteArray rawData(reinterpret_cast<const char *>(&data[0]), 4*data.size());
        buf.append(rawData);
        QByteArray rawCrc(4, 0);
        buf.append(rawCrc);
        return buf;
    }

    QDebug operator<< (QDebug os, const RegIoPacket &p)
    {
        std::ostringstream sstr;
        sstr << p;
        const std::string &stdstr = sstr.str();
        os << QString::fromStdString(stdstr);
        return os;
    }

} // namespace mlink
