/*
**    Copyright 2010-2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MREGDEVICE_H
#define MREGDEVICE_H

#include "fw_version.h"
#include "regio/mlinkdevice.h"

namespace mlink {

class DeviceAttr {
public:
    QString getFirmwareVersionStr() const;

    quint16 device_id = 0;
    bool id_chip_present = false;
    quint64 device_serial = 0;
    double temp = 0;
    double clock = 0;
    fw_version_t fw_version;
    bool mstreamMultiAck = false;
    quint8 mstreamHwBufSize = 8; // notice me
};

class MregDevice : public MlinkDevice
{
public:
    enum {
        REG_DEVICE_ID = 0x42
    };
    MregDevice(RegIOBase &p, int address);
    virtual void print() const;
    quint16 detect() override;
    DeviceAttr getAttr() const { return attr; }

    virtual double readTemp() = 0;
    virtual quint64 readTimestampCurrent() = 0;
    virtual quint64 readTimestampRun() = 0;
    virtual quint64 readTimestampTrig() = 0;
    virtual quint32 writeLiveMagic(quint32 value) = 0;
    virtual quint32 readLiveMagic() = 0;
    quint16 readDeviceId(bool checkAccess = true);
    static double convertOneWireTemp(quint16 rawTemp);

protected:
    quint16 getDeviceId() const { return attr.device_id; }
    quint64 getDeviceSerial() const { return attr.device_serial; }
    virtual quint64 readOneWireId(bool checkAccess = true) = 0;
    virtual bool readFirmwareId() = 0;
    virtual bool readMStreamCfg() = 0;

    DeviceAttr attr;

};

class MregDeviceDummy : public MregDevice
{
public:
    MregDeviceDummy(RegIOBase &link, int address): MregDevice(link, address) {}
    ~MregDeviceDummy() override = default;
    double readTemp() override {return -255;}
    quint64 readTimestampCurrent() override {return 0;}
    quint64 readTimestampRun() override {return 0;}
    quint64 readTimestampTrig() override {return 0;}
    quint32 writeLiveMagic(quint32 value) override {Q_UNUSED(value); return 0;}
    quint32 readLiveMagic() override {return 0;}


protected:
//    virtual quint16 readDeviceId();
    quint64 readOneWireId(bool checkAccess = true) override {Q_UNUSED(checkAccess); return 0;}
    bool readFirmwareId() override {return false;}
    bool readMStreamCfg() override {return false;}
};

class MregDeviceBpm : public MregDevice
{
public:
    enum {
        REG_DEVICE_ID   = 0x42,
        REG_LIVE_MAGIC  = 0x43,
        REG_MSTREAM_CFG  = 0x4F,
        REG64_SERIAL_NUM = 0x50,

        REG_TEMPERATURE = 0x54, // 84

        REG_FW_VER = 0x56, // 86
        REG_FW_REV = 0x57, // 87

        REG_LINK_ERR_CNT_1 = 0x58,
        REG_LINK_ERR_CNT_2 = 0x59,
        REG_LINK_ERR_CNT_3 = 0x5A,
        REG_LINK_CRC_CNT_1 = 0x5B,
        REG_LINK_CRC_CNT_2 = 0x5C,
        REG_LINK_CRC_CNT_3 = 0x5D,

        REG_LINK_STATUS = 0x61,

        REG64_CNT_NCU = 0x70,
        REG64_CNT_KCU = 0x74,
        REG64_CURRENT_TIMESTAMP = 0x78
    };

    MregDeviceBpm(RegIOBase &p, int address);
    ~MregDeviceBpm() override = default;
    double readTemp() override;
    quint64 readTimestampCurrent() override;
    quint64 readTimestampRun() override;
    quint64 readTimestampTrig() override;
    quint32 writeLiveMagic(quint32 value) override;
    quint32 readLiveMagic() override;

protected:
//    virtual quint16 readDeviceId();
    quint64 readOneWireId(bool checkAccess = true) override;
    bool readFirmwareId() override;
    bool readMStreamCfg() override;
};

class MregDeviceAdcm : public MregDevice
{
public:
    enum {
        // Control registers
        ADCM16_CTRL_REG           = 0x40,
        ADCM16_CTRL_DEFAULT     = 0x0000,

        ADCM16_WIN_REG            = 0x41,
        ADCM16_STATUS_REG         = 0x42, /* read-only */
        ADCM16_WIN_DEFAULT      = 0xFFFF,

        ADCM16_TRIGMULT_REG       = 0x43,
        ADCM16_SERIAL_ID_HI_REG   = 0x46, /* read-only */
        ADCM16_LIVE_MAGIC         = 0x48,
        ADCM16_CH_DPM_KS_REG      = 0x4A, /* read-only */
        ADCM16_TEMPERATURE_REG    = 0x4B, /* read-only */
        ADCM16_FW_VER_REG         = 0x4C, /* read-only */
        ADCM16_FW_REV_REG         = 0x4D, /* read-only */
        ADCM16_SERIAL_ID_REG      = 0x4E, /* read-only */
        ADCM16_MSTREAM_CFG        = 0x4F, /* read-only */


        // Timestamp registers
        ADCM16_TS_READ_A          = 0x50,
        ADCM16_TS_SET_REG64       = 0x54,
        ADCM16_TS_READ_B          = 0x58,
        ADCM16_TS_READ_REG64      = 0x5C
    };
/*
    enum {
        REG_DEVICE_ID = 0x42, //
        REG64_SERIAL_NUM = 0x50, //

        REG_TEMPERATURE = 0x54, // 84

        REG_FW_VER = 0x56, // 86
        REG_FW_REV = 0x57, // 87

        REG_LINK_ERR_CNT_1 = 0x58,
        REG_LINK_ERR_CNT_2 = 0x59,
        REG_LINK_ERR_CNT_3 = 0x5A,
        REG_LINK_CRC_CNT_1 = 0x5B,
        REG_LINK_CRC_CNT_2 = 0x5C,
        REG_LINK_CRC_CNT_3 = 0x5D,

        REG_LINK_STATUS = 0x61,
    };
*/
    MregDeviceAdcm(RegIOBase &p, int address);
    ~MregDeviceAdcm() override = default;
    double readTemp() override;
    quint64 readTimestampCurrent() override;
    quint64 readTimestampRun() override;
    quint64 readTimestampTrig() override;
    quint32 writeLiveMagic(quint32 value) override;
    quint32 readLiveMagic() override;

protected:
//    virtual quint16 readDeviceId();
    quint64 readOneWireId(bool checkAccess = true) override;
    bool readFirmwareId() override;
    bool readMStreamCfg() override;
};

} // namespace mlink

#endif // MREGDEVICE_H
