//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "mregdev/qtmregdevice.h"

#include <cstdlib>
#include <stdexcept>

#include <QDateTime>
#include <QNetworkInterface>
#include <QTime>
#include <QTimer>

#include "mregdev/BMStatus.h"
#include "mregdev/WrEpStatus.h"
#include "mregdev/mregdevice.h"
#include "regio/crc8_dallas_maxim.h"
#include "regio/mlinkdevice.h"
#include "regio/regiomlink.h"
#include "regio/regop.h"

enum {
    REG_WR_EP_STATUS_CODE            = 0x0,
    REG_WR_EP_STATUS_RESERVED        = 0x1,
    REG32_WR_EP_STATUS_UPTIME        = 0x2, // 32-bit
    REG64_WR_EP_STATUS_TIMECODE      = 0x4, // 64-bit
    REG32_WR_EP_STATUS_LINK_LOSS_CNT = 0x8, // 32-bit
    REG32_WR_EP_STATUS_SYNC_LOSS_CNT = 0xA, // 32-bit
    REG32_WR_EP_STATUS_CODE_ERR_CNT  = 0xC, // 32-bit
    REG32_WR_EP_STATUS_OTHER_ERR_CNT = 0xE, // 32-bit

    REG64_BM_ID_1  = 0x380,
    REG_BM_TEMP_1  = 0x384,
    REG_BM_READY_1 = 0x385,

    REG64_BM_ID_2  = 0x388,
    REG_BM_TEMP_2  = 0x38C,
    REG_BM_READY_2 = 0x38D,

    REG64_BM_ID_3  = 0x390,
    REG_BM_TEMP_3  = 0x394,
    REG_BM_READY_3 = 0x395,

    REG64_BM_ID_4  = 0x398,
    REG_BM_TEMP_4  = 0x39C,
    REG_BM_READY_4 = 0x39D,

    REG_BM_PWR_STATUS = 0x3BF,
};

const quint8 OW_ID_DS18B20 = 0x28;

const int onlineStatusPollTimeoutMs = 100; // milliseconds
const int offlineStatusPollTimeoutMs = 5000; // milliseconds
const int STARTUP_OFFLINE_TIMEOUT_MS = 1500;

QtMregDevice::QtMregDevice(quint16 devId, quint64 devSerial, QObject *parent) :
    QObject(parent),
    link(new mlink::RegIOMLink(this)),
    mregDevice(createMregDevice(*link, 0xFEFE, devId)),
    timerStatusPoll(new QTimer(this)),
    startupOfflineTimer(new QTimer(this)),
    liveMagic(qrand()%30000+100),
    deviceId(devId),
    deviceSerial(devSerial)
{
    connectedState = false;
    connect(timerStatusPoll, &QTimer::timeout, this, &QtMregDevice::statusPoll);
    connect(this, &QtMregDevice::deviceConnected, this, &QtMregDevice::on_deviceConnected);
    connect(this, &QtMregDevice::deviceDisconnected, this, &QtMregDevice::on_deviceDisconnected);
    timerStatusPoll->setSingleShot(true);
    timerStatusPoll->start(offlineStatusPollTimeoutMs);
    connect(startupOfflineTimer, &QTimer::timeout, this, &QtMregDevice::deviceDisconnected);
    startupOfflineTimer->setSingleShot(true);
    startupOfflineTimer->start(STARTUP_OFFLINE_TIMEOUT_MS);
}

QtMregDevice::~QtMregDevice()
{
    delete timerStatusPoll;
    mregDevice->disable();
    delete mregDevice;
    delete link;
}

mlink::MregDevice *QtMregDevice::createMregDevice(mlink::RegIOBase &link, int address, int id)
{
    if(
            id == QtMregDevice::DEVICE_ID_8CADC ||
            id == QtMregDevice::DEVICE_ID_ADC8E ||
            id == QtMregDevice::DEVICE_ID_ADC16LTCM ||
//            id == QtMregDevice::DEVICE_ID_ADC16LTCS ||
            id == QtMregDevice::DEVICE_ID_ADC32P ||
            id == QtMregDevice::DEVICE_ID_ADC64 ||
            id == QtMregDevice::DEVICE_ID_ADC64S2 ||
            id == QtMregDevice::DEVICE_ID_ADC64WR ||
            id == QtMregDevice::DEVICE_ID_ADC64VE ||
            id == QtMregDevice::DEVICE_ID_ADC64VE_XGE ||
            id == QtMregDevice::DEVICE_ID_ADC64ECAL ||
            id == QtMregDevice::DEVICE_ID_ADCM16 ||
            id == QtMregDevice::DEVICE_ID_ADCM163 ||
            id == QtMregDevice::DEVICE_ID_ADCM164 ||
            id == QtMregDevice::DEVICE_ID_BRIC ||
            id == QtMregDevice::DEVICE_ID_CRU16WR ||
            id == QtMregDevice::DEVICE_ID_TTVXS ||
            id == QtMregDevice::DEVICE_ID_TDC64VHLE ||
            id == QtMregDevice::DEVICE_ID_TDC72VXS ||
            id == QtMregDevice::DEVICE_ID_TQDC16VS_DIG ||
            id == QtMregDevice::DEVICE_ID_VMM128RO ||
            id == QtMregDevice::DEVICE_ID_TQDC16VS_E ) {
        return new mlink::MregDeviceAdcm(link, address);
    } else if(id == QtMregDevice::DEVICE_ID_ADC8BEDRS ||
              id == QtMregDevice::DEVICE_ID_BPM4100 ||
              id == QtMregDevice::DEVICE_ID_CTF6 ||
              id == QtMregDevice::DEVICE_ID_EVADC ||
              id == QtMregDevice::DEVICE_ID_FVME ||
              id == QtMregDevice::DEVICE_ID_HRB6ASD ||
              id == QtMregDevice::DEVICE_ID_HRB128E ||
              id == QtMregDevice::DEVICE_ID_MSC8E ||
              id == QtMregDevice::DEVICE_ID_MSC16VE_E ||
              id == QtMregDevice::DEVICE_ID_MSC16VE_TM ||
              id == QtMregDevice::DEVICE_ID_NUCLORF ||
              id == QtMregDevice::DEVICE_ID_TEST ||
              id == QtMregDevice::DEVICE_ID_TTB9V ||
              id == QtMregDevice::DEVICE_ID_U40VE ||
              id == QtMregDevice::DEVICE_ID_U40VE_8SPI ||
              id == QtMregDevice::DEVICE_ID_U40VE_IDEAS ||
              id == QtMregDevice::DEVICE_ID_U40VE_RC ||
              id == QtMregDevice::DEVICE_ID_U40VE_SI ||
              id == QtMregDevice::DEVICE_ID_UT24VE ||
              id == QtMregDevice::DEVICE_ID_VME_TRIG ) {
          return new mlink::MregDeviceBpm(link, address);
    } else if(id == QtMregDevice::DEVICE_ID_AnyDevice) {
        return new mlink::MregDeviceDummy(link, address);
    }
    return new mlink::MregDeviceDummy(link, address);
}

mlink::DeviceAttr QtMregDevice::getAttr() const
{
    return mregDevice->getAttr();
}

QtMregDevice::DeviceId QtMregDevice::getDeviceId() const
{
    return static_cast<DeviceId>(deviceId);
}

fw_version_t QtMregDevice::getDeviceFwVersion() const
{
    return getAttr().fw_version;
}

bool QtMregDevice::hasMultiAckSupport() const
{
    return getAttr().mstreamMultiAck;
}

quint8 QtMregDevice::getHwBufSize() const
{
    return getAttr().mstreamHwBufSize;
}

quint64 QtMregDevice::getDeviceSerialNumber() const
{
    return deviceSerial;
}

DeviceIndex QtMregDevice::getDeviceIndex() const
{
    return DeviceIndex(deviceId, deviceSerial);
}

QString QtMregDevice::getFirmwareVersionStr() const
{
    return getAttr().getFirmwareVersionStr();
}

QString QtMregDevice::getDeviceIdStr() const
{
    DeviceId id = getDeviceId();
    QString str = getDeviceTypeName(id);
    if(str.isEmpty())
        str = QString("Unknown device %1").arg(id, 4, 16, QChar('0'));

    QString fw_id = getAttr().getFirmwareVersionStr();
    if (!fw_id.isEmpty())
        str += " "+fw_id;
    return str;
}

QString QtMregDevice::getSerialIdStr() const
{
    return getSerialIdStr(deviceSerial);
}

QString QtMregDevice::getSerialIdStr(quint64 serial)
{
    quint16 s3 = (serial >> 32) & 0xFFFF;
    quint16 s2 = (serial >> 16) & 0xFFFF;
    quint16 s1 = serial & 0xFFFF;
    QString str = QString("%1-%2")
            .arg(s2, 4, 16, QChar('0'))
            .arg(s1, 4, 16, QChar('0'));
    if (s3)
        str = QString("%1-%2")
            .arg(s3, 4, 16, QChar('0'))
                .arg(str);
    return str.toUpper();
}

QString QtMregDevice::getDeviceTypeName(quint16 deviceId)
{
    switch (deviceId) {
    case QtMregDevice::DEVICE_ID_8CADC: return "8CADC";
    case QtMregDevice::DEVICE_ID_ADC8E: return "ADC8E";
    case QtMregDevice::DEVICE_ID_ADC8BEDRS: return "ADC8BEDRS";
    case QtMregDevice::DEVICE_ID_ADC16LTCM: return "ADC16LTCM";
//    case QtMregDevice::DEVICE_ID_ADC16LTCS: return "ADC16LTCS";
    case QtMregDevice::DEVICE_ID_ADC32P: return "ADC32P";
    case QtMregDevice::DEVICE_ID_ADC64: return "ADC64";
    case QtMregDevice::DEVICE_ID_ADC64S2: return "ADC64S2";
    case QtMregDevice::DEVICE_ID_ADC64WR: return "ADC64WR";
    case QtMregDevice::DEVICE_ID_ADC64VE: return "ADC64VE";
    case QtMregDevice::DEVICE_ID_ADC64VE_XGE: return "ADC64VE-XGE";
    case QtMregDevice::DEVICE_ID_ADC64ECAL: return "ADC64ECAL";
    case QtMregDevice::DEVICE_ID_ADCM16: return "ADCM16";
    case QtMregDevice::DEVICE_ID_ADCM163: return "ADCM163";
    case QtMregDevice::DEVICE_ID_ADCM164: return "ADCM164";
    case QtMregDevice::DEVICE_ID_BPM4100: return "BPM4100";
    case QtMregDevice::DEVICE_ID_BRIC: return "BRIC";
    case QtMregDevice::DEVICE_ID_CTF6: return "CTF6";
    case QtMregDevice::DEVICE_ID_CRU16WR: return "CRU16WR";
    case QtMregDevice::DEVICE_ID_EVADC: return "EVADC";
    case QtMregDevice::DEVICE_ID_FVME: return "FVME";
    case QtMregDevice::DEVICE_ID_HRB128E: return "HRB128E";
    case QtMregDevice::DEVICE_ID_HRB6ASD: return "HRB6ASD";
    case QtMregDevice::DEVICE_ID_MSC8E: return "MSC8E";
    case QtMregDevice::DEVICE_ID_MSC16VE_E: return "MSC16VE_E";
    case QtMregDevice::DEVICE_ID_MSC16VE_TM: return "MSC16VE_TM";
    case QtMregDevice::DEVICE_ID_NUCLORF: return "NUCLORF";
    case QtMregDevice::DEVICE_ID_TEST: return "TEST";
    case QtMregDevice::DEVICE_ID_TTVXS: return "TTVXS";
    case QtMregDevice::DEVICE_ID_TDC64VHLE: return "TDC64VHLE";
    case QtMregDevice::DEVICE_ID_TDC72VXS: return "TDC72VXS";
    case QtMregDevice::DEVICE_ID_TQDC16VS_DIG: return "TQDC16VS_DIG";
    case QtMregDevice::DEVICE_ID_TQDC16VS_E: return "TQDC16VS_E";
    case QtMregDevice::DEVICE_ID_TTB9V: return "TTB9V";
    case QtMregDevice::DEVICE_ID_U40VE: return "U40VE";
    case QtMregDevice::DEVICE_ID_U40VE_8SPI: return "U40VE-8SPI";
    case QtMregDevice::DEVICE_ID_U40VE_IDEAS: return "U40VE-IDEAS";
    case QtMregDevice::DEVICE_ID_U40VE_RC: return "U40VE-RC";
    case QtMregDevice::DEVICE_ID_U40VE_SI: return "U40VE-SI";
    case QtMregDevice::DEVICE_ID_UT24VE: return "UT24VE";
    case QtMregDevice::DEVICE_ID_VME_TRIG: return "TRIG";
    case QtMregDevice::DEVICE_ID_VMM128RO: return "VMM128RO";
    }
    // QtMregDevice::DEVICE_ID_AnyDevice
    return QString("Unknown device %1").arg(deviceId, 2, 16, QChar('0'));
}

int QtMregDevice::getDeviceChNum(const DeviceIndex &index)
{
    switch (index.getDevId()) {
    case QtMregDevice::DEVICE_ID_TDC72VXS:
        return 72;
    case QtMregDevice::DEVICE_ID_TDC64VHLE:
        return 64;
    case QtMregDevice::DEVICE_ID_TQDC16VS_E:
        return 16;
    }
    return -1;
}

QString QtMregDevice::messagePrefix() const
{
    return QString("[%1 %2]:")
            .arg(getDeviceTypeName(getDeviceId()))
            .arg(getSerialIdStr());
}

QHostAddress QtMregDevice::getAddress() const
{
    return link->getHostAddress();
}

QString QtMregDevice::getAddressStr() const
{
    return getAddress().toString();
}

QString QtMregDevice::getIdent() const
{
    DeviceId id = getDeviceId();
    QString str = getDeviceTypeName(id);
    if (str.isEmpty())
        str = QString("Unknown device %1").arg(id, 4, 16, QChar('0'));
    return str + " " + getSerialIdStr();
}

QString QtMregDevice::getDeviceInfoStr() const
{
    return QString("%1: %2, S/N %4, t %5")
          .arg(getAddressStr())
          .arg(getDeviceIdStr())
          .arg(getSerialIdStr())
          .arg(getTemperature(), 0, 'f', 1);
}

QPair<int, int> QtMregDevice::getHwVer() const
{
    return lastDeviceDescription.getHwVer();
}

double QtMregDevice::getTemperature() const
{
    return mregDevice->getAttr().temp;
}

void QtMregDevice::setBlockLiveMagic(bool block)
{
    blockLiveMagic = block;
}

void QtMregDevice::setHostAddress(const QHostAddress &address)
{
    if (connectedState && (address == link->getHostAddress()))
        return;
    if (connectedState) {
        mregDevice->disable();
        connectedState = false;
        Q_EMIT deviceDisconnected();
        Q_EMIT setStatus(QString("Disconnected"));
    }
    link->setHostAddress(address);
    detect();
}

void QtMregDevice::deviceUpdated(const DeviceDescription &dd)
{
    if(deviceId != dd.device_id || deviceSerial != dd.serial_id)
        return;
    QHostAddress addr(dd.ip_addr);
    if(addr.isNull() || !dd.ip_addr)
        return;

    if(dd.master_locked) {
        QHostAddress master(dd.master_ip);
        if(!QNetworkInterface::allAddresses().contains(master)) {
            if(lastDeviceDescription.master_ip != dd.master_ip)
                qWarning()<<"Device blocked with "<<master.toString();
            lastDeviceDescription = dd;
            return;
        }
    }

    if(getAddress() == addr) {
        lastDeviceDescription = dd;
    } else {
        if(lastDeviceDescription.isOutOfDate()){
            lastDeviceDescription = dd;
            setHostAddress(addr);
        }
    }
}

void QtMregDevice::detect()
{
    QString offlineReason = "Disconnected";
    if(!lastDeviceDescription.isNull() && lastDeviceDescription.isOutOfDate() && !link->getHostAddress().isNull())
        setHostAddress(QHostAddress());

    if (link->getHostAddress().isNull()) return;
    if (mregDevice->isOnline() && connectedState) {
        Q_EMIT setStatus(QString("Reconnecting to %1...").arg(link->getHostAddress().toString()));
    } else {
        Q_EMIT setStatus(QString("Connecting to %1...").arg(link->getHostAddress().toString()));
    }
    mregDevice->enable();
    quint16 id = 0;
    try {
        id = mregDevice->detect();
    }
    catch (...) {}
//    qDebug() << QString("Device ID:%1").arg(id, 4, 16, QChar('0'));
    if (id == 0 || id == 0xFFFF) {
        mregDevice->disable();
        QString msg =  QString("%1 Can't connect to %2").arg(messagePrefix()).arg(getAddressStr());
        qCritical() << msg;
        offlineReason = msg;
        Q_EMIT setStatus(msg);
    } else {
        if (mregDevice->isOnline()) {
            if (deviceId != id || deviceSerial != getAttr().device_serial) {
                QString msg = QString("%1 Unexpected device %2 serial %3 at %4")
                        .arg(messagePrefix())
                        .arg(getDeviceTypeName(id))
                        .arg(getSerialIdStr(getAttr().device_serial))
                        .arg(getAddressStr());
                lastDeviceDescription.clear();
                setHostAddress(QHostAddress());
                qWarning() << msg;
                offlineReason = msg;
                Q_EMIT setStatus(msg);
            } else {
//                mregDevice->enable();
                qDebug() << getDeviceInfoStr();
            }
        }
    }

    supportLiveMagic = !blockLiveMagic;
    try{
        if(supportLiveMagic)
            writeLiveMagic();
    } catch (...) {}

    bool newState = isConnected();
    if (newState) {
        Q_EMIT setStatus(QString("Online"));
        if (!connectedState) {
            Q_EMIT deviceConnected();
        }
    } else {
        Q_EMIT setStatus(QString(offlineReason));
        if (connectedState) {
            Q_EMIT deviceDisconnected();
        }
    }
    connectedState = newState;
}

bool QtMregDevice::writeLiveMagic()
{
    if (!isConnected() || !supportLiveMagic) return false;

    quint16 tryMagic = liveMagic+1;
    if(tryMagic==0)
        tryMagic = qrand()%30000+100;
    quint16 wroteMagic = mregDevice->writeLiveMagic(tryMagic);
    supportLiveMagic = wroteMagic == tryMagic;
    if (!supportLiveMagic) {
        qWarning() << QString("%1 LiveMagic function not supported").arg(messagePrefix());
    }
    liveMagic = wroteMagic;
    return supportLiveMagic;
}

quint16 QtMregDevice::readLiveMagic()
{
    try {
        return mregDevice->readLiveMagic();
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
    return 0;
}

bool QtMregDevice::checkLiveMagic()
{
    if (!isConnected()) return false;
    if (!supportLiveMagic) return true;
    quint16 rd = readLiveMagic();
    if (!isConnected()) return false;
    if (rd != liveMagic) qWarning() << "liveMagic: read" << rd << " expected " << liveMagic;
    return rd == liveMagic;
}

bool QtMregDevice::isConnected() const
{
    return mregDevice->isOnline() && mregDevice->isEnabled();
}

double QtMregDevice::readTemp()
{
    double temp = -255;
    try {
        temp = mregDevice->readTemp();
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }

    if (temp != currentTemp) {
        Q_EMIT TempChanged(temp);
        currentTemp = temp;
    }
    return temp;
}

quint64 QtMregDevice::readTimestampCurrent()
{
    return mregDevice->readTimestampCurrent();
}

quint64 QtMregDevice::readTimestampRun()
{
    return mregDevice->readTimestampRun();
}

quint64 QtMregDevice::readTimestampTrig()
{
    return mregDevice->readTimestampTrig();
}

bool QtMregDevice::hasWrEpStatus() const
{
    constexpr fw_version_t REV_20519 = fw_version_t(1, 0, 20519);
    return (getDeviceId() == DEVICE_ID_TTVXS) ||
           (getDeviceId() == DEVICE_ID_ADC64WR
            && (getDeviceFwVersion().isNull() ||
                getDeviceFwVersion() >= REV_20519));
}

WrEpStatus QtMregDevice::readWrEpStatus()
{
    WrEpStatus wrEpStatus;
    mlink::RegOpVector op;
    const quint16 regWrBase = getDeviceId() == DEVICE_ID_TTVXS ? 0x800 : 0x60;
    op << mlink::RegRead  (regWrBase + REG_WR_EP_STATUS_CODE);
    op << mlink::RegRead  (regWrBase + REG_WR_EP_STATUS_RESERVED);
    op << mlink::RegRead32(regWrBase + REG32_WR_EP_STATUS_UPTIME);
    op << mlink::RegRead64(regWrBase + REG64_WR_EP_STATUS_TIMECODE);
    op << mlink::RegRead32(regWrBase + REG32_WR_EP_STATUS_LINK_LOSS_CNT);
    op << mlink::RegRead32(regWrBase + REG32_WR_EP_STATUS_SYNC_LOSS_CNT);
    op << mlink::RegRead32(regWrBase + REG32_WR_EP_STATUS_CODE_ERR_CNT);
    op << mlink::RegRead32(regWrBase + REG32_WR_EP_STATUS_OTHER_ERR_CNT);

    qint64 t1 = QDateTime::currentMSecsSinceEpoch();
    mlink::RegOpVector result = regOpExec(op);
    qint64 t2 = QDateTime::currentMSecsSinceEpoch();

    wrEpStatus.hasWrEpStatus = hasWrEpStatus();
    wrEpStatus.acquiredAtMs = (t1 + t2) /2;
    wrEpStatus.acquireErrorMs = std::abs(t1 - t2);

    Q_ASSERT(result.size() == 16);
    wrEpStatus.wrStatusCode = result.at(0).data;
    wrEpStatus.wrReserved   = result.at(1).data;
    wrEpStatus.wrUptime     = get32(result, 2);
    wrEpStatus.wrTimecode   = get64(result, 4);
    wrEpStatus.wrLinkError  = get32(result, 8);
    wrEpStatus.wrSyncFail   = get32(result, 10);
    wrEpStatus.wrCodeError  = get32(result, 12);
    wrEpStatus.wrOtherError = get32(result, 14);
    return wrEpStatus;
}

bool oneWireCrcValid(quint64 ow_id)
{
    const quint8 ow_crc = ow_id >> 56;
    // const int testID = OW_ID_DS18B20;
    unsigned char crc = 0;
    for(int i=0; i<7; i++)
        crc = crc8_dallas_maxim((ow_id >> i*8) & 0xFF, crc);
    return crc == ow_crc;
}

BMStatus QtMregDevice::readBMStatus()
{
    BMStatus bmStatus;
    mlink::RegOpVector op;
    const int numSensors = 2;
    for (int i=0; i<numSensors; i++) {
        op << mlink::RegRead64(i*8 + REG64_BM_ID_1);
        op << mlink::RegRead(i*8 + REG_BM_TEMP_1);
        op << mlink::RegRead(i*8 + REG_BM_READY_1);
    }

    op << mlink::RegRead(REG_BM_PWR_STATUS);

    mlink::RegOpVector result = regOpExec(op);
    Q_ASSERT(result.size() == numSensors * 6 + 1);
    if (result.size() < numSensors * 6 + 1)
        return bmStatus;
    for (int i=0; i<numSensors; i++) {
        BMSensor sensor;
        sensor.ow_id = ((quint64)result.at(i*6+3).data << 48) |((quint64)result.at(i*6+2).data << 32) | ((quint64)result.at(i*6+1).data << 16) | result.at(i*6+0).data;
        bool crcValid = oneWireCrcValid(sensor.ow_id);
        bool sensor_valid = crcValid && ((sensor.ow_id & 0xFF) == OW_ID_DS18B20);
        sensor.temp = mlink::MregDevice::convertOneWireTemp(result.at(i*6+4).data);
        sensor.ready = (result.at(i*6+5).data & 0x1);
//        if (sensor.ready)
//            qWarning() << QString("crc:%1, id:%2").arg(crcValid).arg(sensor.ow_id, 8, 16, QChar('0'));
        if (sensor_valid)
            bmStatus.sensors.push_back(sensor);
    }
    bmStatus.pwr_status = result.at(result.length() - 1).data;
    return bmStatus;
}

void QtMregDevice::enable()
{
    mregDevice->enable();
}

void QtMregDevice::disable()
{
    mregDevice->disable();
}

void QtMregDevice::statusPoll()
{
    if (!mregDevice->isOnline() || !checkLiveMagic()) {
        //        Q_EMIT setStatus(QString("Offline"));
        //qDebug() << "re-detect";
        if(mregDevice->isOnline() && isConnected()){
            connectedState = false;
            emit liveMagicFailed(getDeviceIndex());
            Q_EMIT deviceDisconnected();
            Q_EMIT setStatus(QString("LiveMagic failed"));
        }
        detect(); // re-detect
        //        goto finish;
    }
    if (mregDevice->isOnline()) {
        readTemp();
    }

    // restart timer
    if (mregDevice->isOnline())
        timerStatusPoll->start(onlineStatusPollTimeoutMs);
    else
        timerStatusPoll->start(offlineStatusPollTimeoutMs);
}

void QtMregDevice::on_deviceConnected()
{
    qInfo().noquote() << QString("%1 connected to %3")
               .arg(messagePrefix())
               .arg(getAddressStr());
    timerStatusPoll->start(0);
    if(startupOfflineTimer->isActive())
        startupOfflineTimer->stop();
}

void QtMregDevice::on_deviceDisconnected()
{
    qInfo() << QString("%1 disconnected")
          .arg(messagePrefix());
    timerStatusPoll->start(offlineStatusPollTimeoutMs);
}

void QtMregDevice::deleteHardware(const DeviceIndex &index)
{
    if(index.isNull() ||
            (index.getDevId()==deviceId && index.getSerial()==deviceSerial) ){
        mregDevice->disable();
        deleteLater();
    }
}

/// read 16-bit register
quint16 QtMregDevice::regRead(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead %1").arg(regNumber));
        return mregDevice->regRead(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 16-bit register
void QtMregDevice::regWrite(int regNumber, quint16 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

//void QtMregDevice::regWrite(const std::map<quint16, quint16> &data)
//{
//    Q_ASSERT(isConnected());
//    try {
//        Q_EMIT setStatus(QString("regWriteMap"));
//        mregDevice->regWriteBlk(data);
//        Q_EMIT setStatus(QString("Ok"));
//    }
//    catch (std::runtime_error &e) {
//        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
//    }
//}

/// read 32-bit register
quint32 QtMregDevice::regRead32(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead32 %1").arg(regNumber));
        return mregDevice->regRead32(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 32-bit register
void QtMregDevice::regWrite32(int regNumber, quint32 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite32 %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite32(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

/// read 64-bit register
quint64 QtMregDevice::regRead64(int regNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regRead64 %1").arg(regNumber));
        return mregDevice->regRead64(regNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

/// write 64-bit register
void QtMregDevice::regWrite64(int regNumber, quint64 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regWrite64 %1 %2").arg(regNumber).arg(data));
        mregDevice->regWrite64(regNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

std::vector<quint16> QtMregDevice::regReadBlk(const std::vector<int> &registers, bool checkAccess)
{
    std::vector<quint16> result;
    try {
        result = mregDevice->regReadBlk(registers, checkAccess);
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        if(!mregDevice->isOnline())
            detect();
    }
    return result;
}

/// evaluate register operations
mlink::RegOpVector QtMregDevice::regOpExec(const mlink::RegOpVector &data)
{
    mlink::RegOpVector result;
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("regOpExec %1").arg(data.size()));
        result = mregDevice->regOpExec(data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        if(!mregDevice->isOnline())
            detect();
    }
    return result;
}

std::vector<quint32> QtMregDevice::memReadBlk(int startWord, size_t wordCount)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memReadBlk %1 %5").arg(startWord).arg(wordCount));
        return mregDevice->memReadBlk(startWord, wordCount);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return std::vector<quint32>();
    }
}

quint32 QtMregDevice::memRead(int wordNumber)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWrite %1 %5").arg(wordNumber).arg(data));
        return mregDevice->memRead(wordNumber);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
        return 0;
    }
}

void QtMregDevice::memWrite(int wordNumber, quint32 data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWrite %1 %5").arg(wordNumber).arg(data));
        mregDevice->memWrite(wordNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

void QtMregDevice::memWriteBlk(int wordNumber, const std::vector<quint32> &data)
{
//    Q_ASSERT(isConnected());
    try {
//        Q_EMIT setStatus(QString("memWriteBlk %1 %5").arg(wordNumber).arg(data.size()));
        mregDevice->memWriteBlk(wordNumber, data);
//        Q_EMIT setStatus(QString("Ok"));
    }
    catch (std::runtime_error &e) {
        Q_EMIT setStatus(QString("Communication error: %1").arg(e.what()));
    }
}

quint32 QtMregDevice::get32(const mlink::RegOpVector &rx, int startPos)
{
    return (quint32(rx.at(startPos+1).data)<<16) | rx.at(startPos).data;
}

quint64 QtMregDevice::get64(const mlink::RegOpVector &rx, int startPos)
{
    return (quint64(get32(rx,startPos+2))<<32) | get32(rx,startPos);
}

quint32 QtMregDevice::get32(mlink::RegOpVector::const_iterator &regOpIt, bool *ok)
{
    bool lockal_ok = ((regOpIt->addr+1) == (regOpIt+1)->addr)
            && !regOpIt->write
            && !(regOpIt+1)->write;
    Q_ASSERT(lockal_ok);
    if(ok)
        *ok = lockal_ok;
    if(!lockal_ok)
        return 0;
    quint32 retVal = (quint32((regOpIt+1)->data)<<16) | regOpIt->data;
    regOpIt += 2;
    return retVal;
}

quint64 QtMregDevice::get64(mlink::RegOpVector::const_iterator &regOpIt, bool *ok)
{
    bool lockal_ok = regOpIt->addr+2 == (regOpIt+2)->addr;
    Q_ASSERT(lockal_ok);
    if(ok)
        *ok = lockal_ok;
    if(!lockal_ok)
        return 0;
    return (quint64(get32(regOpIt, ok))) | (quint64(get32(regOpIt, ok))<<32);
}
