//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef QTMREGDEVICE_H
#define QTMREGDEVICE_H

#include <vector>

#include <QMap>
#include <QObject>

#include "mldiscover/DeviceDescription.h"
#include "mregdev/fw_version.h"

namespace mlink {
class DeviceAttr;
class MregDevice;
class RegIOBase;
class RegOpVector;
class RegOp;
}

class BMStatus;
class QHostAddress;
class QTimer;
class WrEpStatus;

class QtMregDevice : public QObject
{
    Q_OBJECT
public:
    typedef enum {
        DEVICE_ID_UNKNOWN       = 0,
        DEVICE_ID_VME_TRIG      = 0x0A, //  0x0A00,
        DEVICE_ID_TEST          = 0x55, // 0x5500,
        DEVICE_ID_TQDC16VS_DIG  = 0x89, // 0x8900
        DEVICE_ID_8CADC         = 0xB3, // 0xB300,
        DEVICE_ID_ADCM16        = 0xB4, // 0xB400,
        DEVICE_ID_ADC16LTCM     = 0xB5, // 0xB500,
//        DEVICE_ID_ADC16LTCS     = 0xB5, // 0xB501,
        DEVICE_ID_ADCM163       = 0xB6, // 0xB600,
        DEVICE_ID_BPM4100       = 0xB7, // 0xB700,
        DEVICE_ID_ADC8E         = 0xB8, // 0xB800,
        DEVICE_ID_EVADC         = 0xB9, // 0xB900,
        DEVICE_ID_ADC64         = 0xBA, // 0xBA00,
        DEVICE_ID_NUCLORF       = 0xBB, // 0xBB00,
        DEVICE_ID_BRIC          = 0xBC, // 0xBC00,
        DEVICE_ID_MSC8E         = 0xBD, // 0xBD00,
        DEVICE_ID_ADCM164       = 0xBE, // 0xBE00,
        DEVICE_ID_ADC8BEDRS     = 0xBF, // 0xBF00,
        DEVICE_ID_HRB128E       = 0xC1, // 0xC100,
        DEVICE_ID_HRB6ASD       = 0xC2, // 0xC200,
        DEVICE_ID_ADC32P        = 0xC3, // 0xC300,
        DEVICE_ID_TTB9V         = 0xC4, // 0xC400,
        DEVICE_ID_ADC64S2       = 0xC5, // 0xC500,
        DEVICE_ID_U40VE         = 0xC6, // 0xC600,
        DEVICE_ID_CTF6          = 0xC8, // 0xC800,
        DEVICE_ID_UT24VE        = 0xC9, // 0xC900,
        DEVICE_ID_ADC64WR       = 0xCA, // 0xCA00,
        DEVICE_ID_U40VE_8SPI    = 0xCB, // 0xCB00,
        DEVICE_ID_U40VE_RC      = 0xCC, // 0xCC00,
        DEVICE_ID_U40VE_IDEAS   = 0xCD, // 0xCD00,
        DEVICE_ID_U40VE_SI      = 0xCE, // 0xCE00,
        DEVICE_ID_TTVXS         = 0xCF, // 0xCF00,
        DEVICE_ID_TDC72VXS      = 0xD0, // 0xD000,
        DEVICE_ID_FVME          = 0xD1, // 0xD100,
        DEVICE_ID_TDC64VHLE     = 0xD3, // 0xD300,
        DEVICE_ID_ADC64VE       = 0xD4, // 0xD400,
        DEVICE_ID_TQDC16VS_E    = 0xD6, // 0x8900,
        DEVICE_ID_VMM128RO      = 0xDB, // 0xDB00,
        DEVICE_ID_MSC16VE_TM    = 0xDC, // 0xD800,
        DEVICE_ID_MSC16VE_E     = 0xD8, // 0xD800,
        DEVICE_ID_ADC64VE_XGE   = 0xD9, // 0xD900,
        DEVICE_ID_ADC64ECAL     = 0xDE, // 0xDE00,
        DEVICE_ID_CRU16WR       = 0xE0, // 0xE000,
        DEVICE_ID_AnyDevice     = 0xFF //  0xFFFF
    } DeviceId;

    Q_DISABLE_COPY(QtMregDevice)

    explicit QtMregDevice(quint16 devId, quint64 devSerial, QObject *parent = nullptr);
    ~QtMregDevice() override;
    void detect();
    bool isConnected() const;
    mlink::DeviceAttr getAttr() const;
    DeviceId getDeviceId() const;
    fw_version_t getDeviceFwVersion() const;
    bool hasMultiAckSupport() const;
    quint8 getHwBufSize() const;
    quint64 getDeviceSerialNumber() const;
    DeviceIndex getDeviceIndex() const;
    QString getIdent() const;
    QString getDeviceIdStr() const;
    QString getSerialIdStr() const;
    static QString getSerialIdStr(quint64 serial);
    static QString getDeviceTypeName(quint16 deviceId);
    static int getDeviceChNum(const DeviceIndex &index);
    QString messagePrefix() const;
    QString getFirmwareVersionStr() const;
    QHostAddress getAddress() const;
    QString getAddressStr() const;
    QString getDeviceInfoStr() const;
    QPair<int,int> getHwVer() const;
    double getTemperature() const;
    void setBlockLiveMagic(bool block=true);
    bool writeLiveMagic();
    virtual bool checkLiveMagic();
    bool hasWrEpStatus() const;

    /// read 16-bit register
    quint16 regRead(int regNumber);

    /// write 16-bit register
    void regWrite(int regNumber, quint16 data);

    //    void regWrite(const QMap<quint16, quint16> &data);

    /// read 32-bit register
    quint32 regRead32(int regNumber);

    /// write 32-bit register
    void regWrite32(int regNumber, quint32 data);

    /// read 64-bit register
    quint64 regRead64(int regNumber);

    /// write 64-bit register
    void regWrite64(int regNumber, quint64 data);

    /// read 16-bit registers
    std::vector<quint16> regReadBlk(const std::vector<int> &registers, bool checkAccess = true);

    /// evaluate register operations
    mlink::RegOpVector regOpExec(const mlink::RegOpVector &data);

    /// read 32-bit word from RAM at word offset wordNumber
    quint32 memRead(int wordNumber);

    /// write 32-bit word to RAM at word offset wordNumber
    void memWrite(int wordNumber, quint32 data);

    /// read wordCount 32-bit words from RAM starting from startWord offset
    std::vector<quint32> memReadBlk(int startWord, size_t wordCount);

    /// write 32-bit words to RAM starting from startWord offset
    void memWriteBlk(int wordNumber, const std::vector<quint32> &data);

    static quint32 get32(const mlink::RegOpVector &rx, int startPos);
    static quint64 get64(const mlink::RegOpVector &rx, int startPos);
    static quint32 get32(QVector<mlink::RegOp>::const_iterator &regOpIt, bool *ok=nullptr);
    static quint64 get64(QVector<mlink::RegOp>::const_iterator &regOpIt, bool *ok=nullptr);


Q_SIGNALS:
    void deviceConnected();
    void deviceDisconnected();
    void liveMagicFailed(const DeviceIndex &);
    void setStatus(const QString &);
    void TempChanged(double);

protected Q_SLOTS:
    virtual void statusPoll();
    virtual void on_deviceConnected();
    virtual void on_deviceDisconnected();

public Q_SLOTS:
    virtual void deleteHardware(const DeviceIndex &index);
    void setHostAddress(const QHostAddress &address);
    void deviceUpdated(const DeviceDescription &dd);

protected:
    mlink::MregDevice *createMregDevice(mlink::RegIOBase &p, int address, int id) Q_REQUIRED_RESULT;
    double readTemp();
    quint64 readTimestampCurrent();
    quint64 readTimestampRun();
    quint64 readTimestampTrig();
    WrEpStatus readWrEpStatus();
    BMStatus readBMStatus();

    void enable();
    void disable();
    quint16 readLiveMagic();

    mlink::RegIOBase * const link;

private:
    mlink::MregDevice * const mregDevice; // dev must be after link
    QTimer * const timerStatusPoll;
    QTimer * const startupOfflineTimer;
    bool connectedState = false;
    double currentTemp = 0;
    quint32 liveMagic = 0;
    bool supportLiveMagic = false;
    bool blockLiveMagic = false;
    const quint16 deviceId = 0;
    const quint64 deviceSerial = 0;
    DeviceDescription lastDeviceDescription;
};


#endif // QTMREGDEVICE_H
