//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef REGIO_MLINKDEVICE_H
#define REGIO_MLINKDEVICE_H

#include <cstddef>
#include <vector>
#include <QVector>
#include <string>

#include <regio/regiopacket.h>
#include <QtGlobal>

namespace mlink {

class RegIoPacket;
//class MlinkPacketProtocol;
class RegIOBase;
class RegOpVector;

class MlinkDevice
{
public:
    Q_DISABLE_COPY(MlinkDevice)

    MlinkDevice(RegIOBase &p, int address);

    virtual ~MlinkDevice() = default;
    /// read 16-bit register
    quint16 regRead(int regNumber, bool checkAccess = true);

    /// read two 16-bit registers to 32-bit word
    quint32 regRead32(int regNumber, bool checkAccess = true);

    /// read four 16-bit registers to 64-bit word
    quint64 regRead64(int regNumber, bool checkAccess = true);

    /// write 16-bit register
    void regWrite(int regNumber, quint16 data, bool checkAccess = true);

    /// evaluate register operations
    RegOpVector regOpExec(const RegOpVector &data, bool checkAccess = true);

    /// write 32-bit word to two 16-bit registers, low word even, high word odd
    void regWrite32(int regNumber, quint32 data, bool checkAccess = true);

    /// write 64-bit word to four 16-bit registers, low word even, high word odd
    void regWrite64(int regNumber, quint64 data, bool checkAccess = true);

    /// read regCount 16-bit registers starting from startReg
    std::vector<quint16> regReadBlk(int startReg, size_t regCount, bool checkAccess = true);

    /// read 16-bit registers at given addresses
    std::vector<quint16> regReadBlk(std::vector<int> registers, bool checkAccess = true);

    /// write 16-bit registers starting from startReg
    void regWriteBlk(int startReg, const std::vector<quint16> &data, bool checkAccess = true);

    /// read 32-bit word from EEPROM at word offset wordNumber
    quint32 eepromRead(int wordNumber, bool checkAccess = true);

    /// write 32-bit word to EEPROM at word offset wordNumber
    void eepromWrite(int wordNumber, quint32 data, bool checkAccess = true);

    /// read 32-bit word from RAM at word offset wordNumber
    quint32 memRead(int wordNumber, bool checkAccess = true);

    /// write 32-bit word to RAM at word offset wordNumber
    void memWrite(int wordNumber, quint32 data, bool checkAccess = true);

    /// read wordCount 32-bit words from RAM starting from startWord offset
    std::vector<quint32> memReadBlk(int startWord, size_t wordCount, bool checkAccess = true);

    /// write 32-bit words to RAM starting from startWord offset
    void memWriteBlk(int startWord, const std::vector<quint32> &data, bool checkAccess = true);

    /// maximum number of data words for single RAM read/write operation
    size_t getMaxRamRequestSize() const;

    /// maximum number of 16-bit registers for single register read/write operation
    size_t getMaxRegRequestSize() const;

    /// device address
    int getAddress() const { return myAddress; }
    void setAddress(int address) { myAddress = address; }
    void enable() { enableState = true; }
    void disable() { enableState = false; }
    void setMaster(bool state = true) { masterState = state; }
    bool isEnabled() const { return enableState; }
    bool isOnline() const { return onlineState; }
    bool isMaster() const { return masterState; }
    virtual quint16 detect();
    virtual void setDefaults();
    virtual bool validateIoAck(const RegIoPacket &tx, const RegIoPacket &rx, const std::string &description = std::string()) const;

    //    int64_t readData(char *data, int64_t maxSize);
    RegIOBase &getLink() { return link; } // FIXME
    QString getIdent() const;

protected:
    RegIoPacket ctrlExchangeSingle(const RegIoPacket &); ///< Send, receive and validate control packet
    PacketList ctrlExchange(const RegIoPacket &); ///< Send, receive and validate control packets
    PacketList ctrlExchange(const PacketList &); ///< Send, receive and validate control packets
    bool hardwareAccessAllowed() const { return enableState && onlineState; }
    RegIOBase &link;
//    MlinkPacketProtocol * const packetProtocol;
    bool onlineState = false;
    bool enableState = true;
    bool masterState = false;

private:
    int nextTxPacketSequenceNumber() { return txPacketSequenceNumber++; }
    int myAddress = 0;
    int txPacketSequenceNumber = 0;
};

} // namespace mlink

#endif /* REGIO_MLINKDEVICE_H */
