
#include "regiobase.h"

#include <algorithm>
#include <cmath>
#include <iostream>
#include <sstream>
#include <stdexcept>

#include <QCoreApplication>
#include <QDebug>

#include "TextFormatter.h"
#include "frame_util.h"
#include "mlink/frame_types.h"
#include "mlink/mlink_frame_header.h"
#include "util/gettimeofday.h"

namespace mlink {

double getCurrentTimestamp()
{
    struct timeval tv{};
    gettimeofday(&tv, nullptr);
    return tv.tv_sec + 1e-6*tv.tv_usec;
}

RegIOBase::RegIOBase(QObject *parent, const QHostAddress &address)
    : QObject(parent), deviceAddress(address)
{
}

std::string RegIOBase::getDevName() const
{
    return "mlink-netdev";
}

unsigned long RegIOBase::getDriverVersion() const
{
    return 0;
}

unsigned long RegIOBase::getDriverRevision() const
{
    return 0;
}

void RegIOBase::setHostAddress(const QHostAddress &address)
{
    deviceAddress = address;
}

//const double protocolTimeout = 1.0;

bool RegIOBase::packetTypeIsData(int type)
{
    if (type == FRAME_TYPE_ADC_DATA) return true;
    if (type == FRAME_TYPE_STAT) return true;
    return false;
}

void RegIOBase::sendPackets(const PacketList &pl)
{
    for(PacketList::const_iterator it = pl.begin(); it != pl.end(); ++it) {
        rawSend(*it);
    }
}

void RegIOBase::reset()
{
    rxBufCtrl.clear();
    //    rxBufData.clear();
}

double RegIOBase::getDeviceThroughput() const
{
    return 1e6; // default value
}

size_t RegIOBase::getCtrlPacketCount() const
{
    return rxBufCtrl.size();
}

RegIoPacket RegIOBase::getCtrlPacketBySeq(int seq, double timeoutSec)
{
    double tv1 = getCurrentTimestamp();
    int attempt = 0;
    const int attemptLimit = 10;
    while(true)
    {
        auto waitTimeMs = timeoutSec*1e3/pow(2,attemptLimit-(attempt++));
        updateRxBuf(waitTimeMs);
//        qDebug() << rxBufCtrl.size();
        for(PacketList::const_iterator it = rxBufCtrl.begin(); it != rxBufCtrl.end(); ++it) {
//            qDebug() << "seq = " << it->seq << ", expected " << seq;
            if (it->seq == seq) {
                RegIoPacket packet = *it;
                rxBufCtrl.clear();
//                qDebug() << "Ok: rxBufCtrl=" << rxBufCtrl;
                return packet;
            }
        }
        double tv2 = getCurrentTimestamp();
        double dt = tv2 - tv1;
        if (dt > timeoutSec) {
//            qWarning() << "Protocol timeout, dt=" << dt << " s (getCtrlPacketBySeq(" << seq << ", " << timeout << "))";
//            qDebug() << "rxBufCtrl=" << rxBufCtrl;
            break;
        }
    }
//    const PacketList pl(rxBufCtrl);
//    rxBufCtrl.clear();
//    qDebug()<<attempt<< " "<<qPow(0.5,attemptLimit-attempt+1);
    return RegIoPacket();
}

PacketList RegIOBase::getCtrlPackets(size_t packetCount, double timeout)
{
    double tv1 = getCurrentTimestamp();
    while(getCtrlPacketCount() < packetCount)
    {
        //        receive();
        if (getCtrlPacketCount() >= packetCount) break;
        double tv2 = getCurrentTimestamp();
        double dt = tv2 - tv1;
        if (dt > timeout) {
            qWarning() << "Protocol timeout, dt=" << dt << " s (getControlPackets(" << packetCount << ", " << timeout << "))";
            qDebug() << "rxBufCtrl=" << rxBufCtrl;
            break;
        }
        qApp->processEvents();
    }
    const PacketList pl(rxBufCtrl);
    rxBufCtrl.clear();
    return pl;
}

} // namespace mlink
