#include "ConfigurationManager.h"
#include "ui_ConfigurationManager.h"

#include <QAction>
#include <QActionGroup>
#include <QCloseEvent>
#include <QDialogButtonBox>
#include <QDialog>
#include <QDesktopWidget>
#include <QDebug>
#include <QJsonDocument>
#include <QLayout>
#include <QLabel>
#include <QListWidgetItem>
#include <QMenu>
#include <QModelIndexList>
#include <QShortcut>
#include <QSplitter>
#include <QStack>
#include <QTreeView>
#include <QTreeWidgetItem>
#include <QMessageBox>

namespace {
    const QString TREE_HEADER_STYLE = (
    "   QHeaderView::section::first{               "
    "       background: white;                     "
    "       border: 0px;                           "
    "       border-bottom: 1px solid gray;         "
    "       border-right: 1px solid gray;          "
    "       padding: 4px;                          "
    "       text-align: center;                    "
    "   }                                          "
    "   QHeaderView::section::last {               "
    "       background: white;                     "
    "       border: 0px;                           "
    "       border-bottom: 1px solid gray;         "
    "       padding: 4px;                          "
    "       text-align: center;                    "
    "   }                                          "
    );

    QString TREE_STYLE_1 = (
    "   QTreeView {                             "
    "       background: rgb(216, 231, 235);     "
    "       border: 1px solid gray;             "
    "       border-radius: 5px;                 "
    "       padding: 15px;                      "
    "   }                                       "
    );

    const QString TREE_STYLE = (
    "   QTreeView{                                 "
    "       background: white;                     "
    "       border: 1px solid gray;                "
    "   }                                          "
    );

}

ConfigurationManager::ConfigurationManager(
    QWidget *parent,
    QString program_type,
    RootConfig default_root)
    :
    QDialog(parent),
    ui(new Ui::ConfigurationManager),
    program_type(program_type),
    default_root(default_root)
{
    ui->setupUi(this);
    setWindowTitle("Configuration manager");

    if(program_type.isEmpty())
        return;

    connect(ui->tree_left, &QTreeWidget::currentItemChanged, this, &ConfigurationManager::config_changed);

    //--- splitter
    QSplitter *split_vertical = new QSplitter(this);
    split_vertical->addWidget(ui->tree_left);
    split_vertical->addWidget(ui->tree);
    split_vertical->setStretchFactor(0,6);
    split_vertical->setStretchFactor(1,5);
    ui->horizontalLayout_3->addWidget(split_vertical);

    create_subdialog_delete();
}

ConfigurationManager::~ConfigurationManager()
{
    delete ui;
}

void ConfigurationManager::reset(QString program_index_, QString configuration_name_)
{
    program_index = program_index_;
    configuration_name = configuration_name_;
    program_index_selected = program_index_;
    configuration_name_selected = configuration_name_;
    selected_root = base_config.read_config(program_type, program_index_selected, configuration_name_selected);

    //--- mongo status
    ui->label_db_status->setText("Mongo disconnected");
    QPalette palette = ui->label_db_status->palette();
    palette.setColor(ui->label_db_status->foregroundRole(), Qt::red);
    ui->label_db_status->setPalette(palette);

    #ifdef MONGO_DRIVERS
        if(base_config.mongo_db.try_connection()){
            ui->label_db_status->setText("Mongo connected");
            QColor green;
            green.setCmyk(100,0,100,100);
            QPalette palette = ui->label_db_status->palette();
            palette.setColor(ui->label_db_status->foregroundRole(), green);
            ui->label_db_status->setPalette(palette);
        }
    #endif

    populate_tree(selected_root, ui->tree);
    populate_left_tree(ui->tree_left);
}

void ConfigurationManager::create_subdialog_delete()
{

}

void ConfigurationManager::populate_tree(RootConfig root, QTreeWidget *tree)
{    
    //--- leaflet
    QMap<QString, QVariant> root_leaflets = root.data();
    QMap<QString, QTreeWidgetItem*> tree_leaflets;

    for(int i=0; i<tree->topLevelItemCount(); ++i){
        QTreeWidgetItem *item = tree->topLevelItem(i);
        if(item->data(1, Qt::UserRole).toString() == "leaflet"){
            tree_leaflets[item->data(0, Qt::DisplayRole).toString()] = item;
        }
    }

    for(QString field : tree_leaflets.keys()){
        if(!root_leaflets.keys().contains(field)){
            delete tree_leaflets[field];
        }
    }

    for(QString field : root_leaflets.keys()){
        QTreeWidgetItem * item_to_update;
        if(!tree_leaflets.contains(field)){
            item_to_update = new QTreeWidgetItem();
            tree->addTopLevelItem(item_to_update);
        }
        else{
            item_to_update = tree_leaflets[field];
        }
        item_to_update->setData(0, Qt::DisplayRole, field);
        item_to_update->setData(1, Qt::DisplayRole, root_leaflets[field]);
        item_to_update->setData(1, Qt::UserRole, "leaflet");
    }

    // --- branch
    QStack<root_item> stack;
    QMap<QString, RootConfig> root_branches = root.children();
    QMap<QString, QTreeWidgetItem*> tree_branches;

    for(int i=0; i<tree->topLevelItemCount(); ++i){
        QTreeWidgetItem *item = tree->topLevelItem(i);
        if(item->data(1, Qt::UserRole).toString() == "branch"){
            tree_branches[item->data(0, Qt::DisplayRole).toString()] = item;
        }
    }

    for(QString field : tree_branches.keys()){
        if(!root_branches.keys().contains(field)){
            delete tree_branches[field];
        }
    }

    for(QString field : root_branches.keys()){
        QTreeWidgetItem * item_to_update;
        if(!tree_branches.contains(field)){
            item_to_update = new QTreeWidgetItem();
            tree->addTopLevelItem(item_to_update);
            tree_branches[field] = item_to_update;
        }
        else{
            item_to_update = tree_branches[field];
        }

        item_to_update->setData(0, Qt::DisplayRole, field);
        item_to_update->setData(0, Qt::UserRole, root_branches[field].id);
        item_to_update->setData(1, Qt::UserRole, "branch");
        struct root_item root_item;
        root_item.root = root_branches[field];
        root_item.item = tree_branches[field];

        stack.append(root_item);
    }

    while(!stack.isEmpty()){ // while stack not empty
        root_item root_item = stack.pop(); // take upper node
        RootConfig root = root_item.root;
        QTreeWidgetItem *item = root_item.item;

        // --- leaflet
        QMap<QString, QVariant> root_leaflets = root.data();
        QMap<QString, QTreeWidgetItem*> tree_leaflets;

        for(int i=0; i<item->childCount(); ++i){
            QTreeWidgetItem *item_child = item->child(i);
            if(item_child->data(1, Qt::UserRole).toString() == "leaflet"){
                tree_leaflets[item_child->data(0, Qt::DisplayRole).toString()] = item_child;
            }
        }

        for(QString field : tree_leaflets.keys()){
            if(!root_leaflets.keys().contains(field)){
                delete tree_leaflets[field];
            }
        }

        for(QString field : root_leaflets.keys()){
            QTreeWidgetItem * item_to_update;
            if(!tree_leaflets.contains(field)){
                item_to_update = new QTreeWidgetItem();
                item->addChild(item_to_update);
            }
            else{
                item_to_update = tree_leaflets[field];
            }
            item_to_update->setData(0, Qt::DisplayRole, field);
            item_to_update->setData(1, Qt::DisplayRole, root_leaflets[field]);
            item_to_update->setData(1, Qt::UserRole, "leaflet");
        }

        QMap<QString, RootConfig> root_branches = root.children();
        QMap<QString, QTreeWidgetItem*> tree_branches;

        for(int i=0; i<item->childCount(); ++i){
            QTreeWidgetItem *item_child = item->child(i);
            if(item_child->data(1, Qt::UserRole).toString() == "branch"){
                tree_branches[item_child->data(0, Qt::DisplayRole).toString()] = item_child;
            }
        }

        for(QString field : tree_branches.keys()){
            if(!root_branches.keys().contains(field)){
                delete tree_branches[field];
            }
        }

        for(QString field : root_branches.keys()){
            QTreeWidgetItem * item_to_update;
            if(!tree_branches.contains(field)){
                item_to_update = new QTreeWidgetItem();
                item->addChild(item_to_update);
                tree_branches[field] = item_to_update;
            }
            else{
                item_to_update = tree_branches[field];
            }

            item_to_update->setData(0, Qt::DisplayRole, field);
            item_to_update->setData(0, Qt::UserRole, root_branches[field].id);
            item_to_update->setData(1, Qt::UserRole, "branch");
            struct root_item root_item_child;
            root_item_child.root = root_branches[field];
            root_item_child.item = tree_branches[field];
            stack.append(root_item_child);
        }
    }

    //--- style
    tree->header()->setStyleSheet(TREE_HEADER_STYLE);
    tree->setStyleSheet(TREE_STYLE);

    for(int i=0; i<tree->columnCount(); ++i){
        tree->resizeColumnToContents(i);
    }
}

void ConfigurationManager::populate_left_tree(QTreeWidget *tree)
{
    tree->blockSignals(true);
    tree->clear();

    config_map = base_config.get_config_map(program_type);

    QColor green;
    green.setCmyk(100,0,100,100);

    for(QString program_index_tmp : config_map.keys()){
        QTreeWidgetItem *program_index_item = new QTreeWidgetItem();
        program_index_item->setText(0, program_index_tmp);
        ui->tree_left->addTopLevelItem(program_index_item);

        for(int i=0; i<config_map[program_index_tmp].size(); ++i){
            QTreeWidgetItem *configuration_name_item = new QTreeWidgetItem();
            configuration_name_item->setText(0, config_map[program_index_tmp][i]);
            configuration_name_item->
                    setText(1, base_config.get_date_of_modification(program_type, program_index_tmp, config_map[program_index_tmp][i]).toString("dd MMMM yyyy hh:mm"));
            program_index_item->addChild(configuration_name_item);

            if((program_index_tmp == program_index) &&
               (config_map[program_index][i] == configuration_name)){
                configuration_name_item->setBackgroundColor(0,green);
                configuration_name_item->setBackgroundColor(1,green);
                configuration_name_item->setTextColor(0, Qt::black);
                configuration_name_item->setTextColor(1, Qt::black);  
                configuration_name_item->parent()->setExpanded(true);
            }

            if((program_index_tmp == program_index_selected) &&
               (config_map[program_index_tmp][i] == configuration_name_selected)){
                 tree->setCurrentItem(configuration_name_item);
                 config_changed(configuration_name_item, nullptr);
                 configuration_name_item->parent()->setExpanded(true);
            }
        }
    }
    ui->tree_left->header()->setVisible(true);
    ui->tree_left->header()->setStyleSheet(TREE_HEADER_STYLE);
    ui->tree_left->setStyleSheet(TREE_STYLE);

    for(int i=0; i<ui->tree_left->columnCount(); ++i){
        ui->tree_left->resizeColumnToContents(i);
    }

    tree->blockSignals(false);
}

void ConfigurationManager::config_changed(QTreeWidgetItem *current, QTreeWidgetItem *)
{
    bool sub_item = current && current->parent();
    if(sub_item) {
        program_index_selected = current->parent()->text(0);
        configuration_name_selected = current->text(0);
    }

    bool this_item = (program_index_selected == program_index) &&
       (configuration_name_selected == configuration_name);

    ui->pushButton_new->setEnabled(sub_item);
    ui->pushButton_rename->setEnabled(sub_item);
    ui->pushButton_clone->setEnabled(sub_item);
    ui->pushButton_delete->setEnabled(sub_item && !this_item);
    ui->pushButton_switch_to->setEnabled(sub_item && !this_item);

    if(sub_item) {
        selected_root = base_config.read_config(program_type, program_index_selected, configuration_name_selected);
        populate_tree(selected_root, ui->tree);
    }
}

void ConfigurationManager::on_tree_left_itemDoubleClicked(QTreeWidgetItem *item, int column)
{
    if(item->parent() == nullptr){
        return;
    }

    program_index_selected = item->parent()->text(0);
    configuration_name_selected = item->text(0);

    switch_to(program_index_selected, configuration_name_selected);
}

void ConfigurationManager::on_pushButton_new_clicked()
{
    exec_sub_dialog(CMSubDialog::TYPE_NEW);
}

void ConfigurationManager::on_pushButton_rename_clicked()
{
    exec_sub_dialog(CMSubDialog::TYPE_RENAME);
}

void ConfigurationManager::on_pushButton_clone_clicked()
{
    exec_sub_dialog(CMSubDialog::TYPE_CLONE);
}

void ConfigurationManager::exec_sub_dialog(CMSubDialog::TYPE type)
{
    CMSubDialog sub_dialog(type, type==CMSubDialog::TYPE_NEW ? default_root : selected_root,
                                          program_type, program_index_selected, configuration_name_selected, this);
    connect(&sub_dialog, &CMSubDialog::switch_to, this, &ConfigurationManager::switch_to);
    if(sub_dialog.exec() == QDialog::Accepted) {
        on_pushButtonRefresh_clicked();
    }
}

void ConfigurationManager::switch_to(QString program_index, QString configuration_name)
{
    emit switch_to_config(program_index, configuration_name);
    QDialog::accept();
}

void ConfigurationManager::on_pushButton_delete_clicked()
{
    int ret = QMessageBox::question(this, "Delete configuration", QString("Delete configuration '%1/%2'?")
                                    .arg(program_index_selected)
                                    .arg(configuration_name_selected));
    if(ret == QMessageBox::Yes){
        base_config.transform_to_phantom(selected_root);
        populate_left_tree(ui->tree_left);
    }
}

//---button switch clicked---
void ConfigurationManager::on_pushButton_switch_to_clicked()
{
    switch_to(program_index_selected, configuration_name_selected);
}

//---button close clicked---
void ConfigurationManager::on_pushButton_close_clicked()
{
    QDialog::reject();
}

void ConfigurationManager::on_pushButtonRefresh_clicked()
{
    selected_root = base_config.read_config(program_type, program_index_selected, configuration_name_selected);
    populate_tree(selected_root, ui->tree);
    populate_left_tree(ui->tree_left);
}

