#include "LocalDB.h"
#include "MongoConsts.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QDir>
#include <QDirIterator>
#include <QFileInfo>
#include <QStandardPaths>

namespace {
QString remove_bad_symbols(QString s) {
    return s.replace(QRegExp("[^\\w_]"), "_");
}
void copy_setting_block(QSettings *src, QSettings *dst) {
    for(auto key : src->childKeys()) {
        dst->setValue(key, src->value(key));
    }
    for(auto group : src->childGroups()) {
        dst->beginGroup(group);
        src->beginGroup(group);

        copy_setting_block(src, dst);

        dst->endGroup();
        src->endGroup();
    }
}
}

LocalDB::LocalDB() :
    settings_saved(nullptr),
    settings_history(new QSettings(QSettings::IniFormat,
                                   QSettings::UserScope,
                                   QCoreApplication::organizationName(),
                                   COLLECTION_HISTORY))
{
    local_db_chekup();
}

LocalDB::~LocalDB()
{
    sync();
    if(settings_saved)
        delete settings_saved;
    delete settings_history;
}

void LocalDB::init(QString program_type)
{
    auto glob_settings = new QSettings(QSettings::IniFormat,
                                   QSettings::UserScope,
                                   QCoreApplication::organizationName(),
                                   COLLECTION_SAVED);
    remove_brocken_docs_from_local(glob_settings);
    delete glob_settings;

    split_saved_by_indexes(program_type);

    get_rootconfig_map(program_type);
}

void LocalDB::sync()
{
    if(settings_saved)
        settings_saved->sync();
    settings_history->sync();
}

bool LocalDB::local_db_chekup()
{
    bool to_return = true;
    const auto PERMISSIONS =
            QFile::WriteOwner | QFile::WriteGroup | QFile::WriteUser;
    if(settings_saved) {
        const QFileInfo saved_info(settings_saved->fileName());
        if(saved_info.exists() && !saved_info.permission(PERMISSIONS)){
            qWarning().noquote() << QString("\033[91m") +
                                    QString("Settings file is not writable (%1)! ")
                                    .arg(settings_saved->fileName()) +
                                    QString("\033[0m");
            to_return = false;
        }
    }

    const QFileInfo history_info(settings_history->fileName());
    if(history_info.exists() && !history_info.permission(PERMISSIONS)){
        qWarning().noquote() << QString("\033[91m") +
                                QString("Settings file is not writable (%1)! ")
                                .arg(settings_history->fileName()) +
                                QString("\033[0m");
        to_return = false;
    }
    return to_return;
}

RootConfig LocalDB::read_config(QString program_type, QString program_index, QString config_name)
{
    update_settings_saved(program_type, program_index);

    //---read settings from local
    RootConfig root = RootConfig();
    settings_saved->beginGroup(program_type);
    settings_saved->beginGroup(program_index);
    settings_saved->beginGroup(config_name);
    root = local_to_root(settings_saved);
    settings_saved->endGroup();
    settings_saved->endGroup();
    settings_saved->endGroup();
    return root;
}

bool LocalDB::write_config(RootConfig root, bool doSync)
{
    QString program_type = root.get_data<QString>("program_type");
    QString program_index = root.get_data<QString>("program_index");
    QString configuration_name = root.get_data<QString>("configuration_name");

    update_settings_saved(program_type, program_index);

    settings_saved->beginGroup(program_type);
    settings_saved->beginGroup(program_index);
    settings_saved->beginGroup(configuration_name);
    root_to_local(root, settings_saved);
    settings_saved->endGroup();
    settings_saved->endGroup();
    settings_saved->endGroup();

    if(doSync)
        settings_saved->sync();

    return settings_saved->status() == QSettings::NoError;
}

bool LocalDB::write_history_config(RootConfig root)
{
    QString program_type = root.get_data<QString>("program_type");
    QString run_index = root.get_data<QString>("run_index");
    int run_number = root.get_data<int>("run_number", std::numeric_limits<int>::max());

    settings_history->beginGroup(program_type);
    settings_history->beginGroup(run_index);
    settings_history->beginGroup(QString("%1").arg(run_number, 5, 10, QChar('0')));
    root_to_local(root, settings_history);
    settings_history->endGroup();
    settings_history->endGroup();
    settings_history->endGroup();

    settings_history->sync();
    return settings_history->status() == QSettings::NoError;
}

int LocalDB::check_run_number(QString program_type, QString run_index, int run_number)
{
    Q_ASSERT(!program_type.isEmpty());
    Q_ASSERT(!run_index.isEmpty());

    QVector<int> run_numbers;
    settings_history->beginGroup(program_type);
    settings_history->beginGroup(run_index);
    QStringList run_numbers_str = settings_history->childGroups();
    for(int i=0; i<run_numbers_str.size(); ++i){
        bool ok=true;
        int k=run_numbers_str[i].toInt(&ok);
        if(!ok)
            qWarning()<< "LocalDb check run number warning: unexpected run number";
        if(k>=run_number)
            run_numbers.append(k);
    }

    settings_history->endGroup();
    settings_history->endGroup();

    if(run_numbers.size() ==0)
        return run_number;
    if(run_numbers.size() != 0)
        if(run_number != run_numbers[0]){
            return run_number;
        }

    int k = run_number;
    for(int i=0; i< run_numbers.size(); ++i){
        if(run_numbers[i]-k > 1){
            break;
        }
        k = run_numbers[i];
    }
    return k+1;
}

bool LocalDB::write_device_config(RootConfig root)
{
    //TODO
    Q_UNUSED(root)
    return true;
}

QMap<QString, QStringList> LocalDB::get_config_map(QString program_type)
{
    QMap<QString, QStringList> config_map;
    if(program_type.isEmpty())
        return config_map;

    const QString cfg_dir = (QStringList()
                             << QStandardPaths::writableLocation(
                                 QStandardPaths::ConfigLocation)
                             << QCoreApplication::organizationName()
                             << program_type
                             ).join(QDir::separator());

    QDirIterator it(cfg_dir, QDir::NoDotAndDotDot | QDir::Files);
    while(it.hasNext()) {
        QSettings s(it.next(), QSettings::IniFormat);

        s.beginGroup(program_type);
        for(QString program_index : s.childGroups()) {
            s.beginGroup(program_index);
            QStringList config_names;
            for(QString group_config_name : s.childGroups()){
                config_names.append(group_config_name);
            }
            s.endGroup();
            config_map[program_index] = config_names;
            s.endGroup();
        }
        s.endGroup();
    }

    return config_map;
}

QMap<QString, QMap<QString, RootConfig>> LocalDB::get_rootconfig_map(QString program_type)
{
    QMap<QString, QMap<QString, RootConfig>> to_return;
    Q_ASSERT(!program_type.isEmpty());
    if(program_type.isEmpty())
        return to_return;

    QString cur_prog_type_bak = cur_prog_type;
    QString cur_prog_index_bak = cur_prog_index;
    update_settings_saved(program_type, PROGRAM_CONFIG_DEFAULT);

    QDirIterator it(QFileInfo(settings_saved->fileName()).absolutePath(),
                    QDir::NoDotAndDotDot | QDir::Files);
    while(it.hasNext()) {
        QSettings s(it.next(), QSettings::IniFormat);

        s.beginGroup(program_type);
        for(QString program_index : s.childGroups()) {
            s.beginGroup(program_index);
            QMap<QString, RootConfig> root_config_map;
            for(QString config_name : s.childGroups()) {
                s.beginGroup(config_name);
                auto root = local_to_root(&s);
                if(!root.isEmpty())
                    root_config_map[config_name] = root;
                s.endGroup();
            }
            s.endGroup();

            if(!root_config_map.isEmpty())
                to_return[program_index] = root_config_map;
        }
        s.endGroup();
    }

    if(cur_prog_type_bak.isEmpty()) {
        if(settings_saved) {
            settings_saved->sync();
            delete settings_saved;
            settings_saved = nullptr;
        }
    } else {
        update_settings_saved(cur_prog_type_bak, cur_prog_index_bak);
    }
    return to_return;
}

QDateTime LocalDB::get_date_of_modification(QString program_type,
                                            QString program_index,
                                            QString config_name)
{
    update_settings_saved(program_type, program_index);

    QDateTime to_return;
    settings_saved->beginGroup(program_type);
    settings_saved->beginGroup(program_index);
    settings_saved->beginGroup(config_name);
    to_return = settings_saved->value("date_of_modification").toDateTime();
    settings_saved->endGroup();
    settings_saved->endGroup();
    settings_saved->endGroup();

    return to_return;
}

bool LocalDB::delete_config(QString program_type, QString program_index, QString config_name)
{
    update_settings_saved(program_type, program_index);
    settings_saved->beginGroup(program_type);
    settings_saved->beginGroup(program_index);
    settings_saved->beginGroup(config_name);
    settings_saved->remove("");
    settings_saved->endGroup();
    settings_saved->endGroup();
    settings_saved->endGroup();

    settings_saved->sync();
    return true;
}

void LocalDB::remove_brocken_docs_from_local(QSettings *s)
{
    bool changes = false;
    for(QString program_type : s->childGroups()){
        s->beginGroup(program_type);
        for(QString program_index : s->childGroups()){
            s->beginGroup(program_index);
            for(QString configuration_name : s->childGroups()){
                s->beginGroup(configuration_name);
                auto program_type_ = s->value("program_type").toString();
                auto program_index_ = s->value("program_index").toString();
                auto configuration_name_ = s->value("configuration_name").toString();

                if(program_type!=program_type_ ||
                        program_index!=program_index_ ||
                        configuration_name!=configuration_name_) {
                    changes = true;
                    s->remove("");
                }
                s->endGroup();
            }
            s->endGroup();
        }
        s->endGroup();
    }
    if(changes)
        s->sync();
}

void LocalDB::split_saved_by_indexes(QString prog_type)
{
    QSettings global_settings(QSettings::IniFormat,
                                   QSettings::UserScope,
                                   QCoreApplication::organizationName(),
                                   COLLECTION_SAVED);

    global_settings.beginGroup(prog_type);
    for(auto prog_index : global_settings.childGroups()) {
        global_settings.beginGroup(prog_index);
        update_settings_saved(prog_type, prog_index);
        settings_saved->beginGroup(prog_type);
        settings_saved->beginGroup(prog_index);

        for(auto prog_cfg : global_settings.childGroups()) {
            global_settings.beginGroup(prog_cfg);
            settings_saved->beginGroup(prog_cfg);

            auto global = local_to_root(&global_settings);
            auto separate = local_to_root(settings_saved);
            auto global_date = global.get_delete_time();
            if(global_date.isNull())
                global_date = global.get_modification_time();
            auto separate_date = separate.get_delete_time();
            if(separate_date.isNull())
                separate_date = separate.get_modification_time();

            if(!global_date.isNull() && (separate_date.isNull()
                                         || global_date<separate_date)) {
                copy_setting_block(&global_settings, settings_saved);
            }

            global_settings.remove("");
            global_settings.endGroup();
            settings_saved->endGroup();
        }

        global_settings.endGroup();
        settings_saved->endGroup();
        settings_saved->endGroup();
    }
    global_settings.endGroup();
    global_settings.sync();
    if(settings_saved)
        settings_saved->sync();
}

void LocalDB::root_to_local(RootConfig root, QSettings *settings)
{
    settings->remove("");
    tree_traversing(settings, root);
}

void LocalDB::tree_traversing(QSettings *settings, RootConfig root)
{
    //inspect my fields
    for(QString field : root.data().keys()){
        settings->setValue(field, root.data()[field]);
    }

    //inspect my childes
    for(QString child_name : root.children().keys()){
        settings->beginGroup(child_name);
            tree_traversing(settings, root.child(child_name));
        settings->endGroup();
    }
}

RootConfig LocalDB::local_to_root(QSettings *settings){
    return settings_traversing(settings);
}

RootConfig LocalDB::settings_traversing(QSettings *settings)
{
    QStringList local_fields = settings->childKeys();
    QStringList local_childs = settings->childGroups();

    RootConfig root = RootConfig();

    //inspect my fields
    for(const auto &field : local_fields){
        root.set_data(field, settings->value(field));
    }

    //inspect my childs
    for(const auto &field : local_childs){
        RootConfig child = RootConfig();
        settings->beginGroup(field);
            child = settings_traversing(settings);
            child.item_name = field;
        settings->endGroup();
        root.append_child(child);
    }
    return root;
}

void LocalDB::update_settings_saved(QString type, QString index)
{
    Q_ASSERT(!type.isEmpty());
    if(index.isEmpty())
        index = PROGRAM_INDEX_DEFAULT;
    if(settings_saved && cur_prog_type == type && cur_prog_index == index)
        return;
    if(settings_saved) {
        settings_saved->sync();
        delete settings_saved;
    }
    QString fileName = type + "_" + index;
    QString appName = remove_bad_symbols(type)
            + QDir::separator()
            + remove_bad_symbols(fileName);
    settings_saved = new QSettings(QSettings::IniFormat,
                                   QSettings::UserScope,
                                   QCoreApplication::organizationName(),
                                   appName);
    cur_prog_type = type;
    cur_prog_index = index;
}

//--- garbage ---

void LocalDB::rename_date_of_creation()
{
    // Used in mongo viewer only

    if(!settings_saved)
        return;
    QSettings *settings = settings_saved;

    for(QString program_type : settings->childGroups()){
        settings->beginGroup(program_type);
        for(QString program_index : settings->childGroups()){
            settings->beginGroup(program_index);
            for(QString configuration_name : settings->childGroups()){
                settings->beginGroup(configuration_name);
                QDateTime date_of_creation = settings->value("date_of_creation").toDateTime();
                settings->setValue("date_of_modification", date_of_creation);
                settings->remove("date_of_creation");
                settings->endGroup();
            }
            settings->endGroup();
        }
        settings->endGroup();
    }
}

