#ifndef PNPMESSAGE_H
#define PNPMESSAGE_H

class QByteArray;
class QDomElement;
#include <QHostAddress>
#include <QHostInfo>
#include <QStringList>
#include <QMetaType>
#include <QUuid>

enum ProgramInterfaceType { ProgramInterfaceEmty,
                            ProgramInterfaceRemoteControl,
                            ProgramInterfaceDataFlow,
                            ProgramInterfaceMonitorData,
                            ProgramInterfaceConfigTransfer
                          };

struct ProgramInterfacePeer
{
    ProgramInterfacePeer() : port(0) {}
    ProgramInterfacePeer(QHostAddress h, quint32 p) : host(h), port(static_cast<quint16>(p)) {}
    inline bool operator==(const ProgramInterfacePeer& i) const {
        return host.toIPv4Address() == i.host.toIPv4Address()
                && port == i.port;
    }

    QString toString() const { return QString("%1:%2").arg(host.toString()).arg(port); }
    QHostAddress host;
    quint16 port;
};

class ProgramInterface
{
public:
    ProgramInterface():host(QHostAddress::LocalHost),port(0),enabled(false),isFree(true),id(0) {}
    ProgramInterface(const QHostAddress h, const quint32 p, const QString t, bool free=true, int i=0):
        host(h),port(static_cast<quint16>(p)),type(t),enabled(true),isFree(free),id(i) {}
    static QString getTypeName(ProgramInterfaceType type);
    void setType(ProgramInterfaceType t);
    bool isEmpty() const { return type.isEmpty(); }
    inline bool operator!=(const ProgramInterface& i) const {
        return !(*this == i);
    }

    inline bool operator==(const ProgramInterface& i) const {
        return id==i.id && type==i.type
                && host==i.host && port==i.port
                && enabled==i.enabled && isFree==i.isFree
                && peers==i.peers;
    }
    void setPeer(QHostAddress h, quint32 p);
    void addPeer(QHostAddress h, quint32 p);

    QHostAddress host;
    quint16 port;
    QString type;
    QVector<ProgramInterfacePeer> peers;
    bool enabled;
    bool isFree;
    int id;
};
Q_DECLARE_METATYPE(ProgramInterface)

class ProgramDescription
{
public:
    ProgramDescription() :
        uuid(QUuid::createUuid()),
        seq(0),
        host(QHostAddress::LocalHost),
        hostName(QHostInfo::localHostName())
    {}
    QUuid uuid;
    quint32 seq;
    QString name; // app name
    QString type; // the one from DaqConfig can be empty
    QString index; // the one from DaqConfig can be empty
    QHostAddress host;
    QString hostName;
    QVector<ProgramInterface> interfaces;
    ProgramInterface dummyInterface;
    QMap<QString, QString> options;
    QString ver_hash;
    QString ver_date;

    ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr, int id=-1);
    const ProgramInterface &getProgramInterface(ProgramInterfaceType type, bool *ok=nullptr) const;
    ProgramInterface &getProgramInterface(QString type, bool *ok=nullptr, int id=-1);
    const ProgramInterface &getProgramInterface(QString type, bool *ok=nullptr) const;
    int getProgramInterfaceNum(ProgramInterfaceType type);
    int getProgramInterfaceNum(QString type=QString());
    bool containsProgramInterface(ProgramInterfaceType type, int id);
    bool containsProgramInterface(QString type, int id);
    bool updateProgramInterface(ProgramInterfaceType type, int id, int port, bool isListening, bool isFree, const QHostAddress &peerHost, int peerPort);

    QString getAnnounceStr() const;
    QString getClosingStr() const;
    QString toString() const;
    void clear();
    quint32 getHostIp() const;
    QString getHostName() const;
    QString getName() const { return name.isEmpty() ? type+" #"+index : name; }
    bool isEmpty() const { return type.isEmpty() || index.isEmpty(); }
    bool sameProgram(const ProgramDescription &descr) const {
        return type==descr.type && index==descr.index;
    }
    friend bool operator <(const ProgramDescription &pd1, const ProgramDescription &pd2){
        return pd1.getName()<pd2.getName() || (pd1.getName()==pd2.getName() && pd1.uuid<pd2.uuid);
    }
    inline bool operator==(const ProgramDescription& pd) const {
        return sameProgram(pd)
                && uuid == pd.uuid
                && seq == pd.seq
                && name == pd.name
                && host == pd.host
                && hostName == pd.hostName
                && interfaces == pd.interfaces
                && dummyInterface == pd.dummyInterface
                && options == pd.options;
    }
    inline bool operator!=(const ProgramDescription& pd) const {
        return !this->operator==(pd);
    }
};
Q_DECLARE_METATYPE(ProgramDescription)

enum PNPMessage_Type { PNPMessage_Type_Search, PNPMessage_Type_Discover, PNPMessage_Type_Close};

class PNPMessage
{
public:
    PNPMessage();
    QString toString() const;
    QString getXml() const;
    bool decode(const QByteArray &data, const QHostAddress &host, bool *ok=nullptr);

    void addInterface(const QString &type, const QHostAddress &host, const quint32 port);
    bool decodeOptions(const QDomElement &domEl, const QString &hostName);
    bool decodeInterfaces(const QDomElement &domEl, const QHostAddress &host);

    PNPMessage_Type type;
    QStringList searchTargets; // list program types

    ProgramDescription progDesc;
};

#endif // PNPMESSAGE_H
