// QString s = "<!DOCTYPE pnp_message>"
//             "<discover_request/>"

// QString s = "<!DOCTYPE pnp_message>"
//             "<program name=\"pnp_search\">"
//             " <options>"
//             "  <option name=\"targets\" value=\"all\"/>"
//             " <options>"
//             "</program>";

//    QString s = "<!DOCTYPE pnp_message><program name=\"mstream\">"
//            "<options><option name=\"address\" value=\"10.93.120.171\"/><option name=\"port\" value=\"1222\"/></options>"
//            "<interfaces><interface type=\"data downflow\" host=\"0.0.0.0\" port=\"33511\"/></interfaces></program>";

#include "PNPServer.h"
#include "multicast-listener/MultiCastListener.h"
#include <QDebug>
#include <QTimer>
#include <QNetworkInterface>
#include <QUdpSocket>
#include <arpa/inet.h>
#ifdef QT_DEBUG
#include <QTime>
#endif


namespace {
const int SEARCH_INTERVAL_MS = 5000;
}

PNPServer::PNPServer(QObject *parent):
    QObject(parent),
    searchEnabled(false),
    searchTimer(new QTimer(this)),
    multiCastListener(new MultiCastListener(QHostAddress(PNP_DISCOVER_IP_ADDR), PNP_DISCOVER_PORT, this)),
    multiCastSender(new QUdpSocket(this))
{
    searchTimer->setInterval(SEARCH_INTERVAL_MS);
    connect(searchTimer, &QTimer::timeout, this, &PNPServer::sendSearchRequest);
    connect(multiCastListener, &MultiCastListener::received, this, &PNPServer::decode);
    qRegisterMetaType<ProgramDescription>();
    if(!multiCastSender->bind())
        qFatal("Can't bind UDP socket for PNP multicast");
    u_char ttl=15;
    if(setsockopt(static_cast<int>(multiCastSender->socketDescriptor()),
                  IPPROTO_IP, IP_MULTICAST_TTL, &ttl,sizeof(ttl)) != 0) {
        qCritical("Can't set TTL to %d for PNP multicast", ttl);
    }
}

PNPServer::~PNPServer()
{
    if(searchEnabled)
        stopSearch();
    while(!progDesc.isEmpty()) {
        closeProgram(progDesc.first());
    }
    multiCastSender->flush();
}

QString PNPServer::getSearchStr()
{
    PNPMessage mes;
    mes.type = PNPMessage_Type_Search;
    mes.searchTargets = searchTargets;
    return mes.getXml();
}

bool PNPServer::decode(const QByteArray &data, const QHostAddress &host)
{
    PNPMessage msg;
    bool ok = msg.decode(data, host);
    if(ok){
        switch (msg.type) {
        case PNPMessage_Type_Search:
            //qDebug()<<"Got search request.";
            for (auto desc : progDesc) {
                if(msg.searchTargets.isEmpty() || msg.searchTargets.contains(desc.type, Qt::CaseInsensitive)){
#ifdef QT_DEBUG
                    //qDebug() << QTime::currentTime().toString() << "Sent discover response.";
#endif
                    sendProgDescr(desc);
                }
            }
            break;
        case PNPMessage_Type_Discover:
        case PNPMessage_Type_Close:
            if(searchEnabled) {
                if(searchTargets.isEmpty() || searchTargets.contains(msg.progDesc.type)){
                    bool isMine=false;
                    for (auto pd : progDesc) {
                        if(msg.progDesc.uuid == pd.uuid){
                            isMine = true;
                            break;
                        }
                    }
                    if(!isMine){
//                        print("==========");
//                        print(msg.toString());
                        if(msg.type==PNPMessage_Type_Discover) {
                            emit programDiscovered(msg.progDesc, data.size());
                        } else
                            emit programClosed(msg.progDesc);
                    }
                }
            }
            break;
        default:
            qWarning()<<"Unknown msg type:"<<data;
            ok=false;
            break;
        }
    } else {
        qWarning()<<"Failed decode msg:"<<data;
    }
    return ok;
}

void PNPServer::sendSearchRequest()
{
    if(!searchEnabled)
        return;
    const QString searchStr = getSearchStr();
    sendMulticastMsg(searchStr);
}

void PNPServer::changeProgramDescription(const ProgramDescription &pd)
{
    progDesc.clear();
    if(pd.isEmpty())
        return;
    progDesc.append(pd);
    sendProgDescr(pd);
}

void PNPServer::updateProgramDescription(const ProgramDescription &pd)
{
    if(pd.isEmpty())
        return;

    bool found = false;
    for(int i=0; i<progDesc.size(); ++i){
        ProgramDescription &oldPd = progDesc[i];
        if(oldPd.sameProgram(pd)){
            progDesc[i] = pd;
            found = true;
            break;
        }
    }

    if(!found)
        progDesc.append(pd);

    sendProgDescr(pd);
}

void PNPServer::closeProgram(const ProgramDescription &pd)
{
    const QString closingStr = pd.getClosingStr();
    for(int i=0; i<progDesc.size();) {
        if(progDesc[i].sameProgram(pd)){
            progDesc.removeAt(i);
        } else
            ++i;
    }
//    for (auto p : progDesc) {
//        if(p.sameProgram(pd)){
//            progDesc.removeAt(progDesc.indexOf(p));
//        }
//    }
    sendMulticastMsg(closingStr);
}

void PNPServer::sendProgDescr(ProgramDescription pd)
{
    pd.seq = ++(seqMap[pd.uuid]);
    sendMulticastMsg(pd.getAnnounceStr());
}

void PNPServer::sendMulticastMsg(const QString &msg)
{
    const QByteArray data = msg.toLatin1();
    qint64 nwr = multiCastSender->writeDatagram(data, QHostAddress(PNP_DISCOVER_IP_ADDR), PNP_DISCOVER_PORT);
    if (nwr != data.size()) {
        qCritical() << "[PNPServer::sendMulticastMsg] "
                    << multiCastSender->errorString();
    }
//        qInfo() << "PNPServer Send "<< nwr << " bytes:" << msg;
}

void PNPServer::print(QString s) const
{
#if QT_VERSION < QT_VERSION_CHECK(5, 5, 0)
    qDebug()<<s;
#else
    qInfo()<<s;
#endif
}

void PNPServer::startSearch(QString target, bool single)
{
    startSearch(QStringList(target), single);
}


void PNPServer::startSearch(QStringList searchTargets, bool single)
{
    this->searchTargets = searchTargets;
    searchEnabled = true;
    sendSearchRequest();
    if(!single){
        searchTimer->start();
        print(QString("Pnp_search started with %1ms interval, targets:%2")
                  .arg(SEARCH_INTERVAL_MS)
                  .arg(searchTargets.isEmpty() ? "all" : searchTargets.join(", "))
              );
    } else {
        print(QString("Pnp_search single request for targets:%1")
                  .arg(searchTargets.isEmpty() ? "all" : searchTargets.join(", "))
              );
    }

}

void PNPServer::stopSearch()
{
    searchEnabled = false;
    searchTimer->stop();
    print("Pnp_search stopped");
}

QVector<QHostAddress> PNPServer::getLocalHostAddresses()
{
    QVector<QHostAddress> addresses;
    for (auto interface : QNetworkInterface::allInterfaces()) {
        if ( !(interface.flags() & QNetworkInterface::IsUp) ) continue;
        if ( !(interface.flags() & QNetworkInterface::IsRunning) ) continue;
        if (  (interface.flags() & QNetworkInterface::IsLoopBack) ) continue;
        for (auto addr : interface.allAddresses()) {
            if(addr.isNull() || addr.protocol() != QAbstractSocket::IPv4Protocol)
                continue;
#if QT_VERSION > 0x050000
            if(addr.isLoopback())
                continue;
#endif
            addresses.push_back(addr);
        }
    }
    return addresses;
}

