//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <iostream>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <cassert>

#include "qxw/WaveformData.h"
#include "qxw/qxw_randgen.h"
//#include "hptimer.h"

using namespace qxw;

//! eliminate edge effects
//! t - fine offset (0..1)
/*
double find_fine_integral(std::vector<double> v, double t)
{
//  t-=.5;
  unsigned int n = v.size();
  if (n<2) return v[0]; // one point - no 1st derivative possible
  double i1 = (v[1]+v[0])/2*t; // subtract this
//  double vn = 2*v[n-1] - v[n-2]; // linear extrapolation
//  double i2 = (vn+v[n-1])/2*t; // add this
  double i2 = (2*v[n-1] + t*(v[n-1]-v[n-2]))*t;
  double i = 0;
  for(unsigned int k=0;k<n;k++)
    i += v[k];
//  std::cout << "Integral: i="<< i << ", i1=" << i1 << ", i2=" << i2 << std::endl;
  std::cout << t << " " << i1 << " " << i2 << " "<< i << " ";
  i -= i1*.5;
  i += i2*.5;
  std::cout << i << std::endl;
  return i;
}
*/
double find_fine_integral(std::vector<double> v, double t)
{
    unsigned int n = v.size();
    if (n<2) return v[0]; // one point - no 1st derivative possible
    double dvb = v[1]-v[0];
    double vb = v[0]+dvb*t;
    double dve = v[n-2]-v[n-1];
    double ve = v[n-2]+dve*t;
    double i1 = (v[0]+vb)/2*t; // subtract this
    double i2 = (v[n-1]+ve)/2*(1-t); // subtract this
    double i = 0;
    for(unsigned int k=0;k<n;k++)
        i += v[k];
    //  std::cout << "Integral: i="<< i << ", i1=" << i1 << ", i2=" << i2 << std::endl;
    //  std::cout << t << " " << i1 << " " << i2 << " "<< i << " ";
    i -= i1;
    i += i2;
    //  std::cout << i << std::endl;
    return i;
}

WaveformData::WaveformData(const std::string &dataName, int dataChannel)
    : c(dataChannel), name(dataName)
{
}

WaveformData::WaveformData(const std::vector<double> &vy, const std::string &dataName, int dataChannel)
    : c(dataChannel), name(dataName)
{
    const size_t n = vy.size();
    p.resize(n);
    for(size_t i=0; i<n; i++) {
        p[i].x = i; // x axis is index
        p[i].y = vy[i];
    }
}

WaveformData::WaveformData(const std::vector<double> &vx, const std::vector<double> &vy, const std::string &dataName, int dataChannel)
    : c(dataChannel), name(dataName)
{
    assert(vx.size() == vy.size());
    const size_t n = std::min(vx.size(), vy.size());
    p.resize(n);
    for(size_t i=0; i<n; i++) {
        p[i].x = vx[i];
        p[i].y = vy[i];
    }
}


void WaveformData::simulate(unsigned int len)
{
    if(len<20) len=20;
    int ch = rand() % 8;
    double bias = rangaus(0.01);
    double t_start = ranuni(-3.*len, 3.*len); // waveform start offset
    double dt = 1.;
    // pulse amplitude parameters
    double a = ranuni(0.0, 1.0);
    double rt = 1.5;
    double ft = 5.;
    // TDC hits
    bool has_t0 = 1;//(rand()%100 < 95); // 5% tdc hit is lost
    // 10% - one pileup
    // 2% - two pileup
    unsigned int num_tx = 0;
    int r = rand()%100;
    if (r >= 90) num_tx++;
    if (r >= 98) num_tx++;
    double pulse_t = ranuni(3.0, len-15.0); // pulse time offset inside waveform slide
    t.clear();
    if (has_t0) t.push_back(t_start + pulse_t);
    for(unsigned int i=0;i<num_tx;i++)
        t.push_back(t_start + ranuni(0., len));

    //  if (a<0) a=-a;

    p.resize(len);

    std::vector<double> py(len);
    for(unsigned int i=0;i<len;i++) {
        py[i] = bias + detpulse (a, i*dt-pulse_t, rt, ft) + rangaus(0.001);
    }

    //  detpulse_v (a, t, rt, ft, &py[0], len, dt);

    c = ch;
    for(unsigned int i=0;i<len;i++) {
        p[i].x = t_start + i*dt;
        p[i].y = py[i];
        //	std::cout << "x=" << p[i].x << ", y=" << p[i].y << std::endl;
    }
    //  double m_t = has_t0?(t[0]-floor(t[0])):0;
    //  double m_a = find_fine_integral(py, m_t);
    //  std::cout << "Source t="<<m_t<<", a="<<a<<", found a="<<m_a<<", err="<< (m_a-a)/a << std::endl;
}

void WaveformData::print() const
{
    std::cout << "WaveformData, channel=" << c << ", len=" << p.size() << std::endl;
    //  for(unsigned int i=0;i<p.size();i++)
    //	std::cout << "x=" << p[i].x << ", y=" << p[i].y << std::endl;
}
/*
boost::shared_ptr<vector <double> > WaveformData::x() const
{
}
*/
/*
void WaveformModel::simulate() {
  hpTimer hpt;
  hpt.start();
  for(unsigned int i=0;i<wd.size();i++)
    wd[i].simulate();
  hpt.stop();
  double dt = hpt.elapsed();
  stats->add("simulation time", dt);
//  stats->print();
//  std::cout << "Simulated "<<wd.size()<<" waveforms in " << dt << " s."<< std::endl;
}
*/

WaveformDataAll WaveformDataSimulator::run()
{
    WaveformDataAll wd;
    wd.resize(10);
    //  hpTimer hpt;
    //  hpt.start();
    for(unsigned int i=0;i<wd.size();i++)
        wd[i].simulate();
    //  hpt.stop();
    //  double dt = hpt.elapsed();
    //  stats->add("simulation time", dt);
    return wd;
}
