#include "EvNumChecker.h"

#include <QDebug>
#include <QTimer>
#include <QMetaType>

namespace {
//#define STRONG_NOT_IN_RANGE_SUPPRESSION
// [18:44:04] [GUI thread] "RunControl #ivan event number 0 not in range 33761..33761 "
// [18:44:07] [GUI thread] "RunControl #ivan event number 0 not in range 35638..35638 "
// with defined STRONG_NOT_IN_RANGE_SUPPRESSION 2nd log msg will be suppresed

const int DEVICE_TIMEOUT = 1000; // ms
const int SUBSYSTEM_TIMEOUT = 2000; // ms
const int CHECK_PERIOD = 3000; // ms
const bool DEBUG_INFO = false;
}

EvNumChecker::EvNumChecker(QObject *parent) :
    QObject(parent),
    completed(true),
    checkEnabled(false),
    curCheckIndex(0),
    evNumCheckTimer(new QTimer(this)),
    checkExpireTimer(new QTimer(this))
{
    qRegisterMetaType<EvNumSet> ();
    connect(evNumCheckTimer, &QTimer::timeout, this, &EvNumChecker::startCheck);
    evNumCheckTimer->setSingleShot(true);
    evNumCheckTimer->setInterval(CHECK_PERIOD);

    connect(checkExpireTimer, &QTimer::timeout, this, &EvNumChecker::checkTimeout);
    checkExpireTimer->setSingleShot(true);
}

EvNumChecker::~EvNumChecker()
{
    stopCheck();
}

void EvNumChecker::startCheck()
{
    // Can be run in Local mode only
    stopCheck(false);
    checkEnabled = true;
    if(clientMap.isEmpty() && vmeMap.isEmpty()) return;

    ++curCheckIndex;
    emit getTrigEvNum(curCheckIndex);
    checkExpireTimer->start(DEVICE_TIMEOUT);
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__
                << " checkExpire started for " << DEVICE_TIMEOUT << "ms";
}

void EvNumChecker::set_dev_serials(QVector<quint64> serialList)
{
    bool hasChanges = devSerials != serialList;
    devSerials = serialList;
    if(hasChanges && checkEnabled) {
        startCheck(); // restart check
    }
}

void EvNumChecker::stopCheck(bool clearWarnSet)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__;
    checkEnabled = false;
    completed = true;
    evNumCheckTimer->stop();
    checkExpireTimer->stop();
    if(clearWarnSet)
        warnSet.clear();
}

void EvNumChecker::updateClientMap(QMap<int, QString> map)
{
    bool hasChanges = (map.keys() != clientMap.keys());
    clientMap = map;
    if(hasChanges && checkEnabled) {
        startCheck(); // restart check
    }
}

void EvNumChecker::updateVmeMap(QMap<quint64, QString> map)
{
    bool hasChanges = (map.keys() != vmeMap.keys());
    vmeMap = map;
    if(hasChanges && checkEnabled) {
        startCheck(); // restart check
    }
}

void EvNumChecker::onRemoteEvNumCheckRequest(int checkIndex)
{
    if(completed)
        clearWarn(WARN_DOUBLE_REQUEST);
    else
        setWarn(WARN_DOUBLE_REQUEST, "EvNumChecker new request while previous one hasn't been completed");
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << "checkIndex:"<<checkIndex;

    curCheckIndex = checkIndex;
    if(devSerials.isEmpty() && clientMap.isEmpty() && vmeMap.isEmpty()) {
        emit completedEvNumCheck(curCheckIndex, EvNumSet());
    } else {
        completed = false;
        evNumbers.clear();
        clientEvNum.clear();
        emit evNumRequest(curCheckIndex);
        emit getTrigEvNum(curCheckIndex, false);
        checkExpireTimer->start(clientMap.isEmpty() ? DEVICE_TIMEOUT : SUBSYSTEM_TIMEOUT);
        if(DEBUG_INFO)
            qInfo() << typeid(*this).name() << __func__
                    << " checkExpire started for " << DEVICE_TIMEOUT << "ms";
    }
}

void EvNumChecker::gotTopLevelEvNum(int checkIndex, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:"<<checkIndex
                << "evNum:"<<evNum;
    if(!checkEnabled) return;
    if(curCheckIndex != checkIndex) return;
    if(completed) {
        // 1st stage passed
        completed = false;
        trigEvNum = evNum;
        clearWarn(WARN_MISSING_INITIAL);
        clientEvNum.clear();
        evNumbers.clear();
        emit evNumRequest(curCheckIndex);
        checkExpireTimer->start(SUBSYSTEM_TIMEOUT);
    } else {
        // All collected
        performTopLevelCheck(checkIndex, evNum);
    }
}

void EvNumChecker::gotClientEvNum(int checkIndex, int clientIndex, EvNumSet evNums)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:"<<checkIndex
                << "clientIndex:"<<clientIndex
                << "evNum:"<<evNums;
    if(completed) return;
    if(curCheckIndex != checkIndex) return;
    clientEvNum[clientIndex] = evNums;
    checkAllCollected();
}

void EvNumChecker::gotVmeTrigEvNum(int checkIndex, quint64 vmeSerial, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:"<<checkIndex
                << "vmeSerial:"<<hex<<vmeSerial<<dec
                << "evNum:"<<evNum;
    if(completed) return;
    gotEvNum(checkIndex, vmeSerial, evNum);
}

void EvNumChecker::gotDevEvNum(int checkIndex, quint64 serial, quint64 evNum)
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__ << " checkIndex:" << checkIndex
                << "serial:" << hex << serial << dec << "evNum:" << evNum;
    if(completed) return;
    gotEvNum(checkIndex, serial, evNum);
}

void EvNumChecker::checkTimeout()
{
    if(DEBUG_INFO)
        qInfo() << typeid(*this).name() << __func__;
    if(completed) {
        // 1st stage
        setWarn(WARN_MISSING_INITIAL, "Failed to get initial event number from RC module");
    } else {
        QStringList messList;
        if(!checkAllCollected(true)) {
            // 2nd stage
            for (int clIndex : clientMap.keys()) {
                if(!clientEvNum.contains(clIndex))
                    messList << QString("Failed to get event number from client #%1 %2").arg(clIndex).arg(clientMap[clIndex]);
            }
            for (quint64 ser : vmeMap.keys()) {
                if(!evNumbers.contains(ser))
                    messList << QString("Failed to get event number from vmeTrigModule #%1 %2").arg(ser).arg(vmeMap[ser]);
            }
            for (quint64 ser : devSerials) {
                if(!evNumbers.contains(ser))
                    messList << QString("Failed to get event number from dev 0x%1").arg(ser,8,16,QChar('0'));
            }
        } else {
            // 3rd stage
            messList << "Failed to get final event number from RC module";
        }

        const QString mess = messList.join('\n');
        setWarn(WARN_TIMEOUT, mess);
        emit checkFail(mess);
    }

    if(checkEnabled)
        evNumCheckTimer->start();
}

void EvNumChecker::gotEvNum(int checkIndex, quint64 serial, quint64 evNum)
{
    if(curCheckIndex != checkIndex) return;
    evNumbers[serial] = evNum;
    checkAllCollected();
}

bool EvNumChecker::checkAllCollected(bool info_only)
{
    for (int clIndex : clientMap.keys()) {
        if(!clientEvNum.contains(clIndex))
            return false;
    }
    for (quint64 index : vmeMap.keys()) {
        if(!evNumbers.contains(index))
            return false;
    }
    for (quint64 index : devSerials) {
        if(!evNumbers.contains(index))
            return false;
    }
    if(info_only)
        return true;

    if(checkEnabled) {
        // Local mode: Get last evNum
        emit getTrigEvNum(curCheckIndex);
        checkExpireTimer->start(DEVICE_TIMEOUT);
    } else {
        // Send all collected events to remote
        EvNumSet list = evNumbers.values().toSet();
        for(auto clList : clientEvNum.values())
            list.unite(clList);
        emit completedEvNumCheck(curCheckIndex, list);
        checkExpireTimer->stop();
        completed = true;
    }
    clearWarn(WARN_TIMEOUT);
    return true;
}

void EvNumChecker::performTopLevelCheck(int checkIndex, quint64 evNum)
{
    bool checkSuccess = true;
    if(DEBUG_INFO){
        qInfo() << typeid(*this).name() << __func__;
        qInfo() << "All collected! range from"<<trigEvNum<<" to "<<evNum;
        qInfo() << "clientEvNum:"<<clientEvNum;
        qInfo() << "vmeEvNum:"<<vmeMap;
        qInfo() << "devSerials:"<<devSerials;
    }

    if(evNum < trigEvNum) {
        setWarn(WARN_EV_NUM_DECR, QString("RC trigEvNum decreased %1 -> %2").arg(trigEvNum).arg(evNum));
        checkSuccess = false;
    } else {
        clearWarn(WARN_EV_NUM_DECR);
    }
    for (int clIndex : clientEvNum.keys()) {
        bool clientOk = true;
        for (const quint64 &clEvNum : clientEvNum[clIndex]) {
            if (trigEvNum > clEvNum || clEvNum > evNum) {
                setWarn(WARN_NOT_IN_RANGE, static_cast<quint64>(clIndex), QString("%1 has event number %2 not in range %3..%4 ")
                        .arg(clientMap[clIndex]).arg(clEvNum).arg(trigEvNum).arg(evNum));
                clientOk = false;
            }
        }
        if(clientOk) {
            clearWarn(WARN_NOT_IN_RANGE, static_cast<quint64>(clIndex));
        } else {
            checkSuccess = false;
        }
    }
    for (quint64 serial : vmeMap.keys()) {
        if(trigEvNum > evNumbers[serial] || evNumbers[serial] > evNum) {
            setWarn(WARN_NOT_IN_RANGE, serial, QString("TMWR %1 event number %2 not in range %3..%4")
                    .arg(vmeMap[serial]).arg(vmeMap[serial]).arg(trigEvNum).arg(evNum));
            checkSuccess = false;
        } else {
            clearWarn(WARN_NOT_IN_RANGE, serial);
        }
    }
    Q_UNUSED(checkIndex)
//    if(checkSuccess) {
//        clearWarn(WARN_SUMMARY_ERR);
//    } else {
//        QString mess = QString("EventNumberChecker: %1; CheckIndex=%2 trig %3-%4")
//                .arg(checkSuccess ? "Success" : "Fail")
//                .arg(checkIndex).arg(trigEvNum).arg(evNum);
//        if(!clientEvNum.isEmpty()) {
//            mess.append("; clients:");
//            for (int i : clientEvNum.keys()) {
//                QStringList evStrList;
//                for (quint64 e : clientEvNum[i]) {
//                    evStrList << QString::number(e);
//                }
//                mess += QString(" %1:%2").arg(clientMap[i]).arg(evStrList.join(','));
//            }
//        }
//        if(!vmeMap.isEmpty()) {
//            mess.append("; TMWR:");
//            for (quint64 serial : vmeMap.keys()) {
//                mess += QString(" %1:%2").arg(vmeMap[serial]).arg(vmeMap[serial]);
//            }
//        }
//        if(!warnSet.contains(WARN_SUMMARY_ERR)) {
//            warnSet[WARN_SUMMARY_ERR];
//            qWarning() << mess;
//        }
//        emit checkFail(mess);
//    }
    completed = true;
    checkExpireTimer->stop();
    evNumCheckTimer->start();
}

void EvNumChecker::setWarn(WARN_ID id, quint64 subId, QString msg)
{
    Q_ASSERT(!msg.isNull());
    bool printLog = false;
#ifdef STRONG_NOT_IN_RANGE_SUPPRESSION
    if(id == WARN_NOT_IN_RANGE) {
        printLog = !warnSet[id].contains(subId);
        warnSet[id][subId] = msg;
    } else
#endif
    {
        auto &curMsg = warnSet[id][subId];
        if(curMsg != msg) {
            curMsg = msg;
            printLog = true;
        }
    }

    if(printLog) {
        if(!msg.isNull())
            qWarning() << msg;
    }
}

void EvNumChecker::setWarn(EvNumChecker::WARN_ID id, QString msg)
{
    setWarn(id, 0, msg);
}

void EvNumChecker::clearWarn(EvNumChecker::WARN_ID id, quint64 subId)
{
    warnSet[id][subId].clear();
}
